/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*! \file
    \ingroup occurences_merging_group
    \brief Data structure for efficiently storing elementary trees occurence lists.
*/

#include <boost/foreach.hpp>
// #include <boost/range/adaptors.hpp>  //boost>=1.43

#include<cassert>
#include <limits>
#include<ctime>
#include <iostream>

#include <fstream>
#include <boost/serialization/vector.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>

#include <boost/iterator/iterator_facade.hpp>
#include <boost/iterator/iterator_adaptor.hpp>
#include <boost/type_traits/is_convertible.hpp>
#include <boost/utility/enable_if.hpp>

#include <boost/serialization/split_member.hpp>

#ifndef _HPP_ARRAY_OF_PAIRS_
#define _HPP_ARRAY_OF_PAIRS_

/*!

The ArrayOfPairs<T> templated class is designed to emulate (most of) the access interface of a vector< vector<T> >, with the restriction that each "inner" vector contained in the "outer" vector has always a size of 2. T is also expected to be a numeric type and the maximum size of the "outer" vector is limited by the maximum integer value that can be stored in a variable of type T. Most access operations are supported (including iterator access), but not modifying operations such as "push_back".  

The motivation for this class is that it is more memory efficient than a vector< vector<T> >. It is especially used to store inverted index for the occurrences of labels in a tree database. The first element of a pair is the position of the occurrence of the label, while the second is the position of the parent node.

*/

template<typename T>
class ArrayOfPairs{
    T *array;
    //T* copyArray;
    //bool toDelete;
    public:
    ArrayOfPairs(const ArrayOfPairs<T>& other)
        {
        if (other.array!=NULL)
            {
            T size=other.size();
            size_t memsize=1+size*2;
            array=new T[memsize];
            std::copy(other.array,other.array+memsize,array);
            }
        else
            {
            array=NULL;
            }
        };

        
    void concatenateWith(const ArrayOfPairs<T>& other)
        {
        if (other.array==NULL) return;
        
        if (array==NULL)
            {
            T size=other.size();
            size_t memsize=1+size*2;
            array=new T[memsize];
            std::copy(other.array,other.array+memsize,array);
            }
        else
            {
            T other_size=other.size();
            if (other_size==0) return;
            
            size_t other_memsize=1+other_size*2;
            T self_size=size();
            size_t self_memsize=1+self_size*2;
            
            assert ((unsigned int) self_size + (unsigned int) other_size <= std::numeric_limits<T>::max());
            if ((unsigned int) self_size + (unsigned int) other_size >= std::numeric_limits<T>::max())
                {
                throw std::exception();
                }
            
            T new_size=self_size+other_size;
            size_t new_memsize=1+new_size*2;
            T* new_array=new T[new_memsize];
            new_array[0]=new_size;
            
            assert (new_memsize==self_memsize+other_memsize-1);
            
            std::copy(array+1,array+self_memsize,new_array+1);
            std::copy(other.array+1,other.array+other_memsize,new_array+self_memsize);
            
            delete[] array;
            array=new_array;
            
            }
        }
        
    void swap(ArrayOfPairs<T>& other)
        {
        T* temp;
        temp=array;
        array=other.array;
        other.array=temp; 
        };

     ArrayOfPairs<T>& operator= (ArrayOfPairs<T> const& other)
         {

            if (this == &other) return *this;

            if (array!=NULL)
                {
                delete[] array;
                }

            if (other.array!=NULL )
                {

                T size=other.size();
                size_t memsize=1+size*2;
                array=new T[memsize];
                std::copy(other.array,other.array+memsize,array);
                return *this;
                }
            else
                {
                array=NULL;
                return *this;
                }
         };

    ArrayOfPairs(size_t nbOfPairs){
        assert( nbOfPairs<=std::numeric_limits<T>::max()  );
        try{
            array=new T[nbOfPairs*2+1];
            }
        catch(std::bad_alloc &e)
            {
            array=NULL;
            throw;
            }
        array[0]=nbOfPairs;
        }

    ArrayOfPairs():array(NULL){}
    
    ~ArrayOfPairs()
        {
         //cerr<<" destructor: "<<(int)array<<endl;

        if (array!=NULL)
            {
            delete[] array;
            }
        }

    size_t size() const
        {
        if (array==NULL)
            {
            throw std::exception();
            }
        return array[0];
        }

    T* operator[](size_t i) const{
        return array+1+i*2;

        }

    typedef T* value_type;

    template <class Value>
    class pair_iterator
      : public boost::iterator_adaptor<
            pair_iterator<Value>                // Derived
          , Value                          // Base
          , Value              // Value
          , boost::forward_traversal_tag    // CategoryOrTraversal
          , const Value& //Reference
        >
    {
     private:
        struct enabler {};  // a private type avoids misuse

     public:
        pair_iterator()
          : pair_iterator::iterator_adaptor_(0) {}

        explicit pair_iterator(Value p)
          : pair_iterator::iterator_adaptor_(p) {}

        template <class OtherValue>
        pair_iterator(
            pair_iterator<OtherValue> const& other
          , typename boost::enable_if<
                boost::is_convertible<OtherValue,Value>
              , enabler
            >::type = enabler()
        )
          : pair_iterator::iterator_adaptor_(other.base()) {}

     private:
        friend class boost::iterator_core_access;

        template <class OtherValue>
        bool equal(pair_iterator<OtherValue> const& other) const
        {
            return this->base() == other.base();
        }

        void increment() { this->base_reference() += 2; }
        const Value &dereference() const{return this->base();}
    };

    typedef pair_iterator<T*> iterator;
    typedef pair_iterator<const T* > const_iterator;

    template<class Archive>
    void load(Archive & ar, const unsigned int version)
        {
        T size;
        ar & size;
        resize_with_erase(size);
        for (size_t i=0; i<size;++i)
            {
            ar& array[1+i*2];
            ar& array[1+i*2+1];
            }
        };

    template<class Archive>
    void save(Archive & ar, const unsigned int version) const
    {
        T local_size;  // Strangely, ar & size()  does not work
        local_size=size();
        ar & local_size;// local_size;
        for (size_t i=0; i<local_size;++i)
            {
            ar& array[1+i*2];
            ar& array[1+i*2+1];
            }
    };

    BOOST_SERIALIZATION_SPLIT_MEMBER();


    void resize_with_erase(size_t new_nb_of_pairs){
    
        assert( new_nb_of_pairs<=std::numeric_limits<T>::max()  );
    
        if (array!=NULL)
            {
            delete[] array;
            }
        array=new T[new_nb_of_pairs*2+1];

        array[0]=new_nb_of_pairs;
        }

    void setPair(size_t i,T first, T second)
        {
        assert(i<size());
        array[1+2*i]=first;
        array[1+2*i+1]=second;
        };

    iterator begin(){return iterator(array+1);};
    const_iterator begin() const{return const_iterator(array+1);};
    iterator end()
        {
        T nbOfPairs=array[0];
        assert(nbOfPairs>0);
        return iterator(array+nbOfPairs*2+1);
        };
    const_iterator end() const
        {
        T nbOfPairs=array[0];
        assert(nbOfPairs>0);
        return const_iterator(array+nbOfPairs*2+1);
        };

};

template<typename T>
std::ostream &operator<< (std::ostream &stream,const ArrayOfPairs<T> &ap)
  {

  stream<<"AP[";
  BOOST_FOREACH(const T* pElem,ap)
    {
    stream<<"("<<(int)pElem[0]<<","<<(int)pElem[1]<<")";
    }
  stream<<"]";
  return stream;
  }

#endif

