/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*! \file
    \ingroup treelet_computing_group
    \brief Manage the already computed treelets; these treelets will be used for computing larger treelets.
*/

#include <vector>
#include <map>

#include "TypeDefinitions.hpp"
#include "Treelets.hpp"

#ifndef _HPP_COMPUTED_SUBTREELETS_MANAGER_
#define _HPP_COMPUTED_SUBTREELETS_MANAGER_

/**************************************************************************
 *     Management of the list of computed subtreelets
 */

class TreeletIdNbOccPair
    {
    TreeletId id;
    size_t nbOcc;
    friend std::ostream &operator<< (std::ostream &stream,const TreeletIdNbOccPair &lst);
    public:
    TreeletIdNbOccPair(TreeletId id,size_t nbOcc):id(id),nbOcc(nbOcc){};
    bool operator<(const TreeletIdNbOccPair &other) const
        {
        return nbOcc<other.nbOcc;
        };
    TreeletId getId()
        {
        return id;
        }
    
    };

std::ostream &operator<< (std::ostream &stream,const TreeletIdNbOccPair &lst)
  {
  stream<<"("<<lst.id<<","<<lst.nbOcc<<")";
  return stream;
  }

class SubTreeletsList
    {
    bool isSorted;
    std::vector< TreeletIdNbOccPair> treeletsNbOccPairList ;
    std::vector<TreeletId> sortedSubTreelets; 
    void sortSubTreelets()
        {
        std::sort(treeletsNbOccPairList.begin(),treeletsNbOccPairList.end());
        sortedSubTreelets.clear();
        BOOST_FOREACH(TreeletIdNbOccPair idOcc, treeletsNbOccPairList)
            {
            sortedSubTreelets.push_back(idOcc.getId());
            }
        isSorted=true;
        }
    friend std::ostream &operator<< (std::ostream &stream,const SubTreeletsList &lst);
    public:
        SubTreeletsList():isSorted(true){};
        
        void addSubtreeletId(TreeletId subTreeletIndex,size_t nbOcc)
            {
            assert(nbOcc>0);
            treeletsNbOccPairList.push_back( TreeletIdNbOccPair(subTreeletIndex,nbOcc) );
            isSorted=false;
            }
        const std::vector<TreeletId> &getSubtreeletsIdOrderedByNbOcc() const
            {
            if(!isSorted)
                {
                const_cast<SubTreeletsList *>(this)->sortSubTreelets();
                }
            return sortedSubTreelets;
            }
    };

std::ostream &operator<< (std::ostream &stream,const SubTreeletsList &lst)
  {
  
  stream<<lst.getSubtreeletsIdOrderedByNbOcc();
  stream<<" ";
  stream<<lst.treeletsNbOccPairList;
  return stream;
  }

template<typename TreeletType>
class ComputedSubTreeletsManagerGeneric
    {
    std::map<TreeletType,SubTreeletsList > computedSubTreelets;
    friend std::ostream &operator<< (std::ostream &stream,const ComputedSubTreeletsManagerGeneric<TreeletType> &csm) {
      stream<<csm.computedSubTreelets;
      return stream;
    }
    public:
        void addSubTreeletId(const TreeletType &treelet,TreeletId subTreeletIndex,size_t nbOcc)
            {
            computedSubTreelets[treelet].addSubtreeletId(subTreeletIndex,nbOcc);
            };
        const vector<TreeletId> &getSubtreeletsIdOrderedByNbOcc(const TreeletType &treelet) const
            {
            typename std::map<TreeletType,SubTreeletsList >::const_iterator it=computedSubTreelets.find(treelet);
            if (it==computedSubTreelets.end() )
                {
                assert(false);
                }
            return (it->second).getSubtreeletsIdOrderedByNbOcc();
            };
    };

//typedef ComputedSubTreeletsManagerGeneric<Treelet> ComputedSubTreeletsManager;
    
// std::ostream &operator<< (std::ostream &stream,const ComputedSubTreeletsManager &csm)
  // {
  // stream<<csm.computedSubTreelets;
  // return stream;
  // }

/*
 *     END of Management of the list of computed subtreelets
 **************************************************************************/
 
//int main(){ return 0; }
#endif 
 
