/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*! \file
    \ingroup global_glue_group
    \brief Generic Class for combining Treelet Retrieving and post-processing.
*/

#include "TreeDB.hpp"
#include "ResultPostProcessing.hpp"
#include <boost/smart_ptr.hpp>
#include "TypeDefinitions.hpp" 
#include "StringConversions.hpp"
#include "DefaultDataPath.hpp"
 
#include <vector>

#ifndef _HPP_GLOBAL_TR_
#define _HPP_GLOBAL_TR_

template<typename TreeletType>
class TreeletRetrieverBaseGeneric
    {
    protected:
    typedef TreeDBGeneric<TreeletType> TreeDB;
    typedef ResultsHandlerInterfaceGeneric<TreeletType> ResultsHandlerInterface;
    protected:
    
    // Tree Retrieving components:
    TreeDB treeDB;
    //bool preserveOrder;
    //int nbSiblingsLimit;
    // Request Processing components:
    LabelIndexer indexer;
    //bool outputAllDiscovered;
    public:
   

        TreeletRetrieverBaseGeneric(std::string datapath, std::string side,
                        const SearchParameters& sParams
                        //bool preserveOrder, int nbSiblingsLimit, bool skipTrailingWildcard
                                ):
            treeDB(DataFilenameSufixes::getInvertedIndexFilename(datapath,side).c_str(),sParams)//skipTrailingWildcard), 
                                 
            //preserveOrder(preserveOrder),
            //nbSiblingsLimit(nbSiblingsLimit)
            {
            loadIndexerData(DataFilenameSufixes::getLabelIndexerFilename(datapath,side).c_str());
            }
       
       
        void loadIndexerData(const char *filename)
            {
            std::ifstream ifs(filename,ifstream::binary);
            try
                {
                boost::archive::binary_iarchive ia(ifs);
                ia>>indexer;
                }
            catch (boost::archive::archive_exception e)
                {
                    std::cerr<<"Error loading Label Indexer from file "<<filename<<endl;
                    std::cerr<<"what:"<<e.what()<<endl;
                    throw e;
                }                
            } 
        
        
        void _handleRequest(const std::vector<LabelType>& idxList, const std::vector<PosType>& dependList, ResultsHandlerInterface& resultHandler) const
            {
            boost::shared_ptr<Node>  pQueryTree= createTreeFromDependAndLabelList(idxList , dependList);   
            Node * rawpQueryTree=pQueryTree.get();
            treeDB.retrieveTreelets(rawpQueryTree,
                //preserveOrder, nbSiblingsLimit,
                    resultHandler);
            //treeDB.retrieveTreelets(resProc.getHandler(output) );
            }
            
        void _handleRequestAsString(const std::string& labelListAsString, const std::string& dependListAsString, 
                                            ResultsHandlerInterface& resultHandler) const
            {
            std::vector<PosType> dependList=parseDependList(dependListAsString);
            std::vector< std::string> splittedString = splitUtf8String(labelListAsString);
            std::vector<LabelType> idxList= indexer.convertVectorOfLabelWithOOVId(splittedString);               
            _handleRequest(idxList,dependList,resultHandler);
            }
    };   
    
//typedef TreeletRetrieverBaseGeneric<Treelet> TreeletRetrieverBase;
    
template<typename TreeletType>
class TreeletRetrieverWithContinuousTargetSideGeneric: public TreeletRetrieverBaseGeneric<TreeletType>
    {
    typedef IndividualResultsProcessorGeneric<TreeletType> IndividualResultsProcessor;
    typedef CachedPostProcessingHandlersFactoryGeneric<TreeletType> CachedPostProcessingHandlersFactory;
    typedef typename TreeletRetrieverBaseGeneric<TreeletType>::ResultsHandlerInterface ResultsHandlerInterface;
    protected:
    
    // // Tree Retrieving components:
    // TreeDB treeDB;
    // bool preserveOrder;
    // int nbSiblingsLimit;
    // // Request Processing components:
    // LabelIndexer indexer;
    
    
    // Post-processing components:
    PostProcessingInfos<AdditionalInfosUnpacker> postProcessingInfos;//UsingMassStorage> ;
    CachedPostProcessingHandlersFactory postProcessingHandlerFactory;

    public:
    

        TreeletRetrieverWithContinuousTargetSideGeneric(std::string datapath, std::string side, 

                        //size_t maxNumberOfSamplePerTreelet, 
                        //bool preserveOrder, int nbSiblingsLimit,
                        //bool filterOutDominated,
                        //bool skipTrailingWildcards,
                                //bool keepElementaryWildcard,
                                //bool keepTopWilcardChain
                        const SearchParameters& sParams,
                        const PostProcessingParameters& ppParams
                                
                                ):
            TreeletRetrieverBaseGeneric<TreeletType>(datapath,side,sParams), //preserveOrder,nbSiblingsLimit,skipTrailingWildcards),
            
            postProcessingInfos(DataFilenameSufixes::getTargetInfoFilename(datapath,side),
                                DataFilenameSufixes::getMostFrequentLabelFilename(datapath,side),
                                DataFilenameSufixes::getNumTreeToIdFilename(datapath),
                                DataFilenameSufixes::getAdditionalDatasFilename(datapath,side)
                                ),
                                
            postProcessingHandlerFactory(postProcessingInfos.getTargetInfos(),
                                  postProcessingInfos.getMostFrequentLabelIds(),
                                  //maxNumberOfSamplePerTreelet,
                                  50,   //sizeCache
                                  TreeletRetrieverBaseGeneric<TreeletType>::treeDB,
                                  //filterOutDominated,//keepElementaryWildcard,
                                  //keepTopWilcardChain,
                                  ppParams
                                  )
            {}
       
 
        
        void handleRequestAsStringWithGivenProcessor(const std::string& labelListAsString, 
                                                    const std::string& dependListAsString,
                                                    IndividualResultsProcessor& processor
                                                    ) const
            {
            boost::shared_ptr<ResultsHandlerInterface> postProcessingHandler=
                                            postProcessingHandlerFactory.createHandler(processor);
            TreeletRetrieverBaseGeneric<TreeletType>::_handleRequestAsString(labelListAsString, dependListAsString,*postProcessingHandler);
            }
 
    };

typedef TreeletRetrieverWithContinuousTargetSideGeneric<Treelet> TreeletRetrieverWithContinuousTargetSide;

template<typename TreeletType>
class TreeletRetrieverSimpleGeneric: public TreeletRetrieverBaseGeneric<TreeletType>
    {
    typedef IndividualResultsProcessorGeneric<TreeletType> IndividualResultsProcessor;
    typedef CachedPostProcessingHandlersFactoryGeneric<TreeletType> CachedPostProcessingHandlersFactory;
    typedef typename TreeletRetrieverBaseGeneric<TreeletType>::ResultsHandlerInterface ResultsHandlerInterface;
    protected:

    std::vector<std::string> numToTreeIdVector;
    TargetList emptyTargetList;
    CachedPostProcessingHandlersFactory postProcessingHandlerFactory;

    public:
   
    void loadNumToTree(const std::string& filename)
        {
        std::ifstream ifs(filename.c_str(),std::ifstream::binary);
        try
            {
            boost::archive::binary_iarchive ia(ifs);
            ia>>numToTreeIdVector;
            //std::cout<<"numToTree 132069->"<<numToTreeIdVector.at(132069)<<std::endl;
            }
        catch (boost::archive::archive_exception e)
            {
                cerr<<"Error loading numToTreeIdVector from file "<<filename<<endl;
                cerr<<"what:"<<e.what()<<endl;
                throw e;
            }               
        
        }
        
        TreeletRetrieverSimpleGeneric(std::string datapath, std::string side, 

                        //size_t maxNumberOfSamplePerTreelet, 
                        //bool preserveOrder, int nbSiblingsLimit,
                        //bool filterOutDominated,
                        //bool skipTrailingWildcards,
                        const SearchParameters& sParams,
                        const PostProcessingParameters& ppParams
                        //bool keepElementaryWildcard,
                        //bool keepTopWilcardChain
                                ):
                        TreeletRetrieverBaseGeneric<TreeletType>(datapath,side,sParams),//preserveOrder, nbSiblingsLimit,skipTrailingWildcards),                              
            postProcessingHandlerFactory(
                                  emptyTargetList,
                                  //maxNumberOfSamplePerTreelet,
                                  //filterOutDominated,keepElementaryWildcard,
                                  //keepTopWilcardChain,
                                  ppParams
                                  )
            {
            loadNumToTree(DataFilenameSufixes::getNumTreeToIdFilename(datapath));
            }
       
 
        
        void handleRequestAsStringWithGivenProcessor(const std::string& labelListAsString, 
                                                    const std::string& dependListAsString,
                                                    IndividualResultsProcessor& processor
                                                    ) const
            {
            boost::shared_ptr<ResultsHandlerInterface> postProcessingHandler=
                                            postProcessingHandlerFactory.createHandler(processor);
            _handleRequestAsString(labelListAsString, dependListAsString,*postProcessingHandler);
            }
  


    };
    
    
#endif
