/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*! \file
    \ingroup data_creation_group
    \brief Binary format encoding functions.
*/

#include <vector>
#include <map>
#include <fstream>
#include <cassert>
#include <iostream>
#include <string>
#include <boost/foreach.hpp>
#include <boost/bind.hpp>

#include "utf8cpp/utf8.h" 

#include "TypeDefinitions.hpp"
#include "AdditionalInfos.hpp"

#include <boost/serialization/vector.hpp>
#include <boost/serialization/base_object.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>

#include <boost/smart_ptr.hpp>
#include <boost/smart_ptr/make_shared.hpp>
#include <exception>
#include <limits>

#ifndef _HPP_INDEXER_AND_ENCODING__
#define _HPP_INDEXER_AND_ENCODING__



struct AnalyzedIData
    {
    
    
    std::vector<std::string> labelList;
    std::vector<PosType> depend;
    
    OtherInfosContainer others;

    // std::vector<std::string> phrase_f_list;
    // std::vector<std::string> word_f_list;
    
    /*
    std::vector<std::string> POSList;
    std::vector<std::string> conjList;
    std::vector<std::string> content_p;
    
    std::vector<std::string> catList;
    std::vector<std::string> typeList;
    */
    
    };

std::ostream& operator<<(std::ostream &stream, const AnalyzedIData& aid)
    {
    stream<<"idata: "<< aid.labelList<<" | "<< aid.depend<<" | "<< aid.others;
    return stream;
    }

class IndexerException: public std::exception
{
  virtual const char* what() const throw()
  {
    return "IndexerException";
  }
};

template<typename CodeType>
class OrdererOfLabelByCount
    {
    const std::map<CodeType,size_t>& countOfEachLabel;
    public:
    OrdererOfLabelByCount(const std::map<CodeType,size_t>& countOfEachLabel):countOfEachLabel(countOfEachLabel){};
    bool operator()(LabelType a,LabelType b)
        {
        size_t countOfA,countOfB;
        typename std::map<CodeType,size_t>::const_iterator itA=countOfEachLabel.find(a);
        if ( itA==countOfEachLabel.end() )
            {
            countOfA=0;
            }
        else
            {
            countOfA=itA->second;
            }

        typename std::map<CodeType,size_t>::const_iterator itB=countOfEachLabel.find(b);
        if ( itB==countOfEachLabel.end() )
            {
            countOfB=0;
            }
        else
            {
            countOfB=itB->second;
            }
            
        return countOfA>countOfB;
        }
    
    };

template<typename CodeType> 
class GenericStringIndexer
    {
    std::vector<std::string> idToLabel;
    std::map<std::string,CodeType> labelToId;
    
    std::map<CodeType,size_t> countOfEachCode;
    bool keepTrackOfCodeCount;
    
    friend class boost::serialization::access;
    template<class Archive> void serialize(Archive & ar, const unsigned int version)
        {
            // serialize base class information
            ar & idToLabel;
            ar & labelToId;
        }
        
    template<CodeType> 
    friend std::ostream &operator<< (std::ostream &stream,const GenericStringIndexer<CodeType> &li);
       
    public:
    
        
    
        GenericStringIndexer(bool keepTrackOfCodeCount=false):keepTrackOfCodeCount(keepTrackOfCodeCount){}
        GenericStringIndexer(const std::vector<std::string>& idToLabel,
                            const std::map<std::string,CodeType>& labelToId):
                            idToLabel(idToLabel),labelToId(labelToId),
                            keepTrackOfCodeCount(false){}
        
        CodeType getOOVId() const {
            return std::numeric_limits<CodeType>::max();
        }
        
        std::vector<CodeType>  getCodeVectorOrderedByCount() const
            {
            assert(keepTrackOfCodeCount);
            std::vector<CodeType> orderedCodes;
            typedef typename std::map<CodeType,int>::value_type MapPair;
            BOOST_FOREACH(const MapPair& pair,countOfEachCode)
                {
                orderedCodes.push_back(pair.first);
                }
           // std::transform(countOfEachCode.begin(),countOfEachCode.end(),
           //                         std::back_inserter(orderedCodes),
            //                        boost::bind(& (typename std::map<CodeType,int>::value_type::first),_1));
                                    //std::select1st<std::map<LabelType,int>::value_type>());
            std::sort(orderedCodes.begin(),orderedCodes.end(),OrdererOfLabelByCount<CodeType>(countOfEachCode));
            return orderedCodes;
            }
            
       
        std::pair<std::vector<std::string>, std::vector<CodeType> > getNewOptimalCodingVectors()
            {
            std::vector<CodeType> orderedCodes=getCodeVectorOrderedByCount();
            
            std::vector<std::string> newDecoderVector;
            newDecoderVector.reserve(idToLabel.size());
            
            std::vector<CodeType> toNewCodeTranslationVector(idToLabel.size());
            
            for (CodeType code=0;code<idToLabel.size();++code)
                {
                CodeType oldCode=orderedCodes[code];
                newDecoderVector.push_back(idToLabel[oldCode]);
                toNewCodeTranslationVector[oldCode]=code;
                }
            return std::make_pair(newDecoderVector,toNewCodeTranslationVector);
            }

        GenericStringIndexer<CodeType> getNewOptimalIndexer() const
            {
            std::vector<CodeType> orderedCodes=getCodeVectorOrderedByCount();
            
            std::vector<std::string> newDecoderVector;
            newDecoderVector.reserve(idToLabel.size());
            
            std::map<std::string,CodeType> newLabelToId;
            
            for (CodeType code=0;code<idToLabel.size();++code)
                {
                CodeType oldCode=orderedCodes[code];
                newDecoderVector.push_back(idToLabel[oldCode]);
                newLabelToId[ idToLabel[oldCode] ]=code;
                }
            return GenericStringIndexer<CodeType> (newDecoderVector,newLabelToId);
            }
        
        static void recodeCodeVector(std::vector<CodeType>&vect, const std::vector<CodeType>&toNewCodeTranslationVector)
            {
            for(size_t i=0;i<vect.size();++i)
                {
                vect[i]=toNewCodeTranslationVector[vect[i]];
                }
            }
        
        bool saveAs(const char *filename) const
            {
            std::ofstream ofs(filename,std::ofstream::binary);
            if (ofs.fail())
                {
                std::cerr<<"Error creating "<<filename<<std::endl;
                return false;
                }
            boost::archive::binary_oarchive oa(ofs);
            oa << *this;
            return true;
            }

        const std::vector<std::string>& getIdToLabelVector() const
            {
            return idToLabel;
            }

        CodeType getIdAndUpdateEnsuringUtf8(const std::string& label)
            {
            std::string stringEnsured;
            std::vector<uint32_t> utf32Vector;
            utf8::utf8to32(label.begin(), label.end(), std::back_inserter(utf32Vector));  
            utf8::utf32to8(utf32Vector.begin(), utf32Vector.end(),  std::back_inserter(stringEnsured) );
            return getIdAndUpdate(stringEnsured);
            }
   
        const std::string& getLabel(CodeType id) const
            {
            if (id>idToLabel.size())
                {
                throw IndexerException();
                }
            return idToLabel[id];    
            }
            
        CodeType getId(const std::string& label) const
            {
            typename std::map<std::string,CodeType>::const_iterator it=labelToId.find(label);
            if (it==labelToId.end())
                {
                throw IndexerException();
                }
            return it->second;
            }

        CodeType getIdWithOOVId(const std::string& label) const
            {
            typename std::map<std::string,CodeType>::const_iterator it=labelToId.find(label);
            if (it==labelToId.end())
                {
                return getOOVId();
                }
            return it->second;
            }
            
        
            
        CodeType getIdAndUpdate(const std::string& label)
            {
            typename std::map<std::string,CodeType>::iterator it=labelToId.find(label);
            CodeType code;
            if (it==labelToId.end())
                {
                idToLabel.push_back(label);
                code=idToLabel.size()-1;
                labelToId[label]=code;
                //std::cout<<"not found "<<label<<" new id: "<<id<<std::endl;
                }
            else
                {
                //std::cout<<"found "<<label<<" id: "<<it->second<<std::endl;
                code= it->second;
                }
            if (keepTrackOfCodeCount)
                {
                countOfEachCode[code]++;
                }
            return code;
            }
            
        std::vector<CodeType> convertVectorOfLabel(const std::vector<std::string>& vLabel) const
            {
            std::vector<CodeType> result;
            BOOST_FOREACH(const std::string& label, vLabel)
                {
                result.push_back(getId(label));
                }
            return result;
            }

        std::vector<CodeType> convertVectorOfLabelAndUpdate(const std::vector<std::string>& vLabel)
            {
            std::vector<CodeType> result;
            BOOST_FOREACH(const std::string& label, vLabel)
                {
                result.push_back(getIdAndUpdate(label));
                }
            return result;
            }  
            
        std::vector<CodeType> convertVectorOfLabelWithOOVId(const std::vector<std::string>& vLabel) const
            {
            std::vector<CodeType> result;
            BOOST_FOREACH(const std::string& label, vLabel)
                {
                result.push_back(getIdWithOOVId(label));
                }
            return result;
            }                
    };


template<typename CodeType>
std::ostream &operator<< (std::ostream &stream,const GenericStringIndexer<CodeType> &li)
  {
  stream<<li.labelToId;
  return stream;
  }
   
typedef GenericStringIndexer<LabelType> LabelIndexer;

typedef std::vector<std::string> UIntToStringVector;
typedef GenericStringIndexer<unsigned int> AddInfosIndexer;

/*********************************************************************:
 *  Encoding
 */
 
 
// Convert u to a big endian prefix representation (similar to varint in google protocol buffer)
void writeUintAsPrefixCode(size_t u,std::string &result)
    {
    while (1)
        {
        unsigned char codepoint=u & 127;
//        std::cout<<"u:"<<u<<std::endl;
//        std::cout<<"codepoint:"<<(int) codepoint<<std::endl;

        u>>=7;
        if (u==0)
            {
            result.push_back(codepoint);
            break;
            }
        else
            {
            codepoint|=128;
//            std::cout<<"codepoint up:"<<(int) codepoint<<std::endl;
            result.push_back(codepoint);
            }
        }
    }
    
 
    
// read one uint at position cursor and update cursor
// cursor is set to 0 if no uint could be read before the end of the sentence
inline unsigned int readOneUInt(const std::string& code, size_t& cursor)
    {
    unsigned int result=0;
    unsigned int n=0;
    unsigned char codepoint;
    size_t codeSize=code.size();
    do
        {
        if(cursor>=codeSize)
            {
            cursor=0;
            return 0;
            }
        codepoint=code[cursor];
//        std::cout<<"read codepoint:"<<codepoint<<std::endl;
        result+= ( ((unsigned int)(codepoint&127))<<(7*n) );
//        std::cout<<"result after codepoint:"<<result<<std::endl;
        cursor++;
        n++;        
        }
    while(codepoint&128);
    
    return result;
    }
    
class VectorStringPusher
    {
    std::vector<std::string>& result;
    public:
    VectorStringPusher(std::vector<std::string>& result):result(result){}
    
    void pushString(const std::string& s)
        {
        result.push_back(s);
        }   
    };
/*
class PerlArrayStringPusher
    {
    AV *array;
    public:
        PerlArrayStringPusher(AV* array):array(array){}
        void pushString(const std::string& s)
            {
            
            }
    }
*/   

template<typename StringPusher>
void decodePrefixCodingGeneric(StringPusher& sPusher,
                        const std::string& code, 
                        size_t& startCodePos,
                        const UIntToStringVector& indexer, size_t nbElemToDecode)
    {
    size_t cursor=startCodePos;
    for (size_t i=0;i<nbElemToDecode;++i)
        {
        unsigned int u=readOneUInt(code,cursor);
        assert(cursor!=0);
        assert(u<indexer.size() );
        const std::string& elem=indexer[u];
        sPusher.pushString(elem);        
        }
    startCodePos=cursor;
    }

void decodePrefixCoding(std::vector<std::string>& result,
                        const std::string& code, 
                        size_t& startCodePos,
                        const UIntToStringVector& indexer, size_t nbElemToDecode)
    {
    VectorStringPusher pusher(result);
    decodePrefixCodingGeneric(pusher,code,startCodePos,indexer,nbElemToDecode);
    }

class PrefixEncodeVectors
    {
    std::string& code; 

    public:
    PrefixEncodeVectors(std::string& code):code(code){}
    
    void packUInt(size_t u)
        {
        writeUintAsPrefixCode(u,code);
        }

    void packString(const std::string& str)
        {
        size_t nbElems=str.size();
        packUInt(nbElems);
        code+=str;
        }
        
    void packStringWithUpdateToIndexer(const std::string& str,AddInfosIndexer& indexer)
        {
        unsigned int uCode=indexer.getIdAndUpdate(str);
        packUInt(uCode);        
        }
        
    void packVectorString(const std::vector<std::string>& vString)
        {
        size_t nbElems=vString.size();
        packUInt(nbElems);
        BOOST_FOREACH(std::string elem, vString)
            {
            packString(elem);
            }
        }
        
    void packVectorStringWithUpdateToIndexer(const std::vector<std::string>& vString, AddInfosIndexer& indexer)
        {
        size_t nbElems=vString.size();
        packUInt(nbElems);
        BOOST_FOREACH(std::string elem, vString)
            {
            packStringWithUpdateToIndexer(elem,indexer);
            }
        }
        
    void mergeVectorStringsAndPackWithUpdateToIndexer(const std::vector< std::vector<std::string> >& listOfStringsVect,
                                                      const size_t arrayOfIndices[], const size_t sizeOfArrayOfIndices,
                                                        AddInfosIndexer& indexer)
        {
        
        if (sizeOfArrayOfIndices==1)
            {
            packVectorStringWithUpdateToIndexer(listOfStringsVect[arrayOfIndices[0] ],indexer);
            return;
            }
        
        int vectorLength=-1;
        
        for (size_t i=0;i<sizeOfArrayOfIndices;++i)
            {
            size_t indice=arrayOfIndices[i];
            if (vectorLength==-1)
                {
                vectorLength=listOfStringsVect[indice].size();
                #ifdef NDEBUG
                break;
                #endif
                }
            #ifndef NDEBUG
            else
                {
                assert(vectorLength==(int)listOfStringsVect[indice].size());
                }
            #endif
            }
        assert(vectorLength!=-1);
     
        std::vector<std::string> temp;
        temp.reserve(vectorLength);
        for(size_t j=0;(int)j<vectorLength;++j)
            {
            std::string codedGroup;
            PrefixEncodeVectors subEncoder(codedGroup);
            for (size_t i=0;i<sizeOfArrayOfIndices;++i)
                {
                size_t indice=arrayOfIndices[i];
                subEncoder.packString(listOfStringsVect[indice][j]);
                }
            temp.push_back(codedGroup);
            }
        packVectorStringWithUpdateToIndexer(temp,indexer);     
        }
        
        
    void packPosListUC(const std::vector<unsigned char>& pList)
        {
        size_t nbElems=pList.size();
        packUInt(nbElems);
        code.reserve(code.size()+pList.size());
        std::copy(pList.begin(),pList.end(), std::back_inserter(code)); 
        }

    void packPosSetUC(const std::set<unsigned char>& pList)
        {
        size_t nbElems=pList.size();
        packUInt(nbElems);
        code.reserve(code.size()+pList.size());
        std::copy(pList.begin(),pList.end(), std::back_inserter(code)); 
        }
        
    void packMatchListUC(const std::vector<std::pair<std::vector<unsigned char>,std::vector<unsigned char> >  >& matchList)
        {
        size_t nbPairs=matchList.size();
        packUInt(nbPairs);
        for(size_t i=0;i<nbPairs;++i)
            {
            packPosListUC(matchList[i].first);
            packPosListUC(matchList[i].second);
            }
        }
        

    void addRawDataToCode(const std::string& raw)
        {
        code+=raw;
        }

    };
    
class PrefixDecodeVectors
    {
    const std::string& code; 
    size_t readCursor;    
    public:
    PrefixDecodeVectors(const std::string& code):code(code),readCursor(0){}


    bool isFinished()
        {
        assert(readCursor<=code.size());
        return readCursor==code.size();
        }
    size_t readOnePackedUInt()
        {
        size_t result= readOneUInt(code, readCursor); 
        assert(readCursor>0);//readCursor is set to 0 if unexpected end of string
        return result;     
        }

    std::string unPackStringNoIndexer()
        {
        size_t nbElems=readOnePackedUInt();
        assert(nbElems+readCursor<=code.size());
        size_t start=readCursor;
        readCursor+=nbElems;
        return code.substr(start,nbElems);
        }

    void unpackVectorString(std::vector<std::string>& result)
        {
        size_t nbElems=readOnePackedUInt();
        for(size_t i=0;i<nbElems;++i)
            {
            result.push_back( unPackStringNoIndexer() );
            }
        }
        
    std::string unPackStringWithIndexer(const UIntToStringVector& indexer)
        {
        size_t nbElems=readOnePackedUInt();
        assert(nbElems+readCursor<=code.size());
        size_t start=readCursor;
        readCursor+=nbElems;
        return code.substr(start,nbElems);
        }

    void readPosListUCToVector(std::vector<unsigned char> &result)
        {
        size_t nbElems=readOnePackedUInt();
        assert(nbElems+readCursor<=code.size());
        result.reserve(nbElems);
        for(size_t i=0;i<nbElems;++i)
            {
            result.push_back(code[readCursor+i]);
            }
        readCursor+=nbElems;
        }

    void readPosListUCToSet(std::set<unsigned char> &result)
        {
        size_t nbElems=readOnePackedUInt();
        assert(nbElems+readCursor<=code.size());
        for(size_t i=0;i<nbElems;++i)
            {
            result.insert(code[readCursor+i]);
            }
        readCursor+=nbElems;
        }
        
    void readMatchListUC(std::vector<std::pair<std::vector<unsigned char>,std::vector<unsigned char> >  >& matchList)
        {

        size_t nbPairs=readOnePackedUInt();
        matchList.resize(nbPairs);
        for(size_t i=0;i<nbPairs;++i)
            {
            readPosListUCToVector(matchList[i].first);
            readPosListUCToVector(matchList[i].second);
            }
        }


     void readVectorString(std::vector<std::string>& result, const UIntToStringVector& indexer)
        {
        size_t nbElems=readOnePackedUInt();
        result.reserve(nbElems);
        decodePrefixCoding(result,code,readCursor,indexer,nbElems);
        }
        
    void readMergedVectorStrings(std::vector< std::vector<std::string> >& result,
                                                      const size_t arrayOfIndices[], const size_t sizeOfArrayOfIndices,
                                                        const UIntToStringVector& indexer)
        {
        if (sizeOfArrayOfIndices==1)
            {
            readVectorString(result[arrayOfIndices[0]],indexer);
            return;
            }
        std::vector<std::string> temp;
        readVectorString(temp,indexer);
        size_t vectorLength=temp.size();
        
        for(size_t j=0;j<vectorLength;++j)
            {
            std::string& codedGroup=temp[j];
            PrefixDecodeVectors subDecoder(codedGroup);
            for (size_t i=0;i<sizeOfArrayOfIndices;++i)
                {
                size_t indice=arrayOfIndices[i];
                assert(indice<result.size());
                result[indice].push_back(subDecoder.unPackStringNoIndexer());
                }
            }            
        }
        

    
    };

struct OtherInfosEncoder
    {
    std::vector<AddInfosIndexer> indexVector;
    bool useGroups;
    
    OtherInfosEncoder():indexVector(OtherInfosContainer::nbOthers),useGroups(false)
        {}

    OtherInfosEncoder(bool useGroups):useGroups(useGroups)
        {
        size_t nbVectors;
        if (useGroups)
            {
            nbVectors=OtherInfosContainer::nbGroups;
            //indexVector.resize(OtherInfosContainer::nbGroups);
            }
        else
            {
            nbVectors=OtherInfosContainer::nbOthers;
            //indexVector.resize(OtherInfosContainer::nbOthers);
            }
        indexVector.reserve(nbVectors);
        for(size_t i=0;i<nbVectors;++i)
            {
            indexVector.push_back(AddInfosIndexer(true));
            }
        }

    OtherInfosEncoder(bool useGroups,const std::vector<AddInfosIndexer>& indexVector):useGroups(useGroups)
        {
        for (size_t i=0;i<indexVector.size();++i)
            {
            this->indexVector.push_back(indexVector[i]);
            }
        }

    OtherInfosEncoder getOptimalEncoder() const
        {
        std::vector<AddInfosIndexer> optimalIndexVector;
        for (size_t i=0;i<indexVector.size();++i)
            {
            optimalIndexVector.push_back(indexVector[i].getNewOptimalIndexer());
            }
        return OtherInfosEncoder(useGroups, optimalIndexVector);
        }

    void encode(const OtherInfosContainer &otherInfos , PrefixEncodeVectors& encoder)
        {
        if (useGroups)
            {
            encodeWithGroups(otherInfos , encoder);
            }
        else
            {
            encodeSimple(otherInfos , encoder);
            }
        }
       
    void encodeSimple(const OtherInfosContainer &otherInfos , PrefixEncodeVectors& encoder)
        {
        assert(!useGroups);
        for(size_t i=0;i<OtherInfosContainer::nbOthers;++i)
            {
            encoder.packVectorStringWithUpdateToIndexer(otherInfos.otherInfosVect[i],indexVector[i]);
            }
        }

    void encodeWithGroups(const OtherInfosContainer &otherInfos , PrefixEncodeVectors& encoder)
        {
        assert(useGroups);
        for(size_t numGroup=0;numGroup<OtherInfosContainer::nbGroups;++numGroup)
            {
            encoder.mergeVectorStringsAndPackWithUpdateToIndexer(otherInfos.otherInfosVect,
                                                            OtherInfosContainer::groups[numGroup],OtherInfosContainer::sizeGroups[numGroup],  
                                                                        indexVector[numGroup]);
            }
        }
        
    void saveToBoostArchive(boost::archive::binary_oarchive& oa) const
        {
        for(size_t i=0;i<indexVector.size();++i)
            {
            
            size_t sizeTot=0;
            BOOST_FOREACH(const std::string& elem,indexVector[i].getIdToLabelVector())
                {
                sizeTot+=elem.length();
                }
            std::cout<<"indexeur "<<i<<" has "<<indexVector[i].getIdToLabelVector().size()<<" elements/ total size:"<<sizeTot<<std::endl;
            oa<<indexVector[i].getIdToLabelVector();
            }
        }
    
    };



struct OtherInfosDecoder
    {
    std::vector<UIntToStringVector > indexVector;
    bool useGroups;

    OtherInfosDecoder():indexVector(OtherInfosContainer::nbOthers),useGroups(false)
        {}

    OtherInfosDecoder(bool useGroups):useGroups(useGroups)
        {
        if (useGroups)
            {
            indexVector.resize(OtherInfosContainer::nbGroups);
            }
        else
            {
            indexVector.resize(OtherInfosContainer::nbOthers);
            }
        }

    OtherInfosDecoder(const OtherInfosEncoder& encoder):useGroups(encoder.useGroups)
        {
        for(size_t i=0;i<encoder.indexVector.size();++i)
            {
            this->indexVector.push_back(encoder.indexVector[i].getIdToLabelVector());
            }
        }

    void decode(OtherInfosContainer &otherInfos , PrefixDecodeVectors& decoder ) const
        {
        if (useGroups)
            {
            decodeWithGroups(otherInfos , decoder );
            }
        else
            {
            decodeSimple(otherInfos , decoder );
            }
        }
                
    void decodeSimple(OtherInfosContainer &otherInfos , PrefixDecodeVectors& decoder ) const
        {

        for(size_t i=0;i<OtherInfosContainer::nbOthers;++i)
            {
            decoder.readVectorString(otherInfos.otherInfosVect[i],indexVector[i]);
            }
        }

    void decodeWithGroups(OtherInfosContainer &otherInfos , PrefixDecodeVectors& decoder) const
        {
        assert(useGroups);
        for(size_t numGroup=0;numGroup<OtherInfosContainer::nbGroups;++numGroup)
            {
            decoder.readMergedVectorStrings(otherInfos.otherInfosVect,
                                                            OtherInfosContainer::groups[numGroup],OtherInfosContainer::sizeGroups[numGroup],  
                                                                        indexVector[numGroup]);
            }
        }

    void readFromBoostArchive(boost::archive::binary_iarchive& ia)
        {
        size_t nbIndexers;
        if(useGroups)
            {
            nbIndexers=OtherInfosContainer::nbGroups;
            }
        else
            {
            nbIndexers=OtherInfosContainer::nbOthers;
            }
        for(size_t i=0;i<nbIndexers;++i)
            {
            ia>>indexVector[i];
            }
        }

    void saveToBoostArchive(boost::archive::binary_oarchive& oa) const
        {
        for(size_t i=0;i<indexVector.size();++i)
            {
            oa<<indexVector[i];
            }
        }

        
    };



    
class AdditionalInfosUnpackerBase
    {
    protected:
    OtherInfosDecoder oiDecoderSource;
    OtherInfosDecoder oiDecoderTarget;
    
    public:
    virtual ~AdditionalInfosUnpackerBase(){};
    
    AdditionalInfosUnpackerBase(const char *filename,bool useGroups=OtherInfosContainer::useGroups):
                                    oiDecoderSource(useGroups),oiDecoderTarget(useGroups)
        {
        std::ifstream ifs(filename,std::ifstream::binary);
        if (!ifs.is_open())
            {
            std::cerr<<"Error opening "<<filename<<std::endl;
            }
        boost::archive::binary_iarchive ia(ifs);
        oiDecoderSource.readFromBoostArchive(ia);
        oiDecoderTarget.readFromBoostArchive(ia);

        //ia>>vectAddInfos;
        }

    // for being called by descendant class:
    AdditionalInfosUnpackerBase(bool useGroups=OtherInfosContainer::useGroups):
                                    oiDecoderSource(useGroups),oiDecoderTarget(useGroups)
        {}
      
    void unpack(const std::string& code,AdditionalInfos& ai) const
        {
        PrefixDecodeVectors decoder(code);
        unpackWithGivenDecoder(decoder,ai);
        }

    void unpackWithGivenDecoder(PrefixDecodeVectors& decoder,AdditionalInfos& ai) const
        {
        decoder.readPosListUCToVector(ai.sourceDepend);
        
        oiDecoderSource.decode(ai.sourceOtherInfos,decoder);
        oiDecoderTarget.decode(ai.targetOtherInfos,decoder);
        
        unpackMoreThings(decoder,ai);
        
        }
        
    virtual void unpackMoreThings(PrefixDecodeVectors& decoder,AdditionalInfos& ai) const
        {}
        
    };


        
class AddInfosVectorInterface
    {
    public:
    virtual ~AddInfosVectorInterface(){};
    virtual size_t size() const=0;
    virtual AdditionalInfos at(size_t i) const=0;
    virtual std::string getRawData(size_t i) const=0;
    };
    
class AddInfosVectorAsVectorImplementationInterface: public AddInfosVectorInterface
    {
    public:
    virtual void takeOwnershipOfVectDataAndIndexers(std::vector<std::string>& vectAddInfos,
                                            const OtherInfosDecoder& oiDecoderSource,
                                            const OtherInfosDecoder& oiDecoderTarget)=0;
    virtual const std::string& getRawDataAsRef(size_t i) const=0;
        //{throw std::logic_error("Not Implemented");}
    };
        
typedef long int FileOffset;
    
template<typename NumType>
bool isInAscendingOrder(const std::vector<NumType>& vect)
    {
    for(size_t i=0;i<vect.size()-1;++i)
        {
        if (vect[i]>vect[i+1])
            {
            return false;
            }
        }
    return true;
    }
    
class AdditionalInfosUnpackerUsingMassStorage: public AdditionalInfosUnpackerBase, 
                                                public AddInfosVectorInterface
    {
    protected:
    std::vector<FileOffset> offsetVect;
    std::ifstream datafile;//(filename,std::ifstream::binary);
    FileOffset startOfDataOffset;

    public:
    AdditionalInfosUnpackerUsingMassStorage(const char *filename,bool useGroups=OtherInfosContainer::useGroups):
                                AdditionalInfosUnpackerBase(useGroups),
                                datafile(filename,std::ifstream::binary)
    {
    std::cout<<"loading additional data (with mass storage) from "<<filename<<std::endl;
    if (!datafile.is_open())
        {
        std::cerr<<"Error opening "<<filename<<std::endl;
        }
    boost::archive::binary_iarchive ia(datafile);
    oiDecoderSource.readFromBoostArchive(ia);
    oiDecoderTarget.readFromBoostArchive(ia);
    ia>>offsetVect;
    
    assert(isInAscendingOrder(offsetVect));
    
    startOfDataOffset=datafile.tellg();
    }
    
    size_t size() const
        {
        return offsetVect.size()-1;
        }
    
    AdditionalInfos at(size_t i) const
        {
        AdditionalInfos result;
        unpack(getRawData(i),result);
        return result;
        }   
    
    std::string getRawData(size_t i) const
        {
        //std::cout<<"i: "<<i<<" offsetVector: "<<offsetVect.size() <<std::endl;
        assert(i+1<offsetVect.size());
        FileOffset offset=offsetVect[i];
        FileOffset startPos=startOfDataOffset+offset;
        FileOffset sizeOfData=offsetVect[i+1]-offset;
        
        //std::cout<<"startPos: "<<startPos<<std::endl;
        const_cast<std::ifstream*>(&datafile)->seekg(startPos);
        
        //std::cout<<"sizeOfData: "<<sizeOfData<<std::endl;
        boost::scoped_array<char> buffer(new char[sizeOfData]);
        //std::cout<<"before reading"<<std::endl;
        const_cast<std::ifstream*>(&datafile)->read(buffer.get(),sizeOfData);
        //std::cout<<"after reading"<<std::endl;
        std::string result(buffer.get(),sizeOfData);   
        return result;
        }
    
    
    };
    
void saveVectorStringForMassStorageAccess(const std::vector<std::string>& vect,
                                                boost::archive::binary_oarchive& oa,
                                                std::ofstream& ofs)
    {
    size_t nbStrings=vect.size();
    
    std::vector<FileOffset> offsetVect;
    offsetVect.reserve(nbStrings+1);
    
    FileOffset cur=0;
    offsetVect.push_back(cur);
    BOOST_FOREACH(const std::string& str,vect)
        {
        cur+=str.length();
        offsetVect.push_back(cur);
        }
  
    oa<< offsetVect;
#ifndef NDEBUG    
    FileOffset startOffset=ofs.tellp();
#endif
    for(size_t i=0;i<nbStrings;++i)
        {
        //std::cout<<"start:"<<startOffset<<" tell:"<<ofs.tellp()<<" offset:"<<offsetVect[i]<<std::endl;
        assert(((FileOffset)ofs.tellp())==(startOffset+offsetVect[i]));
        const std::string& str=vect[i];
        ofs.write(str.data(),str.length());
        }
    assert(((FileOffset)ofs.tellp() )==(startOffset+offsetVect[nbStrings]));    
    }
    
    
class AdditionalInfosUnpacker: public AdditionalInfosUnpackerBase, 
                                public AddInfosVectorAsVectorImplementationInterface
    {
    //friend class AdditionalInfosPacker;
    
    protected:
    std::vector<std::string> vectAddInfos;
    
    public:
    bool inactive;
    
    AdditionalInfosUnpacker(const char *filename,bool useGroups=OtherInfosContainer::useGroups):
                                    AdditionalInfosUnpackerBase(useGroups),inactive(false)
        {
        std::ifstream ifs(filename,std::ifstream::binary);
        if (!ifs.is_open())
            {
            std::cerr<<"Error opening "<<filename<<std::endl;
            }
        boost::archive::binary_iarchive ia(ifs);
        oiDecoderSource.readFromBoostArchive(ia);
        oiDecoderTarget.readFromBoostArchive(ia);
        //ia>>sourceWordsIndexer;        
        //ia>>sourcePOSIndexer;
        //ia>>targetWordsIndexer;
        //ia>>targetPOSIndexer;
        ia>>vectAddInfos;
        }


    AdditionalInfosUnpacker():inactive(true)
        {
        }
    
    size_t size() const
        {
        return vectAddInfos.size();
        }
        
    void takeOwnershipOfVectDataAndIndexers(std::vector<std::string>& vectAddInfos,
                                            const OtherInfosDecoder& oiDecoderSource,
                                            const OtherInfosDecoder& oiDecoderTarget)
        {
        inactive=false;
        this->vectAddInfos.swap(vectAddInfos);
        this->oiDecoderSource=oiDecoderSource;
        this->oiDecoderTarget=oiDecoderTarget;
        }
    
        
    AdditionalInfos at(size_t i) const
        {
        AdditionalInfos result;
        unpack(vectAddInfos[i],result);
        return result;
        }   
 
    const std::string& getRawDataAsRef(size_t i) const
        {
        return vectAddInfos[i];
        }
        
    std::string getRawData(size_t i) const
        {
        return std::string( getRawDataAsRef(i) );
        }
        

        
    void save(const char *filename, bool useMassStorageVersion=false) const
        {
        size_t startPositionInFile,endPositionInFile;
        
        std::ofstream ofs(filename,std::ofstream::binary);
        boost::archive::binary_oarchive oa(ofs);
        
        startPositionInFile=ofs.tellp();
        oiDecoderSource.saveToBoostArchive(oa);
        endPositionInFile=ofs.tellp();
        std::cout<<"saved source indexers: "<<endPositionInFile-startPositionInFile<<"bytes"<<std::endl;
        
        startPositionInFile=ofs.tellp();
        oiDecoderTarget.saveToBoostArchive(oa);
        endPositionInFile=ofs.tellp();
        std::cout<<"saved target indexers: "<<endPositionInFile-startPositionInFile<<"bytes"<<std::endl;
        
        startPositionInFile=ofs.tellp();
        
        if (!useMassStorageVersion)
            {
            oa<<vectAddInfos;
            }
        else
            {
            saveVectorStringForMassStorageAccess(vectAddInfos,oa,ofs);
            }
        
        endPositionInFile=ofs.tellp();
        std::cout<<"saved AddInfos: "<<endPositionInFile-startPositionInFile<<"bytes"<<std::endl;
        }      
    };

    

class AdditionalInfosUnpackerWithF: public AdditionalInfosUnpacker
    {
    public:
    AdditionalInfosUnpackerWithF(const char *filename,bool useGroups=OtherInfosContainer::useGroups):
                            AdditionalInfosUnpacker(filename,useGroups){}
 
    AdditionalInfosUnpackerWithF()
        {
        }
 
    void unpackMoreThings(PrefixDecodeVectors& decoder,AdditionalInfos& ai) const
        {
        
        decoder.unpackVectorString(ai.source_phrase_f_list);
        decoder.unpackVectorString(ai.source_word_f_list);
        decoder.unpackVectorString(ai.target_phrase_f_list);
        decoder.unpackVectorString(ai.target_word_f_list);
        }
 
    };
    
    
class AdditionalInfosPacker
    {
    protected:
    
    OtherInfosEncoder oieSource;
    OtherInfosEncoder oieTarget;
    
    std::vector<std::string> vectAddInfos;
    public:
    virtual ~AdditionalInfosPacker(){};
    
    AdditionalInfosPacker(bool useGroups=OtherInfosContainer::useGroups):
                    oieSource(useGroups),oieTarget(useGroups){};
    
    AdditionalInfosPacker(  const OtherInfosEncoder& oieSource,
                            const OtherInfosEncoder& oieTarget
                            ){
                            this->oieSource=oieSource;
                            this->oieTarget=oieTarget;
                            };
    
    AdditionalInfosPacker getNewEmptyOptimalPacker() const
        {
        return AdditionalInfosPacker(oieSource.getOptimalEncoder(),oieTarget.getOptimalEncoder());
        }
        
    
    boost::shared_ptr<AddInfosVectorAsVectorImplementationInterface> getEmptyAddInfoVectorUnpacker() const
        {
        return boost::make_shared< AdditionalInfosUnpacker>();
        }
        
    void optimizeAndSave(const char *filename)
        {
        std::cout<<"optimizing"<<std::endl;
        AdditionalInfosPacker optimalPacker=getNewEmptyOptimalPacker();
        
        boost::shared_ptr<AddInfosVectorAsVectorImplementationInterface> unpacker=
                                                            getEmptyAddInfoVectorUnpacker();
        unpacker->takeOwnershipOfVectDataAndIndexers(vectAddInfos,oieSource,oieTarget);
        for (size_t i=0; i<unpacker->size(); ++i)
            {
            optimalPacker.packAdditionalInfos( unpacker->at(i) );
            }
        std::cout<<"end optimizing"<<std::endl;
        optimalPacker.save(filename);
        
        }
        
    void packAdditionalInfos(const AnalyzedIData& aidSource,const AnalyzedIData& aidTarget)
        {
        std::string result;
        PrefixEncodeVectors encoder(result);
        
        encoder.packPosListUC(aidSource.depend);
        oieSource.encode(aidSource.others, encoder);
        oieTarget.encode(aidTarget.others, encoder);
        
        packMoreThings(aidSource,aidTarget,encoder);
        
        vectAddInfos.push_back(result);
        }

    virtual void packMoreThings(const AnalyzedIData& aidSource,const AnalyzedIData& aidTarget,
                                        PrefixEncodeVectors& encoder)
        {}

    void packAdditionalInfos(const AdditionalInfos& ai)
        {
        std::string result;
        PrefixEncodeVectors encoder(result);
        
        encoder.packPosListUC(ai.sourceDepend);
        oieSource.encode(ai.sourceOtherInfos,encoder);
        oieTarget.encode(ai.targetOtherInfos,encoder);
        
        packMoreThings(ai,encoder);
        
        vectAddInfos.push_back(result);
        }

    virtual void packMoreThings(const AdditionalInfos& ai,
                                        PrefixEncodeVectors& encoder)
        {}
        
    void save(const char *filename, bool useMassStorageVersion=false) const
        {
        size_t startPositionInFile,endPositionInFile;
        
        std::ofstream ofs(filename,std::ofstream::binary);
        boost::archive::binary_oarchive oa(ofs);
        
        startPositionInFile=ofs.tellp();
        oieSource.saveToBoostArchive(oa);
        endPositionInFile=ofs.tellp();
        std::cout<<"saved source indexers: "<<endPositionInFile-startPositionInFile<<"bytes"<<std::endl;
        
        startPositionInFile=ofs.tellp();
        oieTarget.saveToBoostArchive(oa);
        endPositionInFile=ofs.tellp();
        std::cout<<"saved target indexers: "<<endPositionInFile-startPositionInFile<<"bytes"<<std::endl;
        
        startPositionInFile=ofs.tellp();
        
        if (!useMassStorageVersion)
            {
            oa<<vectAddInfos;
            }
        else
            {
            saveVectorStringForMassStorageAccess(vectAddInfos,oa,ofs);
            }
        
        endPositionInFile=ofs.tellp();
        std::cout<<"saved AddInfos: "<<endPositionInFile-startPositionInFile<<"bytes"<<std::endl;
        }
             
    };
   
class  AdditionalInfosPackerWithF: public AdditionalInfosPacker
    {
    public:
    
    AdditionalInfosPackerWithF(bool useGroups=OtherInfosContainer::useGroups):
                AdditionalInfosPacker(useGroups){};
    
    AdditionalInfosPackerWithF(  const OtherInfosEncoder& oieSource,
                            const OtherInfosEncoder& oieTarget
                            ){
                            this->oieSource=oieSource;
                            this->oieTarget=oieTarget;
                            };

    void optimizeAndSave(const char *filename)
        {
        std::cout<<"optimizing"<<std::endl;
        AdditionalInfosPackerWithF optimalPacker=getNewEmptyOptimalPacker();
        
        boost::shared_ptr<AddInfosVectorAsVectorImplementationInterface> unpacker=
                                                            getEmptyAddInfoVectorUnpacker();
        unpacker->takeOwnershipOfVectDataAndIndexers(vectAddInfos,oieSource,oieTarget);
        for (size_t i=0; i<unpacker->size(); ++i)
            {
            optimalPacker.packAdditionalInfos( unpacker->at(i) );
            }
        std::cout<<"end optimizing"<<std::endl;
        optimalPacker.save(filename);
        
        }
                   
    AdditionalInfosPackerWithF getNewEmptyOptimalPacker() const
        {
        return AdditionalInfosPackerWithF(oieSource.getOptimalEncoder(),oieTarget.getOptimalEncoder());
        }
        
    boost::shared_ptr<AddInfosVectorAsVectorImplementationInterface> getEmptyAddInfoVectorUnpacker()
        {
        return boost::make_shared<AdditionalInfosUnpackerWithF>();
        }    

    // virtual void packMoreThings(const AnalyzedIData& aidSource,const AnalyzedIData& aidTarget,
                                        // PrefixEncodeVectors& encoder)
        // {
        // encoder.packVectorString(aidSource.phrase_f_list);
        // encoder.packVectorString(aidSource.word_f_list);
        // encoder.packVectorString(aidTarget.phrase_f_list);
        // encoder.packVectorString(aidTarget.word_f_list);
        // }

    // virtual void packMoreThings(const AdditionalInfos& ai,
                                        // PrefixEncodeVectors& encoder)
        // {
        // encoder.packVectorString(ai.source_phrase_f_list);
        // encoder.packVectorString(ai.source_word_f_list);
        // encoder.packVectorString(ai.target_phrase_f_list);
        // encoder.packVectorString(ai.target_word_f_list);
        // }
        
    };

#endif



