/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*! \file
    \ingroup treelet_computing_group
    \brief Structure storing the result of the computation (ie. the occurrence list) of each treelets.
*/

#include "binaryMerge.hpp"
#include "binaryMerge.cpp"
#include <vector>
#include <map>
#include <boost/smart_ptr.hpp>

#include <boost/smart_ptr/make_shared.hpp>

#include "TypeDefinitions.hpp"
#include "Treelets.hpp"
#include "MergeConstraintsComputation.hpp"

#ifndef _HPP_MATCHING_CONTAINERS_
#define _HPP_MATCHING_CONTAINERS_              


template<typename OccListType1,typename OccListType2>
OccList * _mergeOcc(const OccListType1& occ1, const OccListType2& occ2, const MergeConstraints& constraints)
    {
    return intersectAndMerge(occ1, occ2,   
                        constraints.equalList,constraints.diffList,constraints.orderRes, constraints.preserveOrder );
    }

class NormalMatchingsListContainer;
class CompactPairMatchingsListContainer;

class MatchingsListContainerVisitor
    {
    public:
    virtual ~MatchingsListContainerVisitor(){};
    virtual void visit(NormalMatchingsListContainer*) const=0;
    virtual void visit(CompactPairMatchingsListContainer *) const=0;
    };

typedef boost::shared_ptr<NormalMatchingsListContainer> NormalMatchingsListContainerSmartPtr;

class GenericMatchingsListContainer
    {
    public:
    virtual NormalMatchingsListContainerSmartPtr mergeWith(const GenericMatchingsListContainer& other, const MergeConstraints& constraints) const=0;
    virtual NormalMatchingsListContainerSmartPtr beMergedWith(const NormalMatchingsListContainer& other, const MergeConstraints& constraints) const=0;
    virtual NormalMatchingsListContainerSmartPtr beMergedWith(const CompactPairMatchingsListContainer& other, const MergeConstraints& constraints) const=0;
    virtual size_t getNbSentOcc() const=0;
    virtual void acceptVisitor(MatchingsListContainerVisitor *)=0;
    virtual void printToStream(std::ostream &stream) const=0;
    
    virtual TreeIdType getNumTreeAtIndex(size_t i) const=0;
    virtual size_t getNbOfMappingsAtIndex(size_t i) const=0;
    virtual std::vector<PosType> getMappingWithoutParentAtIndices(size_t i, size_t j) const=0;
    
    virtual std::vector<PosType> getMappingAtIndices(size_t i, size_t j) const=0;
    
    virtual bool isDominatedBy(const GenericMatchingsListContainer& other, const std::vector<SlotPairType>& equalList) const=0;
    virtual bool isDominating(const NormalMatchingsListContainer& other, const std::vector<SlotPairType>& equalList) const=0;
    virtual bool isDominating(const CompactPairMatchingsListContainer& other, const std::vector<SlotPairType>& equalList) const=0;
    
    virtual size_t getTypicalSizeOfMapp() const=0;
    
    virtual ~GenericMatchingsListContainer()=0;
    //virtual void printToStream(std::ostream &stream) const =0;
    };
//template<typename ReturnType>    
GenericMatchingsListContainer::~GenericMatchingsListContainer(){}


std::ostream &operator<< (std::ostream &stream,const GenericMatchingsListContainer &g)
  {
  g.printToStream(stream);
  return stream;
  }
  
std::ostream &operator<< (std::ostream &stream,const boost::shared_ptr<GenericMatchingsListContainer> &pG)
  {
  stream<<"S@"<<pG.get()<<"/"<<pG.use_count()<<":"<<*pG;
  return stream;
  }

typedef GenericMatchingsListContainer MatchingsContainer;

class NormalMatchingsListContainer:  public MatchingsContainer//<boost::shared_ptr<NormalMatchingsListContainer> >
    {
        
   public:
    boost::shared_ptr<OccList> pV;
    
    NormalMatchingsListContainer():pV(new OccList)
        {
        //std::cout<<"Creating NormalMatchingsListContainer @"<<this<<" -> "<<pV<<std::endl;
        }
    NormalMatchingsListContainer(boost::shared_ptr<OccList> pV):pV(pV)
        {
        //std::cout<<"Creating NormalMatchingsListContainer @"<<this<<" -> "<<pV<<std::endl;
        }
    NormalMatchingsListContainer(OccList* raw_pV):pV(raw_pV)
        {
        //std::cout<<"Creating NormalMatchingsListContainer @"<<this<<" -> "<<pV<<std::endl;
        }
    
    ~NormalMatchingsListContainer()
        {
        //std::cout<<"Destructing NormalMatchingsListContainer@"<<this<<" -> "<<pV<< " : "<<*this<<std::endl; 
        }

    boost::shared_ptr<NormalMatchingsListContainer> mergeWith(const MatchingsContainer& other, 
                                                                        const MergeConstraints& constraints) const
        {
        return other.beMergedWith(*this,constraints);
        }

    NormalMatchingsListContainerSmartPtr beMergedWith(const NormalMatchingsListContainer& other, const MergeConstraints& constraints) const;
    NormalMatchingsListContainerSmartPtr beMergedWith(const CompactPairMatchingsListContainer& other, const MergeConstraints& constraints) const;
    
    
    bool isDominatedBy(const GenericMatchingsListContainer& other, const std::vector<SlotPairType>& equalList) const
        {
        return other.isDominating(*this,equalList);
        }
        
    bool isDominating(const NormalMatchingsListContainer& other, const std::vector<SlotPairType>& equalList) const;
    bool isDominating(const CompactPairMatchingsListContainer& other, const std::vector<SlotPairType>& equalList) const;
    
    
    size_t getNbSentOcc() const
        {
        return pV->size();
        }

    void acceptVisitor(MatchingsListContainerVisitor* v)
        {
        v->visit(this);
        }

        
    void printToStream(std::ostream &stream) const
        {
        stream<<"NM@"<<pV.get()<<"/"<<pV.use_count()<<"("<<*pV<<")";
        }
        
    TreeIdType getNumTreeAtIndex(size_t i) const
        {
        return pV->at(i).numTree;
        }
    size_t getNbOfMappingsAtIndex(size_t i) const
        {
        return pV->at(i).mappingList.size();
        }
    
    std::vector<PosType> getMappingWithoutParentAtIndices(size_t i, size_t j) const
        {
        const std::vector<PosType>& mapp=pV->at(i).mappingList.at(j);
        size_t sizeMapping=mapp.size();
        std::vector<PosType> result(sizeMapping-1); //remove parent
        std::copy(mapp.begin(),mapp.end()-1,result.begin());
        return result;
        }

    std::vector<PosType> getMappingAtIndices(size_t i, size_t j) const
        {
        const std::vector<PosType>& mapp=pV->at(i).mappingList.at(j);
        size_t sizeMapping=mapp.size();
        std::vector<PosType> result(sizeMapping); //remove parent
        std::copy(mapp.begin(),mapp.end(),result.begin());
        return result;
        }
      
    size_t getTypicalSizeOfMapp() const
        {
        assert(pV->size()>0);
        return pV->at(0).mappingList.at(0).size();
        }
      
      
    };
    
//typedef boost::shared_ptr<NormalMatchingsListContainer> NormalMatchingsListContainerSmartPtr;

typedef boost::shared_ptr< MatchingsContainer > MatchingsContainerSmartPtr;

class CompactPairMatchingsListContainer:  public MatchingsContainer
    {
    
    public:
    
    const OccListCompactPair* pV;
    CompactPairMatchingsListContainer(const OccListCompactPair* pV):pV(pV)
        {
        //std::cout<<"Creating CompactPairMatchingsListContainer @"<<this<<" -> "<<pV<<" : "<<*pV<<std::endl;
        }


    ~CompactPairMatchingsListContainer()
        {
        //std::cout<<"Destructing CompactPairMatchingsListContainer @"<<this<<" -> "<<pV<<" :"<<*this<<std::endl;
        }

    NormalMatchingsListContainerSmartPtr mergeWith(const MatchingsContainer& other, 
                                                                        const MergeConstraints& constraints) const
        {
        return other.beMergedWith(*this,constraints);
        }

    NormalMatchingsListContainerSmartPtr beMergedWith(const NormalMatchingsListContainer& other, const MergeConstraints& constraints) const
        {
        OccList* pRes=_mergeOcc<OccList,OccListCompactPair>(*other.pV,*pV,constraints);
        return boost::make_shared<NormalMatchingsListContainer>(pRes);        
        }
    NormalMatchingsListContainerSmartPtr beMergedWith(const CompactPairMatchingsListContainer& other, const MergeConstraints& constraints) const
        {
        OccList* pRes=_mergeOcc<OccListCompactPair,OccListCompactPair>(*other.pV,*pV,constraints);
        return boost::make_shared<NormalMatchingsListContainer>(pRes);
        }

    bool isDominatedBy(const GenericMatchingsListContainer& other, const std::vector<SlotPairType>& equalList) const
        {
        return other.isDominating(*this,equalList);
        }
        
    bool isDominating(const NormalMatchingsListContainer& other, const std::vector<SlotPairType>& equalList) const
        {
        return isDominated(*other.pV,*pV,equalList);
        }
        
    bool isDominating(const CompactPairMatchingsListContainer& other, const std::vector<SlotPairType>& equalList) const
        {
        return isDominated(*other.pV,*pV,equalList);
        }
        
    size_t getNbSentOcc() const
        {
        return pV->size();
        }

    void acceptVisitor(MatchingsListContainerVisitor* v)
        {
        v->visit(this);
        }     

    void printToStream(std::ostream &stream) const
        {
        stream<<"CP@"<<pV<<"("<<*pV<<")";
        }
       
    TreeIdType getNumTreeAtIndex(size_t i) const
        {
        return pV->at(i).numTree;
        }
    size_t getNbOfMappingsAtIndex(size_t i) const
        {
        return pV->at(i).mappingList.size();
        }
    
    std::vector<PosType> getMappingWithoutParentAtIndices(size_t i, size_t j) const
        {
        assert(j<getNbOfMappingsAtIndex(i));
        PosType pos=pV->at(i).mappingList[j][0];
        std::vector<PosType> result(1,pos);
        return result;
        }

    std::vector<PosType> getMappingAtIndices(size_t i, size_t j) const
        {
        assert(j<getNbOfMappingsAtIndex(i));
        PosType pos=pV->at(i).mappingList[j][0];
        PosType parent=pV->at(i).mappingList[j][1];
        std::vector<PosType> result(2);
        result[0]=pos;
        result[1]=parent;
        return result;
        }
       
    size_t getTypicalSizeOfMapp() const
        {
        return 2;
        }
       
    };   

    NormalMatchingsListContainerSmartPtr NormalMatchingsListContainer::beMergedWith(const NormalMatchingsListContainer& other, const MergeConstraints& constraints) const
        {
        OccList* pRes=_mergeOcc<OccList,OccList>(*other.pV,*pV,constraints);
        return boost::make_shared<NormalMatchingsListContainer>(pRes);        
        }
    NormalMatchingsListContainerSmartPtr NormalMatchingsListContainer::beMergedWith(const CompactPairMatchingsListContainer& other, const MergeConstraints& constraints) const
        {
        OccList* pRes=_mergeOcc<OccListCompactPair,OccList>(*other.pV,*pV,constraints);
        return boost::make_shared<NormalMatchingsListContainer>(pRes);
        }

bool NormalMatchingsListContainer::isDominating(const NormalMatchingsListContainer& other, const std::vector<SlotPairType>& equalList) const
    {
    return isDominated(*other.pV,*pV,equalList);
    }
    
bool NormalMatchingsListContainer::isDominating(const CompactPairMatchingsListContainer& other, const std::vector<SlotPairType>& equalList) const
    {
    return isDominated(*other.pV,*pV,equalList);
    }

class MatchingNotFoundException: public std::exception
{
  virtual const char* what() const throw()
  {
    return "MatchingNotFoundException";
  }
};


std::ostream &operator<< (std::ostream &stream,const NormalMatchingsListContainer &g)
  {
  g.printToStream(stream);
  return stream;
  }
  
std::ostream &operator<< (std::ostream &stream,const boost::shared_ptr<NormalMatchingsListContainer> &pG)
  {
  stream<<"S@"<<pG.get()<<"/"<<pG.use_count()<<":"<<*pG;
  return stream;
  }

std::ostream &operator<< (std::ostream &stream,const CompactPairMatchingsListContainer &g)
  {
  g.printToStream(stream);
  return stream;
  }
  
std::ostream &operator<< (std::ostream &stream,const boost::shared_ptr<CompactPairMatchingsListContainer> &pG)
  {
  stream<<"S@"<<pG.get()<<"/"<<pG.use_count()<<":"<<*pG;
  return stream;
  }

template<typename TreeletType>
class MatchingsContainerStoreGeneric: public  boost::noncopyable
    {
    std::map<TreeletType, MatchingsContainerSmartPtr> computedMatchings;
    public:
        void addResult(const TreeletType& key,MatchingsContainerSmartPtr pV)
            {
            //computedMatchings[key]=pV;
            //std::cout<<"adding result "<<key<<std::endl;
            //std::cout<<pV<<std::endl;
            assert(computedMatchings.count(key)==0);
            computedMatchings.insert(std::make_pair(key,pV) );
            }
            
    MatchingsContainerSmartPtr getResult(const TreeletType& treelet) const
        {
        typename std::map<TreeletType, MatchingsContainerSmartPtr>::const_iterator it=computedMatchings.find(treelet);
        if (it==computedMatchings.end())
            {
            throw MatchingNotFoundException();
            }
        else
            {
            return it->second;
            }
        }
        
    void printStoredSmartPtr(const TreeletType& treelet) const
        {
        std::cout<<computedMatchings.find(treelet)->second<<std::endl;
        }
        
    const std::map<TreeletType, MatchingsContainerSmartPtr> & getUnderlyingMap() const
        {
        return computedMatchings;
        }
    
    std::map<TreeletType, MatchingsContainerSmartPtr> & getModifyableUnderlyingMap()
        {
        return computedMatchings;
        }
    
    size_t size()
        {
        return computedMatchings.size();
        }
        
    std::pair<size_t,size_t> getStats(int maxNbOfsamplesPerTreelets) const
        {
        size_t nb_of_non_empty_treelets=0;
        size_t total_nb_of_occ=0;
        typedef std::pair<TreeletType, MatchingsContainerSmartPtr> TreeletMatchingPairs;
        BOOST_FOREACH(const TreeletMatchingPairs& tmp,computedMatchings)
            {
            size_t nb_of_occ=tmp.second->getNbSentOcc();
            if (nb_of_occ>0)
                {
                nb_of_non_empty_treelets++;
                if (maxNbOfsamplesPerTreelets>0 && (int)nb_of_occ>maxNbOfsamplesPerTreelets)
                    {
                    total_nb_of_occ+=maxNbOfsamplesPerTreelets;
                    }
                else
                    {
                    total_nb_of_occ+=nb_of_occ;
                    }
                }
            }
        return std::make_pair(nb_of_non_empty_treelets,total_nb_of_occ);
        }
/*            
        void printAll()
            {
            BOOST_FOREACH(const KeyValType& keyVal, mapToV)
                {
                std::cout<<keyVal.first<< " : "<< *(keyVal.second) <<std::endl;
                }
            
            }
*/

    };

// typedef MatchingsContainerStoreGeneric<Treelet> MatchingsContainerStore;

/*
class MatchingsContainerStore: public boost::noncopyable //non copyable, otherwise catastroph could occur when elements are reallocated, since we later access them by raw pointer
    {
    
    class BePushedToHolder: public MatchingsListContainerVisitor
        {
        MatchingsContainerStore *pHolder;
        const Treelet& key;
        public:
        BePushedToHolder(MatchingsContainerStore *pHolder,const Treelet& key):pHolder(pHolder),key(key){};
        void visit(NormalMatchingsListContainer * pVD) const
            {
            pHolder->push(key, *pVD);
            }
        void visit(CompactPairMatchingsListContainer * pVI) const
            {
            pHolder->push(key, *pVI);
            }        
        };
    
    std::map<Treelet, MatchingsContainer *> computedMatchings;
    // Possibly, a slightly STL-implementation dependent assumption is made here:
    //    that no object put in a std::list will ever be reallocated ...
    std::list<CompactPairMatchingsListContainer> _CompactPairMatchingsListContainerKeptInMemory;
    std::list<NormalMatchingsListContainer> _NormalMatchingsListContainerKeptInMemory;
    public:
        void addResult(const Treelet& key,MatchingsContainer *pV)
            {
            BePushedToHolder visitor(this,key);
            pV->acceptVisitor( (MatchingsListContainerVisitor* ) &visitor );
            }
            
        void push(const Treelet& key,const NormalMatchingsListContainer& vd)
            {
            _NormalMatchingsListContainerKeptInMemory.push_back(vd);
            computedMatchings[key]= & (_NormalMatchingsListContainerKeptInMemory.back() );
            }
            
        void push(const Treelet& key, const CompactPairMatchingsListContainer& vi)
            {
            _CompactPairMatchingsListContainerKeptInMemory.push_back(vi);
            computedMatchings[key]= & (_CompactPairMatchingsListContainerKeptInMemory.back() );
            }
            
        MatchingsContainer * getResult(const Treelet& treelet) const
            {
            std::map<Treelet, MatchingsContainer *>::const_iterator it=computedMatchings.find(treelet);
            if (it==computedMatchings.end())
                {
                throw MatchingNotFoundException();
                }
            else
                {
                return it->second;
                }
            }
           
        // void printAll()
            // {
            // BOOST_FOREACH(const KeyValType& keyVal, mapToV)
                // {
                // std::cout<<keyVal.first<< " : "<< *(keyVal.second) <<std::endl;
                // }
            
            // }


    };
*/  
    
    
    
    
    
 
    
// int dummyFunctionSoThatTemplatesAreInstantiated(){
    // OccList * pOcc=NULL;
    // MatchingsContainerSmartPtr c1=boost::make_shared<NormalMatchingsListContainer>(pOcc);
    // MatchingsContainerSmartPtr c2=boost::make_shared<NormalMatchingsListContainer>(pOcc);
    // OccListCompactPair *pOcc2;
    // MatchingsContainerSmartPtr c3=boost::make_shared<CompactPairMatchingsListContainer>(pOcc2);
    // MatchingsContainerSmartPtr c4=boost::make_shared<CompactPairMatchingsListContainer>(pOcc2);
    // MergeConstraints constraints(true);
    // c1->mergeWith(*c2,constraints);
    // c1->mergeWith(*c3,constraints);
    // c4->mergeWith(*c2,constraints);
    // c3->mergeWith(*c4,constraints);
    
    // MatchingsContainerStore crm;
    // //Treelet ti(NULL);
    // //crm.addResult(ti,c1);
    // //crm.addResult(ti,c3);
    // //MatchingsContainerSmartPtr  g=crm.getResult(ti);
    // return 0;}    


#endif


