/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*! \file
    \ingroup treelet_computing_group
    \brief Functions for computing the merge constraints.
*/

#include <vector>
#include <boost/foreach.hpp>

#include "TypeDefinitions.hpp"
#include "SeqOfPosRepresentation.hpp"

#ifndef _HPP_MERGE_CONSTRAINTS_
#define _HPP_MERGE_CONSTRAINTS_

#define __DEBUG__HERE__

struct PosNumSlotStruct
    {
    PosType pos;
    char num;
    SlotType slot;
    PosNumSlotStruct(PosType pos, char num, SlotType slot):pos(pos),num(num), slot(slot) {};
    bool operator<(const PosNumSlotStruct &other) const
        {
        return pos<other.pos;
        }
        
    };

class MergeConstraints
    {
    typedef TreeletBase::SeqOfPosRepresentation SeqOfPosRepresentation;
    
    public:
    std::vector<SlotPairType> equalList;
    std::vector<SlotListPairType> diffList;
    std::vector<MapSlotPair> orderRes;
    bool preserveOrder;
    //vector<PosType> fullSeqOfPos;
    
    
    MergeConstraints(bool preserveOrder):preserveOrder(preserveOrder){};
    
    void computeAll(const SeqOfPosRepresentation &sop1,const SeqOfPosRepresentation &sop2)
        {
        computeEqualList(sop1,sop2);
        if (!preserveOrder)
            {
            computeDiffList(sop1,sop2);
            }
        computeOrderList(sop1,sop2);
        }
    
    void computeEqualList(const SeqOfPosRepresentation &sop1,const SeqOfPosRepresentation &sop2)
        {
        assert(equalList.size()==0);
        //const std::map<PosType,SlotType> &posToSlot1=sop1.getPosToSlotDict();
        //const std::map<PosType,SlotType> &posToSlot2=sop2.getPosToSlotDict();

        BOOST_FOREACH(PosType pos, sop1.getListOfCommonPos(sop2))
            {
            equalList.push_back( SlotPairType(  sop1.getSlotOfPos(pos), sop2.getSlotOfPos(pos) ) );
            }
            
        PosType parent1=sop1.getParent();
        PosType parent2=sop2.getParent();
        SlotType slotParent1=sop1.getSlotOfParent();
        SlotType slotParent2=sop2.getSlotOfParent();
        
        if ( parent1==parent2)
            {
            assert( !sop1.contains(parent2) && !sop2.contains(parent1));
            equalList.push_back( SlotPairType(  slotParent1, slotParent2 ) );
            }
        else if( sop2.contains(parent1) )
            {
            assert( !sop1.contains(parent2) );
            equalList.push_back( SlotPairType(  slotParent1, sop2.getSlotOfPos(parent1) ) );
            }
        else if ( sop1.contains(parent2) )
            {
            assert( !sop2.contains(parent1) );
            equalList.push_back( SlotPairType(  sop1.getSlotOfPos(parent2), slotParent2 ) );
            }
        else
            {
            assert(false);
            }
        }
        
    void computeDiffList(const SeqOfPosRepresentation &sop1,const SeqOfPosRepresentation &sop2)
        {
        assert(diffList.size()==0);
        diffList.push_back(SlotListPairType());
        
        //const std::map<PosType,SlotType> &posToSlot1=sop1.getPosToSlotDict();
        BOOST_FOREACH(PosType pos, sop1.getPosListDifference(sop2))
            {
            diffList.back().first.push_back( sop1.getSlotOfPos(pos) );
            }

        //const std::map<PosType,SlotType> &posToSlot2=sop2.getPosToSlotDict();
        BOOST_FOREACH(PosType pos, sop2.getPosListDifference(sop1))
            {
            diffList.back().second.push_back(sop2.getSlotOfPos(pos));
            }

        
        /*
        diffList=std::pair<>(diffSide1,diffSide2)
        for pos in seqOfPos1[:-1]:
            if pos not in posToSlot2:
                diffSide1.append(posToSlot1[pos])

        diffSide2=[]
        for pos in seqOfPos2[:-1]:
            if pos not in posToSlot1:
                diffSide2.append(posToSlot2[pos])
        diffList=[(diffSide1,diffSide2)]         
        */
        
        }
        
    void computeOrderList(const SeqOfPosRepresentation &sop1,const SeqOfPosRepresentation &sop2)
        {
        assert(orderRes.size()==0);

        std::vector< PosNumSlotStruct  > preOrderRes; 
        
        SlotType slot=0;
        BOOST_FOREACH (  PosType pos ,sop1.getIterableSeqOfPos() ) 
            {
            preOrderRes.push_back(PosNumSlotStruct(pos,0,slot) );
            ++slot;
            }
            
        slot=0;
        BOOST_FOREACH (  PosType pos ,sop2.getIterableSeqOfPos() ) 
            {
            if (!sop1.contains(pos) )
                {
                preOrderRes.push_back(PosNumSlotStruct(pos,1,slot) );
                }
            ++slot;
            }
                        
        /*       
        for slot,pos in enumerate(seqOfPos1[:-1]):
            preOrderRes.append ( (pos,0,slot) )
        for slot,pos in enumerate(seqOfPos2[:-1]):
            if pos not in posToSlot1 or pos==father1:
                preOrderRes.append( (pos,1,slot) )
        */
        
        std::sort(preOrderRes.begin(),preOrderRes.end());
        //preOrderRes.sort()
        //newSeqOfPos=[pos for pos,numMap,slot in preOrderRes]

        
        #ifdef __DEBUG__HERE__
        
        std::vector<PosType> newSeqOfPos;
        BOOST_FOREACH(PosNumSlotStruct &posNumSlot,preOrderRes)
            {
            newSeqOfPos.push_back(posNumSlot.pos);
            }
        SeqOfPosRepresentation sopMerged=sop1.merge(sop2);
        assert ( sopMerged.getSizeOfSeq()==newSeqOfPos.size()  && std::equal(newSeqOfPos.begin(),newSeqOfPos.end(),sopMerged.getIterableSeqOfPos().begin() ) );
        #endif

        BOOST_FOREACH(PosNumSlotStruct &posNumSlot,preOrderRes)
            {
            orderRes.push_back(MapSlotPair(posNumSlot.num,posNumSlot.slot) );
            }
            
        //orderRes=[(numMap,slot) for pos,numMap,slot in preOrderRes]

        PosType parent1=sop1.getParent();
        PosType parent2=sop2.getParent();
        
        
        if ( parent1==parent2 || sop2.contains(parent1) )
            {
            assert( !sop1.contains(parent2) );
            orderRes.push_back( MapSlotPair(1,sop2.getSlotOfParent()) );
            }
        else if ( sop1.contains(parent2) )
            {
            assert( !sop2.contains(parent1) );
            orderRes.push_back( MapSlotPair(0,sop1.getSlotOfParent()) );
            }
        else
            {
            assert(false);
            }
            
        /*
        if father1 in posToSlot2:
            assert father2 not in posToSlot1 or father1==father2
            newSeqOfPos.append(father2)
            orderRes.append( (1,posToSlot2[father2]) )
        elif father2 in posToSlot1:
            assert father1 not in posToSlot2
            newSeqOfPos.append(father1)
            orderRes.append( (0,posToSlot1[father1]) )        
        else:
            print seqOfPos1
            print seqOfPos2
            assert False    # at least one of the father must be in the opposite set

        if totalPos!=set(newSeqOfPos):
            print "*************"
            print seqOfPos1
            print seqOfPos2
            print newSeqOfPos
            assert False        
        */
        }
    
    };
    

std::ostream &operator<< (std::ostream &stream,const MergeConstraints &mc)
  {
 
  stream<<"E:"<<mc.equalList<<" D:"<<mc.diffList<<" O:"<<mc.orderRes<<" po:"<<mc.preserveOrder;
  return stream;
  }

    
//int main(){ return 0; }
      
#endif
            
