/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*! \file
    \ingroup treelet_computing_group
    \brief Class managing the precomputed treelets.
*/

#include "binaryMerge.hpp"

#include <vector>


#include "TypeDefinitions.hpp"
#include "Treelets.hpp"
#include "MatchingsContainers.hpp"
#ifndef _HPP_PRECOMPUTATION_MANAGER_
#define _HPP_PRECOMPUTATION_MANAGER_

template<typename TreeletType>
class PrecomputationManagerGeneric
    {
        typedef typename TreeletType::TreeletSimpleIndexer TreeletSimpleIndexer;

        const std::vector<OccListCompactPair> *pInvertedIndex;
        OccListCompactPair occListWildcard;
        bool wildcardComputed;
        std::vector<boost::shared_ptr<OccList> > occListWildcardOptim;
        // Warning: order of the following two objects important!
        // emptyOccList HAS to be initialized before emptyMatchingsContainer
        //OccList emptyOccList;
        MatchingsContainerSmartPtr emptyMatchingsContainer;
        bool make_cache_of_wildcard_pair_;

    public:


        PrecomputationManagerGeneric(vector<OccListCompactPair> *pInvertedIndex,//OccListCompactPair *pOccListWildcard=NULL
        bool addWildcardInfo=false, bool make_cache_of_wildcard_pair=false
        ):
                    pInvertedIndex(pInvertedIndex),
                    wildcardComputed(addWildcardInfo),
                    emptyMatchingsContainer(boost::make_shared<NormalMatchingsListContainer>(boost::make_shared<OccList>())),
                    make_cache_of_wildcard_pair_(make_cache_of_wildcard_pair)
                        {
                        if (addWildcardInfo)
                            {
                            mergeListOfOccurenceList(*pInvertedIndex,occListWildcard);
                            if (make_cache_of_wildcard_pair_) {
                                mergeOptimListOfOccurenceList(*pInvertedIndex,occListWildcard,occListWildcardOptim);
                            }
                            }
                        };
        PrecomputationManagerGeneric():pInvertedIndex(NULL),wildcardComputed(false),
                    //emptyMatchingsContainer(boost::make_shared<CompactPairMatchingsListContainer>(&emptyOccList))
                    emptyMatchingsContainer(boost::make_shared<NormalMatchingsListContainer>(boost::make_shared<OccList>()))
                        {};

        MatchingsContainerSmartPtr getEmptyMatchingsContainer() const
            {
            return emptyMatchingsContainer;
            }

        bool isPrecomputed(const TreeletType& treelet, const TreeletSimpleIndexer& indexer) const {
            bool isLeftChild;
            LabelType childLabel;
            bool result = treelet.isElementaryTreelet()
                //|| treelet.isWildcardElemPair(*(treeletIndexer.getUnderlyingIndexer()),
                //                              isLeftChild, childLabel);
                || (make_cache_of_wildcard_pair_ && treelet.isWildcardElemPair(indexer, isLeftChild, childLabel));
            //cout << "treelet " << treelet << "is precomputed: " << result << endl;
            return result;
            }

        const boost::shared_ptr<OccList> getWildcardElemPairOccurences(LabelType label, bool isChildOnLeft) const {
            LabelType index = 2 * label;
            if (!isChildOnLeft) {
                index++;
            }
            return occListWildcardOptim[index];
        }

        MatchingsContainerSmartPtr getOccurrences(const TreeletType& treelet, const TreeletSimpleIndexer& indexer) const
            {
                //assert(isPrecomputed(treelet, indexer));
            //cerr<<"treelet.isElementaryWildcard():"<<treelet.isElementaryWildcard()<<" "<<treelet<<endl;
            //cerr<<occListWildcard<<endl;
            bool isChildOnLeft = false;
            LabelType childLabel = 0; // TODO special value to initialize?
            if (treelet.isElementaryWildcard())
                {
                //cout << "elementary wildcard" << endl;
                if (!wildcardComputed)
                    {
                    throw std::logic_error("Wildcard occurrence list not computed");
                    }
                //cerr<<"will return wildcard occ"<<endl;
                return boost::make_shared<CompactPairMatchingsListContainer>(&occListWildcard);
                }
            else if (treelet.isElementaryTreelet()) {
                LabelType label=treelet.getRootLabel();
                if (label >= pInvertedIndex->size() )
                    {
                        return emptyMatchingsContainer;
                    }
                else
                    {
                        const OccListCompactPair *pOccListCompactPair=&((*pInvertedIndex)[label]);
                        return boost::make_shared<CompactPairMatchingsListContainer>(pOccListCompactPair);
                    }
            } else if (make_cache_of_wildcard_pair_ && treelet.isWildcardElemPair(indexer, isChildOnLeft, childLabel)) {
                //cout << "getOccurrences found wildcard-element pair, label=" << childLabel << "left=" << isChildOnLeft << endl;
                const boost::shared_ptr<OccList> occList = getWildcardElemPairOccurences(childLabel, isChildOnLeft);
                //cout << "getOccurrences treelet " << treelet << endl;
                if (occList->size() == 0) {
                    //cout << "getOccurrences return empty occurence list" << endl;
                    return emptyMatchingsContainer;
                }
                //const OccList *pOccList = &((*pOccListWildcardOptim)[label]);
                MatchingsContainerSmartPtr result = boost::make_shared<NormalMatchingsListContainer>(occList);
                //cout << "getOccurrences return non-empty " << *result << endl;
                return result; //boost::make_shared<NormalMatchingsListContainer>(occList);
            } else {
                assert(false);
                return emptyMatchingsContainer;
            }



            }
    };

//typedef PrecomputationManagerGeneric<Treelet> PrecomputationManager;
#endif

