/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*! \file
    \ingroup result_output_group
    \brief Main Result Post Processing Functionalities.
*/

#include<vector>
#include<map>
#include<set>
#include <boost/foreach.hpp>
#include <boost/smart_ptr.hpp>
#include <iostream>
#include <sstream>

#include "TypeDefinitions.hpp"
#include "MatchingsContainers.hpp"
#include "cTargetInfoList.cpp"
#include "StringConversions.hpp"
#include "IndexersAndEncoding.hpp"

#include <fstream>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/map.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>

#include "AdditionalInfos.hpp"

#ifndef _RESULTS_POST_PROCESSING_HPP
#define _RESULTS_POST_PROCESSING_HPP


// typedef std::pair< PosType,std::map<size_t,std::map<Treelet,MatchingsContainerSmartPtr> > > PosTypeMapPair;
// typedef std::pair< size_t,std::map<Treelet,MatchingsContainerSmartPtr> > SizeMapPair;

template<typename TreeletType>
class IndividualResultsProcessorGeneric
    {
    public:
    virtual ~IndividualResultsProcessorGeneric(){};
    virtual void processOneTreelet(TreeIdType numTree,
                            const PosList& mapp,
                            const ProjectedMappingInfos* pProjInfo,
                            size_t count_emitted,
                            MatchingsContainerSmartPtr occ,
                            size_t j,size_t numMapping)=0;
                            
    virtual void initializeForThisTreelet(const TreeletType &treelet){};
    virtual void changeTree(){};
    virtual void finalizeTreelet(){};
    virtual void finalizeAll(){};
    };

//typedef IndividualResultsProcessorGeneric<Treelet> IndividualResultsProcessor;
    
/************************************************************************/

struct PostProcessingParameters
    {
    size_t maxNumberOfSamplePerTreelet;
    bool filterOutDominated;
    bool keepElementaryWildcard;
    bool keepTopWilcardChain;
    bool handleOnlyContinuousProjections;
    bool enforceProjOfGroundIsContinuous;
    bool enforceProjOfGapIsContinuous;
    bool enforceProjOfAllIsContinuous;
    int maxTotalNbOfTmToSend;
    };
    
class PostProcessingInfosInterface
    {
    public:
    virtual ~PostProcessingInfosInterface(){};
    virtual const TargetList& getTargetInfos() const=0;
    virtual const std::vector<LabelType>& getMostFrequentLabelIds() const=0;
    virtual const std::vector<std::string>& getNumToTreeIdVector() const=0;
    virtual const AddInfosVectorInterface& getAdditionalInfos() const=0;
    
    };

template<typename AddInfosVectorImplementation>
class PostProcessingInfos: public PostProcessingInfosInterface,boost::noncopyable
    {
    protected:
    boost::shared_ptr<TargetList> pTgtList;
    std::vector<LabelType> mostFrequentLabelIds;
    std::vector<std::string> numToTreeIdVector;
    AddInfosVectorImplementation aDataVect;
    
    void loadMostFrequentLabels(const std::string& filename)
        {
        std::ifstream ifs(filename.c_str(),std::ifstream::binary);
        try
            {
            boost::archive::binary_iarchive ia(ifs);
            ia>>mostFrequentLabelIds;
            }
        catch (boost::archive::archive_exception e)
            {
                std::cerr<<"Error loading mostFrequentLabelIds from file "<<filename<<endl;
                std::cerr<<"what:"<<e.what()<<endl;
                throw e;
            }               
        }  
    
    void loadNumToTree(const std::string& filename)
        {
        std::ifstream ifs(filename.c_str(),std::ifstream::binary);
        try
            {
            boost::archive::binary_iarchive ia(ifs);
            ia>>numToTreeIdVector;
            //std::cout<<"numToTree 132069->"<<numToTreeIdVector.at(132069)<<std::endl;
            }
        catch (boost::archive::archive_exception e)
            {
                cerr<<"Error loading numToTreeIdVector from file "<<filename<<endl;
                cerr<<"what:"<<e.what()<<endl;
                throw e;
            }               
        
        }
    
    public:
    PostProcessingInfos(const std::string& tgtInfoFilename,
                        const std::string& mostFrequentLabelIdsFilename,
                        const std::string& numToTreeIdFilename,  
                        const std::string& addInfosFilename)
        :pTgtList( loadTargetList(tgtInfoFilename.c_str()) ),aDataVect(addInfosFilename.c_str())
        {
        
        loadNumToTree(numToTreeIdFilename);
        loadMostFrequentLabels(mostFrequentLabelIdsFilename);
        
        }
    
    const TargetList& getTargetInfos() const
        {
        return *pTgtList;
        }
    const AddInfosVectorInterface& getAdditionalInfos() const
        {
        return aDataVect;
        }
    const std::vector<std::string>& getNumToTreeIdVector() const
        {
        return numToTreeIdVector;
        }
    const std::vector<LabelType>& getMostFrequentLabelIds() const
        {
        return mostFrequentLabelIds;
        }
    
    };


/*****************************************************************************/    

/*
class PostProcessingHandlerInterface
    {
    public:
    virtual void handleResults(const MatchingsContainerStore& occStore)=0;
    };    
*/    

template<typename TreeletType> 
class PostProcessingHandlerGeneric: public ResultsHandlerInterfaceGeneric<TreeletType>
    {
    typedef ResultsHandlerInterfaceGeneric<TreeletType> ResultsHandlerInterface;
    typedef IndividualResultsProcessorGeneric<TreeletType> IndividualResultsProcessor;
    typedef MatchingsContainerStoreGeneric<TreeletType> MatchingsContainerStore;
    protected:

    const TargetList& tgtList;
    const std::map<LabelType, MatchingsContainerSmartPtr>& cachedOccWithContinuousProjection; //TODO: switch to unordered_map
    //size_t originalMaxNumberOfSamplePerTreelet;
    
    
    PostProcessingParameters ppParams;
    size_t maxNumberOfSamplePerTreelet;
    /*
    size_t maxNumberOfSamplePerTreelet;
    bool handleOnlyContinuousProjections;
    bool filterOutDominated; 
    bool keepElementaryWildcard;
    bool keepTopWilcardChain;
    */
    
    IndividualResultsProcessor &idp;

    static TargetList emptyTargetListForSimpleCases;
    public:

    PostProcessingHandlerGeneric(const TargetList& tgtList,
                          const std::map<LabelType, MatchingsContainerSmartPtr>& cachedOccWithContinuousProjection,
                          //size_t maxNumberOfSamplePerTreelet,
                          IndividualResultsProcessor &idp,
                          //bool handleOnlyContinuousProjections,
                          //bool filterOutDominated,bool keepElementaryWildcard,
                          //bool keepTopWilcardChain
                          const PostProcessingParameters& ppParams
                          ):
        tgtList( tgtList ), 
        cachedOccWithContinuousProjection(cachedOccWithContinuousProjection),
        
        //originalMaxNumberOfSamplePerTreelet(maxNumberOfSamplePerTreelet),
        
        
        //handleOnlyContinuousProjections(handleOnlyContinuousProjections),
        //filterOutDominated(filterOutDominated),keepElementaryWildcard(keepElementaryWildcard),keepTopWilcardChain(keepTopWilcardChain),
        ppParams(ppParams),
        maxNumberOfSamplePerTreelet(ppParams.maxNumberOfSamplePerTreelet),
        idp(idp)
                {}
             
                
    void handleResults(const MatchingsContainerStore& occStore)
        {
        
        maxNumberOfSamplePerTreelet=ppParams.maxNumberOfSamplePerTreelet;
        int maxTotalNbOfTmToSend=ppParams.maxTotalNbOfTmToSend;
        std::pair<size_t,size_t> nonEmptyTotalOccPair=occStore.getStats(maxNumberOfSamplePerTreelet);
        size_t nb_of_non_empty_treelets=nonEmptyTotalOccPair.first;
        size_t total_nb_of_occurences=nonEmptyTotalOccPair.second;
        assert(total_nb_of_occurences>=nb_of_non_empty_treelets);
        if (maxTotalNbOfTmToSend>0 && (int)total_nb_of_occurences>maxTotalNbOfTmToSend)
            {
            maxNumberOfSamplePerTreelet=maxTotalNbOfTmToSend/nb_of_non_empty_treelets;
	    if (maxNumberOfSamplePerTreelet == 0) {
            maxNumberOfSamplePerTreelet = 1;
	    }
            std::cerr<<"Warning: reduced outputed nb of sampling per treelets from "<<
                    ppParams.maxNumberOfSamplePerTreelet<<" to "<<maxNumberOfSamplePerTreelet<<
                    " due to total number of examples going over the allowed limit of "<<
                       maxTotalNbOfTmToSend<<" ("<<nb_of_non_empty_treelets<<
                       " non empty treelets)"<<std::endl;
            }
        processAll(occStore.getUnderlyingMap());
        }
                
    void processOneOccurrenceList(MatchingsContainerSmartPtr occ,const TreeletType& treelet,bool forceNormalTreatmentForTreeletsWithWildcard) const
        {
        size_t totalNbOfSent=occ->getNbSentOcc();
        size_t count_emitted=0;
        for (size_t j=0;j<totalNbOfSent;++j)
            {
            TreeIdType numTree=occ->getNumTreeAtIndex(j);
            size_t nbOfMappings=occ->getNbOfMappingsAtIndex(j);
            idp.changeTree();
            for (size_t numMapping=0; numMapping<nbOfMappings;++numMapping)
                {
                
                PosList mapp=occ->getMappingWithoutParentAtIndices(j,numMapping);

                if (ppParams.handleOnlyContinuousProjections)
                    {
                    
                    if (!forceNormalTreatmentForTreeletsWithWildcard &&
                            TreeletType::isUsingWildcard() && 
                                    treelet.hasGaps())
                        {
                        pair<PosList,PosList> posListPair=treelet.separateWildcardFromNWInMapp(mapp);
                        const PosList& mappGrounded=posListPair.first;
                        const PosList& mappGaps=posListPair.second;

                        ProjectedMappingInfos projInfoGaps=tgtList.projectMinMaxAndGiveMapping(numTree,mappGaps);
                        ProjectedMappingInfos projInfoGrounded=tgtList.projectMinMaxAndGiveMapping(numTree,mappGrounded);
                        ProjectedMappingInfos projInfoAll=tgtList.projectMinMaxAndGiveMapping(numTree,mapp);
                        if ( 
                            (!ppParams.enforceProjOfGroundIsContinuous ||projInfoGrounded.couldProject) && 
                            (!ppParams.enforceProjOfGapIsContinuous ||projInfoGaps.couldProject) &&
                            (!ppParams.enforceProjOfAllIsContinuous ||projInfoAll.couldProject)
                            )
                            {
                            if (projInfoGrounded.couldProject)
                                {
                                idp.processOneTreelet(numTree,mapp,&projInfoGrounded,count_emitted,occ,j,numMapping);
                                }
                            else
                                {
                                idp.processOneTreelet(numTree,mapp,NULL,count_emitted,occ,j,numMapping);
                                }
                            count_emitted+=1;
                            }
                        }
                    
                    else
                        {
                    
                        ProjectedMappingInfos projInfo=tgtList.projectMinMaxAndGiveMapping(numTree,mapp);
                        if (projInfo.couldProject)
                            {   
                            //std::string idTreeAsString=convertTreeId(numTree);
                            idp.processOneTreelet(numTree,mapp,&projInfo,count_emitted,occ,j,numMapping);
                            count_emitted+=1;
                            }
                        }
                    //else std::cout<<"rejected one occ of "<<numTree<<std::endl;
                    }
                else
                    {
                    //std::string idTreeAsString=convertTreeId(numTree);
                    idp.processOneTreelet(numTree,mapp,NULL,count_emitted,occ,j,numMapping);
                    count_emitted+=1;
                    }                      
                
                if (maxNumberOfSamplePerTreelet>0 && count_emitted>=maxNumberOfSamplePerTreelet)
                    {
                    return;
                    }
                    
                //output<<  (seqOfPos,numTree,pl,proj) ;
                }
            }
        }   


    void processOneTreelet(const TreeletType& treelet, MatchingsContainerSmartPtr occ) const
        {            
        idp.initializeForThisTreelet(treelet);
        
        if (ppParams.handleOnlyContinuousProjections && treelet.isElementaryTreelet() && !treelet.isElementaryWildcard() && cachedOccWithContinuousProjection.count(treelet.getRootLabel())>0)
            {
            occ=cachedOccWithContinuousProjection.find(treelet.getRootLabel())->second;
            }
        processOneOccurrenceList(occ,treelet,false);     
        }

    void processAll(const std::map<TreeletType,MatchingsContainerSmartPtr>& treeletToOccMap) const//std::ostream& output,bool addEndlineAtEachTreelet)
        {
        typedef std::pair<const TreeletType, const MatchingsContainerSmartPtr> ConstTreeletOccPair;
        BOOST_FOREACH(const ConstTreeletOccPair& treeletOccPair, treeletToOccMap )
            {
//            size_t nbOfTreelets=occStore.size();
//            for (size_t i=0;i<nbOfTreelets;++i)
//                {               
//                std::pair<Treelet, MatchingsContainerSmartPtr> treeletOccPair= occStore.get(i);
            
            const TreeletType& treelet=treeletOccPair.first;
            if (!ppParams.keepElementaryWildcard && treelet.isElementaryWildcard())
                {
                continue;
                }
                    
            if (!ppParams.keepTopWilcardChain &&treelet.hasTopWildcardChain())
                {
                continue;
                }
            
            if (treelet.getNbSentOcc()==0)
                {
                continue;
                }
            //cerr<<"filterOutDominated:"<<filterOutDominated<<" treelet.isDominated:"<<treelet.isDominated()<<endl;
            if (ppParams.filterOutDominated && treelet.isDominated())
                {
                continue;
                }
            //std::cerr<<"sending treelet "<<treelet<<std::endl;
            
            MatchingsContainerSmartPtr occ = treeletOccPair.second;
            processOneTreelet(treelet,occ);
            //printOneTreeletToStream(treelet, occ, output);
            idp.finalizeTreelet();
            }
            idp.finalizeAll();
            //occ.sendTxtConversionToStream(socket,treelet,nbExamplesPerTreelets);
            /*
            if (addEndlineAtEachTreelet)
                {
                output<<std::endl;
                }
            } 
            */           
        } 
        
    };
//typedef PostProcessingHandlerGeneric<Treelet> PostProcessingHandler;
    
/************************************************************************************/

template<typename TreeletType>
class CacheCreationProcessorGeneric: public IndividualResultsProcessorGeneric<TreeletType>
    {
    private:
    boost::shared_ptr<NormalMatchingsListContainer> cachedContinuousOcc;
    bool alreadyAddedMappingGroup;
    public:
    CacheCreationProcessorGeneric():alreadyAddedMappingGroup(false)
        {
        cachedContinuousOcc=boost::make_shared<NormalMatchingsListContainer>();
        }
    
    boost::shared_ptr<NormalMatchingsListContainer> getContinuousOcc()
        {
        return cachedContinuousOcc;
        }

   
    void changeTree()
        {
        alreadyAddedMappingGroup=false;
        }
        
    void processOneTreelet(TreeIdType numTree,
                            const PosList& mapp,
                            const ProjectedMappingInfos* pProjInfo,
                            size_t count_emitted,
                            MatchingsContainerSmartPtr occ,
                            size_t j,size_t numMapping)
        {
        if (!alreadyAddedMappingGroup)
            {
            MappingsPerTree mpt;
            mpt.numTree=numTree;
            cachedContinuousOcc->pV->push_back( mpt);
            alreadyAddedMappingGroup=true;
            }
        PosList mappWithParent=occ->getMappingAtIndices(j,numMapping);
        cachedContinuousOcc->pV->back().mappingList.push_back(mappWithParent);
        }
     
    };
    
template<typename TreeletType>
class PostProcessingHandlersFactoryInterfaceGeneric
    {
    typedef ResultsHandlerInterfaceGeneric<TreeletType> ResultsHandlerInterface;
    typedef IndividualResultsProcessorGeneric<TreeletType> IndividualResultsProcessor;
    public:
    virtual ~PostProcessingHandlersFactoryInterfaceGeneric(){};
    virtual boost::shared_ptr<ResultsHandlerInterface> createHandler(IndividualResultsProcessor &processor) const=0;
    };


template<typename TreeletType>
class CachedPostProcessingHandlersFactoryGeneric: public PostProcessingHandlersFactoryInterfaceGeneric<TreeletType>
    {
    typedef PostProcessingHandlerGeneric<TreeletType> PostProcessingHandler;
    typedef CacheCreationProcessorGeneric<TreeletType> CacheCreationProcessor;
    typedef IndividualResultsProcessorGeneric<TreeletType> IndividualResultsProcessor;
    typedef ResultsHandlerInterfaceGeneric<TreeletType> ResultsHandlerInterface;
    typedef TreeDBGeneric<TreeletType> TreeDB;
    
    const TargetList& tgtList;
    //const std::vector<LabelType>& mostFrequentLabelIds;
    std::map<LabelType, MatchingsContainerSmartPtr> cachedOccWithContinuousProjection;
    size_t sizeCache;
    
    PostProcessingParameters ppParams;
    /*
    size_t maxNumberOfSamplePerTreelet;
    //bool handleOnlyContinuousProjections;
    //bool filterDominatedTreelets;
    bool handleOnlyContinuousProjections;
    bool filterOutDominated;
    bool keepElementaryWildcard;
    bool keepTopWilcardChain;
    */
    public: 
    
    CachedPostProcessingHandlersFactoryGeneric(const TargetList& tgtList,
                                        const std::vector<LabelType>& mostFrequentLabelIds,
                                        //size_t maxNumberOfSamplePerTreelet,
                                        size_t sizeCache,
                                        const TreeDB& treeDB,
                                        const PostProcessingParameters& ppParams
                                        //bool filterOutDominated,
                                        //bool keepElementaryWildcard,
                                        //bool keepTopWilcardChain
                                        ):
                                        
                                        tgtList(tgtList),
                                        //mostFrequentLabelIds(mostFrequentLabelIds),
                                        //maxNumberOfSamplePerTreelet(maxNumberOfSamplePerTreelet),
                                        sizeCache(sizeCache),
                                        ppParams(ppParams)
                                        //handleOnlyContinuousProjections(true),filterOutDominated(filterOutDominated),
                                        //keepElementaryWildcard(keepElementaryWildcard),
                                        //keepTopWilcardChain(keepTopWilcardChain)

        {
        createCache(treeDB,mostFrequentLabelIds);
        }

    CachedPostProcessingHandlersFactoryGeneric(const TargetList& tgtList,
                                        //size_t maxNumberOfSamplePerTreelet,
                                        //bool filterOutDominated,
                                        //bool keepElementaryWildcard,
                                        //bool keepTopWilcardChain
                                        const PostProcessingParameters& ppParams
                                        ):    
                                        tgtList(tgtList),
                                        ppParams(ppParams)
                                        //filterOutDominated(filterOutDominated),//keepElementaryWildcard(keepElementaryWildcard),
                                        //keepTopWilcardChain(keepTopWilcardChain),
                                        //maxNumberOfSamplePerTreelet(//maxNumberOfSamplePerTreelet),
                                        //handleOnlyContinuousProjections(false)

        {}
        
    boost::shared_ptr<ResultsHandlerInterface> createHandler(IndividualResultsProcessor &processor) const
        {
        PostProcessingHandler* pH=new PostProcessingHandler(tgtList,
                     cachedOccWithContinuousProjection,
                     //maxNumberOfSamplePerTreelet,
                     processor,
                     ppParams
                     //handleOnlyContinuousProjections,
                     //filterOutDominated,keepElementaryWildcard,
                     //keepTopWilcardChain
                     );
                                                         
        boost::shared_ptr<ResultsHandlerInterface> spH(pH);
        return spH;
        
        // return boost::make_shared<PostProcessingHandler>(tgtList,
                                                         // cachedOccWithContinuousProjection,
                                                         // maxNumberOfSamplePerTreelet,
                                                         // processor);
        }
        
    void createCache(const TreeDB& treeDB,const std::vector<LabelType>& mostFrequentLabelIds)
        {
#ifdef WITH_MESSAGES
            std::cerr<<"INFO: Caching output... ";
            fflush(stderr);
#endif //WITH_MESSAGES
        size_t max=sizeCache;
        if (sizeCache>mostFrequentLabelIds.size())
            {
            max=mostFrequentLabelIds.size();
            }
        for(size_t i=0;i<max;++i)
            {
            LabelType label=mostFrequentLabelIds.at(i);
            cacheOutputForLabel(label,treeDB.getOccFromInvertedIndex(label) );
            }
#ifdef WITH_MESSAGES
            std::cerr<<"done."<<std::endl;
#endif //WITH_MESSAGES
        }

    void cacheOutputForLabel(LabelType label, MatchingsContainerSmartPtr occ)
        {
        std::map<LabelType, MatchingsContainerSmartPtr> emptyCache;
        CacheCreationProcessor ccp;
        PostProcessingHandler handler(tgtList,
                                       emptyCache,
                                       //maxNumberOfSamplePerTreelet,
                                       ccp,
                                       //handleOnlyContinuousProjections,
                                       //filterOutDominated,keepElementaryWildcard,
                                       //keepTopWilcardChain
                                       ppParams
                                       );
        Node dummyTree(label,0);
        TreeletType dummyTreelet(&dummyTree);        
        handler.processOneOccurrenceList(occ,dummyTreelet,true);
        cachedOccWithContinuousProjection[label]=ccp.getContinuousOcc();
        }   
        
    };
 
//typedef CachedPostProcessingHandlersFactoryGeneric<Treelet> CachedPostProcessingHandlersFactory;

// class SimplePostProcessingHandlersFactory: public PostProcessingHandlersFactoryInterface
    // {
    // size_t maxNumberOfSamplePerTreelet;
    // public: 
    
    // SimplePostProcessingHandlersFactory(size_t maxNumberOfSamplePerTreelet,
                                        // ):maxNumberOfSamplePerTreelet(maxNumberOfSamplePerTreelet) 
                                        // {}
    
    // boost::shared_ptr<ResultsHandlerInterface> createHandler(IndividualResultsProcessor &processor)
        // {
        // PostProcessingHandler* pH=new PostProcessingHandler(
                                                         // maxNumberOfSamplePerTreelet,
                                                         // processor);
                                                         
        // boost::shared_ptr<ResultsHandlerInterface> spH(pH);
        // return spH;
        
        // // return boost::make_shared<PostProcessingHandler>(tgtList,
                                                         // // cachedOccWithContinuousProjection,
                                                         // // maxNumberOfSamplePerTreelet,
                                                         // // processor);
        // }

        
    // };
 
#endif

