/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*! \file
    \ingroup tree_data_struct_groups
    \brief Representation of treelets as a sequence of position in the query tree.
*/

#include <vector>
#include <map>
#include <set>
#include <algorithm>
#include <boost/foreach.hpp>
#include <cassert>

#include <exception>

#include "TypeDefinitions.hpp"
#include "SeqOfPosRepresentationBase.hpp"
#ifndef _HPP_SEQOFPOS_WILDCARD_
#define _HPP_SEQOFPOS_WILDCARD_

//using boost::BOOST_FOREACH;

//template <typename T>
//class SetAsSortedVector
//    {
//    std::vector<>
//    }

namespace TreeletWildcard
{

class SeqOfPosRepresentation: public TreeletBase::SeqOfPosRepresentation
    {
    friend std::ostream &operator<< (std::ostream &stream,const SeqOfPosRepresentation &sop);
    
    protected:
    std::set<PosType> orderedSetOfNonWildcardPos;
    
    public:
    
        SeqOfPosRepresentation(): TreeletBase::SeqOfPosRepresentation() {}   
        
        SeqOfPosRepresentation(PosType rootPos,PosType parent=POSTYPE_SPECIAL_VALUE,bool isRootWildcard=false):TreeletBase::SeqOfPosRepresentation(rootPos,parent)
            {
            if (!isRootWildcard)
                {
                orderedSetOfNonWildcardPos.insert(rootPos);
                }
            }
        
        SeqOfPosRepresentation(const SeqOfPosRepresentation &objectToCopy):
           TreeletBase::SeqOfPosRepresentation(objectToCopy)//copy constructor
            {
            orderedSetOfNonWildcardPos=objectToCopy.orderedSetOfNonWildcardPos;
            isInAConsistentState=false;
            }

        SeqOfPosRepresentation(const TreeletBase::SeqOfPosRepresentation &objectToCopy):
           TreeletBase::SeqOfPosRepresentation(objectToCopy)//copy constructor
            {
            isInAConsistentState=false;
            }
            
        SeqOfPosRepresentation& operator=(const SeqOfPosRepresentation &objectToCopy)
            {
            // no self-assignment problem here
            TreeletBase::SeqOfPosRepresentation::operator=(objectToCopy);
            orderedSetOfNonWildcardPos=objectToCopy.orderedSetOfNonWildcardPos;
            isInAConsistentState=false;
            return *this;           
            }
    
    
    const std::set<PosType>& getIterableSeqOfPosNonWildcard() const
        {
        return orderedSetOfNonWildcardPos;
        }
    
    void printReprAsStringToStream(std::ostream &stream) const
        {
        BOOST_FOREACH(PosType pos,orderedSetOfPos)
            {
            stream<<(int)pos;
            if (orderedSetOfNonWildcardPos.count(pos)==0)
                {
                stream<<"*";
                }
            stream<<" ";
            }
        }
    
    bool isIncludedIn(SeqOfPosRepresentation other) const
        {
        bool allIncluded=TreeletBase::SeqOfPosRepresentation::isIncludedIn(other);
        if (!allIncluded) return false; //optimisation
        bool wildcardIncluded=includes(other.orderedSetOfNonWildcardPos.begin(),other.orderedSetOfNonWildcardPos.end(),orderedSetOfNonWildcardPos.begin(),orderedSetOfNonWildcardPos.end());
        return allIncluded && wildcardIncluded;
        }
    
    bool isEquivalentTo(const SeqOfPosRepresentation & other) const
        {
        bool allEquivalent=TreeletBase::SeqOfPosRepresentation::isEquivalentTo(other);
        if (!allEquivalent) return false; //optimisation
        
        if ( other.orderedSetOfNonWildcardPos.size()!=orderedSetOfNonWildcardPos.size() )
            {
            return false;
            }
        else
            {
            return std::equal(orderedSetOfNonWildcardPos.begin(),orderedSetOfNonWildcardPos.end(),other.orderedSetOfNonWildcardPos.begin());
            }
        }
    
    void mergeWith(const SeqOfPosRepresentation & other)
        {
        TreeletBase::SeqOfPosRepresentation::mergeWith(other);
        std::set<PosType> temp;
        std::set_union(orderedSetOfNonWildcardPos.begin(),orderedSetOfNonWildcardPos.end(),other.orderedSetOfNonWildcardPos.begin(),other.orderedSetOfNonWildcardPos.end(),
                        std::inserter(temp,temp.end()));
        orderedSetOfNonWildcardPos.swap(temp);
        isInAConsistentState=false;    
        }
    
    SeqOfPosRepresentation merge(const SeqOfPosRepresentation & other) const
    {
    TreeletBase::SeqOfPosRepresentation halfResult=TreeletBase::SeqOfPosRepresentation::merge(other);
    SeqOfPosRepresentation result(halfResult);
    std::set_union(orderedSetOfNonWildcardPos.begin(),orderedSetOfNonWildcardPos.end(),
                        other.orderedSetOfNonWildcardPos.begin(),other.orderedSetOfNonWildcardPos.end(),
                                std::inserter(result.orderedSetOfNonWildcardPos,result.orderedSetOfNonWildcardPos.end()));
    result.isInAConsistentState=false;
    return result;
    }
    
    pair<std::vector<PosType> ,std::vector<PosType> > separateWildcardFromNWInMapp(const std::vector<PosType>& mapp) const
        {
        std::vector<PosType> posListW;
        std::vector<PosType> posListNW;
        unsigned int num=0;
        BOOST_FOREACH(PosType pos,orderedSetOfPos)
            {
            if (std::find(orderedSetOfNonWildcardPos.begin(),orderedSetOfNonWildcardPos.end(),pos)!=orderedSetOfNonWildcardPos.end())
                {
                posListNW.push_back(mapp.at(num));
                }
            else
                {
                posListW.push_back(mapp.at(num));
                }
            num++;
            }
        assert(num==mapp.size());
        return std::make_pair(posListNW,posListW);
        }
    
    };

  
std::ostream &operator<<(std::ostream &stream,const std::set<unsigned char> &s)
  {
  stream<<"Set(";
  BOOST_FOREACH(unsigned char elem,s)
    {
    stream<<(int)elem<<",";
    }
  stream<<")";
  return stream;
  }
  
std::ostream &operator<< (std::ostream &stream,const SeqOfPosRepresentation &sop)
  {
  //using namespace baseDefinitions;
  
  stream<<"SOP(";
  stream<<(int)sop.parent;
  stream<<"|";
  //set<PosType> copy(sop.orderedSetOfPos);
  stream<<sop.orderedSetOfPos;
  stream<<"|";
  stream<<sop.orderedSetOfNonWildcardPos;
  stream<<")";
  return stream;
  }
  
}; //namespace TreeletWildcard 


  


    
#endif






