/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/



#include "EBMTppConversionCode.hpp"
   
/// \brief TMRetrievalResult is an implementation of a TRResult interface
//class TMRetrievalResult;

/// \brief TreeletRetrieverForCPP is the class handling the search engine and translation examples
template<typename TreeletType>
class TreeletRetrieverForCPPGeneric: public TreeletRetrieverWithContinuousTargetSideGeneric<TreeletType>, public TRInterface {
public:
    /// \brief Constructor
    TreeletRetrieverForCPPGeneric(const std::string& datapath, const std::string& side, 
                           const SearchParameters& sParams,
                           const PostProcessingParameters& ppParams
        ):                    
        TreeletRetrieverWithContinuousTargetSideGeneric<TreeletType>(datapath.c_str(),side.c_str(),
                                                 sParams,
                                                 ppParams
            ){}
    
    /// \brief Main method for processing a query
    TRResultPtr retrieve(const std::string& labelListAsString, const std::string& dependListAsString);
    
private:
    
    void handleRequestAsString(const std::string& labelListAsString, const std::string& dependListAsString, TMRetrievalResult &result);
    // std::ptr_map<TreeIdType, ParaSentence> cached_frequent_para_sentences;
    // void fill_frequent_para_sentences 
};


template<typename TreeletType>
class TMMapGenerationProcessorGeneric : public IndividualResultsProcessorGeneric<TreeletType> {
    //typedef Treelet TreeletType;
protected:
    TMRetrievalResult &result;
    const AddInfosVectorInterface& aDataVect;
    const TargetList& targetInfoList;
    const std::vector<std::string>& numToTreeIdVector;
    std::map<TreeIdType, ParaSentence*> constructed_para_sentences_cache;

    size_t nbMatchingsForCurrentTreelet;
    size_t nbTreelets;
    
    std::set<PosType> current_seq_of_pos;
    
    std::set<PosType> current_seq_of_wildcard; // only used in wildcard case
    
public:
    TMMapGenerationProcessorGeneric(TMRetrievalResult& result,
                             const AddInfosVectorInterface& aDataVect,
                             const TargetList& targetInfoList,
                             const std::vector<std::string>& numToTreeIdVector) :
        result(result), aDataVect(aDataVect),   targetInfoList(targetInfoList), numToTreeIdVector(numToTreeIdVector), nbMatchingsForCurrentTreelet(0), nbTreelets(0) {}
    
    void initializeForThisTreelet(const TreeletType &treelet) {
        //std::cerr<<"initialize treelet "<<treelet<<std::endl;
        const std::set<PosType>& seqOfPos = treelet.getSeqOfPosRepresentation().getIterableSeqOfPos();
        
        
        current_seq_of_pos=seqOfPos;
        
        if (TreeletType::isUsingWildcard()) {
            const std::set<PosType>& seqOfPosNW = treelet.getSeqOfPosRepresentation().getIterableSeqOfPosNonWildcard();
            current_seq_of_wildcard = std::set<PosType>();
            if (seqOfPosNW.size() < seqOfPos.size()) {
                BOOST_FOREACH(PosType pos, seqOfPos) {
                    if (seqOfPosNW.count(pos) == 0) {
                        current_seq_of_wildcard.insert(pos);
                    }
                }
            }
            assert(current_seq_of_wildcard.size() + seqOfPosNW.size() == seqOfPos.size());
            result.setCurrentSeqOfPos(seqOfPosNW);
        } else {
            result.setCurrentSeqOfPos(seqOfPos);
        }
        
        nbTreelets += 1;
        nbMatchingsForCurrentTreelet = 0;
    }
    
    
    ParaSentence& get_para_sentence_for_num_tree(TreeIdType numTree) {
        if (constructed_para_sentences_cache.count(numTree)!=0) {
            //std::cerr<<"duplicate para_sentence for "<<numTree<<std::endl;
            return *(constructed_para_sentences_cache[numTree]);
        }
        const TargetInfo& targetInfo = targetInfoList.at(numTree);
        const AdditionalInfos& additional_info = aDataVect.at(numTree);
        const std::string& treeId = numToTreeIdVector.at(numTree);    
        const PosList& tgt_depend_list=targetInfo.depend;
        std::vector<std::pair<std::vector<PosType>,std::vector<PosType> > > match_vector=
                    targetInfo.getMatchingInfosAsVector();
        ParaSentence& para_sentence = result.push_one_empty_para_sentence_and_return_ref();
        
        TMBuilder::build_para_sentence_in_place(tgt_depend_list, match_vector, additional_info,
                    treeId, para_sentence);
        constructed_para_sentences_cache[numTree]=&para_sentence;
        return para_sentence;
    
    }
    
    void processOneTreelet(TreeIdType numTree,
                           const PosList& mapp,
                           const ProjectedMappingInfos* pProjInfo,
                           size_t count_emitted,
                           MatchingsContainerSmartPtr occ,
                           size_t j,size_t numMapping) {
        if (pProjInfo == NULL) {
            throw std::exception();
        }
            
        ParaSentence& para_sentence=get_para_sentence_for_num_tree(numTree);
            
		result.push_one_tm_with_para_sentence(para_sentence);
		TM& tm=result.back();
			            
        TMBuilder::build_tm_in_place_without_para_sentence(current_seq_of_pos, mapp, 
                *(pProjInfo->pConnectedImage),
                pProjInfo->additionalNodes, tm);
        
        if (TreeletType::isUsingWildcard()) {
            TMBuilder::set_wildcards_of_tm(current_seq_of_wildcard, tm);
        }
        
        if (nbMatchingsForCurrentTreelet<=1) { //only need to to this once
            result.set_real_total_number_of_tm(current_seq_of_pos, occ->getNbSentOcc());
        }
        
        nbMatchingsForCurrentTreelet += 1;
    }
        
    void finalizeTreelet() {
        result.clean_entry_if_empty();
    }
        
    void finalizeAll() {
        result.finalize();
    }
};

template<typename TreeletType>
void TreeletRetrieverForCPPGeneric<TreeletType>::handleRequestAsString(const std::string& labelListAsString, const std::string& dependListAsString, TMRetrievalResult &result) {        
    //using TreeletRetrieverWithContinuousTargetSideGeneric<TreeletType>;
    TMMapGenerationProcessorGeneric<TreeletType> processor(result,
            TreeletRetrieverWithContinuousTargetSideGeneric<TreeletType>::postProcessingInfos.getAdditionalInfos(),
            TreeletRetrieverWithContinuousTargetSideGeneric<TreeletType>::postProcessingInfos.getTargetInfos(),
            TreeletRetrieverWithContinuousTargetSideGeneric<TreeletType>::postProcessingInfos.getNumToTreeIdVector());
                            
    TreeletRetrieverWithContinuousTargetSideGeneric<TreeletType>::handleRequestAsStringWithGivenProcessor(labelListAsString, dependListAsString, processor);

}           

template<typename TreeletType>
TRResultPtr TreeletRetrieverForCPPGeneric<TreeletType>::retrieve(const std::string& labelListAsString, const std::string& dependListAsString) {
    boost::shared_ptr<TMRetrievalResult> result_ptr = boost::make_shared<TMRetrievalResult>();
    handleRequestAsString(labelListAsString, dependListAsString, *result_ptr);
    return result_ptr;
};  
