/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "tr_interface.hpp"

#include <boost/serialization/vector.hpp>
#include <boost/serialization/map.hpp>
#include <boost/serialization/set.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>

#include <boost/ptr_container/ptr_vector.hpp>
#include <boost/ptr_container/ptr_map.hpp>
//#include <boost/ptr_container/serialize.hpp>
#include <glog/logging.h>

class TMResultException : public std::exception {
    virtual const char* what() const throw()
        {
            return "Trying to to retrieve a Pattern or a TM that do not exist in a TMRetrievalResult object";
        }
};

PatternType convert_pos_list(const PosList& pos_list) {
    PatternType result;
    BOOST_FOREACH(PosType pos, pos_list) {
        if (pos == POSTYPE_SPECIAL_VALUE) {
            assert(false);
        }
        result.insert(pos);
    }
    return result;
}

PosList convert_pattern_type(const PatternType& pattern) {
    PosList result;
    BOOST_FOREACH (WordIndex w_index, pattern) {
        if (w_index > POSTYPE_MAXIMAL_VALUE || w_index == POSTYPE_SPECIAL_VALUE) {
            assert(false);
        }
        result.push_back(w_index);
    }
    return result;
}    

typedef boost::ptr_vector<TM> TMContainer;
typedef boost::ptr_vector<ParaSentence> ParaSentenceContainer;
typedef boost::ptr_map<PosList, TMContainer> PatternToTMListMapType;

class TMRetrievalResult : public TRResult {
    ParaSentenceContainer created_para_sentences;
    PatternToTMListMapType main_map;
    TMContainer* current_vector;
    PosList current_pattern;
    std::vector<PatternType> vector_of_pattern;
    std::map<PatternType, size_t> real_total_number_of_tm_per_pattern;
    bool finalized;
    
    friend std::ostream &operator<< (std::ostream &stream, const TMRetrievalResult &result);
    friend class boost::serialization::access;
    
    template<class Archive>
    void serialize(Archive & ar, const unsigned int version)
    {
        //ar & main_map;
        ar & current_pattern;
        ar & finalized;
        ar & vector_of_pattern;
        //ar & current_vector;
    }
    
public:
    //constructor
    TMRetrievalResult() : current_vector(NULL), finalized(false){};
    
    ~TMRetrievalResult() {
        DLOG(INFO) << "TRResult object is being destroyed (No possible reference to Tms past this point)." << std::endl;
    }
    
    // for adding TMs: (not for user)
    void setCurrentSeqOfPos(const set<PosType>& seqOfPos);
    void setCurrentSeqOfPos(const set<PosType>& seqOfPos, const set<PosType>& seqOfPosNW);
    //void push_back(const TM&tm);

    void push_one_empty_tm();
    TM& back();


    void push_one_tm_with_para_sentence(ParaSentence& para_sentence);

    ParaSentence& push_one_empty_para_sentence_and_return_ref();
    
    void clean_entry_if_empty();
    void finalize();
    
    // for debugging:
    bool is_in_vector_of_pattern(const PatternType& pattern) const;
    
    // for accessing values: 
    // (defining a real interface would be better)
    const PatternToTMListMapType& get_internal_map() const {return main_map;}
    
    // API:
    const vector<PatternType>& get_vector_of_pattern() const;
    size_t get_nb_of_pattern() const;
    size_t get_nb_of_tm_for_pattern(const PatternType& pattern) const;
    const TM& get_tm(const PatternType& pattern,size_t index) const;
    const TMList& get_tm_list(const PatternType& pattern) const;
    TM& push_one_empty_tm_and_return_ref(const PatternType& pattern);
    
    void save_to_stream( std::ofstream& ofs ) {
        try {
            boost::archive::binary_oarchive oa(ofs);
            oa<<*this;
        }
        catch (boost::archive::archive_exception e) {
                cerr<<"Error saving TRResult"<<endl;
                cerr<<"what:"<<e.what()<<endl;
        }
    }
    
    size_t get_real_total_number_of_tm(const PatternType& pattern) const {
        if (real_total_number_of_tm_per_pattern.count(pattern) != 0) {
            return real_total_number_of_tm_per_pattern.at(pattern);
        } else {
            return 1;
        }
    }
    
    void set_real_total_number_of_tm(const set<PosType>& seqOfPos, size_t num) {
        PosList poslist(seqOfPos.begin(),seqOfPos.end());
        PatternType pattern = convert_pos_list(poslist);
        real_total_number_of_tm_per_pattern[pattern] = num;
    }
    
};



std::ostream &operator<< (std::ostream &stream,const TMRetrievalResult &result);
    


void TMRetrievalResult::setCurrentSeqOfPos(const set<PosType>& seqOfPos) {
    PosList pos_list(seqOfPos.begin(),seqOfPos.end());
    current_vector = &(main_map[pos_list]);
    current_pattern = pos_list;
}
    
void TMRetrievalResult::setCurrentSeqOfPos(const set<PosType>& seqOfPos, const set<PosType>& seqOfPosNW) {
    assert(false); //not implemented here
}
    
// void TMRetrievalResult::push_back(const TM&tm) {
    // assert(current_vector != NULL);
    // current_vector->push_back(tm);
// }
    
void TMRetrievalResult::push_one_tm_with_para_sentence(
        ParaSentence& para_sentence) {
    assert(current_vector != NULL);
    current_vector->push_back(new TM(para_sentence));
}

ParaSentence& TMRetrievalResult::push_one_empty_para_sentence_and_return_ref() {
    created_para_sentences.push_back(new ParaSentence);
    ParaSentence& para_sentence=created_para_sentences.back();
    return para_sentence;
}

void TMRetrievalResult::push_one_empty_tm() {
    assert(current_vector != NULL);
    created_para_sentences.push_back(new ParaSentence);
    ParaSentence& para_sentence=created_para_sentences.back();
    current_vector->push_back(new TM(para_sentence));
}

TM& TMRetrievalResult::back() {
	assert(current_vector != NULL && current_vector->size()>=1);
	return current_vector->back();
}
	
void TMRetrievalResult::clean_entry_if_empty() {
    assert(current_vector != NULL);
    if (current_vector->size() == 0) {
        PatternToTMListMapType::iterator it = main_map.find(current_pattern);
        assert(it != main_map.end());
        main_map.erase(it);
    }
}

void TMRetrievalResult::finalize() {
    assert(!finalized);
    assert(vector_of_pattern.size() == 0);
    
    
    for (PatternToTMListMapType::const_iterator it=main_map.begin();
        it!=main_map.end(); ++it) {
        PatternType pattern = convert_pos_list(it->first);
        vector_of_pattern.push_back(pattern);        
    }
    //typedef std::pair<PosList,TMContainer > KeyValuePair;
    // BOOST_FOREACH (const KeyValuePair& key_value_pair, main_map) {
        // PatternType pattern = convert_pos_list(key_value_pair.first);
        // vector_of_pattern.push_back(pattern);
    // }
    finalized = true;
};

const vector<PatternType>& TMRetrievalResult::get_vector_of_pattern() const {
    assert(finalized);
    return vector_of_pattern;
};

size_t TMRetrievalResult::get_nb_of_pattern() const {
    assert(finalized);
    return vector_of_pattern.size();        
};
    
size_t TMRetrievalResult::get_nb_of_tm_for_pattern(const PatternType& pattern) const {
    assert(finalized);
    PosList pos_list = convert_pattern_type(pattern);
    PatternToTMListMapType::const_iterator it = main_map.find(pos_list);
    if (it == main_map.end()) {
        return 0;
    }
    else {
        return (it->second)->size();
    }
};
    
const TM& TMRetrievalResult::get_tm(const PatternType& pattern, size_t index) const {
    assert(finalized);
    PosList pos_list = convert_pattern_type(pattern);
    PatternToTMListMapType::const_iterator it = main_map.find(pos_list);
    if (it == main_map.end()) {
        //cerr<<"c1"<<endl;
        throw TMResultException();
    }
    else if ((it->second)->size() <= index) {
        //cerr<<"c2 "<<index<<" >="<<endl;
        throw TMResultException();
    }
    else {
        const TMContainer& tmc =  *(it->second);
        return tmc[index];
    }        
};

TM& TMRetrievalResult::push_one_empty_tm_and_return_ref(const PatternType& pattern) {
    assert(finalized);
    PosList pos_list = convert_pattern_type(pattern);
    PatternToTMListMapType::iterator it = main_map.find(pos_list);

    created_para_sentences.push_back(new ParaSentence);
    ParaSentence& para_sentence=created_para_sentences.back();
    
    if (it == main_map.end()) {
        assert(!is_in_vector_of_pattern(pattern));
        vector_of_pattern.push_back(pattern);
        main_map[pos_list].push_back(new TM(para_sentence));
        return main_map[pos_list].back();
    }
    else {
        assert(is_in_vector_of_pattern(pattern));
        (it->second)->push_back(new TM(para_sentence));
        return (it->second)->back();
    }
}

bool TMRetrievalResult::is_in_vector_of_pattern(const PatternType& pattern) const
    {
    assert(finalized);
    BOOST_FOREACH(const PatternType& pattern_already_in,vector_of_pattern)
        {
        if (pattern_already_in==pattern)
            {
            return true;
            }
        }
    return false;
    }
    
const TMList& TMRetrievalResult::get_tm_list(const PatternType& pattern) const {
    assert(finalized);
    PosList pos_list = convert_pattern_type(pattern);
    PatternToTMListMapType::const_iterator it = main_map.find(pos_list);
    if (it == main_map.end()) {
        throw TMResultException();
    }
    else {
        return *(it->second);
    }        
};
    
std::ostream &operator<< (std::ostream &stream, const TMRetrievalResult &result) {
    stream << "RESULT(" << result.main_map.size() << ")" << std::endl;
    stream << "NOT IMPLEMENTED ";
    // typedef std::pair<PosList,TMContainer > KeyValuePair;
    // BOOST_FOREACH (const KeyValuePair& key_value_pair, result.main_map) {
        
        // //stream<<key_value_pair.first<<" -> "<<key_value_pair.second<<std::endl;
    // }
    stream << "END RESULT" << std::endl;
    return stream;
}

TRResultPtr load_tr_result_from_stream(std::ifstream& ifs) {       
    boost::shared_ptr<TMRetrievalResult> result = boost::make_shared<TMRetrievalResult>();
    try {
        boost::archive::binary_iarchive ia(ifs);
        ia>>(*result);
    }
    catch (boost::archive::archive_exception e) {
        cerr<<"Error loading TRResult "<<endl;
        cerr<<"what:"<<e.what()<<endl;
        return boost::shared_ptr<TMRetrievalResult>();
    }
    return result;
}

