/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*! \file
    \ingroup server_group
    \brief Generic code for server.
*/

#include <ctime>
#include <iostream>
#include <string>
#include <vector>
#include <boost/asio.hpp>
#include <boost/foreach.hpp>
#include "utf8cpp/utf8.h"
#include "TypeDefinitions.hpp"
#include "GlobalTreeletRetrieving.hpp"
  
#include "StringConversions.hpp"

using boost::asio::ip::tcp;

class SerializationOverStreamProcessor: public IndividualResultsProcessor
    {
    protected:
    std::iostream& stream;
    size_t nbMatchingsForCurrentTreelet;
    size_t nbTreelets;
    public:
    SerializationOverStreamProcessor(std::iostream& stream):stream(stream),nbMatchingsForCurrentTreelet(0)
                        {}
    
    void initializeForThisTreelet(const Treelet &treelet)
        {
        
        std::set<PosType> seqOfPos=treelet.getSeqOfPosRepresentation().getIterableSeqOfPos();
        nbTreelets+=1;
        }
    
    virtual void addAdditionalInfos(TreeIdType numTree)
        {}
    
    void processOneTreelet(TreeIdType numTree,
                            const PosList& mapp,
                            const ProjectedMappingInfos* pProjInfo,
                            size_t count_emitted,
                            MatchingsContainerSmartPtr occ,
                            size_t j,size_t numMapping)
        {
        if (pProjInfo==NULL)
            {
            throw std::exception();
            }
            
        std::string code;
        PrefixEncodeVectors encoder(code);
        encoder.packUInt(numTree);
        encoder.packPosListUC(mapp);
        encoder.packPosSetUC( *(pProjInfo->pConnectedImage) );
        encoder.packPosListUC(pProjInfo->additionalNodes);
        encoder.packMatchListUC(pProjInfo->matchList);
        
        addAdditionalInfos(numTree,encoder);
        
        
        char bufferMessageLength[nbBytesForEncodingMessageLength];
        encodeMessageLength(code.size(),unsigned char *bufferMessageLength)
        stream.write(bufferMessageLength,nbBytesForEncodingMessageLength);
        stream.write(code.data(),code.size())
        }
        
    void finalizeTreelet()
        {
        char buffer[nbBytesForEncodingMessageLength];
        encodeMessageLength(0,buffer);
        stream.write(buffer,nbBytesForEncodingMessageLength);
        
        encodeMessageLength(nbMatchingsForCurrentTreelet,buffer);
        stream.write(buffer,nbBytesForEncodingMessageLength);
        }
        
    void finalizeAll()
        {
        char buffer[nbBytesForEncodingMessageLength];
        encodeMessageLength(nbTreelets,buffer);
        stream.write(buffer,nbBytesForEncodingMessageLength);        
        }
    };

class SerializationOverStreamProcessorWithAddInfos: public SerializationOverStreamProcessor
    {
    protected:
    const AdditionalInfosUnpacker& aDataVect;
    const TargetList& targetInfoList;
    public:
    SerializationOverStreamProcessorAddInfos(
                        std::iostream &stream;
                        const AdditionalInfosUnpacker& aDataVect,
                        const TargetList& targetInfoList):
                        SerializationOverStreamProcessor(stream),aDataVect(aDataVect),targetInfoList(targetInfoList){}
    
    void addAdditionalInfos(TreeIdType numTree,PrefixEncodeVectors& encoder)
        {
        const TargetInfo& targetInfo=targetInfoList[numTree];
        encoder.packPosListUC(targetInfo.depend);
        encoder.packMatchListUC( targetInfo.getMatchingInfosAsVector() ;
        
        encoder.addRawDataToCode(aDataVect.getRawData(numTree));
        }
    
    };


class ResultsHandlerNewProtocol: public ResultsHandlerInterface
    {
    const ResultsProcessor& resProc;
    std::iostream &stream;
    public:
    ResultsHandlerNewProtocol(ResultsProcessor& resProc, std::iostream &stream): 
                        resProc(resProc),stream(stream){}
                        
    void handleResults(const MatchingsContainerStore& occStore)
        {
        SerializationOverStreamProcessor sosp(stream);
        resProc.iterateAndActOnEveryTreeletOcc(occStore.getUnderlyingMap() ,sosp);
        }
    };
    
class ResultsHandlerNewProtocolWithAddInfos: public ResultsHandlerInterface
    {
    const ResultsProcessor& resProc;
    std::iostream &stream;
    public:
    ResultsHandlerNewProtocolWithAddInfos(ResultsProcessor& resProc, std::iostream &stream): 
                        resProc(resProc),stream(stream){}
                        
    void handleResults(const MatchingsContainerStore& occStore)
        {
        SerializationOverStreamProcessorWithAddInfos sosp(stream,resProc.getAddInfoVect(), resProc.getTargetInfoList());
        resProc.iterateAndActOnEveryTreeletOcc(occStore.getUnderlyingMap() ,sosp);
        }
    };
    

    
class ResultsProcessorNewProtocol: public ResultsProcessor
    {
    public:
    ResultsProcessorNewProtocol(const char *filename, 
                    const char *numToTreeIdFilename,
                    size_t maxNumberOfSamplePerTreelet,
                    const char *addDatasFilename)
                :ResultsProcessor(filename, numToTreeIdFilename,maxNumberOfSamplePerTreelet,addDatasFilename){}

    ResultsProcessorNewProtocol(const char *filename, 
                    const char *numToTreeIdFilename,
                    size_t maxNumberOfSamplePerTreelet)
                :ResultsProcessor(filename, numToTreeIdFilename,maxNumberOfSamplePerTreelet){}
                    
    ResultsHandlerNewProtocol getHandlerNewProtocol(std::iostream &stream)
        {
        //std::cout<<"requested handler"<<std::endl;
        return ResultsHandlerNewProtocol(*this,stream);
        }

    ResultsHandlerNewProtocolWithAddInfos getHandlerNewProtocolWithAddInfos(std::iostream &stream)
        {
        //std::cout<<"requested handler"<<std::endl;
        assert(!aDataVect.inactive);
        return ResultsHandlerNewProtocolWithAddInfos(*this,stream);
        }

    };

class TreeletRetrieverNewProtocol: public TreeletRetrieverBase<ResultsProcessorNewProtocol>
    {
    public:
    TreeletRetrieverPerlXS(const char *datapath, const char *side, 
                        size_t maxNumberOfSamplePerTreelet, bool preserveOrder
                                ):TreeletRetrieverBase<ResultsProcessorPerlXS>(datapath,side,maxNumberOfSamplePerTreelet,preserveOrder){}

    void handleRequestAsStringAndUseNewProtocol(const std::string& labelListAsString, const std::string& dependListAsString,std::iostream &stream)
        {
        ResultsHandlerNewProtocol resultHandler=resProc.getHandlerNewProtocol(stream);
        _handleRequestAsString(labelListAsString, dependListAsString,resultHandler);

        }
    
   void handleRequestAsStringAndUseNewProtocolWithAddInfos(const std::string& labelListAsString, const std::string& dependListAsString,std::iostream &stream)
        {
        ResultsHandlerNewProtocolWithAddInfos resultHandler=resProc.getHandlerNewProtocoWithAddInfos(stream);
        _handleRequestAsString(labelListAsString, dependListAsString,resultHandler);

        }                         
    };


int main(int argc, char *argv[])
    {
    if (argc!=5)
        {
        std::cerr<<"Usage: "<<argv[0]<< " /path/to/data"<<" side port max_number_of_sample_per_treelet"<<std::endl;
        return 1;
        }
    std::string datapath(argv[1]);
    std::string side(argv[2]);
    
    int port=atoi(argv[3]);
    int maxSample=atoi(argv[4]);
    //std::string portAsString(argv[3]);
    //std::string maxSampleAsString(argv[4]);
    
    if (side!="1" and side!="2")
        {
        std::cerr<<"side parameter should be 1 or 2, and not "<<side<<std::endl;
        return 1;
        }    
 
  //TreeDB("") treeDB;
  
  std::cout<<"Connecting to port "<<port<<" with database "<<datapath<< " ("<<side<<")"<<"  maxSample:"<<maxSample<<std::endl;
  try
  {
    boost::asio::io_service io_service;

    tcp::acceptor acceptor(io_service, tcp::endpoint(tcp::v4(), port));

    //LabelIndexer idx;
    //TreeDB treeDB("/home/fabien/progs/chiang/unittests_db/someBasicTests4_occidx1.boost_CP");

    TreeletRetrieverNewProtocol tr(datapath,side,
                        maxSample,
                        true);



    //for (int count=0;count<2;++count)
    while (1)
    {
      std::cout<<"ready"<<std::endl;
      tcp::socket socket(io_service);
      acceptor.accept(socket);
        
      std::cout<<"got connection"<<std::endl;
    
    while (1)
        {
        boost::system::error_code connection_error;
        boost::asio::streambuf b;
        size_t readSize=boost::asio::read_until(socket, b, "EOS\n",connection_error);
        std::cout<<"read "<<readSize<<" bytes"<<std::endl;
        if (connection_error==boost::asio::error::eof)
            {
            std::cout<<"EOF. End of connection"<<std::endl;
            break;
            }

        std::istream is(&b);
        std::string line;
        std::getline(is, line); 
        std::cout<<"received line "<<line<<std::endl;
        std::string labelListAsString=line;
     
        std::getline(is, line); 
        std::cout<<"received line "<<line<<std::endl;
        std::string dependListAsString=line;
    
        std::getline(is, line); 
        std::cout<<"received line "<<line<<std::endl;   
     
        boost::asio::streambuf message;
        std::ostream message_stream(&message);


        tr.handleRequestAsStringAndUseNewProtocolWithAddInfos(labelListAsString,dependListAsString,message_stream);

       boost::system::error_code send_error;
       boost::asio::write(socket, message,
       boost::asio::transfer_all(), send_error);
       if (send_error)
          {
            std::cout<<"Send error"<<std::endl;
          }
       } 
    }
  }
  catch (std::exception& e)
  {
    std::cerr << e.what() << std::endl;
  }

  return 0;
}
