/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*! \file
    \ingroup server_group
    \brief Main code for TRServer.
*/

#include <ctime>
#include <iostream>
#include <string>
#include <vector>
#include <boost/asio.hpp>
#include <boost/foreach.hpp>

#include <boost/program_options.hpp>
namespace po = boost::program_options;

#include "TreeletRetrieverNewProtocol.hpp" 
#include <sstream>

#include <boost/thread/thread.hpp>
//#include <boost/thread/thread_group.hpp>

using boost::asio::ip::tcp;


bool read_write_buffers_and_compute(std::istream& is, std::ostream& message_stream, TreeletRetrieverNewProtocol& tr, bool simpleTextMode) {
    std::string line;
    std::getline(is, line); 
    std::cerr<<"received line "<<line<<std::endl;
    std::string labelListAsString=line;
 
    std::getline(is, line); 
    std::cerr<<"received line "<<line<<std::endl;
    std::string dependListAsString=line;

    std::getline(is, line); 
    std::cerr<<"received line "<<line<<std::endl;   
   
    std::cerr<<"start handling"<<std::endl;
    try{
        if (simpleTextMode)
            {
            tr.handleRequestAsStringAndUseSimpleTextProtocol(labelListAsString,dependListAsString,message_stream);
            }
        else
            {
            tr.handleRequestAsStringAndUseNewProtocolWithAddInfos(labelListAsString,dependListAsString,message_stream);
            }
        }
    catch (std::invalid_argument e)
        {
        return false;
        }
    
    message_stream.flush();
    return true;
}

class tcp_connection
  : public boost::enable_shared_from_this<tcp_connection>
{
private:
    boost::asio::streambuf b;
    TreeletRetrieverNewProtocol& tr;
    tcp::socket socket_;
    std::string message_;
    bool simpleTextMode;

public:
    typedef boost::shared_ptr<tcp_connection> pointer;

    static pointer create(boost::asio::io_service& io_service, TreeletRetrieverNewProtocol& tr, bool simpleTextMode) {
    return pointer(new tcp_connection(io_service, tr, simpleTextMode));
    }

    tcp::socket& socket() {
    return socket_;
    }

    bool start() {
        boost::asio::async_read_until(socket_, b, "EOS\n",
                boost::bind(&tcp_connection::handle_read, shared_from_this(),
            boost::asio::placeholders::error,
            boost::asio::placeholders::bytes_transferred));
    }
    
        ~tcp_connection() {
    std::cerr<<"connection destroyed"<<std::endl;
    }

private:
  tcp_connection(boost::asio::io_service& io_service, TreeletRetrieverNewProtocol& tr, bool simpleTextMode)
    : socket_(io_service), tr(tr), simpleTextMode(simpleTextMode) {}

    void handle_write(boost::shared_ptr<boost::asio::streambuf> p_message,
                    const boost::system::error_code& error, size_t bytes_transferred) {
        //passing p_message as an argument here, so that it stays alive until the end of the copy
        std::cerr<<"finish transfered"<<std::endl; 
        if (error) {
            std::cerr<<"Send error"<<std::endl;
            std::cerr<<error.message()<<std::endl;
        }
        start();
    }
    
    void handle_read(const boost::system::error_code& error, size_t bytes_transferred) {
        std::cerr<<"read "<<bytes_transferred<<" bytes"<<std::endl;
        if (error==boost::asio::error::eof)
            {
            std::cerr<<"EOF. End of connection"<<std::endl;
            return;
            }
        
        std::istream is(&b);
        
        boost::shared_ptr<boost::asio::streambuf> p_message=make_shared<boost::asio::streambuf>();
        //message=boost::asio::streambuf();
        std::ostream message_stream(p_message.get());
        
        bool success=read_write_buffers_and_compute(is, message_stream, tr, simpleTextMode);
        if (!success) {
            std::cerr<<"Invalid argument received. Interrupting connection"<<std::endl;
            return;
        }
        
        std::cerr<<"finished handling"<<std::endl; 

        // synchronous write code
        boost::system::error_code send_error;
        boost::asio::write(socket_, *p_message,
        boost::asio::transfer_all(), send_error);
        std::cerr<<"finish transfered"<<std::endl; 
        if (send_error)
            {
                std::cerr<<"Send error"<<std::endl;
                std::cerr<<send_error.message()<<std::endl;
            }
        start(); //process next request
        
        // end synchronous write code
        
        // Asynchronous write (disabled until we can get prioritized handlers)
        // boost::asio::async_write(socket_, *p_message,
        // boost::bind(&tcp_connection::handle_write, shared_from_this(),p_message,
          // boost::asio::placeholders::error,
          // boost::asio::placeholders::bytes_transferred));

    }

};  
//////////////////////////////////////

class tcp_server {
public:
    bool simpleTextMode;
    TreeletRetrieverNewProtocol &tr;
    tcp::acceptor& acceptor_;

public:
  tcp_server(boost::asio::io_service& io_service, tcp::acceptor& acceptor_, TreeletRetrieverNewProtocol& tr, bool simpleTextMode)
    : acceptor_(acceptor_), tr(tr), simpleTextMode(simpleTextMode)
  {
    start_accept();
  }
  ~tcp_server() {
    std::cerr<<"Server destroyed"<<std::endl;
  }

private:
  void start_accept()
  {
    tcp_connection::pointer new_connection =
      tcp_connection::create(acceptor_.get_io_service(), tr, simpleTextMode);

    acceptor_.async_accept(new_connection->socket(),
        boost::bind(&tcp_server::handle_accept, this, new_connection,
          boost::asio::placeholders::error));
  }

  void handle_accept(tcp_connection::pointer new_connection,
      const boost::system::error_code& error) {
    if (!error){
        std::cerr<<"starting connection..."<<std::endl;
        new_connection->start();
    }
    else {
        std::cerr<<"Asynchronous Connection Error"<<std::endl;
        std::cerr<<error.message()<<std::endl;
    }
    start_accept();
    }

};





void serve_connection(tcp::socket& socket, TreeletRetrieverNewProtocol& tr, bool simpleTextMode)
    {
    while (1)
        {
        boost::system::error_code connection_error;
        boost::asio::streambuf b;
        size_t readSize=boost::asio::read_until(socket, b, "EOS\n",connection_error);
        std::cerr<<"read "<<readSize<<" bytes"<<std::endl;
        if (connection_error==boost::asio::error::eof)
            {
            std::cerr<<"EOF. End of connection"<<std::endl;
            break;
            }
        
        std::istream is(&b);
        boost::asio::streambuf message;
        std::ostream message_stream(&message);
        
        bool success=read_write_buffers_and_compute(is, message_stream, tr, simpleTextMode);
        if (!success) {
            std::cerr<<"Invalid argument received. Interrupting connection"<<std::endl;
            break;
        }
        
        std::cerr<<"finished handling"<<std::endl; 
        
        boost::system::error_code send_error;
        boost::asio::write(socket, message,
        boost::asio::transfer_all(), send_error);
        std::cerr<<"finish transfered"<<std::endl; 
        if (send_error)
            {
                std::cerr<<"Send error"<<std::endl;
            }
       
       }
    
    }

void synchronous_serve_loop(TreeletRetrieverNewProtocol &tr, bool simpleTextMode, 
                                boost::asio::io_service& io_service, int port) {
    
    tcp::acceptor acceptor(io_service, tcp::endpoint(tcp::v4(), port));
    while (1) {
        std::cerr<<"ready"<<std::endl;
        tcp::socket socket(io_service);
        acceptor.accept(socket);

        std::cerr<<"got connection"<<std::endl;

        serve_connection(socket, tr, simpleTextMode);      
    }
}

int main(int argc, char *argv[])
    {
    // Declare the supported options.
    po::options_description postProcessingOptions("Result Filtering Options");
    po::options_description searchOptions("Search Options");
    po::options_description desc("TRServer++ datapath port side");
    
    searchOptions.add_options()
        ("limit_siblings", po::value<int>()->default_value(0), "limit siblings (0: no limit  1:weak limit 2:strong limit)")
        ("keep_trailing_wildcards", po::bool_switch(), "search even treelets with trailing wildcards")
        ("no_grouping", po::bool_switch(), "disable grouping of dominated treelets")      
        ("max_nb_gaps", po::value<int>()->default_value(-1), "max number of gaps allowed per treelet (more precisely: the limit is on the sum of the sizes of all gaps)")
    ;
        
    postProcessingOptions.add_options()
        ("max_nb_sample", po::value<int>()->default_value(200), "max number of sample per treelet")
        ("output_all_non_empty_discovered", po::bool_switch(), "output all treelets that were discovered and non empty")
        ("keep_elementary_wildcards", po::bool_switch(), "output even elementary wildcards")
        ("keep_top_wildcards",po::bool_switch(), "output even top wildcards chain")
        ("no_filtering_for_target_continuity",po::bool_switch(), "output even occurences without target continuity")
        ("no_enforce_proj_ground_continuous",po::bool_switch(), " (For Wildcard mode): output even occurences without continuity in the target projection of the grounded positions")
        ("enforce_proj_gap_continuous",po::bool_switch(), " (For Wildcard mode): output only occurences with continuity in the target projection of the gap positions")
        ("enforce_proj_all_continuous",po::bool_switch(), " (For Wildcard mode): output only occurences with continuity in the target projection of all (gap and ground) positions")
        ("max_total_nb_of_tm_to_send", po::value<int>()->default_value(50000), "max total number of sample to send per query (-1 for no limit)")
    ;
        
    desc.add_options()
        ("help,h", "produce help message")
        ("datapath", po::value<string>()->default_value(""), "/path/to/data")
        ("port", po::value<int>()->default_value(0), "port to listen to")
        ("side", po::value<string>()->default_value("1"), "language side")
        //("simple_text_mode", po::value<bool>()->default_value(false), "simple text mode")
        ("simple_text_mode", po::bool_switch(), "simple text mode")
        ("serve_once_and_finish", po::bool_switch(), "end program after first reply")
        ("nb_threads", po::value<int>()->default_value(1), "number of threads to use (experimental)")
    ;
    
    desc.add(searchOptions).add(postProcessingOptions);
    
    po::positional_options_description p;
    p.add("datapath", 1).add("side",1).add("port",1).add("max_nb_sample",1);
    
    po::variables_map vm;
    //po::store(po::parse_command_line(argc, argv, desc), vm);
    po::store(po::command_line_parser(argc, argv).
          options(desc).positional(p).run(), vm);
    po::notify(vm);    

    if (vm.count("help")) {
        cout << desc << "\n";//<<p<<"\n";
        return 1;
    }    
    
    std::string datapath=vm["datapath"].as<string>();
    std::string side=vm["side"].as<string>();
    int port=vm["port"].as<int>();
    int nb_threads=vm["nb_threads"].as<int>();
    int maxSample=vm["max_nb_sample"].as<int>();
    
    int maxNbWildcard=vm["max_nb_gaps"].as<int>();
    int maxTotalNbOfTmToSend=vm["max_total_nb_of_tm_to_send"].as<int>();
    int nbSiblingsLimit=vm["limit_siblings"].as<int>();
    bool simpleTextMode=vm["simple_text_mode"].as<bool>();
    bool serveOnce=vm["serve_once_and_finish"].as<bool>();
    bool outputAllDiscovered=vm["output_all_non_empty_discovered"].as<bool>();
    bool keepTrailingWildcards=vm["keep_trailing_wildcards"].as<bool>();
    bool keepElementaryWildcard=vm["keep_elementary_wildcards"].as<bool>();
    bool keepTopWilcardChain=vm["keep_top_wildcards"].as<bool>();
    bool noFilteringTargetContinuity=vm["no_filtering_for_target_continuity"].as<bool>();

    bool noEnforceProjOfGroundIsContinuous=vm["no_enforce_proj_ground_continuous"].as<bool>();
    bool enforceProjOfGapIsContinuous=vm["enforce_proj_gap_continuous"].as<bool>();
    bool enforceProjOfAllIsContinuous=vm["enforce_proj_all_continuous"].as<bool>();
    bool noGrouping=vm["no_grouping"].as<bool>();
    
    if (datapath.size()==0) {
        cout << "error: invalid datapath" << "\n";//<<p<<"\n";
        return 1;
    } 

    
    if (side!="1" and side!="2")
        {
        std::cerr<<"side parameter should be 1 or 2, and not "<<side<<std::endl;
        return 1;
        }    
 
  //TreeDB("") treeDB;
  
  std::cerr<<"Connecting to port "<<port<<" with database "<<datapath<< " ("<<side<<")"<<"  maxSample:"<<maxSample<<"  siblings limit:"<<nbSiblingsLimit<<"  simpleTextMode:"<<simpleTextMode<<" serveOnce:"<<serveOnce<<" outputAllDiscovered:"<<outputAllDiscovered<<" keepTrailingWildcards:"<<keepTrailingWildcards<<" keepTopWilcardChain:"<<keepTopWilcardChain<<" noFilteringTargetContinuity:"<<noFilteringTargetContinuity<<std::endl;
  try
  {
    boost::asio::io_service io_service;
    tcp::acceptor acceptor_(io_service, tcp::endpoint(tcp::v4(), port));
    std::cerr<<"listening on port "<<port<<std::endl;
    std::cerr<<"will start answering requests when database is loaded..."<<std::endl;
    
    SearchParameters sParams;
    sParams.nbSiblingsLimit=nbSiblingsLimit;
    sParams.preserveOrder=true;
    sParams.skipTrailingWildcard=!keepTrailingWildcards;
    sParams.maxNbOfWildcard=maxNbWildcard;
    sParams.reduceDominated=!noGrouping;
    
    PostProcessingParameters ppParams;
    ppParams.maxNumberOfSamplePerTreelet=maxSample;
    ppParams.filterOutDominated=!outputAllDiscovered;
    ppParams.keepElementaryWildcard=keepElementaryWildcard;
    ppParams.keepTopWilcardChain=keepTopWilcardChain;
    ppParams.handleOnlyContinuousProjections=!noFilteringTargetContinuity;
    ppParams.enforceProjOfGroundIsContinuous=!noEnforceProjOfGroundIsContinuous;
    ppParams.enforceProjOfGapIsContinuous=enforceProjOfGapIsContinuous;
    ppParams.enforceProjOfAllIsContinuous=enforceProjOfAllIsContinuous;
    ppParams.maxTotalNbOfTmToSend=maxTotalNbOfTmToSend;
    
    TreeletRetrieverNewProtocol tr(datapath,side,
                        sParams,
                        ppParams
                        );

    
    //synchronous_serve_loop(tr, simpleTextMode, io_service, port);
    tcp_server server(io_service, acceptor_, tr, simpleTextMode);
    std::cerr<<"Ready to answer queries!"<<std::endl;
    
    if (nb_threads>1) {
        boost::thread_group threads;
        for (int i=0; i<nb_threads; i++) {
            threads.create_thread(boost::bind(&boost::asio::io_service::run, &io_service));
        }
        io_service.run();
        threads.join_all();
    }
    else {
        io_service.run();
    }
    
    
    std::cerr<<"Server closed"<<std::endl;
  }
  catch (std::exception& e)
  {
    std::cerr<< "Exception!!!" <<std::endl;
    std::cerr << e.what() << std::endl;
  }

  return 0;
}



