/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*! \file
    \ingroup tree_data_struct_groups
    \brief Representation of trees.
*/

#include <vector>
#include <boost/foreach.hpp>
#include<boost/smart_ptr.hpp>
#include <cassert>
#include <set>

#include "TypeDefinitions.hpp"

#include<iostream>

#ifndef _HPP_MY_TREE_
#define _HPP_MY_TREE_

using namespace std;
using namespace boost;



class Node;

typedef vector<Node *> NodePointersList;

class Node
    {
    private:
    LabelType label;
    NodePointersList leftChildren;
    NodePointersList rightChildren;
    
    PosType position;
    
    Node *parent;
    size_t posInParent;
    
    void _appendNodeListByPostOrderTraversal(vector<const Node *> &nodeList) const; 
    friend std::ostream &operator<< (std::ostream &stream,const Node &node);
    public:
        Node(LabelType label, PosType position): 
                    label(label), position(position), parent(NULL),posInParent(0) {};
        
        ~Node()
            {
//            std::cout<<"Destructing node@"<<this<<" position:"<<(int)position<<" label:"<<label <<" lc:"<<leftChildren.size()<<" rc:"<<rightChildren.size<<std::endl;
            freeAllSubtree();
//            std::cout<<"done destructing node@"<<this<<" position:"<<(int)position<<" label:"<<label <<" lc:"<<leftChildren.size()<<" rc:"<<rightChildren.size<<std::endl;            
            }
        
        vector<const Node *> getNodePointersListByPostOrderTraversal() const
            {
            vector<const Node *> result;
            _appendNodeListByPostOrderTraversal(result);
            return result;
            }
            
        Node *getChildNodeByNum(size_t num) const
            {
            if (num<leftChildren.size())
                {
                return leftChildren[num];
                }
            else {
                assert (num< leftChildren.size()+ rightChildren.size() );
                return rightChildren[ num-leftChildren.size() ];
                }
                
            }
        
        size_t getNbOfChildren() const
            {
            return leftChildren.size()+rightChildren.size();
            }  

        size_t getNbOfChildrenLeft() const
            {
            return leftChildren.size();
            }
            
        void setParent(Node *pNode)
            {
            parent=pNode;
            }

        void setPosInParent(size_t index)
            {
            posInParent=index;
            }
            
        PosType getPosition() const
            {
            return position;
            }
            
        size_t getPosInParent() const
            {
            return posInParent;
            }
            
            
        void pushToLeftChildren(Node * pNode)
            {
            leftChildren.push_back(pNode);
            }

        void pushToRightChildren(Node * pNode)
            {
            rightChildren.push_back(pNode);
            }
            
        void updateInfosParentsChildren();
        
        void freeAllSubtree();
        
        bool hasParent() const
            {
            return parent!=NULL;
            }
            
        Node *getParent() const
            {
            return parent;
            }
            
        LabelType getLabel() const
            {
            return label;
            }

    };

void Node::freeAllSubtree()
    {
    BOOST_FOREACH(Node * pNode,leftChildren)
        {
        delete(pNode);
        }
    BOOST_FOREACH(Node * pNode,rightChildren)
        {
        delete(pNode);
        }
    }

void Node::updateInfosParentsChildren()
    {
    for (size_t index=0; index<leftChildren.size(); ++index)
        {
        Node *pChild=leftChildren[index];
        pChild->setParent(this);
        pChild->setPosInParent(index);
        }
    for (size_t index=0; index<rightChildren.size(); ++index)
        {
        Node *pChild=rightChildren[index];
        pChild->setParent(this);
        //std::cout<<"set RightChildren pos in parent: "<<index+leftChildren.size()<<std::endl;
        pChild->setPosInParent(index+leftChildren.size());
        }
//    if recursive:
//        for child in self.children:
//            child.updateInfosFatherOfChildren(recursive=True)
    }

void Node::_appendNodeListByPostOrderTraversal(vector<const Node *> &nodeList) const
    {
    BOOST_FOREACH(Node * pNode,leftChildren)
        {
        pNode->_appendNodeListByPostOrderTraversal(nodeList);
        }
    BOOST_FOREACH(Node * pNode,rightChildren)
        {
        pNode->_appendNodeListByPostOrderTraversal(nodeList);
        }
    nodeList.push_back(this);
    }

std::ostream &operator<< (std::ostream &stream,const Node &node)
  {

  stream<<"Node@"<<&node<<"("<<node.label<<":"<<(int)node.position<<"@"<<(int)node.posInParent<<"[";
  BOOST_FOREACH(const Node* child, node.leftChildren)
    {
    stream<<*child<<",";
    }
  stream<<"|";
  BOOST_FOREACH(const Node* child, node.rightChildren)
    {
    stream<<*child<<",";
    }
    stream<<"] )";  
  return stream;
  }

bool validateDependAndLabelList(const vector<LabelType>& labelList , const vector<PosType>& dependList,bool verbose)
    {
    size_t treeSize=labelList.size();
    if (treeSize!=dependList.size())
        {
        if (verbose) std::cerr<<"Error: dependency list and label list have different size"<<endl;
        return false;
        }
        
    int slotRoot=-1;
    for (size_t slot=0;slot<treeSize;++slot)
        {
        if (dependList[slot]==POSTYPE_SPECIAL_VALUE)
            {
            if (slotRoot!=-1)
                {
                if (verbose) std::cerr<<"Error: 2 roots in dependency list"<<endl;
                return false;
                }
            slotRoot=slot;
            }
        }
    if (slotRoot==-1)
        {
        if (verbose) std::cerr<<"Error: no root in dependency list"<<endl;
        return false;
        }

    for (size_t slot=0;slot<treeSize;++slot)
        {
        PosType pos=slot;
        std::set<PosType> seenPos;
        do
            {
            if (seenPos.count(pos)>0)
                {
                if (verbose) std::cerr<<"Error: cycle detected in dependency list"<<endl;
                return false;
                }
            if (pos>=treeSize)
                {
                if (verbose) std::cerr<<"Error: illegal position in dependency list: "<<(int)pos<<"  /max:"<<treeSize <<endl;
                return false;
                }
            seenPos.insert(pos);
            pos=dependList[pos];
            }
        while (pos!=POSTYPE_SPECIAL_VALUE);
        }        
    return true;
    }

boost::shared_ptr<Node> createTreeFromDependAndLabelList(const vector<LabelType>& labelList , const vector<PosType>& dependList)
    {
    size_t treeSize=labelList.size();
    
    //std::cout<<labelList<<std::endl;
    //std::cout<<dependList<<std::endl;
    
    if (!validateDependAndLabelList(labelList ,dependList, true) )
        {
        std::string errorMsg="The label list and dependency list are not consistent";
        throw std::invalid_argument(errorMsg);
        }
    
    NodePointersList nodeList;
    nodeList.reserve(treeSize);
    
    for ( PosType pos=0; pos< treeSize; ++pos )
        {
        assert (pos!=POSTYPE_SPECIAL_VALUE);
        Node * pNode=new Node(labelList[pos],pos);
        nodeList.push_back(pNode);
        }
        
    PosType posRoot=POSTYPE_SPECIAL_VALUE;
        
    for ( PosType pos=0; pos< treeSize; ++pos)
        {
        PosType posParent=dependList[pos];
        if (posParent!=POSTYPE_SPECIAL_VALUE)
            {
        
            if (pos<posParent)
                {
                nodeList[posParent]->pushToLeftChildren(nodeList[pos]);
                }
            else if (pos>posParent)
                {
                nodeList[posParent]->pushToRightChildren(nodeList[pos]);
                }
            else
                {
                assert(false);
                }
            }
        else
            {
            assert (posRoot==POSTYPE_SPECIAL_VALUE);
            posRoot=pos;
            nodeList[pos]->setParent(NULL);
            }
        }
    assert (posRoot!=POSTYPE_SPECIAL_VALUE);    
    BOOST_FOREACH(Node *pNode, nodeList)
        {
        pNode->updateInfosParentsChildren();
        }
        
    return boost::shared_ptr<Node>( nodeList[posRoot] );
    }
    
//int main(){ return 0; }
#endif

