/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*! \file
    \ingroup global_glue_group
    \brief Complete Tree Database.
*/

#include <boost/utility.hpp>
#include "PrecomputationManager.hpp"
#include "binaryMerge.hpp"

#include <vector>
#include "TypeDefinitions.hpp"

#include "TreeletDiscoveryManager.hpp"
#include "TreeletComputationManager.hpp"
#include "Tree.hpp"

//#include "params.hpp"
//#include "ResultsHandler.hpp"

#include <iostream>

#ifndef _HPP_TREE_DB_
#define _HPP_TREE_DB_

// For Post processing:

template<typename TreeletType>
class ResultsHandlerInterfaceGeneric
    {
    typedef MatchingsContainerStoreGeneric<TreeletType> MatchingsContainerStore;
    public:
    virtual ~ResultsHandlerInterfaceGeneric(){};
    virtual void handleResults(const MatchingsContainerStore& occStore)=0;
    };

struct SearchParameters
    {
    SearchParameters() : wildcardOnlyAtRoot(false), onlyContiguousSiblings(false), use_wildcard_pair_cache(false) {}
    bool preserveOrder;
    int nbSiblingsLimit;
    bool skipTrailingWildcard;
    bool wildcardOnlyAtRoot;
    bool onlyContiguousSiblings;
    int maxNbOfWildcard;
    bool reduceDominated;
    bool use_wildcard_pair_cache;
    };

// Main Class:
template<typename TreeletType>
class TreeDBGeneric: boost::noncopyable
    {
    typedef PrecomputationManagerGeneric<TreeletType> PrecomputationManager;
    typedef TreeletComputationManagerGeneric<TreeletType> TreeletComputationManager;
    typedef TreeletDiscoveringManagerGeneric<TreeletType> TreeletDiscoveringManager;
    typedef MatchingsContainerStoreGeneric<TreeletType> MatchingsContainerStore;
    typedef ResultsHandlerInterfaceGeneric<TreeletType> ResultsHandlerInterface;

    public:
    std::vector<OccListCompactPair> *pInvertedIndex;
    PrecomputationManager pcm;

    SearchParameters sParams;

    //ResultsHandler resultsHandler;

        TreeDBGeneric(std::string filename,//bool skipTrailingWildcard
                        const SearchParameters& sParams):
            sParams(sParams)
            //skipTrailingWildcard(skipTrailingWildcard)
            {
            initializeInvertedIndex(filename);
            }

   //     TreeDB(const char *filename, std::ostream& output): resultsHandler(output, 0)
   //         {
   //         initializeInvertedIndex(filename);
   //         }

        ~TreeDBGeneric()
            {
            delete pInvertedIndex;
            }

         MatchingsContainerSmartPtr getOccFromInvertedIndex(LabelType label) const
            {
            assert(label < pInvertedIndex->size() );
            const OccListCompactPair *pOccListCompactPair=&(pInvertedIndex->at(label) );
            return boost::make_shared<CompactPairMatchingsListContainer>(pOccListCompactPair);
            }


        void initializeInvertedIndex(std::string filename)
            {

#ifdef WITH_MESSAGES //TO DO: add a real logging system
                std::cerr<<"INFO: Loading index from "<<filename<<"... ";
                fflush(stderr);
#endif //WITH_MESSAGES
            pInvertedIndex=loadVectorOccListCompactPair(filename.c_str());
            if (pInvertedIndex==NULL)
                {
                std::cerr<<"Error loading from "<<filename<<"  ->exiting"<<std::endl;
                exit(1);
                }
#ifdef WITH_MESSAGES
                std::cerr<<"done."<<std::endl;
#endif //WITH_MESSAGES
            // needToComputeWildcardOcc() is defined in the namespace of the type of Treelet used
            pcm=PrecomputationManager(pInvertedIndex, TreeletType::needToComputeWildcardOcc(), sParams.use_wildcard_pair_cache);
            //std::cout<<*pInvertedIndex<<std::endl;
            }

        void retrieveTreelets(const Node* pQueryTree,
                                    //bool preserveOrder,
                                    //int nbSiblingsLimit,
                                    ResultsHandlerInterface& resultsHandler) const //TODO: change architecture so that it is possible to put back the const here?
            {
            TreeletComputationManager tcm(pcm,sParams.preserveOrder);
            TreeletDiscoveringManager tdm(pQueryTree,tcm,sParams.nbSiblingsLimit,sParams.skipTrailingWildcard,
                            sParams.wildcardOnlyAtRoot, sParams.onlyContiguousSiblings,
                            sParams.maxNbOfWildcard,sParams.reduceDominated);

            tdm.computeAllTreelets();

            tdm.addDominationInfos(tcm.getModifyableResults());
            resultsHandler.handleResults(tcm.getResults());
            //return tcm.result();
            }


        //void retrieveTreelets( (ResultsProcessor::ResultsHandler)& resultsHandler) const
        //    {
        //    }

    };

//typedef TreeDBGeneric<Treelet> TreeDB;

/*
MappingsList TreeDB::retrieveListOfTreelets(TreeList queryTreeList)
    {
    PreComputationManager pcm(this);
    FOREACH queryTree
        {
        retrieveTreelets(queryTree,pcm);
        }
    }
*/

#endif
