/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*! \file
    \ingroup treelet_computing_group
    \brief Manage the computation of discovered treelets.
*/

#include "binaryMerge.hpp"

#include <vector>
#include <map>

#include "TypeDefinitions.hpp"
#include "Treelets.hpp"
#include "PrecomputationManager.hpp"
#include "MatchingsContainers.hpp"
#include "ComputedSubtreeletsManager.hpp"

#ifndef _HPP_TREELET_COMPUTATION_MANAGER_
#define _HPP_TREELET_COMPUTATION_MANAGER_



/**************************************************************************
 * Treelet indexer for TCM
 */

template<typename TreeletType>
class TreeletIndexerForTCM
    {
    typedef typename TreeletType::TreeletSimpleIndexer TreeletSimpleIndexer;
    TreeletSimpleIndexer *pIndexer;
    public:
        TreeletIndexerForTCM(): pIndexer(NULL){};
        void setIndexer(TreeletSimpleIndexer *pIndexer)
            {
            this->pIndexer=pIndexer;
            }
        TreeletType getTreelet(TreeletId id) const
            {
            assert(pIndexer!=NULL);
            return pIndexer->getTreelet(id);
            }
    const TreeletSimpleIndexer *getUnderlyingIndexer() const
        {
        return pIndexer;
        }
/*        TreeletId getIdOfTreelet(Treelet treelet)
            {
            return treelet.getId();
            }
*/    
    };

/*
 *
 ***************************************************************************/


class CouldNotFindCoverageOfTreeletError: public std::exception
{
  virtual const char* what() const throw()
  {
    return "CouldNotFindCoverageOfTreeletError";
  }
};



/**************************************************************************
*     TreeletComputationManager
*/

template<typename TreeletType>
class TreeletComputationManagerGeneric
    {
    typedef PrecomputationManagerGeneric<TreeletType> PrecomputationManager;
    typedef MatchingsContainerStoreGeneric<TreeletType> MatchingsContainerStore;
    typedef typename TreeletType::SeqOfPosRepresentation_ SeqOfPosRepresentation;
    typedef typename TreeletType::TreeletStoreByChildrenIndices TreeletStoreByChildrenIndices;
    typedef typename TreeletType::TreeletSimpleIndexer TreeletSimpleIndexer;
    typedef ComputedSubTreeletsManagerGeneric<TreeletType> ComputedSubTreeletsManager;
    
    MatchingsContainerStore computedMatchings;    
    ComputedSubTreeletsManager subtreeletsManager;
    
    //TreeletIndexerForTCM indexer;
    std::set<TreeletType> setOfKnownEmptyTreelets;
    
    const PrecomputationManager &pcm; 
    TreeletIndexerForTCM<TreeletType> treeletIndexer;
    
    map<string,MatchingsContainerSmartPtr> cachedComputation;
    
    bool preserveOrder;
    
    enum MergeStatus {REJECTED,NOT_MERGEABLE_NOW,MERGED,FINISHED};
    
    void __increment_j_to_next_pos(size_t& j,const set<size_t>& iToSkip, size_t strictUpperValue) const;
    MergeStatus __tryToMergeSeqOfPos(size_t j, const vector<TreeletId>& orderedList,
                SeqOfPosRepresentation& totalSeqOfPos,const SeqOfPosRepresentation& seqOfPosOfFinalTreelet,
                vector< std::pair<TreeletType,SeqOfPosRepresentation> >& result) const;        
    public:
        TreeletComputationManagerGeneric(const PrecomputationManager& pcm,bool preserveOrder):pcm(pcm),preserveOrder(preserveOrder){};
    
    bool updatesubTreeletsWithLeftRightSplittedSubtreelets(const TreeletType& treelet,const TreeletStoreByChildrenIndices& mapTreeletToId,int index);
    
    void computeTreelet(TreeletType& t,const TreeletStoreByChildrenIndices* pMapTreeletToId=NULL,int index=-1);
    void _computeTreelet(TreeletType& t,const TreeletStoreByChildrenIndices* pMapTreeletToId=NULL,int index=-1);
    
    vector< std::pair<TreeletType,SeqOfPosRepresentation> > getOptimalSubtreeletsCoverage(const TreeletType& treelet) const; 
    
    void setIndexer(TreeletSimpleIndexer *pIndexer)
        {
        treeletIndexer.setIndexer(pIndexer);
        }
    
    void rememberTreeletIsEmpty(const TreeletType& treelet)
        {
        setOfKnownEmptyTreelets.insert(treelet);
        }
        
    bool knownAsEmpty(const TreeletType& treelet) const
        {
        return setOfKnownEmptyTreelets.count(treelet)>0;
        }
        
    void addComputedSubTreeletId(const TreeletType &treelet,TreeletId subTreeletIndex,size_t nbOcc)
        {
        subtreeletsManager.addSubTreeletId(treelet,subTreeletIndex, nbOcc);
        }
    
    const MatchingsContainerStore& getResults()
        {
        return computedMatchings;
        }

    MatchingsContainerStore& getModifyableResults()
        {
        return computedMatchings;
        }
        
    bool isDominatedBy(const TreeletType& treelet, const TreeletType& superTreelet) const
        {
        size_t nbSentOccSubtreelet= treelet.getNbSentOcc();
        size_t nbSentOccSuper=superTreelet.getNbSentOcc();
        
        //assert( superTreelet.getSize()>treelet.getSize() );
        assert(typename TreeletType::CompareTreeletsBySize()(superTreelet,treelet));
        assert(nbSentOccSuper<=nbSentOccSubtreelet);        

        if (nbSentOccSuper<nbSentOccSubtreelet) // if there is at least one sentence in which we have sub but not super...
            {
            return false;
            }
        else
            {
            MatchingsContainerSmartPtr occSub=computedMatchings.getResult(treelet);
            MatchingsContainerSmartPtr occSuper=computedMatchings.getResult(superTreelet);
            
            SeqOfPosRepresentation sopSub=treelet.getSeqOfPosRepresentation();
            SeqOfPosRepresentation sopSuper=superTreelet.getSeqOfPosRepresentation();
            
            std::vector< SlotPairType > equalList;
            BOOST_FOREACH(PosType pos, sopSub.getListOfCommonPos(sopSuper))
                {
                equalList.push_back( SlotPairType(  sopSub.getSlotOfPos(pos), sopSuper.getSlotOfPos(pos) ) );
                }            
            assert(equalList.size()==treelet.getSize());
            return occSub->isDominatedBy(*occSuper,equalList);
            
            }
        }
        
    std::vector<TreeletId> getImmediateDominatingTreeletsFromListUsingOwnIndexer(
                                    TreeletId treeletIndex, const std::vector<TreeletId>& superTreeletIdList) const
        {
        std::vector<TreeletId> result;
        TreeletType treelet=treeletIndexer.getTreelet(treeletIndex);

        BOOST_FOREACH( TreeletId superTreeletIndex,superTreeletIdList)
            {
            TreeletType superTreelet=treeletIndexer.getTreelet(superTreeletIndex);

            if (isDominatedBy(treelet,superTreelet))
                {
                result.push_back(superTreeletIndex);
                }
            }
        return result;
        }
           
    };

template<typename TreeletType> 
void TreeletComputationManagerGeneric<TreeletType>::computeTreelet(TreeletType &treelet,const TreeletStoreByChildrenIndices* pMapTreeletToId,int index)
    {
    //cerr<<"index:"<<index<<endl;
    //cerr<<"before compute type "<<treelet<<endl;
    treelet.computeTypeRepresentation(*(treeletIndexer.getUnderlyingIndexer()));
    //cerr<<"after compute type "<<treelet<<endl;    
    //if (cachedComputation.count(typeRepresentation)

    if (pcm.isPrecomputed(treelet, *(treeletIndexer.getUnderlyingIndexer())))
        {
        //std::cout<<treelet<<" is precomputed"<<std::endl;
        MatchingsContainerSmartPtr pOcc=pcm.getOccurrences(treelet, *(treeletIndexer.getUnderlyingIndexer()));
        //cerr<<"returned: "<<*pOcc<<endl;
        treelet.setNbSentOcc( pOcc->getNbSentOcc() );
        computedMatchings.addResult(treelet, pOcc );
        
        if(treelet.getNbSentOcc()>0)
            {
            //cout << "size of treelet= " << treelet.getSize() << endl;
            assert(treelet.getSize()==pOcc->getTypicalSizeOfMapp()-1);
            }
        return;
        }

    const std::string& typeRepresentation=treelet.getTypeRepresentation();
    map<string,MatchingsContainerSmartPtr>::const_iterator it=cachedComputation.find(typeRepresentation);
    if (it!=cachedComputation.end())
        {
        MatchingsContainerSmartPtr pOcc=it->second;
        if (pOcc->getNbSentOcc()>0)
            {
            treelet.setNbSentOcc(pOcc->getNbSentOcc());
            computedMatchings.addResult(treelet, pOcc );
            }
        else
            {
            treelet.setNbSentOcc(0);
            }
        return;
        }

    // if(treelet.isWildcardElemPair(xxx)) {
        

    // }

    if ( knownAsEmpty(treelet) )
        {
        //std::cout<<treelet<<" is known as empty"<<std::endl;
        treelet.setNbSentOcc(0); //TODO:???
        cachedComputation[typeRepresentation]=pcm.getEmptyMatchingsContainer();
        }
    else
        {
        //std::cout<<treelet<<" need to be computed"<<std::endl;
        _computeTreelet(treelet,pMapTreeletToId,index);
        cachedComputation[typeRepresentation]=computedMatchings.getResult(treelet);
        }
    }

    
    


template<typename TreeletType>
bool TreeletComputationManagerGeneric<TreeletType>::updatesubTreeletsWithLeftRightSplittedSubtreelets
            (const TreeletType& treelet,const TreeletStoreByChildrenIndices& mapTreeletToId,int index)
    {
    assert(index>=0);
    // no more necessary after changes in discovering process:
    //TreeletId leftTreeletId=mapTreeletToId.getLeftTreeletId(treelet,index);
    //size_t nbOccLeft=treeletIndexer.getTreelet(leftTreeletId).getNbSentOcc();
    
    TreeletId rightTreeletId=treelet.getChildIndex(index);
    size_t nbOccRight=treeletIndexer.getTreelet(rightTreeletId).getNbSentOcc();
    
    //in case of empty subtreelet
    if ( nbOccRight==0) return true; //nbOccLeft==0 ||
    
    //addComputedSubTreeletId(treelet,leftTreeletId,nbOccLeft);
    addComputedSubTreeletId(treelet,rightTreeletId,nbOccRight);
    return false;
    }
    
template<typename TreeletType>
void TreeletComputationManagerGeneric<TreeletType>::_computeTreelet(TreeletType &treelet,const TreeletStoreByChildrenIndices* pMapTreeletToId,int index)
    {
    //treelet.ensureUpdatedSeqOfPos(treeletIndexer);
    
    // seqOfTreeletCoverage contains a sequence of pair of treelet and SeqOfPosRepresentation that will be used for computing the new treelet
    // Once it is filled in the following lines, seqOfTreeletCoverage should contain a list of already computed treelets such that their union
    // is equal to the treelet that need to be computed. This imply that the new treelet can be computed by intersecting all the occLists of the
    // treelets in seqOfTreeletCoverage. Since in general there are several seqOfTreeletCoverage that can be used to compute a given treelet, we
    // heuristically try to find one that will result in the least amount of computation.
    vector< std::pair<TreeletType,SeqOfPosRepresentation> > seqOfTreeletCoverage;  //TODO: instantiation inutile optimisable?
    
    //std::cout<<"computedSubtreelets:"<<std::endl;
    //std::cout<<subtreeletsManager<<std::endl;

    // The following try block computes an optimal seqOftreeletCoverage.
    try
        {
        try
            {
            seqOfTreeletCoverage=getOptimalSubtreeletsCoverage(treelet); //,subT_computed[t]);
            }
        catch (CouldNotFindCoverageOfTreeletError ex)
            {
            //cerr<<"trying to add splitted subtreelets at index "<<index<<endl;
            if (pMapTreeletToId!=NULL && index>=0)
                {
                bool hasEmptySubtreelet=updatesubTreeletsWithLeftRightSplittedSubtreelets(treelet,*pMapTreeletToId,index);
                if (hasEmptySubtreelet)
                    {
                    treelet.setNbSentOcc(0);
                    return;
                    }
                seqOfTreeletCoverage=getOptimalSubtreeletsCoverage(treelet);
                }
            else
                {
                throw;
                }
            }
        }
    catch (CouldNotFindCoverageOfTreeletError ex)
        {
        std::cerr<<"Error computing coverage of treelet "<<treelet<<std::endl;
        std::cerr<<"Subtreelets: "<<subtreeletsManager<<std::endl;
        assert(false);
        }
    

    //std::cout<<"Optimal coverage:"<<std::endl;
    //std::cout<<seqOfTreeletCoverage<<std::endl;
    

    // we now intersect the occlist of all treelets in seqOftreeletCoverage
    bool firstLoop=true;
    MatchingsContainerSmartPtr mergedOcc;// TODO: ajouter seertion pour eviter utilisation non initialisee
    typedef std::pair<TreeletType,SeqOfPosRepresentation> TreeletSeqOfPosRepresentationPair;
    BOOST_FOREACH( TreeletSeqOfPosRepresentationPair subTreeletSopMergedPair  ,seqOfTreeletCoverage)
        {
        TreeletType &subTreelet=subTreeletSopMergedPair.first;
        MatchingsContainerSmartPtr occSubtreelet2=computedMatchings.getResult(subTreelet);
        if (firstLoop)
            {
            mergedOcc=occSubtreelet2;
            firstLoop=false;
            continue;
            }
        else
            {
            const SeqOfPosRepresentation &sopMerged=subTreeletSopMergedPair.second;
            const SeqOfPosRepresentation &sop2=subTreelet.getSeqOfPosRepresentation();
            
            MergeConstraints mergeConstraints(preserveOrder);
            mergeConstraints.computeAll(sopMerged,sop2);
            //std::cout<<"merging "<<mergedOcc<<" with "<<*occSubtreelet2<<" constraints:"<<mergeConstraints<<std::endl;
            mergedOcc=mergedOcc->mergeWith(*occSubtreelet2,mergeConstraints);
            //std::cout<<"---->"<<mergedOcc<<std::endl;
            }
        }
    treelet.setNbSentOcc(mergedOcc->getNbSentOcc()); //TODO: check that nbOcc du bon treelet est initialisee...
    computedMatchings.addResult(treelet,mergedOcc);   
    if(mergedOcc->getNbSentOcc()>0)
        {
        assert(treelet.getSize()==mergedOcc->getTypicalSizeOfMapp()-1);
        }
    //TODO: assert t.seqOfPos==newSeqOfPos
    }
    
template<typename TreeletType>
typename TreeletComputationManagerGeneric<TreeletType>::MergeStatus TreeletComputationManagerGeneric<TreeletType>::__tryToMergeSeqOfPos(size_t j, const vector<TreeletId>& orderedList,
                SeqOfPosRepresentation& totalSeqOfPos,const SeqOfPosRepresentation& seqOfPosOfFinalTreelet,
                vector< std::pair<TreeletType,SeqOfPosRepresentation> >& result) const
    {
    TreeletId subTreeletIndex=orderedList[j];
    TreeletType subTreelet =treeletIndexer.getTreelet(subTreeletIndex);
    const SeqOfPosRepresentation &seqOfPos=subTreelet.getSeqOfPosRepresentation();
    if ( seqOfPos.isIncludedIn(totalSeqOfPos) ) // notIncluded(setOfPos,curUnionTreeletSetOfPos) )
        {
        return REJECTED;
        }
    else
        {
        if (totalSeqOfPos.isIncludedIn(seqOfPos) ) //notIncluded(curUnionTreeletSetOfPos,setOfPos)
            {
            //#unlikely or impossible case?
            result.clear();
            result.push_back(std::pair<TreeletType,SeqOfPosRepresentation>(subTreelet,SeqOfPosRepresentation()));//   subTreeletIndex);
            totalSeqOfPos=seqOfPos;
            assert (!totalSeqOfPos.isEquivalentTo(seqOfPosOfFinalTreelet) );
            return MERGED;
            }        
        else
            {
            if (totalSeqOfPos.hasJunctionWith(seqOfPos)) 
                {
                result.push_back( std::pair<TreeletType,SeqOfPosRepresentation>(subTreelet,totalSeqOfPos)  );

                totalSeqOfPos=totalSeqOfPos.merge( seqOfPos );

                if (totalSeqOfPos.isEquivalentTo(seqOfPosOfFinalTreelet))
                    {
                    return FINISHED;
                    }
                return MERGED;
                }
            else
                {
                return NOT_MERGEABLE_NOW;
                /*
                if (firstJumpedPosition==-1)
                    {
                    firstJumpedPosition=j;
                    }
                */
                }//???
            }//???

        }
    }

template<typename TreeletType>
void TreeletComputationManagerGeneric<TreeletType>::__increment_j_to_next_pos(size_t& j,const set<size_t>& iToSkip, size_t strictUpperValue) const
    {
    while (1)
        {//???
        j+=1;
        if ( iToSkip.count(j)>0 )
            {
            continue;
            }
        if (j>=strictUpperValue )
            {
            throw CouldNotFindCoverageOfTreeletError(); //("j:%i>=%i"%(j,len(orderedList)))
            }
        break;
        }//???    
    }

template<typename TreeletType>
vector< std::pair<TreeletType, typename TreeletType::SeqOfPosRepresentation_> > TreeletComputationManagerGeneric<TreeletType>::getOptimalSubtreeletsCoverage(const TreeletType& treelet) const//,listOfSubtreelets,nbOccFunction,convertTreelet):
    /*
    treelet: treelet needing to be computed
    listOfSubtreelets: list of computed subtreelets of treelet
    return: list of subtreelets t1,t2,.... tn such that:
        * Pour tout j: U (i<j) ti inclut strictement dans U(i<=j) ti  [condition2]
        * Pour tout j: U (i<j) ti n'est pas inclut dans tj      [condition1]
        * Ui ti =treelet
        * |occ(ti)|<=|occ(ti+1)|
        * U(i<j) (ti+father) inter (tj) !=empty or U(i<j) (ti) inter (tj+father) !=empty   [condition3]
    */
    {
    
    //std::cout<<"Computing optimal coverage for "<<treelet<<std::endl;
    
    const vector<TreeletId>& orderedList= subtreeletsManager.getSubtreeletsIdOrderedByNbOcc(treelet);
    
    SeqOfPosRepresentation totalSeqOfPos;
    const SeqOfPosRepresentation &seqOfPosOfFinalTreelet=treelet.getSeqOfPosRepresentation();
    
    
    
    //vector<PosType> seqOfPosTotal=treelet.getSeqOfPos();

    vector< std::pair<TreeletType,SeqOfPosRepresentation> > result;
    

    set<size_t> iToSkip;
    // //positionsJumped=set()
    size_t i=0;
    while (1)
        {
        size_t j=i;
        int firstJumpedPosition=-1;
        while (1)
            {
            MergeStatus status=__tryToMergeSeqOfPos(j,orderedList,totalSeqOfPos,seqOfPosOfFinalTreelet,result);
            switch(status)
                {
                case REJECTED:
                    iToSkip.insert(j);
                    break;
                    
                case NOT_MERGEABLE_NOW:
                    if (firstJumpedPosition==-1)
                        {
                        firstJumpedPosition=j;
                        }
                    break;
                    
                case FINISHED:
                    return result;
                    break;
                    
                case MERGED:
                    break;
                }
            
            __increment_j_to_next_pos(j,iToSkip,orderedList.size());

            }
                
                
        if (firstJumpedPosition!=-1)
            {
            i=firstJumpedPosition;
            }
        else
            {
            i=j+1;
            }
            
            
        }
    }
   
//typedef TreeletComputationManagerGeneric<Treelet> TreeletComputationManager;
    
#endif



