/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*! \file
    \ingroup treelet_discovering_group
    \brief Manage the discovering of treelets.
*/

#include <vector>
#include <map>
#include <queue>

#include <boost/foreach.hpp>

#include "TypeDefinitions.hpp"
#include "Tree.hpp"
#include "Treelets.hpp"
#include "DescendingSuperTreeletsManager.hpp"
#include "TreeletComputationManager.hpp"

#ifndef _HPP_TREELET_DISCOVERY_MANAGER_
#define _HPP_TREELET_DISCOVERY_MANAGER_


template<typename TreeletType>
class PriorityQueueTreelet
    {
    typedef typename TreeletType::CompareTreeletsBySize CompareTreeletsBySize;
    std::priority_queue<TreeletType, vector<TreeletType>, CompareTreeletsBySize > priorityQueue;
    std::set<TreeletType> insertedTreeletSet;
    friend std::ostream &operator<< (std::ostream &stream,const PriorityQueueTreelet &queue) {
        stream<<queue.insertedTreeletSet ;
        return stream;
    }
    public:
        void push(const TreeletType& treelet)
            {
            if (insertedTreeletSet.count(treelet)==0)
                {
                priorityQueue.push(treelet);
                insertedTreeletSet.insert(treelet);
                }
            else
                {
                //std::cout<<treelet<<" already inserted"<<std::endl;
                
                }
            }
            
        TreeletType pop()  // probablement tres optimisable (avec smart pointers ?)
            {
            TreeletType result=priorityQueue.top();
            priorityQueue.pop();
            size_t nbErased=insertedTreeletSet.erase(result);
#ifdef NDEBUG            
            (void) nbErased;
#endif
            assert(nbErased==1);
            return result;
            }
            
        bool empty()
            {
            assert ( insertedTreeletSet.empty()==priorityQueue.empty() );
            return priorityQueue.empty();
            }
    };


//typedef   PriorityQueueTreelet;
//typedef std::map<PriorityQueueTreelet> PriorityQueuesByNode;

/*!
 This Class manages the enumeration of the set of treelets of the query tree. When a treelet is discovered, it is put in a priority queue, and later computed through the TreeletComputationManager. The enumeration has to be careful to: 1-not enumerate the same treelet twice, 2-avoid enumerating empty treelets, 3-avoid enumerating dominated treelets.
*/

template<typename TreeletType>
class TreeletDiscoveringManagerGeneric
    {
    typedef TreeletComputationManagerGeneric<TreeletType> TreeletComputationManager;
    typedef DescendingSuperTreeletsManagerGeneric<TreeletType> DescendingSuperTreeletsManager;
    typedef MatchingsContainerStoreGeneric<TreeletType> MatchingsContainerStore;
    typedef typename TreeletType::TreeletSimpleIndexer TreeletSimpleIndexer;
    typedef typename TreeletType::TreeletStoreByChildrenIndices TreeletStoreByChildrenIndices;
    
    const Node* pQueryTree;
    std::map<PosType,PriorityQueueTreelet<TreeletType> > priorityQueuesByNode;
    
    DescendingSuperTreeletsManager descendingSuperTreelets;
    
    TreeletSimpleIndexer treeletIndexer;
    TreeletComputationManager &tcm;
    bool skipTrailingWildcard;
    bool wildcardOnlyAtRoot;
    bool onlyContiguousSiblings;
    
    int maxNbGaps;
    bool reduceDominated;
    public:
        TreeletDiscoveringManagerGeneric(const Node* pQueryTree, TreeletComputationManager& tcm,int nbSiblingsLimit,bool skipTrailingWildcard, bool wildcardOnlyAtRoot, bool onlyContiguousSiblings, int maxNbGaps,bool reduceDominated):pQueryTree(pQueryTree),descendingSuperTreelets(nbSiblingsLimit),tcm(tcm),
            skipTrailingWildcard(skipTrailingWildcard), wildcardOnlyAtRoot(wildcardOnlyAtRoot), onlyContiguousSiblings(onlyContiguousSiblings),
            maxNbGaps(maxNbGaps),reduceDominated(reduceDominated)
            {
            tcm.setIndexer(&treeletIndexer);
            }
            
        const map<TreeletId,vector<TreeletId> >& getDominatedBy() const
            {
            return descendingSuperTreelets.getDominatedBy();
            };
        
        void addDominationInfos(MatchingsContainerStore& occStore) const
            {
            std::map<TreeletType,MatchingsContainerSmartPtr>& treeletToOccMap=occStore.getModifyableUnderlyingMap();
            const map<TreeletId,vector<TreeletId> >& dominatedBy=descendingSuperTreelets.getDominatedBy();
            typedef pair<TreeletId,vector<TreeletId> > DomMapElem;
            BOOST_FOREACH(const DomMapElem& elem,dominatedBy)
                {
                const TreeletType& treelet=treeletIndexer.getTreelet(elem.first);
                assert(treeletToOccMap.count(treelet)==1);
                // const_cast OK here since we don't modify the values
                // influencing the ordering of treelet (?)
                TreeletType& tRef=const_cast<TreeletType& > (treeletToOccMap.find(treelet)->first);
                tRef.setDominated(true);
                }
            }
        
        void computeAllTreelets();
        void updateDiscoveredAndSuperTreelesAndTCM(TreeletType& treelet);
        void updateDiscoveredAndSuperTreelesAndTCMLatteralStrategy(TreeletType & treelet,int curIndex,
        
        //map< vector<TreeletId>, TreeletId> &mapTreeletToId,
        TreeletStoreByChildrenIndices &mapTreeletToId,
vector<TreeletType>& listOfComputedNonEmptyTreelets,map<TreeletId,int>& mapTreeletIdToCurIndex);
        void computeAllTreeletsLatteralStrategy();
        void initForCurrentIndex(int curLevel,const Node * pRootNodeInQueryTree, PriorityQueueTreelet<TreeletType>& discovered);
        
    };
    
template<typename TreeletType>
void TreeletDiscoveringManagerGeneric<TreeletType>::computeAllTreelets()
    {
    computeAllTreeletsLatteralStrategy();
    return;
/*    
    BOOST_FOREACH(const Node *pNode,pQueryTree->getNodePointersListByPostOrderTraversal())
        {
        //std::cout<<"Processing "<<*pNode<<std::endl;
        
        
        
        Treelet treeletBase(pNode);
        //std::cout<<"TreeletBase: "<< treeletBase <<std::endl;
        
        
        PosType positionOfNode=pNode->getPosition();
        descendingSuperTreelets.initializeNodeToBaseForThisPosition(positionOfNode);
        
        PriorityQueueTreelet &discovered=priorityQueuesByNode[positionOfNode];
        //std::cout<<"discovered avant treeletbase:"<< discovered <<std::endl;
        discovered.push(treeletBase);
        
        //std::cout<<"Initial discovered:"<< discovered <<std::endl;
        while ( !discovered.empty() )
            {
            Treelet treelet=discovered.pop();
            //std::cout<<"Poping "<<treelet<<" from "<< discovered <<std::endl;
            treelet.buildSeqOfPos(treeletIndexer); //TODO: optimisable
            tcm.computeTreelet(treelet);
            //std::cout<<"result: "<<treelet<< " has: "<<treelet.getNbSentOcc()<<" sent-occurences"<<std::endl;
            updateDiscoveredAndSuperTreelesAndTCM(treelet);
            //std::cout<<"----------- finished processing "<<treelet<<std::endl;   
            }
            //Cleaning of dominated treelets
        //std::cout<<"treeletIndexer"<<std::endl;
        //std::cout<<treeletIndexer<<std::endl;
        descendingSuperTreelets.reduceDescendingSuperTreeletsList(positionOfNode,tcm);  
        }
    */   
    }

template<typename TreeletType>
void TreeletDiscoveringManagerGeneric<TreeletType>::initForCurrentIndex(int curLevel,const Node * pRootNodeInQueryTree, PriorityQueueTreelet<TreeletType>& discovered)
    {
    if (curLevel==-1)
        {
        TreeletType treeletBase(pRootNodeInQueryTree);
        discovered.push(treeletBase);
        return;
        }
    assert(curLevel>=0);
    
    PosType posCurLevel=pRootNodeInQueryTree->getChildNodeByNum(curLevel)->getPosition();
    const vector<TreeletId>& topRightTreeletId=descendingSuperTreelets.getBaseTreeletTable(posCurLevel);
    //cout<<"topRightTreeletId@"<< ((unsigned int)posCurLevel)<< ": "<<topRightTreeletId<<endl;
    vector<TreeletId> leftTreeletIdList=descendingSuperTreelets.getBaseTreeletTable(pRootNodeInQueryTree->getPosition());//descendingSuperTreelets.getListOfDescendingIds(pRootNodeInQueryTree->getPosition());
    BOOST_FOREACH(TreeletId treeletId,topRightTreeletId)
        {

        //int sizeOfRightTreelet=descendingSuperTreelets.getSizeOfTreelet(treeletId);
        //Treelet treeletStart(pRootNodeInQueryTree,pRootNodeInQueryTree->getNbOfChildren(),curLevel,treeletId,sizeOfRightTreelet);
        //discovered.push(treeletStart);
        
        BOOST_FOREACH (TreeletId leftTreeletId,leftTreeletIdList)
            {
            const TreeletType & leftTreelet=treeletIndexer.getTreelet(leftTreeletId);
            
            //Treelet superTreelet=leftTreelet.getExtendedDownwardTreelet(curLevel,treeletId,sizeOfRightTreelet);
            TreeletType superTreelet=leftTreelet.getExtendedDownwardTreelet(curLevel,treeletId,descendingSuperTreelets.getTreeletSizeInfos());
            
            if (maxNbGaps>=0 && superTreelet.nbGaps()>maxNbGaps)
                {
                continue;
                }
                
            if (leftTreelet.getNbSentOcc()>0)
                {
                discovered.push(superTreelet);
                //descendingSuperTreelets.addAscendingSubtreeletRelationship( leftTreeletId,superTreelet );
                
                //tcm.addComputedSubTreeletId(superTreelet,leftTreeletId, leftTreelet.getNbSentOcc());
                
                
                tcm.addComputedSubTreeletId(superTreelet,treeletId, treeletIndexer.getTreelet(treeletId).getNbSentOcc());
                }
            else
                {
                tcm.rememberTreeletIsEmpty(superTreelet);
                }
            }
        }

    }
    
template<typename TreeletType>
void TreeletDiscoveringManagerGeneric<TreeletType>::computeAllTreeletsLatteralStrategy()
    {
    BOOST_FOREACH(const Node *pNode,pQueryTree->getNodePointersListByPostOrderTraversal())
        {

        PosType positionOfNode=pNode->getPosition();
        descendingSuperTreelets.initializeNodeToBaseForThisPosition(positionOfNode);
        
        PriorityQueueTreelet<TreeletType> &discovered=priorityQueuesByNode[positionOfNode];
        //std::cerr<<"processing node@"<<((unsigned int)positionOfNode)<<std::endl;
        int nbTopChildren=pNode->getNbOfChildren();
        //map< vector<TreeletId>, TreeletId> mapTreeletToId;
        TreeletStoreByChildrenIndices mapTreeletToId;
        map<TreeletId,int> mapTreeletIdToCurIndex;
        TreeletId curWildcard=TREELET_ID_SPECIAL_VALUE;
        
        std::set<TreeletId> list_of_treelets_with_wildcard_roots;
        
        for (int curIndex=-1; curIndex<nbTopChildren; curIndex++)
            {
            //std::cerr<<"curIndex:"<<curIndex<<std::endl;
            //discovered.push(treeletBase);
            initForCurrentIndex(curIndex,pNode,discovered);
            vector<TreeletType> listOfComputedNonEmptyTreelets;
            while ( !discovered.empty() )
                {
                //cerr<<treeletIndexer<<endl;
                //std::cout<<"disc:"<<discovered<<std::endl;
                TreeletType treelet = discovered.pop();
                
                if (TreeletType::isUsingWildcard() && onlyContiguousSiblings && 
                    treelet.isRootWildcard() && !treelet.hasContiguousChildrenBeforeIndex(curIndex+1)) {
                    continue;
                }
                
                if (descendingSuperTreelets.hasInitialRightPart(treelet,curIndex)) 
                    {
                    TreeletId leftTreeletId=mapTreeletToId.getLeftTreeletId(treelet,curIndex); //optimisable?
                    size_t nbOccLeft=treeletIndexer.getTreelet(leftTreeletId).getNbSentOcc();
                    
                    if (nbOccLeft>0)
                        {
                        descendingSuperTreelets.addAscendingSubtreeletRelationship( leftTreeletId,treelet );

                        tcm.addComputedSubTreeletId(treelet,leftTreeletId, nbOccLeft);                         
                        }
                    else
                        {
                        tcm.rememberTreeletIsEmpty(treelet);
                        }

                    }
                //std::cout<<"poped "<<treelet<<std::endl;
                treelet.buildSeqOfPos(treeletIndexer); //TODO: optimisable
                //cerr<<"curIndex beforecall to compute:"<<curIndex<<endl;
                tcm.computeTreelet(treelet,&mapTreeletToId,curIndex);
                updateDiscoveredAndSuperTreelesAndTCMLatteralStrategy(treelet,curIndex,mapTreeletToId,listOfComputedNonEmptyTreelets,mapTreeletIdToCurIndex);
                
                if (TreeletType::isUsingWildcard() && curIndex==-1 && treelet.isElementaryWildcard())
                    {
                    assert(curWildcard==TREELET_ID_SPECIAL_VALUE);
                    curWildcard=treelet.getId();
                    }
                    
                if (TreeletType::isUsingWildcard() && wildcardOnlyAtRoot && treelet.hasGaps())
                    {
                    try{
                        TreeletId tId = treelet.getId();
                        assert(list_of_treelets_with_wildcard_roots.count(tId) == 0);
                        list_of_treelets_with_wildcard_roots.insert(tId);
                    }
                    catch (const TreeletBase::UsingNonSetValueInTreeletException& ){
                        // the treelet was empty
                    }   
                    }
                    
                }
                                
                //Cleaning of dominated treelets
            BOOST_FOREACH(const TreeletType &treelet,listOfComputedNonEmptyTreelets)
                {
                descendingSuperTreelets.pushCompletedTreelet(treelet);
                }
            
            // We can do a full reduction either when the node is fully processed
            // (because there is no more any child node to the right that can match a reduced child)
            // or when the left children have just all be fully processed
            // (because no child to the right can have a similar match to any child to the right, even with wildcard)
            if (reduceDominated)
                {
                if (curIndex==(int)nbTopChildren-1||curIndex==(int)(pNode->getNbOfChildrenLeft())-1)
                    {
                    descendingSuperTreelets.reduceDescendingSuperTreeletsList(positionOfNode,tcm);
                    }
                else
                    {
                    descendingSuperTreelets.reduceDescendingSuperTreeletsList(positionOfNode,tcm,curIndex,&mapTreeletIdToCurIndex);
                    }
                }
            }
        if (TreeletType::isUsingWildcard() && skipTrailingWildcard)
            {
            assert(curWildcard!=TREELET_ID_SPECIAL_VALUE);
            descendingSuperTreelets.removeTopElementaryWildcard(positionOfNode, curWildcard);
            }
            
        if (TreeletType::isUsingWildcard() && wildcardOnlyAtRoot)
            descendingSuperTreelets.removeWildcardRootedTreeletsFromNodeToBaseTable(
                positionOfNode, list_of_treelets_with_wildcard_roots);
        }
    //cerr<<treeletIndexer<<endl;
    }

template<typename TreeletType>
void TreeletDiscoveringManagerGeneric<TreeletType>::updateDiscoveredAndSuperTreelesAndTCM(TreeletType& treelet)
    {
    size_t nbOcc=treelet.getNbSentOcc();
    if (nbOcc>0)
        {
        TreeletId idTreelet=treeletIndexer.addTreeletAndGetId(treelet);
        treelet.setId(idTreelet);
        BOOST_FOREACH (const TreeletType &superTreelet, descendingSuperTreelets.computeListOfSuperTreelets(treelet,true))
            {
            //std::cout<<treelet<<" has occ>0 and added to computedSubtreelets of "<<superTreelet<<" as "<<idTreelet<<std::endl;
            tcm.addComputedSubTreeletId(superTreelet,idTreelet, nbOcc); //TODO: check that it is OK to move this inside the if ...
            
            if (superTreelet.wasGeneratedAsDescendingSuperTreelet() )
                {
                descendingSuperTreelets.addAscendingSubtreeletRelationship( idTreelet,superTreelet );
                priorityQueuesByNode[superTreelet.getRootPos()].push(superTreelet);
                }
            }
        descendingSuperTreelets.pushCompletedTreelet(treelet);
        }
    else
        {
        // ascending supertreelet should not need to be marked as empty, as it should not be reachable by the discovering process
        BOOST_FOREACH (const TreeletType &superTreelet, descendingSuperTreelets.computeListOfSuperTreelets(treelet,false))
            {
            //std::cout<<"remember is empty: "<<superTreelet<<std::endl;
            tcm.rememberTreeletIsEmpty(superTreelet);
            }        
        } 
    }

/*!
 \brief define < so that vector<TreeletId> can be keys in a map
*/
bool operator<(const vector<TreeletId>& v1,const vector<TreeletId>& v2)
    {
    if (v1.size()!=v2.size())
        {
        throw logic_error("Compare Error") ;
        }
    else
        {
        return std::lexicographical_compare(v1.begin(),v1.end(),
                                v2.begin(),v2.end());
        }
    }
    
template<typename TreeletType>
void TreeletDiscoveringManagerGeneric<TreeletType>::updateDiscoveredAndSuperTreelesAndTCMLatteralStrategy(TreeletType & treelet,int curIndex,
    //map< vector<TreeletId>, TreeletId> &mapTreeletToId,
    TreeletStoreByChildrenIndices &mapTreeletToId,
vector<TreeletType>& listOfComputedNonEmptyTreelets,map<TreeletId,int>& mapTreeletIdToCurIndex)
    {
    size_t nbOcc=treelet.getNbSentOcc();
    
    // if (curIndex==-1)
        // {
        // if (nbOcc>0)
            // {
            // TreeletId idTreelet=treeletIndexer.addTreeletAndGetId(treelet);
            // treelet.setId(idTreelet);
            // assert (mapTreeletToId.count(treelet.getChildrenVector())==0);
            // mapTreeletToId[treelet.getChildrenIndices()]=idTreelet;
            // BOOST_FOREACH (const Treelet &superTreelet, 
                    // descendingSuperTreelets.computeListOfSuperTreeletsLatteralStrategy(
                        // treelet,true,curIndex,mapTreeletToId,treeletIndexer))
                // {
                // tcm.addComputedSubTreeletId(superTreelet,idTreelet, nbOcc); 
                // }
        // return;
        // }
    // assert(curLevel>=0);
    if (nbOcc>0)
        {
        TreeletId idTreelet=treeletIndexer.addTreeletAndGetId(treelet);
        treelet.setId(idTreelet);
        mapTreeletIdToCurIndex[idTreelet]=curIndex;
        mapTreeletToId.addTreelet(treelet,idTreelet);
        //assert (mapTreeletToId.count(treelet.getChildrenIndices())==0);
        //mapTreeletToId[treelet.getChildrenIndices()]=idTreelet;
        
        BOOST_FOREACH (const TreeletType &superTreelet, descendingSuperTreelets.computeListOfSuperTreeletsLatteralStrategy(treelet,true,curIndex,mapTreeletToId,treeletIndexer,maxNbGaps))
            {
            //std::cerr<<"ST: "<<superTreelet<<endl;
            //std::cout<<treelet<<" has occ>0 and added to computedSubtreelets of "<<superTreelet<<" as "<<idTreelet<<std::endl;
            tcm.addComputedSubTreeletId(superTreelet,idTreelet, nbOcc); //TODO: check that it is OK to move this inside the if ...
            
            if (superTreelet.wasGeneratedAsDescendingSuperTreelet() )
                {
                descendingSuperTreelets.addAscendingSubtreeletRelationship( idTreelet,superTreelet );
                priorityQueuesByNode[superTreelet.getRootPos()].push(superTreelet);
                }
            }
        listOfComputedNonEmptyTreelets.push_back(treelet);
        //descendingSuperTreelets.pushCompletedTreelet(treelet);
        }
    else
        {
        // ascending supertreelet should not need to be marked as empty, as it should not be reachable by the discovering process
        BOOST_FOREACH (const TreeletType &superTreelet, descendingSuperTreelets.computeListOfSuperTreeletsLatteralStrategy(treelet,false,curIndex,mapTreeletToId,treeletIndexer,maxNbGaps))
            {
            //std::cout<<"remember is empty: "<<superTreelet<<std::endl;
            tcm.rememberTreeletIsEmpty(superTreelet);
            }        
        } 
    }
    
//typedef TreeletDiscoveringManagerGeneric<Treelet> TreeletDiscoveringManager;
    
#endif



