/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "TRClientServerCommonValues.hpp"
#include "utf8cpp/utf8.h" 
#include "TypeDefinitions.hpp"
#include "GlobalTreeletRetrieving2.hpp"

#include "StringConversions.hpp"


#ifndef __TREELET_RETRIEVER_NEW_PROTOCOL_HPP__
#define __TREELET_RETRIEVER_NEW_PROTOCOL_HPP__

void printPosListToStreamAsString(std::ostream& output, const std::vector<PosType>& posList,const std::string& delim )
    {
    for (size_t i=0; i<posList.size();++i)
        {
        if (i>0)
            {
            output<<delim;
            }
        output<< (int) (posList[i]);
        }
    }

void printPosSetToStreamAsString(std::ostream& output, const std::set<PosType>& posSet,const std::string& delim )
    {
    bool first=true;
    BOOST_FOREACH(PosType pos, posSet)
        {
        if (first)
            {
            first=false;
            }
        else
            {
            output<<delim;
            }
        output<< (int) pos;
        }
    }
 
class SimpleTextSerializationOverStreamProcessor: public IndividualResultsProcessor
    {
    protected:
    std::ostream& stream;
    size_t nbMatchingsForCurrentTreelet;
    size_t nbTreelets;
    
    const std::vector<std::string>& numToTreeIdVector;
    
    public:
    SimpleTextSerializationOverStreamProcessor(std::ostream& stream,
    const std::vector<std::string>& numToTreeIdVector):stream(stream),nbMatchingsForCurrentTreelet(0),nbTreelets(0),numToTreeIdVector(numToTreeIdVector)
                        {}
    
    void initializeForThisTreelet(const Treelet &treelet)
        {
        //std::cerr<<"initialize treelet "<<treelet<<std::endl;
        //std::set<PosType> seqOfPos=treelet.getSeqOfPosRepresentation().getIterableSeqOfPos();
        //printPosSetToStreamAsString(stream, seqOfPos," "); 
        
        treelet.getSeqOfPosRepresentation().printReprAsStringToStream(stream);       
        stream<<" : ";       
        nbTreelets+=1;
        nbMatchingsForCurrentTreelet=0;
        }
    
    virtual void addAdditionalInfos(TreeIdType numTree,PrefixEncodeVectors& encoder)
        {}
    
    void processOneTreelet(TreeIdType numTree,
                            const PosList& mapp,
                            const ProjectedMappingInfos* pProjInfo,
                            size_t count_emitted,
                            MatchingsContainerSmartPtr occ,
                            size_t j,size_t numMapping)
        {
        
        stream<<numToTreeIdVector.at(numTree)<<"@<";
        printPosListToStreamAsString(stream, mapp," ");
        stream<<"> ";
       
        nbMatchingsForCurrentTreelet+=1;
        //std::cerr<<"<< processOneTreelet"<<std::endl;
        }
        
    void finalizeTreelet()
        {
        stream<<endl;
        }
        
    void finalizeAll()
        {
        stream<<"EOS"<<endl;
        }
    };


class SerializationOverStreamProcessor: public IndividualResultsProcessor
    {
    protected:
    std::ostream& stream;
    size_t nbMatchingsForCurrentTreelet;
    size_t nbTreelets;
    public:
    SerializationOverStreamProcessor(std::ostream& stream):stream(stream),nbMatchingsForCurrentTreelet(0),nbTreelets(0)
                        {}
    
    void initializeForThisTreelet(const Treelet &treelet)
        {
        //std::cerr<<"initialize treelet "<<treelet<<std::endl;
        const std::set<PosType>& seqOfPos=treelet.getSeqOfPosRepresentation().getIterableSeqOfPos();
        std::string code;
        PrefixEncodeVectors encoder(code);
        encoder.packPosSetUC(seqOfPos);
        
        if (isUsingWildcard())
            {
            const std::set<PosType>& seqOfPosNW=treelet.getSeqOfPosRepresentation().getIterableSeqOfPosNonWildcard();
            //cerr<<"adding "<<seqOfPosNW<<endl;
            encoder.packPosSetUC(seqOfPosNW);
            }
        //cerr<<"code: "<<code<<endl;
        unsigned char bufferMessageLength[nbBytesForEncodingMessageLength];
        encodeMessageLength(code.size(),bufferMessageLength);
        stream.write((char *)bufferMessageLength,nbBytesForEncodingMessageLength);
        stream.write(code.data(),code.size());
        
        nbTreelets+=1;
        nbMatchingsForCurrentTreelet=0;
        }
    
    virtual void addAdditionalInfos(TreeIdType numTree,PrefixEncodeVectors& encoder)
        {}
    
    void processOneTreelet(TreeIdType numTree,
                            const PosList& mapp,
                            const ProjectedMappingInfos* pProjInfo,
                            size_t count_emitted,
                            MatchingsContainerSmartPtr occ,
                            size_t j,size_t numMapping)
        {
        if (pProjInfo==NULL)
            {
            throw std::exception();
            }
        //std::cerr<<">> processOneTreelet "<<numTree<<std::endl;
        std::string code;
        PrefixEncodeVectors encoder(code);
        encoder.packUInt(numTree);
        encoder.packPosListUC(mapp);
        encoder.packPosSetUC( *(pProjInfo->pConnectedImage) );
        encoder.packPosListUC(pProjInfo->additionalNodes);
        encoder.packMatchListUC(pProjInfo->matchList);
        
        addAdditionalInfos(numTree,encoder);
        
        
        unsigned char bufferMessageLength[nbBytesForEncodingMessageLength];
        encodeMessageLength(code.size(),bufferMessageLength);
        stream.write((char *)bufferMessageLength,nbBytesForEncodingMessageLength);
        stream.write(code.data(),code.size());
        
        nbMatchingsForCurrentTreelet+=1;
        
        //std::cerr<<"<< processOneTreelet"<<std::endl;
        
        }
        
    void finalizeTreelet()
        {
        //std::cerr<<">> finalizeTreelet"<<std::endl;
        unsigned char buffer[nbBytesForEncodingMessageLength];
        encodeMessageLength(0,buffer); // no more matchings
        stream.write((char *)buffer,nbBytesForEncodingMessageLength);
        
        encodeMessageLength(nbMatchingsForCurrentTreelet,buffer);
        stream.write((char *)buffer,nbBytesForEncodingMessageLength);
        //std::cerr<<"<< finalizeTreelet"<<std::endl;
        }
        
    void finalizeAll()
        {
        //std::cerr<<">> finalizeAll"<<std::endl;
        unsigned char buffer[nbBytesForEncodingMessageLength];
        
        encodeMessageLength(0,buffer);  //no more treelets
        stream.write((char *)buffer,nbBytesForEncodingMessageLength);
        
        encodeMessageLength(nbTreelets,buffer);
        stream.write((char *)buffer,nbBytesForEncodingMessageLength);
        //std::cerr<<"<< finalizeAll"<<std::endl;    
        }
    };

class SerializationOverStreamProcessorWithAddInfos: public SerializationOverStreamProcessor
    {
    protected:
    const AddInfosVectorInterface& aDataVect;
    const TargetList& targetInfoList;
    public:
    SerializationOverStreamProcessorWithAddInfos(
                        std::ostream &stream,
                        const AddInfosVectorInterface& aDataVect,
                        const TargetList& targetInfoList):
                        SerializationOverStreamProcessor(stream),aDataVect(aDataVect),targetInfoList(targetInfoList){}
    
    void addAdditionalInfos(TreeIdType numTree,PrefixEncodeVectors& encoder)
        {
        const TargetInfo& targetInfo=targetInfoList[numTree];
        encoder.packPosListUC(targetInfo.depend);
        encoder.packMatchListUC( targetInfo.getMatchingInfosAsVector() ) ;
        
        encoder.addRawDataToCode(aDataVect.getRawData(numTree));
        }
    
    };

    
// class HandlingToStreamInterface
    // {
    // public:
        // virtual void handleRequestAsStringToStream(const std::string& labelListAsString, const std::string& dependListAsString,std::ostream &stream)=0;
    // }
    
class TreeletRetrieverNewProtocol: public TreeletRetrieverWithContinuousTargetSide

    {
    public:
    TreeletRetrieverNewProtocol(const std::string& datapath, const std::string& side, 
                        //size_t maxNumberOfSamplePerTreelet, 
                        //bool preserveOrder, 
                        //int nbSiblingsLimit,
                                //bool filterOutDominated,
                                //bool skipTrailingWildcards,
                                //bool keepElementaryWildcard,
                                //bool keepTopWilcardChain
                        const SearchParameters& sParams,
                        const PostProcessingParameters& ppParams
                                ):
                                
          TreeletRetrieverWithContinuousTargetSide(datapath.c_str(),side.c_str(),
          //maxNumberOfSamplePerTreelet,
          //preserveOrder,nbSiblingsLimit,//filterOutDominated,
          //skipTrailingWildcards,
          //keepElementaryWildcard,keepTopWilcardChain
          sParams,
          ppParams
          ){}

    void handleRequestAsStringAndUseNewProtocol(const std::string& labelListAsString, const std::string& dependListAsString,std::ostream &stream)
        {
        SerializationOverStreamProcessor processor(stream);
        handleRequestAsStringWithGivenProcessor(labelListAsString, dependListAsString, processor);
        }

    void handleRequestAsStringAndUseSimpleTextProtocol(const std::string& labelListAsString, const std::string& dependListAsString,std::ostream &stream)
        {
        SimpleTextSerializationOverStreamProcessor processor(stream,
        postProcessingInfos.getNumToTreeIdVector());
        handleRequestAsStringWithGivenProcessor(labelListAsString, dependListAsString, processor);
        }
        
   void handleRequestAsStringAndUseNewProtocolWithAddInfos(const std::string& labelListAsString, const std::string& dependListAsString,std::ostream &stream)
        {        
        SerializationOverStreamProcessorWithAddInfos processor(stream,
                     postProcessingInfos.getAdditionalInfos(),
                     postProcessingInfos.getTargetInfos());
                                
        handleRequestAsStringWithGivenProcessor(labelListAsString, dependListAsString, processor);

        }                         
    };

#endif //__TREELET_RETRIEVER_NEW_PROTOCOL_HPP__
