/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*! \file
 \ingroup tree_data_struct_groups
 \brief Description of the Treelet Class.
 */

#include <vector>
#include <algorithm>

#include "TypeDefinitions.hpp"
#include "Tree.hpp"

#include "SeqOfPosRepresentationBase.hpp"
#include <boost/utility.hpp>

#ifndef _HPP_MY_TREELETS_
#define _HPP_MY_TREELETS_

namespace TreeletBase {

int getSafeNbSiblingsForGivenNodeWidth(int n, int safety) {
    if (safety <= 1) {
        if (n <= 17)
            return n;
        if (n <= 18)
            return 8;
        if (n <= 20)
            return 7;
        if (n <= 23)
            return 6;
        if (n <= 29)
            return 5;
        if (n <= 44)
            return 4;
        if (n <= 96)
            return 3;
        return 2;
    } else {
        if (n <= 10)
            return n;
        if (n <= 11)
            return 7;
        if (n <= 12)
            return 6;
        if (n <= 13)
            return 5;
        if (n <= 14)
            return 4;
        if (n <= 15)
            return 3;
        if (n <= 16)
            return 2;
        return 1;
    }
}

class UsingNonSetValueInTreeletException: public std::exception {
    virtual const char* what() const throw () {
        return "UsingNonSetValueInTreeletException";
    }
};

// class Treelet;

// void setIdOfTreelet(Treelet &treelet,TreeletId id);

class DescendingSuperTreeletsNotFoundException: public std::exception {
    virtual const char* what() const throw () {
        return "Descending SuperTreelet Not Found";
    }
};

// size_t getSizeOfTreeletFunction(const Treelet&); //to solve forwarding problems

/*
 class TreeletWithWildcard: public Treelet
 {


 virtual void extendChildrenDownwardInPlace(size_t childNumber, TreeletId newChildId, size_t sizeOfNewChild) =0
 {
 //std::cout<<"extending downward "<<*this<<" to ";
 //Treelet result(*this);
 assert( (int)size - sizeOfChildrenVector[childNumber] +  sizeOfNewChild >=0 );
 size=size - sizeOfChildrenVector[childNumber] +  sizeOfNewChild ;
 sizeOfChildrenVector[childNumber]=sizeOfNewChild;
 recursiveRepr.replaceChild(childNumber,newChildId);

 seqOfPosIsConsistent=false;
 nbSentOccIsSet=false;
 treeletIdIsSet=false;
 nbSentOcc=0;
 index=0;
 isElementaryTreeletFlag=false;
 //std::cout<<*this<<std::endl;
 }

 void buildSeqOfPos	(	const TreeletSimpleIndexer & 	treeletIndexer	)
 {
 //std::cout<<"building seqOfPos for "<<*this<<std::endl;
 SeqOfPosRepresentation seqOfPos(recursiveRepr.getRootPointer()->getPosition(),treelet.rootIsStar());//X

 if ( recursiveRepr.getRootPointer()->hasParent() )
 {
 seqOfPos.addParent( recursiveRepr.getParentNode()->getPosition() );
 }
 BOOST_FOREACH(TreeletId treeletId, recursiveRepr.getChildrenIndices())
 {
 if (treeletId==TREELET_ID_SPECIAL_VALUE) continue;
 const SeqOfPosRepresentation& seqOfPosChild= treeletIndexer.getTreelet(treeletId).getSeqOfPosRepresentation();
 seqOfPos.mergeWith(seqOfPosChild);
 }
 this->seqOfPos=seqOfPos;
 seqOfPosIsConsistent=true;
 }

 };


 class TreeletWithWildcardAsRecursiveRepresentation: public TreeletAsRecursiveRepresentation
 {
 bool isRootWildcard;
 public:

 TreeletWithWildcardAsRecursiveRepresentation(const Node *pRootNodeInQueryTree,std::vector<TreeletId> childrenIndices,bool rootIsWildcard):
 pRootNodeInQueryTree(pRootNodeInQueryTree),childrenIndices(childrenIndices),isRootWildcard(rootIsWildcard) {};

 // Constructor for elementary treelets:
 TreeletWithWildcardAsRecursiveRepresentation(const Node *pRootNodeInQueryTree,size_t nbChildren,bool rootIsWildcard):
 TreeletAsRecursiveRepresentation(pRootNodeInQueryTree,nbChildren),
 isRootWildcard(rootIsWildcard){};
 }
 */
class TreeletAsRecursiveRepresentation {
protected:
    const Node *pRootNodeInQueryTree;
    std::vector<TreeletId> childrenIndices;

    friend std::ostream &operator<<(std::ostream &stream,
            const TreeletAsRecursiveRepresentation &treeletRecRepr);

public:

    TreeletAsRecursiveRepresentation(const Node *pRootNodeInQueryTree, std::vector<TreeletId> childrenIndices) :
            pRootNodeInQueryTree(pRootNodeInQueryTree), childrenIndices(childrenIndices) {
    }
    ;

    // Constructor for elementary treelets:            
    TreeletAsRecursiveRepresentation(const Node *pRootNodeInQueryTree, size_t nbChildren) :
            pRootNodeInQueryTree(pRootNodeInQueryTree), childrenIndices(nbChildren, TREELET_ID_SPECIAL_VALUE) {
    }
    ;

    bool hasContiguousChildrenBeforeIndex(int index) const {
        bool in_children_sequence = false;
        assert(index <= (int ) childrenIndices.size());
        for (int i = 0; i < index; i++) {
            TreeletId id = childrenIndices[i];
            if (id == TREELET_ID_SPECIAL_VALUE) {
                if (in_children_sequence)
                    return false;
            } else {
                assert(!in_children_sequence);
                in_children_sequence = true;
            }
        }
        return true;
    }

    size_t getNbChildrenLeft() const {
        return pRootNodeInQueryTree->getNbOfChildrenLeft();
    }

    // return the number of (non-empty) children of the root of the treelet
    unsigned int getNbTopSiblings() const {
        unsigned int nbTopSiblings = 0;
        BOOST_FOREACH(TreeletId tId,childrenIndices) {
            if (tId != TREELET_ID_SPECIAL_VALUE) {
                nbTopSiblings++;
            }
        }
        return nbTopSiblings;
    }

    // return the top sibling if there is only one.
    // return TREELET_ID_SPECIAL_VALUE otherwise (no top sibing or more than 1)
    TreeletId getTopSibling(bool& isOnLeft) const {
        bool foundTopSiblings = false;
        TreeletId childTreeletId = TREELET_ID_SPECIAL_VALUE;
        size_t indexOfTopChild = 0;

        for (size_t i = 0; i < childrenIndices.size(); ++i) {
            TreeletId tId = childrenIndices[i];
            if (tId != TREELET_ID_SPECIAL_VALUE) {
                if (foundTopSiblings)
                    return TREELET_ID_SPECIAL_VALUE; // more than one child
                foundTopSiblings = true;
                childTreeletId = tId;
                indexOfTopChild = i;
            }
        }

        if (!foundTopSiblings)
            return TREELET_ID_SPECIAL_VALUE; // No child

        //const Node *rootNode = pRootNodeInQueryTree;
        size_t leftRightLimit = pRootNodeInQueryTree->getNbOfChildrenLeft();
        isOnLeft = indexOfTopChild < leftRightLimit;
        // cout << "getTopSibling indexOfTopChild= " << indexOfTopChild 
        // << " leftRightLimit= " << leftRightLimit
        // << " childOnLeft= " << isOnLeft
        // << endl;
        return childTreeletId;
    }

    int getNodeWidth() const {
        return childrenIndices.size();
    }

    bool hasStrictlyMoreThanNTopSiblings(int n_) const {
        assert(n_ >= 0);
        unsigned int n = (unsigned int) n_;
        if (childrenIndices.size() <= n) {
            return false;
        }
        unsigned int nbTopSiblings = getNbTopSiblings();
        return nbTopSiblings > n;
    }

    inline bool operator<(const TreeletAsRecursiveRepresentation &other) const {
        if (pRootNodeInQueryTree != other.pRootNodeInQueryTree) //Warning: nodes are compared by their memory adress: potentially dangerous...
                {
            return pRootNodeInQueryTree < other.pRootNodeInQueryTree;
        } else {
            if (childrenIndices.size() != other.childrenIndices.size()) {
                return childrenIndices.size() < other.childrenIndices.size();
            } else {
                return std::lexicographical_compare(childrenIndices.begin(), childrenIndices.end(),
                        other.childrenIndices.begin(), other.childrenIndices.end());
            }
        }
    }
    ;

    const std::vector<TreeletId>& getChildrenIndices() const {
        return childrenIndices;
    }

    void replaceChild(size_t childNumber, TreeletId newChildId) {
        childrenIndices[childNumber] = newChildId;

    }
    ;

    bool hasUpwardExtension() const {
        return pRootNodeInQueryTree->hasParent();
    }
    ;

    const Node *getParentNode() const {
        return pRootNodeInQueryTree->getParent();
    }
    ;

    size_t getNbOfChildren() const {
        return childrenIndices.size();
    }
    ;

    TreeletId getChildIndex(size_t num) const {
        return childrenIndices[num]; // .getChildIndex(num);
    }
    PosType getPosNodeAtChildNum(size_t num) const {
        return pRootNodeInQueryTree->getChildNodeByNum(num)->getPosition();
    }

    const Node *getRootPointer() const {
        return pRootNodeInQueryTree;
    }

    LabelType getRootLabel() const {
        return pRootNodeInQueryTree->getLabel();
    }

};

std::ostream &operator<<(std::ostream &stream, const TreeletAsRecursiveRepresentation &treeletRecRepr) {

    stream << "t" << "[(" << (int) treeletRecRepr.pRootNodeInQueryTree->getPosition() << ") ";

    bool first = true;
    BOOST_FOREACH(TreeletId id, treeletRecRepr.childrenIndices) {
        if (first) {
            first = false;
        } else {
            stream << " ";
        }
        if (id == TREELET_ID_SPECIAL_VALUE) {
            stream << "N";
        } else {
            stream << id;
        }
    }
    stream << "]";
    return stream;
}

class TreeletBaseClass {
protected:
    TreeletId index;
    size_t size;
    size_t nbSentOcc;
    bool nbSentOccIsSet;
    bool treeletIdIsSet;
    bool wasGeneratedAsDescendingSuperTreeletFlag;

    bool isElementaryTreeletFlag;

    std::string typeRepresentation;
    bool isTypeRepresentationComputed;

    bool seqOfPosIsConsistent;

    bool isDominatedFlag;

public:

    TreeletBaseClass(size_t size, bool isElementaryTreeletFlag) :
            size(size), nbSentOccIsSet(false), treeletIdIsSet(false), wasGeneratedAsDescendingSuperTreeletFlag(
                    false), isElementaryTreeletFlag(isElementaryTreeletFlag),

            isTypeRepresentationComputed(false), seqOfPosIsConsistent(false), isDominatedFlag(false) {
    }

    bool isDominated() const {
        return isDominatedFlag;
    }

    void setDominated(bool flag) {
        isDominatedFlag = flag;
    }

    bool isTreeletIdSet() const {
        return treeletIdIsSet;
    }

    bool isElementaryTreelet() const {
        return isElementaryTreeletFlag;
    }

    bool isElementaryWildcard() const {
        return false;
    }
    ;
    bool isDescendingSupertreesRoot() const {
        return isElementaryTreelet();
    }
    void setGeneratedAsDescendingSuperTreeletFlag(bool b) {
        wasGeneratedAsDescendingSuperTreeletFlag = b;
    }

    bool wasGeneratedAsDescendingSuperTreelet() const {
        return wasGeneratedAsDescendingSuperTreeletFlag;
    }

    inline size_t getSize() const {
        return size;
    }
    ;

    TreeletId getId() const {
        if (!treeletIdIsSet) {
            throw UsingNonSetValueInTreeletException();
        }
        return index;
    }

    void setId(TreeletId id) {
        index = id;
        treeletIdIsSet = true;
    }

    size_t getNbSentOcc() const {
        if (!nbSentOccIsSet) {
            throw UsingNonSetValueInTreeletException();
        }
        return nbSentOcc;
    }

    void setNbSentOcc(size_t nbOcc) {
        nbSentOccIsSet = true;
        nbSentOcc = nbOcc;
    }

    std::string getTypeRepresentation() const {
        if (!isTypeRepresentationComputed) {
            throw std::logic_error("type representation not computed");
        }
        return typeRepresentation;
    }
    ;

    void resetTypeRepresentation() {
        typeRepresentation = "";
        isTypeRepresentationComputed = false;
    }
    ;

};

class Treelet: public TreeletBaseClass {
public:
    typedef SeqOfPosRepresentation SeqOfPosRepresentation_;
    //bool checkIfRootWildcard() const {return false;}
private:

    //vector<PosType> seqOfPos;
    //PosType parent;
    //TreeletSignature signature; // such that t1.signature==t2.signature => occ(t1)==occ(t2)

    SeqOfPosRepresentation seqOfPos;
    TreeletAsRecursiveRepresentation recursiveRepr;

    /*
     TreeletId index;
     size_t size;
     size_t nbSentOcc;
     bool nbSentOccIsSet;
     bool treeletIdIsSet;
     bool seqOfPosIsConsistent;

     // flag used for descending supertreelets management:
     bool wasGeneratedAsDescendingSuperTreeletFlag;

     bool isElementaryTreeletFlag;

     std::string typeRepresentation;
     bool isTypeRepresentationComputed;
     */

    std::vector<size_t> sizeOfChildrenVector;

    friend std::ostream &operator<<(std::ostream &stream, const Treelet &treelet);
public:

    class TreeletSimpleIndexer  //TODO: check if more complex version is necessary
    {
        std::vector<Treelet> treeletList;
        //map<Treelet,TreeletId> treeletToId;
        friend std::ostream &operator<<(std::ostream &stream, const TreeletSimpleIndexer &treeletIdx);
    public:
        TreeletId addTreeletAndGetId(const Treelet &treelet) {
            treeletList.push_back(treelet);
            TreeletId id = treeletList.size() - 1;
            treeletList.back().setId(id);
            return id;
        }
        ;

        Treelet &getTreelet(TreeletId treeletIndex) const {
            assert(treeletIndex!=TREELET_ID_SPECIAL_VALUE);
            const Treelet& result = treeletList.at(treeletIndex);
            // const_cast so that returned treelet can be modified, while still enforcing the fact that no treelet is added to treeletList
            // TODO: check if necessary...
            return const_cast<Treelet &>(result);
        }
        ;
    };

    class TreeletSizeInfos {
        map<TreeletId, size_t> sizeOfTreeletsMap;
    public:
        size_t getSizeOfTreelet(TreeletId id) const {
            map<TreeletId, size_t>::const_iterator it = sizeOfTreeletsMap.find(id);
            if (it == sizeOfTreeletsMap.end()) {
                throw DescendingSuperTreeletsNotFoundException();
            } else {
                return it->second;
            }
        }

        void addTreelet(TreeletId treeletId, const Treelet& treelet) {
            assert(sizeOfTreeletsMap.count(treeletId) == 0);
            sizeOfTreeletsMap[treeletId] = treelet.getSize();
        }
    };

    class TreeletStoreByChildrenIndices: public boost::noncopyable {
        map<vector<TreeletId>, TreeletId> mapTreeletsToId;
    public:

        void addTreelet(const Treelet& treelet, TreeletId idTreelet) {
            assert(mapTreeletsToId.count(treelet.getChildrenIndices()) == 0);
            mapTreeletsToId[treelet.getChildrenIndices()] = idTreelet;
        }

        TreeletId getLeftTreeletId(const Treelet& treelet, int index) const {
            assert(treelet.checkTreeletHasAllTopChildrenEmptyAfterIndex(index));
            vector<TreeletId> childVector = treelet.getChildrenIndices();
            childVector[index] = TREELET_ID_SPECIAL_VALUE;

            map<vector<TreeletId>, TreeletId>::const_iterator it1 = mapTreeletsToId.find(childVector);
            if (it1 == mapTreeletsToId.end()) {
                throw logic_error("bug");
            }

            TreeletId leftTreeletId = it1->second;
            return leftTreeletId;

        }
    };

    class CompareTreeletsBySize {
    public:
        bool operator()(const Treelet& t1, const Treelet& t2) const   // Returns true if t1 is smaller than t2
                {
            return t1.getSize() > t2.getSize();  // or the opposite?
        }
    };

    static bool needToComputeWildcardOcc() {
        return false;
    }

    static bool isUsingWildcard() {
        return false;
    }

    bool isWildcardElemPair(const TreeletSimpleIndexer& tsi, bool& isLeftChild, LabelType& childLabel) const {
        return false;
    }

    //General constructor
    Treelet(const Node *pRootNodeInQueryTree, std::vector<TreeletId> childrenIndices, size_t size,
            std::vector<size_t> sizeOfChildrenVector) :
            TreeletBaseClass(size, false),
            //size(size),
            recursiveRepr(pRootNodeInQueryTree, childrenIndices), sizeOfChildrenVector(sizeOfChildrenVector)
    //,seqOfPosIsConsistent(false),treeletIdIsSet(false),nbSentOccIsSet(false),isElementaryTreeletFlag(false),
    //wasGeneratedAsDescendingSuperTreeletFlag(false),isTypeRepresentationComputed(false)
    {
    }
    ;

    Treelet clone() const {
        return Treelet(recursiveRepr.getRootPointer(), recursiveRepr.getChildrenIndices(), size,
                sizeOfChildrenVector);
    }
    ;

    // Constructor for elementary treelets
    Treelet(const Node *pRootNodeInQueryTree) :
            TreeletBaseClass(1, true),
            //size(1),
            recursiveRepr(pRootNodeInQueryTree, pRootNodeInQueryTree->getNbOfChildren()), sizeOfChildrenVector(
                    pRootNodeInQueryTree->getNbOfChildren(), 0)
    //,seqOfPosIsConsistent(false),treeletIdIsSet(false),nbSentOccIsSet(false),isElementaryTreeletFlag(true),
    //wasGeneratedAsDescendingSuperTreeletFlag(false),isTypeRepresentationComputed(false)
    {
    }
    ;

    // Constructor for treelet with only one child at the root
    Treelet(const Node *pRootNodeInQueryTree, size_t nbChildren, size_t numChild, TreeletId uniqueChildId,
            size_t sizeOfUniqueChild) :
            TreeletBaseClass(1, false),
            //size(1),
            recursiveRepr(pRootNodeInQueryTree, nbChildren), sizeOfChildrenVector(nbChildren, 0)
    //,seqOfPosIsConsistent(false),treeletIdIsSet(false),nbSentOccIsSet(false),isElementaryTreeletFlag(false),
    //wasGeneratedAsDescendingSuperTreeletFlag(false),isTypeRepresentationComputed(false)
    {
        extendChildrenDownwardInPlace(numChild, uniqueChildId, sizeOfUniqueChild);
    }
    ;

    //TO DO:
    /*
     void resetTypeRepresentation()
     {
     typeRepresentation="";
     isTypeRepresentationComputed=false;
     };
     */
    void computeTypeRepresentation(const TreeletSimpleIndexer& tsi) {
        std::ostringstream typeAsStream;
        typeAsStream << getRootLabel();
        typeAsStream << "(";
        const vector<TreeletId>& childrenIndices = recursiveRepr.getChildrenIndices();
        for (size_t i = 0; i < childrenIndices.size(); i++) {
            if (i == recursiveRepr.getNbChildrenLeft()) {
                typeAsStream << "| ";
            }
            TreeletId childId = childrenIndices[i];

            if (childId == TREELET_ID_SPECIAL_VALUE) {
                continue;
            }
            typeAsStream << tsi.getTreelet(childId).getTypeRepresentation() << " ";
        }
        typeAsStream << ")";
        typeRepresentation = typeAsStream.str();
        isTypeRepresentationComputed = true;
    }
    ;

    /*
     std::string getTypeRepresentation() const
     {
     if (!isTypeRepresentationComputed)
     {
     throw std::logic_error("type representation not computed");
     }
     return typeRepresentation;
     };

     */

    /*
     bool isTreeletIdSet() const
     {
     return treeletIdIsSet;
     }
     */
    bool hasStrictlyMoreThanNTopSiblings(int n) const {
        return recursiveRepr.hasStrictlyMoreThanNTopSiblings(n);
    }

    bool hasTopWildcardChain() const {
        return false;
    }

    const vector<TreeletId>& getChildrenIndices() const {
        return recursiveRepr.getChildrenIndices();
    }

    const SeqOfPosRepresentation& getSeqOfPosRepresentation() const {
        if (!seqOfPosIsConsistent) {
            throw UsingNonSetValueInTreeletException();
        }
        return seqOfPos;
    }

    void buildSeqOfPos(const TreeletSimpleIndexer &treeletIndexer) {
        //std::cout<<"building seqOfPos for "<<*this<<std::endl;
        SeqOfPosRepresentation seqOfPos(recursiveRepr.getRootPointer()->getPosition());
        if (recursiveRepr.getRootPointer()->hasParent()) {
            seqOfPos.addParent(recursiveRepr.getParentNode()->getPosition());
        }
        BOOST_FOREACH(TreeletId treeletId, recursiveRepr.getChildrenIndices()) {
            if (treeletId == TREELET_ID_SPECIAL_VALUE)
                continue;
            const SeqOfPosRepresentation& seqOfPosChild =
                    treeletIndexer.getTreelet(treeletId).getSeqOfPosRepresentation();
            seqOfPos.mergeWith(seqOfPosChild);
        }
        this->seqOfPos = seqOfPos;
        seqOfPosIsConsistent = true;
    }

//        LabelType getRootLabel() const
//            {
//            return recursiveRepr.getRootLabel();
    //           }
    /*
     bool isElementaryTreelet() const
     {
     return isElementaryTreeletFlag;
     }
     */

    /*
     bool isElementaryWildcard() const
     {
     return false;
     };
     */

    bool isDescendingSupertreesRoot() const {
        return isElementaryTreelet();
    }

    bool hasContiguousChildrenBeforeIndex(int index) const {
        return recursiveRepr.hasContiguousChildrenBeforeIndex(index);
    }

    /*
     void setGeneratedAsDescendingSuperTreeletFlag(bool b)
     {
     wasGeneratedAsDescendingSuperTreeletFlag=b;
     }

     bool wasGeneratedAsDescendingSuperTreelet() const
     {
     return wasGeneratedAsDescendingSuperTreeletFlag;
     }
     */
    void extendChildrenDownwardInPlace(size_t childNumber, TreeletId newChildId, size_t sizeOfNewChild) {
        //std::cout<<"extending downward "<<*this<<" to ";
        //Treelet result(*this);
        assert((int )size - sizeOfChildrenVector[childNumber] + sizeOfNewChild >= 0);
        size = size - sizeOfChildrenVector[childNumber] + sizeOfNewChild;
        sizeOfChildrenVector[childNumber] = sizeOfNewChild;
        recursiveRepr.replaceChild(childNumber, newChildId);

        seqOfPosIsConsistent = false;
        nbSentOccIsSet = false;
        treeletIdIsSet = false;
        nbSentOcc = 0;
        index = 0;
        isElementaryTreeletFlag = false;
        //std::cout<<*this<<std::endl;
    }

    Treelet getExtendedDownwardTreelet(size_t childNumber, TreeletId newChildId,
            const TreeletSizeInfos &treeletSizeInfos) const {
        Treelet result = this->clone();
        size_t sizeOfNewChild = treeletSizeInfos.getSizeOfTreelet(newChildId);
        result.extendChildrenDownwardInPlace(childNumber, newChildId, sizeOfNewChild);
        return result;
    }

    Treelet getExtendedUpwardTreelet() const {
        const Node *pParentNode = recursiveRepr.getParentNode();
        size_t nbChildrenOfParent = pParentNode->getNbOfChildren();
        const Node *thisNode = recursiveRepr.getRootPointer();
        size_t posInParent = thisNode->getPosInParent();
        //std::cout<<"extending upward:"<< *this << " id:"<< getId()<<" size:"<<size<<" posInParent:"<<posInParent<<std::endl;
        return Treelet(pParentNode, nbChildrenOfParent, posInParent, getId(), size);
    }

    bool hasUpwardExtension() const {
        return recursiveRepr.hasUpwardExtension();
    }
    ;

    bool hasSpecialSuperTreelets(int index) const {
        assert(index == -1);
        return false;
    }
    ;

    bool isRootWildcard() const {
        return false;
    }

    void addSpecialSupertreelets(vector<Treelet>& result) const {
    }
    ;

    inline size_t getSize() const {
        return size;
    }
    ;

    inline bool operator<(const Treelet &other) const {
        return recursiveRepr < other.recursiveRepr;
    }
    ;

    TreeletId getId() const {
        if (!treeletIdIsSet) {
            throw UsingNonSetValueInTreeletException();
        }
        return index;
    }

    void setId(TreeletId id) {
        index = id;
        treeletIdIsSet = true;
    }

    size_t getNbOfChildren() const {
        return recursiveRepr.getNbOfChildren();
    }

    TreeletId getChildIndex(size_t num) const {
        return recursiveRepr.getChildIndex(num);
    }
    PosType getPosNodeAtChildNum(size_t num) const {
        return recursiveRepr.getPosNodeAtChildNum(num);
    }

    bool safeToExtendTopSiblings(int safety) const {
        int nbTopSiblings = recursiveRepr.getNbTopSiblings();
        int nodeWidth = recursiveRepr.getNodeWidth();
        return nbTopSiblings <= getSafeNbSiblingsForGivenNodeWidth(nodeWidth, safety);
    }

    const Node *getRootPointer() const {
        return recursiveRepr.getRootPointer();
    }

    PosType getRootPos() const {
        return getRootPointer()->getPosition();
    }

    bool hasGaps() const {
        return false;
    }
    int nbGaps() const {
        return 0;
    }

    bool rootIsWildcardAndHasDiscontiguousSiblings() const {
        return false;
    }

    std::pair<std::vector<PosType>, std::vector<PosType> > separateWildcardFromNWInMapp(
            const std::vector<PosType>& mapp) const {
        throw std::logic_error("Not implemented in TreeletBase");
    }
    /*
     size_t getNbSentOcc() const
     {
     if (!nbSentOccIsSet )
     {
     throw UsingNonSetValueInTreeletException();
     }
     return nbSentOcc;
     }

     void setNbSentOcc(size_t nbOcc)
     {
     nbSentOccIsSet=true;
     nbSentOcc=nbOcc;
     }
     */
    LabelType getRootLabel() const {
        return recursiveRepr.getRootLabel();
    }

    bool checkTreeletHasAllTopChildrenEmptyAfterIndex(int index) const {
        assert(index < (int )getNbOfChildren());
        assert(index >= -1);
        for (int i = index + 1; i < (int) getNbOfChildren(); i++) {
            if (getChildIndex(i) != TREELET_ID_SPECIAL_VALUE) {
                return false;
            }
        }
        return true;
    }

};

// size_t getSizeOfTreeletFunction(const Treelet& treelet)
// {
// return treelet.getSize();
// }

std::ostream &operator<<(std::ostream &stream, const Treelet &treelet) {

    stream << treelet.recursiveRepr << "(s:" << treelet.size;
    if (treelet.treeletIdIsSet)
        stream << "#" << treelet.getId();
    if (treelet.nbSentOccIsSet)
        stream << " |" << treelet.getNbSentOcc();
    stream << "|";
    stream << ")";
    if (treelet.seqOfPosIsConsistent)
        stream << treelet.getSeqOfPosRepresentation();
    // if (treelet.isTreeletIdSet())
    // {
    // stream<<"#"<<treelet.getId()<<"# ";
    // }
    return stream;
}
//int main(){ return 0; }

std::ostream &operator<<(std::ostream &stream, const Treelet::TreeletSimpleIndexer &treeletIdx) {
    for (TreeletId idx = 0; idx < treeletIdx.treeletList.size(); ++idx) {
        stream << idx << " : " << treeletIdx.treeletList[idx] << "\n";
    }
    return stream;
}

}
;
//namespace TreeletBase
#endif

