/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*! \file
 \ingroup occurences_merging_group
 \brief Algorithms for efficient merging of occurences.
 */

#include <glog/logging.h>
#include <exception>
#include <set>

#include "binaryMerge.hpp"

// As there is templates, we will need to include this file directly:
#ifndef CPP_BINARYMERGE_CPP
#define CPP_BINARYMERGE_CPP

//using std::set;
using std::endl;
using std::cout;
using std::cerr;

void ensureNoEmptyMapPerTree(OccListCompactPair& result, const std::set<TreeIdType>& seenTreeId) {
    if (result.size() == seenTreeId.size())
        return;
    std::cerr << "The data seem to contains empty trees, which is not supported at this time..." << std::endl;
    assert(false);
    throw std::exception();
}

void mergeOptimListOfOccurenceList(std::vector<OccListCompactPair>& occListVector,
        OccListCompactPair& wildcardOccListVector, std::vector<boost::shared_ptr<OccList> >& result) {
    //set<TreeIdType> seenTreeId;
    //result.resize(occListVector.size()*2);
    //int ir = 0;
    cout << "start mergeOptimListOfOccurenceList" << endl;
    result.reserve(occListVector.size() * 2);
    BOOST_FOREACH(const OccListCompactPair& occList, occListVector) {
        boost::shared_ptr<OccList> occListWcLeft = boost::make_shared<OccList>(); // 2i
        boost::shared_ptr<OccList> occListWcRight = boost::make_shared<OccList>(); // 2i+1
        BOOST_FOREACH(const MappingsPerTreeCompactPair& mapPerTree,occList) {
            //cout << "mapPerTree Original = " << mapPerTree << endl;
            TreeIdType numTree = mapPerTree.numTree;

            vector<Mapping> triplesLeft;
            vector<Mapping> triplesRight;

            const ArrayOfPairs<PosType>& mappingListAsArrayOfPair = mapPerTree.mappingList;
            BOOST_FOREACH(const PosType* pairOfPos ,mappingListAsArrayOfPair) {

                PosType wordPos = pairOfPos[0];
                PosType parentPos = pairOfPos[1];
                if (parentPos == POSTYPE_SPECIAL_VALUE) {
                    continue;
                }

                bool leftSide = wordPos < parentPos;
                if (leftSide) {
                    triplesLeft.push_back(Mapping());
                } else {
                    triplesRight.push_back(Mapping());
                }

                Mapping& triple = leftSide ? triplesLeft.back() : triplesRight.back();

                if (leftSide) {
                    // child on left
                    triple.push_back(wordPos);
                    triple.push_back(parentPos);
                } else {
                    // child on right
                    triple.push_back(parentPos);
                    triple.push_back(wordPos);
                }
                // find parent of parent in wildcard occurrence list
                MappingsPerTreeCompactPair& wcMPT = wildcardOccListVector[numTree];
                assert(wcMPT.numTree == numTree);
                BOOST_FOREACH(const PosType* wcPos, wcMPT.mappingList) {
                    if (parentPos == *wcPos) {
                        triple.push_back(wcPos[1]);
                        break;
                    }
                }

            }

            if (triplesRight.size() > 0) {
                occListWcRight->push_back(MappingsPerTree());
                occListWcRight->back().numTree = numTree;
                occListWcRight->back().mappingList.swap(triplesRight);

                //cout << "mapPerTree Right   = " << mapPerTree << endl;
            }
            if (triplesLeft.size() > 0) {
                occListWcLeft->push_back(MappingsPerTree());
                occListWcLeft->back().numTree = numTree;
                occListWcLeft->back().mappingList.swap(triplesLeft);
                //cout << "mapPerTree Left    = " << mapPerTree << endl;
            }
            //cout << endl;
        }
        result.push_back(occListWcLeft);
        result.push_back(occListWcRight);
        //ir++;
        //break; //debug
    }
    //ensureNoEmptyMapPerTree(result,seenTreeId);  
    cout << "finish mergeOptimListOfOccurenceList" << endl;
}

/*!
 Merge a list of OccListCompactPair. (Can be used to generate wildcard occurences)
 */
void mergeListOfOccurenceList(std::vector<OccListCompactPair>& occListVector, OccListCompactPair& result) {
    std::set<TreeIdType> seenTreeId;
    BOOST_FOREACH(const OccListCompactPair& occList,occListVector) {
        BOOST_FOREACH(const MappingsPerTreeCompactPair& mapPerTree,occList) {
            TreeIdType numTree = mapPerTree.numTree;
            seenTreeId.insert(numTree);
            assert(numTree >= 0);
            if (result.size() <= (unsigned int) numTree) {
                result.resize(numTree + 1);
            }
            MappingsPerTreeCompactPair& resultMapPerTree = result[numTree];
            resultMapPerTree.numTree = numTree;
            resultMapPerTree.mappingList.concatenateWith(mapPerTree.mappingList);
        }
    }
    ensureNoEmptyMapPerTree(result, seenTreeId);
}

//@+node:fabien.20100629171036.1820:vectUtil
/*
 template<typename T>
 std::ostream &operator<< (std::ostream &stream,const std::vector<T> &vect)
 {

 stream<<"Vect("<<vect.size()<< "/"<<vect.capacity()<<")<";
 for (unsigned int i=0;i<vect.size();++i)
 {
 stream<<vect[i]<<" ";
 }

 stream<<">";

 return stream;
 }
 */
//@-node:fabien.20100629171036.1820:vectUtil

//@+node:fabien.20100629171036.1821:areMappingsCompatibleNoOrder
/*!
 \brief Check the compatibility of 2 occurence lists.
 */

template<typename MappingType1, typename MappingType2>
bool areMappingsCompatibleNoOrder(const MappingType1 & mapping1, const MappingType2& mapping2,
        const vector<SlotPairType>& equal, const vector<SlotListPairType>& diff) {
    //cout<<"areMappingsCompatibleNoOrder"<<endl;
    BOOST_FOREACH(const SlotPairType& pp,equal) {
        if (mapping1[pp.first] != mapping2[pp.second]) {
            return false;
        }
    }

    BOOST_FOREACH(const SlotListPairType& plp,diff) {
        BOOST_FOREACH(SlotType pos1,plp.first) {
            BOOST_FOREACH(SlotType pos2,plp.second)  //optimizable aussi...
            {
                if (mapping1[pos1] == mapping2[pos2]) {
                    return false;
                }
            }

        }
    }
    return true;

}

//@-node:fabien.20100629171036.1821:areMappingsCompatibleNoOrder

//@+node:fabien.20100707115414.1854:areMappingCompatiblePreserveOrder
template<typename MappingType1, typename MappingType2>
bool areMappingsCompatiblePreserveOrder(const MappingType1 & mapping1, const MappingType2& mapping2,
        const vector<SlotPairType>& equal, const vector<MapSlotPair>& orderRes) {

    //cout<<"areMappingsCompatiblePreserveOrder"<<endl;

    BOOST_FOREACH(const SlotPairType& pp,equal) {
        if (mapping1[pp.first] != mapping2[pp.second]) {
            return false;
        }
    }

    size_t sizeOfOrderRes = orderRes.size();
    assert(sizeOfOrderRes >= 1);

    int prevPos = -1;
    int curPos = -1;

    vector<MapSlotPair>::const_iterator start = orderRes.begin();
    vector<MapSlotPair>::const_iterator end = start + sizeOfOrderRes - 1;
    for (; start != end; ++start)

    //for( size_t numSlot=0; numSlot < sizeOfOrderRes-1; numSlot++)
    // BOOST_FOREACH(const MapSlotPair& mp,orderRes)
            {
        // if (numSlot>=sizeOfOrderRes-1)  // on ne compare pas le dernier slot contenant le father
        // {
        // break;
        // }
        // numSlot++;

        //const MapSlotPair& mp = orderRes[numSlot];

        if (start->numMapping == 0) {
            curPos = mapping1[start->slot];
        } else {
            assert(start->numMapping == 1);
            curPos = mapping2[start->slot];
        }

        if (curPos <= prevPos) {
            return false;
        }
        prevPos = curPos;
    }

    return true;

}
//@-node:fabien.20100707115414.1854:areMappingCompatiblePreserveOrder

//@+node:fabien.20100629171036.1822:mergeCompatibleMappingsPair
template<typename MappingType1, typename MappingType2>
void mergeCompatibleMappingsPair(const MappingType1 & mapping1, const MappingType2& mapping2,
        vector<Mapping>& res, const vector<MapSlotPair>& orderRes) {
    res.push_back(Mapping());
    Mapping &newMapping = res.back();
    newMapping.reserve(orderRes.size());
    BOOST_FOREACH(const MapSlotPair& mp,orderRes) {

        if (mp.numMapping == 0) {
            newMapping.push_back(mapping1[mp.slot]);
        } else {
            assert(mp.numMapping == 1);
            newMapping.push_back(mapping2[mp.slot]);
        }
    }
}

//@-node:fabien.20100629171036.1822:mergeCompatibleMappingsPair

template<class Mapping1Type, class Mapping2Type>
class CompareMappingsEquality {
    bool operator()(const Mapping1Type& mapping1, const Mapping2Type& mapping2) const {
        for (auto slot_pair : equal_) {
            if (mapping1[slot_pair.first] == mapping2[slot_pair.second])
                continue;
            if (mapping1[slot_pair.first] < mapping2[slot_pair.second]) {
                return true;
            } else {
                return false;
            }
        }
        return false; // equals
    }
private:
    const vector<SlotPairType>& equal_;
};

//@+node:fabien.20100629171036.1823:mergeMappings
template<class MappingListType1, class MappingListType2>
void mergeMappings(const MappingListType1& mappingList1, const MappingListType2& mappingList2,
        TreeIdType numTree, OccList& res, const vector<SlotPairType>& equal,
        const vector<SlotListPairType>& diff, const vector<MapSlotPair>& orderRes, bool preserveOrder) {
    //if (numTree==25020)
    //     {
    //   cout<<"found 25020"<<endl;
    //    }
    //typename GenericOccList1::value_type MappingType1;
    //typename GenericOccList2::value_type MappingType2;
    typedef typename std::iterator_traits<typename MappingListType1::const_iterator>::value_type Mapping1Type;
    typedef typename std::iterator_traits<typename MappingListType2::const_iterator>::value_type Mapping2Type;

    res.push_back(MappingsPerTree());
    BOOST_FOREACH(const Mapping1Type& mapping1, mappingList1) // optimisable avec hashage des positions et compare equal
    {
        BOOST_FOREACH(const Mapping2Type& mapping2, mappingList2) {

            if ((preserveOrder && areMappingsCompatiblePreserveOrder(mapping1, mapping2, equal, orderRes))
                    || (!preserveOrder && areMappingsCompatibleNoOrder(mapping1, mapping2, equal, diff)))

                    {
                mergeCompatibleMappingsPair(mapping1, mapping2, res.back().mappingList, orderRes);
            }
        }
    }
    if (res.back().mappingList.size() > 0) {
        res.back().numTree = numTree;
    } else {
        res.pop_back();
    }
}

/*!
 Intersect 2 occurrence lists using a linear merge algorithm (faster than binary merge when the 2 lists have similar sizes).
 */
template<typename GenericOccList1, typename GenericOccList2>
void intersectLinearMerge(const GenericOccList1 &set1, const GenericOccList2 &set2, OccList &res,
        const vector<SlotPairType>& equal, const vector<SlotListPairType>& diff,
        const vector<MapSlotPair>& orderRes, const vector<SlotPairType>& invEqual,
        const vector<SlotListPairType>& invDiff, const vector<MapSlotPair>& invOrderRes, bool preserveOrder) {

    typename GenericOccList1::size_type n1 = set1.size();
    typename GenericOccList2::size_type n2 = set2.size();
    if (n1 == 0 || n2 == 0)
        return;

    size_t cur1 = 0;
    size_t cur2 = 0;

    TreeIdType numTree1 = set1[0].numTree;
    TreeIdType numTree2 = set2[0].numTree;

    while (1)   // (cur1<n1 && cur2<n2)
    {
        if (numTree1 < numTree2) {
            ++cur1;
            if (cur1 >= n1)
                break;
            numTree1 = set1[cur1].numTree;
        } else if (numTree2 < numTree1) {
            ++cur2;
            if (cur2 >= n2)
                break;
            numTree2 = set2[cur2].numTree;
        } else {
            mergeMappings(set1[cur1].mappingList, set2[cur2].mappingList, numTree1, res, equal, diff,
                    orderRes, preserveOrder);
            ++cur1;
            ++cur2;
            if (cur1 >= n1 || cur2 >= n2)
                break;
            numTree1 = set1[cur1].numTree;
            numTree2 = set2[cur2].numTree;
        }
    }

}

/*!
 Intersect 2 occurrence lists using a binary merge algorithm (faster than linear merge when the 2 lists have very different sizes).
 */
template<typename GenericOccList1, typename GenericOccList2>
void intersectBinaryMerge(const GenericOccList1 &set1, const GenericOccList2 &set2, OccList &res,
        typename GenericOccList1::size_type start1, typename GenericOccList1::size_type end1,
        typename GenericOccList2::size_type start2, typename GenericOccList2::size_type end2,
        const vector<SlotPairType>& equal, const vector<SlotListPairType>& diff,
        const vector<MapSlotPair>& orderRes, const vector<SlotPairType>& invEqual,
        const vector<SlotListPairType>& invDiff, const vector<MapSlotPair>& invOrderRes, bool preserveOrder) {
    assert(end1 > start1);
    assert(end2 > start2);
    typename GenericOccList1::size_type n1 = end1 - start1;
    typename GenericOccList2::size_type n2 = end2 - start2;

    if (n1 == 0 || n2 == 0) //if one of the occurrence set is empty
            {
        return;
    }

    OccList::size_type med;

    typedef typename GenericOccList1::value_type MappingsPerTree1;
    typedef typename GenericOccList2::value_type MappingsPerTree2;

    MappingsPerTree1 medianIn1;
    MappingsPerTree2 medianIn2;

    if (n2 > n1) {
        intersectBinaryMerge(set2, set1, res, start2, end2, start1, end1, invEqual, invDiff, invOrderRes,
                equal, diff, orderRes, preserveOrder);
        return;
    }
    //cout<<"__intersect2"<<endl;
    //cout<<"     size1:"<<set1.size()<<" start1:"<<start1<<" end1:"<<end1<<endl;
    //cout<<"     size2:"<<set2.size()<<" start2:"<<start2<<" end2:"<<end2<<endl;
    //cout<<"     size_res:"<<res.size()<<endl;
    bool foundMedian = false;
    std::pair<typename GenericOccList1::const_iterator, typename GenericOccList1::const_iterator> startEnd;

    med = start2 + n2 / 2;
    medianIn2 = set2[med];
    typename GenericOccList1::const_iterator beginOfSearch = set1.begin() + start1;
    typename GenericOccList1::const_iterator endOfSearch = set1.begin() + end1;
    startEnd = equal_range(beginOfSearch, endOfSearch, medianIn2);  //optimize??

    ptrdiff_t medIn1_signed = startEnd.first - set1.begin();
    assert(medIn1_signed >= 0);
    unsigned int medIn1 = (unsigned int) medIn1_signed;
    //cout<<"     med:"<<med<<" medianIn2:"<<medianIn2.numTree<<"   start1:"<<start1<<"  end1:"<<end1<<"  start2:"<<start2<<"  end2:"<<end2<< endl;

    if (startEnd.first != startEnd.second)  // median a ete trouve
            {
        assert(startEnd.first + 1 == startEnd.second);   //verifie qu il n y en a qu un

        medianIn1 = *startEnd.first;
        foundMedian = true;

        //cout<<"     found medianIn2 at medIn1:"<<medIn1<<endl;
    } else {
        //cout<<"     not found medianIn2 at medIn1:"<<medIn1<<endl;
    }

    if (med > start2 && medIn1 > start1) // on verifie qu'il reste des elements a comparer
                                         //TODO: ajouter possibilite de repartit sur Merge sequentiel si ratio de taille proche de 1 ??
            {
        intersectBinaryMerge(set1, set2, res, start1, medIn1, start2, med, equal, diff, orderRes, invEqual,
                invDiff, invOrderRes, preserveOrder);
    }
    if (foundMedian) {
        //cout<<" output:"<<medianIn1.numTree<<endl;
        mergeMappings(medianIn1.mappingList, medianIn2.mappingList, medianIn1.numTree, res, equal, diff,
                orderRes, preserveOrder);
    }

    if ((medIn1 + (foundMedian ? 1 : 0)) < end1 && (med + 1) < end2) {
        intersectBinaryMerge(set1, set2, res, medIn1 + (foundMedian ? 1 : 0), end1, med + 1, end2, equal,
                diff, orderRes, invEqual, invDiff, invOrderRes, preserveOrder);
    }
}

/*!
 Main function for merging 2 occurence lists. Will call either the linear merge function or the binary merge function, depending on the relative size of each list.
 */
template<typename GenericOccList1, typename GenericOccList2>
OccList *intersectAndMerge(const GenericOccList1 &set1, const GenericOccList2 &set2,
        const vector<SlotPairType>& equal, const vector<SlotListPairType>& diff,
        const vector<MapSlotPair>& orderRes, bool preserveOrder) {
    OccList *res = new OccList;
    typename GenericOccList1::size_type n1 = set1.size();
    typename GenericOccList2::size_type n2 = set2.size();

    if (n1 == 0 || n2 == 0) {
        return res;
    }

    vector<SlotPairType> invEqual;
    BOOST_FOREACH(const SlotPairType& pp,equal) {
        invEqual.push_back(SlotPairType(pp.second, pp.first));
    }

    vector<SlotListPairType> invDiff;
    if (preserveOrder) {
        assert(diff.size() == 0);
    } else {
        BOOST_FOREACH(const SlotListPairType& pp,diff) {
            invDiff.push_back(SlotListPairType(pp.second, pp.first));
        }
    }

    vector<MapSlotPair> invOrderRes;
    BOOST_FOREACH(const MapSlotPair& mp,orderRes) {
        invOrderRes.push_back(MapSlotPair(1 - mp.numMapping, mp.slot));
    }

    if (RATIO_BINARY_MERGE * n1 > n2 && RATIO_BINARY_MERGE * n2 > n1) //grosso modo: possiility to optimize values here
            {
        intersectLinearMerge(set1, set2, *res, equal, diff, orderRes, invEqual, invDiff, invOrderRes,
                preserveOrder);
    } else {
        intersectBinaryMerge(set1, set2, *res, 0, n1, 0, n2, equal, diff, orderRes, invEqual, invDiff,
                invOrderRes, preserveOrder);
    }

    //cout<<"preserveOrder:"<<preserveOrder<<endl;
    return res;

}
//@-node:fabien.20100629171036.1825:intersectAndMerge
//@+node:fabien.20110526095404.2555:isMappingListDominated
template<class MappingListType1, class MappingListType2>
bool isMappingListDominated(const MappingListType1& mappingList1, const MappingListType2& mappingList2,
        const vector<SlotPairType>& equal) {
    typedef typename std::iterator_traits<typename MappingListType1::const_iterator>::value_type MappingType1;

    typedef typename std::iterator_traits<typename MappingListType2::const_iterator>::value_type MappingType2;

    BOOST_FOREACH(const MappingType1& mapping1,mappingList1) // optimisable avec hashage des positions et compare equal
    {
        bool foundDominant = false;
        BOOST_FOREACH(const MappingType2& mapping2,mappingList2) // for each mapping1, we check all possible mapping2
        {

            //We check if mapping1 has the same occurence as mapping2
            bool compatible = true;
            BOOST_FOREACH(const SlotPairType& pp,equal) {
                if (mapping1[pp.first] != mapping2[pp.second]) {
                    compatible = false;
                    break;
                }
            }
            if (compatible) {
                foundDominant = true;
                break;
            }
        }
        if (!foundDominant)
            return false;  // no dominant found for mapping1
    }
    return true; // all mapping1 have a dominant in mapingList2
}
//@-node:fabien.20110526095404.2555:isMappingListDominated
//@+node:fabien.20110526095404.2554:isDominated
template<typename GenericOccList1, typename GenericOccList2>
bool isDominated(GenericOccList1 &set1, GenericOccList2 &set2, const vector<SlotPairType>& equal) {

    typename GenericOccList1::size_type n1 = set1.size();
    typename GenericOccList2::size_type n2 = set2.size();

    if (n1 != n2)
        return false;

    for (typename GenericOccList1::size_type i = 0; i < n1; ++i) {
        TreeIdType numTree1 = set1[i].numTree;
        TreeIdType numTree2 = set2[i].numTree;
        if (numTree1 != numTree2)
            return false;

        if (!isMappingListDominated(set1[i].mappingList, set2[i].mappingList, equal))
            return false;
    }
    return true;
}
;
//@-node:fabien.20110526095404.2554:isDominated
//@+node:fabien.20100702173129.1831:load/save
vector<OccList> *loadVectorOccList(char *filename) {
    std::ifstream ifs(filename, std::ifstream::binary);
    vector<OccList> *res = new vector<OccList>();

    try {
        boost::archive::binary_iarchive ia(ifs);
        ia >> (*res);
    } catch (const boost::archive::archive_exception& e) {
        delete res;
        LOG(FATAL)<< "Error loading vector<OccList> from file " << filename << " what: " << e.what() <<endl;
        return NULL;
    }
    return res;
}

void saveVectorOccList(const vector<OccList> &oList, char *filename) {
    std::ofstream ofs(filename, std::ofstream::binary);

    try {
        boost::archive::binary_oarchive oa(ofs);
        oa << oList;
    } catch (const boost::archive::archive_exception& e) {
        LOG(FATAL) << "Error saving vector<OccList> to file " << filename << " what:" << e.what() << endl;
    }

}

vector<vector<MappingsPerTreeCompactPair> > *loadVectorOccListCompactPair(const char *filename) {
    std::ifstream ifs(filename, std::ifstream::binary);
    vector<vector<MappingsPerTreeCompactPair> > *res = new vector<vector<MappingsPerTreeCompactPair> >();
    try {
        boost::archive::binary_iarchive ia(ifs);
        ia >> (*res);
    } catch (const boost::archive::archive_exception& e) {
        LOG(FATAL) << "Error loading vector< vector<MappingsPerTreeCompactPair> > from file " << filename << " what:" << e.what() <<
                "\nDid you run 'make sample' from the root directory?" << endl;
        delete res;
        return NULL;
    }
    return res;
}
;

void saveVectorOccListCompactPair(const vector<vector<MappingsPerTreeCompactPair> > &oList,
        const char *filename) {
    std::ofstream ofs(filename, std::ofstream::binary);

    try {
        boost::archive::binary_oarchive oa(ofs);
        oa << oList;
    } catch (const boost::archive::archive_exception& e) {
        LOG(FATAL) << "Error saving vector< vector<MappingsPerTreeCompactPair> > to file " << filename << " what:" << e.what() << endl;
    }

}
;

#endif
