/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*! \file
 \ingroup target_info_group
 \brief Code for checking "target-side continuity" and computing the "target image" of a treelet occurrence.
 */

//@+leo-ver=4-thin
//@+node:fabien.20100607195349.1650:@thin cTargetInfoList.cpp
//@@language c
//#include "/home/fabien/orchid/local/include/python2.5/Python.h"
//#include <Python.h>
#include <vector>
#include <set>
#include <stdexcept>
//#include <multimap>
#include <iostream>
#include <fstream>
#include <sstream>

//#include <hash_map>
#include <string>

#include <boost/serialization/vector.hpp>
#include <boost/serialization/base_object.hpp>

//#include "/home/fabien/orchid/boost2/include/boost-1_34_1/boost/serialization/map.hpp"

#include <boost/foreach.hpp>

#include <boost/smart_ptr.hpp>

#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
//#include "/home/fabien/orchid/boost2/include/boost-1_34_1/boost/archive/binary_iarchive.hpp"
//#include "/home/fabien/orchid/boost2/include/boost-1_34_1/boost/archive/binary_oarchive.hpp"

#include <climits>
#include <cwchar>
#include <limits>
#include <boost/static_assert.hpp>
#include <cassert>

#include "TypeDefinitions.hpp"

#ifndef _HPP_TARGET_INFO_
#define _HPP_TARGET_INFO_

using namespace std;
// using namespace __gnu_cxx;

//#define POS_NONE 255
// assert  POS_NONE can be represented by PosType
//BOOST_STATIC_ASSERT(std::numeric_limits<PosType>::digits >= POS_NONE);

typedef std::vector<PosType> PosList;
typedef std::set<PosType> PosSet;

struct LinkWithGaps {
    std::vector<PosType> first;
    std::vector<PosType> second;
    std::vector<PosType> add_i_side;
    std::vector<PosType> add_j_side;
    void invertSide() {
        first.swap(second);
        add_i_side.swap(add_j_side);
    }
};

//@+others
//@+node:fabien.20100612171957.1702:contains
bool contains(const PosList *posList, PosType pos) //could be faster (asymptotically) by assuming sorted list
        {
    BOOST_FOREACH(PosType p,*posList) {
        if (p == pos) {
            return true;
        }
    }
    return false;
}
;

std::vector<PosType> getGrouping(PosType pos0, const std::vector<PosType>& grouping) {
    PosType pos = pos0;
    std::vector<PosType> result;
    do {
        result.push_back(pos);
        pos = grouping[pos];
    } while (pos != pos0);
    return result;
}

struct ProjectedMappingInfos {
    bool couldProject;
    boost::shared_ptr<std::set<PosType> > pConnectedImage;
    std::vector<std::pair<std::vector<PosType>, std::vector<PosType> > > matchList;
    std::vector<PosType> additionalNodes;

    ProjectedMappingInfos() :
            couldProject(false) {
    }
    ProjectedMappingInfos(std::set<PosType> *pConnectedImage) :
            couldProject(true), pConnectedImage(pConnectedImage) {
    }
};

std::ostream &operator<<(std::ostream &stream, const ProjectedMappingInfos &pmi) {
    if (pmi.couldProject) {
        stream << "OK:" << pmi.couldProject << "  cSet:" << *(pmi.pConnectedImage) << "  add:"
                << pmi.additionalNodes;
    } else {
        stream << "[NO Proj]";
    }

    return stream;
}

class TargetInfo {
    friend class boost::serialization::access;
    template<class Archive> void serialize(Archive & ar, const unsigned int version) {
        // serialize base class information
        ar & depend;
        ar & grouping1;
        ar & grouping2;
        ar & mapping;
        ar & partsMapping;
    }

public:
    PosList depend; // target dependency list
    PosList grouping1; // cycle representation of the source aligned groups
    PosList grouping2; // cycle representation of the target aligned groups
    PosList mapping; // mapping between lowest positions of source and target groups
    PosList partsMapping;
    //new for gap

    bool isPartOfGappedLink(PosType pos) const {
        return partsMapping[pos] != POSTYPE_SPECIAL_VALUE;
    }

    // end new for gap

    std::vector<std::pair<std::vector<PosType>, std::vector<PosType> > > getMatchingInfosAsVector() const {

        std::vector<PosType> listOfMin;

#ifndef NDEBUG
        std::set<PosType> insertedTargetPos;
#endif

        for (size_t i = 0; i < mapping.size(); ++i) {
            assert(isValidNonSpecialPosTypeValue(i));
            if (mapping[i] != POSTYPE_SPECIAL_VALUE) {
#ifndef NDEBUG
                assert(insertedTargetPos.count(mapping[i]) == 0);
                insertedTargetPos.insert(mapping[i]);
#endif

                listOfMin.push_back((PosType) i);
            }
        }

        std::vector<std::pair<std::vector<PosType>, std::vector<PosType> > > result;

        for (size_t i = 0; i < listOfMin.size(); ++i) {
            PosType minPos = listOfMin[i];
            PosType mappedMinPos = mapping[minPos];
            result.push_back(make_pair(getGrouping(minPos, grouping1), getGrouping(mappedMinPos, grouping2)));
        }

        return result;
    }

    bool isConnected(const PosList *posList) const;

    PosType getHigher(PosType pos, const PosList *posList) const;
    PosType getHigherWithIntermediate(PosType pos, const PosList *posList, PosSet &intermediate) const;

    PosList *project(PosSet &posSet) const;
    PosList *project_debug(PosSet &posSet) const;

    bool isProjectionConnected(PosSet &posSet) const;
    PosList * returnProjectionIfConnected(PosSet &posSet) const;
    PosSet * returnConnectedImage(const PosList *posList) const;
    bool returnConnectedImageAndMappings(PosSet &posSet) const;

    PosList *projectAndAddMappings(PosSet &posSet, PosList &mappingsMin) const;
    PosSet *returnMinAndMaxConnectedImage(const PosList *posList, vector<PosType> &maxAdditionalNodes) const;

    void moveHigherEmptyNodesToMaxAdditionalNodes(PosSet &intermediate, const PosList &higherImageNodes,
            PosList &maxAdditionalNodes) const;

    PosType getHigherWithIntermediateForHighest(PosType pos, PosList &intermediate) const;

    ProjectedMappingInfos projectMinMaxAndGiveMapping(const std::vector<PosType>& posList) const;

    PosSet * returnMinAndMaxMultipleConnectedImage(const PosList *posList,
            vector<PosType> &maxAdditionalNodes, const vector<PosType>& map2ToMin) const;

    pair<PosType, PosType> getHigherAndHighestWithIntermediate(PosType pos, const PosList *posList,
            PosSet &intermediate) const;

    ProjectedMappingInfos projectMinMaxAndGiveMappingWithGap(const std::vector<PosType>& posList) const;

    PosList * projectAndAddMappingsWithMap2ToMin(PosSet &posSet, PosList &mappingsMin,
            PosList &map2ToMin) const;

    void set(const std::vector<std::pair<std::vector<PosType>, std::vector<PosType> > >& matchList,
            const std::vector<PosType>& depend, size_t sourceSize);

    void set_gap(const std::vector<LinkWithGaps>& matchList, const std::vector<PosType>& depend,
            size_t sourceSize);

};

std::ostream &operator<<(std::ostream &stream, const TargetInfo &ti) {

    stream << "D:" << ti.depend << "  G1:" << ti.grouping1 << "  G2:" << ti.grouping2 << "  M:" << ti.mapping
            << " P:" << ti.partsMapping;

    return stream;
}

void TargetInfo::set_gap(const std::vector<LinkWithGaps>& matchList, const std::vector<PosType>& depend,
        size_t sourceSize) {
    size_t targetSize = depend.size();
    this->depend = depend;
    grouping1 = PosList(sourceSize, POSTYPE_SPECIAL_VALUE);
    grouping2 = PosList(targetSize, POSTYPE_SPECIAL_VALUE);
    mapping = PosList(sourceSize, POSTYPE_SPECIAL_VALUE);
    partsMapping = PosList(targetSize, POSTYPE_SPECIAL_VALUE);

    //std::cout<< "start:"<<std::endl;
    for (size_t i = 0; i < matchList.size(); ++i) {
        LinkWithGaps link = matchList[i];
        assert(0!=POSTYPE_SPECIAL_VALUE);
        //std::cout<< "add:"<<i<<std::endl;
        //std::cout<< link.add_j_side<<std::endl;
        //std::cout<< link.add_i_side<<std::endl;
        // We set partsMapping and copy link.add_j_link to link.second
        if (link.add_j_side.size() > 0) {
            BOOST_FOREACH(PosType pos,link.second) {
                partsMapping[pos] = 0;
            }
            for (size_t numAdd = 0; numAdd < link.add_j_side.size(); numAdd++) {
                assert((numAdd+1)!=POSTYPE_SPECIAL_VALUE);
                PosType pos = link.add_j_side[numAdd];
                partsMapping[pos] = numAdd + 1;
                link.second.push_back(pos);
            }
        }

        std::sort(link.first.begin(), link.first.end());
        std::sort(link.second.begin(), link.second.end());

        assert(link.first.back() < sourceSize);
        assert(link.second.back() < targetSize);
        assert(link.first.size() > 0 && link.second.size() > 0);
        assert(mapping[link.first[0]]==POSTYPE_SPECIAL_VALUE);

        mapping[link.first[0]] = link.second[0];
        for (size_t j = 0; j < link.first.size(); ++j) {
            assert(grouping1[link.first[j]]==POSTYPE_SPECIAL_VALUE);
            if (j + 1 < link.first.size()) {
                grouping1[link.first[j]] = link.first[j + 1];
            } else {
                grouping1[link.first[j]] = link.first[0];
            }
        }

        for (size_t j = 0; j < link.second.size(); ++j) {
            assert(grouping2[link.second[j]]==POSTYPE_SPECIAL_VALUE);
            if (j + 1 < link.second.size()) {
                grouping2[link.second[j]] = link.second[j + 1];
            } else {
                grouping2[link.second[j]] = link.second[0];
            }
        }
    }
}

void TargetInfo::set(const std::vector<std::pair<std::vector<PosType>, std::vector<PosType> > >& matchList,
        const std::vector<PosType>& depend, size_t sourceSize) {
    size_t targetSize = depend.size();
    this->depend = depend;
    grouping1 = PosList(sourceSize, POSTYPE_SPECIAL_VALUE);
    grouping2 = PosList(targetSize, POSTYPE_SPECIAL_VALUE);
    mapping = PosList(sourceSize, POSTYPE_SPECIAL_VALUE);
    for (size_t i = 0; i < matchList.size(); ++i) {
        std::pair<std::vector<PosType>, std::vector<PosType> > link = matchList[i];
        std::sort(link.first.begin(), link.first.end());
        std::sort(link.second.begin(), link.second.end());

        assert(link.first.back() < sourceSize);
        assert(link.second.back() < targetSize);
        assert(link.first.size() > 0 && link.second.size() > 0);
        assert(mapping[link.first[0]]==POSTYPE_SPECIAL_VALUE);

        mapping[link.first[0]] = link.second[0];
        for (size_t j = 0; j < link.first.size(); ++j) {
            assert(grouping1[link.first[j]]==POSTYPE_SPECIAL_VALUE);
            if (j + 1 < link.first.size()) {
                grouping1[link.first[j]] = link.first[j + 1];
            } else {
                grouping1[link.first[j]] = link.first[0];
            }
        }

        for (size_t j = 0; j < link.second.size(); ++j) {
            assert(grouping2[link.second[j]]==POSTYPE_SPECIAL_VALUE);
            if (j + 1 < link.second.size()) {
                grouping2[link.second[j]] = link.second[j + 1];
            } else {
                grouping2[link.second[j]] = link.second[0];
            }
        }

    }

}

//@+node:fabien.20100901091806.2010:Connection
//@+node:fabien.20100612171957.1694:TargetInfo::isConnected

ProjectedMappingInfos TargetInfo::projectMinMaxAndGiveMapping(const std::vector<PosType>& posList) const {
    //using the version with allowed gap instead
    if (true) {
        return projectMinMaxAndGiveMappingWithGap(posList);
    }

    //#get list of minPos for projection
    std::vector<PosType> listOfMin;
    //cdef vector[PosType] *pProjectedPosSet
    std::set<PosType> posSet(posList.begin(), posList.end());

    std::vector<PosType>* pProjectedPosSet = projectAndAddMappings(posSet, listOfMin);
    if (pProjectedPosSet == NULL) {
        return ProjectedMappingInfos();
    }
    //#ConnectProjectedPosSet
    //#print "listOfMin",listOfMin.size()

    std::vector<PosType> maxAdditionalNodes;
    std::set<PosType>* pConnected = returnMinAndMaxConnectedImage(pProjectedPosSet, maxAdditionalNodes);
    delete pProjectedPosSet;

    if (pConnected == NULL) {
        return ProjectedMappingInfos();
    }
    ProjectedMappingInfos result(pConnected);
    result.additionalNodes = maxAdditionalNodes;

    for (size_t i = 0; i < listOfMin.size(); ++i) {
        PosType minPos = listOfMin[i];
        PosType mappedMinPos = mapping[minPos];
        result.matchList.push_back(
                make_pair(getGrouping(minPos, grouping1), getGrouping(mappedMinPos, grouping2)));
    }

    return result;
}

ProjectedMappingInfos TargetInfo::projectMinMaxAndGiveMappingWithGap(
        const std::vector<PosType>& posList) const {

    //#get list of minPos for projection
    std::vector<PosType> listOfMin;
    //cdef vector[PosType] *pProjectedPosSet
    std::set<PosType> posSet(posList.begin(), posList.end());

    std::vector<PosType> map2ToMin(grouping2.size(), POSTYPE_SPECIAL_VALUE);

    // We try to compute the target image of posList
    std::vector<PosType>* pProjectedPosSet = projectAndAddMappingsWithMap2ToMin(posSet, listOfMin, map2ToMin);

    if (pProjectedPosSet == NULL) {
        // posList contains incomplete aligned group, or no aligned group at all: we cannot project
        return ProjectedMappingInfos();
    }
    //#ConnectProjectedPosSet
    //#print "listOfMin",listOfMin.size()

    std::vector<PosType> maxAdditionalNodes;

    // We check whether the image is connected, adding non-aligned positions if necessary
    // maxAdditionalNodes contains non-aligned positions that can be added while retaining the connectedness
    PosSet* pConnected = returnMinAndMaxMultipleConnectedImage(pProjectedPosSet, maxAdditionalNodes,
            map2ToMin);

    delete pProjectedPosSet;

    if (pConnected == NULL) {
        // the target image is disconnected: we reject
        //cout<<"not connected"<<endl;
        return ProjectedMappingInfos();
    }
    //cout<<"pCon: "<<*pConnected<<endl;

    // We could project posList into a connected image; we now build the result object
    ProjectedMappingInfos result(pConnected);
    result.additionalNodes = maxAdditionalNodes;

    for (size_t i = 0; i < listOfMin.size(); ++i) {
        PosType minPos = listOfMin[i];
        PosType mappedMinPos = mapping[minPos];
        result.matchList.push_back(
                make_pair(getGrouping(minPos, grouping1), getGrouping(mappedMinPos, grouping2)));
    }

    return result;
}

bool TargetInfo::isConnected(const PosList *posList) const {
    PosType higher = POSTYPE_SPECIAL_VALUE;
    //bool higherHasBeenSet=false;
    PosType h;
    BOOST_FOREACH(PosType pos, *posList) {
        h = getHigher(pos, posList);
        if (h != POSTYPE_SPECIAL_VALUE) {
            if (higher == POSTYPE_SPECIAL_VALUE) {
                higher = h;
            } else if (higher != h) {
                return false;
            }
        }

    }
    if (higher == POSTYPE_SPECIAL_VALUE) {
        assert(false);
        return false;
    } else {
        return true;
    }

}
//@nonl
//@-node:fabien.20100612171957.1694:TargetInfo::isConnected
//@+node:fabien.20100901091806.2004:TargetInfo::returnConnectedImage
PosSet * TargetInfo::returnConnectedImage(const PosList *posList) const {
    PosType higher = POSTYPE_SPECIAL_VALUE;
    //bool higherHasBeenSet=false;
    PosType h;
    PosSet * pIntermediate = new PosSet();
    BOOST_FOREACH(PosType pos, *posList) {
        h = getHigherWithIntermediate(pos, posList, *pIntermediate);
        if (h != POSTYPE_SPECIAL_VALUE) {
            if (higher == POSTYPE_SPECIAL_VALUE) {
                higher = h;
            } else if (higher != h)  // 2 different "higher" nodes --> not connected
                    {
                delete pIntermediate;
                return NULL;
            }
        }

    }
    if (higher == POSTYPE_SPECIAL_VALUE) {
        delete pIntermediate;
        assert(false);
        return NULL;
    } else {
        return pIntermediate;
    }

}
//@nonl
//@-node:fabien.20100901091806.2004:TargetInfo::returnConnectedImage
//@+node:fabien.20100901091806.2013:TargetInfo::returnMinAndMaxConnectedImage

/**************************************************
 * moveHigherEmptyNodesToMaxAdditionalNodes
 * move the non-aligned nodes in the upper part of "intermediate" to maxAdditionalNodes
 */

void TargetInfo::moveHigherEmptyNodesToMaxAdditionalNodes(PosSet &intermediate,
        const PosList &higherImageNodes, PosList &maxAdditionalNodes) const {
    size_t nbHigherNodes = higherImageNodes.size();

    // intermediateList will contain a vector of sequence of nodes
    // each sequence being a sequence of node from a "higher aligned node" to the (possibly non-aligned) highest node in intermediate
    vector<vector<PosType> > intermediateList;
    BOOST_FOREACH(PosType pos, higherImageNodes) {
        intermediateList.push_back(vector<PosType>());
        getHigherWithIntermediateForHighest(pos, intermediateList.back());
    }

    vector<vector<PosType>::reverse_iterator> itArrayStart;
    vector<vector<PosType>::reverse_iterator> itArrayEnd;
    BOOST_FOREACH(vector<PosType> &vectInterm,intermediateList) {
        itArrayStart.push_back(vectInterm.rbegin());
        itArrayEnd.push_back(vectInterm.rend());
    }

    //cout<<intermediateList<<endl;

    // the sequences of node are compared in reverse
    // to find the "lowest" common ancestor of the higher nodes
    bool finishLoop = false;
    PosList commonPos;
    while (!finishLoop) {
        PosType curPos;
        PosType thisPos;
        for (size_t i = 0; i < nbHigherNodes; ++i) {
            if (itArrayStart[i] == itArrayEnd[i])  // one of the list (list i) is finished
                    {
                //cout<<"finishLoop end array: curPos: "<<(int)curPos<<" thisPos: "<<(int)thisPos<<" i: "<<i<<endl;
                finishLoop = true;
                break;
            }
            thisPos = *(itArrayStart[i]);
            //cout<<"after deref: curPos: "<<(int)curPos<<" thisPos: "<<(int)thisPos<<" i: "<<i<<endl;
            if (i == 0)   //first list: initialisation
                    {
                curPos = thisPos;
                //cout<<"after init: curPos: "<<(int)curPos<<" thisPos: "<<(int)thisPos<<" i: "<<i<<endl;
            } else if (curPos != thisPos) // one of the list list(i) is different from the first list at current position
                    {
                //cout<<"finishLoop: curPos: "<<(int)curPos<<" thisPos: "<<(int)thisPos<<" i: "<<i<<endl;
                finishLoop = true;
                break;
            }
            ++itArrayStart[i];

        }
        if (finishLoop)
            break;
        //cout<<"push: "<<(int)curPos<<endl;
        commonPos.push_back(curPos);
    }

    assert(commonPos.size() >= 1);
    //cout<<"commonPos.size() "<<commonPos.size()<<endl;
    //cout<<"nbHigherNodes "<<nbHigherNodes<<endl;
    if (commonPos.size() > 1) {
        for (size_t i = 0; i < commonPos.size() - 1; ++i) {
            intermediate.erase(commonPos[i]);
            maxAdditionalNodes.push_back(commonPos[i]);
        }

    }

}

void extendToNonImageNodes(PosType pos, PosList &maxAdditionalNodes,
        const vector<vector<PosType> > &invDepend, const PosList &grouping2) {
    if (grouping2[pos] == POSTYPE_SPECIAL_VALUE)  // if pos is null-aligned
    {
        maxAdditionalNodes.push_back(pos); //add pos
        BOOST_FOREACH(PosType child,invDepend[pos]) //iterate over children of pos
        {
            extendToNonImageNodes(child, maxAdditionalNodes, invDepend, grouping2);
        }
    }

}

vector<vector<PosType> > buildInvDepend(const vector<PosType>& depend) {
    vector<vector<PosType> > invDepend(depend.size() + 1);
    for (size_t i = 0; i < depend.size(); ++i) {
        PosType posFather = depend[i];
        if (posFather == POSTYPE_SPECIAL_VALUE) {
            invDepend[depend.size()].push_back(i);
        } else {
            invDepend[posFather].push_back(i);
        }
    }
    return invDepend;
}

PosSet * TargetInfo::returnMinAndMaxConnectedImage(const PosList *posList,
        vector<PosType> &maxAdditionalNodes) const {
    PosType higher = POSTYPE_SPECIAL_VALUE;
    //bool higherHasBeenSet=false;
    PosType h;
    PosSet * pIntermediate = new PosSet();

    vector<PosType> higherImageNodes;

    //Step 1: moving up from each node in posList to the highest admissible node
    //        an admissible node is either a node in posList or a non-aligned node connected to posList through non-aligned nodes 
    BOOST_FOREACH(PosType pos, *posList) {
        h = getHigherWithIntermediate(pos, posList, *pIntermediate);
        if (h != POSTYPE_SPECIAL_VALUE) {
            if (higher == POSTYPE_SPECIAL_VALUE) {
                higher = h;
                higherImageNodes.push_back(pos);
            } else if (higher != h)  // 2 different "higher" nodes --> not connected
                    {
                delete pIntermediate;
                return NULL;
            } else    // higher identique
            {
                higherImageNodes.push_back(pos);

            }
        }

    }

    // Step 2: downward extension of pIntermediate through
    //         non-aligned nodes
    vector<vector<PosType> > invDepend = buildInvDepend(depend);

    BOOST_FOREACH(PosType pos, *pIntermediate) {
        BOOST_FOREACH(PosType child,invDepend[pos]) {
            if (pIntermediate->find(child) == pIntermediate->end()) {
                extendToNonImageNodes(child, maxAdditionalNodes, invDepend, grouping2);
            }
        }

    }

    // Step 3: reduce pIntermediate by moving the optional upper non-aligned nodes to maxAdditionalNodes
    moveHigherEmptyNodesToMaxAdditionalNodes(*pIntermediate, higherImageNodes, maxAdditionalNodes);

    if (higher == POSTYPE_SPECIAL_VALUE) {
        delete pIntermediate;
        assert(false);
        return NULL;
    } else {
        return pIntermediate;
    }

}

struct NumPart {
    PosType num;
    PosType part;

    NumPart(PosType num, PosType part) :
            num(num), part(part) {
    }

    bool operator<(const NumPart& other) const {
        return (num != other.num) ? (num < other.num) : (part < other.part);
    }

};

std::ostream &operator<<(std::ostream &stream, const NumPart &ti) {

    stream << "[num:" << (int) ti.num << " part:" << (int) ti.part << "]";

    return stream;
}

typedef set<NumPart> GapLinkSet;

struct ComponentInfo {
    PosSet connectedSet;
    vector<PosType> listOfHigherAlignedPos;
    GapLinkSet gapLinkSet;
};

std::ostream &operator<<(std::ostream &stream, const ComponentInfo &ti) {

    stream << "cSet" << ti.connectedSet << " higherList:" << ti.listOfHigherAlignedPos << " gapLink:"
            << ti.gapLinkSet;

    return stream;
}

bool areComponentGapConnected(vector<GapLinkSet> listOfGapLinkSet) {
    //cout<<">>areComponentGapConnected"<<endl;
    GapLinkSet gapLinkSeenSoFar;
    PosSet gapLinkNumVisited;

    bool sthgHasChanged(false);

    BOOST_FOREACH (const NumPart& numPart, listOfGapLinkSet.back()) {
        gapLinkNumVisited.insert(numPart.num);
        gapLinkSeenSoFar.insert(numPart);
    }
    listOfGapLinkSet.pop_back();
    while (1) {
        sthgHasChanged = false;
        vector<GapLinkSet> notYetLinked;
        BOOST_FOREACH(GapLinkSet& gapLinkSet,listOfGapLinkSet) {
            bool connected(false);
            GapLinkSet gapLinkSeenSoFar_temp;
            BOOST_FOREACH (const NumPart& numPart, gapLinkSet) {
                if (gapLinkNumVisited.count(numPart.num) > 0) {
                    if (gapLinkSeenSoFar.count(numPart) > 0) {
                        //cout<<"a part is spread over two components" <<endl;
                        return false; //a part is spread over two components
                    } else {
                        sthgHasChanged = true;
                        gapLinkNumVisited.insert(numPart.num);
                        gapLinkSeenSoFar_temp.insert(numPart);
                        connected = true;
                    }
                }
            }
            gapLinkSeenSoFar.insert(gapLinkSeenSoFar_temp.begin(), gapLinkSeenSoFar_temp.end());
            if (!connected) {
                notYetLinked.push_back(gapLinkSet);
            }
        }
        if (notYetLinked.size() == 0) {
            //cout<<"Linked" <<endl;
            return true;
        } else {
            if (sthgHasChanged) {
                listOfGapLinkSet = notYetLinked;
            } else {
                //cout<<"Components cannot be linked" <<endl;
                return false; //Components cannot be linked
            }
        }
    }

}

typedef pair<const PosType, ComponentInfo> ComponentMapElemType;

PosSet * TargetInfo::returnMinAndMaxMultipleConnectedImage(const PosList *posList,
        vector<PosType> &maxAdditionalNodes, const vector<PosType>& map2ToMin) const {
    //PosType higher=POSTYPE_SPECIAL_VALUE;
    //set<PosType> rootOfConnectedParts;

    //bool higherHasBeenSet=false;
    //PosSet * pIntermediate=new PosSet();

    //map: highest->(connectedSet,listOfHigherAlignedPos,gapLinkSet)
    map<PosType, ComponentInfo> componentMap;

    //Step 1: moving up from each node in posList to the highest admissible node
    //        an admissible node is either a node in posList or a non-aligned node connected to posList through non-aligned nodes 
    BOOST_FOREACH(PosType pos, *posList) {
        PosSet intermediatePos;

        pair<PosType, PosType> higher_highest = getHigherAndHighestWithIntermediate(pos, posList,
                intermediatePos);
        PosType higher = higher_highest.first;
        PosType highest = higher_highest.second;
        //cout<<(int)pos<<" higher:"<<(int)higher<<" highest:"<<(int)highest<<endl;
        ComponentInfo& componentInfo = componentMap[highest];

        componentInfo.connectedSet.insert(intermediatePos.begin(), intermediatePos.end());

        if (isPartOfGappedLink(pos)) {

            PosType num = map2ToMin[pos];
            assert(num!=POSTYPE_SPECIAL_VALUE);

            PosType part = partsMapping[pos];
            assert(part!=POSTYPE_SPECIAL_VALUE);

            componentInfo.gapLinkSet.insert(NumPart(num, part));
        }
        if (higher != POSTYPE_SPECIAL_VALUE) {
            assert(higher == highest);
            componentInfo.listOfHigherAlignedPos.push_back(pos);
        }
    }

    //Step1b: Test if components are connected by gapLinks
    assert(componentMap.size() > 0);
    //cout<<componentMap<<endl;
    if (componentMap.size() > 1) {
        vector<GapLinkSet> listOfGapLinkSet;

        BOOST_FOREACH(ComponentMapElemType& met,componentMap) {
            listOfGapLinkSet.push_back(met.second.gapLinkSet);
        }

        BOOST_FOREACH(GapLinkSet& gapLinkSet,listOfGapLinkSet) {
            //if any gapLinkSet has size 0, reject
            //cout<<"reject gapLinkSet.size()==0"<<endl;
            if (gapLinkSet.size() == 0)
                return NULL;
        }
        //cout<<listOfGapLinkSet<<endl;
        //cout<<"GapCon?:"<<areComponentGapConnected(listOfGapLinkSet)<< (areComponentGapConnected(listOfGapLinkSet)?"con":"discon")<<!areComponentGapConnected(listOfGapLinkSet)<<endl;
        if (!areComponentGapConnected(listOfGapLinkSet)) {
            //cout<<"reject: not gap connected"<<endl;
            return NULL;
        }

    }

    //cout<<"accepted"<<endl;
    vector<vector<PosType> > invDepend = buildInvDepend(depend);

    BOOST_FOREACH (ComponentMapElemType& cmet,componentMap) {
        PosSet& intermediate = cmet.second.connectedSet;
        vector<PosType>& higherImageNodes = cmet.second.listOfHigherAlignedPos;
        // Step 2: downward extension of intermediate through
        //         non-aligned nodes
        BOOST_FOREACH(PosType pos, intermediate) {
            BOOST_FOREACH(PosType child,invDepend[pos]) {
                if (intermediate.count(child) == 0) {
                    extendToNonImageNodes(child, maxAdditionalNodes, invDepend, grouping2);
                }
            }

        }
        // Step 3: reduce intermediate by moving the optional upper non-aligned nodes to maxAdditionalNodes
        moveHigherEmptyNodesToMaxAdditionalNodes(intermediate, higherImageNodes, maxAdditionalNodes);
    }

    PosSet *pConnected = new PosSet;
    BOOST_FOREACH (ComponentMapElemType& cmet,componentMap) {
        pConnected->insert(cmet.second.connectedSet.begin(), cmet.second.connectedSet.end());

    }

    return pConnected;

}

//@-node:fabien.20100901091806.2013:TargetInfo::returnMinAndMaxConnectedImage
//@+node:fabien.20100612171957.1695:TargetInfo::getHigher
PosType TargetInfo::getHigher(PosType pos, const PosList *posList) const {
    PosType h = pos;
    PosType father;
    //assert pos in posList ?
    while (1) {
        father = depend[h];

        if (father == POSTYPE_SPECIAL_VALUE)  //  h is the position of the root
        {
            return h;
        } else if (grouping2[father] == POSTYPE_SPECIAL_VALUE) // the father of h is not an image node
        {
            h = father;
        } else if (contains(posList, father)) // the father of h is an image node and is in posList
                {
            return POSTYPE_SPECIAL_VALUE;
        } else  // the father of h is an image node but is not in posList
        {
            return h;
        }
    }

}
//@nonl
//@-node:fabien.20100612171957.1695:TargetInfo::getHigher
//@+node:fabien.20100901091806.2002:TargetInfo::getHigherWithIntermediate
PosType TargetInfo::getHigherWithIntermediate(PosType pos, const PosList *posList,
        PosSet &intermediate) const {
    PosType h = pos;
    PosType father;
    //assert pos in posList ?
    while (1) {
        father = depend[h];
        intermediate.insert(h);

        if (father == POSTYPE_SPECIAL_VALUE)  //  h is the position of the root
        {
            return h;
        } else if (grouping2[father] == POSTYPE_SPECIAL_VALUE) // the father of h is not an image node
        {
            h = father;
        } else if (contains(posList, father)) // the father of h is an image node and is in posList
                {
            return POSTYPE_SPECIAL_VALUE;
        } else  // the father of h is an image node but is not in posList
        {
            return h;
        }
    }

}

pair<PosType, PosType> TargetInfo::getHigherAndHighestWithIntermediate(PosType pos, const PosList *posList,
        PosSet &intermediate) const {
    PosType h = pos;
    PosType father;
    PosType higher = POSTYPE_SPECIAL_VALUE;
    PosType highest = POSTYPE_SPECIAL_VALUE;
    //assert pos in posList ?

    std::vector<PosType> seenNodes;
    bool higherFound(false);
    bool done(false);
    while (!done) {
        father = depend[h];
        intermediate.insert(h);

        if (father == POSTYPE_SPECIAL_VALUE)  //  h is the position of the root
        {
            if (!higherFound) {
                higher = h;
                higherFound = true;
            }
            highest = h;
            done = true;
        } else if (grouping2[father] == POSTYPE_SPECIAL_VALUE) // the father of h is not an image node
        {
            h = father;
        } else if (contains(posList, father)) // the father of h is an image node and is in posList
                {
            if (!higherFound) {
                higher = POSTYPE_SPECIAL_VALUE;
                higherFound = true;
            }
            h = father;
        } else  // the father of h is an image node but is not in posList
        {
            if (!higherFound) {
                higher = h;
                higherFound = true;
            }
            highest = h;
            done = true;
        }
    }
    return make_pair(higher, highest);

}

PosType TargetInfo::getHigherWithIntermediateForHighest(PosType pos, PosList &intermediate) const {
    PosType h = pos;
    PosType father;
    //assert pos in posList ?
    while (1) {
        father = depend[h];
        intermediate.push_back(h);

        if (father == POSTYPE_SPECIAL_VALUE)  //  h is the position of the root
        {
            return h;
        } else if (grouping2[father] == POSTYPE_SPECIAL_VALUE) // the father of h is not an image node
        {
            h = father;
        } else  // the father of h is an image node but is not in posList
        {
            return h;
        }
    }

}
//@-node:fabien.20100901091806.2002:TargetInfo::getHigherWithIntermediate
//@-node:fabien.20100901091806.2010:Connection
//@+node:fabien.20100901091806.2009:projection
//@+node:fabien.20100612171957.1696:TargetInfo::project
PosList *TargetInfo::project(PosSet &posSet) const {
    PosType pos0, pos, min, pi0, pi;
    PosSet::iterator it;
    PosSet::iterator end = posSet.end();
    PosList *res = new PosList;

    while (!posSet.empty()) {
        it = posSet.begin();
        pos0 = *it;
        posSet.erase(it);
        if (grouping1.at(pos0) == POSTYPE_SPECIAL_VALUE) {
            continue;
        }
        pos = pos0;
        min = pos0;
        while (1) {
            pos = grouping1.at(pos);     //  get next pos in cycle
            if (pos == pos0)
                break;  //  cycle completed: finished
            if (min > pos)
                min = pos;  //  update min
            it = posSet.find(pos);
            if (it == end) {
                delete res;
                return NULL;  //  incomplete group:  no projection
            }
            posSet.erase(it);  // pop out pos
        }
        pi0 = mapping.at(min);
        pi = pi0;
        do {
            res->push_back(pi);
            pi = grouping2.at(pi);
        } while (pi != pi0);
    }
    if (res->size() == 0) {
        delete res;
        return NULL;
    }
    return res;
}
//@nonl
//@-node:fabien.20100612171957.1696:TargetInfo::project
//@+node:fabien.20100612171957.1704:TargetInfo::project_debug
PosList *TargetInfo::project_debug(PosSet &posSet) const {
    PosType pos0, pos, min, pi0, pi;
    PosSet::iterator it;
    PosSet::iterator end = posSet.end();
    PosList *res = new PosList;

    while (!posSet.empty()) {
        it = posSet.begin();
        pos0 = *it;
        posSet.erase(it);
        try {
            if (grouping1.at(pos0) == POSTYPE_SPECIAL_VALUE) {
                continue;
            }
        } catch (std::out_of_range exc) {
            cerr << "grouping1.at(pos0): " << (int) pos0 << ":" << grouping1.size() << endl;
            throw exc;
        }
        pos = pos0;
        min = pos0;

        try {
            while (1) {
                pos = grouping1.at(pos);     //  get next pos in cycle
                if (pos == pos0)
                    break;  //  cycle completed: finished
                if (min > pos)
                    min = pos;  //  update min
                it = posSet.find(pos);
                if (it == end) {
                    delete res;
                    return NULL;  //  incomplete group:  no projection
                }
                posSet.erase(it);  // pop out pos
            }
        } catch (std::out_of_range exc) {
            cerr << "grouping1.at(pos): " << (int) pos << ":" << grouping1.size() << endl;
            throw exc;
        }

        try {
            pi0 = mapping.at(min);
        } catch (std::out_of_range exc) {
            cerr << "mapping.at(min): " << (int) min << ":" << mapping.size() << endl;
            throw exc;
        }

        pi = pi0;

        try {

            do {
                res->push_back(pi);
                pi = grouping2.at(pi);
                if (res->size() > depend.size()) {
                    throw std::out_of_range("bad cycle in grouping2");
                }
            } while (pi != pi0);
        } catch (std::out_of_range exc) {
            cerr << "grouping2.at(pi): " << (int) pi << ":" << grouping2.size() << endl;
            cerr << "A grouping2:" << grouping2 << std::endl;
            throw exc;
        }
    }
    if (res->size() == 0) {
        delete res;
        return NULL;
    }
    return res;
}
//@nonl
//@-node:fabien.20100612171957.1704:TargetInfo::project_debug
//@+node:fabien.20100901091806.2008:TargetInfo::projectAndAddMappings
PosList *TargetInfo::projectAndAddMappings(PosSet &posSet, PosList &mappingsMin) const {
    PosType pos0, pos, min, pi0, pi;
    PosSet::iterator it;
    //PosSet::iterator end=posSet.end();
    PosList *res = new PosList;

    while (!posSet.empty()) {
        it = posSet.begin();
        pos0 = *it;
        posSet.erase(it);
        try {
            if (grouping1.at(pos0) == POSTYPE_SPECIAL_VALUE) {
                continue;
            }
        } catch (std::out_of_range exc) {
            cerr << "grouping1.at(pos0): " << (int) pos0 << ":" << grouping1.size() << endl;
            throw exc;
        }
        pos = pos0;
        min = pos0;

        try {
            while (1) {
                pos = grouping1.at(pos);   //  get next pos in cycle
                if (pos == pos0)
                    break;   //  cycle completed: finished
                if (min > pos)
                    min = pos;  //  update min
                it = posSet.find(pos);
                if (it == posSet.end()) {
                    delete res;
                    return NULL;  //  incomplete group:  no projection
                }
                posSet.erase(it);  // pop out pos
            }
        } catch (std::out_of_range exc) {
            cerr << "grouping1.at(pos): " << (int) pos << ":" << grouping1.size() << endl;
            throw exc;
        }

        mappingsMin.push_back(min);
        try {
            pi0 = mapping.at(min);
        } catch (std::out_of_range exc) {
            cerr << "mapping.at(min): " << (int) min << ":" << mapping.size() << endl;
            throw exc;
        }

        pi = pi0;

        try {

            do {
                res->push_back(pi);
                pi = grouping2.at(pi);
                if (res->size() > depend.size()) {
                    throw std::out_of_range("bad cycle in grouping2");
                }
            } while (pi != pi0);
        } catch (std::out_of_range exc) {
            cerr << "grouping2.at(pi): " << (int) pi << ":" << grouping2.size() << endl;
            cerr << "B grouping2:" << grouping2 << std::endl;
            throw exc;
        }
    }
    if (res->size() == 0) {
        delete res;
        return NULL;
    }
    return res;
}

/*!
 * This method return the image position of its parameter posSet. (Warning:posSet is modified during this operation)
 * It also set mappingsMin to the set of "group-minimal-representative source positions" in posSet
 * map2ToMin is set to map target image positions to the "group-minimal-representative source positions"
 */

PosList *TargetInfo::projectAndAddMappingsWithMap2ToMin(PosSet &posSet, PosList &mappingsMin,
        PosList &map2ToMin) const {
    PosType pos0, pos, min, pi0, pi;
    PosSet::iterator it;
    //PosSet::iterator end=posSet.end();
    PosList *res = new PosList;
    assert(map2ToMin.size() == grouping2.size());
    while (!posSet.empty()) {

        // Poping out one of the position out of posSet into pos0
        it = posSet.begin();
        pos0 = *it;
        posSet.erase(it);

        // if pos0 is not aligned, we continue
        try {
            if (grouping1.at(pos0) == POSTYPE_SPECIAL_VALUE) {
                continue;
            }
        } catch (std::out_of_range exc) {
            cerr << "grouping1.at(pos0): " << (int) pos0 << ":" << grouping1.size() << endl;
            throw exc;
        }

        // pos0 is aligned, so we check whether all of the positions of its group belong to posSet
        pos = pos0;
        min = pos0;

        try {
            while (1) {
                pos = grouping1.at(pos);   //  get next pos in cycle
                if (pos == pos0)
                    break;   //  cycle completed: finished
                if (min > pos)
                    min = pos;  //  update min
                it = posSet.find(pos);
                if (it == posSet.end()) {
                    delete res;
                    return NULL;  //  incomplete group:  no projection
                }
                posSet.erase(it);  // pop out pos
            }
        } catch (std::out_of_range exc) {
            cerr << "grouping1.at(pos): " << (int) pos << ":" << grouping1.size() << endl;
            throw exc;
        }

        // here, pos0 is aligned, and all of the position of its group belong to posSet
        // also, min contains the first position of the group

        mappingsMin.push_back(min);

        // We now proceed to retrieve the aligned target image of the pos0 group
        try {
            pi0 = mapping.at(min);
        } catch (std::out_of_range exc) {
            cerr << "mapping.at(min): " << (int) min << ":" << mapping.size() << endl;
            throw exc;
        }

        // pi0 is the first position of the image group

        // We now proceed to add the positions of the image group to res
        pi = pi0;

        try {

            do {
                res->push_back(pi);
                map2ToMin[pi] = min;
                pi = grouping2.at(pi);
                if (res->size() > depend.size()) {
                    throw std::out_of_range("bad cycle in grouping2");
                }
            } while (pi != pi0);
        } catch (std::out_of_range exc) {
            cerr << "grouping2.at(pi): " << (int) pi << ":" << grouping2.size() << endl;
            cerr << "C grouping2:" << grouping2 << std::endl;
            throw exc;
        }
    }

    // if no element of posSet was aligned, there is no image and we return NULL
    if (res->size() == 0) {
        delete res;
        return NULL;
    }

    // We could find the image of posSet, we return it
    return res;
}

//@nonl
//@-node:fabien.20100901091806.2008:TargetInfo::projectAndAddMappings
//@-node:fabien.20100901091806.2009:projection
//@+node:fabien.20100901091806.2011:Global
//@+node:fabien.20100612171957.1697:TargetInfo::isProjectionConnected
bool TargetInfo::isProjectionConnected(PosSet &posSet) const {
    bool res;

    PosList *image = project_debug(posSet);
    if (image == NULL)  // cannot project
    {
        return false;
    }
    res = isConnected(image);
    delete image;
    return res;
}
//@-node:fabien.20100612171957.1697:TargetInfo::isProjectionConnected
//@+node:fabien.20100901091806.2006:TargetInfo::returnConnectedImageAndMappings
bool TargetInfo::returnConnectedImageAndMappings(PosSet &posSet) const {
    PosSet *res;

    PosList *image = project_debug(posSet);
    if (image == NULL)  // cannot project
    {
        return false;
    }
    res = returnConnectedImage(image);
    delete image;
    return res;
}
//@-node:fabien.20100901091806.2006:TargetInfo::returnConnectedImageAndMappings
//@+node:fabien.20100830110037.1997:TargetInfo::returnProjectionIfConnected
PosList *TargetInfo::returnProjectionIfConnected(PosSet &posSet) const {
    PosList *image = project_debug(posSet);
    if (image == NULL)  // cannot project
    {
        return NULL;
    }
    if (isConnected(image)) {
        return image;
    } else {
        delete image;
        return NULL;
    }
}
//@-node:fabien.20100830110037.1997:TargetInfo::returnProjectionIfConnected
//@-node:fabien.20100901091806.2011:Global
//@-node:fabien.20100612171957.1698:TargetInfo
//@+node:fabien.20100612171957.1699:TargetList
class TargetList: public std::vector<TargetInfo> {
    friend class boost::serialization::access;
    template<class Archive> void serialize(Archive & ar, const unsigned int version) {
        // serialize base class information
        ar & boost::serialization::base_object<std::vector<TargetInfo> >(*this);
    }

public:

    void pushElem() {
        push_back(TargetInfo());
    }
    ;
    void pushDependToLast(PosType pos) {
        back().depend.push_back(pos);
    }
    ;
    void pushGrouping1ToLast(PosType pos) {
        back().grouping1.push_back(pos);
    }
    ;
    void pushGrouping2ToLast(PosType pos) {
        back().grouping2.push_back(pos);
    }
    ;
    void pushMappingToLast(PosType pos) {
        back().mapping.push_back(pos);
    }
    ;

    bool serializeTargetInfoList(const TargetList &lst, const char *filename) {
        std::ofstream ofs(filename, ofstream::binary);
        if (ofs.fail()) {
            std::cerr << "Error creating " << filename << std::endl;
            return false;
        }

        boost::archive::binary_oarchive oa(ofs);
        oa << lst;
        return true;
    }
    ;

    bool saveTo(const char *filename) {
        return serializeTargetInfoList(*this, filename);
    }
    ;

    int isProjectionConnected(unsigned int numTree, PosSet posSet) {
        return (*this)[numTree].isProjectionConnected(posSet);
    }
    ;

    PosList * returnProjectionIfConnected(unsigned int numTree, PosSet posSet) {
        return (*this)[numTree].returnProjectionIfConnected(posSet);
    }
    ;

    ProjectedMappingInfos projectMinMaxAndGiveMapping(TreeIdType idTree,
            const std::vector<PosType>& posList) const;

};

ProjectedMappingInfos TargetList::projectMinMaxAndGiveMapping(TreeIdType idTree,
        const std::vector<PosType>& posList) const {
    //#get list of minPos for projection
    //std::cout<<"IDTree:"<<idTree<<std::endl;
    const TargetInfo& tgtInfo = at(idTree);
    return tgtInfo.projectMinMaxAndGiveMapping(posList);
}

//@+node:fabien.20100612171957.1700:load
TargetList *loadTargetList(const char *filename) {
    std::ifstream ifs(filename, ifstream::binary);
    TargetList *res = new TargetList;

    try {
        boost::archive::binary_iarchive ia(ifs);
        ia >> (*res);
    } catch (boost::archive::archive_exception e) {
        cerr << "Error loading TargetList from file " << filename << endl;
        cerr << "what:" << e.what() << endl;
        delete res;
        return NULL;
    }
    return res;
}
;

//int main(){};

#endif

//@-node:fabien.20100607195349.1650:@thin cTargetInfoList.cpp
//@-leo
