/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/


#include "para_sentence.hpp"
#include "dtree.hpp"
#include "utf8cpp/utf8.h"
#include "TypeDefinitions.hpp"
#include "GlobalTreeletRetrieving2.hpp"
#include "StringConversions.hpp"

#include "TRClientGeneric.hpp"
#include "tr_interface.hpp"
#include "TRResultObject.cpp"
#include "EBMTppConversionCode.hpp"
#include "TRForEBMT.hpp"

#include "TreeletsWildcard.hpp"
///////////////////////////////////////////////////////////////////////

/// \brief These 2 functions provide default parameter objects for the construction of a TR object
PostProcessingParameters get_default_post_processing_parameters();
SearchParameters get_default_search_params(bool use_faster_constrained_sibling_retrieving);

typedef TreeletRetrieverForCPPGeneric<Treelet> TreeletRetrieverForCPP;
typedef TreeletRetrieverForCPPGeneric<TreeletWildcard::Treelet> TreeletRetrieverForCPPWC;

TRInterfacePtr get_default_TR_WC(std::string datapath, std::string side, size_t max_nb_tm_per_pattern, bool use_faster_constrained_sibling_retrieving,
                                    bool use_wildcard_pair_cache)
    {
    PostProcessingParameters pp_params=get_default_post_processing_parameters();
    SearchParameters s_params=get_default_search_params(use_faster_constrained_sibling_retrieving);

    pp_params.maxNumberOfSamplePerTreelet = max_nb_tm_per_pattern;

    s_params.maxNbOfWildcard = 1;
    //pp_params. = ;
    s_params.use_wildcard_pair_cache = use_wildcard_pair_cache;

    TRInterfacePtr pTR=boost::make_shared<TreeletRetrieverForCPPWC>(
                                                            datapath,
                                                            side,
                                                            s_params,
                                                            pp_params
        );
    return pTR;
    }

TRInterfacePtr get_default_TR(std::string datapath, std::string side, size_t max_nb_tm_per_pattern)
    {
    PostProcessingParameters pp_params=get_default_post_processing_parameters();
    SearchParameters s_params=get_default_search_params(false);

    pp_params.maxNumberOfSamplePerTreelet = max_nb_tm_per_pattern;

    TRInterfacePtr pTR=boost::make_shared<TreeletRetrieverForCPP>(
                                                            datapath,
                                                            side,
                                                            s_params,
                                                            pp_params
        );
    return pTR;
    }

/*******************************:
 *   Implementations (from now on, do not look, oh user)
 */



SearchParameters get_default_search_params(bool use_faster_constrained_sibling_retrieving) {
    bool keepTrailingWildcards = false;
    int  maxNbWildcard         = -1;
    int  nbSiblingsLimit       = 0;
    bool noGrouping            = false;

    SearchParameters sParams;
    sParams.nbSiblingsLimit      = nbSiblingsLimit;
    sParams.preserveOrder        = true;
    sParams.skipTrailingWildcard = !keepTrailingWildcards;
    sParams.maxNbOfWildcard      = maxNbWildcard;
    sParams.reduceDominated      = !noGrouping;
    if (use_faster_constrained_sibling_retrieving) {
        sParams.wildcardOnlyAtRoot = true;
        sParams.onlyContiguousSiblings = true;
    } else {
        sParams.wildcardOnlyAtRoot = false;
        sParams.onlyContiguousSiblings = false;
    }
    return sParams;
}

PostProcessingParameters get_default_post_processing_parameters() {
    int  maxSample                         = 200;
    bool outputAllDiscovered               = false;
    bool keepElementaryWildcard            = false;
    bool keepTopWilcardChain               = false;
    bool noFilteringTargetContinuity       = false;
    bool noEnforceProjOfGroundIsContinuous = false;
    bool enforceProjOfGapIsContinuous      = false;
    bool enforceProjOfAllIsContinuous      = false;

    PostProcessingParameters ppParams;
    ppParams.maxNumberOfSamplePerTreelet     = maxSample;
    ppParams.filterOutDominated              = !outputAllDiscovered;
    ppParams.keepElementaryWildcard          = keepElementaryWildcard;
    ppParams.keepTopWilcardChain             = keepTopWilcardChain;
    ppParams.handleOnlyContinuousProjections = !noFilteringTargetContinuity;
    ppParams.enforceProjOfGroundIsContinuous = !noEnforceProjOfGroundIsContinuous;
    ppParams.enforceProjOfGapIsContinuous    = enforceProjOfGapIsContinuous;
    ppParams.enforceProjOfAllIsContinuous    = enforceProjOfAllIsContinuous;
    ppParams.maxTotalNbOfTmToSend=-1;
    return ppParams;
}


class AllMatchingsToEBMTppConverter
    {
    private:
    TMRetrievalResult* pResult;
    PosSet currentSeqOfPos;
    std::string treeId;
    PosList mapp;
    PosSet image;
    PosList addNodes;

    public:
    AllMatchingsToEBMTppConverter(TMRetrievalResult* pResult):pResult(pResult){}

    AllMatchingsToEBMTppConverter():pResult(NULL){}

    void pushTreelet(const PosSet& seqOfPos)
        {
        pResult->setCurrentSeqOfPos(seqOfPos);
        currentSeqOfPos=seqOfPos;
        }

    void pushTreeletWithWildcard(const PosSet& seqOfPosTotal,const PosSet& seqOfPosNW)
        {
        throw std::runtime_error("Not Implemented");
        }

    void pushMatching(const std::string& treeId,
                            const PosList& mapp,
                            const PosSet& image,
                            const PosList& addNodes,
                            const std::vector< std::pair<std::vector<PosType>,std::vector<PosType> > >& matchList)
        {
        // optimizable:
        this->treeId=treeId;
        this->image=image;
        this->mapp=mapp;
        this->addNodes=addNodes;

        }

    void addAdditionalInfosToCurrentMatching(const AdditionalInfos& additional_info,
                                               const PosList& tgt_depend_list,
                                            const std::vector< std::pair<std::vector<PosType>,std::vector<PosType> > >& match_vector)
        {
        pResult->push_one_empty_tm();

        TM& tm=pResult->back();

        TMBuilder::build_tm_in_place(tgt_depend_list, match_vector, additional_info,
                treeId, currentSeqOfPos, mapp, image,
                addNodes, tm);

        //pResult->push_back(tm);
        }

    void finalize()
        {
        pResult->finalize();
        }
    };

class TRClientCPP: public TRClientGeneric<StreamDeserializerGeneric<AllMatchingsToEBMTppConverter,TMRetrievalResult*>, TMRetrievalResult* >, public TRInterface
    {
    public:

    TRClientCPP(const std::string& datapath,const std::string& side,const std::string& host,const std::string& port): TRClientGeneric<StreamDeserializerGeneric<AllMatchingsToEBMTppConverter,TMRetrievalResult*>,TMRetrievalResult* >(datapath,side,host,port)
        {}

    TRResultPtr retrieve(const std::string& labelListAsString, const std::string& dependListAsString)
        {
        sendRequest(labelListAsString,dependListAsString);

        boost::shared_ptr<TMRetrievalResult> result_ptr = boost::make_shared<TMRetrievalResult>();
        readResult(result_ptr.get());
        return result_ptr;
        }
    };

TRInterfacePtr get_client_TR(std::string datapath, std::string side, std::string host, std::string port)
    {
    TRInterfacePtr pTR=boost::make_shared<TRClientCPP>(
                                                            datapath,
                                                            side,
                                                            host,
                                                            port
        );
    return pTR;
    }


