/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*! \file
  \ingroup global_glue_group
  \brief Defining a convenient interface to be called from C++
*/

#include <fstream>
#include <stdlib.h>
#include "tm.hpp"
#include "def.hpp"
#include <boost/smart_ptr.hpp>
#include <set>
#include <boost/ptr_container/ptr_vector.hpp>

#ifndef __TR_INTERFACE_HPP__
#define __TR_INTERFACE_HPP__

/*******************************:
 *   Declaration and Interfaces:
 */

/// \brief TMList is the type of a list of TM (implemented as a std::vector, but could be changed)
typedef boost::ptr_vector<TM> TMList;

/// \brief Interface of the Result object return by a TR search
class TRResult {
public:

    /// \brief virtual destructor
    virtual ~TRResult(){};

    /// \brief get a std::vector of pattern of the input sentence having at least one matched TM
    virtual const std::vector<PatternType>& get_vector_of_pattern() const = 0;

    /// \brief get the number of pattern of the input sentence having at least one matched TM
    virtual size_t get_nb_of_pattern() const = 0;

    /// \brief get the number of TM matched to a given pattern
    virtual size_t get_nb_of_tm_for_pattern(const PatternType& pattern) const = 0;

    /// \brief get the *index*th TM matched to a given pattern
    virtual const TM& get_tm(const PatternType& pattern,size_t index) const = 0;

    /// \brief get a list of all the TM matched to a given pattern
    virtual const TMList& get_tm_list(const PatternType& pattern) const = 0;

    /// \brief get the total number of tm that were found to match the given pattern
    virtual size_t get_real_total_number_of_tm(const PatternType& pattern) const = 0;

    /// \brief add a default-constructed TM to a given pattern, and return a reference to the created empty tm.
    //! (warning1: pattern are only compared as sequences, so <1 3 4> will be considered a different pattern than <1 4 3>
    //! (warning2: the returned reference should be used to initialize the TM -> leaving an un-initialized TM in the result object could be dangerous)
    //! This function is for now mainly provided for adding OOV pseudo-TMs
    virtual TM& push_one_empty_tm_and_return_ref(const PatternType& pattern) =0;

    /// \brief save content to an output stream.
    virtual void save_to_stream( std::ofstream& ofs ) = 0;
};

/// \brief This exception can be thrown when trying to access a pattern that do not exist in the result object
class TMResultException;

/// \brief TRResultPtr is defined as a smart pointer to an object of type TRResult
typedef boost::shared_ptr<TRResult> TRResultPtr;

/// \brief Load a previously saved TRResult object from an input stream.
TRResultPtr load_tr_result_from_stream(std::ifstream& ifs);

class TRInterface
    {
    public:
    virtual ~TRInterface(){};
    virtual TRResultPtr retrieve(const std::string& labelListAsString, const std::string& dependListAsString)=0;
    };

typedef boost::shared_ptr<TRInterface> TRInterfacePtr;

/// \brief create a default TRInterface object.
TRInterfacePtr get_default_TR(std::string datapath, std::string side, size_t max_nb_tm_per_pattern = 200);

/// \brief create a default TRInterface object that can match with wildcards
TRInterfacePtr get_default_TR_WC(std::string datapath, std::string side, size_t max_nb_tm_per_pattern = 200, bool use_faster_constrained_sibling_retrieving = false,
                                        bool use_wildcard_pair_cache = false);

/// \brief create and connectt a client TRInterface object.
TRInterfacePtr get_client_TR(std::string datapath, std::string side, std::string host, std::string port);

/*
 *  End of Declaration and interfaces
 *********************************************************************/

#endif // __TR_INTERFACE_HPP__
