/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "additionals.hpp"
#include <glog/logging.h>
#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>
#include <cassert>
#include <string>
#include <vector>
#include "params.hpp"

void offset_wordindex_vector(const std::vector<WordIndex> &in, std::vector<WordIndex> &out, int offset) {
    size_t size = in.size();
    out.reserve(size);
    for (size_t i = 0; i < size; i++) {
        out.push_back(in[i] + offset);
    }
}

const bool kEnforceMonotonicAdditionalCombination = true;

void offset_wordindex_vector_conditional(const std::vector<WordIndex> &in, 
                                         std::vector<WordIndex> &out, WordIndex insertion_point, int offset,
                                         bool is_reference_insertion, int parent_rel_position,
                                         bool inserted_is_pre, bool add_is_pre) {
    assert(parent_rel_position == -1 || parent_rel_position == 0 || 
           parent_rel_position == 1 || parent_rel_position == -2);
    if (is_reference_insertion) {
        offset--;
        assert(offset >= -1); // if inserted t_string is empty, offset can be -1
    } else {
        assert(offset >= 0);
    }

    size_t size = in.size();
    out.reserve(size);
    for (size_t i = 0; i < size; i++) {
        if (in[i] < insertion_point) {
            out.push_back(in[i]);
        } else if (in[i] == insertion_point) { 
            // the case when an insertion position of the additional is equal to the insertion point
            // is the trickiest.
            // if is_reference_insertion, we can keep the old additional insertion position
            if (is_reference_insertion) {
                assert(parent_rel_position == -2);
                out.push_back(in[i]);
            } else { //otherwise, we have to choose where will be the updated additional insertion position
                assert(parent_rel_position != -2);
                if (parent_rel_position == 0) { 
                    // if parent is the same, they should be both pre or both post
                    // we add an insertion position before and after
                    assert(inserted_is_pre == add_is_pre);
                    if (!kEnforceMonotonicAdditionalCombination) out.push_back(in[i]);
                    out.push_back(in[i] + offset);
                    assert(offset + static_cast<int>(in[i]) >= 0);
                } else if ((inserted_is_pre && !add_is_pre) || 
                            (parent_rel_position == 1 && (inserted_is_pre == add_is_pre))) {
                    // cases where we should just create the insertion position before the newly inserted hypothesis
                    if (!kEnforceMonotonicAdditionalCombination) {
                        out.push_back(in[i]);
                    } else {
                        out.push_back(in[i] + offset);
                    }
                } else if ((!inserted_is_pre && add_is_pre) ||
                            (parent_rel_position == -1 && (inserted_is_pre == add_is_pre))) {
                    // cases where we should just create the insertion position after the newly inserted hypothesis
                    out.push_back(in[i] + offset);
                    assert(offset + static_cast<int>(in[i]) >= 0);
                } else {
                    // We should never get here.
                    LOG(FATAL) << "bug here\n";
                }
            }
        } else {
            out.push_back(in[i] + offset);
            assert(offset + static_cast<int>(in[i]) >= 0);
        }
    }
}

void update_and_combine_additionals_inside(const std::vector<Additional> &additionals,
        int non_terminal_index, std::vector<Additional> &result) {
    assert(non_terminal_index >= 0);
    size_t additionals_size = additionals.size();
    for (size_t i = 0; i < additionals_size; i++) {
        result.push_back(Additional());
        Additional& temp = result.back();
        const Additional& additional = additionals[i];
        temp.set_cnum(additional.cnum());
        temp.set_dpnd(additional.dpnd() + non_terminal_index);
        temp.set_pos(additional.pos());
        offset_wordindex_vector(additional.prelist(), temp.get_modifiable_prelist(), non_terminal_index);
        offset_wordindex_vector(additional.postlist(), temp.get_modifiable_postlist(), non_terminal_index);
    } 
}

int sign_of(int x) {
    if (x < 0) return -1;
    if (x > 0) return 1;
    return 0;
}

void restrict_additionals(const WordIndex added, const WordIndex temp, const WordIndex insertion_point,
                          std::vector<WordIndex>& list) {
    // std::cerr << added << " " << temp << " " << insertion_point << std::endl;
    if (temp < added) {
        for (std::vector<WordIndex>::iterator it = list.begin(); it != list.end(); ) {
            if (*it > insertion_point) {
                // std::cerr << "delete: " << *it << std::endl;
                it = list.erase(it);
            } else {
                it++;
            }
        }
    } else {
        for (std::vector<WordIndex>::iterator it = list.begin(); it != list.end(); ) {
            if (*it <= insertion_point) {
                // std::cerr << "delete: " << *it << std::endl;
                it = list.erase(it);
            } else {
                it++;
            }
        }
    }
}

void update_and_combine_additionals_outside(const std::vector<Additional> &additionals,
        int non_terminal_index, int non_terminal_size, std::vector<Additional> &result,
        bool is_reference_insertion, int index_of_used_additional) {
    if (is_reference_insertion) {
        assert(index_of_used_additional == -1);
    } else {
        assert(index_of_used_additional >= 0);
    }
    size_t additionals_size = additionals.size();
    
    // getting the parent of the inserted additional
    int inserted_dpnd = -2;
    bool inserted_is_pre = false;
    if (!is_reference_insertion) {
        inserted_dpnd = static_cast<int>(additionals[index_of_used_additional].dpnd());
        inserted_is_pre = non_terminal_index <= inserted_dpnd;
        assert(inserted_dpnd >= 0);
    }
    
    for (size_t i = 0; i < additionals_size; i++) {
        if (!is_reference_insertion && i == (size_t) index_of_used_additional) {
            continue;
        }
        result.push_back(Additional());
        Additional& temp = result.back();
        
        const Additional& additional = additionals[i];
        
        // find the relative position of the parent of the inserted additionals
        int parent_rel_position = -2;
        if (!is_reference_insertion) {
            assert(additional.dpnd() >= 0);
            parent_rel_position = sign_of(static_cast<int>(additional.dpnd()) - inserted_dpnd);
        }
        
        temp.set_cnum(additional.cnum());
        if (non_terminal_index <= additional.dpnd()) {
            if (is_reference_insertion) {
                temp.set_dpnd(additional.dpnd() + non_terminal_size - 1);
            } else {
                temp.set_dpnd(additional.dpnd() + non_terminal_size);
            } 
        } else {
            temp.set_dpnd(additional.dpnd());
        }
        temp.set_pos(additional.pos());
        
        offset_wordindex_vector_conditional(additional.prelist(), temp.get_modifiable_prelist(), 
                non_terminal_index, non_terminal_size, is_reference_insertion, parent_rel_position,
                inserted_is_pre, true);
        offset_wordindex_vector_conditional(additional.postlist(), temp.get_modifiable_postlist(), 
                non_terminal_index, non_terminal_size, is_reference_insertion, parent_rel_position,
                inserted_is_pre, false);

        // restrict additionals so as to keep input order
        if (!Params::Translator::reconstruct_final_hypotheses) {
            if (!is_reference_insertion && parent_rel_position == 0) {
                // std::cerr << "hoge" << std::endl;
                if (inserted_is_pre) {
                    restrict_additionals(additionals[index_of_used_additional].cnum(), temp.cnum(),
                                         non_terminal_index, temp.get_modifiable_prelist());
                } else {
                    restrict_additionals(additionals[index_of_used_additional].cnum(), temp.cnum(),
                                         non_terminal_index, temp.get_modifiable_postlist());
                }
            }
        }
    }    
}

// void update_and_combine_additionals_in_place(
        // const std::vector<Additional> &additionals1, const std::vector<Additional> &additionals2, 
        // int non_terminal_index, int non_terminal_size, std::vector<Additional>& result,
        // bool is_reference_insertion) {
    // CALLED_UPDATE_AND_COMBINE_ADDITIONALS++;
    // update_and_combine_additionals_inside(additionals1, non_terminal_index, non_terminal_size, result);
    // update_and_combine_additionals_outside(additionals2, non_terminal_index, non_terminal_size, 
            // result, is_reference_insertion);
            

bool Additional::is_dpnd_consistent() const {
    if (static_cast<int>(dpnd_) < 0) return false;
    BOOST_FOREACH(WordIndex idx, prelist_) {
        if (static_cast<int>(idx) > dpnd_) return false;
    }
    BOOST_FOREACH(WordIndex idx, postlist_) {
        if (static_cast<int>(idx) <= dpnd_) return false;
    }
    return true;
}
            
bool Additional::check_consistency_with_length(size_t length) const {
    BOOST_FOREACH(const WordIndex &index, prelist_) {
        if (index > length) {
            return false;
        }
    }
    BOOST_FOREACH(const WordIndex &index, postlist_) {
        if (index > length) {
            return false;
        }
    }
    return true;
}
 
std::ostream& operator<<(std::ostream& out, const Additional& add) {
    out << "Additional<" << add.cnum_ << ":" << add.dpnd_ << ":";
    BOOST_FOREACH(WordIndex idx, add.prelist_) {
        out << idx << "_";
    }
    out << ":";
    BOOST_FOREACH(WordIndex idx, add.postlist_) {
        out << idx << "_";
    }
    out << ":" << add.pos_;
    out << ">";
    return out;
}
 
AdditionalsList::AdditionalsList(const AdditionalsList& add_list_outside, const AdditionalsList& add_list_inside, 
                    WordIndex insertion_point, int inserted_size, bool is_reference_insertion,
                    int index_of_used_additional) {
    // checking for negative overflow conversion
    assert(static_cast<int>(insertion_point) >= 0); 
    update_and_combine_additionals_inside(add_list_inside, insertion_point, *this);
    update_and_combine_additionals_outside(add_list_outside, insertion_point, inserted_size, 
            *this, is_reference_insertion, index_of_used_additional);
}

bool AdditionalsList::check_consistency_with_length(size_t length) const {
    BOOST_FOREACH(const Additional &add, *this) {
        if (!add.check_consistency_with_length(length)) {
            return false;
        }
    }
    return true;
}

int AdditionalsList::find_compatible_additional_index(const PatternType &list) const {
    for (int i = size() - 1; i >= 0; i--) {
        if (list.count((*this)[i].cnum()) == 0) {
            continue;
        }
        WordIndex cnum = (*this)[i].cnum();
        if (list.count(cnum) != 0) {
            return i;
        }
    }
    return -1;
}

std::ostream& operator<<(std::ostream& out, const AdditionalsList& add_list) {
    out << "A[";
    BOOST_FOREACH(const Additional &add, add_list) {
        out << add << " ";
    }
    out << "]";
    return out;
}

void Additional::hash(std::string *result) const {
    *result += boost::lexical_cast<std::string>(cnum_) + "/";
    for (size_t i = 0; i < prelist_.size(); i++) {
        *result += boost::lexical_cast<std::string>(prelist_[i]) + ",";
    }
    *result += "/";
    for (size_t i = 0; i < postlist_.size(); i++) {
        *result += boost::lexical_cast<std::string>(postlist_[i]) + ",";
    }
}
