/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <set>
#include <string>
#include <vector>
#include <cassert>
#include "glog/logging.h"
#include "def.hpp"
#include "partial_translation_string.hpp"

#ifndef __ADDITIONALS__HPP__
#define __ADDITIONALS__HPP__

class Additional {
    friend std::ostream& operator<<(std::ostream& out, const Additional& add);
    friend class HumanReadableHypothesisSerialization;
  public: 
    Additional() {}
    
    // Constructor merging two Additionale
    Additional(WordIndex cnum, std::vector<WordIndex> prelist, std::vector<WordIndex> postlist, 
            Dpnd dpnd, POS pos)
        : cnum_(cnum), prelist_(prelist), postlist_(postlist), dpnd_(dpnd), pos_(pos) {
        DLOG_IF(FATAL, !is_dpnd_consistent()) << "Created an inconsistent additional: " << *this << "\n";
        assert(is_dpnd_consistent());
        }

    bool is_dpnd_consistent() const;
    bool is_pre_or_post_list_non_empty() const {
        return !(prelist_.size() == 0 && postlist_.size() == 0);
    }

    WordIndex cnum() const {return cnum_;}
    void set_cnum(WordIndex cn) {cnum_ = cn;}
    const std::vector<WordIndex>& prelist() const {return prelist_;}
    const std::vector<WordIndex>& postlist() const {return postlist_;}
    Dpnd dpnd() const {return dpnd_;}
    void set_dpnd(Dpnd dpnd) {dpnd_ = dpnd;}
    const POS &pos() const {return pos_;}
    void set_pos(const POS &pos) {pos_ = pos;}
    
    std::vector<WordIndex>& get_modifiable_prelist() {return prelist_;}
    std::vector<WordIndex>& get_modifiable_postlist() {return postlist_;}
    
    bool check_consistency_with_length(size_t length) const;
    
    inline const std::vector<WordIndex>& get_pre_or_post_list(ParentBondRelation pbr) const {
        if (pbr == PRE) {
            return prelist_;
        } else if (pbr == POST) {
            return postlist_;
        } else {
            // if unsure, arbitrarily choose prelist
            return prelist_;
        }
    }
    void hash(std::string *result) const;
    
    // bool operator<(const Additional& other) const { //for use in std::map
    // if (cnum_ < other.cnum_) return true;
    // if (cnum_ > other.cnum_) return false;
    // if (prelist_.size() < other.prelist_.size()) return true;
    // if (prelist_.size() > other.prelist_.size()) return false;
    // if (postlist_.size() < other.postlist_.size()) return true;
    // if (postlist_.size() > other.postlist_.size()) return false;
    // for (size_t i=0; i<prelist_.size(); i++) {
    // if (prelist_[i]<other.prelist_[i]) return true;
    // if (prelist_[i]>other.prelist_[i]) return false;
    // }
    // for (size_t i=0; i<postlist_.size(); i++) {
    // if (postlist_[i]<other.postlist_[i]) return true;
    // if (postlist_[i]>other.postlist_[i]) return false;
    // }
    // return false; // the additionals are actually identical in that case
    // }
    int compare3(const Additional& other) const { //for use in std::map
        if (cnum_ < other.cnum_) return -1;
        if (cnum_ > other.cnum_) return 1;
        if (prelist_.size() < other.prelist_.size()) return -1;
        if (prelist_.size() > other.prelist_.size()) return 1;
        if (postlist_.size() < other.postlist_.size()) return -1;
        if (postlist_.size() > other.postlist_.size()) return 1;
        for (size_t i = 0; i < prelist_.size(); i++) {
            if (prelist_[i] < other.prelist_[i]) return -1;
            if (prelist_[i] > other.prelist_[i]) return 1;
        }
        for (size_t i = 0; i < postlist_.size(); i++) {
            if (postlist_[i] < other.postlist_[i]) return -1;
            if (postlist_[i] > other.postlist_[i]) return 1;
        }
        return 0; // the additionals are actually identical in that case
    }
    bool is_duplicate(const Additional& other) const {
        if (cnum_ != other.cnum_) return false;
        if (prelist_ != other.prelist_) return false;
        if (postlist_ != other.postlist_) return false; 
        if (dpnd_ != other.dpnd_) return false;
        if (pos_ != other.pos_) return false;
        return true;
    }

  private:
    WordIndex                   cnum_;
    std::vector<WordIndex>      prelist_;
    std::vector<WordIndex>      postlist_;
    Dpnd                        dpnd_;
    POS                         pos_;
};
std::ostream& operator<<(std::ostream& out, const Additional& add);

class AdditionalsList : public std::vector<Additional> {
  public:
    AdditionalsList(const AdditionalsList& add_list_outside, const AdditionalsList& add_list_inside, 
                    WordIndex insertion_point, int inserted_size, bool is_reference_insertion,
                    int index_of_used_additional);
    AdditionalsList(): std::vector<Additional>() {}
    AdditionalsList(const AdditionalsList& add_list): std::vector<Additional>(add_list) {}
    
    std::set<WordIndex> get_set_of_insertion_positions() const {
        std::set<WordIndex> result;
        BOOST_FOREACH(const Additional& add, *this) {
            BOOST_FOREACH(WordIndex i, add.prelist()) {
                result.insert(i);
            }
            BOOST_FOREACH(WordIndex i, add.postlist()) {
                result.insert(i);
            }
        }
        return result;
    }
    
    bool check_consistency_with_length(size_t length) const;
    
    void check_pre_or_post_list_non_empty() const {
#ifndef NDEBUG
        BOOST_FOREACH(const Additional& add, *this) {
            assert(add.is_pre_or_post_list_non_empty());
        }
#endif
    }

    AdditionalsList& operator=(const std::vector<Additional>& add) {
        std::vector<Additional>& this_as_vect = *this;
        this_as_vect = add;
        return *this;
    }

    int find_compatible_additional_index(const PatternType &list) const;
    
    int find_additional_index_for_ref(InputInsertionId requested_ref) const {
        for (int i = size() - 1; i >= 0; i--) {
            WordIndex cnum = (*this)[i].cnum();
            if (requested_ref.compatible_with(cnum)) {
                return i;
            }
        }
        return -1;
    }

    std::vector<int> find_additional_index_vector_for_ref(InputInsertionId requested_ref) const {
        std::vector<WordIndex> cnum_list;
        for (size_t i = 0; i <size(); i++) {
            cnum_list.push_back((*this)[i].cnum());
        }
        return requested_ref.compatible_with_cnum_additionals(cnum_list);
    }
    
    bool remove_add_if_exists_with_cnum(WordIndex idx) {
        for (AdditionalsList::iterator it = begin(); it != end(); it++) {
            if (it->cnum() == idx) {
                erase(it);
                return true;
            }
        }
        return false;
    }
    
    int compare3(const AdditionalsList& other) const { //for use in std::map
        if (size() < other.size()) return -1;
        if (size() > other.size()) return 1;
        for (size_t i = 0; i < size(); i++) {
            int compare_int = (*this)[i].compare3(other[i]);
            if (compare_int != 0) return compare_int;
        }
        return 0; //identical
    }
    bool is_duplicate(const AdditionalsList& other) const {
        if (size() != other.size()) {
            return false;
        }
        for (size_t i = 0; i < size(); i++) {
            if (!(*this)[i].is_duplicate(other[i])) {
                return false;
            }
        }
        return true;
    }
};
std::ostream& operator<<(std::ostream& out, const AdditionalsList& add_list);

#endif //__ADDITIONALS__HPP__
