/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include <map>
#include <limits>
#include <vector>
#include <algorithm>
#include <cmath>
#include "bleu_computer.hpp"

bool StringVectorRange::operator<(const StringVectorRange& other) const {
    if (size() < other.size())
        return true;
    if (size() > other.size())
        return false;
    VS::const_iterator cur1 = start;
    VS::const_iterator cur2 = other.start;
    while (cur1 != end) {
        assert(cur2 != other.end);
        if (*cur1 != *cur2)
            return *cur1 < *cur2;
        ++cur1;
        ++cur2;
    }
    assert(cur2 == other.end);
    return false;
}

bool StringVectorRange::operator==(const StringVectorRange& other) const {
    if (size() != other.size())
        return false;
    VS::const_iterator cur1 = start;
    VS::const_iterator cur2 = other.start;
    while (cur1 != end) {
        assert(cur2 != other.end);
        if (*cur1 != *cur2)
            return false;
        ++cur1;
        ++cur2;
    }
    assert(cur2 == other.end);
    return true;
}

void build_ngram_map(const VS& sentence, std::map<StringVectorRange, int>* map,
        int max_ngram_size) {
    // Warning: the VS refered to by sentence should not be modified for the whole lifetime of map! (or bad things will happen)
    assert(map->size() == 0);
    VS::const_iterator begin = sentence.begin();
    for (int n_size = 1; n_size <= max_ngram_size; n_size++) {
        int last_starting_pos = sentence.size() - n_size;
        for (int start_pos = 0; start_pos <= last_starting_pos; ++start_pos) {
            (*map)[StringVectorRange(begin + start_pos,
                    begin + start_pos + n_size)]++;
        }
    }
}

double compute_BLEU_from_counts(const std::vector<int>& ngram_counts,
        const std::vector<int>& total_ngram_counts, int size_ref,
        int size_cand) {
    assert(ngram_counts.size() == 4);
    assert(total_ngram_counts.size() == 4);

    if (size_cand == 0) { //special case
        return (size_ref == 0) ? 1 : 0;
    }

    double precision_product = 1.0;
    for (int n = 0; n < 4; ++n) {
        assert(ngram_counts[n] <= total_ngram_counts[n]);
        double modified_precision;
        if (total_ngram_counts[n] == 0) {
            modified_precision = 0;
        } else {
            modified_precision = static_cast<double>(ngram_counts[n])
                / total_ngram_counts[n];
        }
        precision_product *= modified_precision;
    }
    double bleu = pow(precision_product, 0.25);

    double bp = 1.0;
    if (size_cand < size_ref) {
        bp = exp(1.0 - (static_cast<double>(size_ref) / size_cand));
    }
    //std::cerr << "bp:" << bp <<"   \r";
    bleu *= bp;
    return bleu;
}

void set_4_vector(std::vector<int>* vect, int i1, int i2, int i3, int i4) {
    vect->clear();
    vect->reserve(4);
    vect->push_back(i1);
    vect->push_back(i2);
    vect->push_back(i3);
    vect->push_back(i4);
}

void increment_vector(std::vector<int>* vect) {
    for (std::vector<int>::iterator it = vect->begin(); it != vect->end();
            ++it) {
        (*it)++;
    }
}

ModifiedNgramPrecisionComputer::ModifiedNgramPrecisionComputer(
        const VS& reference) :
        reference_(reference), max_ngram_(4), is_valid_(true), current_bleu_(
                -std::numeric_limits<double>::max()) {
    build_ngram_map(reference, &ref_count_, max_ngram_);
}

void ModifiedNgramPrecisionComputer::set_current_counts(int ng1, int ng2,
        int ng3, int ng4, int tng1, int tng2, int tng3, int tng4, int size_ref,
        int size_cand) {
    set_4_vector(&current_ngram_counts_, ng1, ng2, ng3, ng4);
    set_4_vector(&current_total_ngram_counts_, tng1, tng2, tng3, tng4);
    current_size_ref_ = size_ref;
    current_size_ = size_cand;
    current_bleu_ = compute_BLEU_from_counts(current_ngram_counts_,
            current_total_ngram_counts_, current_size_ref_, current_size_);
}

std::vector<int> ModifiedNgramPrecisionComputer::compute_modified_ngram_counts(
        const VS& sentence) const {
    assert(is_valid());
    std::vector<int> result(max_ngram_);
    NGramCounts ngram_counts;
    build_ngram_map(sentence, &ngram_counts, max_ngram_);
    BOOST_FOREACH(const NGramCounts::value_type& span_count_pair, ngram_counts) {
        NGramCounts::const_iterator ref_val = ref_count_.find(
                span_count_pair.first);
        if (ref_val != ref_count_.end()) {
            int modified_val = std::min(ref_val->second,
                    span_count_pair.second);
            assert(modified_val >= 1);
            result[span_count_pair.first.size() - 1] += modified_val;
        }
    }
    return result;
}

double ModifiedNgramPrecisionComputer::compute_BLEU_delta(
        const VS& partial_sentence,
        const std::vector<int>& current_ngram_counts,
        const std::vector<int>& current_total_ngram_counts,
        int current_size_ref, int current_size, double ratio_size) const {
    assert(max_ngram_ == 4);
    double current_bleu = compute_BLEU_from_counts(current_ngram_counts,
            current_total_ngram_counts, current_size_ref, current_size);

    std::vector<int> delta_ngram = compute_modified_ngram_counts(
            partial_sentence);

    std::vector<int> new_ngram_counts(current_ngram_counts);
    std::vector<int> new_total_ngram_counts(current_total_ngram_counts);
    for (int n = 0; n < 4; ++n) {
        new_ngram_counts[n] += delta_ngram[n];
        new_total_ngram_counts[n] += partial_sentence.size() - n;
    }

    int delta_size_ref = std::max(1,
            static_cast<int>(reference_.size() * ratio_size));
    int new_size_ref = current_size_ref + delta_size_ref;
    int new_size = current_size + partial_sentence.size();

    double new_bleu = compute_BLEU_from_counts(new_ngram_counts,
            new_total_ngram_counts, new_size_ref, new_size);

    return new_bleu - current_bleu;
}

double ModifiedNgramPrecisionComputer::compute_BLEU_delta_preset(
        const VS& partial_sentence, double ratio_size) const {
    assert(max_ngram_ == 4);
    assert(current_bleu_ != -std::numeric_limits<double>::max());

    std::vector<int> delta_ngram = compute_modified_ngram_counts(
            partial_sentence);

    std::vector<int> new_ngram_counts(current_ngram_counts_);
    std::vector<int> new_total_ngram_counts(current_total_ngram_counts_);
    for (int n = 0; n < 4; ++n) {
        new_ngram_counts[n] += delta_ngram[n];
        new_total_ngram_counts[n] += partial_sentence.size() - n;
    }

    int delta_size_ref = std::max(1,
            static_cast<int>(reference_.size() * ratio_size));
    int new_size_ref = current_size_ref_ + delta_size_ref;
    int new_size = current_size_ + partial_sentence.size();

    double new_bleu = compute_BLEU_from_counts(new_ngram_counts,
            new_total_ngram_counts, new_size_ref, new_size);

    return new_bleu - current_bleu_;
}

double ModifiedNgramPrecisionComputer::compute_BLEU(const VS& sentence,
        bool plus_one) const {
    assert(max_ngram_ == 4);
    std::vector<int> ngrams = compute_modified_ngram_counts(sentence);
    std::vector<int> total_ngrams(max_ngram_);
    for (int n = 0; n < max_ngram_; ++n) {
        total_ngrams[n] = std::max(static_cast<int>(sentence.size()) - n, 0);
    }

    if (plus_one) {
        increment_vector(&ngrams);
        increment_vector(&total_ngrams);
    }
    return compute_BLEU_from_counts(ngrams, total_ngrams, reference_.size(),
            sentence.size());
}

void ModifiedNgramPrecisionComputer::update_counts(const VS& sentence,
        std::vector<int>* modified_ngram_counts,
        std::vector<int>* total_ngram_counts, int* total_size, int *total_ref_size) const {
    assert(max_ngram_ == 4);
    std::vector<int> ngrams = compute_modified_ngram_counts(sentence);
    for (int n = 0; n < max_ngram_; ++n) {
        (*modified_ngram_counts)[n] += ngrams[n];
        (*total_ngram_counts)[n] += std::max(static_cast<int>(sentence.size()) - n, 0);
    }
    *total_size += sentence.size();
    *total_ref_size += reference_.size();
}



