/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef BLEU_COMPUTER_HPP_
#define BLEU_COMPUTER_HPP_

#include <algorithm>
#include <deque>
#include <limits>
#include <map>
#include <string>
#include <vector>
#include "def.hpp"

inline VS segment_sentence_along_spaces(const std::string& sentence) {
    std::stringstream ss(sentence);
    std::istream_iterator<std::string> begin(ss);
    std::istream_iterator<std::string> end;
    return VS(begin, end);
}

double compute_BLEU_from_counts(const std::vector<int>& ngram_counts,
        const std::vector<int>& total_ngram_counts, int size_ref,
        int size_cand);

struct StringVectorRange {
    VS::const_iterator start;
    VS::const_iterator end;

    size_t size() const {return end-start;}

    bool operator<(const StringVectorRange& other) const;
    bool operator==(const StringVectorRange& other) const;

    StringVectorRange(VS::const_iterator start, VS::const_iterator end): start(start), end(end) {}
};

class ModifiedNgramPrecisionComputer {
    typedef std::map<StringVectorRange, int> NGramCounts;
    VS reference_;
    int max_ngram_;
    NGramCounts ref_count_;
    bool is_valid_;

    std::vector<int> current_ngram_counts_;
    std::vector<int> current_total_ngram_counts_;
    int current_size_ref_;
    int current_size_;
    double current_bleu_;

  public:
    explicit ModifiedNgramPrecisionComputer(const VS& reference);

    ModifiedNgramPrecisionComputer(): max_ngram_(4) , is_valid_(false), current_bleu_(-std::numeric_limits<double>::max()) {}

    bool is_valid() const { return is_valid_; }

    void set_current_counts(int ng1, int ng2, int ng3, int ng4, int tng1, int tng2, int tng3, int tng4, int size_ref, int size_cand);

    std::vector<int> compute_modified_ngram_counts(const VS& sentence) const;

    double compute_BLEU_delta(const VS& partial_sentence,
                const std::vector<int>& current_ngram_counts, const std::vector<int>& current_total_ngram_counts,
                int current_size_ref, int current_size, double  ratio_size) const;

    double compute_BLEU_delta_preset(const VS& partial_sentence, double ratio_size) const;

    double compute_BLEU(const VS& sentence, bool plus_one = false) const;

    void update_counts(const VS& sentence, std::vector<int>* modified_ngram_counts, std::vector<int>* total_ngram_counts, int* total_size,
            int *total_ref_size) const;
};

class BleuInfo {
  public:
    friend std::ostream& operator<<(std::ostream& stream, const BleuInfo& info);

    BleuInfo(): modified_ngram_counts_(4), total_ngram_counts_(4), total_size_(0), total_ref_size_(0) {}

    double bleu() const {
        double current_bleu = compute_BLEU_from_counts(modified_ngram_counts_, total_ngram_counts_, total_ref_size_, total_size_);
        return current_bleu;
    }

    void update_with_sentence(const ModifiedNgramPrecisionComputer& ngrams, const VS& sentence) {
        ngrams.update_counts(sentence, &modified_ngram_counts_,
                    &total_ngram_counts_, &total_size_, &total_ref_size_);
    }

    double get_delta_bleu(const ModifiedNgramPrecisionComputer& ngrams, const VS& sentence) const {
        BleuInfo copy(*this);
        copy.update_with_sentence(ngrams, sentence);
        return copy.bleu() - bleu();
    }

    void add(const BleuInfo& other) {
        for (int n = 0; n < 4; ++n) {
            modified_ngram_counts_[n] += other.modified_ngram_counts_[n];
            total_ngram_counts_[n] += other.total_ngram_counts_[n];
        }
        total_size_ += other.total_size_;
        total_ref_size_ += other.total_ref_size_;
    }

    void remove(const BleuInfo& other) {
        for (int n = 0; n < 4; ++n) {
            modified_ngram_counts_[n] -= other.modified_ngram_counts_[n];
            total_ngram_counts_[n] -= other.total_ngram_counts_[n];
            assert(total_ngram_counts_[n] >= 0);
            assert(modified_ngram_counts_[n] >= 0);
        }
        total_size_ -= other.total_size_;
        total_ref_size_ -= other.total_ref_size_;
        assert(total_size_ >= 0);
        assert(total_ref_size_>= 0);
    }

//    st    std::vector<int> modified_ngram_counts_;
//    std::vector<int> total_ngram_counts_;
//    int total_size_;
//    int total_ref_size_;d::string to_string() const {
//        std::stringstream ss;
//
//        ss << bleu() << "(";
//        for (int n = 0; n < 4; ++n) {
//            ss << modified_ngram_counts_[n] << "/" << total_ngram_counts_[n] << " ";
//        }
//        ss << ") [" << total_size_ << ":" << total_ref_size_ <<"]";
//        return ss.str();
//    }
  private:
    std::vector<int> modified_ngram_counts_;
    std::vector<int> total_ngram_counts_;
    int total_size_;
    int total_ref_size_;
};

inline std::ostream& operator<<(std::ostream& stream, const BleuInfo& info) {
    stream << info.bleu() << "(";
            for (int n = 0; n < 4; ++n) {
                stream << info.modified_ngram_counts_[n] << "/" << info.total_ngram_counts_[n] << " ";
            }
            stream << ") [" << info.total_size_ << ":" << info.total_ref_size_ <<"]";
    return stream;
}

class RollingBleuInfo {
  public:
    friend std::ostream& operator<<(std::ostream& stream, const RollingBleuInfo& infos);

    explicit RollingBleuInfo(size_t max_size) : max_size_(max_size) {}

    void add_sentence(const ModifiedNgramPrecisionComputer& ngrams, const VS& sentence) {
        BleuInfo info;
        info.update_with_sentence(ngrams, sentence);
        add_info(info);
    }

    void add_info(const BleuInfo& info) {
        if (rolling_infos_.size() >= max_size_) {
            remove_first();
        }
        rolling_infos_.push_back(info);
        current_bleu_info_.add(info);
    }

    void remove_first() {
        current_bleu_info_.remove(rolling_infos_.front());
        rolling_infos_.pop_front();
    }

    const BleuInfo& current_bleu_info() {
        return current_bleu_info_;
    }

  private:
    std::deque<BleuInfo> rolling_infos_;
    BleuInfo current_bleu_info_;
    size_t max_size_;
};

inline std::ostream& operator<<(std::ostream& stream, const RollingBleuInfo& infos) {
    stream << "bleu on last " << infos.rolling_infos_.size() << " sentences : " << infos.current_bleu_info_;
    return stream;
}

inline double compute_bleu_of_single_sentence(const VS& sentence, const VS& reference) {
    ModifiedNgramPrecisionComputer ngram_computer(reference);
    BleuInfo bleu_info;
    bleu_info.update_with_sentence(ngram_computer, sentence);
    return bleu_info.bleu();
}

#endif /* BLEU_COMPUTER_HPP_ */
