/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include <algorithm>
#include <cassert>
#include <set>
#include <string>
#include <vector>

#include "augmented_pattern.hpp"
#include "../serialization_helpers.hpp"

/****************************************************************
 *  AugmentedPattern class implementation
 */ 

ConnectedAugmentedPattern::ConnectedAugmentedPattern(const PatternType& pat, const Dtree& i_dtree): 
                        PatternType(pat),
                        parent((WordIndex)(-1)),
                        root((WordIndex)(-1)) {
    // std::cerr<<"creating augmented pattern"<<std::endl;
    BOOST_FOREACH(WordIndex index, pat) {
        WordIndex parent_index = i_dtree.get_parent_index(index);
        // std::cerr<<"index: "<<index<<" parent:"<<parent_index<<" count:"<<pat.count(parent_index)<<std::endl;
        if (parent_index == (WordIndex)(-1) || pat.count(parent_index) == 0) {
            // check root hasn't been set before
            assert(root == (WordIndex)(-1)); 
            parent = parent_index;
            root = index;
            assert(root != parent);
        }
        BOOST_FOREACH(WordIndex child_index, i_dtree.get_children_as_wordindex(index)) {
            if (pat.count(child_index) == 0) {
                outer_frontier.insert(child_index);
                inner_frontier[index].insert(child_index);
            }
        }
    }
    assert(check_if_consistent());
}

ConnectedAugmentedPattern::ConnectedAugmentedPattern(const ConnectedAugmentedPattern& pat_above, const ConnectedAugmentedPattern& pat_below): 
        PatternType(pat_above),
        parent(pat_above.parent),
        root(pat_above.root),
        inner_frontier(pat_above.inner_frontier),
        outer_frontier(pat_above.outer_frontier) {
    assert(pat_below.parent != (WordIndex)(-1));
    assert(!pat_above.is_complete());
    assert(pat_above.inner_frontier.count(pat_below.parent) == 1);
    assert(pat_above.outer_frontier.count(pat_below.root) == 1);
    
    insert(pat_below.begin(), pat_below.end());
    
    inner_frontier[pat_below.parent].erase(pat_below.root);
    if (inner_frontier[pat_below.parent].size() == 0) {
        inner_frontier.erase(pat_below.parent);
    }
    outer_frontier.erase(pat_below.root);
    inner_frontier.insert(pat_below.inner_frontier.begin(),
                                pat_below.inner_frontier.end());
    outer_frontier.insert(pat_below.outer_frontier.begin(),
                                pat_below.outer_frontier.end());
    assert(check_if_consistent());
}

bool ConnectedAugmentedPattern::is_complete() const {
    bool result = (inner_frontier.size() == 0);
    assert(result || outer_frontier.size() > 0);
    return result;
}

WordIndex ConnectedAugmentedPattern::get_left_most_outer_frontier_index() const {
    assert(!is_complete());
#ifndef NDEBUG
    // const std::set<WordIndex>&dummy = *this; 
    // checking we are actually using a derivation of std::set 
#endif  // NDEBUG
    WordIndex result=*(outer_frontier.begin());
    assert(result!= (WordIndex) -1);
    return result;
}

WordIndex ConnectedAugmentedPattern::get_left_most_outer_frontier_index_or_nothing() const {
    if (outer_frontier.size() == 0) return static_cast<WordIndex>(-1);
    WordIndex result=*(outer_frontier.begin());
    assert(result!= (WordIndex) -1);
    return result;
}

WordIndex ConnectedAugmentedPattern::get_root() const {
    assert(root != (WordIndex)(-1));
    return root;
}

bool ConnectedAugmentedPattern::contains_root() const {
  return parent == (WordIndex)(-1);
}

bool ConnectedAugmentedPattern::check_if_consistent() const {
    if (root == (WordIndex)(-1)) return false;
    if (root == parent) return false;
    if (count(parent) != 0) return false;
    if (count(root) != 1) return false;
    size_t size_outer_frontier = 0;
    BOOST_FOREACH(const WordIndexToWordIndexSetMap::value_type& index_set_pair, inner_frontier) {
        if (index_set_pair.first == (WordIndex)(-1)) return false;
        if (count(index_set_pair.first) == 0) return false;
        if (index_set_pair.second.size() == 0) return false;
        BOOST_FOREACH(WordIndex index_outer, index_set_pair.second) {
            if (outer_frontier.count(index_outer) != 1) return false;
        }
        size_outer_frontier += index_set_pair.second.size();
    }
    if (size_outer_frontier != outer_frontier.size()) return false;
    BOOST_FOREACH(WordIndex index_outer, outer_frontier) {
        if (count(index_outer) == 1) return false;
        if (index_outer == (WordIndex)(-1)) return false;
    }
    return true;
}

void ConnectedAugmentedPattern::serialize_one(std::ostream *stream) const {
    const ConnectedAugmentedPattern& a_pattern = *this;
    serialize_wordindex_container(a_pattern, ' ', *stream);
    // BOOST_FOREACH(WordIndex pos, a_pattern) {
        // stream << static_cast<int>(pos) <<" ";
    // }
    *stream << "|";
    *stream << static_cast<int>(a_pattern.root);
    *stream << "|" << static_cast<int>(a_pattern.parent);
    *stream << "|";
    serialize_wordindex_container(a_pattern.outer_frontier, ' ', *stream);
    // BOOST_FOREACH(WordIndex pos, a_pattern.outer_frontier) {
        // stream << static_cast<int>(pos) <<" ";
    // }
    *stream << "|";
    size_t cursor_ext = 0;
    BOOST_FOREACH(const WordIndexToWordIndexSetMap::value_type& idx_set_pair, a_pattern.inner_frontier) {
        *stream << static_cast<int>(idx_set_pair.first) << ":";
        size_t cursor = 0;
        BOOST_FOREACH(WordIndex pos, idx_set_pair.second) {
            *stream << static_cast<int>(pos);
            if (cursor + 1 < idx_set_pair.second.size()) {
                *stream << "_";
            }
            cursor++;
        }
        if (cursor_ext + 1 < a_pattern.inner_frontier.size()) {
            *stream << " ";
        }
        cursor_ext++;
    }
}

void ConnectedAugmentedPattern::serialize(std::ostream *stream) const {
    *stream << "P|";
    serialize_one(stream);
    *stream << "\n";
}

void deserialize_inner_frontier_element(char *string, WordIndexToWordIndexSetMap* inner_frontier) {
    char *rest = NULL;
    
    WordIndex inner_index = static_cast<WordIndex>(read_one_int_delimited_by(string, ':', &rest));
    
    char* token = my_strtok_r(NULL, ':', &rest);
    read_wordindex_set(token, '_', &((*inner_frontier)[inner_index]));
    
    LOG_IF(FATAL, rest != NULL) << "Parsing Error: inner of pattern has too many colon.\n";
}

void deserialize_inner_frontier(char *string, WordIndexToWordIndexSetMap* inner_frontier) {
    if (*string == '\0') return; // empty inner frontier
    char *rest = NULL;
    
    char* token = my_strtok_r(string, ' ', &rest);
    //if (*token == '\0') return; // empty inner frontier
    deserialize_inner_frontier_element(token, inner_frontier);
    
    while ((token = my_strtok_r(NULL, ' ', &rest))) {
        deserialize_inner_frontier_element(token, inner_frontier);
    }    
}

PatternType ConnectedAugmentedPattern::deserialize_one(char* serialized_pattern, 
                                                                    std::vector<ConnectedAugmentedPattern>* augmented_pattern_vect) {
    VLOG(2) << "deserializing augmented pattern:'" << serialized_pattern << "'" << std::endl;
    
    PatternType pat;
    
    char* rest = NULL;
    
    char* token = my_strtok_r(serialized_pattern, '|', &rest);
    LOG_IF(FATAL, token == NULL) << "Error Parsing Pattern" << "\n";
    read_wordindex_set(token, ' ', &pat);
    
    augmented_pattern_vect->push_back(ConnectedAugmentedPattern());
    augmented_pattern_vect->back().insert(pat.begin(), pat.end());
    
    augmented_pattern_vect->back().root = static_cast<WordIndex>(read_one_int_delimited_by(NULL, '|', &rest));
    augmented_pattern_vect->back().parent = static_cast<WordIndex>(read_one_int_delimited_by(NULL, '|', &rest));
    
    token = my_strtok_r(NULL, '|', &rest);
    LOG_IF(FATAL, token == NULL) << "Error Parsing Pattern rest:" << *rest << "\n";
    read_wordindex_set(token, ' ', &(augmented_pattern_vect->back().outer_frontier));
    
    token = my_strtok_r(NULL, '|', &rest);
    LOG_IF(FATAL, token == NULL) << "Error Parsing Pattern" << "\n";
    deserialize_inner_frontier(token, &(augmented_pattern_vect->back().inner_frontier));
    
    LOG_IF(FATAL, rest != NULL) << "Error Parsing Pattern" << "\n";

    return pat;
}

PatternType ConnectedAugmentedPattern::deserialize(char* serialized_pattern, 
                                                                    std::vector<ConnectedAugmentedPattern>* augmented_pattern_vect) {
    VLOG(2) << "deserializing augmented pattern:'" << serialized_pattern << "'" << std::endl;
    
    PatternType pat;
    
    char* rest = NULL;
    
    char* token = my_strtok_r(serialized_pattern, '|', &rest);
    LOG_IF(FATAL, token == NULL) << "Error Parsing Pattern" << "\n";
    LOG_IF(FATAL, token[0] != 'P' || token[1] != '\0') << "Error Parsing Pattern" << "\n";
    
    return deserialize_one(rest, augmented_pattern_vect);
}

std::ostream& operator<<(std::ostream& stream, const ConnectedAugmentedPattern& pattern) {
    const std::set<WordIndex>& this_as_set = pattern;
    stream << "P<" << this_as_set << "| " << pattern.outer_frontier << ">";
    return stream;
}

/**********************************************************************************
 *  LatterallyDisconnectedAugmentedPattern
 */
 
template<class Set1, class Set2> 
bool is_disjoint(const Set1 &set1, const Set2 &set2) {
    if (set1.empty() || set2.empty()) return true;

    typename Set1::const_iterator 
        it1 = set1.begin(), 
        it1End = set1.end();
    typename Set2::const_iterator 
        it2 = set2.begin(), 
        it2End = set2.end();

    if (*it1 > *set2.rbegin() || *it2 > *set1.rbegin()) return true;

    while (it1 != it1End && it2 != it2End) {
        if (*it1 == *it2) {
            return false;
        }
        if (*it1 < *it2) { 
            it1++; 
        } else { 
            it2++;
        }
    }
    return true;
}

std::vector<std::set<WordIndex> > find_connected_parts(const PatternType& pat, const Dtree& i_dtree) {
    std::set<WordIndex> remaining = pat;
    std::vector<std::set<WordIndex> > result;
    
    while (remaining.size() > 0) {
        std::set<WordIndex> connected_part;
        std::vector<WordIndex> stack;
        stack.push_back(*remaining.begin());
        remaining.erase(remaining.begin());
        while (stack.size() > 0) {
            WordIndex idx = stack.back();
            stack.pop_back();
            assert(connected_part.count(idx) == 0);
            connected_part.insert(idx);
            WordIndex parent_index = i_dtree.get_parent_index(idx);
            if (remaining.count(parent_index) == 1) {
                assert(remaining.count(parent_index) == 1);
                stack.push_back(parent_index);
                remaining.erase(parent_index);
            }
            BOOST_FOREACH(WordIndex child_index, i_dtree.get_children_as_wordindex(idx)) {
                if (remaining.count(child_index) == 1) {
                    assert(remaining.count(child_index) == 1);
                    stack.push_back(child_index);
                    remaining.erase(child_index);
                }
            }
        }
        result.push_back(connected_part);
    }
    return result;
}

class CompareByRoot {
  public:
    bool operator()(const ConnectedAugmentedPattern& pat1, const ConnectedAugmentedPattern& pat2) {
        return pat1.get_root() < pat2.get_root();
    }
};

LatterallyDisconnectedAugmentedPattern::LatterallyDisconnectedAugmentedPattern(const PatternType& pat, const Dtree& i_dtree)
        : PatternType(pat) {
    std::vector<std::set<WordIndex> > parts = find_connected_parts(pat, i_dtree);
    // std::cerr<<"creating augmented pattern"<<std::endl;
    BOOST_FOREACH(const PatternType& connected, parts) {
        connected_patterns_list_.push_back(ConnectedAugmentedPattern(connected, i_dtree));
    }
    std::sort(connected_patterns_list_.begin(), connected_patterns_list_.end(), CompareByRoot());
    assert(check_if_consistent());
}

LatterallyDisconnectedAugmentedPattern::LatterallyDisconnectedAugmentedPattern(const LatterallyDisconnectedAugmentedPattern& pat_above, 
                    const LatterallyDisconnectedAugmentedPattern& pat_below): PatternType(pat_above) {
    assert(is_disjoint(pat_below, pat_above));
    insert(pat_below.begin(), pat_below.end());
    connected_patterns_list_ = pat_above.connected_patterns_list_;

    BOOST_FOREACH(const ConnectedAugmentedPattern& connected_below, pat_below.connected_patterns_list_) {
        WordIndex root = connected_below.get_root();
        size_t i = pat_above.get_index_of_connected_pattern_with_outer_frontier_element(root);
        const ConnectedAugmentedPattern& connected_above = connected_patterns_list_[i];
        connected_patterns_list_[i] = ConnectedAugmentedPattern(connected_above, connected_below);
    }
    assert(check_if_consistent());
}

std::vector<WordIndex> LatterallyDisconnectedAugmentedPattern::get_roots() const {
    std::vector<WordIndex> result;
    BOOST_FOREACH(const ConnectedAugmentedPattern& pat, connected_patterns_list_) {
        result.push_back(pat.get_root());
    }
    return result;
}

WordIndex LatterallyDisconnectedAugmentedPattern::get_root() const {
    assert(connected_patterns_list_.size() >= 1);
    return connected_patterns_list_[0].get_root();
}

bool LatterallyDisconnectedAugmentedPattern::is_complete() const {
    BOOST_FOREACH(const ConnectedAugmentedPattern& pat, connected_patterns_list_) {
        if (!pat.is_complete()) return false;
    }
    return true;
}

bool LatterallyDisconnectedAugmentedPattern::contains_root() const {
    bool result = false;
    BOOST_FOREACH(const ConnectedAugmentedPattern& pat, connected_patterns_list_) {
        if (pat.contains_root()) {
            assert(result == false);
            result = true;
        }
    }
    return result;
}

WordIndex LatterallyDisconnectedAugmentedPattern::get_left_most_outer_frontier_index() const {
    assert(!is_complete());
    WordIndex result = static_cast<WordIndex>(-1);
    bool first_iter = true;
    BOOST_FOREACH(const ConnectedAugmentedPattern& pat, connected_patterns_list_) {
        WordIndex left_most = pat.get_left_most_outer_frontier_index_or_nothing();
        if (left_most == static_cast<WordIndex>(-1)) continue;
        if (first_iter || left_most < result) result = left_most;
        first_iter = false;
    }
    assert(!first_iter);
    assert(result != static_cast<WordIndex>(-1));
    return result;
}

size_t LatterallyDisconnectedAugmentedPattern::get_index_of_connected_pattern_with_outer_frontier_element(WordIndex idx) const {
    for (size_t i = 0; i < connected_patterns_list_.size(); i++) {
        const ConnectedAugmentedPattern& connected = connected_patterns_list_[i];
        if (connected.get_outer_frontier().count(idx) != 0) {
            return i;
        }
    }
    LOG(FATAL) << "Bug here!\n"; // we should only call this function when an index can be found
}

bool LatterallyDisconnectedAugmentedPattern::check_if_consistent() const {
    std::set<WordIndex> cumulated_positions;
    std::set<WordIndex> cumulated_outer_frontier;
    WordIndex previous_root = static_cast<WordIndex>(-1);
    bool first_iter = true;
    for (size_t i = 0; i < connected_patterns_list_.size(); i++) {
        const ConnectedAugmentedPattern& connected = connected_patterns_list_[i];
        if (!connected.check_if_consistent()) return false;
        BOOST_FOREACH(WordIndex idx, connected) {
            if (cumulated_positions.count(idx) != 0) return false; // non-overlapping test
            cumulated_positions.insert(idx);
        }
        BOOST_FOREACH(WordIndex idx, connected.outer_frontier) {
            if (cumulated_outer_frontier.count(idx) != 0) return false; // non-overlapping test
            cumulated_outer_frontier.insert(idx);
        }
        WordIndex root = connected.get_root();
        if (!first_iter && root <= previous_root) return false; //test that roots are strictly increasing
        previous_root = root;
        first_iter = false;
    }
    if (!is_disjoint(cumulated_positions, cumulated_outer_frontier)) return false;
    const PatternType& as_pattern_type = *this;
    if (as_pattern_type != cumulated_positions) return false;
    return true;
}

void LatterallyDisconnectedAugmentedPattern::serialize(std::ostream *stream) const {
    *stream << "P|";
    bool first = true;
    BOOST_FOREACH(const ConnectedAugmentedPattern& pat, connected_patterns_list_) {
        if (!first) {
            *stream << "U";
        }
        pat.serialize_one(stream);
        first = false;
    }
    *stream << "\n";
}

PatternType LatterallyDisconnectedAugmentedPattern::deserialize(char* serialized_pattern, 
                                           std::vector<LatterallyDisconnectedAugmentedPattern>* augmented_pattern_vect) {
    VLOG(2) << "deserializing augmented pattern:'" << serialized_pattern << "'" << std::endl;
    
    PatternType pat;
    
    char* rest = NULL;
    
    char* token = my_strtok_r(serialized_pattern, '|', &rest);
    LOG_IF(FATAL, token == NULL) << "Error Parsing Pattern" << "\n";
    LOG_IF(FATAL, token[0] != 'P' || token[1] != '\0') << "Error Parsing Pattern" << "\n";
    
    augmented_pattern_vect->push_back(LatterallyDisconnectedAugmentedPattern());
    
    while (char* connected_as_string = my_strtok_r(NULL, 'U', &rest)) {
        PatternType pat_partial = ConnectedAugmentedPattern::deserialize_one(connected_as_string, &augmented_pattern_vect->back().connected_patterns_list_);
        assert(is_disjoint(pat_partial, pat));
        pat.insert(pat_partial.begin(), pat_partial.end());
    }
    
    PatternType& as_pattern_type =  augmented_pattern_vect->back();
    as_pattern_type = pat;
    assert(augmented_pattern_vect->back().check_if_consistent());
    return pat;
}

std::ostream& operator<<(std::ostream& stream, const LatterallyDisconnectedAugmentedPattern& pattern) {
    stream << "LDP{";
    BOOST_FOREACH(const ConnectedAugmentedPattern& pat, pattern.connected_patterns_list_) {
        stream << pat << ", ";
    }
    stream << "}";
    return stream;
}
