/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*! \file
    \ingroup translation_core
    \brief Define the Augmented Pattern Class
*/

#ifndef __AUGMENTED_PATTERN_HPP__
#define __AUGMENTED_PATTERN_HPP__

#include <glog/logging.h>
#include <boost/foreach.hpp>
#include <map>
#include <set>
#include <vector>
#include "../def.hpp"
#include "../dtree.hpp"

typedef std::map<WordIndex, std::set<WordIndex> > WordIndexToWordIndexSetMap;

/*! \brief AugmentedPattern is a class deriving from PatternType.
 *  It allows to keep track of the parents, roots, inner and outer frontiers of each pattern.
 * In the following graph, we can see an input tree containing a pattern "a b c e".
 * The root of the pattern is a. The parent of the pattern is h. 
 * The outer frontier of the pattern are the nodes not in the pattern but child of
 * nodes in the pattern (d and f). The inner frontier are the nodes of the pattern
 * having childs in the outer frontier (c and b).
 * (Graphviz dot might set the nodes in a strange layout, so the tree might have an unusual representation).
 \dot
 digraph{
    compound=true;
    subgraph cluster_pattern
        {
        label="Pattern";
        a->b; a->c; c->e;
        }
    subgraph cluster_outer
        {
        label="Outer Frontier";
        c->d; b->f;
        }
    h->a; h->i;
    d->g;
    }
  \enddot
 */
class ConnectedAugmentedPattern: public PatternType {
  public:
    //! \brief Construct an AugmentedPattern from a PatternType and an input tree.
    ConnectedAugmentedPattern(const PatternType& pat, const Dtree& i_dtree);
    
    //! Default constructor (needed mainly so that AugmentedPattern can be inserted in a map)
    ConnectedAugmentedPattern() {}
    
    /*! \brief Construct an AugmentedPattern from two AugmentedPattern.
     *  pat_above and pat_below should be compatible, in the sense that the root of pat_below
     *  should be in the outer frontier of pat_above
     */
    ConnectedAugmentedPattern(const ConnectedAugmentedPattern& pat_above, const ConnectedAugmentedPattern& pat_below);

    /*! \brief Check if the pattern is complete.
     *  Here, complete means that the pattern cover a full subtree of the input tree.
     */    
    bool is_complete() const;

    /*! \brief Get the lowest WordIndex of the outer frontier.
     *  Note: The outer frontier should not be empty.
     */     
    WordIndex get_left_most_outer_frontier_index() const;

    
    
    /*! \brief Get the root of the pattern. */   
    WordIndex get_root() const;
    
    bool contains_root() const;

    const PatternType& to_pattern() const {
        return *this;
    }
    
    bool is_discontiguous() const {
        return false;
    }
    
    const std::set<WordIndex>& get_outer_frontier() const {
        return outer_frontier;
    }
    
    bool additional_compatibility_test(const ConnectedAugmentedPattern& pat) const {
        return true;
    }
    
    void serialize(std::ostream *stream) const;
    
    static PatternType deserialize(char* serialized_pattern, std::vector<ConnectedAugmentedPattern>* augmented_pattern_vect);
    
  private:
    WordIndex parent;
    WordIndex root;
    WordIndexToWordIndexSetMap inner_frontier;
    std::set<WordIndex> outer_frontier;
    
    void serialize_one(std::ostream *stream) const;
    static PatternType deserialize_one(char* serialized_pattern, 
                                std::vector<ConnectedAugmentedPattern>* augmented_pattern_vect);
    
    bool check_if_consistent() const;
    WordIndex get_left_most_outer_frontier_index_or_nothing() const;
    //friend class HumanReadableHypothesisSerialization;
    friend std::ostream& operator<<(std::ostream& stream, const ConnectedAugmentedPattern& pattern);
    friend class LatterallyDisconnectedAugmentedPattern;
};

/*! \brief LatterallyDisconnectedAugmentedPattern is a class that represents the concept
 *  of "laterally disconnected" patterns. It is implemented as a vector of AugmentedPattern.
 * Latterally disconnected means that the connected parts should be "cousins" but not ancestors/descendant.
 */

class LatterallyDisconnectedAugmentedPattern: public PatternType {
  public:  
    LatterallyDisconnectedAugmentedPattern(const LatterallyDisconnectedAugmentedPattern& pat_above, 
                        const LatterallyDisconnectedAugmentedPattern& pat_below);
    
    LatterallyDisconnectedAugmentedPattern(const PatternType& pat, const Dtree& i_dtree);
    
    LatterallyDisconnectedAugmentedPattern() {}
    
    std::vector<WordIndex> get_roots() const;
    
    WordIndex get_root() const;
    
    bool is_complete() const;

    bool contains_root() const;
    
    WordIndex get_left_most_outer_frontier_index() const;
    
    void serialize(std::ostream *stream) const;
    
    PatternType to_pattern() const {
        assert(connected_patterns_list_.size() >= 1);
        if (connected_patterns_list_.size() == 1) {
            return connected_patterns_list_[0].to_pattern();
        } else {
            PatternType result;
            size_t total_size = 0;
            BOOST_FOREACH(const ConnectedAugmentedPattern& pat, connected_patterns_list_) {
                result.insert(pat.begin(), pat.end());
                total_size += pat.size();
            }
            assert(total_size == result.size());
            return result;
        }
    }
    
    bool is_discontiguous() const {
        return connected_patterns_list_.size() > 1;
    }
    
    bool additional_compatibility_test(const LatterallyDisconnectedAugmentedPattern& pat) const {
        BOOST_FOREACH(WordIndex root, get_roots()) {
            if (!pat.is_in_outer_frontier(root)) return false;
        }        
        return true;
    }
    
    bool is_in_outer_frontier(WordIndex idx) const {
        BOOST_FOREACH(const ConnectedAugmentedPattern& pat, connected_patterns_list_) {
            if (pat.outer_frontier.count(idx) >= 1) return true;
        }
        return false;
    }
    
    static PatternType deserialize(char* serialized_pattern, std::vector<LatterallyDisconnectedAugmentedPattern>* augmented_pattern_vect);
    
  private:
    std::vector<ConnectedAugmentedPattern> connected_patterns_list_;
    
    bool check_if_consistent() const;
    size_t get_index_of_connected_pattern_with_outer_frontier_element(WordIndex idx) const;
    friend std::ostream& operator<<(std::ostream& stream, const LatterallyDisconnectedAugmentedPattern& pattern);
};

std::ostream& operator<<(std::ostream& stream, const LatterallyDisconnectedAugmentedPattern& pattern);

//typedef ConnectedAugmentedPattern AugmentedPattern;
typedef LatterallyDisconnectedAugmentedPattern AugmentedPattern;

#endif //__AUGMENTED_PATTERN_HPP__

