/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "base_lattice.hpp"
#include <map>
#include <vector>
#include <string>
#include <set>

namespace latt {

std::ostream& operator<<(std::ostream& stream, const PinedEdge& edge) {
    stream << edge.id_start_ << " -> " << edge.edge_ << " -> " << edge.id_end_;
    return stream;
}

std::ostream& operator<<(std::ostream& stream, const std::vector<PinedEdge*>& vect) {
    for (int i = 0; i < static_cast<int>(vect.size()); ++i) {
        stream << i << " : " << *vect[i] << ", ";
    }
    return stream;
}

void Path::add_vertex(VertexId vertex) {
    //std::cerr << "Path push vertex " << vertex << "\n";
    vertice_.push_back(vertex);
    elements_sequence_.push_back(PathElement(vertice_.size()-1, PathElement::VERTEX));
//    if (vertex == kInitialVertex) {
//        level_ -= 1;
//        assert(level_ >= 0);
//        if (level_ == 0) vertex_sequence_.push_back(VertexLevelPair(vertex, level_));
//    } else if (vertex == kFinalVertex) {
//        if (level_ == 0) vertex_sequence_.push_back(VertexLevelPair(vertex, level_));
//        level_ += 1;
//    } else {
//        vertex_sequence_.push_back(VertexLevelPair(vertex, level_));
//    }
}

void Path::add_edge(const FreeEdge& edge, double score, StateFeatures st_feat) {
    // std::cerr << "Path push edge " << edge << "\n";
    edges_.push_back(EdgeStats(edge, st_feat, score));
    elements_sequence_.push_back(PathElement(edges_.size()-1, PathElement::EDGE));
//    edge_sequence_.push_back(edge);
//    edge_scores_.push_back(score);
//    st_feat_sequence_.push_back(st_feat);
    cumulative_score_ += score;
}

//void Path::add_offset(double score) {
//    offset_sequence_.push_back(score);
//    cumulative_score_ += score;
//}

void Path::set_score(double score) {
    score_ = score;
}

template<typename Visitor>
void Path::accept(Visitor* visitor) const {
    assert(!is_special());
    for (int i = static_cast<int>(elements_sequence_.size()) - 1; i >= 0; --i) {
        const PathElement& elem = elements_sequence_[i];
        switch (elem.type_) {
        case PathElement::VERTEX:
            visitor->visit_vertex(vertice_[elem.index_]);
            break;
        case PathElement::EDGE:
            visitor->visit_edge(edges_[elem.index_]);
            break;
        default:
            assert(false);
            break;
        }
    }
}

class GlobalAdditiveFeatureBuilderVisitor {
  public:
    explicit GlobalAdditiveFeatureBuilderVisitor(GlobalAdditiveFeature* result) :
            result_(result) {
    }
    void visit_vertex(VertexId vertex) {}
    void visit_edge(const EdgeStats& edge) {
        edge.edge.add_to_additive_features(result_);
        result_->merge_in(edge.st_feat);
    }
  private:
    GlobalAdditiveFeature* result_;
};

GlobalAdditiveFeature Path::get_additive_features() const {
    assert(!is_special());
    GlobalAdditiveFeature result;
    GlobalAdditiveFeatureBuilderVisitor visitor(&result);
    accept(&visitor);
//    for (int i = static_cast<int>(edge_sequence_.size()) - 1; i >= 0; --i) {
//        edge_sequence_[i].add_to_additive_features(&result);
//    }
    return result;
}

class PrintToStreamAsSentenceVisitor {
  public:
    explicit PrintToStreamAsSentenceVisitor(std::ostream* stream) :
            stream_(stream) {
    }

    void visit_vertex(VertexId vertex) {}

    void visit_edge(const EdgeStats& edge) {
        if (edge.edge.type() == FreeEdge::WORD) {
            (*stream_) << edge.edge.word() << " ";
        }
    }
  private:
    std::ostream* stream_;
};

void Path::print_as_sentence(std::ostream* stream) const {
    if (is_special()) {
        (*stream) << "#SPATH#";
    } else {
        PrintToStreamAsSentenceVisitor visitor(stream);
        accept(&visitor);
//        for (int i = static_cast<int>(edge_sequence_.size()) - 1; i >= 0; --i) {
//            if (edge_sequence_[i].type() == FreeEdge::WORD) {
//                stream << edge_sequence_[i].word() << " ";
//            }
//        }
    }
}

class MakeSentenceAsVSAndHypVisitor {
  public:
    explicit MakeSentenceAsVSAndHypVisitor(std::vector<WordHypPair>* result) :
            result_(result) {
        bond_stack_.push_back(Bond());
    }

    void visit_vertex(VertexId vertex) {
        if (vertex == kFinalVertex) {
            assert(hyp_stack_.size() > 0);
            assert(bond_stack_.size() > 0);
            hyp_stack_.pop_back();
            bond_stack_.pop_back();
        }
    }

    void visit_edge(const EdgeStats& edge) {
        if (edge.edge.type() == FreeEdge::WORD) {
            assert(hyp_stack_.size() > 0);
            result_->push_back(WordHypPair(edge.edge.word(), hyp_stack_.back(), false, bond_stack_.back()));
        } else if (edge.edge.type() == FreeEdge::EPSILON) {
            if (edge.edge.concatenable_information().hyp() != NULL) {
                hyp_stack_.push_back(edge.edge.concatenable_information().hyp());
            } else {
                assert(hyp_stack_.size() > 0);
                result_->push_back(WordHypPair(std::string(), hyp_stack_.back(), true, bond_stack_.back() ));
            }
        } else if (edge.edge.type() == FreeEdge::BOND) {
            bond_stack_.push_back(edge.edge.bond());
        }
    }

    bool empty_hyp_stack() const {
        return hyp_stack_.size() == 0;
    }

  private:
    std::vector<const Hypothesis*> hyp_stack_;
    std::vector<Bond> bond_stack_;
    std::vector<WordHypPair>* result_;
};

std::vector<WordHypPair> Path::get_sentence_as_VS_hyp() const {
    std::vector<WordHypPair> result;
    if (!is_special()) {
        MakeSentenceAsVSAndHypVisitor visitor(&result);
        accept(&visitor);
    }
    return result;
}

class MakeSentenceAsVSVisitor {
  public:
    explicit MakeSentenceAsVSVisitor(std::vector<std::string>* result) :
            result_(result) {
    }
    void visit_vertex(VertexId vertex) {}
    void visit_edge(const EdgeStats& edge) {
        if (edge.edge.type() == FreeEdge::WORD) {
            result_->push_back(edge.edge.word());
        }
    }
  private:
    std::vector<std::string>* result_;
};

std::vector<std::string> Path::get_sentence_as_VS() const {
    std::vector<std::string> result;
    if (!is_special()) {
        MakeSentenceAsVSVisitor visitor(&result);
        accept(&visitor);
//        for (int i = static_cast<int>(edge_sequence_.size()) - 1; i >= 0; --i) {
//            if (edge_sequence_[i].type() == FreeEdge::WORD) {
//                result.push_back(edge_sequence_[i].word());
//            }
//        }
    }
    return result;
}


class GetAdditionalInfosVisitor {
  public:
    explicit GetAdditionalInfosVisitor(InformationAccumulator* result) :
            result_(result) {
    }
    void visit_vertex(VertexId vertex) {}
    void visit_edge(const EdgeStats& edge) {
        result_->add_infos(edge.edge.concatenable_information());
    }
  private:
    InformationAccumulator* result_;
};

InformationAccumulator Path::get_additional_informations() const {
    InformationAccumulator result;
    GetAdditionalInfosVisitor visitor(&result);
    accept(&visitor);
    return result;
}

class PrintToStreamVisitor {
  public:
    explicit PrintToStreamVisitor(std::ostream* stream) :
            stream_(stream) {
    }

    void visit_vertex(VertexId vertex) {
        (*stream_) << "[" << vertex << "]";
    }

    void visit_edge(const EdgeStats& edge) {
        (*stream_) << "--[" << edge.edge << ":" << edge.score << "]-->";
    }

  private:
    std::ostream* stream_;
};

double Path::get_score() const {
    return score_;
}

std::ostream& operator<<(std::ostream& stream, const Path& path) {
    //path.check_is_consistent_full_path();
    if (path.is_special()) {
        stream << "#SPATH#";
    } else {
        PrintToStreamVisitor visitor(&stream);
        path.accept(&visitor);
    }
//        for (int i = static_cast<int>(path.elements_sequence_.size()) - 1; i >= 0; --i) {
//        //assert(path.edge_sequence_.size() + 1 == path.vertex_sequence_.size());
//        //assert(path.edge_scores_.size() ==path.edge_sequence_.size());
//        stream << path.vertex_sequence_.back();
//        for (int i = static_cast<int>(path.edge_sequence_.size()) - 1; i >= 0; --i) {
//            stream << " --[" << path.edge_sequence_[i] << ":" << path.edge_scores_[i] << "]--> " <<
//                    path.vertex_sequence_[i];
//        }
//        stream << " offsets: ";
//        BOOST_FOREACH(double offset, path.offset_sequence_) {
//            stream << offset << " ";
//        }
//        stream << " (sc:" << path.score_ << " == " << path.cumulative_score_ << ")";
//    }
    return stream;
}

void BaseLattice::build_topologically_sorted_vertices() {
    topologically_sorted_vertices_.clear();
    topologically_sorted_vertices_.reserve(nb_vertices());

    std::set<VertexId> set_of_vertices_with_no_incoming_edges;
    VertexToEdgeMap incoming_edges_copy = incoming_edges_;
    //std::set<PinedEdge> removed_edges;

    set_of_vertices_with_no_incoming_edges.insert(kInitialVertex);
    while (set_of_vertices_with_no_incoming_edges.size() > 0) {
        VertexId vertex = *set_of_vertices_with_no_incoming_edges.begin();
        set_of_vertices_with_no_incoming_edges.erase(set_of_vertices_with_no_incoming_edges.begin());
        //std::cerr << "inserting " << vertex << std::endl;
        topologically_sorted_vertices_.push_back(vertex);

        BOOST_FOREACH(EdgeId num_edge, outgoing_edges_[vertex]) {
            const PinedEdge* edge = &get_edge(num_edge);
            std::vector<EdgeId>& edge_list = incoming_edges_copy[edge->v_end()];
            std::vector<EdgeId>::iterator it = find(edge_list.begin(), edge_list.end(), num_edge);
            assert(it != edge_list.end());
            edge_list.erase(it);
            if (edge_list.size() == 0) {
                set_of_vertices_with_no_incoming_edges.insert(edge->v_end());
            }
        }
    }
#ifndef NDEBUG
    BOOST_FOREACH(const VertexToEdgeMap::value_type& v_e_pair, incoming_edges_copy) {
        //std::cerr << v_e_pair.first << v_e_pair.second << std::endl;
        if (v_e_pair.second.size() > 0) {
            std::cerr << *this;
            std::cerr << v_e_pair.second.size() << std::endl;
            std::cerr << v_e_pair.first << std::endl;
        }
        assert(v_e_pair.second.size() == 0); // otherwise, not a DAG
    }
#endif //NDEBUG
}

VertexId vertex_conversion(VertexId old_vertex, VertexId start, VertexId end, int vertex_translation) {
    VertexId new_vertex;
    switch (old_vertex) {
    case kInitialVertex:
        new_vertex = start;
        break;
    case kFinalVertex:
        new_vertex = end;
        break;
    default:
        new_vertex = old_vertex + vertex_translation;
    }
    return new_vertex;
}

void BaseLattice::merge_in_lattice(const BaseLattice& sublattice, VertexId start, VertexId end) {
    assert(this != &sublattice); // better not merge a lattice into itself
    sublattice.check_vertice_consistency();
    check_vertice_consistency();
    assert(finalized_);
    assert(sublattice.finalized_);

    int vertex_translation = vertices_.size() - 2;
    int edge_translation = edges_.size();
    int new_number_of_vertices = vertices_.size() + sublattice.vertices_.size() - 2;

    // adding edges
    edges_.reserve(edges_.size() + sublattice.edges_.size());
    BOOST_FOREACH(const PinedEdge& sub_edge, sublattice.edges_) {
        VertexId new_start = vertex_conversion(sub_edge.v_start(), start, end, vertex_translation);
        VertexId new_end = vertex_conversion(sub_edge.v_end(), start, end, vertex_translation);
        edges_.push_back(PinedEdge(sub_edge.free_edge(), new_start, new_end));
    }

    // adding new vertices
    BOOST_FOREACH(VertexId old_vertex, sublattice.vertices_) {
        if (old_vertex == kInitialVertex || old_vertex == kFinalVertex) {
            continue;
        }
        VertexId new_vertex = old_vertex + vertex_translation;
        assert(vertices_.count(new_vertex) == 0);
        vertices_.insert(new_vertex);
    }
    assert(static_cast<int>(vertices_.size()) == new_number_of_vertices);
    (void) new_number_of_vertices; // for avoiding compiler warning about unused variable when NDEBUG is set

    _update_edges_from_sublattice(sublattice.incoming_edges_, &incoming_edges_, start, end,
            vertex_translation, edge_translation);
    _update_edges_from_sublattice(sublattice.outgoing_edges_, &outgoing_edges_, start, end,
            vertex_translation, edge_translation);

    merge_in_topologically_sorted_vertices(end, vertex_translation,
            sublattice.topologically_sorted_vertices_);

    check_vertice_consistency();
}

void BaseLattice::merge_in_topologically_sorted_vertices(int end, int vertex_translation,
        const std::vector<VertexId>& other_topo) {
    int nb_of_original_vertices = topologically_sorted_vertices_.size();
    int new_number_of_vertices = nb_of_original_vertices + other_topo.size() - 2;

    topologically_sorted_vertices_.resize(new_number_of_vertices); // resize for new number of vertices
//
//        int nb_sub_vertices = sublattice.topologically_sorted_vertices_.size();
//        int nb_sub_vertices_to_add = nb_sub_vertices -2; // minus kFinal and kInitial

    std::vector<VertexId>::iterator it = topologically_sorted_vertices_.begin();

    // set it to position of end
    for (; *it != end; ++it) {
        assert(it != topologically_sorted_vertices_.end());
    }

    // move elements
    int nb_of_elements_that_do_not_have_to_move = (it - topologically_sorted_vertices_.begin());
    int nb_of_elem_to_move = nb_of_original_vertices - nb_of_elements_that_do_not_have_to_move;
    assert(nb_of_elem_to_move >= 1);
    (void) nb_of_elem_to_move; // for avoiding compiler warning about unused variable when NDEBUG is set

    std::vector<VertexId>::iterator it_end_big = topologically_sorted_vertices_.end();
    std::vector<VertexId>::iterator it_end_small = topologically_sorted_vertices_.begin()
            + nb_of_original_vertices;
    do {
        --it_end_big;
        --it_end_small;
        *it_end_big = *it_end_small;
    } while (it_end_small != it);

    //insert new vertices
    std::vector<VertexId>::const_iterator it_sub = other_topo.begin();
    std::vector<VertexId>::const_iterator it_sub_end = other_topo.end();
    for (; it_sub != it_sub_end; ++it_sub) {
        VertexId old_vertex = *it_sub;
        if (*it_sub == kInitialVertex || old_vertex == kFinalVertex) {
            continue;
        }
        *it = old_vertex + vertex_translation;
        ++it;
    }
}


void BaseLattice::_update_edges_from_sublattice(const VertexToEdgeMap& source_edges_map,
        VertexToEdgeMap* target_edges_map, VertexId start, VertexId end, int vertex_translation,
        int edge_translation) {
    BOOST_FOREACH(const VertexToEdgeMap::value_type& vertex_edges, source_edges_map) {
        VertexId old_vertex = vertex_edges.first;
        const std::vector<EdgeId>& old_edges = vertex_edges.second;
        VertexId new_vertex = vertex_conversion(old_vertex, start, end, vertex_translation);
        assert(new_vertex == start || new_vertex == end || target_edges_map->count(new_vertex) == 0);

        std::vector<EdgeId>& new_edges = (*target_edges_map)[new_vertex];
        new_edges.reserve(new_edges.size() + old_edges.size());
        BOOST_FOREACH(EdgeId num_edge, old_edges) {
            EdgeId new_num_edge = num_edge + edge_translation;
            new_edges.push_back(new_num_edge);
        }
        assert(new_vertex == start || new_vertex == end || new_edges.size() == old_edges.size());
    }
}

void BaseLattice::check_vertice_consistency() const {
    assert(vertices_.count(kInitialVertex) == 1);
    assert(vertices_.count(kFinalVertex) == 1);

    std::set<VertexId>::const_iterator it = vertices_.begin();
    for (int i = 0; i < static_cast<int>(vertices_.size()); ++i) {
        assert(*it == i);
        ++it;
    }
}

void BaseLattice::check_global_consistency() const {
    check_vertice_consistency();
    assert(incoming_edges_.count(kInitialVertex) == 0);
    assert(incoming_edges_.size() == vertices_.size() -1);
    assert(outgoing_edges_.count(kFinalVertex) == 0);
    assert(outgoing_edges_.size() == vertices_.size() -1);

    std::set<EdgeId> seen_edges_in_incoming;
    // check incoming
    BOOST_FOREACH(VertexId vertex, vertices_) {
        //std::cerr << vertex << std::endl;
        if (vertex == kInitialVertex) continue;
        assert(incoming_edges_.at(vertex).size() > 0);
        BOOST_FOREACH(EdgeId edge_id, incoming_edges_.at(vertex)) {
            assert(seen_edges_in_incoming.count(edge_id) == 0);
            seen_edges_in_incoming.insert(edge_id);
            const PinedEdge& edge = edges_.at(edge_id);
            (void) edge; // for avoiding compiler warning about unused variable when NDEBUG is set
            assert(!edge.is_disconnected());
            assert(edge.v_end() == vertex);
//            const std::vector<EdgeId> edge_list& = outgoin;
        }
    }

    std::set<EdgeId> seen_edges_in_outgoing;
        // check incoming
        BOOST_FOREACH(VertexId vertex, vertices_) {
            if (vertex == kFinalVertex) continue;
            assert(outgoing_edges_.at(vertex).size() > 0);
            BOOST_FOREACH(EdgeId edge_id, outgoing_edges_.at(vertex)) {
                assert(seen_edges_in_outgoing.count(edge_id) == 0);
                seen_edges_in_outgoing.insert(edge_id);
                const PinedEdge& edge = edges_.at(edge_id);
                (void) edge; // for avoiding compiler warning about unused variable when NDEBUG is set
                assert(!edge.is_disconnected());
                assert(edge.v_start() == vertex);
    //            const std::vector<EdgeId> edge_list& = outgoin;
            }
        }

    assert(seen_edges_in_outgoing == seen_edges_in_incoming);

    for (EdgeId edge_id = 0; edge_id < static_cast<int>(edges_.size()); ++edge_id) {
        const PinedEdge& edge = edges_.at(edge_id);
        (void) edge; // for avoiding compiler warning about unused variable when NDEBUG is set
        assert(edge.is_disconnected() || (seen_edges_in_outgoing.count(edge_id) == 1));
    }
}

struct VertexEdgePair {
    VertexId vertex;
    const FreeEdge& free_edge;

    VertexEdgePair(VertexId vertex, const FreeEdge& free_edge): vertex(vertex), free_edge(free_edge) {}

    bool operator<(const VertexEdgePair& other) const {
        if (vertex != other.vertex) return vertex < other.vertex;
        return free_edge < other.free_edge;
    }
};

typedef std::set<VertexEdgePair> VertexEdgePairSet;
typedef std::map<VertexEdgePairSet, std::vector<VertexId> > EquivalentVertexFinderMap;

void remove_edge_id_from_vector(EdgeId edge_id, std::vector<EdgeId>* edge_list) {
    std::vector<EdgeId>::iterator it = std::find(edge_list->begin(), edge_list->end(), edge_id);
    assert(it != edge_list->end());
    edge_list->erase(it);
}

void BaseLattice::disconnect_edge(EdgeId edge_id) {
    PinedEdge& edge = edges_[edge_id];
    VertexId start = edge.v_start();
    VertexId end = edge.v_end();
    remove_edge_id_from_vector(edge_id, &incoming_edges_.at(end));
    remove_edge_id_from_vector(edge_id, &outgoing_edges_.at(start));
    edge.disconnect();
}

void BaseLattice::change_end_of_edge(EdgeId edge_id, VertexId new_end) {
    PinedEdge& edge = edges_[edge_id];
    //VertexId start = edge.v_start();
    VertexId end = edge.v_end();
    remove_edge_id_from_vector(edge_id, &incoming_edges_.at(end));
    //remove_edge_id_from_vector(edge_id, &outgoing_edges_[start]);
    assert(std::find(incoming_edges_.at(new_end).begin(), incoming_edges_.at(new_end).end(), edge_id) ==
            incoming_edges_.at(new_end).end());
    incoming_edges_.at(new_end).push_back(edge_id);
    edge.reset_end(new_end);
}

void BaseLattice::change_start_of_edge(EdgeId edge_id, VertexId new_start) {
    PinedEdge& edge = edges_[edge_id];
    //VertexId start = edge.v_start();
    VertexId start = edge.v_start();
    remove_edge_id_from_vector(edge_id, &outgoing_edges_.at(start));
    //remove_edge_id_from_vector(edge_id, &outgoing_edges_[start]);
    assert(std::find(outgoing_edges_.at(new_start).begin(), outgoing_edges_.at(new_start).end(), edge_id) ==
            outgoing_edges_.at(new_start).end());
    outgoing_edges_.at(new_start).push_back(edge_id);
    edge.reset_start(new_start);
}

void BaseLattice::remove_isolated_vertex(VertexId vertex) {
        assert(vertex != kInitialVertex);
        assert(vertex != kFinalVertex);
        assert(outgoing_edges_.at(vertex).size() == 0);
        assert(incoming_edges_.at(vertex).size() == 0);
        assert(vertices_.count(vertex) > 0);
        vertices_.erase(vertex);
        outgoing_edges_.erase(vertex);
        incoming_edges_.erase(vertex);
    }

void BaseLattice::renumerate_vertices() {
    assert(vertices_.count(kInitialVertex) == 1);
    assert(vertices_.count(kFinalVertex) == 1);
    assert(kInitialVertex == 0 && kFinalVertex == 1);

    std::map<VertexId, VertexId> vertex_map;
    std::set<VertexId> new_vertices;


    VertexToEdgeMap new_incoming;
    VertexToEdgeMap new_outgoing;

    VertexId count = 0;
    BOOST_FOREACH(VertexId vertex, vertices_) {
        if (count == kInitialVertex) assert(vertex == kInitialVertex);
        if (count == kFinalVertex) assert(vertex == kFinalVertex);
        vertex_map[vertex] = count;

        if (vertex != kInitialVertex) {
            new_incoming[count].swap(incoming_edges_.at(vertex));
        }

        if (vertex != kFinalVertex) {
            new_outgoing[count].swap(outgoing_edges_.at(vertex));
        }
        new_vertices.insert(count);
        count++;
    }
    assert(vertex_map.at(kInitialVertex) == kInitialVertex);
    assert(vertex_map.at(kFinalVertex) == kFinalVertex);

    assert(vertices_.size() == new_vertices.size());
    vertices_.swap(new_vertices);

    assert(incoming_edges_.size() == new_incoming.size());
    assert(outgoing_edges_.size() == new_outgoing.size());
    incoming_edges_.swap(new_incoming);
    outgoing_edges_.swap(new_outgoing);

    BOOST_FOREACH(PinedEdge& edge, edges_) { //NOLINT
        if (edge.is_disconnected()) continue;
        edge.reset_end(vertex_map.at(edge.v_end()));
        edge.reset_start(vertex_map.at(edge.v_start()));
    }

//    std::map<EdgeId, EdgeId> edge_map;
//    std::vector<PinedEdge> new_edges;
//    for (int original_edge_id = 0; original_edge_id < static_cast<EdgeId>(edges_.size()); ++original_edge_id) {
//        if
//    }
//
//    edges_.swap(new_edges);
}

class BaseLatticeAdapter {
  public:
    BaseLatticeAdapter(BaseLattice* lattice, bool inverted) :lattice_(lattice), inverted_(inverted) {}
    void minimize_lattice();

  private:
    VertexId v_end(const PinedEdge& edge) {
        return inverted_?edge.v_start():edge.v_end();
    }

    VertexId v_start(const PinedEdge& edge) const {
        return inverted_?edge.v_end():edge.v_start();
    }

    void change_end_of_edge(EdgeId edge_id, VertexId new_end) {
        if (inverted_) {
            lattice_->change_start_of_edge(edge_id, new_end);
        } else {
            lattice_->change_end_of_edge(edge_id, new_end);
        }
    }

    const VertexToEdgeMap& outgoing_edges() {
        return inverted_?(lattice_->incoming_edges_):(lattice_->outgoing_edges_);
    }

    const VertexToEdgeMap& incoming_edges() {
        return inverted_?(lattice_->outgoing_edges_):(lattice_->incoming_edges_);
    }

    VertexId initial_vertex() const {
        return inverted_?kFinalVertex:kInitialVertex;
    }
    const VertexId final_vertex() const {
        return inverted_?kInitialVertex:kFinalVertex;
    }

    const std::vector<PinedEdge>& edges() {
        return lattice_->edges_;
    }

    void disconnect_edge(EdgeId edge_id) {
        lattice_->disconnect_edge(edge_id);
    }

    void remove_isolated_vertex(VertexId vertex) {
        lattice_->remove_isolated_vertex(vertex);
    }

  private:
    BaseLattice* lattice_;
    bool inverted_;
};

bool edge_list_has_no_duplicate(const std::vector<EdgeId>& lst) {
    std::set<EdgeId> set(lst.begin(), lst.end());
    return lst.size() == set.size();
}

void BaseLatticeAdapter::minimize_lattice() {
    std::set<VertexId> candidate_set;
    std::set<VertexId> processed_set;
    VertexToEdgeMap outgoing_edges_copy = outgoing_edges();
    //std::set<PinedEdge> removed_edges;

    candidate_set.insert(final_vertex());

    while (candidate_set.size() > 0) {
        // first build next candidate set
        std::set<VertexId> new_candidate_set;
        BOOST_FOREACH(VertexId vertex, candidate_set) {
            if (vertex == initial_vertex()) continue; // no incoming edges on initial vertex
            BOOST_FOREACH(EdgeId edge_id, incoming_edges().at(vertex)) {
                const PinedEdge& edge = edges().at(edge_id);
                assert(v_end(edge) == vertex);
                VertexId start_vertex = v_start(edge);
                std::vector<EdgeId>& edge_list = outgoing_edges_copy[start_vertex];
                remove_edge_id_from_vector(edge_id, &edge_list);
                if (edge_list.size() == 0) {
                    new_candidate_set.insert(start_vertex);
                }
            }
        }

        // Remove equivalent edges
        BOOST_FOREACH(VertexId vertex, candidate_set) {
            if (vertex == final_vertex()) continue; // no outgoing edge
            typedef std::map<VertexEdgePair, EdgeId> VertexEdgePairMap;
            VertexEdgePairMap outgoing_edges_duplicate_check_set;
            assert(edge_list_has_no_duplicate(outgoing_edges().at(vertex)));
            std::vector<EdgeId> edge_list_copy = outgoing_edges().at(vertex);
            BOOST_FOREACH(EdgeId edge_id, edge_list_copy) {
                // std::cerr << vertex << " " << edge_id << std::endl;
                const PinedEdge& edge = edges().at(edge_id);
                assert(v_start(edge) == vertex);
                VertexId vertex_end = v_end(edge);
                assert(std::find(incoming_edges().at(vertex_end).begin(), incoming_edges().at(vertex_end).end(),
                        edge_id) != incoming_edges().at(vertex_end).end());

                VertexEdgePair edge_end_pair(vertex_end, edge.free_edge());
                if (outgoing_edges_duplicate_check_set.count(edge_end_pair) > 0) {
//                    std::cerr << "original: " << outgoing_edges_duplicate_check_set.at(edge_end_pair) << " " <<
//                                    edges()[outgoing_edges_duplicate_check_set.at(edge_end_pair)] << std::endl;
//                    std::cerr << "duplicated by " << edge_id << " " << edges()[edge_id] << std::endl;
                    disconnect_edge(edge_id);
                    assert(outgoing_edges().at(vertex).size() > 0);
                    assert(incoming_edges().at(vertex_end).size() > 0);
                } else {
                    //outgoing_edges_duplicate_check_set.insert(edge_end_pair);
                    outgoing_edges_duplicate_check_set[edge_end_pair] = edge_id;
                }
            }
        }

        // find equivalent states
        EquivalentVertexFinderMap equivalent_state_finder;
        BOOST_FOREACH(VertexId vertex, candidate_set) {
            if (vertex == final_vertex()) continue; // final_vertex is not equivalent to any state
            VertexEdgePairSet key;
            BOOST_FOREACH(EdgeId edge_id, outgoing_edges().at(vertex)) {
                const PinedEdge& edge = edges().at(edge_id);
                assert(v_start(edge) == vertex);
                key.insert(VertexEdgePair(v_end(edge), edge.free_edge()));
            }

            assert(key.size() == outgoing_edges().at(vertex).size()); // since duplicate edges have been removed
            equivalent_state_finder[key].push_back(vertex);
        }

        BOOST_FOREACH(const EquivalentVertexFinderMap::value_type& equi_set, equivalent_state_finder) {
            if (equi_set.second.size() == 1) continue; // no equivalent vertex
            assert(equi_set.second.size() != 0); // should not be empty
            VertexId representative_vertex = equi_set.second[0];

            for (int i = 1; i < static_cast<int>(equi_set.second.size()); ++i) {
                VertexId other_vertex = equi_set.second[i];
                assert(other_vertex != representative_vertex);

                std::vector<EdgeId> incoming_edge_list_copy = incoming_edges().at(other_vertex);
                BOOST_FOREACH(EdgeId edge_id, incoming_edge_list_copy) {
                    change_end_of_edge(edge_id, representative_vertex);
                }

                std::vector<EdgeId> outgoing_edge_list_copy = outgoing_edges().at(other_vertex);
                BOOST_FOREACH(EdgeId edge_id, outgoing_edge_list_copy) {
                    disconnect_edge(edge_id);
                }

                assert(outgoing_edges().at(other_vertex).size() == 0);
                assert(incoming_edges().at(other_vertex).size() == 0);
                remove_isolated_vertex(other_vertex);
            }
        }
        candidate_set.swap(new_candidate_set);
    }
}

void BaseLattice::_minimize_lattice_from_start() {
    BaseLatticeAdapter adapter(this, true);
    adapter.minimize_lattice();
    renumerate_vertices();
#ifndef NDEBUG
    check_global_consistency();
#endif
}

void BaseLattice::_minimize_lattice_from_end2() {
    BaseLatticeAdapter adapter(this, false);
    adapter.minimize_lattice();
    renumerate_vertices();
#ifndef NDEBUG
    check_global_consistency();
#endif
}
//
//void BaseLattice::_minimize_lattice_from_end() {
//    std::set<VertexId> candidate_set;
//    std::set<VertexId> processed_set;
//    VertexToEdgeMap outgoing_edges_copy = outgoing_edges_;
//    //std::set<PinedEdge> removed_edges;
//
//    candidate_set.insert(kFinalVertex);
//
//    while (candidate_set.size() > 0) {
//        // first build next candidate set
//        std::set<VertexId> new_candidate_set;
//        BOOST_FOREACH(VertexId vertex, candidate_set) {
//            if (vertex == kInitialVertex) continue; // no incoming edges on initial vertex
//            BOOST_FOREACH(EdgeId edge_id, incoming_edges_.at(vertex)) {
//                const PinedEdge& edge = edges_.at(edge_id);
//                assert(edge.v_end() == vertex);
//                VertexId start_vertex = edge.v_start();
//                std::vector<EdgeId>& edge_list = outgoing_edges_copy[start_vertex];
//                remove_edge_id_from_vector(edge_id, &edge_list);
//                if (edge_list.size() == 0) {
//                    new_candidate_set.insert(start_vertex);
//                }
//            }
//        }
//
//        // Remove equivalent edges
//        BOOST_FOREACH(VertexId vertex, candidate_set) {
//            if (vertex == kFinalVertex) continue; // no outgoing edge
//            VertexEdgePairSet outgoing_edges_duplicate_check_set;
//            BOOST_FOREACH(EdgeId edge_id, outgoing_edges_.at(vertex)) {
//                const PinedEdge& edge = edges_.at(edge_id);
//                assert(edge.v_start() == vertex);
//                VertexId vertex_end = edge.v_end();
//                VertexEdgePair edge_end_pair(vertex_end, edge.free_edge());
//                if (outgoing_edges_duplicate_check_set.count(edge_end_pair) > 0) {
//                    disconnect_edge(edge_id);
//                    assert(outgoing_edges_.at(vertex).size() > 0);
//                    assert(incoming_edges_.at(vertex_end).size() > 0);
//                } else {
//                    outgoing_edges_duplicate_check_set.insert(edge_end_pair);
//                }
//            }
//        }
//
//        // find equivalent states
//        EquivalentVertexFinderMap equivalent_state_finder;
//        BOOST_FOREACH(VertexId vertex, candidate_set) {
//            if (vertex == kFinalVertex) continue; // kFinalVertex is not equivalent to any state
//            VertexEdgePairSet key;
//            BOOST_FOREACH(EdgeId edge_id, outgoing_edges_.at(vertex)) {
//                const PinedEdge& edge = edges_.at(edge_id);
//                assert(edge.v_start() == vertex);
//                key.insert(VertexEdgePair(edge.v_end(), edge.free_edge()));
//            }
//
//            assert(key.size() == outgoing_edges_.at(vertex).size()); // since duplicate edges have been removed
//            equivalent_state_finder[key].push_back(vertex);
//        }
//
//        BOOST_FOREACH(const EquivalentVertexFinderMap::value_type& equi_set, equivalent_state_finder) {
//            if (equi_set.second.size() == 1) continue; // no equivalent vertex
//            assert(equi_set.second.size() != 0); // should not be empty
//            VertexId representative_vertex = equi_set.second[0];
//
//            for (int i = 1; i < static_cast<int>(equi_set.second.size()); ++i) {
//                VertexId other_vertex = equi_set.second[i];
//                assert(other_vertex != representative_vertex);
//
//                BOOST_FOREACH(EdgeId edge_id, incoming_edges_.at(other_vertex)) {
//                    change_end_of_edge(edge_id, representative_vertex);
//                }
//
//                BOOST_FOREACH(EdgeId edge_id, outgoing_edges_.at(other_vertex)) {
//                    disconnect_edge(edge_id);
//                }
//
//                assert(outgoing_edges_.at(other_vertex).size() == 0);
//                assert(incoming_edges_.at(other_vertex).size() == 0);
//                remove_isolated_vertex(other_vertex);
//            }
//
//        }
//
//        candidate_set.swap(new_candidate_set);
//    }
//    renumerate_vertices();
//#ifndef NDEBUG
//    check_global_consistency();
//#endif
//}


std::ostream& operator<<(std::ostream& stream, const BaseLattice& lattice) {
    stream << "Begin Base Lattice:\n";
    BOOST_FOREACH(const PinedEdge& edge, lattice.edges_) {
        stream << edge << "\n";
    }
    if (lattice.finalized_) {
        stream << "Topo order: ";
        BOOST_FOREACH(VertexId vertex, lattice.topologically_sorted_vertices_) {
            stream << vertex << " ";
        }
        stream << "\n";
    }
    stream << "End Base Lattice\n";
    return stream;
}

} // namespace latt

