/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef BASE_LATTICE_HPP_
#define BASE_LATTICE_HPP_

#include <boost/foreach.hpp>
#include <boost/noncopyable.hpp>
#include <boost/scoped_ptr.hpp>
#include <boost/ptr_container/ptr_vector.hpp>

#include <cassert>
#include <vector>
#include <map>
#include <set>
#include <list>
#include <limits>
#include <deque>
#include <queue>
#include <string>

#include "free_edge.hpp"

namespace latt {
// TODO(fabien): optimisations for probably frequent case of a simple linear lattice

typedef int VertexId;

const VertexId kInitialVertex = 0;
const VertexId kFinalVertex = 1;
const VertexId kSpecialVertex = std::numeric_limits<int>::min();

class PinedEdge {
  public:
    friend std::ostream& operator<<(std::ostream& stream, const PinedEdge& edge);
    PinedEdge(const FreeEdge& edge, VertexId start, VertexId end) :
            edge_(edge), id_start_(start), id_end_(end) {
        assert(id_start_ != kSpecialVertex);
        assert(id_end_ != kSpecialVertex);
    }

    const FreeEdge& free_edge() const {
        return edge_;
    }

    VertexId v_start() const {
        return id_start_;
    }

    VertexId v_end() const {
        return id_end_;
    }

    void reset_end(VertexId end) {
        id_end_ = end;
    }

    void reset_start(VertexId start) {
        id_start_ = start;
    }

    void disconnect() {
        id_start_ = kSpecialVertex;
        id_end_ = kSpecialVertex;
    }

    bool is_disconnected() const {
        return (id_start_ == kSpecialVertex && id_end_ == kSpecialVertex);
    }

  private:
    FreeEdge edge_;
    VertexId id_start_;
    VertexId id_end_;
};

std::ostream& operator<<(std::ostream& stream, const std::vector<PinedEdge*>& vect);

//struct VertexLevelPair {
//    VertexLevelPair(VertexId vertex, int level): vertex(vertex), level(level) {}
//    VertexId vertex;
//    int level;
//};
//
//inline std::ostream& operator<<(std::ostream& stream, const VertexLevelPair& pair) {
//    if (pair.vertex == kInitialVertex) stream << "I";
//    else if (pair.vertex == kFinalVertex) stream << "F";
//    else stream << pair.vertex;
//    stream << "(" << pair.level << ")";
//    return stream;
//}

class PathElement {
  public:
    friend class Path;
  private:
    enum Type {
        EDGE, VERTEX
    };
    PathElement(size_t index, Type type) :
            index_(index), type_(type) {
    }
    size_t index_;
    Type type_;
};

struct EdgeStats {
    FreeEdge edge;
    StateFeatures st_feat;
    double score;
    EdgeStats(FreeEdge edge, StateFeatures st_feat, double score) :
            edge(edge), st_feat(st_feat), score(score) {
    }
};

//std::ostream& operator<<(std::ostream& stream, const EdgeStats& obj) {
//    stream << "EdgeStats<" << obj.edge << ", " << obj.st_feat << ">";
//    return stream;
//}

struct WordHypPair {
    std::string word;
    const Hypothesis* hyp;
    bool epsilon;
    Bond in_bond;

    WordHypPair(std::string word, const Hypothesis* hyp, bool epsilon, Bond in_bond) :
            word(word), hyp(hyp), epsilon(epsilon), in_bond(in_bond) {
    }
//    bool operator<(const WordHypPair& other) const {
//        if (word != other.word)
//            return word < other.word;
//        return hyp < other.hyp;
//    }
};

inline std::ostream& operator<<(std::ostream& stream, const WordHypPair& obj) {
    stream << "WordHypPair<" << obj.word << ", " << obj.hyp << ", " << obj.in_bond << ">";
    return stream;
}

class Path {
  public:
    friend std::ostream& operator<<(std::ostream& stream, const Path& path);

    Path() :
            score_(0), cumulative_score_(0) { //, level_(0){
    }

    void add_vertex(VertexId vertex);

    void add_edge(const FreeEdge& edge, double score, StateFeatures st_feat);

//    void add_edge(const FreeEdge& edge, double score) {
//        assert(false);
//    } //deprecated

//    void add_offset(double score);

    void set_score(double score);

    double get_score() const;

    void print_as_sentence(std::ostream* stream) const;

    std::vector<std::string> get_sentence_as_VS() const;

    std::vector<WordHypPair> get_sentence_as_VS_hyp() const;

    InformationAccumulator get_additional_informations() const;

    void check_is_consistent_full_path() const {
        //assert(edge_sequence_.size() + 1 == vertex_sequence_.size());
        //assert(vertex_sequence_.size() >= 2);
        //assert(vertex_sequence_[0].vertex == kFinalVertex);
        //assert(vertex_sequence_.back().vertex == kInitialVertex);
        assert(vertice_.size() >= 2);
        assert(elements_sequence_.size() >= 2);
        assert(elements_sequence_[0].type_ == PathElement::VERTEX);
        assert(elements_sequence_.back().type_ == PathElement::VERTEX);
        assert(vertice_[0] == kFinalVertex);
        assert(vertice_.back() == kInitialVertex);
    }

    static Path make_special() {
        Path res;
        res.score_ = -std::numeric_limits<double>::max();
        return res;
    }

    bool is_special() const {
        return edges_.size() == 0 && vertice_.size() == 0 && score_ == -std::numeric_limits<double>::max();
    }

    GlobalAdditiveFeature get_additive_features() const;

    // visitor pattern
    template<typename Visitor>
    void accept(Visitor* visitor) const;

  private:
//    std::vector<FreeEdge> edge_sequence_;
//    std::vector<double> edge_scores_;
//    std::vector<VertexLevelPair > vertice_;
//    std::vector<StateFeatures> st_feat_sequence_;
    std::vector<EdgeStats> edges_;
    std::vector<VertexId> vertice_;
    std::vector<PathElement> elements_sequence_;

    double score_;
    double cumulative_score_;
//    int level_;
};

typedef int EdgeId;
#define EDGE_ID_SPECIAL (std::numeric_limits<int>::min())

//typedef std::tr1::unordered_map<VertexId, std::vector<PinedEdge*> > VertexToEdgeMap; //TODO(fabien): unordered should be more efficient
typedef std::map<VertexId, std::vector<EdgeId> > VertexToEdgeMap;

class BaseLattice: boost::noncopyable {
  public:
    friend std::ostream& operator<<(std::ostream& stream, const BaseLattice& lattice);
    friend class BaseLatticeAdapter;

    BaseLattice() :
            finalized_(false) {
    }

    void disconnect_edge(EdgeId edge_id);

    void change_end_of_edge(EdgeId edge_id, VertexId new_end);

    void change_start_of_edge(EdgeId edge_id, VertexId new_end);

    void remove_isolated_vertex(VertexId vertex);

    void renumerate_vertices();

    void insert_final_EOS_edge(double score, const kenlm::ngram::Vocabulary& vocab) {
        assert(!finalized_);
        assert(nb_vertices() >= 2);
        VertexId pre_final_state_id = nb_vertices();
        assert(vertices_.count(pre_final_state_id) == 0);
        vertices_.insert(pre_final_state_id);

        BOOST_FOREACH(EdgeId edge_id , incoming_edges_[kFinalVertex]) {
            assert(edges_[edge_id].v_end() == kFinalVertex);
            edges_[edge_id].reset_end(pre_final_state_id);
        }

        incoming_edges_[pre_final_state_id].swap(incoming_edges_[kFinalVertex]);

        FreeEdge edge_EOS(score, vocab);
        add_edge(edge_EOS, pre_final_state_id, kFinalVertex);

        assert(outgoing_edges_[pre_final_state_id].size() == 1);
        assert(incoming_edges_[kFinalVertex].size() == 1);
    }

    const std::vector<VertexId>& get_topologically_sorted_vertices() const {
        assert(static_cast<int>(topologically_sorted_vertices_.size()) == nb_vertices());
        assert(finalized_);
        return topologically_sorted_vertices_;
    }

    int nb_vertices() const {
        return vertices_.size();
    }

    size_t nb_connected_edges() const {
        size_t result = 0;
        BOOST_FOREACH(const PinedEdge& edge, edges_) {
            if (!edge.is_disconnected())
                result++;
        }
        return result;
    }

    VertexId get_start_id() const {
        return 0;
    }

    const std::vector<EdgeId>& outgoing_edges(VertexId vertex) const {
        return outgoing_edges_.at(vertex);
    }

    const PinedEdge& get_edge(EdgeId num_edge) const {
        assert(num_edge != EDGE_ID_SPECIAL);
        assert(num_edge < static_cast<int>(edges_.size()));
        return edges_[num_edge];
    }

    void add_edge(const FreeEdge& edge, VertexId start, VertexId end) {
        assert(start != kFinalVertex);
        assert(end != kInitialVertex);
        edges_.push_back(PinedEdge(edge, start, end));
        EdgeId num_edge = edges_.size() - 1;
        outgoing_edges_[start].push_back(num_edge);
        incoming_edges_[end].push_back(num_edge);
        vertices_.insert(start);
        vertices_.insert(end);
    }

    void finalize() {
        check_vertice_consistency();
        assert(incoming_edges_.count(kInitialVertex) == 0);
        assert(outgoing_edges_.count(kFinalVertex) == 0);
        build_topologically_sorted_vertices();
        finalized_ = true;
    }

    void merge_in_lattice(const BaseLattice& sublattice, VertexId start, VertexId end);

    //void _minimize_lattice_from_end();
    void _minimize_lattice_from_end2();
    void _minimize_lattice_from_start();
//
//    void minimize_lattice() {
//        // merge equivalent edges
//        // push back epsilons?
//        // minimize start
//
//        processed = {initial node}
//        found_but_not_processed
//        candidates =  nodes st all incoming is from processed
//                group by {incoming edge/state} ->merge -> put in processed
//
//
//
//        for node in topo order
//            if node2 is such that incoming_edges_[node1] == incoming_edges_[node2] (comparing start + free_edge):
//                merge node2, node1
//        // split nodes for better combi? n1(e1+e2+e3) // n2(e1+e2+e4) -> n1(e3) + n2(e4) + n3(n1+n2+n3) (pas forcement interessant... less edges/ more states )
//        //minimize end
//        for node in inv topo order
//            if node2 is such that outcoming_edges_[node1] == outcoming_edges_[node2] (comparing start + free_edge):
//            merge node2, node1
//    }

    void check_vertice_consistency() const;
    void check_global_consistency() const;

  private:
    std::vector<PinedEdge> edges_; // TODO(fabien): list or vector?
    VertexToEdgeMap outgoing_edges_;
    VertexToEdgeMap incoming_edges_;
    std::set<VertexId> vertices_;
    std::vector<VertexId> topologically_sorted_vertices_;
    bool finalized_;

    void merge_in_topologically_sorted_vertices(int end, int vertex_translation,
            const std::vector<VertexId>& other_topo);

    void _update_edges_from_sublattice(const VertexToEdgeMap& source_edges_map,
            VertexToEdgeMap* target_edges_map, VertexId start, VertexId end, int vertex_translation,
            int edge_translation);

    void build_topologically_sorted_vertices();
};

} //namespace latt

#endif /* BASE_LATTICE_HPP_ */
