/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "combi_search.hpp"

#include <boost/foreach.hpp>
#include <boost/noncopyable.hpp>
#include <boost/ptr_container/ptr_vector.hpp>
#include <boost/smart_ptr/make_shared.hpp>
#include <boost/smart_ptr/scoped_ptr.hpp>
#include <boost/smart_ptr/shared_ptr.hpp>
#include <glog/logging.h>
//#include <ncurses.h>
#include <stddef.h>
#include <algorithm>
#include <cassert>
#include <iostream>
//#include <queue>
#include <set>
#include <string>
#include <utility>
#include <vector>

#include "lm/model.hh"
#include "../additionals.hpp"
#include "../dtree.hpp"
#include "../hypothesis.hpp"
//#include "../hypothesis_serialization.hpp"
#include "../input_insertion_id.hpp"
//#include "../lm_computer.hpp"
#include "../params.hpp"
#include "../post_processor.hpp"
#include "../score_computer.hpp"
//#include "analysis.hpp"
#include "base_lattice.hpp"
#include "state_lattice.hpp"
#include "hypotheses_to_lattice.hpp"
#include "print_utils.hpp"

/************************************************************************************
 * Code for hypothesis reconstruction.
 * Only used to reconstruct final old-style hypothesis.
 * Not relevant for decoding
 */

/*************************************************************************************
 * InsertionPoints: used by get_next_insertion_point_of_hypothesis to find the correct insertion point
 *  for each hypothesis merging
 */
struct InsertionPoints {
    double fractional_insertion_point;
    WordIndex insertion_point;
    bool replace_ref;

    // only relevant for additionals
    int index_of_used_additional;
    bool is_pre;
    WordIndex cnum;

    bool operator==(const InsertionPoints& other) const {
        if (fractional_insertion_point != other.fractional_insertion_point) return false;
        if (insertion_point != other.insertion_point) return false;
        if (replace_ref != other.replace_ref) return false;
        if (index_of_used_additional != other.index_of_used_additional) return false;
        if (is_pre != other.is_pre) return false;
        if (cnum != other.cnum) return false;
        return true;
    }
    bool operator<(const InsertionPoints& other) const {
        if (fractional_insertion_point < other.fractional_insertion_point) return true;
        if (other.fractional_insertion_point < fractional_insertion_point) return false;
        assert(other == *this || (this->replace_ref == false && other.replace_ref == false));

        if (replace_ref < other.replace_ref) return true;
        if (other.replace_ref < replace_ref) return false;

        if (other.insertion_point != insertion_point) return insertion_point < other.insertion_point;
        if ((!is_pre) < (!other.is_pre)) return true; // is_pre = 1 should come first
        if ((!other.is_pre) < (!is_pre)) return false;
        return cnum < other.cnum;
    }
};


void get_next_insertion_point_of_hypothesis(const Hypothesis* hyp, WordIndex* insertion_point,
        bool* replace_ref, int* index_of_used_additional) {
    const VS& t_string = hyp->t_string();

    std::vector<InsertionPoints> ins_pts;
    for (int i = 0; i < static_cast<int>(t_string.size()); ++i) {
        RefType ref_id = RefType::get_ref_id(t_string[i]);
        if (ref_id.is_valid()) {
            ins_pts.push_back(InsertionPoints());

            ins_pts.back().fractional_insertion_point = i;
            ins_pts.back().insertion_point = i;
            ins_pts.back().replace_ref = true;

            // not relevant for ref
            ins_pts.back().index_of_used_additional = -1;
            ins_pts.back().is_pre = false;
            ins_pts.back().cnum = 0;
        }
    }

    hyp->additionals().check_pre_or_post_list_non_empty();
    for (size_t num_additional = 0; num_additional < hyp->additionals().size(); ++num_additional) {
        assert(num_additional < hyp->additionals().size());
        const Additional& add = hyp->additionals()[num_additional];
        ins_pts.push_back(InsertionPoints());
        if (add.prelist().size() > 0) {
            ins_pts.back().fractional_insertion_point = static_cast<double>(add.prelist()[0])-0.5;
            ins_pts.back().insertion_point = add.prelist()[0];
            ins_pts.back().is_pre = true;

        } else {
            assert(add.postlist().size() > 0);
            ins_pts.back().fractional_insertion_point = static_cast<double>(add.postlist()[0])-0.5;
            ins_pts.back().insertion_point = add.postlist()[0];
            ins_pts.back().is_pre = false;
        }
        ins_pts.back().index_of_used_additional = num_additional;
        ins_pts.back().replace_ref = false;
        ins_pts.back().cnum = add.cnum();
    }
    assert(ins_pts.size() > 0);
    std::sort(ins_pts.begin(), ins_pts.end());

    *insertion_point = ins_pts[0].insertion_point;
    *replace_ref = ins_pts[0].replace_ref;
    *index_of_used_additional = ins_pts[0].index_of_used_additional;
}

/*****************************************************************************
 * reconstruct_hypotheses takes as input a vector of initial hypotheses.
 * It will use the Hypothesis merge constructor for iteratively merging these initial hypotheses
 * As output, it will return a shared_ptr of ptr_vector of Hypothesis, that contains all intermediate hypotheses.
 * The final reconstructed hypothesis is the last one in the vector.
 */

boost::shared_ptr<boost::ptr_vector<Hypothesis> > reconstruct_hypotheses(const std::vector<const Hypothesis*>& initial_hyps) {
    assert(initial_hyps.size() > 1); // do not use if there is only one initial hypothesis (then we do not need to combine anyway)

    // A shared_ptr object to take ownership of all the intermediate hypotheses created
    boost::shared_ptr<boost::ptr_vector<Hypothesis> > new_hyps = boost::make_shared<boost::ptr_vector<Hypothesis> >();

    // The parameters needed by the Hypothesis merging constructor
    WordIndex insertion_point;
    bool replace_ref;
    int index_of_used_additional;

    // A small check that the input data is as expected
    BOOST_FOREACH(const Hypothesis* hyp, initial_hyps) {
        hyp->additionals().check_pre_or_post_list_non_empty();
    }

    // The first two hupotheses are merged
    get_next_insertion_point_of_hypothesis(initial_hyps[0], &insertion_point, &replace_ref, &index_of_used_additional);
    new_hyps->push_back(
            new Hypothesis(*initial_hyps[0], *initial_hyps[1], insertion_point, replace_ref, index_of_used_additional));
//    std::cerr << initial_hyps[0] << " + " << initial_hyps[1] << " ip:" << insertion_point << " r:" << replace_ref << " iu:" << index_of_used_additional << std::endl;

    // if there is more than 2 initial hypotheses, we iterate:
    for (int i = 2; i < static_cast<int>(initial_hyps.size()); ++i) {
        get_next_insertion_point_of_hypothesis(&new_hyps->back(), &insertion_point, &replace_ref, &index_of_used_additional);
//        std::cerr << new_hyps->back().t_string() << new_hyps->back().additionals() << " + " << initial_hyps[i] << " ip:" << insertion_point << " r:" << replace_ref << " iu:" << index_of_used_additional << std::endl;
        new_hyps->push_back(
                new Hypothesis(new_hyps->back(), *initial_hyps[i], insertion_point, replace_ref, index_of_used_additional));
            }

    return new_hyps;
}

/* End of Code for hypothesis reconstruction.
 **************************************************************************************/

/// Utility function for flattening t_strings into strings
std::string flatten_t_string(const VS &t_string) {
    std::string result = "";
    for (size_t i = 0; i < t_string.size(); i++) {
        result += t_string[i] + " ";
    }
    return result;
}

/***************************************************************************************
 * NBestTranslationResultBuilder:  a class for generating a NBestTranslationResult object, the nbest-list
 *   object expected to be outputed by the decoder.
 * It is in charge of post-processing and formatting the final results, and also to remove duplicates
 * (although the decoder should have efficiently filtered most of the nbest duplicates before that).
 */


class NBestTranslationResultBuilder : boost::noncopyable {
  public:
    NBestTranslationResultBuilder(size_t n, const PostProcessor& post_processor)
        : n_(n), post_processor_(post_processor) {}

    bool enough() const {
        return result_.size() >= n_;
    }

    NBestTranslationResult get() {
        return result_;
    }

    size_t current_size() const {
        return result_.size();
    }

    double best_score() const {
        assert(result_.size() > 0);
        return result_[0].score;
    }

    double last_score() const {
        assert(result_.size() > 0);
        return result_.back().score;
    }

    void push_translation(const VS& translation, double score, const latt::GlobalAdditiveFeature& add_feat, 
            const std::vector<const Hypothesis*>& initial_hyps, const std::vector<latt::WordHypPair> &details_vector) {
        if (result_.size() >= n_) return;

        std::pair<VS, latt::GlobalAdditiveFeature> vs_feat_pair(translation, add_feat);
        if (duplicates_VS_feat_check_set_.count(vs_feat_pair) > 0) return;
        duplicates_VS_feat_check_set_.insert(vs_feat_pair);

//        if(duplicates_VS_check_set_.count(translation) > 0) return;
//        duplicates_VS_check_set_.insert(translation);

        std::string final_str = flatten_t_string(translation);
        final_str = post_processor_.post_process(final_str);

        result_.push_back(SingleNBestTranslationResult());
        SingleNBestTranslationResult&  tl_result = result_.back();
        tl_result.t_string_flat = final_str;
        tl_result.score = score;
        add_feat.set_feature_map(&tl_result.features);

        assert(initial_hyps.size() > 0);
        tl_result.initial_hypothesis_list = initial_hyps;

        if (Params::Translator::reconstruct_final_hypotheses) {
            if (initial_hyps.size() == 1) {
                tl_result.hypothesis = initial_hyps[0];
            } else {
                tl_result.intermediate_hypotheses_holder = reconstruct_hypotheses(initial_hyps);
                tl_result.hypothesis = &(tl_result.intermediate_hypotheses_holder->back());
            }
            tl_result.dpnd = tl_result.hypothesis->get_t_dpnd();
        } else {
            tl_result.hypothesis = NULL;
        }

        tl_result.score_no_lm = tl_result.score;
        tl_result.t_string = translation;
        tl_result.details_vector = details_vector;
    }

  private:
    NBestTranslationResult result_;
    std::set<std::string> duplicates_strings_check_set_;
    std::set<VS> duplicates_VS_check_set_;
    std::set<std::pair<VS, latt::GlobalAdditiveFeature> > duplicates_VS_feat_check_set_;
    size_t n_;
    const PostProcessor& post_processor_;
};


/************************************************************************************************
 * Implementation of lattice_search
 * (explanations of parameters in the .hpp)
 */

NBestTranslationResult lattice_search(const std::vector<AugmentedPattern>& initial_patterns,
        InitialHypothesesStorage *trans_cand, const lm::ngram::RestProbingModel& lm_rest,
        const WeightVector& weights, t_log::TranslationLog* translation_log) {
    assert(!trans_cand->has_empty_stack());

    latt::AdditiveFeaturesWeight add_weights(weights);
    latt::StateFeaturesWeight st_weights(weights);

    int time0; // just for logging times

    // Creation of the Computation Manager object, that will be in charge of the decoding
    time0 = t_log::get_time_in_ms();
    LatticeComputationManager comp_manager(initial_patterns,
                trans_cand, st_weights, add_weights,
                lm_rest, translation_log);
    translation_log->set_lattice_setup_time(time0);

    // Creation of the final BiLmLatticeComputer object
    time0 = t_log::get_time_in_ms();
    latt::StateLatticeComputer& final_comp = comp_manager.get_final_lattice_computer();
    translation_log->set_lattice_building_time(time0);

    // Calling get_bilm_lattice() on the BiLmLatticeComputer object actually trigger the decoding
    //  (decoding was delayed until that point by using "lazy" calls)
    time0 = t_log::get_time_in_ms();
    latt::StateLattice* final_latt = final_comp.get_bilm_lattice();
    translation_log->set_lattice_decoding_time(time0);

    //std::cout << final_latt->get_kbest_size() << " paths" << std::endl;

    // Building of the NBestTranslationResultBuilder, that will create the NBestTranslationResult object
    PostProcessor post_processor(Params::Translator::target_language);
    NBestTranslationResultBuilder result_builder(Params::Translator::n_best_length, post_processor);

    int time_before_building_nbest = t_log::get_time_in_ms();
    for (size_t k = 0; k < final_latt->get_kbest_size(); ++k) {
        // Extracting the path in the lattice corresponding to the k-th translation
        latt::Path k_path = final_latt->get_kbest_path(k);
        if (k_path.is_special()) break; // no more translations to extract

        // Getting all needed features from k_path (features, list of initial hypotheses,...)
        latt::GlobalAdditiveFeature add_features = k_path.get_additive_features();
        latt::InformationAccumulator infos = k_path.get_additional_informations();
        const std::vector<const Hypothesis*>& initial_hyps = infos.initial_hyps();
        assert(initial_hyps.size() > 0);

//        std::vector<latt::WordHypPair> word_hyp_vect = k_path.get_sentence_as_VS_hyp();
//        size_t count_no_epsilon = 0;
//        BOOST_FOREACH(const latt::WordHypPair& word_hyp, word_hyp_vect) {
//            if (!word_hyp.epsilon) {
//                count_no_epsilon += 1;
//            }
//        }
//        assert(count_no_epsilon == k_path.get_sentence_as_VS().size());

        // Getting more details about the candidate translation:
        // details_vector contains a WordHypPair object for each words (INCLUDING the removed optional words)
        // in the candidate translation
        // details_vector[i].word is the (possibly removed) i-th word as a string
        // details_vector[i].epsilon is a boolean that is true if the i-th word is a removed optional
        // details_vector[i].hyp is a pointer to the initial hypothesis from which thei-th word is coming
        // details_vector[i].in_bond contains information about the bond in which the hypothesis of the i-th word has been inserted
        std::vector<latt::WordHypPair> details_vector = k_path.get_sentence_as_VS_hyp();

        // Adding the translation to result_builder
        result_builder.push_translation(k_path.get_sentence_as_VS(), k_path.get_score(), add_features, initial_hyps, details_vector);
        if (result_builder.enough()) break; // if our nbest list is full, we stop here
    }

    // logging informations
    translation_log->set_nbest_gen_time(time_before_building_nbest);
    translation_log->set_best_model_score(result_builder.best_score());
    translation_log->set_last_model_score(result_builder.last_score());
    translation_log->set_nb_nbest_produced(result_builder.current_size());

    return result_builder.get();
}
