/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef FEATURES_STRUCTS_HPP_
#define FEATURES_STRUCTS_HPP_

#include <boost/foreach.hpp>
#include <iostream>
#include <map>
#include <string>
#include <utility>
#include <cassert>

namespace latt {
// a value of -infinity for kImpossibleScore would probably be better, but it seem to interact badly with
// some assertions in the code. TODO(fabien): check this.
static const double kImpossibleScore = -10000000000000.0; //- std::numeric_limits<double>::infinity();

inline void feature_needed_or_die(const std::map<std::string, double>& weights, const char* feat_name) {
    if (weights.count(feat_name) == 0) {
        std::cerr << "Error: weight for feature " << feat_name << " HAS to be specified\n";
        exit(1);
    }
}

static const char* kOovFeatureName = "real_oov";
static const char* kNbTargetWordsFeatureName = "real_target_size";

class AdditiveFeaturesWeight {
  public:
    friend class AdditiveFeatures;
    friend std::ostream& operator<<(std::ostream& stream, const AdditiveFeaturesWeight& add);

    AdditiveFeaturesWeight(int nb_words, int nb_oov) :
            nb_words_(nb_words), nb_oov_(nb_oov) {
    }

    explicit AdditiveFeaturesWeight(const std::map<std::string, double>& weights) {
        feature_needed_or_die(weights, kNbTargetWordsFeatureName);
        feature_needed_or_die(weights, kOovFeatureName);
        feature_needed_or_die(weights, "nb_optionals_skipped");

        nb_words_ = weights.at(kNbTargetWordsFeatureName);
        nb_oov_ = weights.at(kOovFeatureName);
        BOOST_FOREACH(const SparseMap::value_type& name_val, weights) {
            if (name_val.first == kNbTargetWordsFeatureName || name_val.first == kOovFeatureName || name_val.first == "lm") continue;
            assert(sparse_features_.count(name_val.first) == 0);
            sparse_features_.insert(name_val);
        }
    }

    void set_sparse_feature_weight(std::string name, double val) {
        sparse_features_[name] = val;
    }

  private:
    double nb_words_;
    double nb_oov_;
    typedef std::map<std::string, double> SparseMap;
    SparseMap sparse_features_;
};

class AdditiveFeatures {
  public:
    friend std::ostream& operator<<(std::ostream& stream, const AdditiveFeatures& add);

    AdditiveFeatures() :
            nb_words_(0), nb_oov_(0), has_impossible_edge_(false) {
    }

    AdditiveFeatures(int nb_words, int nb_oov) :
            nb_words_(nb_words), nb_oov_(nb_oov), has_impossible_edge_(false) {
    }

    explicit AdditiveFeatures(const std::map<std::string, double>& feat) :
            nb_words_(0), nb_oov_(0), has_impossible_edge_(false) {
        set_sparse_features_from_map(feat);
    }

    static AdditiveFeatures make_impossible() {
        AdditiveFeatures result;
        result.has_impossible_edge_ = true;
        return result;
    }

    void set_sparse_feature(std::string name, double val) {
        assert(name != kOovFeatureName && name !=kNbTargetWordsFeatureName && name != "lm");
        sparse_features_[name] = val;
    }

    void set_sparse_features_from_map(const std::map<std::string, double>& weights) {
        assert(weights.count(kNbTargetWordsFeatureName) == 0);
        assert(weights.count(kOovFeatureName) == 0);
        assert(weights.count("lm") == 0);
        assert(weights.count("nb_optionals_skipped") == 0);
        BOOST_FOREACH(const SparseMap::value_type& name_val, weights) {
            sparse_features_.insert(name_val);
        }
    }

//    double get_sparse_feature(std::string& name) const {
//        return sparse_features_.at(name);
//    }

    void set_feature_map(std::map<std::string, double>* feat_map) const {
        assert(feat_map->count(kOovFeatureName) == 0);
        assert(feat_map->count(kNbTargetWordsFeatureName) == 0);

        (*feat_map)[kOovFeatureName] = nb_oov_;
        (*feat_map)[kNbTargetWordsFeatureName] = nb_words_;

        BOOST_FOREACH(const SparseMap::value_type& name_val, sparse_features_) {
            assert(feat_map->count(name_val.first) == 0);
            feat_map->insert(name_val);
        }
    }

    void merge_in(const AdditiveFeatures& other) {
        nb_words_ += other.nb_words_;
        nb_oov_ += other.nb_oov_;
        has_impossible_edge_ |= other.has_impossible_edge_;
        BOOST_FOREACH(const SparseMap::value_type& name_val, other.sparse_features_) {
            sparse_features_[name_val.first] += name_val.second;
        }
    }

    double compute_score(const AdditiveFeaturesWeight& weight) const {
        if (has_impossible_edge_)
            return kImpossibleScore;
        double result = 0;
        result += nb_oov_ * weight.nb_oov_;
        result += nb_words_ * weight.nb_words_;
        BOOST_FOREACH(const SparseMap::value_type& name_val, sparse_features_) {
            AdditiveFeaturesWeight::SparseMap::const_iterator it = weight.sparse_features_.find(
                    name_val.first);
            if (it != weight.sparse_features_.end()) {
                result += name_val.second * it->second;
            } else {
                std::cerr << "unknown feature: " << name_val.first << std::endl;
                assert(false);
            }
        }
        return result;
    }

    bool operator<(const AdditiveFeatures& other) const {
        if (nb_words_ != other.nb_words_)
            return nb_words_ < other.nb_words_;
        if (nb_oov_ != other.nb_oov_)
            return nb_oov_ < other.nb_oov_;
        if (has_impossible_edge_ != other.has_impossible_edge_)
            return has_impossible_edge_ < other.has_impossible_edge_;
        return sparse_features_ < other.sparse_features_;
    }

  private:
    //REMINDER: when adding new members, update the operator< method ...
    int nb_words_;
    int nb_oov_;
    bool has_impossible_edge_;
    typedef std::map<std::string, double> SparseMap;
    SparseMap sparse_features_;
};

static const char* kAdditionalPRMismatchFeatureName = "additional_pr_mismatch";
static const char* kReferencePRMismatchFeatureName = "reference_pr_mismatch";

class StateFeaturesWeight {
  public:
    friend class StateFeatures;
    explicit StateFeaturesWeight(double lm_weight) :
            lm_weight_(lm_weight), additional_PR_mismatch_weight_(0), reference_PR_mismatch_weight_(0) {
    }

    explicit StateFeaturesWeight(const std::map<std::string, double>& weights): additional_PR_mismatch_weight_(0), reference_PR_mismatch_weight_(0) {
        feature_needed_or_die(weights, "lm");
        lm_weight_ = weights.at("lm");
#ifdef USE_PR
        feature_needed_or_die(weights, kAdditionalPRMismatchFeatureName);
        feature_needed_or_die(weights, kReferencePRMismatchFeatureName);
        additional_PR_mismatch_weight_ = weights.at(kAdditionalPRMismatchFeatureName);
        reference_PR_mismatch_weight_ = weights.at(kReferencePRMismatchFeatureName);
#endif // USE_PR
    }

  private:
    double lm_weight_;
    double additional_PR_mismatch_weight_;
    double reference_PR_mismatch_weight_;
};

class StateFeatures {
  public:
    friend class GlobalAdditiveFeature;

    explicit StateFeatures(double lm) :
            lm_(lm), additional_PR_mismatch_(0), reference_PR_mismatch_(0) {
    }

    void set_feature_map(std::map<std::string, double>* feat_map) const {
        assert(feat_map->count("lm") == 0);
        assert(feat_map->count(kAdditionalPRMismatchFeatureName) == 0);
        assert(feat_map->count(kReferencePRMismatchFeatureName) == 0);
        (*feat_map)["lm"] = lm_;
        (*feat_map)[kAdditionalPRMismatchFeatureName] = additional_PR_mismatch_;
        (*feat_map)[kReferencePRMismatchFeatureName] = reference_PR_mismatch_;
    }

    double compute_score(const StateFeaturesWeight& weight) const {
        double result = 0;
        result += lm_ * weight.lm_weight_;
        result += additional_PR_mismatch_ * weight.additional_PR_mismatch_weight_;
        result += reference_PR_mismatch_ * weight.reference_PR_mismatch_weight_;
        return result;
    }

    void merge_in(const StateFeatures& other) {
        lm_ += other.lm_;
        additional_PR_mismatch_ += other.additional_PR_mismatch_;
        reference_PR_mismatch_ += other.reference_PR_mismatch_;
    }

    bool operator<(const StateFeatures& other) const {
        if (additional_PR_mismatch_ < other.additional_PR_mismatch_) return true;
        if (other.additional_PR_mismatch_ < additional_PR_mismatch_) return false;
        if (reference_PR_mismatch_ < other.reference_PR_mismatch_) return true;
        if (other.reference_PR_mismatch_ < reference_PR_mismatch_) return false;
        return lm_ < other.lm_;
    }

    void set_additional_PR_mismatch(int i) {
        additional_PR_mismatch_ = i;
    }

    void set_reference_PR_mismatch(int i) {
        reference_PR_mismatch_ = i;
    }

  private:
    StateFeatures()
      : lm_(0), additional_PR_mismatch_(0), reference_PR_mismatch_(0) {
    }

    double lm_;
    int additional_PR_mismatch_;
    int reference_PR_mismatch_;
};

class GlobalAdditiveFeature {
  public:
    void merge_in(const AdditiveFeatures& add) {
        add_feat_.merge_in(add);
    }

    void merge_in(const StateFeatures& st) {
        st_feat_.merge_in(st);
    }

    void set_feature_map(std::map<std::string, double>* feat_map) const {
        assert(feat_map->size() == 0);
        add_feat_.set_feature_map(feat_map);
        st_feat_.set_feature_map(feat_map);
    }

    bool operator<(const GlobalAdditiveFeature& other) const {
        if (st_feat_ < other.st_feat_ || other.st_feat_ < st_feat_)
            return st_feat_ < other.st_feat_;
        return add_feat_ < other.add_feat_;
    }

  private:
    AdditiveFeatures add_feat_;
    StateFeatures st_feat_;
};

} // namespace latt

#endif /* FEATURES_STRUCTS_HPP_ */
