/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef FREE_EDGE_HPP_
#define FREE_EDGE_HPP_

#include <iostream>
#include <map>
#include <string>
#include <vector>

#include "lm/model.hh"
#include "../hypothesis.hpp"
#include "features_structs.hpp"

#include "../input_insertion_id.hpp"

namespace latt {
// TODO(fabien): optimisations for probably frequent case of a simple linear lattice

namespace kenlm = ::lm;

// typedef kenlm::ngram::Model LanguageModel;

typedef kenlm::ngram::RestProbingModel LanguageModel;
class StateLatticeComputer;

enum ParentRelationship {BEFORE, AFTER, UNKNOWN, UNSET};

void print_parent_relationship_to_stream(std::ostream* stream, ParentRelationship pr);

class Bond {
  public:
    enum Type {REFERENCE, ADDITIONAL, INVALID};
    friend std::ostream& operator<<(std::ostream& stream, const Bond& bond);

    Bond(): bilm_lattice_(NULL), type_(INVALID), parent_relationship_(UNSET) {}

    Bond(StateLatticeComputer* bilm_lattice, Type type, ParentRelationship parent_relationship)
        : bilm_lattice_(bilm_lattice), type_(type), parent_relationship_(parent_relationship) {
        assert(type != INVALID);
        assert(parent_relationship_ != UNKNOWN);
        assert(parent_relationship_ != UNSET);
    }

    bool operator<(const Bond& other) const {
        if (type_ == INVALID && other.type_ == INVALID) return false;
#ifdef USE_PR
        if (type_ < other.type_) return true;
        if (other.type_ < type_) return false;
        if (parent_relationship_ < other.parent_relationship_) return true;
        if (other.parent_relationship_ < parent_relationship_) return false;
#endif //USE_PR
        if (bilm_lattice_ < other.bilm_lattice_) return true;
        if (other.bilm_lattice_ < bilm_lattice_) return false;
        return false; //identical
    }

    bool is_valid() const {
        return type_ != INVALID;
    }

    StateLatticeComputer* bilm_lattice() const {
        assert(is_valid());
        return bilm_lattice_;
    }

    Type type() const {
        return type_;
    }

    ParentRelationship parent_relationship() const {
        assert(is_valid());
        assert(parent_relationship_ != UNSET);
        return parent_relationship_;
    }

    InputInsertionId insertion_id() const {
        return insertion_id_;
    }

    void set_insertion_id(InputInsertionId id) {
        insertion_id_ = id;
    }

  private:
    StateLatticeComputer* bilm_lattice_;
    Type type_;
    ParentRelationship parent_relationship_;
    InputInsertionId insertion_id_;
};

class ConcatenableInformation {
  public:
    friend class InformationAccumulator;

    ConcatenableInformation(): hyp_(NULL) {}
    explicit ConcatenableInformation(const Hypothesis* hyp)
        : hyp_(hyp) {assert(hyp != NULL);}

    bool operator<(const ConcatenableInformation& other) const {
        return hyp_ < other.hyp_;
        //return false; // it should be ok to accept that edges with different hyp are mergeable (keeping only the first hyp)
        //return hyp_ < other.hyp_;
    }

    void set_hyp(const Hypothesis *hyp) {
        assert(hyp != NULL);
        hyp_ = hyp;
    }

    const Hypothesis* hyp() const {
        return hyp_;
    }

  private:
    const Hypothesis *hyp_;
};

class InformationAccumulator {
  public:
    void add_infos(const ConcatenableInformation& info) {
        if (info.hyp_ != NULL) initial_hyps_.push_back(info.hyp_);
    }

    const std::vector<const Hypothesis*>& initial_hyps() const {
        return initial_hyps_;
    }

  private:
    std::vector<const Hypothesis*> initial_hyps_;
};

class FreeEdge {
  public:
    friend std::ostream& operator<<(std::ostream& stream, const FreeEdge& edge);
    enum Type {
        WORD, EPSILON, BOND, EOS
    }; // REF, SUBLATTICE};
    typedef std::string POSType; // type of Part Of Speech

    FreeEdge(std::string word, double score, const LanguageModel::Vocabulary& vocab)
            : word_(word), score_(score), idx_(vocab.Index(word)), type_(WORD), prefered_parent_relationship_(UNSET),
                    additive_features_(
                    1, (idx_ == vocab.NotFound()) ? 1 : 0) {
    }

    FreeEdge(double score, const LanguageModel::Vocabulary& vocab) :
            score_(score), idx_(vocab.EndSentence()), type_(EOS), prefered_parent_relationship_(UNSET), additive_features_() {
    }

    static FreeEdge make_impossible() {
        FreeEdge result(kImpossibleScore);
        result.additive_features_ = AdditiveFeatures::make_impossible();
        return result;
    }

    explicit FreeEdge(double score) :
            score_(score), idx_(0), type_(EPSILON), prefered_parent_relationship_(UNSET), additive_features_() {
    } //Epsilon constructor

    explicit FreeEdge(Bond bond, double score) :
            score_(score), idx_(0), type_(BOND), prefered_parent_relationship_(UNSET), bond_(bond), additive_features_() {
    }

    void add_to_additive_features(GlobalAdditiveFeature* out) const {
        out->merge_in(additive_features_);
    }
//    FreeEdge(const EdgeListFacade& edge_list, double score) :
//            score_(score), idx_(0), edge_list_(&edge_list), type_(LIST) {
//    } //BaseLattice constructor

    double score() const {
        return score_;
    }

    void set_score_from_weights(const AdditiveFeaturesWeight& weight) {
        score_ = additive_features_.compute_score(weight);
    }

    void set_sparse_feature(std::string name, double val) {
        additive_features_.set_sparse_feature(name, val);
    }

    void set_sparse_feature_from_map(const std::map<std::string, double>& weights) {
        additive_features_.set_sparse_features_from_map(weights);
    }

//    double get_k_score(int k) const {
//        assert(edge_list_ != NULL);
//        assert(type_ == LIST);
//        return edge_list_->get_k_score(k);
//    }

    kenlm::WordIndex word_idx() const {
        assert(type_ == WORD || type_ == EOS);
        return idx_;
    }

    std::string word() const {
        assert(type_ == WORD);
        return word_;
    }

    Type type() const {
        return type_;
    }

    const Bond& bond() const {
        assert(bond_.is_valid());
        assert(type_ == BOND);
        return bond_;
    }

    bool operator<(const FreeEdge& other) const {
        if (other.type_ != type_)
            return type_ < other.type_;
        if (other.idx_ != idx_)
            return idx_ < other.idx_;
        if (other.score_ != score_)
            return score_ < other.score_;
        if (other.word_ != word_)
            return word_ < other.word_;


        if (bond_ < other.bond_) return true;
        if (other.bond_ < bond_) return false;

#ifdef USE_PR
        if (prefered_parent_relationship_ < other.prefered_parent_relationship_) return true;
        if (other.prefered_parent_relationship_ < prefered_parent_relationship_) return false;
#endif // USE_PR

        if (concatenable_information_ < other.concatenable_information_) return true;
        if (other.concatenable_information_ < concatenable_information_) return false;
        return additive_features_ < other.additive_features_;
    }

    void set_prefered_parent_relationship(ParentRelationship pr) {
        prefered_parent_relationship_ = pr;
    }

    ParentRelationship prefered_parent_relationship() const {
        return prefered_parent_relationship_;
    }

    ConcatenableInformation* get_modifiable_concatenable_information() {
        return &concatenable_information_;
    }

    const ConcatenableInformation& concatenable_information() const {
        return concatenable_information_;
    }

//    const EdgeListFacade* lattice() const {
//        assert(edge_list_ != NULL);
//        return edge_list_;
//    }
//    const POSType& pos() const {
//        assert(type_ == WORD);
//        return pos_;
//    }

  private:
    //REMINDER: when adding new members, update the operator< method ...
    std::string word_;
    double score_;
    kenlm::WordIndex idx_;
//    const EdgeListFacade* edge_list_;
    Type type_;

//    POSType pos_; // Part Of Speech
    ParentRelationship prefered_parent_relationship_;

    Bond bond_;

    AdditiveFeatures additive_features_;

    ConcatenableInformation concatenable_information_;

//    kenlm::WordIndex upper_cased_idx_;
//    bool is_oov_;
//    int ref_num_;
};

} // namespace latt

#endif /* FREE_EDGE_HPP_ */
