/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "hypotheses_to_lattice.hpp"
#include <string>
#include <algorithm>
#include <utility>
#include <vector>
#include <set>
#include <map>

/*************************************
 * Some utility functions
 */

bool is_word_optional(const std::string& w) {
    if (w.size() >= 2 && w[w.size() - 1] == '*')
        return true;
    return false;
}

std::string clean_optional_word(const std::string& word) {
    assert(is_word_optional(word));
    return std::string(word.begin(), word.end() - 1);
}

std::vector<BondInfos> make_bond_infos(const VS& t_string, const std::vector<Dpnd>& depend) {
    std::vector<BondInfos> result;
    assert(t_string.size() == depend.size());
    for (int i = 0; i < static_cast<int>(t_string.size()); ++i) {
        assert(i != depend[i]);
        result.push_back(BondInfos());
        RefType ref_id = RefType::get_ref_id(t_string[i]);
        if (ref_id.is_valid()) {
            assert(i >= 0 && i < static_cast<int>(t_string.size()));
            result.back().type = latt::Bond::REFERENCE;
            result.back().pr = (i < depend[i])? latt::BEFORE : latt::AFTER;
        } else {
            result.back().type = latt::Bond::INVALID;
            result.back().pr = latt::UNSET;
        }
    }
    return result;
}

struct AddInsertionStruct {
    ::WordIndex cnum;
    ::WordIndex insertion_pos;
    bool is_pre;

    bool operator<(const AddInsertionStruct& other) const {
        if (other.insertion_pos != insertion_pos) return insertion_pos < other.insertion_pos;
        if (!is_pre < !other.is_pre) return true; // is_pre = 1 should come first
        if (!other.is_pre < !is_pre) return false;
        return cnum < other.cnum;
    }
};

std::pair<VS, std::vector<BondInfos> > make_t_string_from_hyp_with_additionals(const Hypothesis* hyp) {
    std::vector<AddInsertionStruct> vect_of_add_insertion;
    BOOST_FOREACH(const Additional& add, hyp->additionals()) {
        vect_of_add_insertion.push_back(AddInsertionStruct());
        vect_of_add_insertion.back().cnum = add.cnum();
        if (add.prelist().size() > 0) {
            vect_of_add_insertion.back().insertion_pos = add.prelist()[0];
            vect_of_add_insertion.back().is_pre = true;
        } else {
            assert(add.postlist().size() > 0);
            vect_of_add_insertion.back().insertion_pos = add.postlist()[0];
            vect_of_add_insertion.back().is_pre = false;
        }
    }

    std::pair<VS, std::vector<BondInfos> > result;
    std::vector<BondInfos>& bond_infos = result.second;
    VS& new_t_string = result.first;

    new_t_string = (hyp->t_string());
    bond_infos = make_bond_infos(hyp->t_string(), hyp->get_t_dpnd());

    assert(new_t_string.size() == bond_infos.size());

    std::sort(vect_of_add_insertion.begin(), vect_of_add_insertion.end());
    for (int i = static_cast<int>(vect_of_add_insertion.size()) - 1; i>= 0; --i) {
        std::stringstream ss;
        ss << "[X" << vect_of_add_insertion.at(i).cnum << "]";
        std::string ref_string = ss.str();
        new_t_string.insert(new_t_string.begin()+vect_of_add_insertion[i].insertion_pos, ref_string);
        BondInfos b_info;
        b_info.type = latt::Bond::ADDITIONAL;
        b_info.pr = vect_of_add_insertion[i].is_pre? latt::BEFORE : latt::AFTER;
        bond_infos.insert(bond_infos.begin()+vect_of_add_insertion[i].insertion_pos, b_info);
    }
    assert(new_t_string.size() == bond_infos.size());

    return result;
}

struct VSNbPair {
    int nb_add;
    VS sent;
    VSNbPair(int nb_add, VS sent) :
            nb_add(nb_add), sent(sent) {
    }
    bool operator<(const VSNbPair& other) const {
        if (nb_add != other.nb_add)
            return nb_add < other.nb_add;
        return sent < other.sent;
    }
};


/********************************************************************************************************
 * Implementation of LatticeComputationManager
 */

LatticeComputationManager::LatticeComputationManager(const std::vector<AugmentedPattern>& initial_patterns,
            const InitialHypothesesStorage *trans_cand, const latt::StateFeaturesWeight& lm_weight, const latt::AdditiveFeaturesWeight& weights,
            const latt::LanguageModel& lm, t_log::TranslationLog* translation_log)
                : lm_weight_(lm_weight), weights_(weights), vocab_(lm.GetVocabulary()), lm_(lm), state_updater_(lm),
                  translation_log_(translation_log) {
    // We add each initial hypothesis to the relevant ref_to_stack_
    BOOST_FOREACH(const AugmentedPattern& pattern, initial_patterns) {
        RefType ref(pattern); // construct the ref from the augmented pattern (note that different augmented pattern can give the same ref)

        // Then get the corresponding list of initial hypotheses
        std::vector<Hypothesis*> hyp_list = trans_cand->_get_cand_list_for_pattern(pattern);

        // Insert all the initial hypotheses into ref_to_stack_[ref]
        ref_to_stack_[ref].insert(ref_to_stack_[ref].end(), hyp_list.begin(), hyp_list.end());

        if (pattern.contains_root()) { // if the pattern contains the root, then the corresponding ref is the final ref
            assert(!final_ref_.is_valid() || final_ref_ == ref); // we cannot have 2 different final_ref
            final_ref_ = ref;
        }
    }
    assert(final_ref_.is_valid());
}

latt::StateLatticeComputer& LatticeComputationManager::get_final_lattice_computer() {
    return get_lattice_computer_for_ref(final_ref_);
}

latt::ParentRelationship convert_pr(ParentBondRelation pr_original) {
    latt::ParentRelationship pr;
    switch (pr_original) {
    case UNK:
        pr = latt::UNKNOWN;
        break;
    case PRE:
        pr = latt::BEFORE;
        break;
    case POST:
        pr = latt::AFTER;
        break;
    default:
        LOG(FATAL) << "Bug here\n";
    }
    return pr;
}

void LatticeComputationManager::make_lattice_computer_for_ref(const RefType& ref) {
    //std::cerr << "making lattice for ref " << ref << std::endl;
    assert(map_to_sublattice_computer_.count(ref) == 0);

    // create a BaseLattice sublattice for the ref
    base_lattice_vector_.push_back(new latt::BaseLattice);
    latt::BaseLattice& lattice = base_lattice_vector_.back();

    // We then populate the sublattice with edges using the initial hypotheses
    if (ref_to_stack_.count(ref) > 0) { // in general, we should have some hypotheses for the given ref
        std::set<VSNbPair> duplicate_detection_set; // used for removing duplicate hypotheses
        // TODO(fabien): duplicate detection should be done offline; also currently duplicate selection is not checking for features
        assert(ref_to_stack_.count(ref) > 0);
        assert(ref_to_stack_.at(ref).size() > 0);

        BOOST_FOREACH(const Hypothesis* hyp, ref_to_stack_.at(ref)) {
            //double score = hyp->feature().compute_score(weights_);
            const FeatureVector& feat = hyp->feature();
            assert(feat.at("example_penalty") == 1);

            latt::ParentRelationship pr = convert_pr(hyp->parent_bond_relation());

            if (hyp->additionals().size() == 0) { // if no additionals, we straightforwardly add the hypothesis
                VSNbPair vs_nb_pair(0, hyp->t_string());
                if (duplicate_detection_set.count(vs_nb_pair) == 0) {
                    duplicate_detection_set.insert(vs_nb_pair);
                    std::vector<BondInfos> bond_infos = make_bond_infos(hyp->t_string(), hyp->get_t_dpnd());
                    add_base_hyp_to_lattice(hyp->t_string(), &lattice, feat, hyp, bond_infos, pr);
                }
            } else { // if hyp has additionals, we convert it to a non-additionals version
                assert(feat.at("nb_additionals") == hyp->additionals().size());
                // TODO(fabien): create more complex lattice of hypotheses for handling additionals
                std::pair<VS, std::vector<BondInfos> > t_string_bond_infos = make_t_string_from_hyp_with_additionals(hyp);
                VS& new_t_string = t_string_bond_infos.first;

                VSNbPair vs_nb_pair(hyp->additionals().size(), new_t_string);
                if (duplicate_detection_set.count(vs_nb_pair) == 0) {
                    duplicate_detection_set.insert(vs_nb_pair);
                    add_base_hyp_to_lattice(new_t_string, &lattice, feat, hyp, t_string_bond_infos.second, pr);
                }
            }
        }

        if (ref == final_ref_) lattice.insert_final_EOS_edge(0.0, lm_.GetVocabulary());
    } else { // If we have no ref, we still add a "bridging" edge with high penalty
        assert(!(ref == final_ref_));
        DLOG(WARNING) << "No hypotheses existing for ref " << ref << std::endl; //TODO(fabien): check this for sibling case
        latt::FreeEdge edge_epsilon_impossible = latt::FreeEdge::make_impossible();
        edge_epsilon_impossible.set_prefered_parent_relationship(latt::UNKNOWN);
        lattice.add_edge(edge_epsilon_impossible, latt::kInitialVertex, latt::kFinalVertex);
    }
    // The base sublattice has now been constructed.

#ifndef NDEBUG
    lattice.check_global_consistency();
#endif

    translation_log_->add_nb_original_vertice(lattice.nb_vertices());

    // We minimize the lattice
    if (!Params::Translator::disable_lattice_reduction) {
        lattice._minimize_lattice_from_end2();
        lattice._minimize_lattice_from_start();
    }

    translation_log_->add_nb_reduced_vertice(lattice.nb_vertices());

//    std::cerr << "Reduced Ref start: " << ref << " vertices: " << lattice.nb_vertices() <<
//                " edges: " << lattice.nb_connected_edges() << std::endl;

    lattice.finalize();

    // The last step is to actually create the BiLmLatticeComputer object by giving it the sublattice
    RefType key = ref;
    map_to_sublattice_computer_.insert(key, new latt::StateLatticeComputer(lattice, state_updater_, lm_weight_, (ref == final_ref_), translation_log_));

    std::stringstream ss;
    ss << ref;
    map_to_sublattice_computer_.at(key).set_debug_id(ss.str());
    //std::cerr << "just created " << map_to_sublattice_computer_.at(key) << std::endl;
}

void LatticeComputationManager::add_base_hyp_to_lattice(const VS& t_string,
        latt::BaseLattice* lattice, const std::map<std::string, double>& feat, const Hypothesis* hyp,
        const std::vector<BondInfos>& bond_infos, latt::ParentRelationship pr) {
    //assert(t_string.size() > 0);
    assert(t_string.size() == bond_infos.size());

    // As a "convention", we put all of the "global" information of the hypothesis in an epsilon-edge at the begining
    latt::FreeEdge initial_epsilon_score(0.0);
    initial_epsilon_score.set_sparse_feature_from_map(feat);

    initial_epsilon_score.get_modifiable_concatenable_information()->set_hyp(hyp);
    initial_epsilon_score.set_prefered_parent_relationship(pr);

    initial_epsilon_score.set_score_from_weights(weights_);


    if (t_string.size() == 0) { // case of an empty t_string
        //full_epsilon.set_sparse_feature("empty_example", 1.0);
        lattice->add_edge(initial_epsilon_score, latt::kInitialVertex, latt::kFinalVertex);
        return;
    }

    // We define a vertex intermediate_kInitialVertex, and connect the initial_epsilon_score from kInitialVertex to intermediate_kInitialVertex
    int offset_vect_id = lattice->nb_vertices();
    if (offset_vect_id == 0) offset_vect_id = 2; // taking into account kInitialVertex/ kFinalVertex
    assert(offset_vect_id != latt::kInitialVertex && offset_vect_id != latt::kFinalVertex);
    latt::VertexId intermediate_kInitialVertex = offset_vect_id;
    lattice->add_edge(initial_epsilon_score, latt::kInitialVertex, intermediate_kInitialVertex);

    // For each word or ref in t_string, we add a relevant edge
    for (int i = 0; i < static_cast<int>(t_string.size()); ++i) {
        latt::VertexId v_start = (i == 0) ? intermediate_kInitialVertex : (i + offset_vect_id);
        latt::VertexId v_end = (i == (static_cast<int>(t_string.size()) - 1)) ? latt::kFinalVertex : i + offset_vect_id + 1;

        assert(i == 0 || v_start != latt::kInitialVertex);
        assert(i == (static_cast<int>(t_string.size()) - 1) || v_end != latt::kFinalVertex);

        RefType ref_id = RefType::get_ref_id(t_string[i]);

        if (ref_id.is_valid()) { // case of a ref
            assert(bond_infos[i].type != latt::Bond::INVALID);
            latt::Bond bond(&get_lattice_computer_for_ref(ref_id), bond_infos[i].type, bond_infos[i].pr);
            bond.set_insertion_id(ref_id);
            latt::FreeEdge edge_X(bond, 0.0);
            edge_X.set_score_from_weights(weights_);
            lattice->add_edge(edge_X, v_start, v_end);
        } else {  // case of a word
            std::string word = t_string[i];
            bool is_optional = is_word_optional(word);
            if (is_optional) {
                word = clean_optional_word(word);
            }

            latt::FreeEdge free_edge(word, 0.0, vocab_);
            free_edge.set_score_from_weights(weights_);
            lattice->add_edge(free_edge, v_start, v_end);

            if (is_optional) { // if the word is optional, we need to add an epsilon edge to allows the bypassing of the word
                //const latt::FreeEdge& free_edge = free_edge_register->get_epsilon_edge(0.0);
                latt::FreeEdge free_edge_epsilon(0.0);
                free_edge_epsilon.set_sparse_feature("nb_optionals_skipped", 1.0);
                free_edge_epsilon.set_score_from_weights(weights_);
                lattice->add_edge(free_edge_epsilon, v_start, v_end);
//                if (v_start == intermediate_kInitialVertex && v_end == latt::kFinalVertex) {
//                    DLOG(WARNING) << "Optional reducing to epsilon hypothesis." << std::endl;
//                }
            }
        }
    }
}

latt::StateLatticeComputer& LatticeComputationManager::get_lattice_computer_for_ref(const RefType& ref) {
    if (map_to_sublattice_computer_.count(ref) == 0) {
        make_lattice_computer_for_ref(ref);
    }
    return map_to_sublattice_computer_.at(ref);
}

