/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include <boost/foreach.hpp>
#include <boost/ptr_container/ptr_map.hpp>
#include <boost/ptr_container/ptr_vector.hpp>

#include <string>
#include <map>
#include <iostream>
#include <vector>

#include "lm/model.hh"
#include "../hypothesis.hpp"
#include "base_lattice.hpp"
#include "../hypotheses_storage.hpp"
#include "state_lattice.hpp"
#include "translation_log.hpp"

#ifndef HYPOTHESES_TO_LATTICE_HPP_
#define HYPOTHESES_TO_LATTICE_HPP_

typedef InputInsertionId RefType;

struct BondInfos {
    latt::Bond::Type type;
    latt::ParentRelationship pr;
};


/*******************************************************************************************
 * Class in charge of translating a sentence. It essentially manages the creation of the sublattices
 * corresponding to each subtree of the input sentence.
 */
class LatticeComputationManager {
  public:
    /*! Constructor
     *   @params initial_patterns: a list of AugmentedPatterns associated with each pattern in trans_cand
     *   @params trans_cand: a TranslationCandidate object storing the initial hypotheses
     *   @params weights: the weights for non-local stateful feature
     *   @params weights: the weights to for local features
     *   @params lm: language model object, including rest costs
     *   @params translation_log: a log object (just for logging, has no effect on result)
     */
    LatticeComputationManager(const std::vector<AugmentedPattern>& initial_patterns,
            const InitialHypothesesStorage *trans_cand, const latt::StateFeaturesWeight&  lm_weight, const latt::AdditiveFeaturesWeight& weights,
            const latt::LanguageModel& lm, t_log::TranslationLog* translation_log);


    /*! Return a BiLmLatticeComputer object, that hold a delayed computation of the translated lattice of the input sentence
     * Calling get_final_lattice_computer will initialize the lattices, but the actual decoding is delayed through a "lazy" paradigm
     * Calling the get_bilmlattice on the returned BiLmLatticeComputer object will actually perform the decoding
     */
    latt::StateLatticeComputer& get_final_lattice_computer();


  private:
    boost::ptr_map<RefType, latt::StateLatticeComputer> map_to_sublattice_computer_;
    boost::ptr_vector<latt::BaseLattice> base_lattice_vector_;

    typedef std::map<RefType, std::vector<Hypothesis*> > RefToStackMap;
    RefToStackMap ref_to_stack_;

    RefType final_ref_;
    latt::StateFeaturesWeight lm_weight_;
    latt::AdditiveFeaturesWeight weights_;

    const latt::LanguageModel::Vocabulary& vocab_;
    const latt::LanguageModel& lm_; // language model
    const latt::StateUpdater state_updater_;

    t_log::TranslationLog* translation_log_; // for logging

    /// return a BiLmLatticeComputer object representing the delayed lazy decoding of the subtree rooted at ref
    latt::StateLatticeComputer& get_lattice_computer_for_ref(const RefType& ref);

    /// Create a BiLmLatticeComputer object representing the delayed lazy decoding of the subtree rooted at ref
    void make_lattice_computer_for_ref(const RefType& ref);

    /*! add_base_hyp_to_lattice is used when constructing the initial lattice. It converts an initial hypothesis
     * into a lattice form and add it to the BaseLattice object representing the lattice of translations of a subtree
     *  @params t_string: t_string of the added initial hypothesis
     *  @params lattice: sub_lattice to which the hypothesis is added
     *  @params feat: feature map of the initial hypothesis
     *  @params hyp: pointer to the initial hypothesis (used for hypothesis reconstruction)
     *  @params bond_infos: additional informations about the bonds in the t_string
     *  @params pr: prefered parent relationship of the hypothesis
     */
    void add_base_hyp_to_lattice(const VS& t_string,
                                    latt::BaseLattice* lattice, const std::map<std::string, double>& feat, const Hypothesis* hyp,
                                    const std::vector<BondInfos>& bond_infos, latt::ParentRelationship pr);
};

#endif /* HYPOTHESES_TO_LATTICE_HPP_ */
