/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _LATT_STATE_HPP_
#define _LATT_STATE_HPP_

#include <iostream>

#include "base_lattice.hpp"
#include "lattice_generic.hpp" // for generic::StateInterface
#include "lm_history.hpp"

namespace latt {

/***************************************************************************************************
 * BiLmState is the state class used during the decoding of BaseLattice into a BiLmLattice.
 * Such a state should contains all the stateful information needed to compute a score when
 * traversing an edge of the BaseLattice.
 * In particular, it should contain the n-gram history needed for computing the language model.
 */
class LatticeState: public generic::StateInterface {
  public:
//    friend class StateUpdateManagement;
//    friend class StateLatticeComputer;

//    template<typename U> friend class StateManager<U>;

    friend std::ostream& operator<<(std::ostream& stream, const LatticeState& state);

    const VertexId& v_id() const {
        return id_;
    }

    const History& history() const {
        return history_;
    }

    History& get_modifyable_history() {
        return history_;
    }

    /// As BiLmStates are currently stored in std::map, the merging of equivalent states will depend
    /// on the correct definition of operator<
    bool operator<(const LatticeState& other) const {
        if (history_ < other.history_) return true;
        if (other.history_ < history_) return false;

#ifdef USE_PR
        if (prefered_parent_relationship_ < other.prefered_parent_relationship_) return true;
        if (other.prefered_parent_relationship_ < prefered_parent_relationship_) return false;
#endif // USE_PR
//        if (pos_of_head_ < other.pos_of_head_) return true;
//        if (other.pos_of_head_ < pos_of_head_) return false;

        return false;
    }

    /// to_stream function, for debugging
    void to_stream(std::ostream* stream) const {
        (*stream) << "<" << id_ << "|" << history_ << ">";
    }

    void set_prefered_parent_relationship(ParentRelationship pr) {
        prefered_parent_relationship_ = pr;
    }

    ParentRelationship prefered_parent_relationship() const {
        return prefered_parent_relationship_;
    }

    static LatticeState* make_initial_state(bool starts_with_BOS, LatticeState* result) {
        //LatticeState* result = new LatticeState(kInitialVertex);
        result->history_ = History(starts_with_BOS);
        //result->prefered_parent_relationship_ = UNKNOWN;
        return result;
    }
//    static LatticeState* make_initial_state(bool starts_with_BOS) {
//        LatticeState* result = new LatticeState(kInitialVertex);
//        result->history_ = History(starts_with_BOS);
//        //result->prefered_parent_relationship_ = UNKNOWN;
//        return result;
//    }

    explicit LatticeState(const VertexId& id) : 
            id_(id), history_(), prefered_parent_relationship_(UNSET) {
    }

  private:
    VertexId id_; // vertex in the BaseLattice to which this state belong
    History history_; // ngram history (left + right)
    ParentRelationship prefered_parent_relationship_; //prefered parent relationship of the current path
    //std::string pos_of_head_;



//    BiLmState(const VertexId& id, const History& history) :
//            id_(id), history_(history), pos_of_head_(), parent_rel_(UNKNOWN){
//    }
};


/***************************************************************************************************
 * StateUpdater is in charge of updating a state after going through an edge
 */

class StateUpdater {
  public:
    explicit StateUpdater(const LanguageModel& lm): lm_(lm) {}
    /// update the state starting_state when going through free_edge. The result is written in new_state
    StateFeatures compute_state_transition_simple(const LatticeState& starting_state, const FreeEdge& free_edge,
                        LatticeState* new_state) const;
    /// update the state starting_state when going through an edge set compacted as a transition_state. The result is written in new_state
    StateFeatures compute_state_transition(const LatticeState& starting_state, const LatticeState& transition_state, const FreeEdge& free_edge,
                        LatticeState* new_state) const;

  private:
    const LanguageModel& lm_;
};

} //namespace latt

#endif /* BILM_STATE_HPP_ */
