/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "lattice_generic.hpp"
#include <limits>
#include <vector>
#include <cassert>
#include <iostream>


namespace latt {
namespace generic {

KBestComputingUnit::KBestComputingUnit(const std::vector<SourceEdgePair>& incoming_edges, const StateInterface* state)
        : incoming_edges_(incoming_edges),
          priority_queue_of_prev_k_best_edge_(incoming_edges.size() + 1), //optimisation: after the first call to get_k_score, priority queue should have this size
          size_(0), state_(state), exhausted_(false) {
    //TODO(fabien): consider reserving here, as allowed by trick in
     // http://stackoverflow.com/questions/3666387/c-priority-queue-underlying-vector-container-capacity-resize

    // TODO(fabien): lazy initialization ?

    EdgeKStruct ek(0, 0, 0, 0); // for optimisation purpose, we just create one ek and edit the num_edge and score fields before pushing
    for (ek.num_edge = 0; ek.num_edge < static_cast<int>(incoming_edges_.size()); ++ek.num_edge) {
        SourceEdgePair src_edge = incoming_edges[ek.num_edge];

        assert(src_edge.source->size() != 0);
        assert(src_edge.edge->size() != 0);
        assert(src_edge.source->get_k_score(0) != kBadScore);
        assert(src_edge.edge->get_k_score(0) != kBadScore);

        size_ += src_edge.source->size() * src_edge.edge->size();
        ek.score = src_edge.source->get_k_score(0) + src_edge.edge->get_k_score(0);
        assert(ek.score != kBadScore);


       //std::cerr << *state_ << " initial push: " << ek << std::endl;
       priority_queue_of_prev_k_best_edge_.push(ek);
        }
        // std::cerr << "done initialize_priority_queue_of_prev_k_best_edge_ " << *this << std::endl;
    }

void KBestComputingUnit::make_first_hash() {
    assert(computed_k_best_edges_.size() == 1);
    assert(computed_hash_.size() == 0);
    assert(duplicate_check_set_.size() == 0);
    EdgeKStruct ek0 = get_k_edge(0);
    assert(ek0.k_of_previous == 0);
    assert(ek0.sub_edge == 0);
    AssociativePathHash* hash_of_this = incoming_edges_[ek0.num_edge].edge->get_initial_hash();
    incoming_edges_[ek0.num_edge].edge->set_k_hash_by_left(0, hash_of_this);
    incoming_edges_[ek0.num_edge].source->set_k_hash_by_left(0, hash_of_this);
    add_hash(hash_of_this);
    assert(computed_k_best_edges_.size() == 1);
    assert(computed_hash_.size() == 1);
    assert(duplicate_check_set_.size() == 1);
}

void KBestComputingUnit::add_hash(AssociativePathHash* hash) {
    assert(duplicate_check_set_.count(hash) == 0U);
    duplicate_check_set_.insert(hash);
    computed_hash_.push_back(hash);
}

bool KBestComputingUnit::is_hash_duplicate(const AssociativePathHash* hash) const {
    return duplicate_check_set_.find(hash) != duplicate_check_set_.end();
}

bool KBestComputingUnit::update_kbest_to(size_t k) {
//    assert(!initial_);
    // computed_k_best_edges_.reserve(k); TODO(fabien): could this do more harm than good if k incremented by 1 each call?
    // std::cerr << "in update_prev_k_best_edge_to_k for " <<  k << " " << *this << std::endl;

    int current_duplicate_run = 0;
    while (k >= computed_k_best_edges_.size() && !priority_queue_of_prev_k_best_edge_.empty() && !exhausted_) {
        EdgeKStruct ek = priority_queue_of_prev_k_best_edge_.top();
        priority_queue_of_prev_k_best_edge_.pop();
        //std::cerr << *state_ << " poped " << ek << " left: " << priority_queue_of_prev_k_best_edge_.size() << std::endl;

        bool is_duplicate = false;
        if (computed_k_best_edges_.size() > 0) { // if it was first element, we know no duplicate
            if (duplicate_check_set_.size() == 0) {
                assert(computed_k_best_edges_.size() == 1);
                //we have skiped the creation of the hash of the first element, so we create it here
                make_first_hash();
            }
            assert(computed_k_best_edges_.size() == computed_hash_.size());
            assert(computed_k_best_edges_.size() == duplicate_check_set_.size());

            AssociativePathHash* hash_of_this = incoming_edges_[ek.num_edge].edge->get_initial_hash();
            incoming_edges_[ek.num_edge].edge->set_k_hash_by_left(ek.sub_edge, hash_of_this);
            incoming_edges_[ek.num_edge].source->set_k_hash_by_left(ek.k_of_previous, hash_of_this);

            if (is_hash_duplicate(hash_of_this)) {
                is_duplicate = true;
                delete hash_of_this;
            } else {
                add_hash(hash_of_this);
            }
        } else {
            assert(duplicate_check_set_.size() == 0);
            assert(computed_hash_.size() == 0);
            assert(duplicate_check_set_.size() == 0);
        }

        //is_duplicate = false;

        if (!is_duplicate) {
            computed_k_best_edges_.push_back(ek);
            current_duplicate_run = 0;
        } else {
            current_duplicate_run += 1;
            if (current_duplicate_run > 100) {
                //TODO(fabien): try a more subtle strategy, for example by considering duplicate runs on an
                //    incoming edges basis, and skip ahead instead of stoping
                exhausted_ = true;
                break;
            }
        }

        double current_source_score = incoming_edges_[ek.num_edge].source->get_k_score(ek.k_of_previous);
        double current_edge_score = incoming_edges_[ek.num_edge].edge->get_k_score(ek.sub_edge);
        assert(current_source_score != kBadScore);
        assert(current_edge_score != kBadScore);
        assert(ek.score == current_source_score + current_edge_score);

        double next_source_score = incoming_edges_[ek.num_edge].source->get_k_score(ek.k_of_previous +1);
        assert(next_source_score <= current_source_score); // TODO(fabien): might not be true when approximations are done, check for better conditions

        if (next_source_score != kBadScore) {
            EdgeKStruct new_ek(ek.num_edge, ek.sub_edge, ek.k_of_previous + 1,
                                    next_source_score + current_edge_score);
            //std::cerr << *state_ << " pushing inc-source " << new_ek << std::endl;
            priority_queue_of_prev_k_best_edge_.push(new_ek);
        }

        double next_edge_score = incoming_edges_[ek.num_edge].edge->get_k_score(ek.sub_edge +1);
        assert(next_edge_score <= current_edge_score); // TODO(fabien): might not be true when approximations are done, check for better conditions
        if (next_edge_score != kBadScore) {
            EdgeKStruct new_ek(ek.num_edge, ek.sub_edge + 1, ek.k_of_previous,
                                next_edge_score + current_source_score);
            //std::cerr << *state_ << " pushing inc-edge " << new_ek << std::endl;
            priority_queue_of_prev_k_best_edge_.push(new_ek);
        }
    }
//    std::cerr << "leaving update_prev_k_best_edge_to_k for " <<  k << " " << *this << std::endl;
    if (k >= computed_k_best_edges_.size()) {
        assert(priority_queue_of_prev_k_best_edge_.empty() || exhausted_);
        return false;
    } else {
        return true;
    }
}

bool KBestComputingUnit::set_k_path(int k, PathInterface *path) {
//    if (initial_) {
//        return true;
//    }
    EdgeKStruct ek = get_k_edge(k);
    //std::cerr << "set_k_best_path k:" << k << " v:" << vertex_id_ << " ini:" << is_initial << " " << ek
    //        << " P:";
//    if (!is_initial)
//        std::cerr << *path;
//    std::cerr << std::endl;

    if (ek.is_special()) {
        return false;
    }

    path->add_state(state_);

    assert(ek.num_edge < static_cast<int>(incoming_edges_.size()) && ek.num_edge >= 0);
    bool edge_ok = incoming_edges_[ek.num_edge].edge->set_k_path(ek.sub_edge, path);
    assert(edge_ok);
    (void) edge_ok; // for avoiding compiler warning about unused variable when NDEBUG is set

    bool prev_ok = incoming_edges_[ek.num_edge].source->set_k_path(ek.k_of_previous, path);
    assert(prev_ok);
    return prev_ok;
}

//void GenericLatticeManager::set_topological_order(const std::vector<const StateInterface*>& states) {
//        topological_order_ = states;
////        std::cerr << "set_topological_order " << incoming_edges_.size() << " " << topological_order_.size() << "\n";
////        BOOST_FOREACH(const StateInterface* st, states) {
////            std::cerr << *st << std::endl;
////        }
////        std::cerr << *this;
//        assert(initial_state_ != NULL);
//        assert(incoming_edges_.size()+1 == topological_order_.size());
//        assert(topological_order_[0] == initial_state_);
//        for (int i = 1; i < static_cast<int>(topological_order_.size()); ++i) {
//            assert(incoming_edges_.count(topological_order_[i]) > 0);
//        }
//        topological_order_has_been_set_ = true;
//    }

//void GenericLatticeManager::make_kbest_for_state(const StateInterface* st) {
//    assert(initial_state_ != NULL);
//    assert(state_to_kbest_.count(st) == 0);
//
////    std::cerr << "making kbest for " << st << std::endl;
//    if (st == initial_state_) {
//        assert(incoming_edges_.count(st) == 0);
//        assert(!starting_edge_);
//        starting_edge_.reset(new StartingEdge(*st));
////            k_bests_.push_back(new KBestComputingUnit(st));
//        state_to_kbest_[st] = starting_edge_.get();
//    } else {
//        assert(incoming_edges_.count(st) > 0 && incoming_edges_.at(st).size() > 0); // we do not want to deal with unreachale states here
//
//        const std::vector<StateEdgePair>& state_edge_vector = incoming_edges_.at(st);
//        std::vector<SourceEdgePair>& source_edge_vect = incoming_edges_as_edge_interfaces_[st];
//        source_edge_vect.reserve(state_edge_vector.size());
//        BOOST_FOREACH(const StateEdgePair& state_edge, state_edge_vector) {
////            if (state_to_kbest_.count(state_edge.state) == 0) std::cerr << "error: no kbest computed: " << state_edge.state << std::endl;
//            assert(state_to_kbest_.count(state_edge.state) > 0); // checking all previous nodes have been initialized
//            source_edge_vect.push_back(SourceEdgePair(state_to_kbest_[state_edge.state], state_edge.edge));
//        }
//
//        k_bests_.push_back(new KBestComputingUnit(source_edge_vect, st));
//        state_to_kbest_[st] = &k_bests_.back();
//    }
//}

//double GenericLatticeManager::get_best_score_for_state(const StateInterface* st) const {
//    assert(initial_state_ != NULL);
//    assert(state_to_kbest_.count(st) == 0);
//
//    if (st == initial_state_) {
//        assert(incoming_edges_.count(st) == 0);
//        return 0.0;
//    } else {
//        assert(incoming_edges_.count(st) > 0 && incoming_edges_.at(st).size() > 0); // we do not want to deal with unreachale states here
//        double max_score = - std::numeric_limits<double>::infinity();
//        const std::vector<StateEdgePair>& state_edge_vector = incoming_edges_.at(st);
//        BOOST_FOREACH(const StateEdgePair& state_edge, state_edge_vector) {
//            assert(state_to_kbest_.count(state_edge.state) > 0); // checking all previous nodes have been initialized
//            double edge_score = state_to_kbest_.at(state_edge.state)->get_k_score(0) + state_edge.edge->get_k_score(0);
//            if (edge_score > max_score) max_score = edge_score;
//        }
//        assert(max_score != - std::numeric_limits<double>::infinity());
//        return max_score;
//    }
//}

//void GenericLatticeManager::make_k_best_for_all() {
//    assert(topological_order_has_been_set_);
//    assert(initial_state_ != NULL);
//    assert(!starting_edge_);
//
//    //std::cerr << "make_k_best_for_all :\n" << *this;
//    BOOST_FOREACH(const StateInterface* st, topological_order_) {
//        //std::cerr << "make_k_best_for_all st: " << *st << std::endl;
//        make_kbest_for_state(st);
//    }
//    assert(incoming_edges_.count(initial_state_) == 0);
//    assert(state_to_kbest_.size() == incoming_edges_.size() + 1);
//    assert(topological_order_.size() == state_to_kbest_.size());
//    assert(incoming_edges_as_edge_interfaces_.size() == incoming_edges_.size());
//}


std::ostream& operator<<(std::ostream& stream, const GenericLatticeManager& lattice) {
    //assert(lattice.topological_order_has_been_set_);
    assert(lattice.initial_state_ != NULL);
    stream << "Begin GLattice";
//    if (lattice.topological_order_has_been_set_) stream << " topo-is-set";
//    stream << "\n";
//    if (lattice.topological_order_has_been_set_) {
//        BOOST_FOREACH(const StateInterface* st, lattice.topological_order_) {
//            if (st == lattice.initial_state_) {
//                assert(lattice.incoming_edges_.count(st) == 0);
//                continue;
//            }
//            BOOST_FOREACH(StateEdgePair st_edge, lattice.incoming_edges_.at(st)) {
//                stream << *st_edge.state << " --[" << *st_edge.edge << "]--> " << *st << "\n";
//            }
//        }
//    } else {
//        BOOST_FOREACH(const GenericLatticeManager::StateEdgesMap::value_type& key_val, lattice.incoming_edges_) {
//            if (key_val.first == lattice.initial_state_) {
//                assert(lattice.incoming_edges_.count(key_val.first) == 0);
//                continue;
//            }
//            BOOST_FOREACH(StateEdgePair st_edge, key_val.second) {
//                stream << *st_edge.state << " --[" << *st_edge.edge << "]--> " << *key_val.first << "\n";
//            }
//        }
//    }

    stream << "End GLattice\n";
    return stream;
}

} // namespace latt
} // namespace generic
