/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef LATTICE_GENERIC_HPP_
#define LATTICE_GENERIC_HPP_

#include <boost/foreach.hpp>
#include <boost/noncopyable.hpp>
#include <boost/ptr_container/ptr_vector.hpp>
#include <boost/scoped_ptr.hpp>
#include <boost/unordered_map.hpp>

#include <cassert>
#include <map>
#include <vector>
#include <limits>
#include <queue>
#include <set>

namespace latt {
namespace generic {

class PathInterface;

class AssociativePathHash {
  public:
    virtual ~AssociativePathHash() {}
    //virtual uint64_t hash() const = 0;
    virtual bool operator<(const AssociativePathHash& other) const = 0;
    virtual void add_by_left(AssociativePathHash* other) const = 0;
};

struct CompareAssociativePathHashPointers {
    bool operator()(const AssociativePathHash* a, const AssociativePathHash* b) const {
        return (*a) < (*b);
    }
};

static const double kBadScore = -std::numeric_limits<double>::max();

class EdgeInterface: public boost::noncopyable {
  public:
    friend std::ostream& operator<<(std::ostream& stream, const EdgeInterface& edge);
    virtual ~EdgeInterface() {}

    virtual size_t size() const = 0;
    virtual double get_k_score(int k) = 0;
    virtual bool set_k_path(int k, PathInterface* path) = 0;

    //for duplicate management
    virtual void set_k_hash_by_left(int k, AssociativePathHash*) = 0;
    virtual AssociativePathHash* get_initial_hash() const = 0;

    virtual void to_stream(std::ostream* stream) const {
        (*stream) << "E<" << size() << ">";
    }
};

inline std::ostream& operator<<(std::ostream& stream, const EdgeInterface& edge) {
    edge.to_stream(&stream);
    return stream;
}


class StateInterface: public boost::noncopyable {
  public:
    virtual ~StateInterface() {}

    virtual void to_stream(std::ostream* stream) const {
        (*stream) << "<ST>";
    }
};

template<typename Value>
struct StateMap {
    //typedef boost::unordered_map<const StateInterface*, Value> type;
    typedef std::map<const StateInterface*, Value> type;
};

inline std::ostream& operator<<(std::ostream& stream, const StateInterface& state) {
    state.to_stream(&stream);
    return stream;
}

class SpecialInfo: public boost::noncopyable {
  public:
    virtual ~SpecialInfo() {}
};

class PathInterface {
  public:
    virtual ~PathInterface() {}
    virtual void add_edge(EdgeInterface* edge) = 0;
    virtual void add_state(const StateInterface* state) = 0;
    virtual void add_special_info(const SpecialInfo* info) =0;
};

struct EdgeKStruct {
    int num_edge;
    int sub_edge;
    int k_of_previous;
    double score;

    EdgeKStruct(int num_edge, int sub_edge, int k_of_previous, double score) :
            num_edge(num_edge), sub_edge(sub_edge), k_of_previous(k_of_previous), score(score) {
        //assert(!is_special() && !is_initial());
    }

    bool is_special() const {
        return (num_edge == std::numeric_limits<int>::min() && sub_edge == std::numeric_limits<int>::min() &&
                k_of_previous == std::numeric_limits<int>::min()
                && score == kBadScore);
    }


    static EdgeKStruct make_special() {
        EdgeKStruct res(std::numeric_limits<int>::min(), std::numeric_limits<int>::min(),
                std::numeric_limits<int>::min(),
                kBadScore);
        assert(res.is_special());
        return res;
    }
};

inline std::ostream& operator<<(std::ostream& stream, const EdgeKStruct& ek) {
    stream << "EK<" << ek.num_edge << "." << ek.sub_edge << "." << ek.k_of_previous << ", " << ek.score << ">";
    return stream;
}

struct CompareScoreEdgeKStruct {
    bool operator()(const EdgeKStruct& first, const EdgeKStruct& other) {
        return first.score < other.score;
    }
};

class EdgeKReservableQueue: public std::priority_queue<EdgeKStruct, std::vector<EdgeKStruct>, CompareScoreEdgeKStruct> {
  public:
    explicit EdgeKReservableQueue(size_type capacity = 0) { reserve(capacity); }
    void reserve(size_type capacity) {this->c.reserve(capacity);}
    size_type capacity() const {return this->c.capacity(); }
};


struct SourceEdgePair {
    SourceEdgePair(EdgeInterface* source, EdgeInterface* edge): source(source), edge(edge) {}
    EdgeInterface* source;
    EdgeInterface* edge;
};



class KBestComputingUnit: public EdgeInterface {
  public:
    // k_best_size_(0), k_best_has_been_initialized_(false)
friend std::ostream& operator<<(std::ostream& stream, const KBestComputingUnit& kbc);

    KBestComputingUnit(const std::vector<SourceEdgePair>& incoming_edges, const StateInterface* state);

    size_t size() const {
        return size_;
    }

    double get_k_score(int k) {
        //std::cerr << "in get_k_best_previous_score for " <<  k << " " << *this << std::endl;

//        if (initial_) {
//            return (k==0)?0.0:- std::numeric_limits<double>::max();
//        }

        if (k >= static_cast<int>(computed_k_best_edges_.size())) {
            bool could_update = update_kbest_to(k);
            if (!could_update) {
                return kBadScore;
            }
        }
        return computed_k_best_edges_[k].score;
    }

    bool set_k_path(int k, PathInterface *path);

    void to_stream(std::ostream* stream) const {
        (*stream) << "<<" << size() << ":" << *state_ << ">>";
    }

    void set_k_hash_by_left(int k, AssociativePathHash* result) {
        assert(k >= 0);
//        if (computed_k_best_edges_.size() > 0) {
//            if (duplicate_check_set_.size() == 0) {
//                assert(computed_k_best_edges_.size() == 1);
//                make_first_hash();
//                assert(duplicate_check_set_.size() == 1);
//            }
//        }
        assert(computed_k_best_edges_.size() == computed_hash_.size()
                || (computed_k_best_edges_.size() == 1 && computed_hash_.size() == 0));
        update_kbest_to(static_cast<size_t>(k));
        if (duplicate_check_set_.size() == 0) {
            assert(computed_k_best_edges_.size() == 1);
            make_first_hash();
            assert(duplicate_check_set_.size() == 1);
        }
        assert(static_cast<size_t>(k) < computed_hash_.size());
        computed_hash_[k].add_by_left(result);
    }

    AssociativePathHash* get_initial_hash() const {
        assert(incoming_edges_.size() > 0);
        return incoming_edges_[0].edge->get_initial_hash();
    }

  private:
//    bool initial_;
    const std::vector<SourceEdgePair>& incoming_edges_;
    std::vector<EdgeKStruct> computed_k_best_edges_;

    //std::priority_queue<EdgeKStruct, std::vector<EdgeKStruct>, CompareScore> priority_queue_of_prev_k_best_edge_;
    EdgeKReservableQueue priority_queue_of_prev_k_best_edge_;
    size_t size_; // can overflow, but probably not a problem in practice
    const StateInterface* state_;
    bool exhausted_;

    boost::ptr_vector<AssociativePathHash> computed_hash_;
    std::set<const AssociativePathHash*, CompareAssociativePathHashPointers> duplicate_check_set_;

    void add_hash(AssociativePathHash* hash);

    bool is_hash_duplicate(const AssociativePathHash* hash) const;

    void make_first_hash();

    bool update_kbest_to(size_t k);

    EdgeKStruct get_k_edge(int k) {
//        assert(!initial_);
        if (k >= static_cast<int>(computed_k_best_edges_.size())) {
            bool could_update = update_kbest_to(k);
            if (!could_update) {
                return EdgeKStruct::make_special();
            }
        }
        return computed_k_best_edges_[k];
    }
};

struct StateEdgePair {
    StateEdgePair(const StateInterface* state, EdgeInterface* edge): state(state), edge(edge) {}
    const StateInterface* state;
    EdgeInterface* edge;
};

class StartingEdge: public EdgeInterface {
  public:
    explicit StartingEdge(const StateInterface& state): state_(state) {}

    size_t size() const { return 1;}

    double get_k_score(int k) {
        return (k == 0)?0.0:kBadScore;
    }

    bool set_k_path(int k, generic::PathInterface* path) {
        if (k != 0) return false;
        path->add_state(&state_);
        return true;
    }


    void set_k_hash_by_left(int k, AssociativePathHash* result) {}

    AssociativePathHash* get_initial_hash() const {
        assert(false); //should never be called
        return NULL;
    }

  private:
    const StateInterface& state_;
};



struct StateDelayEdgePair {
    EdgeInterface* edge;
    EdgeInterface* source_state_kbest;
};

class KBestComputingUnitDelay: public EdgeInterface {
  public:
    explicit KBestComputingUnitDelay(const StateInterface* st) : st_(st),
            best_score_(-std::numeric_limits<double>::infinity()) {
#ifndef NDEBUG
        should_not_add_more_edges_ = false;
#endif //NDEBUG
    }

    void ensure_kbest() const {
        if (kbest_) return;
#ifndef NDEBUG
        should_not_add_more_edges_ = true;
#endif //NDEBUG
        kbest_.reset(new KBestComputingUnit(incoming_edges_, st_));
    }

    void add_incoming_edge(EdgeInterface* source_kbest,  EdgeInterface* edge) {
        assert(!should_not_add_more_edges_);
        incoming_edges_.push_back(SourceEdgePair(source_kbest, edge));
        double score = source_kbest->get_k_score(0) + edge->get_k_score(0);
        if (score > best_score_) best_score_ = score;
    }

     size_t size() const {
         ensure_kbest();
         return kbest_->size();
     }

     size_t nb_incoming_edges() const {
         return incoming_edges_.size();
     }

     double get_k_score(int k) {
#ifndef NDEBUG
        should_not_add_more_edges_ = true;
#endif //NDEBUG
        if (k == 0) return best_score_;
        ensure_kbest();
        return kbest_->get_k_score(k);
     }

     bool set_k_path(int k, PathInterface* path) {
         ensure_kbest();
         return kbest_->set_k_path(k, path);
     }

     //for duplicate management
     void set_k_hash_by_left(int k, AssociativePathHash* h) {
         ensure_kbest();
         return kbest_->set_k_hash_by_left(k, h);
     }
     AssociativePathHash* get_initial_hash() const {
         ensure_kbest();
         return kbest_->get_initial_hash();
     }

     void to_stream(std::ostream* stream) const {
         if (!kbest_) {
             (*stream) << "Delay<>";
         } else {
             (*stream) << "Delay<";
             kbest_->to_stream(stream);
             (*stream) << ">";
         }
     }

  private:
    const StateInterface* st_;
    double best_score_;
    std::vector<SourceEdgePair> incoming_edges_;
    mutable boost::scoped_ptr<EdgeInterface> kbest_;

#ifndef NDEBUG
    mutable bool should_not_add_more_edges_;
#endif //NDEBUG
};

class GenericLatticeManager: boost::noncopyable {
  public:
    friend std::ostream& operator<<(std::ostream& stream, const GenericLatticeManager& latt);

    GenericLatticeManager():
        initial_state_(NULL) {} //, topological_order_has_been_set_(false) {}

    void set_initial_state(const StateInterface* st) {
        initial_state_ = st;
        assert(!starting_edge_);
        starting_edge_.reset(new StartingEdge(*st));
    }

    void add_edge(const StateInterface* st1, const StateInterface* st2, EdgeInterface* edge) {
//        std::cerr << "adding edge between " << st1 << " and " << st2 << std::endl;
        //assert(state_to_kbest_.count(st1) > 0);
        EdgeInterface* source_kbest = get_kbest_of_state(st1);
        KBestComputingUnitDelay* kbest_st2 = get_or_create_kbest_for_state(st2);
        kbest_st2->add_incoming_edge(source_kbest, edge);
        //topological_order_has_been_set_ = false;
    }



    double get_best_score_for_state(const StateInterface* st) const {
        return get_kbest_of_state(st)->get_k_score(0);
    }

//    void set_topological_order(const std::vector<const StateInterface*>& states);

//    void compute_topological_order() { assert(false);} //not implemented yet

//    size_t size();
//    double get_k_score(int k);
//    void set_k_path(int k, PathInterface* path);

    EdgeInterface& edge_interface_from_state(const StateInterface* st1) {
        return *get_kbest_of_state(st1);
    }

    //void make_k_best_for_all();

    //void make_kbest_for_state(const StateInterface* st);

    size_t get_nb_incoming_edges_for_state(const StateInterface* st) const {
        if (st == initial_state_) {
//            assert(incoming_edges_.count(st) == 0);
            return 0;
        }
        return state_to_kbest_.at(st)->nb_incoming_edges();
    }

  private:
    const StateInterface* initial_state_;

    //typedef std::map<const StateInterface*, std::vector<StateEdgePair> > StateEdgesMap;
    //StateEdgesMap incoming_edges_;

    //std::map<const StateInterface*, std::vector<SourceEdgePair> > incoming_edges_as_edge_interfaces_;
    //typedef StateMap<std::vector<SourceEdgePair> >::type StateToSourceEdgeMap;
    //StateToSourceEdgeMap incoming_edges_as_edge_interfaces_;

    //std::vector<const StateInterface*> topological_order_;
    //bool topological_order_has_been_set_;

    typedef std::map<const StateInterface*, KBestComputingUnitDelay*> StateKBestMap;
    StateKBestMap state_to_kbest_;
    boost::ptr_vector<KBestComputingUnitDelay> k_bests_;

    boost::scoped_ptr<StartingEdge> starting_edge_;

    KBestComputingUnitDelay* get_or_create_kbest_for_state(const StateInterface* st) {
        assert(initial_state_ != NULL);
        assert(initial_state_ != st);

        StateKBestMap::iterator it = state_to_kbest_.find(st);
        if (it != state_to_kbest_.end()) {
            return it->second;
        } else {
            k_bests_.push_back(new KBestComputingUnitDelay(st));
            state_to_kbest_[st] = &k_bests_.back();
            return &k_bests_.back();
        }
    }

    EdgeInterface* get_kbest_of_state(const StateInterface* st) const {
        assert(initial_state_ != NULL);
        assert(starting_edge_);
        if (st == initial_state_) {
            return starting_edge_.get();
        } else {
            assert(state_to_kbest_.count(st) > 0);
            return state_to_kbest_.at(st);
        }
    }
};

} // namespace latt
} // namespace generic

#endif /* LATTICE_GENERIC_HPP_ */
