/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef LM_EDGE_HPP_
#define LM_EDGE_HPP_

#include <stddef.h>
#include <cassert>
#include <iostream>
#include <limits>
#include <string>
#include <typeinfo>

#include "lm/vocab.hh"
#include "base_lattice.hpp"
#include "lattice_generic.hpp"
#include "matrix_hash.hpp"

namespace latt {

//struct LmVertexId {
//    VertexId original_id;
//    int subscript;
//
//    static LmVertexId make_special() {
//        LmVertexId result;
//        result.original_id = kSpecialVertex;
//        result.subscript = std::numeric_limits<int>::min();
//        return result;
//    }
//
//    bool is_special() const {
//        return original_id == kSpecialVertex && subscript == std::numeric_limits<int>::min();
//    }
//};
//
//std::ostream& operator<<(std::ostream& stream, const LmVertexId& vertex);


class MatrixHashFacade: public generic::AssociativePathHash {
  public:
    friend std::ostream& operator<<(std::ostream& stream, const MatrixHashFacade& mat);
    MatrixHashFacade() {}
    explicit MatrixHashFacade(std::string word) {
        uint64_t hash = kenlm::ngram::detail::HashForVocab(word.c_str(), word.size());
        mat_ = m_hash::MatrixHash(hash);
//        seq_.push_back(std::make_pair(word, mat_));
    }
    bool operator<(const AssociativePathHash& other) const {
        if (typeid(other) == typeid(MatrixHashFacade)) {
            const MatrixHashFacade* other_mat = static_cast<const MatrixHashFacade*>(&other);
            return mat_ < other_mat->mat_;
        } else {
            assert(false);
            return false;
        }
    }
    void add_by_left(AssociativePathHash* result) const {
        if (typeid(*result) == typeid(MatrixHashFacade)) {
            MatrixHashFacade* mat_result = static_cast<MatrixHashFacade*>(result);
            mat_.multiply_result_from_left(&(mat_result->mat_));
//            mat_result->seq_.insert(mat_result->seq_.begin(), seq_.begin(), seq_.end());
        } else {
            assert(false);
        }
    }

    bool operator==(const MatrixHashFacade& other) const {
        return mat_ == other.mat_;
    }

  private:
    m_hash::MatrixHash mat_;
//    std::deque<std::pair<std::string, m_hash::MatrixHash> > seq_;
};


class LmEdgeAdapter: public generic::EdgeInterface {
  public:
    LmEdgeAdapter(const FreeEdge& edge, StateFeatures feat,
                const StateFeaturesWeight& weights_): edge_(edge), feat_(feat) {
        score_ = edge.score() + feat_.compute_score(weights_);

        if (edge.type() == FreeEdge::WORD) {
            mat_hash_ = MatrixHashFacade(edge.word());
        }
    }

    size_t size() const { return 1;}

    double get_k_score(int k) {
        assert(score_ != generic::kBadScore);
        return (k == 0)?score_:generic::kBadScore;
    }
    bool set_k_path(int k, generic::PathInterface* path) {
        if (k != 0) return false;
        path->add_edge(this);
        return true;
    }

    void to_stream(std::ostream* stream) const {
         (*stream) << edge_ << "(" << score_ << ")";
     }

    const FreeEdge& free_edge() const {return edge_;}

    double score() const {
        assert(score_ != generic::kBadScore);
        return score_;}

    const StateFeatures& state_features() const {
         return feat_;
     }

    generic::AssociativePathHash* get_initial_hash() const {
        return new MatrixHashFacade;
    }

    void set_k_hash_by_left(int k, generic::AssociativePathHash* result) {
        mat_hash_.add_by_left(result);
    }

  private:
    const FreeEdge& edge_;
    StateFeatures feat_;
    double score_;
    MatrixHashFacade mat_hash_;
};

class EdgeSimpleConnectors: public generic::EdgeInterface {
  public:
    size_t size() const { return 1;}

    double get_k_score(int k) {
        return (k == 0)?0.0:generic::kBadScore;
    }

    bool set_k_path(int k, generic::PathInterface* path) {
        if (k != 0) return false;
        return true;
    }

    void to_stream(std::ostream* stream) const {
        (*stream) << "<-->";
    }

    void set_k_hash_by_left(int k, generic::AssociativePathHash* result) {}

    generic::AssociativePathHash* get_initial_hash() const {
        return new MatrixHashFacade;
    }
};


class FinalLmState: public generic::StateInterface {
    // This class is only used for generating final states
    void to_stream(std::ostream* stream) const {
        (*stream) << "<F>";
    }
};


class EdgeOffsetAdapter: public generic::EdgeInterface {
  public:
    EdgeOffsetAdapter(generic::EdgeInterface* ei, const StateFeatures& st_feat,
            const StateFeaturesWeight& st_weight, const FreeEdge& enveloping_free_edge): ei_(ei),
            st_feat_(st_feat), enveloping_free_edge_(enveloping_free_edge) {
        offset_ = st_feat_.compute_score(st_weight) + enveloping_free_edge_.score();
    }

    size_t size() const { return ei_->size();}

    double get_k_score(int k) {
        assert(offset_ != generic::kBadScore);
        double score = ei_->get_k_score(k);
        if (score == generic::kBadScore) return generic::kBadScore;
        return score + offset_;
    }
    bool set_k_path(int k, generic::PathInterface* path);

    void to_stream(std::ostream* stream) const {
        (*stream) << "[[offset: " << offset_ << "]";
        ei_->to_stream(stream);
        (*stream) << "]";
    }

    generic::EdgeInterface* wrapped_edge() {return ei_;}
    double offset() const {return offset_;}

    const FreeEdge& enveloping_free_edge() const {
        return enveloping_free_edge_;
    }

    const StateFeatures& state_features() const {
        return st_feat_;
    }

    generic::AssociativePathHash* get_initial_hash() const {
        return new MatrixHashFacade;
    }

    void set_k_hash_by_left(int k, generic::AssociativePathHash* result) {
        ei_->set_k_hash_by_left(k, result);
    }

  private:
    generic::EdgeInterface* ei_;
    StateFeatures st_feat_;
    const FreeEdge& enveloping_free_edge_;
    double offset_;
};

class EndOfEdgeOffsetMarker: public generic::EdgeInterface {
};

inline bool EdgeOffsetAdapter::set_k_path(int k, generic::PathInterface* path) {
    bool could_find_path = ei_->set_k_path(k, path);
    path->add_edge(this);
    return could_find_path;

//        if (typeid(*path) == typeid(PathFacadeBiLm)) {
//            PathFacadeBiLm* as_path_facade = static_cast<PathFacadeBiLm *>(path);
//            as_path_facade->add_offset(offset_);
//            bool could_find_path = ei_->set_k_path(k, path);
//            as_path_facade->add_offset(-offset_);
//            return could_find_path;
//        } else {
//            std::cerr << typeid(PathFacadeBiLm).name() << " != " << typeid(*path).name() << std::endl;
//            assert(false);
//            return false;
//        }
    }


class PathFacade: public generic::PathInterface {
  public:
    explicit PathFacade(Path* path): path_(path) {}

    void add_edge(generic::EdgeInterface* edge) {
        if (typeid(*edge) == typeid(LmEdgeAdapter)) {
            LmEdgeAdapter* lm_edge = static_cast<LmEdgeAdapter*>(edge);
            path_->add_edge(lm_edge->free_edge(), lm_edge->score(), lm_edge->state_features());
        } else if (typeid(*edge) == typeid(EdgeOffsetAdapter)) {
            EdgeOffsetAdapter* lm_edge = static_cast<EdgeOffsetAdapter*>(edge);
            path_->add_edge(lm_edge->enveloping_free_edge(), lm_edge->offset(), lm_edge->state_features());
            //path_.add_offset(lm_edge->offset());
//            bool could_find_k_path = lm_edge->wrapped_edge()->set_k_path(k, this);
//            current_score_offset_ -= lm_edge->offset();
        } else if (typeid(*edge) == typeid(EdgeSimpleConnectors)) {
            //do nothing
        } else {
            assert(false);
        }
    }

    void add_state(const generic::StateInterface* state) {
        //        if (typeid(*state) == typeid(LmState)) {
        //            const LmState* lm_state = static_cast<const LmState*>(state);
        //            path_->add_vertex(lm_state->id().original_id);
        //        } else
        if (typeid(*state) == typeid(FinalLmState)) {
            // do nothing
        } else {
            assert(false);
        }
    }

    void add_special_info(const generic::SpecialInfo* info) {
    }

//    void add_offset(double offset) {
//        current_score_offset_ += offset;
//    }
  protected:
    Path* path_;
//    double current_score_offset_;
};

} // namespace latt

#endif /* LM_EDGE_HPP_ */
