/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "lm_history.hpp"

#include <cassert>
#include <limits>

#include "lm/left.hh"
//#include "../../KenLM/lm/virtual_interface.hh"

#include "lm_stream.hpp"

namespace latt {

History::History(): type_(UNSET) {}

History::History(bool starts_with_BOS) {
    starts_with_BOS_ = starts_with_BOS;
    ends_with_EOS_ = false;
    if (starts_with_BOS) {
        type_ = BOS;
    } else {
        type_ = INIT;
    }
}

bool History::operator<(const History& other) const {
    if (other.type_ != type_) {
        return type_ < other.type_;
    }
    if (type_ == CHART) {
        return chart_state_ < other. chart_state_;
    } else {
        return false; //identical
    }
}

double History::set_next(const LanguageModel& model, lm::WordIndex word_index, History* new_history) const {
    new_history->type_ = CHART;
    new_history->starts_with_BOS_ = starts_with_BOS_;

    assert(!ends_with_EOS_);
    if (word_index == model.GetVocabulary().EndSentence()) {
        new_history->ends_with_EOS_ = true;
    } else {
        new_history->ends_with_EOS_ = false;
    }

    lm::ngram::RuleScore<LanguageModel> scorer(model, new_history->chart_state_);
    switch (type_) {
    case BOS:
        scorer.BeginSentence();
        break;
    case INIT:
        break;
    case CHART:
        scorer.BeginNonTerminal(chart_state_);
        break;
    case UNSET:
    default:
        assert(false);
    }
    scorer.Terminal(word_index);
    double prob = scorer.Finish();
    return prob;
}

double History::set_next(const LanguageModel& model, const History& hist_edge, History* new_history) const {
    if (hist_edge.type_ == INIT) { // case of an empty edge
        *new_history = *this;
        return 0.0;
    }
    assert(hist_edge.type_ == CHART);
    new_history->type_ = CHART;

    assert(!(starts_with_BOS_ && hist_edge.starts_with_BOS_) );
    new_history->starts_with_BOS_ = starts_with_BOS_;
    assert(!ends_with_EOS_);
    new_history->ends_with_EOS_ = hist_edge.ends_with_EOS_;

    kenlm::ngram::RuleScore<LanguageModel> scorer(model, new_history->chart_state_);
    switch (type_) {
    case BOS:
        scorer.BeginSentence();
        break;
    case INIT:
        new_history->starts_with_BOS_ = hist_edge.starts_with_BOS_;
        break;
    case CHART:
        scorer.BeginNonTerminal(chart_state_);
        break;
    case UNSET:
    default:
        assert(false);
    }
    scorer.NonTerminal(hist_edge.chart_state_);
    double prob = scorer.Finish();
    return prob;
}

double History::eval_delta_with_BOS_EOS(const LanguageModel& model) const {
    kenlm::ngram::ChartState c_state;
    kenlm::ngram::RuleScore<LanguageModel> scorer(model, c_state);
    switch (type_) {
    case BOS:
        break;
    case INIT:
        scorer.BeginSentence();
        break;
    case CHART:
        if (!starts_with_BOS_) {
            scorer.BeginSentence();
        }
        scorer.NonTerminal(chart_state_);
        break;
    case UNSET:
    default:
        assert(false);
    }
    if (!ends_with_EOS_) {
        scorer.Terminal(model.GetVocabulary().EndSentence());
    }
    double prob = scorer.Finish();
    return prob;
}

double History::eval_delta_with_OOV(const LanguageModel& model) const {
    if (starts_with_BOS_) return 0.0; // no delta if already starts with BOS
    switch (type_) {
    case BOS:
        return 0.0;
        break;
    case INIT:
        return 0.0;
        break;
    case CHART:
        double base_oov_proba, prob;
        {
            kenlm::ngram::ChartState c_state;
            kenlm::ngram::RuleScore<LanguageModel> scorer(model, c_state);
            scorer.Terminal(model.GetVocabulary().NotFound());
            base_oov_proba = scorer.Finish();
        }
        {
            kenlm::ngram::ChartState c_state;
            kenlm::ngram::RuleScore<LanguageModel> scorer(model, c_state);
            scorer.Terminal(model.GetVocabulary().NotFound());
            scorer.NonTerminal(chart_state_);
            prob = scorer.Finish() - base_oov_proba;
        }
        return prob;
        break;
    case UNSET:
    default:
        assert(false);
        return std::numeric_limits<double>::max();
    }
}

std::ostream& operator<<(std::ostream& stream, const History& hist) {
    switch (hist.type_) {
    case History::BOS:
        stream << "<BOS>";
        break;
    case History::INIT:
        stream << "<INIT>";
        break;
    case History::CHART:
        if (hist.starts_with_BOS_)
            stream << "BOS>";
        stream << hist.chart_state_;
        if (hist.ends_with_EOS_)
            stream << "<EOS";
        break;
    case History::UNSET:
        stream << "<UNSET>";
        break;
    default:
        assert(false);
    }
    return stream;
}

} //namespace latt

