/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdint.h>
#include <cassert>
#include <iostream>

#ifndef MATRIX_HASH_HPP_
#define MATRIX_HASH_HPP_

namespace m_hash {

const uint32_t k32BitPrime = 4294967291; //(2**32-5) is prime

inline uint32_t compute_modulo_ab_plus_cd(uint32_t a, uint32_t b, uint32_t c, uint32_t d) {
    uint64_t ab = a;
    ab *= b;
    ab %= k32BitPrime;

    uint64_t cd = c;
    cd *= d;
    cd %= k32BitPrime;

    ab+=cd;
    ab %= k32BitPrime;
    assert((ab>>32) == 0);
    return static_cast<uint32_t>(ab);
}

class MatrixHash {
  public:
    friend std::ostream& operator<<(std::ostream& stream, const MatrixHash& mat);
    MatrixHash(): a1_(1), a2_(0), b1_(0), b2_(1) {}

    // TODO(fabien): since a2 is set to 0 all the time, all multiplications also give 0 in a2 (-> could remove a2?)
    // also note that a1 and b2 have then commutative multiplication;

    explicit MatrixHash(uint64_t idx) {
        uint64_t higher = static_cast<uint32_t>(idx >> 32);
        uint64_t lower = static_cast<uint32_t>((idx & 0x00000000FFFFFFFF));
        higher ^= lower;
        higher %= k32BitPrime;
        a1_ = 13;
        a2_ = 0;
        b1_ = static_cast<uint32_t>(higher);
        b2_ = static_cast<uint32_t>((lower >> 2) + 1);
    }

    void multiply_result_from_left(MatrixHash* result) const {
        uint32_t new_a1, new_a2, new_b1, new_b2;

        new_a1 = compute_modulo_ab_plus_cd(a1_, result->a1_, b1_, result->a2_);
        new_b2 = compute_modulo_ab_plus_cd(a2_, result->b1_, b2_, result->b2_);
        new_a2 = compute_modulo_ab_plus_cd(a2_, result->a1_, b2_, result->a2_);
        new_b1 = compute_modulo_ab_plus_cd(a1_, result->b1_, b1_, result->b2_);

        result->a1_ = new_a1;
        result->a2_ = new_a2;
        result->b1_ = new_b1;
        result->b2_ = new_b2;
    }

//    bool operator<(const MatrixHash& other) const {
//
//    }
//    add_to(AssociativePathHash*);

    bool operator==(const MatrixHash& other) const {
        return (a1_ == other.a1_) && (a2_ == other.a2_) && (b1_ == other.b1_) && (b2_ == other.b2_);
    }

    bool operator<(const MatrixHash& other) const {
        return (b1_ == other.b1_)?
                    ((a1_ == other.a1_)?
                          ((b2_ == other.b2_)?
                                ((a2_ == other.a2_)?false:(a2_ < other.a2_))
                                 :(b2_ < other.b2_)
                          )
                          :(a1_ < other.a1_)
                     )
                    :(b1_ < other.b1_);
    }

  private:
    uint32_t a1_, a2_, b1_, b2_;
};

inline std::ostream& operator<<(std::ostream& stream, const MatrixHash& mat) {
    stream << "Matx {\n";
    stream << mat.a1_ << " " << mat.b1_ << "\n" << mat.a2_ << " " << mat.b2_;
    stream << "}\n";
    return stream;
}

} //namespace m_hash;

#endif /* MATRIX_HASH_HPP_ */
