/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "state_lattice.hpp"

#include <boost/foreach.hpp>
//#include <boost/mpl/aux_/preprocessed/gcc/and.hpp>
//#include <boost/mpl/aux_/preprocessed/gcc/or.hpp>
//#include <boost/ptr_container/detail/void_ptr_iterator.hpp>
//#include <boost/ptr_container/ptr_set_adapter.hpp>
//#include <boost/type_traits/is_const.hpp>
#include <algorithm>
//#include <cstdlib>
#include <set>
#include <string>
#include <typeinfo>
#include <vector>

#include "lm/left.hh"
#include "lm/model.hh"
#include "lm/return.hh"
#include "lm/state.hh"
#include "../params.hpp"

namespace latt {

namespace kenlm = ::lm;

template<typename Model>
void eval_lm_with_different_techniques(const std::vector<std::string>& sentence, const Model& model) {
    const kenlm::ngram::Vocabulary &vocab = model.GetVocabulary();
    {
        kenlm::ngram::State state(model.BeginSentenceState());
        kenlm::ngram::State out_state;
        kenlm::FullScoreReturn ret;
        double total = 0;
        std::cout << "Classic with BOS: ";
        for (int i = 0; i < static_cast<int>(sentence.size()); ++i) {
            const std::string& word = sentence[i];
            ret = model.FullScore(state, vocab.Index(word), out_state);
            state = out_state;
            total += ret.prob;
            std::cout << "+" << ret.prob << "(" << word << ")  |";
        }
        std::cout << "before EOS: " << total << " |";
        ret = model.FullScore(state, vocab.EndSentence(), out_state);
        total += ret.prob;
        std::cout << "Final: " << total << std::endl;
    }
    {
        std::cout << "RuleScorer/ BOS + EOS";
        kenlm::ngram::ChartState c_state0;
        kenlm::ngram::RuleScore<Model> scorer0(model, c_state0);
        scorer0.BeginSentence();
        for (int i = 0; i < static_cast<int>(sentence.size()); ++i) {
            const std::string& word = sentence[i];
            scorer0.Terminal(vocab.Index(word));
        }
        scorer0.Terminal(vocab.EndSentence());
        double total2 = scorer0.Finish();
        std::cout << total2 << std::endl;
    }
    {
        std::cout << "RuleScorer/ noBOS ";
        kenlm::ngram::ChartState c_state1;
        kenlm::ngram::RuleScore<Model> scorer(model, c_state1);
        for (int i = 0; i < static_cast<int>(sentence.size()); ++i) {
            const std::string& word = sentence[i];
            scorer.Terminal(vocab.Index(word));
        }
        double total2 = scorer.Finish();
        std::cout << total2 << "| ";

        kenlm::ngram::ChartState c_state2;
        kenlm::ngram::RuleScore<Model> scorer2(model, c_state2);
        scorer2.BeginSentence();
        scorer2.NonTerminal(c_state1);
        double total3 = scorer2.Finish();
        std::cout << "with BOS: " << total2 + total3 << "|";

        kenlm::ngram::ChartState c_state3;
        kenlm::ngram::RuleScore<Model> scorer3(model, c_state3);
        scorer3.BeginNonTerminal(c_state2);
        scorer3.Terminal(vocab.EndSentence());
        double total4 = scorer3.Finish();
        std::cout << "with BOS+EOS: " << total2 + total3 + total4 << "|" << std::endl;
    }
}

template
void eval_lm_with_different_techniques<kenlm::ngram::ProbingModel>(const std::vector<std::string>& sentence,
        const kenlm::ngram::ProbingModel& model);

template
void eval_lm_with_different_techniques<kenlm::ngram::RestProbingModel>(
        const std::vector<std::string>& sentence, const kenlm::ngram::RestProbingModel& model);


StateEdgeScoreStruct::StateEdgeScoreStruct()
    : state(NULL), edge(NULL) , score(generic::kBadScore) {
    //assert(false); // should never be invoked
}

StateEdgeScoreStruct::StateEdgeScoreStruct(const LatticeState& state, generic::EdgeInterface* edge, double score)
    : state(&state), edge(edge), score(score) {
    //score = edge->get_k_score(0);
}

bool StateEdgeScoreStruct::operator<(const StateEdgeScoreStruct& other) const {
    assert(edge != NULL);
    return score > other.score;
}

std::ostream& operator<<(std::ostream& stream, const StateScorePair& obj) {
    stream << "StateScorePair<" << obj.state << ", " << obj.score << ">";
    return stream;
}


void StateUpdateManagement::make_new_state(const std::vector<StateScorePair>& sorted_starting_states, size_t nb_max_starting_states, const PinedEdge& edge) {
    assert(nb_max_starting_states <= sorted_starting_states.size());

    VertexId v_end_id = edge.v_end();
    const FreeEdge& free_edge = edge.free_edge();

    if (free_edge.type() == FreeEdge::BOND) {
        free_edge.bond().bilm_lattice()->make_all_edges_from_starting_state(sorted_starting_states, nb_max_starting_states, v_end_id, free_edge,
                                                state_updater_, state_weights_, edges_, state_manager_,
                                                      lattice_manager_);
    } else {
        for (size_t i = 0; i < nb_max_starting_states; ++i) {
            const LatticeState& starting_state = *(sorted_starting_states[i].state);
            assert(edge.v_start() == starting_state.v_id());

            //LatticeState* new_state = new LatticeState(v_end_id);
            LatticeState* new_state = state_manager_->get_state_buffer(v_end_id);

            //create state features
            StateFeatures st_feat = state_updater_.compute_state_transition_simple(starting_state, free_edge, new_state);

            // make new edge
            edges_->push_back(new LmEdgeAdapter(free_edge, st_feat, state_weights_));

            // add to lattice manager
//            LatticeState& state = state_manager_->add_or_get_identical_state(v_end_id, new_state);
            LatticeState* state = state_manager_->add_buffer();

            generic::EdgeInterface& new_edge = edges_->back();

//            lattice_manager_->add_edge(&starting_state, &state, &new_edge);
            lattice_manager_->add_edge(&starting_state, state, &new_edge);
        }
    }
}

void StateUpdateManagement::make_initial_state(bool starts_with_BOS) {
    //History history(starts_with_BOS);
//    LatticeState& initial_state = state_manager_->add_or_get_identical_state(kInitialVertex, LatticeState::make_initial_state(starts_with_BOS));


    LatticeState* initial_state = state_manager_->get_state_buffer(kInitialVertex);
    LatticeState::make_initial_state(starts_with_BOS, initial_state);
    initial_state = state_manager_->add_buffer();

//    lattice_manager_->set_initial_state(&initial_state);
    lattice_manager_->set_initial_state(initial_state);
}

void StateLattice::make_final_ei() {
    //std::cerr << "make final" << std::endl;
    assert(final_ei_ == NULL);
    generic::EdgeInterface& empty_edge = empty_free_edge_;


    const StateManager<LatticeState>::IterableStateContainer& state_list_final = state_manager_.get_states_for_vertex(kFinalVertex);
    StateManager<LatticeState>::IterableStateContainer::const_iterator it_final = state_list_final.begin();
    StateManager<LatticeState>::IterableStateContainer::const_iterator it_final_end = state_list_final.end();
    for (; it_final != it_final_end; ++it_final) {
//    BOOST_FOREACH(const BiLmState& state, state_manager_.get_states_for_vertex(kFinalVertex)) {
        lattice_manager_.add_edge(&(*it_final), state_manager_.final_state(), &empty_edge);
    }

    //std::cerr << "computing kbest for final state " << state_manager_.final_state() << std::endl;
//    lattice_manager_.make_kbest_for_state(state_manager_.final_state());

    final_ei_ = &(lattice_manager_.edge_interface_from_state(state_manager_.final_state()));
}

StateLattice::StateLattice(const BaseLattice& lattice, const StateUpdater& state_updater, StateFeaturesWeight lm_weight, bool starts_with_BOS, t_log::TranslationLog* translation_log)
        : state_manager_(lattice.nb_vertices()), state_updater_(state_updater, lm_weight, &state_manager_, &edges_, &lattice_manager_)
                , original_lattice_topologically_sorted_vertices_(
                lattice.get_topologically_sorted_vertices()), final_ei_(NULL), translation_log_(translation_log), fully_computed_(false) {
    state_updater_.make_initial_state(starts_with_BOS);

//    std::vector<const BiLmState*> final_states;

    assert(Params::Translator::beam_width >= 0);
    unsigned int beam_as_unsigned = static_cast<unsigned int>(Params::Translator::beam_width);

    BOOST_FOREACH(VertexId vertex, original_lattice_topologically_sorted_vertices_) {
        std::vector<StateScorePair> sorted_states;

        const StateManager<LatticeState>::IterableStateContainer& state_list = state_manager_.get_states_for_vertex(vertex);
        StateManager<LatticeState>::IterableStateContainer::const_iterator it = state_list.begin();
        StateManager<LatticeState>::IterableStateContainer::const_iterator it_end = state_list.end();
//        BOOST_FOREACH(const StateManager::IterableStateContainer::value_type& starting_state, state_list) {
        for (; it != it_end; ++it) {
            double best_score_of_state = lattice_manager_.get_best_score_for_state(&(*it));
            assert(best_score_of_state != generic::kBadScore);
            sorted_states.push_back(StateScorePair(&(*it), best_score_of_state));
        }

        std::sort(sorted_states.begin(), sorted_states.end());

        size_t max_nb_of_states = static_cast<int>(sorted_states.size());

        if (max_nb_of_states > beam_as_unsigned) max_nb_of_states = beam_as_unsigned;

//        for (size_t i = 0; i < max_nb_of_states; ++i) {
//            lattice_manager_.make_kbest_for_state(sorted_states[i].state);
//         }

#ifndef NDEBUG
        assert(sorted_states.size() >= max_nb_of_states);
        translation_log_->add_nb_states_discarded(sorted_states.size() - max_nb_of_states);
        translation_log_->add_nb_states_kept(max_nb_of_states);

        for (size_t i = max_nb_of_states; i < sorted_states.size(); ++i) {
            translation_log_->add_nb_lm_edges_discarded(lattice_manager_.get_nb_incoming_edges_for_state(sorted_states[i].state));
        }
        for (size_t i = 0; i < max_nb_of_states; ++i) {
            translation_log_->add_nb_lm_edges_kept(lattice_manager_.get_nb_incoming_edges_for_state(sorted_states[i].state));
        }
#endif //NDEBUG

        if (vertex == kFinalVertex) {
            assert(final_states_.size() == 0);
            final_states_.reserve(max_nb_of_states);
            for (size_t i = 0; i < max_nb_of_states; ++i) {
                final_states_.push_back(sorted_states[i].state);
             }
        }

        BOOST_FOREACH(EdgeId num_edge, lattice.outgoing_edges(vertex)) {
            const PinedEdge& edge = lattice.get_edge(num_edge);
            state_updater_.make_new_state(sorted_states, max_nb_of_states, edge);
        }
    }
    assert(final_states_.size() > 0);

    make_final_history_edges();

    fully_computed_ = true;

#ifndef NDEBUG
    assert(translation_log_ != NULL);
    translation_log_->add_nb_states_created(state_manager_.nb_created_states_);
    translation_log_->add_nb_states_merged(state_manager_.nb_merged_states_);
#endif
}

class PathFacadeBiLm: public PathFacade {
  public:
    explicit PathFacadeBiLm(Path* path)
        : PathFacade(path) {
    }

    void add_state(const generic::StateInterface* state) {
        if (typeid(*state) == typeid(LatticeState)) {
            const LatticeState* lm_state = static_cast<const LatticeState*>(state);
            path_->add_vertex(lm_state->v_id());
        } else {
            PathFacade::add_state(state);
        }
    }
};

struct CompareStatePointers {
    bool operator()(const LatticeState* a, const LatticeState* b) const {
        return *a < *b;
    }
};

void StateLattice::make_final_history_edges() {
    //assert(states_[kFinalVertex].size() > 0); // We should have at least one final state
//    const boost::ptr_vector<BiLmState>& final_states_list = state_manager_.get_final_states_list();

    final_history_edges_pair_list_.reserve(final_states_.size());
    for (int i = 0; i < static_cast<int>(final_states_.size()); ++i) {
        const LatticeState& st = *(final_states_[i]);
        generic::EdgeInterface* edge = &(lattice_manager_.edge_interface_from_state(&st));
        StateEdgeScoreStruct hist_edge(st, edge, edge->get_k_score(0)); // + st.history().eval_delta_with_BOS_EOS(lm_)

        final_history_edges_pair_list_.push_back(hist_edge);
    }

#ifndef NDEBUG
    std::set<const LatticeState*, CompareStatePointers> unicity_test_set;
    BOOST_FOREACH(const StateEdgeScoreStruct& state_edge, final_history_edges_pair_list_) {
        assert(unicity_test_set.count(state_edge.state) == 0);
        unicity_test_set.insert(state_edge.state);
    }
#endif //NDEBUG

    assert(final_history_edges_pair_list_.size() > 0);
}

Path get_kbest_path_to_edge(int k, generic::EdgeInterface* ei) {
    assert(ei != NULL);
    Path result;
    PathFacadeBiLm path_f(&result);
    bool could_find_k_path = ei->set_k_path(k, &path_f);
    if (could_find_k_path) {
        //std::cout << "Path: " << result << std::endl;
        result.check_is_consistent_full_path();
        assert(!result.is_special());
        result.set_score(ei->get_k_score(k));
        return result;
    } else {
        return Path::make_special();
    }
}

Path StateLattice::get_kbest_path(int k) {
    if (final_ei_ == NULL) make_final_ei();
    assert(final_ei_ != NULL);
    return get_kbest_path_to_edge(k, final_ei_);
}

//Path BiLmLattice::get_kbest_path(int k, const History& hist) {
//    generic::EdgeInterface* ei = get_final_edge_interface_for_history(hist);
//    return get_kbest_path_to_edge(k, ei);
//}

std::ostream& operator<<(std::ostream& stream, const StateLattice& lattice) {
    stream << "Begin BiLmLattice " << lattice.edges_.size() << "\n";
    stream << lattice.lattice_manager_;
    stream << "End BiLmLattice\n";
    return stream;
}

StateLatticeComputer::StateLatticeComputer(const BaseLattice& base_lattice, const StateUpdater& state_updater,
        const latt::StateFeaturesWeight& lm_weight, bool starts_with_BOS, t_log::TranslationLog* translation_log) :
        base_lattice_(base_lattice), state_updater_(state_updater), lm_weight_(lm_weight), starts_with_BOS_(starts_with_BOS),  translation_log_(translation_log) {
    //std::cerr << "Creating lattice computer for lattice of nb of states " << base_lattice.nb_vertices() << std::endl;
}

struct StateEdgeScore {
    const StateScorePair* sc_pair;
    const StateEdgeScoreStruct* edge;
    StateEdgeScore(const StateScorePair* sc_pair, const StateEdgeScoreStruct* edge) :
            sc_pair(sc_pair), edge(edge) {}
    bool operator<(const StateEdgeScore& other) const {
          return (other.sc_pair->score + other.edge->score) < (sc_pair->score + edge->score);
    }
};


void StateLatticeComputer::make_all_edges_from_starting_state(
                    const std::vector<StateScorePair>& sorted_starting_states,
                    size_t nb_max_starting_states,
                    VertexId v_end_id,
                    const FreeEdge& free_edge,
                    const StateUpdater& state_updater,
                    const StateFeaturesWeight state_weights_,
                    boost::ptr_vector<generic::EdgeInterface>* edges,
                    StateManager<LatticeState>* state_manager,
                    generic::GenericLatticeManager* lattice_manager) {
    StateLattice* sub_lattice = get_bilm_lattice();

    if (Params::Translator::lattice_beam_pre_truncation > 0) {
        std::vector<StateEdgeScore> state_edges_combinations;
        for (size_t i = 0; i < nb_max_starting_states; ++i) {
            BOOST_FOREACH(const StateEdgeScoreStruct& hist_edge, sub_lattice->get_final_history_edges()) {
                state_edges_combinations.push_back(StateEdgeScore(&sorted_starting_states[i], &hist_edge));
            }
        }
        std::sort(state_edges_combinations.begin(), state_edges_combinations.end());

        assert(Params::Translator::lattice_beam_pre_truncation >= 1);
        assert(Params::Translator::beam_width > 0);
        size_t pre_truncation_limit = static_cast<size_t>(Params::Translator::beam_width * Params::Translator::lattice_beam_pre_truncation);
        for (size_t i = 0; (i < state_edges_combinations.size() && i < pre_truncation_limit); ++i) {
            make_one_edge(*state_edges_combinations[i].sc_pair->state, *state_edges_combinations[i].edge,
                    v_end_id, free_edge, state_updater, state_weights_, edges, state_manager, lattice_manager);
        }
    } else {
        for (size_t i = 0; i < nb_max_starting_states; ++i) {
            const LatticeState& starting_state = *(sorted_starting_states[i].state);
            //assert(edge.v_start() == starting_state.v_id());
            BOOST_FOREACH(const StateEdgeScoreStruct& hist_edge, sub_lattice->get_final_history_edges()) {
                make_one_edge(starting_state, hist_edge, v_end_id, free_edge, state_updater, state_weights_, edges, state_manager, lattice_manager);
            }
        }
    }
}

void StateLatticeComputer::make_one_edge(
                    const LatticeState& starting_state,
                    const StateEdgeScoreStruct& state_edge,
                    VertexId v_end_id,
                    const FreeEdge& free_edge,
                    const StateUpdater& state_updater,
                    const StateFeaturesWeight state_weights_,
                    boost::ptr_vector<generic::EdgeInterface>* edges,
                    StateManager<LatticeState>* state_manager,
                    generic::GenericLatticeManager* lattice_manager) {
    // create new state
//    LatticeState* new_state = new LatticeState(v_end_id);
    LatticeState* new_state = state_manager->get_state_buffer(v_end_id);

    StateFeatures st_feat = state_updater.compute_state_transition(starting_state, *state_edge.state, free_edge, new_state);

    // create new edge
    edges->push_back(new EdgeOffsetAdapter(state_edge.edge, st_feat, state_weights_, free_edge));

    // adding all to lattice_manager
    generic::EdgeInterface* ei_mod = &(edges->back());
//    const generic::StateInterface& state = state_manager->add_or_get_identical_state(v_end_id, new_state);
    const generic::StateInterface* state = state_manager->add_buffer();
//    lattice_manager->add_edge(&starting_state, &state, ei_mod);
    lattice_manager->add_edge(&starting_state, state, ei_mod);
}


StateLattice* StateLatticeComputer::get_bilm_lattice() {
    if (!bilm_lattice_) {
        //std::cerr << "Request computing for lattice of nb of states " << base_lattice_.nb_vertices() << std::endl;
        //std::cerr << "Using computer:  " << *this << std::endl;
        bilm_lattice_.reset(new StateLattice(base_lattice_, state_updater_, lm_weight_, starts_with_BOS_, translation_log_));
        //std::cerr << "Done computing for lattice of nb of states " << base_lattice_.nb_vertices() << std::endl;
        //std::cerr << "size of history: " << bilm_lattice_->get_final_history().size() << std::endl;
    }
    return bilm_lattice_.get();
}

std::ostream& operator<<(std::ostream& stream, const StateLatticeComputer& comp) {
    stream << "LattComp<" << comp.debug_id_ << " | nb_v: " << comp.base_lattice_.nb_vertices() << ">";
    return stream;
}

}  // namespace latt

