/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _STATE_LATTICE_HPP_
#define _STATE_LATTICE_HPP_

//#include <boost/foreach.hpp>

#include <boost/noncopyable.hpp>
//#include <boost/ptr_container/detail/reversible_ptr_container.hpp>
#include <boost/ptr_container/ptr_vector.hpp>
#include <boost/ptr_container/ptr_set.hpp>
#include <boost/smart_ptr/scoped_ptr.hpp>
#include <stddef.h>
#include <cassert>
#include <iostream>
//#include <limits>
//#include <map>
#include <string>
#include <vector>

//#include "../../KenLM/lm/left.hh"
//#include "../../KenLM/lm/state.hh"
//#include "../../KenLM/lm/virtual_interface.hh"
//#include "../../KenLM/lm/word_index.hh"
#include "base_lattice.hpp"
#include "lattice_generic.hpp"
#include "lm_edge.hpp"
#include "translation_log.hpp"

#include "state_storage.hpp"
#include "latt_state.hpp"

namespace latt {

namespace kenlm = ::lm;

//template<typename Model>
//void eval_lm_with_different_techniques(const std::vector<std::string>& sentence, const Model& model);

struct StateScorePair {
    const LatticeState* state;
    double score;
    StateScorePair(const LatticeState* state, double score)
        : state(state), score(score) {
    }
    bool operator<(const StateScorePair& other) const {
        return other.score < score;
    }
};

std::ostream& operator<<(std::ostream& stream, const StateScorePair& obj);

struct StateEdgeScoreStruct {
    const LatticeState* state;
    generic::EdgeInterface* edge;
    double score;

    StateEdgeScoreStruct();

    StateEdgeScoreStruct(const LatticeState& state, generic::EdgeInterface* edge, double score);

    bool operator<(const StateEdgeScoreStruct& other) const;
};

class StateUpdateManagement {
  public:
    StateUpdateManagement(const StateUpdater& lm, StateFeaturesWeight state_weights,
            StateManager<LatticeState>* state_manager,
            boost::ptr_vector<generic::EdgeInterface>* edges,
            generic::GenericLatticeManager* lattice_manager)
                    : state_updater_(lm), state_weights_(state_weights),
                      state_manager_(state_manager), edges_(edges),
                      lattice_manager_(lattice_manager) {}

    void make_new_state(const std::vector<StateScorePair>& sorted_starting_states,
            size_t nb_max_starting_states,
            const PinedEdge& edge);

    void make_initial_state(bool starts_with_BOS);

  private:
    const StateUpdater& state_updater_;
    StateFeaturesWeight state_weights_;

    StateManager<LatticeState>* state_manager_;
    boost::ptr_vector<generic::EdgeInterface>* edges_;
    generic::GenericLatticeManager* lattice_manager_;

//    generic::EdgeInterface& make_new_edge(double lm_cost, const FreeEdge& free_edge);
};

/**************************************************************************
 * StateLattice create a state lattice that decode a given BaseLattice
 */

class StateLattice: boost::noncopyable {
  public:
    friend std::ostream& operator<<(std::ostream& stream, const StateLattice& lattice);
    friend class StateLatticeComputer;

    StateLattice(const BaseLattice& lattice, const StateUpdater& state_updater,
                StateFeaturesWeight lm_weight, bool starts_with_BOS = false, t_log::TranslationLog* translation_log = NULL);


    size_t get_kbest_size() {
        if (final_ei_ == NULL) make_final_ei();
        assert(final_ei_ != NULL);
        return final_ei_->size();
    }

    Path get_kbest_path(int k);

  private:
    StateManager<LatticeState> state_manager_;
    StateUpdateManagement state_updater_;

    std::vector<StateEdgeScoreStruct> final_history_edges_pair_list_;

    boost::ptr_vector<generic::EdgeInterface> edges_;
    generic::GenericLatticeManager lattice_manager_;

    std::vector<const LatticeState*> final_states_;

    std::vector<VertexId> original_lattice_topologically_sorted_vertices_;

    //FinalLmState final_state_;
    generic::EdgeInterface* final_ei_;
    EdgeSimpleConnectors empty_free_edge_;

    t_log::TranslationLog* translation_log_;

    bool fully_computed_;

    void make_final_history_edges();

    void make_final_ei();

    std::vector<StateEdgeScoreStruct>& get_final_history_edges() {
         assert(fully_computed_);
         assert(final_history_edges_pair_list_.size() > 0);
         return final_history_edges_pair_list_;
     }

//    double update_lm_history(const kenlm::ngram::ChartState& history, kenlm::WordIndex word_index,
//            kenlm::ngram::ChartState* new_history);
};

/**************************************************************************
 * BiLmLatticeComputer is in charge of delaying the decoding of a BaseLattice (which happen when the BilmLattice constructor is called)
 */

class StateLatticeComputer: boost::noncopyable {
  public:
    friend std::ostream& operator<<(std::ostream& stream, const StateLatticeComputer& comp);

    StateLatticeComputer(const BaseLattice& base_lattice, const StateUpdater& state_updater,
            const StateFeaturesWeight& lm_weight, bool starts_with_BOS = false, t_log::TranslationLog* translation_log = NULL);

    void set_debug_id(std::string id) {
        debug_id_ = id;
    }

    void make_all_edges_from_starting_state(
                        const std::vector<StateScorePair>& sorted_starting_states,
                        size_t nb_max_starting_states,
                        VertexId v_end_id,
                        const FreeEdge& free_edge,
                        const StateUpdater& state_updater,
                        const StateFeaturesWeight state_weights_,
                        boost::ptr_vector<generic::EdgeInterface>* edges,
                        StateManager<LatticeState>* state_manager,
                        generic::GenericLatticeManager* lattice_manager);

    StateLattice* get_bilm_lattice();

  private:
    const BaseLattice& base_lattice_;
    const StateUpdater& state_updater_;
    StateFeaturesWeight lm_weight_;
    bool starts_with_BOS_;
    std::string debug_id_;
    boost::scoped_ptr<StateLattice> bilm_lattice_;
    t_log::TranslationLog* translation_log_;

    void make_one_edge(
                        const LatticeState& starting_state,
                        const StateEdgeScoreStruct& hist_edge,
                        VertexId v_end_id,
                        const FreeEdge& free_edge,
                        const StateUpdater& state_updater,
                        const StateFeaturesWeight state_weights_,
                        boost::ptr_vector<generic::EdgeInterface>* edges,
                        StateManager<LatticeState>* state_manager,
                        generic::GenericLatticeManager* lattice_manager);
};

} // namespace latt

#endif /* BILM_LATTICE_HPP_ */
