/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef TRANSLATION_LOG_HPP_
#define TRANSLATION_LOG_HPP_

#include <stddef.h>
#include <sys/time.h>
#include <cassert>
#include <iostream>
#include <limits>

namespace t_log {
typedef int TimeUnit;
const TimeUnit kBadTime = std::numeric_limits<int>::min();

inline TimeUnit get_time_in_ms() {
    struct timeval tv;
    gettimeofday(&tv, NULL);
    int ret = tv.tv_usec;
    /* Convert from micro seconds (10^-6) to milliseconds (10^-3) */
    ret /= 1000;
    /* Adds the seconds (10^0) after converting them to milliseconds (10^-3) */
    ret += (tv.tv_sec * 1000);
    return ret;
}

class TimePrinter {
  public:
    friend std::ostream& operator<<(std::ostream& stream, const TimePrinter& time);
    explicit TimePrinter(TimeUnit time): time_(time) {}

  private:
    TimeUnit time_;
};

inline std::ostream& operator<<(std::ostream& stream, const TimePrinter& time) {
    if (time.time_ == kBadTime) {
        stream << "## ";
    } else {
        stream << time.time_ << " ms";
    }
    return stream;
}

const double kBadScore = -std::numeric_limits<double>::max();

#define TLOG_ADD_METHOD(name) void add_##name(unsigned int nb) {\
    name##_ += nb;}\
    void print_##name(std::ostream* stream) const {\
        *stream << #name << ": " << name##_ << " ";}\
    private:\
    unsigned int name##_;\
    public: //NOLINT
//NOLINT

class TranslationLog {
  public:
    friend std::ostream& operator<<(std::ostream& stream, const TranslationLog& log);

    TranslationLog()
        : nb_original_vertice_(0), nb_reduced_vertice_(0), nb_states_created_(0), nb_states_merged_(0),
          nb_states_discarded_(0), nb_states_kept_(0), nb_lm_edges_discarded_(0), nb_lm_edges_kept_(0),
          tr_retrieving_time_(kBadTime), loading_time_(kBadTime), lattice_setup_time_(kBadTime), lattice_building_time_(kBadTime),
          lattice_decoding_time_(kBadTime), nbest_gen_time_(kBadTime), total_decoding_time_(kBadTime), best_model_score_(kBadScore), last_model_score_(kBadScore),
          nb_nbest_produced_(0) {}

//    virtual ~TranslationLog() {}
//    virtual void set_nb_states_created(int) = 0;
//    virtual void set_nb_states_discarded(int) = 0;
//    virtual void set_nb_original_vertice(int) = 0;

    static TranslationLog zero_translation_log() {
        TranslationLog result;
        result.tr_retrieving_time_ = 0;
        result.tr_retrieving_time_ = 0;
        result.loading_time_ = 0;
        result.lattice_setup_time_ = 0;
        result.lattice_building_time_ = 0;
        result.lattice_decoding_time_ = 0;
        result.nbest_gen_time_ = 0;
        result.total_decoding_time_ = 0;

        result.best_model_score_ = 0;
        result.last_model_score_ = 0;

        return result;
    }

    TLOG_ADD_METHOD(nb_original_vertice)
    TLOG_ADD_METHOD(nb_reduced_vertice)

    TLOG_ADD_METHOD(nb_states_created)
    TLOG_ADD_METHOD(nb_states_merged)
    TLOG_ADD_METHOD(nb_states_discarded)
    TLOG_ADD_METHOD(nb_states_kept)
    TLOG_ADD_METHOD(nb_lm_edges_discarded)
    TLOG_ADD_METHOD(nb_lm_edges_kept)

//    void add_nb_states_created(unsigned int nb) {
//        nb_states_created_ += nb;
//    }
//
//    void add_nb_states_merged(unsigned int nb) {
//        nb_states_merged_ += nb;
//    }
//
//    void add_nb_original_vertice(int nb) {
//        nb_original_vertice_ += nb;
//    }
//
//    void add_nb_reduced_vertice(int nb) {
//        nb_reduced_vertice_ += nb;
//    }

    void set_best_model_score(double score) {
        best_model_score_ = score;
    }

    void set_last_model_score(double score) {
        last_model_score_ = score;
    }

    void set_nb_nbest_produced(int nb) {
        nb_nbest_produced_ = nb;
    }

    void set_tr_retrieving_time(TimeUnit start_time) {
        assert(tr_retrieving_time_ == kBadTime);
        tr_retrieving_time_ = get_time_in_ms() - start_time;
    }

    void set_loading_time(TimeUnit start_time) {
        assert(loading_time_ == kBadTime);
        loading_time_ = get_time_in_ms() - start_time;
    }

    void set_lattice_setup_time(TimeUnit start_time) {
        assert(lattice_setup_time_ == kBadTime);
        lattice_setup_time_ = get_time_in_ms() - start_time;
    }

    void set_lattice_building_time(TimeUnit start_time) {
        assert(lattice_building_time_ == kBadTime);
        lattice_building_time_ = get_time_in_ms() - start_time;
    }

    void set_lattice_decoding_time(TimeUnit start_time) {
        assert(lattice_decoding_time_ == kBadTime);
        lattice_decoding_time_ = get_time_in_ms() - start_time;
    }

    void set_nbest_gen_time(TimeUnit start_time) {
        assert(nbest_gen_time_ == kBadTime);
        nbest_gen_time_ = get_time_in_ms() - start_time;
    }

    void set_total_decoding_time(TimeUnit start_time) {
        assert(total_decoding_time_ == kBadTime);
        total_decoding_time_ = get_time_in_ms() - start_time;
    }

    void add_in(const TranslationLog& other) {
        tr_retrieving_time_ += other.tr_retrieving_time_;
        loading_time_ += other.loading_time_;
        lattice_setup_time_ += other.lattice_setup_time_;
        lattice_building_time_ += other.lattice_building_time_;
        lattice_decoding_time_ += other.lattice_decoding_time_;
        nbest_gen_time_ += other.nbest_gen_time_;
        total_decoding_time_ += other.total_decoding_time_;
        best_model_score_ += other.best_model_score_;
        last_model_score_ += other.last_model_score_;
        nb_nbest_produced_ += other.nb_nbest_produced_;
        nb_original_vertice_ += other.nb_original_vertice_;
        nb_reduced_vertice_ += other.nb_reduced_vertice_;
        nb_states_created_ += other.nb_states_created_;
        nb_states_merged_ += other.nb_states_merged_;

        nb_lm_edges_discarded_ += other.nb_lm_edges_discarded_;
        nb_lm_edges_kept_ += other.nb_lm_edges_kept_;

        nb_states_discarded_ += other.nb_states_discarded_;
        nb_states_kept_ += other.nb_states_kept_;
     }

  private:
    TimeUnit tr_retrieving_time_;
    TimeUnit loading_time_;
    TimeUnit lattice_setup_time_;
    TimeUnit lattice_building_time_;
    TimeUnit lattice_decoding_time_;
    TimeUnit nbest_gen_time_;
    TimeUnit total_decoding_time_;

    double best_model_score_;
    double last_model_score_;
    int nb_nbest_produced_;

//    unsigned int nb_original_vertice_;
//    unsigned int nb_reduced_vertice_;
//
//    unsigned int nb_states_created_;
//    unsigned int nb_states_merged_;
//
//    unsigned int nb_states_discarded_;
//    unsigned int nb_states_kept_;
//
//    unsigned int nb_lm_edges_discarded_;
//    unsigned int nb_lm_edges_kept_;
};

inline std::ostream& operator<<(std::ostream& stream, const TranslationLog& log) {
    stream << "Model score: " << log.best_model_score_ << " -- " << log.last_model_score_ << " (" << log.nb_nbest_produced_ << ") // ";
    stream << "Vertice: " << log.nb_original_vertice_ << " -> " << log.nb_reduced_vertice_ << " ";
    stream << "States: " << log.nb_states_created_ << " (+" << log.nb_states_merged_ << ") // ";
    log.print_nb_states_discarded(&stream);
    log.print_nb_states_kept(&stream);
    log.print_nb_lm_edges_discarded(&stream);
    log.print_nb_lm_edges_kept(&stream);
//    stream << log.nb_lm_edges_kept_ << " ";
    stream << "// ";
    stream << "TR: " << TimePrinter(log.tr_retrieving_time_) << " ";
    stream << "Loading hyp: " << TimePrinter(log.loading_time_) << " ";
    stream << "Latt setup: " << TimePrinter(log.lattice_setup_time_) << " ";
    stream << "Latt building: " << TimePrinter(log.lattice_building_time_) << " ";
    stream << "Latt decoding: " << TimePrinter(log.lattice_decoding_time_) << " ";
    stream << "Nbest gen: " << TimePrinter(log.nbest_gen_time_) << " ";
    stream << "Total decoding: " << TimePrinter(log.total_decoding_time_) << " ";
    return stream;
}

} //namespace t_log

#endif /* TRANSLATION_LOG_HPP_ */
