/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "dtree.hpp"
#include <boost/lexical_cast.hpp>
#include <set>
#include <string>
#include <vector>
#include <algorithm>
#include "utils.hpp"
#include "glog/logging.h"

/* ノードの親子関係と木のrootを設定 */
/* set parent-child relations to every node and find the root node */
int Dtree::format_dtree() {
    assert(nodes_.size() > 0);
    // dpnd_ref, pre_children_ref, post_children_refを作る
    int num_of_nodes = nodes_.size();
    bool multiple_root = false;
    int current_root = -1;
    for (int i = 0; i < num_of_nodes; i++) {
        nodes_[i].pre_children_ref.clear();
        nodes_[i].post_children_ref.clear();
        nodes_[i].num = i;
    }
    for (int i = 0; i < num_of_nodes; i++) {
        if (nodes_[i].dpnd == -1) {
            if (current_root != -1) {
                nodes_[current_root].dpnd = i;
                nodes_[current_root].dpnd_ref = &nodes_[i];
                nodes_[i].pre_children_ref.push_back(&(nodes_[current_root]));
                multiple_root = true;
            }
            nodes_[i].dpnd_ref = NULL;
            root_ = &nodes_[i];
            current_root = i;
        } else {
            nodes_[i].dpnd_ref = &(nodes_[nodes_[i].dpnd]);
            if (i < nodes_[i].dpnd) {
                nodes_[nodes_[i].dpnd].pre_children_ref.push_back(&(nodes_[i]));
            } else {
                nodes_[nodes_[i].dpnd].post_children_ref.push_back(&(nodes_[i]));
            }
        }
    }
    if (multiple_root) {
        LOG(WARNING) << "Warning: There are more than one ROOT node whose dpnd are set -1.";
    }
    assert(root_ != NULL);
    mark_tree_depth(root_, 0);
    return 0;
}

/* ノードの深さを記録 */
void Dtree::mark_tree_depth(DtreeNode *node, int depth) {
    node->depth = depth;
    std::vector<DtreeNode*>::const_iterator i;
    for (i = node->pre_children_ref.begin(); i != node->pre_children_ref.end(); i++) {
        mark_tree_depth(*i, depth + 1);
    }
    for (i = node->post_children_ref.begin(); i != node->post_children_ref.end(); i++) {
        mark_tree_depth(*i, depth + 1);
    }
}

/* find the root node among given nodes */
const DtreeNode *Dtree::get_root_p(const std::vector<const DtreeNode*> &list) const {
    if (list.size() == 0) {
        return NULL;
    }
    const DtreeNode *now = list.at(0);
    for (size_t i = 1; i < list.size(); i++) {
        if (now->depth > list.at(i)->depth) {
            now = list.at(i);
        }
    }
    while (1) {
        while (now->dpnd_ref != NULL && grep(const_cast<const DtreeNode*>(now->dpnd_ref), list).size() != 0) {
            now = now->dpnd_ref;
        }
        if (now->num == size() - 1 ||
            now->dpnd != node_at(now->num+1)->dpnd ||
            grep(node_at(now->num+1), list).size() == 0) {
            break;
        }
        now = node_at(now->num+1);
    }
    return now;
}
const WordIndex Dtree::get_root(const std::vector<WordIndex> &list) const {
    WordIndex now = list.at(0);
    for (size_t i = 1; i < list.size(); i++) {
        if (node_at(now)->depth > node_at(list.at(i))->depth) {
            now = list.at(i);
        }
    }
    while (1) {
        while (node_at(now)->dpnd != -1 && grep((WordIndex)node_at(now)->dpnd, list).size() != 0) {
            now = node_at(now)->dpnd;
        }
        if (now == size() - 1 ||
            node_at(now)->dpnd != node_at(now+1)->dpnd ||
            grep(now+1, list).size() == 0) {
            break;
        }
        now++;
    }
    return now;
}

/* 木を描画 */
/* Tree printing functions */
void Dtree::print_dtree(std::ostream *os, const DtreeNode *node_r, const int *bond) const {
    VS buffer;
    std::vector<char> mark;
    print_dtree(node_r, mark, buffer, false, bond);
    int buffer_size = buffer.size();
    for (int s = 0; s < buffer_size; s++) {
        *os << buffer[s] << std::endl;
    }
}
void Dtree::print_dtree(const DtreeNode *node_r,
                        std::vector<char> &mark,
                        VS& output_buffer,
                        bool wnum,
                        const int *bond,
                        bool ast) const {
    std::string local_buffer;
    std::vector<const DtreeNode*> children;

    // 前の子の表示
    // print pre-children
    int pre_children_ref_size = node_r->pre_children_ref.size();
    for (int i = 0; i < pre_children_ref_size; i++) {
        if (bond == NULL ||
            (ast && bond[node_r->pre_children_ref[i]->num] == 0) ||
            (!ast && bond[node_r->pre_children_ref[i]->num] != -1)) {
            children.push_back(node_r->pre_children_ref[i]);
        }
    }
    pre_children_ref_size = children.size();
    for (int i = 0; i < pre_children_ref_size; i++) {
        if (i == 0) {
            std::vector<char> mark_tmp = mark;
            mark_tmp.push_back('L');
            print_dtree(children[i], mark_tmp, output_buffer, wnum, bond, ast);
        } else {
            std::vector<char> mark_tmp = mark;
            mark_tmp.push_back('l');
            print_dtree(children[i], mark_tmp, output_buffer, wnum, bond, ast);
        }
    }

    // 自分の表示
    // print self
    int mark_size = mark.size();
    for (int m = 0; m < mark_size; m++) {
        if (m == mark_size-1) {
            if (mark[m] == 'L') {
                local_buffer += "┌";
            } else if (mark[m] == 'R') {
                local_buffer += "└";
            } else {
                local_buffer += "├";
            }
        } else {
            if (mark[m] == 'l' || mark[m] == 'r' ||
                    (mark[m] == 'L' && (mark[m+1] == 'r' || mark[m+1] == 'R')) ||
                    (mark[m] == 'R' && (mark[m+1] == 'l' || mark[m+1] == 'L'))) {
                local_buffer += "│";
            } else {
                local_buffer += "　";
            }
        }
    }

    // bond info
    if (bond != NULL) {
        if (bond[node_r->num] == 1) {
            local_buffer += "*";
        } else if (bond[node_r->num] == 2) {
            local_buffer += "@";
        } else if (bond[node_r->num] == -1) { // wildcard
            local_buffer += "#";
        }
    }
    
    // alignment info
    if (node_r->align_index != -1) {
        local_buffer += "[" + 
            boost::lexical_cast<std::string>(node_r->align_index) + "] ";
    }
    
    local_buffer += node_r->word.content;

    if (wnum) {
        size_t p_num = output_buffer.size();
        if (p_num < 10) {
            output_buffer.push_back(boost::lexical_cast<std::string>(p_num) + 
                                    "  " + local_buffer);
        } else {
            output_buffer.push_back(boost::lexical_cast<std::string>(p_num) + 
                                    " " + local_buffer);
        }
    } else {
            output_buffer.push_back(local_buffer);
    }

    // 後ろの子の表示
    // print post-children
    children.clear();
    int post_children_ref_size = node_r->post_children_ref.size();
    for (int i = 0; i < post_children_ref_size; i++) {
        if (bond == NULL ||
            (ast && bond[node_r->post_children_ref[i]->num] == 0) ||
            (!ast && bond[node_r->post_children_ref[i]->num] != -1)) {
            children.push_back(node_r->post_children_ref[i]);
        }
    }
    post_children_ref_size = children.size();
    for (int i = 0; i < post_children_ref_size; i++) {
        if (i == post_children_ref_size-1) {
            std::vector<char> mark_tmp = mark;
            mark_tmp.push_back('R');
            print_dtree(children[i], mark_tmp, output_buffer, wnum, bond, ast);
        } else {
            std::vector<char> mark_tmp = mark;
            mark_tmp.push_back('r');
            print_dtree(children[i], mark_tmp, output_buffer, wnum, bond, ast);
        }
    }
}

void Dtree::print(std::ostream *os) const {
    VS buffer;
    std::vector<char> mark;
    print_dtree(root_, mark, buffer, true);
    for (VS::iterator it = buffer.begin(); it != buffer.end(); ++it) {
        *os << *it << std::endl;
    }
}

std::string Dtree::sentence() const {
    std::string sent = "";
    size_t num_of_nodes = nodes_.size();
    for (size_t i = 0; i < num_of_nodes; i++) {
        sent += nodes_[i].word.content;
        if (i != num_of_nodes - 1) {
            sent += " ";
        }
    }
    return sent;
}

void Dtree::get_neighbors(const std::vector<const DtreeNode*> &node_list, 
        std::set<const DtreeNode*> *neighbors, bool parent, bool children, 
        bool siblings) const {
    for (size_t i = 0; i < node_list.size(); i++) {
        if (parent) {
            get_parent(node_list[i], neighbors);
        }
        if (children) {
            get_children(node_list[i], neighbors);
        }
        if (siblings) {
            get_siblings(node_list[i], neighbors);
        }
    }
    for (size_t i = 0; i < node_list.size(); i++) {
        neighbors->erase(node_list[i]);
    }
}

void Dtree::get_neighbors(const std::vector<WordIndex> &node_list, 
        std::set<const DtreeNode*> *neighbors, bool parent, bool children, 
        bool siblings) const {
    for (size_t i = 0; i < node_list.size(); i++) {
        if (parent) {
            get_parent(node_at(node_list[i]), neighbors);
        }
        if (children) {
            get_children(node_at(node_list[i]), neighbors);
        }
        if (siblings) {
            get_siblings(node_at(node_list[i]), neighbors);
        }
    }
    for (size_t i = 0; i < node_list.size(); i++) {
        neighbors->erase(node_at(node_list[i]));
    }
}

void Dtree::get_neighbors(const DtreeNode *node, std::set<const DtreeNode*> *neighbors,
                          bool parent, bool children, bool siblings) const {
    if (parent) {
        get_parent(node, neighbors);
    }
    if (children) {
        get_children(node, neighbors);
    }
    if (siblings) {
        get_siblings(node, neighbors);
    }
    neighbors->erase(node);
}

void Dtree::get_parent(const DtreeNode *node, std::set<const DtreeNode*> *parent) const {
    if (node->dpnd_ref == NULL) {
        return;
    }
    parent->insert(node->dpnd_ref);
}

WordIndex Dtree::get_parent_index(WordIndex child) const {
    for (size_t i = 0; i < nodes_.size(); i++) {
        if (nodes_[i].num == child) {
            if (nodes_[i].dpnd_ref == NULL) {
                return -1;
            } else {
                return nodes_[i].dpnd_ref->num;
            }
        }
    }
    return -1;
}

void Dtree::get_children(const DtreeNode *node, std::set<const DtreeNode*> *children) const {
    for (size_t i = 0; i < node->pre_children_ref.size(); i++) {
        if (node->pre_children_ref[i] != NULL) {
            children->insert(node->pre_children_ref[i]);
        }
    }
    for (size_t i = 0; i < node->post_children_ref.size(); i++) {
        if (node->post_children_ref[i] != NULL) {
            children->insert(node->post_children_ref[i]);
        }
    }
}

std::vector<WordIndex> Dtree::get_children_as_wordindex(WordIndex word_index) const {
    std::vector<WordIndex> result;
    assert(word_index != (WordIndex) -1 && word_index < size());
    const DtreeNode &node = nodes_[word_index];
    for (size_t i = 0; i < node.pre_children_ref.size(); i++) {
        if (node.pre_children_ref[i] != NULL) {
            result.push_back(node.pre_children_ref[i]->num);
        }
    }
    for (size_t i = 0; i < node.post_children_ref.size(); i++) {
        if (node.post_children_ref[i] != NULL) {
            result.push_back(node.post_children_ref[i]->num);
        }
    }
    return result;
}

void Dtree::get_siblings(const DtreeNode *node, std::set<const DtreeNode*> *siblings) const {
    DtreeNode *parent = node->dpnd_ref;
    if (parent == NULL) {
        return;
    }
    for (size_t i = 0; i < parent->pre_children_ref.size(); i++) {
        if ((parent->pre_children_ref[i]->num == node->num - 1 || 
             parent->pre_children_ref[i]->num == node->num + 1) &&
            parent->pre_children_ref[i] != node && 
            parent->pre_children_ref[i] != NULL) {
            siblings->insert(parent->pre_children_ref[i]);
        }
    }
    for (size_t i = 0; i < parent->post_children_ref.size(); i++) {
        if ((parent->post_children_ref[i]->num == node->num - 1 || 
             parent->post_children_ref[i]->num == node->num + 1) &&
            parent->post_children_ref[i] != node &&
            parent->post_children_ref[i] != NULL) {
            siblings->insert(parent->post_children_ref[i]);
        }
    }
}

/*
 * obsolete
bool Dtree::is_tree(const std::vector<const DtreeNode*> &nodes) const {
    std::set<const DtreeNode*> nodes_reached;
    std::set<const DtreeNode*> nodes_todo;
    nodes_todo.insert(*(nodes.begin()));
    while (!nodes_todo.empty()) {
        const DtreeNode *node = *(nodes_todo.begin());
        nodes_reached.insert(node);
        get_neighbors(node, &nodes_todo, true, true, true);
        for (std::set<const DtreeNode*>::iterator it = nodes_reached.begin(); 
             it != nodes_reached.end(); it++) {
            nodes_todo.erase(*it);
        }
    }
    for (size_t i = 0; i < nodes.size(); i++) {
        if (nodes_reached.find(nodes[i]) == nodes_reached.end()) {
            return false;
        }
    }
    return true;
}
*/

std::ostream &operator<<(std::ostream &os, Dtree &t) {
    t.print(&os);
    return os;
}

