/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*! @file
  This file contains the definition of Dtree class
*/
#ifndef DTREE_INCLUDED
#define DTREE_INCLUDED

#include <boost/serialization/vector.hpp>
#include <boost/utility.hpp>
#include <iostream>
#include <set>
#include <string>
#include <vector>
#include "def.hpp"
#include "utils.hpp"
#include "word.hpp"

/*! This class contains information of a dependency tree. The member
  fields are nodes_ for a list of pointers to DtreeNodes and root_ for a
  pointer to the root node. 
  It is made noncopyable by inheriting from boost::noncopyable. This means that 
  if one tries to copy or assign an object of this class, the compiler will complain
  with a message containing something like: 
  "Error: noncopyable(const noncopyable&) constructor is private" (+ a lot of template instanciation mess)  
*/
class Dtree : boost::noncopyable {
    // TMBuilder class is implemented in TR++/EBMTppConversionCode.hpp
    friend class TMBuilder;
  public:
    //! default constructor
    Dtree() : root_(NULL) {}

    /*! constructor with the number of nodes specified.
      This constructor allocates memory for nodes.
      \param[in] nb_nodes the number of nodes to be allocated
    */
    explicit Dtree(int nb_nodes) : nodes_(nb_nodes), root_(NULL) {
        for (int i = 0; i < nb_nodes; i++) {
            nodes_.push_back(DtreeNode());
            nodes_.back().num = i;
        }
    };

    // ~Dtree() {
        // std::cerr << "Destroying dtree at " << (size_t) this << std::endl;
    // }
    
    // accessors
    //! returns a pointer to a modifiable node at i
    //! @param i index of a node
    DtreeNode *get_modifiable_node_at(int i) {
        return &(nodes_[i]);
    }
    //! returns a const pointer to a node at i
    //! @param i index of a node
    const DtreeNode *node_at(WordIndex i) const {return &(nodes_[i]);}
    //! returns a const reference to the list of nodes
    const std::vector<DtreeNode> &nodes() const {return nodes_;}
    //! returns a const reference to the root of dtree
    const DtreeNode *root() const {return root_;}
    //! set the root node
    //! @param reference the new root node
    void set_root(DtreeNode *root) {root_ = root;}
    //! returns the number of nodes in the dtree
    const size_t size() const {return nodes_.size();}

    //! add one node
    void push_node(const DtreeNode &node) {nodes_.push_back(node);}

    //! print dependency tree as vector of string
    void print_dtree(std::ostream *os, const DtreeNode *node_r, const int *bond) const;
    void print_dtree(const DtreeNode *node_r,
                     std::vector<char> &mark,
                     VS& output_buffer,
                     bool wnum = false,
                     const int *bond = NULL,
                     bool ast = false) const;
    //! print dependency tree 
    void print(std::ostream *os) const;
    //! returns sentence as string
    std::string sentence() const;

    //! formatting Dtree. this function sets pre/post children and
    //! parent to each node and find the root node of the Dtree
    int format_dtree();
    //! set the depth of each node from the root
    void mark_tree_depth(DtreeNode *node, Depth depth);

    //! returns the root node among the list of nodes passed (returns a pointer)
    const DtreeNode *get_root_p(const std::vector<const DtreeNode*> &nodes) const;
    //! returns the root node among the list of nodes passed (returns WordIndex)
    const WordIndex get_root(const std::vector<WordIndex> &words) const;

    //! find the neighboring nodes (parent, children and siblings)
    //! the input is a list of node pointers
    void get_neighbors(const std::vector<const DtreeNode*> &node_list,
                       std::set<const DtreeNode*> *neighbors, bool parent, bool children, bool siblings) const;
    //! find the neighboring nodes (parent, children and siblings)
    //! the input is a node pointer
    void get_neighbors(const DtreeNode *node,
                       std::set<const DtreeNode*> *neighbors, bool parent, bool children, bool siblings) const;
    //! find the neighboring nodes (parent, children and siblings)
    //! the input is a list of WordIndex
    void get_neighbors(const std::vector<WordIndex> &word_list,
                       std::set<const DtreeNode*> *neighbors, bool parent, bool children, bool siblings) const;

    //! find the parent node and set to &parent
    void get_parent(const DtreeNode *node, std::set<const DtreeNode*> *parent) const;
    //! return the WordIndex of the parent node
    WordIndex get_parent_index(WordIndex child) const;

    //! find child nodes and set to &children
    void get_children(const DtreeNode *node, std::set<const DtreeNode*> *children) const;
    //! return a list of child node index
    std::vector<WordIndex> get_children_as_wordindex(WordIndex word_index) const;

    //! find sibling nodes and set to &siblings
    void get_siblings(const DtreeNode *node, std::set<const DtreeNode*> *siblings) const;

    //! check if the specified nodes form a contiguous tree
    bool is_tree(const std::vector<const DtreeNode*> &nodes) const;

    //! small function for sanity check
    //! (used in the construction of dtree by the TR code)
    bool is_default_constructed() const {
        return nodes_.size() == 0 && root_ == NULL;
    }
     
    //! Used for in-place creation of OOV pseudo-TM
    void set_to_size_1_for_OOV() {
        assert(is_default_constructed());
        nodes_.push_back(DtreeNode());
    }
    
 private:
    std::vector<DtreeNode> nodes_;   //!< list of nodes
    DtreeNode *root_;   //!< pointer to the root node
};
std::ostream &operator<<(std::ostream &os, Dtree &t);

#endif
