/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "hypothesis.hpp"
#include <boost/regex.hpp>
#include <algorithm>
#include <limits>
#include <map>
#include <set>
#include <string>
#include <utility>
#include <vector>
#include "params.hpp"

Hypothesis::Hypothesis(const Hypothesis&         hypothesis_outside,
                       const Hypothesis&         hypothesis_inside,
                       WordIndex                 insertion_point,
                       bool                      replace_ref,
                       int                       index_of_used_additional) :
        score_(hypothesis_outside.score_ + hypothesis_inside.score_),  
        score_no_lm_(hypothesis_outside.score_no_lm_ + hypothesis_inside.score_no_lm_),
        //feature_(hypothesis_outside.feature_, hypothesis_inside.feature_),
        t_string_(hypothesis_outside.t_string_,
                  hypothesis_inside.t_string_,
                  insertion_point, replace_ref),
        additionals_(hypothesis_outside.additionals_,
                     hypothesis_inside.additionals_,
                     insertion_point, hypothesis_inside.t_string_.size(), 
                     replace_ref, index_of_used_additional),
        parent_bond_relation_(hypothesis_outside.parent_bond_relation_),
        outside_parent_hypothesis_(&hypothesis_outside),
        inside_parent_hypothesis_(&hypothesis_inside),
        tm_(NULL), data_for_tm_(NULL), 
        starts_with_BOS_(hypothesis_outside.starts_with_BOS_), 
        ends_with_EOS_(hypothesis_outside.ends_with_EOS_),
        //lm_state_has_been_evaluated_(false),
        count_(std::numeric_limits<int>::min()),
        real_size_(hypothesis_inside.real_size_ + hypothesis_outside.real_size_) {
// Calculate dependency information from inside/outside parents.
    init_t_dpnd(hypothesis_outside, hypothesis_inside, insertion_point, replace_ref, index_of_used_additional);
    
    if (Params::Global::keep_tms_in_memory) {
        // Initialize initial hypotheses information (which initial hypothesis did
        // each terminal come from?).
        init_initial_hypotheses(hypothesis_outside, hypothesis_inside, insertion_point, replace_ref);
    }
    
    // Set root POS tag as root POS tag of outside parent.
    root_pos_ = hypothesis_outside.get_root_pos();
    // Copy bond POS map from outside parent, removing bond used to 
    // attach inside parent for generating this hypothesis.
    init_bond_pos(hypothesis_outside.get_bond_pos_map(), 
            insertion_point, replace_ref, hypothesis_inside.get_t_dpnd().size(),
            index_of_used_additional);
            
    assert(check_consistency());
}

inline double max(double a, double b) {
    return a > b ? a : b;
}

void Hypothesis::print(std::ostream *os) const {
    *os << "Hypothesis: ";
    if (tm_ != NULL) *os << tm_->id() << " ";
    *os << "< ";
    // for (VS::iterator it = t_string_.begin(); it != t_string_.end(); ++it)
    // *os << *it << " ";
    *os << t_string_;
    *os << "> ";
    for (std::vector<Additional>::const_iterator it = additionals_.begin(); it != additionals_.end(); ++it) {
        *os << "< ";
        *os << (*it).cnum() << " : ";
        for (std::vector<WordIndex>::const_iterator it2 = it->prelist().begin(); it2 != it->prelist().end(); ++it2)
            *os << *it2 << " ";
        *os << "/ ";
        for (std::vector<WordIndex>::const_iterator it2 = it->postlist().begin(); it2 != it->postlist().end(); ++it2)
            *os << *it2 << " ";
        *os << "> ";
    }
    *os << score_ << " : " << parent_bond_relation_;
}

std::ostream &operator<<(std::ostream &stream, const Hypothesis &hyp) {
    //if (!stream) return stream; //shortcut
    
    stream << "H<";
    if (hyp.outside_parent_hypothesis_ == NULL) {
        assert(hyp.inside_parent_hypothesis_ == NULL);
        stream << "initial|";
    } else {
        if (hyp.outside_parent_hypothesis_ == NULL) {
            assert(hyp.inside_parent_hypothesis_ == NULL);
            stream << "initial|";
        } else {
            stream << "combined|";
        }
    }
    stream << hyp.score_ << "|" << hyp.score_no_lm_ << "|";
    //if (hyp.tm_ != NULL) {
    //    // TODO(fabien/nakazawa): Removed logging of id for now, as it can cause crash when TMs have been previously removed from memory). Maybe enable it condiationally based on Params:: fields?
    //    //stream << hyp.tm_->id() << "|"; 
    // } else {
    //    stream << "combined|";
    // }
    // stream << hyp.score_ << "|";
    BOOST_FOREACH(const std::string& word, hyp.t_string_) {
        stream << word << " ";
    }
    stream << "|";
    BOOST_FOREACH(WordIndex idx, hyp.t_dpnd_) {
        stream << static_cast<int>(idx) << " ";
    }
    stream << "|P:" << hyp.parent_bond_relation_ << "|";
    stream << hyp.additionals_ << "|";
    stream << hyp.root_pos_ << "|";
    for (std::map<WordIndex, POS>::const_iterator it = hyp.bond_pos_map_.begin();
            it != hyp.bond_pos_map_.end(); ++it) {
        stream << static_cast<int>(it->first) << ":" << it->second << " ";
    }
    if (hyp.starts_with_BOS_ || hyp.ends_with_EOS_) stream << "|";
    if (hyp.starts_with_BOS_) stream << "BOS ";
    if (hyp.ends_with_EOS_) stream << "EOS ";
    stream << "|";
    stream << hyp.feature_;
//    if (hyp.lm_state_has_been_evaluated_) {
//        stream << " | " << hyp.lm_state_ << " lm_p: " << hyp.lm_proba_;
//    }
    stream << ">";
    return stream;
}

std::string Hypothesis::real_str() {
    VS str;
    const boost::regex reg2("\\[X[\\d\\,]+\\]");
    for (int i = static_cast<int>(t_string_.size()) - 1; i >= 0; i--) {
        if (t_string_[i].size() == 1 ||
            (t_string_[i].find("*") != t_string_[i].size()-1 &&
             !boost::regex_match(t_string_[i], reg2)))
            str.push_back(t_string_[i]);
    }
    sort(str.begin(), str.end());
    return join(";", str);
}

std::string Hypothesis::lm_str() {
    VS str;
    for (int i = static_cast<int>(t_string_.size()) - 1; i >= 0; i--) {
        if (t_string_[i].size() != 1 && t_string_[i].find("*") == t_string_[i].size() - 1) {
            str.push_back(t_string_[i]);
        }
    }
    sort(str.begin(), str.end());
    return join(";", str);
}

FeatureVector Hypothesis::get_consolidated_features() const {
    if (inside_parent_hypothesis_ == NULL) {
        assert(outside_parent_hypothesis_ == NULL);
        return feature_;
    }
    assert(inside_parent_hypothesis_ != NULL && outside_parent_hypothesis_ != NULL);
    FeatureVector result;
    std::vector<const Hypothesis*> stack;
    result.merge_in(feature_);
    stack.push_back(inside_parent_hypothesis_);
    stack.push_back(outside_parent_hypothesis_);
    while (stack.size() != 0) {
        const Hypothesis *parent = stack.back();
        assert(parent != NULL);
        stack.pop_back();
        result.merge_in(parent->feature_);
        if (parent->inside_parent_hypothesis_ == NULL) {
            assert(parent->outside_parent_hypothesis_ == NULL);
            // std::cout << parent->get_feature_value("size") << ",";
            // parent->print(&std::cout);
            // std::cout << std::endl;
        } else {
            assert(parent->inside_parent_hypothesis_ != NULL && parent->outside_parent_hypothesis_ != NULL);
            stack.push_back(parent->inside_parent_hypothesis_);
            stack.push_back(parent->outside_parent_hypothesis_);
        }
    }
    // std::cout << std::endl;
    return result;
}

std::vector<const Hypothesis*> Hypothesis::get_ancestor_hypotheses() const {
    std::vector<const Hypothesis*> result;
    result.push_back(this);
    if (inside_parent_hypothesis_ == NULL) {
        assert(outside_parent_hypothesis_ == NULL);
        return result;
    }
    assert(inside_parent_hypothesis_ != NULL && outside_parent_hypothesis_ != NULL);
    std::vector<const Hypothesis*> stack;
    stack.push_back(inside_parent_hypothesis_);
    stack.push_back(outside_parent_hypothesis_);
    while (stack.size() != 0) {
        const Hypothesis *parent = stack.back();
        assert(parent != NULL);
        stack.pop_back();
        result.push_back(parent);
        if (parent->inside_parent_hypothesis_ == NULL) {
            assert(parent->outside_parent_hypothesis_ == NULL);
        } else {
            assert(parent->inside_parent_hypothesis_ != NULL && parent->outside_parent_hypothesis_ != NULL);
            stack.push_back(parent->inside_parent_hypothesis_);
            stack.push_back(parent->outside_parent_hypothesis_);
        }
    }
    return result;
}

// --- obsolete ---
std::vector<const Hypothesis*> Hypothesis::get_list_of_initial_hypothesis_old() const {
    std::vector<const Hypothesis*> result;
    
    if (inside_parent_hypothesis_ == NULL) {
        assert(outside_parent_hypothesis_ == NULL);
        result.push_back(this);
    } else {  
        assert(inside_parent_hypothesis_ != NULL && outside_parent_hypothesis_ != NULL);
        std::vector<const Hypothesis*> stack;
        stack.push_back(inside_parent_hypothesis_);
        stack.push_back(outside_parent_hypothesis_);
        while (stack.size() != 0) {
            const Hypothesis *parent = stack.back();
            assert(parent != NULL);
            stack.pop_back();
            if (parent->inside_parent_hypothesis_ == NULL) {
                assert(parent->outside_parent_hypothesis_ == NULL);
                result.push_back(parent);
            } else {
                assert(parent->inside_parent_hypothesis_ != NULL && parent->outside_parent_hypothesis_ != NULL);
                stack.push_back(parent->inside_parent_hypothesis_);
                stack.push_back(parent->outside_parent_hypothesis_);
            }
        }
    }
    // std::cout << std::endl;
    return result;
}

void Hypothesis::get_used_tm(std::vector<std::pair<const TM*, const DataForTM*> > *result) const {
    if (inside_parent_hypothesis_ == NULL) {
        assert(outside_parent_hypothesis_ == NULL);
        if (data_for_tm_ != NULL) {
            assert(tm_ != NULL);
            result->push_back(std::make_pair(tm_, data_for_tm_));
        }
        return;
    }
    assert(inside_parent_hypothesis_ != NULL &&
           outside_parent_hypothesis_ != NULL &&
           tm_ == NULL);
    std::vector<const Hypothesis*> stack;
    stack.push_back(inside_parent_hypothesis_);
    stack.push_back(outside_parent_hypothesis_);
    while (stack.size() != 0) {
        const Hypothesis *parent = stack.back();
        assert(parent != NULL);
        stack.pop_back();
        if (parent->inside_parent_hypothesis_ == NULL) {
            assert(parent->outside_parent_hypothesis_ == NULL);
            if (parent->data_for_tm_ != NULL) {
                assert(parent->tm_ != NULL);
                result->push_back(std::make_pair(parent->tm_, parent->data_for_tm_));
            }
        } else {
            assert(parent->inside_parent_hypothesis_ != NULL && parent->outside_parent_hypothesis_ != NULL && parent->tm_ == NULL);
            stack.push_back(parent->inside_parent_hypothesis_);
            stack.push_back(parent->outside_parent_hypothesis_);
        }
    }
}

void Hypothesis::hash(std::string *result) const {
    t_string_.hash(result); 
    *result += " || ";
    for (size_t i = 0; i < additionals_.size(); i++) {
        additionals_[i].hash(result);
        *result += " | ";
    }
}

bool Hypothesis::compare_for_duplicate(const Hypothesis& other) const {
    int t_string_compare = t_string_.compare3(other.t_string_);
    if (t_string_compare == -1) return true;  // optimizable
    if (t_string_compare == 1) return false;
    int add_compare = additionals_.compare3(other.additionals_);
    if (add_compare == -1) return true;
    if (add_compare == 1) return false;
    return false;  // identical
}

bool Hypothesis::compare_for_strict_equality(const Hypothesis& other) const {
    int t_string_compare = t_string_.compare3(other.t_string_);
    if (t_string_compare == -1) return true;  // optimizable
    if (t_string_compare == 1) return false;
    int add_compare = additionals_.compare3(other.additionals_);
    if (add_compare == -1) return true;
    if (add_compare == 1) return false;
    // TODO(fabien): 3-way comparison like above would be more efficient:
    if (t_dpnd_ < other.t_dpnd_) return true;
    if (t_dpnd_ > other.t_dpnd_) return false;
    if (parent_bond_relation_ < other.parent_bond_relation_) return true;
    if (parent_bond_relation_ > other.parent_bond_relation_) return false;
    if (feature_ < other.feature_) return true;
    if (feature_ > other.feature_) return false;
    return false;  // identical
}

bool Hypothesis::is_duplicate(const Hypothesis *other) const {
    if (score_ != other->score_) return false;
    int t_string_compare = t_string_.compare3(other->t_string_);
    if (t_string_compare == -1) return false;  // optimizable
    if (t_string_compare == 1) return false;
    if (t_dpnd_ != other->t_dpnd_) return false;
    if (parent_bond_relation_ != other->parent_bond_relation_) return false;
    if (!additionals_.is_duplicate(other->additionals_)) return false;
    if (root_pos_ != other->root_pos_) return false;
    for (std::map<WordIndex, POS>::const_iterator it = bond_pos_map_.begin();
            it != bond_pos_map_.end(); ++it) {
        if (other->bond_pos_map_.count(it->first) == 0 || it->second != other->bond_pos_map_.at(it->first)) {
            return false;
        }
    }
    for (std::map<std::string, double>::const_iterator it = feature_.begin();
            it != feature_.end(); ++it) {
        if (other->feature_.count(it->first) == 0 || it->second != other->feature_.at(it->first)) {
            return false;
        }
    }
    return true;
}

void Hypothesis::init_t_dpnd(const Hypothesis &hypothesis_outside, 
        const Hypothesis &hypothesis_inside, WordIndex insertion_point, bool replace_ref,
        int index_of_used_additional) {
    const DpndList &out_dpnd = hypothesis_outside.get_t_dpnd();
    const DpndList &in_dpnd = hypothesis_inside.get_t_dpnd();
    if (replace_ref) {
        // Replacement of [X*].
        for (size_t i = 0; i < insertion_point; i++) {
            Dpnd dpnd = out_dpnd[i];
            if (dpnd != -1 && dpnd > static_cast<int>(insertion_point)) {
                t_dpnd_.push_back(dpnd + in_dpnd.size() - 1);
            } else {
                t_dpnd_.push_back(dpnd);
            }
        }
        for (size_t i = 0; i < in_dpnd.size(); i++) {
            Dpnd dpnd = in_dpnd[i];
            if (dpnd == -1) {
                if (out_dpnd[insertion_point] > static_cast<int>(insertion_point)) {
                    t_dpnd_.push_back(out_dpnd[insertion_point] + in_dpnd.size() - 1);
                } else {
                    t_dpnd_.push_back(out_dpnd[insertion_point]);
                }
            } else {
                t_dpnd_.push_back(dpnd + insertion_point);
            }
        }
        for (size_t i = insertion_point + 1; i < out_dpnd.size(); i++) {
            Dpnd dpnd = out_dpnd[i];
            if (dpnd != -1 && dpnd > static_cast<int>(insertion_point)) {
                t_dpnd_.push_back(dpnd + in_dpnd.size() - 1);
            } else {
                t_dpnd_.push_back(dpnd);
            }
        }
    } else {
        // Additional insertion.
        for (size_t i = 0; i < insertion_point; i++) {
            Dpnd dpnd = out_dpnd[i];
            if (dpnd != -1 && dpnd >= static_cast<int>(insertion_point)) {
                t_dpnd_.push_back(dpnd + in_dpnd.size());
            } else {
                t_dpnd_.push_back(dpnd);
            }
        }
        for (size_t i = 0; i < in_dpnd.size(); i++) {
            Dpnd dpnd = in_dpnd[i];
            if (dpnd == -1) {
                Dpnd additional_dpnd = hypothesis_outside.additional(index_of_used_additional).dpnd();
                if (additional_dpnd >= static_cast<int>(insertion_point)) {
                    t_dpnd_.push_back(additional_dpnd + in_dpnd.size());
                } else {
                    t_dpnd_.push_back(additional_dpnd);
                }
            } else {
                t_dpnd_.push_back(dpnd + insertion_point);
            }
        }
        for (size_t i = insertion_point; i < out_dpnd.size(); i++) {
            if (replace_ref && i == insertion_point) {
                continue;
            }
            Dpnd dpnd = out_dpnd[i];
            if (dpnd != -1 && dpnd >= static_cast<int>(insertion_point)) {
                t_dpnd_.push_back(dpnd + in_dpnd.size());
            } else {
                t_dpnd_.push_back(dpnd);
            }
        }
    }
}

void Hypothesis::init_bond_pos(const std::map<WordIndex, POS> &parent_bond_pos_map,
        WordIndex insertion_point, bool replace_ref, int inserted_size,
        int index_of_used_additional) {
    if (replace_ref) {
        // X-bond combination.
        for (std::map<WordIndex, POS>::const_iterator it = parent_bond_pos_map.begin();
                it != parent_bond_pos_map.end(); ++it) {
            if (it->first < insertion_point) {
                bond_pos_map_[it->first] = it->second;
            } else if (it->first == insertion_point) {
                // Do not copy replaced bond.
            } else {
                bond_pos_map_[it->first + inserted_size - 1] = it->second;
            }
        }
    } else {
        // Additional combination.
        for (std::map<WordIndex, POS>::const_iterator it = parent_bond_pos_map.begin();
                it != parent_bond_pos_map.end(); ++it) {
            if (it->first < insertion_point) {
                bond_pos_map_[it->first] = it->second;
            } else {
                bond_pos_map_[it->first + inserted_size] = it->second;
            }
        }
    }
}

void Hypothesis::set_initial_hypotheses_as_self() {
    for (size_t i = 0; i < t_dpnd_.size(); i++) {
        initial_hypotheses_.push_back(this);
    }
}

void Hypothesis::init_initial_hypotheses(const Hypothesis &hypothesis_outside, 
        const Hypothesis &hypothesis_inside, WordIndex insertion_point, bool replace_ref) {
    const InitialHypothesisList &out_init_hyp = hypothesis_outside.get_initial_hypotheses();
    const InitialHypothesisList &in_init_hyp = hypothesis_inside.get_initial_hypotheses();
    for (size_t i = 0; i < insertion_point; i++) {
        initial_hypotheses_.push_back(out_init_hyp[i]);
    }
    for (size_t i = 0; i < in_init_hyp.size(); i++) {
        initial_hypotheses_.push_back(in_init_hyp[i]);
    }
    if (replace_ref) {
        for (size_t i = insertion_point + 1; i < out_init_hyp.size(); i++) {
            initial_hypotheses_.push_back(out_init_hyp[i]);
        }
    } else {
        for (size_t i = insertion_point; i < out_init_hyp.size(); i++) {
            initial_hypotheses_.push_back(out_init_hyp[i]);
        }
    }
}

void Hypothesis::set_count(int count) {
     count_ = count;
     //score_ += count;
}

