/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef PARTIAL_HYPOTHESIS_INCLUDED
#define PARTIAL_HYPOTHESIS_INCLUDED

#include <boost/shared_ptr.hpp>
#include <boost/ptr_container/ptr_vector.hpp>
#include <algorithm>
#include <cassert>
#include <map>
#include <string>
#include <utility>
#include <vector>
#include <set>
#include <limits>
#include "glog/logging.h"
#include "lm/state.hh"
#include "def.hpp"
#include "tm.hpp"
#include "additionals.hpp"
#include "features.hpp"
#include "params.hpp"
#include "partial_translation_string.hpp"
#include "score_computer.hpp"

typedef PartialTranslationStringFactory PartialTranslationFactory;
typedef PartialTranslationString PartialTranslation;

class Hypothesis;
typedef std::vector<const Hypothesis*> InitialHypothesisList;

class Hypothesis {
    friend class HumanReadableHypothesisSerialization;
    friend std::ostream &operator<<(std::ostream &stream,
            const Hypothesis &hyp);
    friend class HypothesisTStringCompare;

  public:
    //! default constructor
    Hypothesis() :
        score_(0.0), score_no_lm_(0.0), //score_averaged_features_total_(0.0),
        nb_combinations_to_generate_(0), parent_bond_relation_(UNK),
        outside_parent_hypothesis_(NULL), inside_parent_hypothesis_(NULL),
        tm_(NULL), data_for_tm_(NULL), starts_with_BOS_(false),
        ends_with_EOS_(false), has_wildcard_(false),
        count_(std::numeric_limits<int>::min()), real_size_(0) {
        // default feature values
        set_feature_value("example_penalty", 1.0);
        set_feature_value("NULL_content", 0.0);
        set_feature_value("NULL_function", 0.0);
        set_feature_value("abnormal_child_bond", 0.0);
        set_feature_value("both_ROS", 0.0);
        set_feature_value("child_bond_count", 0.0);
        set_feature_value("child_bond_similarity", 0.0);
        set_feature_value("different_parent_cat", 0.0);
        set_feature_value("no_redundant_child", 0.0);
        set_feature_value("numeral_mismatch", 0.0);
        set_feature_value("one_side_ROS", 0.0);
        set_feature_value("oov", 0.0);
        set_feature_value("parent_bond_available", 0.0);
        set_feature_value("parent_bond_LD", 0.0);
        set_feature_value("parent_bond_similarity", 0.0);
        set_feature_value("redundant_child_in_input", 0.0);
        set_feature_value("redundant_child_in_source", 0.0);
        set_feature_value("root_LD", 0.0);
        set_feature_value("same_parent_cat", 0.0);
        set_feature_value("same_root_content", 0.0);
        set_feature_value("trans_lm_prob", 0.0);
        set_feature_value("trans_prob", 0.0);
        set_feature_value("lex_trans_prob_s_t", -8.0);
        set_feature_value("lex_trans_prob_t_s", -8.0);
    }

    //! constructor merging two hypotheses
    Hypothesis(const Hypothesis&                hypothesis_outside,
                      const Hypothesis&         hypothesis_inside,
                      WordIndex                 insertion_point,
                      bool                      replace_ref,
                      int                       index_of_used_additional = -1);


//    //! constructor merging two hypotheses with update of lm_score
//    Hypothesis(ScoreComputer* score_comp, const Hypothesis& hypothesis_outside,
//            const Hypothesis& hypothesis_inside, WordIndex insertion_point,
//            bool replace_ref, int index_of_used_additional = -1);

    //! constructor with score, ts_string , additionals and p_b_rel

//    Hypothesis(double score, const PartialTranslation& t_string,
//            const AdditionalsList& additionals,
//            const ParentBondRelation& parent_bond_relation) :
//            score_(score), score_no_lm_(score), t_string_(t_string), additionals_(
//                    additionals), parent_bond_relation_(parent_bond_relation), outside_parent_hypothesis_(
//                    NULL), inside_parent_hypothesis_(NULL), tm_(NULL), data_for_tm_(
//                    NULL), starts_with_BOS_(false), ends_with_EOS_(false), has_wildcard_(
//                    false), count_(std::numeric_limits<int>::min()) {
//#ifdef ANALYZE
//        Analysis::hypotheses_generated_total++;
//        Analysis::tstring_length_total += t_string_.size();
//        Analysis::nb_additionals_total += additionals_.size();
//        if (additionals_.size() != 0) {
//            Analysis::nb_hypotheses_with_additionals++;
//        }
//#endif
//    }
//
//    //! constructor with score and additionals and p_b_rel
//    Hypothesis(double score, const ParentBondRelation& parent_bond_relation) :
//            score_(score), score_no_lm_(score), parent_bond_relation_(
//                    parent_bond_relation), outside_parent_hypothesis_(NULL), inside_parent_hypothesis_(
//                    NULL), tm_(NULL), data_for_tm_(NULL), starts_with_BOS_(
//                    false), ends_with_EOS_(false), has_wildcard_(false), count_(std::numeric_limits<int>::min()) {
//#ifdef ANALYZE
//        Analysis::hypotheses_generated_total++;
//        Analysis::tstring_length_total += t_string_.size();
//        Analysis::nb_additionals_total += additionals_.size();
//        if (additionals_.size() != 0) {
//            Analysis::nb_hypotheses_with_additionals++;
//        }
//#endif
//    }

    ~Hypothesis() {
        if (data_for_tm_ != NULL) {
            DataForTM::decref_and_maybe_delete(const_cast<DataForTM*>(data_for_tm_));
            //delete data_for_tm_;
            data_for_tm_ = NULL;
        }
    }

    bool check_consistency() {
        if (!t_string_.check_no_duplicate_ref())
            return false;
        if (!additionals_.check_consistency_with_length(t_string_.size()))
            return false;
        return true;
    }

//    void set_initial_scores(ScoreComputer* score_comp);

    void set_score(const double& s) {
        score_ = s;
    }

    void set_additional(const std::vector<Additional>& add) {
        additionals_ = add;
    }

    void set_additional(const AdditionalsList& add) {
        additionals_ = add;
    }

    void set_t_string(const VS &t_string,
            const PartialTranslationFactory& partial_translation_factory) {
        t_string_ = partial_translation_factory.make_partial_translation(
                t_string);
    }

    DpndList *get_t_dpnd_ptr() {
        return &t_dpnd_;
    }

    const DpndList &get_t_dpnd() const {
        return t_dpnd_;
    }

//    const lm::ngram::ChartState& lm_state() const {
//        assert(lm_state_has_been_evaluated_);
//        return lm_state_;
//    }

    void set_parent_bond_relation(const ParentBondRelation& pbr) {
        parent_bond_relation_ = pbr;
    }
    void set_feature_value(const std::string &fn, const double val) {
        if (Params::Features::all_features.count(fn) == 0) {
            LOG(FATAL) << "Error: trying to set feature '" << fn << "' not defined in weight file.";
        }
        feature_[fn] = val;
    }
    void set_tm(const TM* tm) {
        tm_ = tm;
    }
    void set_data_for_tm(const DataForTM* data) {
        data_for_tm_ = data;
    }

    double get_feature_value(const std::string &fn) const {
        FeatureVector::const_iterator it = feature_.find(fn);
        if (it != feature_.end()) {
            return feature_.at(fn);
        } else {
            assert(false);
            return 0.0;
        }
    }

    const FeatureVector& feature() const {
        return feature_;
    }
    double score() const {
        return score_;
    }
    double score_no_lm() const {
        return score_no_lm_;
    }

    /// Compute the Global FeatureVector by moving up to the initial ancestors hypothesis
    FeatureVector get_consolidated_features() const;

    std::vector<const Hypothesis*> get_ancestor_hypotheses() const;

    // --- obsolete, as now can use initial_hypotheses_ ---
    // Compute the list of initial hypothesis that led to the creation of this hypothesis
    std::vector<const Hypothesis*> get_list_of_initial_hypothesis_old() const;

    // possibly a bit costly function

    VS get_t_string_as_VS() const {
        return t_string_.to_VS();
    }

    const PartialTranslation& get_partial_translation() const {
        return t_string_;
    }
    const PartialTranslation& t_string() const {
        return t_string_;
    }

    const Additional& additional(int i) const {
        return additionals_[i];
    }

    const AdditionalsList& additionals() const {
        return additionals_;
    }
    std::vector<Additional>& get_modifiable_additionals() {
        return additionals_;
    }

    int n_additionals() const {
        return additionals_.size();
    }
    ParentBondRelation parent_bond_relation() const {
        return parent_bond_relation_;
    }

    void print(std::ostream *os) const;
    std::string real_str();
    std::string lm_str();
    void hash(std::string *result) const;

    /// Compare using t_string and additionals
    bool compare_for_duplicate(const Hypothesis& other) const;


    /// Compare using all relevant attributes (t_string, additionals, dpnd, parent_bond_rel, features)
    bool compare_for_strict_equality(const Hypothesis& other) const;

    /// Is this hypothesis a duplicate of *other?
    bool is_duplicate(const Hypothesis *other) const;
    
    /// Get the set of source positions that can be merged into this hypothesis
    std::set<WordIndex> get_full_set_of_ref() const {
        std::set<WordIndex> result = t_string_.get_set_of_ref();
        BOOST_FOREACH(const Additional& add, additionals_) {
            result.insert(add.cnum());
        }
        return result;
    }

    int get_nb_full_set_of_ref() const {
        // TODO(John) we don't need to actually get the references
        std::set<WordIndex> result = t_string_.get_set_of_ref();
        BOOST_FOREACH(const Additional& add, additionals_) {
            result.insert(add.cnum());
        }
        return result.size();
    }

    bool is_translating_complete_subtree() const {
        return ((!t_string_.is_incomplete()) && additionals_.size() == 0);
    }

    bool has_optional_words() const {
        return t_string_.has_optional_words();
    }

    int get_nb_additionals() const {
        return additionals_.size();
    }

    int target_size() const {
        return t_string_.get_target_nb_of_words();
    }

    int get_nb_of_optional_words() const {
        return t_string_.get_nb_of_optional_words();
    }

    void get_used_tm(std::vector<std::pair<const TM*, const DataForTM*> > *result) const;

    const POS &get_root_pos() const {
        return root_pos_;
    }
    void set_root_pos(const POS &pos) {
        root_pos_ = pos;
    }
    const POS &get_bond_pos(WordIndex index) const {
        assert(bond_pos_map_.count(index) != 0);
        return bond_pos_map_.at(index);
    }
    const std::map<WordIndex, POS> &get_bond_pos_map() const {
        return bond_pos_map_;
    }
    void set_bond_pos_map(const std::map<WordIndex, POS> &bond_pos_map) {
        bond_pos_map_ = bond_pos_map;
    }

    void set_initial_hypotheses_as_self();
    const InitialHypothesisList &get_initial_hypotheses() const {
        return initial_hypotheses_;
    }

    const TM* tm() const {
        return tm_;
    }
    const DataForTM* data_for_tm() const {
        return data_for_tm_;
    }

//    double estimate_best_score(ScoreComputer *score_comp) const;
    int get_nb_combinations_to_generate() const {
        return nb_combinations_to_generate_;
    }

    bool remove_add_if_exists_with_cnum(WordIndex idx) {
        return additionals_.remove_add_if_exists_with_cnum(idx);
    }

    void set_has_wildcard(bool flag) {
        has_wildcard_ = flag;
    }

    bool has_wildcard() const {
        return has_wildcard_;
    }

//    void set_lm_state(const lm::ngram::ChartState& lm_state, double lm_score,
//            int nb_oov) {
//        lm_state_has_been_evaluated_ = true;
//        lm_state_ = lm_state;
//        lm_proba_ = lm_score;
//        nb_oov_ = nb_oov;
//        //score_ = score_no_lm_ + lm_proba_;
//    }

//    double lm_proba() const {
//        return lm_proba_;
//    }

//    void update_score_with_current_lm(ScoreComputer* score_computer, int size_pattern);
//
//    void update_score_with_oracle(ScoreComputer* score_computer, int pattern_size);

    void uppercase_1st_word() {
        if (t_string_.size() > 0 && t_string_[0].size() > 0) {
            t_string_[0][0] = toupper(t_string_[0][0]);
        } else {
            DLOG(WARNING) << "empty tstring/ empty word";
        }
    }

    Hypothesis* duplicate_with_new_t_string(const VS& new_t_string,
            const DpndList& new_t_dpnd) const {
        assert(additionals_.size() == 0);
        assert(get_nb_full_set_of_ref() == 0);

        Hypothesis* result = new Hypothesis();
        result->score_ = score_;
        result->score_no_lm_ = score_no_lm_;
//        result->score_averaged_features_total_ = score_averaged_features_total_;
        result->nb_combinations_to_generate_ = nb_combinations_to_generate_;
        result->feature_ = feature_;

        result->t_string_ = new_t_string;
        result->t_dpnd_ = new_t_dpnd;

        result->additionals_ = additionals_;
        result->initial_hypotheses_ = initial_hypotheses_;
        result->parent_bond_relation_ = parent_bond_relation_;
        result->outside_parent_hypothesis_ = outside_parent_hypothesis_;
        result->inside_parent_hypothesis_ = inside_parent_hypothesis_;
        result->tm_ = tm_;

        if (data_for_tm_ == NULL) {
            result->data_for_tm_ = NULL;
        } else {
            result->data_for_tm_ = const_cast<DataForTM*>(data_for_tm_)->duplicate_with_incref();
        }

        result->root_pos_ = root_pos_;
        result->bond_pos_map_ = bond_pos_map_;
        result->starts_with_BOS_ = starts_with_BOS_;
        result->ends_with_EOS_ = ends_with_EOS_;
        result->has_wildcard_ = has_wildcard_;

        return result;
    }

    bool starts_with_BOS() const {
        return starts_with_BOS_;
    }

    bool ends_with_EOS() const {
        return ends_with_EOS_;
    }

    void set_BOS_EOS() {
        starts_with_BOS_ = true;
        ends_with_EOS_ = true;
    }

    void set_count(int count);

    size_t real_size() const {
        return real_size_;
    }

    void set_real_size(size_t s) {
        real_size_ = s;
    }

  private:
    double score_;
    double score_no_lm_;
//    double score_averaged_features_total_;
    int nb_combinations_to_generate_;
    FeatureVector feature_;
    PartialTranslation t_string_;
    DpndList t_dpnd_;
    AdditionalsList additionals_;
    InitialHypothesisList initial_hypotheses_;
    ParentBondRelation parent_bond_relation_;
    const Hypothesis* outside_parent_hypothesis_; //!< pointer to the hypothesis which imports the other hypothesis
    const Hypothesis* inside_parent_hypothesis_; //!< pointer to the hypothesis which is imported to the other hypothesis
    const TM* tm_; //!< pointer to the original TM (NULL if it isn't an initial Hypothesis)
    const DataForTM* data_for_tm_; //!< pointer to the additional data for TM (NULL if it isn't an initial Hypothesis)
    POS root_pos_;
    std::map<WordIndex, POS> bond_pos_map_;
    bool starts_with_BOS_;
    bool ends_with_EOS_;
    bool has_wildcard_;
    int count_; //nb of time this hypothesis appears in a stack

    size_t real_size_; // nb of non-reference words in the t_string

//    bool lm_state_has_been_evaluated_;
//    lm::ngram::ChartState lm_state_;
//    double lm_proba_;
//    int nb_oov_;


    void init_t_dpnd(const Hypothesis &hypothesis_outside,
            const Hypothesis &hypothesis_inside, WordIndex insertion_point,
            bool replace_ref, int index_of_used_additional);
    void init_bond_pos(const std::map<WordIndex, POS> &parent_bond_pos_map,
            WordIndex insertion_point, bool replace_ref, int insertion_size,
            int index_of_used_additional);
    void init_initial_hypotheses(const Hypothesis &hypothesis_outside,
            const Hypothesis &hypothesis_inside, WordIndex insertion_point,
            bool replace_ref);

//    HypothesisSequence make_hypothesis_sequence() const;

//    VS make_VS_with_gap_for_lm_estimate() const;
};

class HypothesisPointerCompareForDuplicate {
  public:
    bool operator()(Hypothesis* a, Hypothesis* b) const {
        //std::cerr<<"comparing "<<(size_t) a<<" with "<<(size_t) b<<" --> "<<a->compare_for_duplicate(*b)<<std::endl;
        return a->compare_for_duplicate(*b);
    }
};

class HypothesisStrictEqualityCompare {
  public:
    bool operator()(Hypothesis* a, Hypothesis* b) const {
        //std::cerr<<"comparing "<<(size_t) a<<" with "<<(size_t) b<<" --> "<<a->compare_for_duplicate(*b)<<std::endl;
        return a->compare_for_strict_equality(*b);
    }
};

class HypothesisTStringCompare {
  public:
    bool operator()(Hypothesis* a, Hypothesis* b) const {
        //std::cerr<<"comparing "<<(size_t) a<<" with "<<(size_t) b<<" --> "<<a->compare_for_duplicate(*b)<<std::endl;
        return a->t_string_.compare3(b->t_string_) == -1;
    }
};

struct HypothesisPointerScoreCompare {
     bool operator()(const Hypothesis *rLeft, const Hypothesis *rRight) {
        return rLeft->score() > rRight->score();
    }
};

inline void sort_hypothesis_pointers_vector_by_score(std::vector<Hypothesis*>* vect) {
    HypothesisPointerScoreCompare compare;
        std::sort(vect->begin(), vect->end(), compare);
}

#endif
