/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include <limits>
#include <map>
#include <string>
#include <vector>

#include "hypothesis_serialization.hpp"
#include "serialization_helpers.hpp"
#include "utils.hpp"

/*******************************************
 *  Serialization
 */
 
void HumanReadableHypothesisSerialization::serialize_stack(const std::vector<Hypothesis*>& stack, const std::map<std::string, int>& map_feature_name_to_int, std::ostream& stream) {
    // std::vector<Hypothesis*> stack_no_dup;
    // std::map<Hypothesis*, int, HypothesisStrictEqualityCompare> hyp_to_count_map;
    // BOOST_FOREACH(Hypothesis* hyp, stack) {
        // if (hyp_to_count_map.count(hyp) == 0) {
            // stack_no_dup.push_back(hyp);
        // }
        // hyp_to_count_map[hyp]++;
    // }    
    // BOOST_FOREACH(Hypothesis* hyp, stack_no_dup) {
    BOOST_FOREACH(Hypothesis* hyp, stack) {
        serialize_hypothesis_to_stream(*hyp, map_feature_name_to_int, stream);
    }
}

void HumanReadableHypothesisSerialization::serialize_hypothesis_to_stream(
            const Hypothesis& hyp, const std::map<std::string, int>& map_feature_name_to_int, std::ostream& stream) {
    stream << hyp.score_<< "|";
    
    size_t num_pos = 0;
    BOOST_FOREACH(const std::string& str, hyp.t_string_) {
        std::string escaped;
        escape_string(str, &escaped);
        stream << escaped;
        if (num_pos + 1 < hyp.t_string_.size()) stream << " ";
        num_pos++;
    } 
    stream << "|";
    
    serialize_dpnd_vector(hyp.t_dpnd_, ' ', stream);
    stream << "|";
    
    switch (hyp.parent_bond_relation_) {
        case PRE:
        stream<< "E";
        break;
        case POST:
        stream<< "O";
        break;        
        case UNK:
        stream<< "U";
        break;
        default:
        LOG(FATAL) << "unexpected parent_bond_relation\n";
        break;
    }
    stream << "|";
    
    size_t cursor = 0;
    BOOST_FOREACH(const Additional& add, hyp.additionals_) {
        serialize_additional(add, stream);
        if (cursor + 1 < hyp.additionals_.size()) stream << " ";
        cursor++;
    }
    stream << "|";
  
    std::string pos_escaped;
    escape_string(hyp.root_pos_, &pos_escaped); 
    stream << pos_escaped << "|";
    
    bool first_feature_outputed = true;
    for (std::map<WordIndex, POS>::const_iterator it = hyp.bond_pos_map_.begin();
            it != hyp.bond_pos_map_.end(); ++it) {
        if (!first_feature_outputed) {
            stream << " ";
        } else {
            first_feature_outputed = false;
        }
        std::string pos_escaped;
        escape_string(it->second, &pos_escaped); 
        stream << it->first << ":" << pos_escaped;
    }
    stream << "|";

    first_feature_outputed = true;
    BOOST_FOREACH(const FeatureVector::value_type& name_val_pair, hyp.feature_) {
        if (name_val_pair.second != 0) {
            if (!first_feature_outputed) {
                stream << " ";
            } else {
                first_feature_outputed = false;
            }
            std::map<std::string, int>::const_iterator it = map_feature_name_to_int.find(name_val_pair.first);
            LOG_IF(FATAL, it == map_feature_name_to_int.end()) << "Trying to serialize a feature not in the weight file: " << name_val_pair.first;
            stream << it->second << ":" << name_val_pair.second;
        }
    }
    
    stream << "\n";
}

void HumanReadableHypothesisSerialization::serialize_additional(const Additional& add, std::ostream& stream) {
    stream << add.cnum_ << ":";
    stream << add.dpnd_ << ":";
    serialize_wordindex_vector(add.prelist_, '_', stream);
    stream << ":";
    serialize_wordindex_vector(add.postlist_, '_', stream);
    stream << ":";
    std::string pos_escaped;
    escape_string(add.pos_, &pos_escaped);
    stream << pos_escaped; 
}

void HumanReadableHypothesisSerialization::serialize_translation_candidate(const InitialHypothesesStorage& tc, const WeightVector& weights, 
                                                            const std::vector<AugmentedPattern>& augmented_patterns, size_t input_size, std::ostream& stream) {
    const InitialHypothesesStorage::HypothesesMapType& translation_candidate_ = tc.hypotheses_map_;
    
    stream //<< std::scientific
           //<< std::fixed
           << std::setprecision(std::numeric_limits<double>::digits10 + 2);
    
    std::map<PatternType, AugmentedPattern> pattern_to_augmented_map;
    BOOST_FOREACH(AugmentedPattern a_pat, augmented_patterns) {
        pattern_to_augmented_map[a_pat] = a_pat;
    }
    
    stream << "START " << input_size << "\n";
    
    std::map<std::string, int> map_feature_name_to_int;
    size_t num_feature = 0;
    BOOST_FOREACH(const WeightVector::value_type& name_val_pair, weights) {
        map_feature_name_to_int[name_val_pair.first] = num_feature;
        stream << name_val_pair.first;
        if (num_feature +1 < weights.size()) {
            stream << " ";
        }
        num_feature++;
    }
    stream << "\n";
    
    BOOST_FOREACH(const InitialHypothesesStorage::HypothesesMapType::value_type& pattern_stack_pair, translation_candidate_) {
        // stream << "P| ";
        // BOOST_FOREACH(WordIndex pos, pattern_stack_pair.first) {
            // stream << pos <<" ";
        // }
        assert(pattern_to_augmented_map.count(pattern_stack_pair.first)>0);
        AugmentedPattern a_pattern = pattern_to_augmented_map[pattern_stack_pair.first];
        a_pattern.serialize(&stream);
        //stream << ": " << pattern_stack_pair.second.candidate_.size();
        serialize_stack(pattern_stack_pair.second, map_feature_name_to_int, stream);
    }
    stream << "END\n";
}

/*******************************************
 *  DE-Serialization FASTER
 */

void read_dpnd_vector(char* string, char delimiter, DpndList *result) {
    DVLOG(3) << "read vect from '" << string << "'.\n";
    if (*string == '\0') return; //empty vector
    char* rest = NULL;
    char* token = my_strtok_r(string, delimiter, &rest);
    result->push_back(atoi(token));
    while ((token = my_strtok_r(NULL, delimiter, &rest))) {
        result->push_back(atoi(token));   
    }
}

void deserialize_t_string(char* string, VS* t_string) {
    if (*string == '\0') return; //empty t_string
    char* rest = NULL;
    char* token = my_strtok_r(string, ' ', &rest);
    //if (token == NULL) return;
    t_string->push_back(std::string());
    //t_string.back().reserve(rest-token); TODO(fabien): check if this optimization would work/ be efficient
    de_escape_cstring(token, &t_string->back());
    while ((token = my_strtok_r(NULL, ' ', &rest))) {
        t_string->push_back(std::string());
        //t_string.back().reserve(rest-token);
        de_escape_cstring(token, &t_string->back());
    }
}

void deserialize_bond_pos_pair(char* string, std::map<WordIndex, POS> *bond_pos_map) {
    std::string delme = string;
    DVLOG(3) << "deserializing bond POS pair:'" << string << "'" << std::endl;
    char* rest = NULL;
    
    char* token = my_strtok_r(string, ':', &rest);
    LOG_IF(FATAL, token == NULL) << "Parsing error in bond POS pair\n";
    WordIndex index = static_cast<WordIndex>(atoi(token));

    token = my_strtok_r(NULL, ':', &rest);
    LOG_IF(FATAL, token == NULL) << "Parsing error in bond POS pair\n";
    std::string pos_de_escaped;
    de_escape_string(token, &pos_de_escaped);
    POS pos = pos_de_escaped;

    LOG_IF(FATAL, rest != NULL) << "Parsing error: more than one colon in bond POS pair '" << delme << "'\n";

    (*bond_pos_map)[index] = pos;
}

void deserialize_bond_pos_map(char* string, std::map<WordIndex, POS> *bond_pos_map) {
    DVLOG(3) << "deserializing bond POS list:'" << string << "'" << std::endl;
    if (*string == '\0') return; //empty features
    char* rest = NULL;
    char* token = my_strtok_r(string, ' ', &rest);
    //if (token == NULL) return;
    deserialize_bond_pos_pair(token, bond_pos_map);
    while ((token = my_strtok_r(NULL, ' ', &rest))) {
        deserialize_bond_pos_pair(token, bond_pos_map);
    }
}

void deserialize_feature_pair(char* string, const std::vector<std::string>& feature_names, FeatureVector* features) {
    DVLOG(3) << "deserializing feature pair:'" << string << "'" << std::endl;
    char* rest = NULL;
    
    char* token = my_strtok_r(string, ':', &rest);
    LOG_IF(FATAL, token == NULL) << "Parsing error in feature pair\n";
    int feature_ref = atoi(token);
    
    token = my_strtok_r(NULL, ':', &rest);
    LOG_IF(FATAL, token == NULL) << "Parsing error in feature pair\n";
    double value = atof(token);
    LOG_IF(FATAL, rest != NULL) << "Parsing error: more than one colon in feature pair\n";

    LOG_IF(FATAL, feature_ref < 0 || feature_ref > static_cast<int>(feature_names.size())) << "Parsing error: feature ref not consistent\n";
    (*features)[feature_names[feature_ref]] = value;
}

void deserialize_features(char* string, const std::vector<std::string>& feature_names, FeatureVector* features) {
    DVLOG(3) << "deserializing feature list:'" << string << "'" << std::endl;
    if (*string == '\0') return; //empty features
    char* rest = NULL;
    char* token = my_strtok_r(string, ' ', &rest);
    //if (token == NULL) return;
    deserialize_feature_pair(token, feature_names, features);
    while ((token = my_strtok_r(NULL, ' ', &rest))) {
        deserialize_feature_pair(token, feature_names, features);
    }
}

Hypothesis* HumanReadableHypothesisSerialization::deserialize_hypothesis_from_string(char* hyp_as_string, 
            const std::vector<std::string>& feature_names, const FeatureVector& empty_feature_vector) {
    VLOG(2) << "deserializing hypothesis:" << hyp_as_string << std::endl;
    Hypothesis* hyp = new Hypothesis();
    
    char* rest = NULL;
    
    char* token = my_strtok_r(hyp_as_string, '|', &rest);
    LOG_IF(FATAL, token == NULL) << "Error Parsing score" << "\n";
    hyp->score_ = atof(token);

    token = my_strtok_r(NULL, '|', &rest);
    LOG_IF(FATAL, token == NULL) << "Error Parsing t_string" << "\n";
    deserialize_t_string(token, &hyp->t_string_);
    
    token = my_strtok_r(NULL, '|', &rest);
    LOG_IF(FATAL, token == NULL) << "Error Parsing dpnd" << "\n";
    read_dpnd_vector(token, ' ', &hyp->t_dpnd_);
    
    token = my_strtok_r(NULL, '|', &rest);
    LOG_IF(FATAL, token == NULL) << "Error Parsing parent_bond_relation" << "\n";
    LOG_IF(FATAL, token[1] != '\0') << "Error Parsing parent_bond_relation" << "\n";
    switch (token[0]) {
        case 'E':
        hyp->parent_bond_relation_ = PRE;
        break;
        case 'O':
        hyp->parent_bond_relation_ = POST;
        break;        
        case 'U':
        hyp->parent_bond_relation_ = UNK;
        break;
        default:
        LOG(FATAL) << "Parsing Error: unexpected parent_bond_relation\n";
        break;
    }

    token = my_strtok_r(NULL, '|', &rest);
    LOG_IF(FATAL, token == NULL) << "Error Parsing additionals" << "\n";
    deserialize_additional_list(token, &hyp->additionals_);

    std::string root_pos_de_escaped;
    de_escape_string(read_one_string_delimited_by(NULL, '|', &rest), &root_pos_de_escaped);
    hyp->root_pos_ = root_pos_de_escaped;

    token = my_strtok_r(NULL, '|', &rest);
    LOG_IF(FATAL, token == NULL) << "Error Parsing bond POS map" << "\n";
    deserialize_bond_pos_map(token, &hyp->bond_pos_map_);

    //hyp->feature_ = empty_feature_vector;
    token = my_strtok_r(NULL, '|', &rest);
    LOG_IF(FATAL, token == NULL) << "Error Parsing features" << "\n";
    deserialize_features(token, feature_names, &hyp->feature_);

    LOG_IF(FATAL, rest != NULL) << "Error Parsing hypothesis: too many fields" << "\n";
    
    return hyp;
}

void HumanReadableHypothesisSerialization::deserialize_additional(char* add_as_string, Additional* add) {
    char *rest = NULL;
    
    add->cnum_ = static_cast<WordIndex>(read_one_int_delimited_by(add_as_string, ':', &rest));
    add->dpnd_ = read_one_int_delimited_by(NULL, ':', &rest);
    
    char* token_prelist = my_strtok_r(NULL, ':', &rest);
    LOG_IF(FATAL, token_prelist == NULL) << "Error Parsing additional" << "\n";
    read_wordindex_vector(token_prelist, '_', &add->prelist_);
    
    char* token_postlist = my_strtok_r(NULL, ':', &rest);
    LOG_IF(FATAL, token_postlist == NULL) << "Error Parsing additional" << "\n";
    read_wordindex_vector(token_postlist, '_', &add->postlist_);

    std::string pos_de_escaped;
    de_escape_string(read_one_string_delimited_by(NULL, ':', &rest), &pos_de_escaped);
    add->pos_ = pos_de_escaped;
}

void HumanReadableHypothesisSerialization::deserialize_additional_list(char* string, AdditionalsList* add_list) {
    DVLOG(3) << "deserializing add list:'" << string << "'" << std::endl;
    if (*string == '\0') return; //empty additionals
    char* rest = NULL;
    char* token = my_strtok_r(string, ' ', &rest);
    //if (*token == '\0') return; // empty additionals
    add_list->push_back(Additional());
    deserialize_additional(token, &add_list->back());
    while ((token = my_strtok_r(NULL, ' ', &rest))) {
        add_list->push_back(Additional());
        deserialize_additional(token, &add_list->back());
    }
}


int deserialize_first_line(char *first_line) {
    char *rest = NULL;
    
    char* token = my_strtok_r(first_line, ' ', &rest);
    LOG_IF(FATAL, token == NULL) << " Hypothesis Parsing Error on first line"  << "\n";
    LOG_IF(FATAL, strcmp(token, "START") != 0) << " Hypothesis Parsing Error on first line"  << "\n";
    
    token = my_strtok_r(NULL, ' ', &rest);
    LOG_IF(FATAL, token == NULL) << " Hypothesis Parsing Error on first line"  << "\n";
    int input_size = atoi(token);
    
    LOG_IF(FATAL, rest != NULL) << " Hypothesis Parsing Error on first line"  << "\n";
    
    return input_size;
}

void deserialize_second_line(char *line, std::vector<std::string>* feature_names, FeatureVector* empty_feature_vector) {
    if (*line == '\0') return; //empty line -> no features
    char *rest = NULL;
    char* token = my_strtok_r(line, ' ', &rest);
    //if (*token == '\0') return; //empty line -> no features
    std::string name(token);
    feature_names->push_back(std::string(token));
    (*empty_feature_vector)[name] = 0;
    while ((token = my_strtok_r(NULL, ' ', &rest))) {
        std::string name(token);
        feature_names->push_back(std::string(token));
        (*empty_feature_vector)[name] = 0;
    }
}

void HumanReadableHypothesisSerialization::deserialize_translation_candidate(char* serialized_tc, 
                InitialHypothesesStorage* tc, std::vector<AugmentedPattern>* augmented_pattern_vect, size_t* input_size) {
    char *rest = NULL; 

    char *first_line = my_strtok_r(serialized_tc, '\n', &rest);
    LOG_IF(FATAL, first_line == NULL) << " Hypothesis Parsing Error on first line"  << "\n";
    *input_size = deserialize_first_line(first_line);
    
    std::vector<std::string> feature_names;
    FeatureVector empty_feature_vector;
    char *second_line = my_strtok_r(NULL, '\n', &rest);
    LOG_IF(FATAL, second_line == NULL) << " Hypothesis Parsing Error on first line"  << "\n";
    deserialize_second_line(second_line, &feature_names, &empty_feature_vector);
    
    InitialHypothesesStorage::HypothesesList* current_stack = NULL;
    bool end_line_met = false;
    
    while (char* line = my_strtok_r(NULL, '\n', &rest)) {
        if (line[0] == 'P') {
            PatternType pattern = AugmentedPattern::deserialize(line, augmented_pattern_vect);
            LOG_IF(FATAL, tc->hypotheses_map_.count(pattern) != 0) << "Parsing Error: pattern found several times\n";
            current_stack = & (tc->hypotheses_map_[pattern]);
        } else if (line[0] == 'E') {
            LOG_IF(FATAL, line[1] != 'N' || line[2] != 'D') << "Parsing Error: unrecognized line\n";
            end_line_met = true;
        } else if (line[0] == '\0') { // empty line: skip
        } else {
            LOG_IF(FATAL, current_stack == NULL) << "Parsing Error: no pattern was defined\n";
            current_stack->push_back(deserialize_hypothesis_from_string(line, feature_names, empty_feature_vector) );
        }
    }
    LOG_IF(FATAL, !end_line_met) << "Parsing Error: Finished without meeting END\n";
    // // loop till my_strtok_r returns NULL.
    // while (token = my_strtok_r(ptr, '\n', &rest)) {
            // printf("%s\n", token); // print the token returned.
            // ptr = rest; // rest contains the left over part..assign it to ptr...and start tokenizing again.    
    // }
}






