/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include <vector>
#include <map>
#include <string>
#include "hypotheses_storage.hpp"
#include "decoding/augmented_pattern.hpp"

#ifndef __HYPOTHESIS_SERIALIZATION_HPP__
#define __HYPOTHESIS_SERIALIZATION_HPP__

/*! Defines a serialization/deserialization method for the set of hypothesis generated after the first step of a translation.
    The serialization format is designed to be human-readable but reasonably efficient.
    The general outline of a generated file is:
    START "size_of_input
    list of feature names
    Pattern1
    <list_of_hypothesis>
    Pattern2
    <list_of_hypothesis>
    ...
    END\n
    
    A Pattern is represented by a line:
    P|pattern|root|parent|outer_frontier|inner_frontier
    pattern and outer_frontier are space-separated integers. root and parent are integers. 
    inner_frontier is a sequence of string formatted as a:b_c_d where a is the word index of an inner frontier position and b, c and d are outer_frontier nodes children of a.
    
    An hypothesis is represented by one line:
    score|t_string|dependency|parent_bond_relationship|additionals_list|features_values
    t_string is the space-delimited escaped t_string of the hypothesis. (escape rules: space-> \_, \->\\, |->\p, EOL->\n, empty_string-> \0)
    dependency is the space-delimited dependency list.
    parent_bond_relationship is one of E (for PRE), O (for POST) or U(for UNK)
    additionals_list is a space-delimited list of additionals.
    Each additional as the format cnum:dpnd:prelist:postlist, where prelist and postlist are word index sequences delimited with "_".
    Finally, features_values represents all of the features values different of zero. The feature name is replaced by an index in the list of features_names written on the
    first line of the file.
    
    Warning to anybody that would like to modify the hypothesis files: the loading code is not flexible at all: even a trailing space is enough to generate a parsing error.
    
    example:
    START 37
    NULL_content NULL_function abnormal_child_bond both_ROS child_bond_count child_bond_similarity different_parent_cat example_penalty lm no_redundant_child numeral_mismatch one_side_ROS parent_bond_LD parent_bond_available parent_bond_similarity redundant_child_in_input redundant_child_in_source root_LD same_parent_cat same_root_content size trans_lm_prob trans_prob trans_prob_multiplied_by_size
    P|0|0|1||
    0|this|-1|E||7:1 9:1 13:1 19:1 21:1 22:1 23:1
    0|this|-1|E||7:1 9:1 13:1 19:1 21:1 22:1 23:1
    0|this|-1|E||7:1 9:1 18:1 19:1 21:1 22:1 23:1
    0|this|-1|E||7:1 9:1 13:1 19:1 21:1 22:1 23:1
    0|this|-1|E||7:1 9:1 13:1 19:1 21:1 22:1 23:1
    0|this|-1|E||7:1 9:1 13:1 19:1 21:1 22:1 23:1
    0|this|-1|E||7:1 9:1 13:1 19:1 21:1 22:1 23:1
    0|this|-1|E||7:1 9:1 12:1 13:1 19:1 21:1 22:1 23:1
    0|these|-1|E||7:1 9:1 13:1 19:1 21:1 22:0.0280899 23:0.0280899
    0|this|-1|E||7:1 9:1 12:1 13:1 19:1 21:1 22:1 23:1
    P|6 7|7|8|5|6:5
    0|[X5] 40 on*|1 -1 1|E||1:1 4:1 7:1 9:1 10:1 12:1 13:1 19:1 20:1 21:0.00694444 22:1 23:2
    0|a* [X5] 40|2 2 -1|E||1:1 4:1 7:1 9:1 10:1 12:1 13:1 19:1 20:1 21:0.0277778 22:1 23:2
    0|the* [X5] 40|2 2 -1|E||1:1 4:1 6:1 7:1 9:1 10:1 19:1 20:1 21:0.569444 22:1 23:2
    P|19 20|20|21|11 12 18|20:11_12_18
    0|wire guides|1 -1|E|11:1:0_1:2 12:1:0_1:2 18:1:0_1:2|7:1 13:1 14:1 15:3 19:1 20:1 21:0.666667 22:0.0909091 23:0.181818
    0|wire guides|1 -1|E|11:1:0_1:2 12:1:0_1:2 18:1:0_1:2|7:1 15:3 18:1 19:1 20:1 21:0.666667 22:0.0909091 23:0.181818
    0|the* wire guides|2 2 -1|E|11:2:0_1_2:3 12:2:0_1_2:3 18:2:0_1_2:3|7:1 15:3 17:-0.3 18:1 19:1 20:1 21:0.333333 22:0.0909091 23:0.181818
    0|corresponding* wire guide|1 2 -1|E|11:2:0_2:3 12:2:0_2:3 18:2:0_2:3|7:1 13:1 15:3 17:-0.3 19:1 20:1 21:0.030303 22:1 23:2
    0|the* wire guide|2 2 -1|E|11:2:0_1_2:3 12:2:0_1_2:3 18:2:0_1_2:3|7:1 15:3 17:-0.3 18:1 19:1 20:1 21:0.242424 22:1 23:2
    END
*/

class HumanReadableHypothesisSerialization {
  public:
    static void serialize_translation_candidate(const InitialHypothesesStorage& translation_candidate_, const WeightVector& weights, 
                                                        const std::vector<AugmentedPattern>& augmented_patterns, size_t input_size, std::ostream& stream);
    
    /*! \brief Faster deserialization code, that takes a c-string (ie null-terminated) as input.
     *  Warning: the input c-string will be modified by the function.
     */
    static void deserialize_translation_candidate(char* serialized_tc, 
                InitialHypothesesStorage* tc, std::vector<AugmentedPattern>* augmented_pattern_vect, size_t* input_size);
    
    private:
    // Serialization
    static void serialize_stack(const std::vector<Hypothesis*>& stack, const std::map<std::string, int>&, std::ostream& stream);
    static void serialize_hypothesis_to_stream(const Hypothesis& hyp, const std::map<std::string, int>&, std::ostream& stream);
    static void serialize_additional(const Additional& add, std::ostream& stream);

    // Faster De-Serialization (using c-strings that are modified in-place)
    static Hypothesis* deserialize_hypothesis_from_string(char* hyp_as_string, const std::vector<std::string>& feature_names,
    const FeatureVector& empty_feature_vector);
    static void deserialize_additional(char* add_as_string, Additional*);     
    static void deserialize_additional_list(char* string, AdditionalsList* add_list);                                                                
};


#endif //__HYPOTHESIS_SERIALIZATION_HPP__


