/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "input_handler.hpp"

#include <boost/iostreams/stream.hpp>
#include <boost/algorithm/string.hpp>
#include <string>
#include <cstdio>
#include <sstream>
#include <iostream>
#include <algorithm>

#include "utils.hpp"
#include "pugixml/src/pugixml.hpp"

void parse_sentence_into_xml(const std::string& sentence, std::string* parsed_sentence) {
    LOG(INFO) << "SENTENCE:" << sentence;
    std::string cmd = Params::Translator::parse_command;
    std::string escaped_sentence(sentence.c_str());
    boost::algorithm::replace_all(escaped_sentence, "\\", "");
    boost::algorithm::replace_all(escaped_sentence, "'", "\\'");
    boost::algorithm::replace_all(cmd, "%SENTENCE%", escaped_sentence);
    exec_cmd(cmd.c_str(), parsed_sentence);
    LOG_IF(ERROR, parsed_sentence->empty()) << "Failed to parse. parse_command=" << cmd;
}


void parse_flat_xml_into_xml_tree(const std::string& para_sentence_buf, std::string* out, std::string* id) {
    pugi::xml_document doc;
    pugi::xml_parse_result result = doc.load_buffer(para_sentence_buf.c_str(), para_sentence_buf.size()+1);
    LOG_IF(FATAL, result.status != pugi::status_ok) << "Error parsing xml input file : " << result.description() << std::endl;

    *id = doc.child("sentence").attribute("id").value();

    // parsing i_sentence
    pugi::xml_node i_sentence = doc.child("sentence").child("i_sentence");
    LOG_IF(FATAL, i_sentence.empty()) << "i_sentence tag is empty while input_mode is xml_nonparsed\n";
    const char *sentence = i_sentence.child_value();
    LOG(INFO) << "i_sentence:" << sentence;
    parse_sentence_into_xml(sentence, out);
}

//    if (!out.empty()) {
//        LOG(INFO) << "Parser:" << cmd;
//        LOG(INFO) << "Output:" << out;
//
//        std::istringstream is(out);
//        StreamInputHandler input_handler(&is);
//        std::string tmp_id;
//        return input_handler.read_one_sentence(dtree, &tmp_id);
//    } else {
//        return false;
//    }

void read_sentence_tag_enclosed_text(std::istream* stream, std::string* para_sentence_buf) {
    std::string line_buf;
    bool in_sentence = false;
    while (!stream->eof()) {
        std::getline(*stream, line_buf);
        if (line_buf.find("<sentence") != std::string::npos) {
            in_sentence = true;
        }
        if (in_sentence) {
            *para_sentence_buf += line_buf;
        }
        if (line_buf.find("</sentence") != std::string::npos) {
            in_sentence = false;
            break;
        }
    }
    LOG_IF(FATAL, in_sentence == true) << "Found opening <sentence> tag unmatched by closing </sentence> tag in input stream";
}

void get_xml(std::istream* input_stream, std::string *para_sentence_buf, std::string *id) {
    if (Params::Global::input_mode == "xml_nonparsed") {
        std::string non_parsed;
        read_sentence_tag_enclosed_text(input_stream, &non_parsed);
        if (non_parsed != "") {
            std::string parsed;
            parse_flat_xml_into_xml_tree(non_parsed, &parsed, id);
            std::stringstream stream(parsed);
            read_sentence_tag_enclosed_text(&stream, para_sentence_buf);
        }
    } else if (Params::Global::input_mode == "plain") {
        std::string plain_line;
        std::getline(*input_stream, plain_line);
        //std::cerr << "received line " << plain_line << std::endl;
        if (plain_line != "") {
            std::string parsed;
            parse_sentence_into_xml(plain_line, &parsed);
            std::stringstream stream(parsed);
            read_sentence_tag_enclosed_text(&stream, para_sentence_buf);
//        para_sentence_buf.swap(parsed);
        }
    } else if (Params::Global::input_mode == "xml") {
        read_sentence_tag_enclosed_text(input_stream, para_sentence_buf);
    } else {
        LOG(FATAL) << "unrecognized stream input mode: " << Params::Global::input_mode;
    }
}

void parse_xml_tree(const std::string& para_sentence_buf, pugi::xml_document *input_data) {
    pugi::xml_parse_result result = input_data->load_buffer(para_sentence_buf.c_str(), para_sentence_buf.size()+1);
    LOG_IF(FATAL, result.status != pugi::status_ok) << "Error parsing xml input file : " << result.description() << std::endl;
}

bool get_dtree(std::istream* input_stream, pugi::xml_document *input_data, pugi::xml_node *i_data, Dtree *dtree, std::string *id) {
    *i_data = i_data->next_sibling("i_data");
    while (*i_data == NULL) {
        std::string para_sentence_buf;
        get_xml(input_stream, &para_sentence_buf, id);
        if (para_sentence_buf == "") {
            return false;
        }
        parse_xml_tree(para_sentence_buf, input_data);
        *i_data = input_data->child("sentence").child("i_data");
    }

    if (*id == "") *id = input_data->child("sentence").attribute("id").value();

    for (pugi::xml_node i_phrase = i_data->first_child();
         i_phrase; i_phrase = i_phrase.next_sibling()) {
        // skip tags other than "phrase"
        if (strcmp(i_phrase.name(), "phrase") != 0) continue;
        pugi::xml_node i_word = i_phrase.child("word");
        DtreeNode node;
        node.dpnd             = atoi(i_phrase.attribute("dpnd").value());
        node.num              = dtree->nodes().size();
        node.align_index      = -1;
        node.word.content     = i_phrase.child_value("word");
        node.word.lem         = i_word.attribute("lem").value();
        node.word.pos         = i_word.attribute("pos").value();
        node.word.content_p   = atoi(i_word.attribute("content_p").value());
        node.word.cat         = i_phrase.attribute("cat").value();
        node.word.type        = i_phrase.attribute("type").value();
        dtree->push_node(node);
    }
    dtree->format_dtree();
    return true;
}

bool InputHandler::read_one_sentence(Dtree *dtree, std::string *id) {
    assert(input_stream_ != NULL);
    if (get_dtree(input_stream_, &input_data_, &current_i_data_, dtree, id)) {
        return true;
    } else {
        return false;
    }
}

bool ConcatenatedInputHandler::read_one_sentence(Dtree *dtree, std::string *id) {
    // check the current input handler is readable
    if (current_input_handler_ == NULL) {
        bool could_update = update_input_handler();
        // if we could not update, nothing more to be done
        if (!could_update) return false;
    }
    if (current_input_handler_->read_one_sentence(dtree, id)) {
        // if we could read, we are done
        return true;
        // TODO(Fabien): beware of InputHandler::read_one_sentence setting partially dtree and id
        // and then returning false
    }
    // if we get to this point, we could not read a sentence
    // therefore, we try to update the input_handler
    bool could_update = update_input_handler();
    // if we could not update, nothing more to be done
    if (!could_update) return false;
    // otherwise, we try the read process again
    return read_one_sentence(dtree, id);
};

bool ConcatenatedInputHandler::update_input_handler() {
    if (next_filename_index_ >= filename_vector_.size()) {
        return false;
    }
    std::string next_filename = filename_vector_[next_filename_index_];
    if (current_input_handler_ != NULL) {
        delete current_input_handler_;
    }
    current_input_handler_ = new InputHandler(next_filename);
    next_filename_index_++;
    
    if (current_input_handler_->is_ready_to_read()) {
        return true;
    } else {
        return update_input_handler();
    }
}
