/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*! @file
  This file contains the definitions of classes which manage the input
  to the decoder. */

#ifndef INPUT_HANDLER
#define INPUT_HANDLER

#include <boost/shared_ptr.hpp>
#include <boost/make_shared.hpp>
#include <boost/noncopyable.hpp>
#include <glog/logging.h>
#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include "def.hpp"
#include "dtree.hpp"
#include "params.hpp"
#include "pugixml/src/pugixml.hpp"

/*! Interface class for input handlers */
class InputHandlerInterface : public boost::noncopyable {
  public:
    //! destructor
    virtual ~InputHandlerInterface() {}

    //! read one input sentence from the input stream (should be
    //! implemented in the child classes)
    virtual bool read_one_sentence(Dtree *dtree, std::string *id) = 0;
};

/*! This class handles only one input stream */
class InputHandler : public InputHandlerInterface {
  public:
    //! default constructor
    InputHandler() : input_stream_(NULL) {}
    
    //! constructor with input filename
    explicit InputHandler(const std::string &filename) : 
        file_input_stream_(boost::make_shared<std::ifstream>(filename.c_str())) {
        LOG(INFO) << "Trying to open file '" << filename << "'.";
        input_stream_ = file_input_stream_.get();

        if (file_input_stream_->is_open() && is_ready_to_read()) {
            LOG(INFO) << "Success!";
        } else {
            LOG(FATAL) << "Failed to open file " << filename << "!";
        }
    }

    explicit InputHandler(std::istream* stream) :
        input_stream_(stream) {
        if (is_ready_to_read()) {
            LOG(INFO) << "Success!";
        } else {
            LOG(FATAL) << "Failed to read input stream!";
        }
    }
    
    //! check if the input_stream_ is correctly opened
    bool is_ready_to_read() {
        if (input_stream_ == NULL) return false;
        assert(input_stream_ != NULL);
        return input_stream_  && !input_stream_->fail() && !input_stream_->bad() && !input_stream_->eof();
    }

    //! read one input sentence from the input stream
    bool read_one_sentence(Dtree *dtree, std::string *id);
    
  private:
    //! input stream
    std::istream* input_stream_;
    boost::shared_ptr<std::ifstream> file_input_stream_;
    pugi::xml_document input_data_;
    pugi::xml_node current_i_data_;
};

/*! This class can handle multiple of input streams */
class ConcatenatedInputHandler : public InputHandlerInterface {
  public:
    //! constructor with a list of input files
    explicit ConcatenatedInputHandler(std::vector<std::string> f_vec) :
        filename_vector_(f_vec), next_filename_index_(0), current_input_handler_(NULL) {}

    //! read one input sentence from the input stream    
    bool read_one_sentence(Dtree *dtree, std::string *id);

  private:
    //! list of input files
    VS filename_vector_;
    //! index to the next input file
    size_t next_filename_index_;
    //! currently using input handler
    InputHandler* current_input_handler_;
    //! update the current_input_handler_ if it is not available any
    //! more. this will move the handler to the next file.
    bool update_input_handler();
};


// plain
// std::cout << "Enter: ";
//char sentence[1024];
//std::cin.getline(sentence, sizeof(sentence));
//if (std::cin.eof()) {
//    break;
//}
//if (strlen(sentence) == 0) {
//    continue;
//}
//
//
//
//if (!parsed_sentence->empty()) {
//    LOG(INFO) << "Parser:" << cmd;
//    LOG(INFO) << "Output:" << *parsed_sentence;
//
//    std::istringstream is(out);
//    StreamInputHandler input_handler(&is);
//    bool could_read_sentence = input_handler.read_one_sentence(input_sent.get(), &id);
//    if (!could_read_sentence) {
//        break;
//    }
//} else {
//    LOG(ERROR) << "Failed to parse. parse_command=" << cmd;
//    continue;
//}


    
#endif
