/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include <boost/foreach.hpp>
#include <iostream>
#include <vector>
#include <string>
#include <set>
#include <cassert>
#include "def.hpp"
#include "decoding/augmented_pattern.hpp"
#include "hypothesis.hpp"

#ifndef __INPUT_INSERTION_ID_HPP__
#define __INPUT_INSERTION_ID_HPP__


class InputInsertionIdSimple {
  public:
    InputInsertionIdSimple(): index_(static_cast<WordIndex>(-1)) {}

    explicit InputInsertionIdSimple(WordIndex index): index_(index) {}
    
    explicit InputInsertionIdSimple(const ConnectedAugmentedPattern& pat): index_(pat.get_root()) {}  
    
    InputInsertionIdSimple(const InputInsertionIdSimple& other): index_(other.index_) {}
    
    bool compatible_with(WordIndex index) const {return index_ == index;}
    bool is_valid() const {
        return index_ != static_cast<WordIndex>(-1);
    }
    bool is_included_in_pattern(const PatternType& pat) {
        return pat.count(index_) > 0;
    }
    void add_refs_to_set(std::set<WordIndex>* set) const {
        set->insert(index_);
    }
    
    bool has_intersection_with_set(const std::set<WordIndex>& set) const {
        return set.count(index_) != 0;
    }
    bool operator==(const InputInsertionIdSimple& other) const {return index_ == other.index_;}
    bool operator< (const InputInsertionIdSimple& other) const {return index_ < other.index_;}
    
    static InputInsertionIdSimple get_ref_id(const std::string& str) {
        // a bit hackish
        // since get_ref_id get called exceedingly often, it is actually
        // interesting to have a very optimized version
        size_t size_ = str.size();
        if (size_ < 3 || str[0] != '[' || str[size_ - 1] != ']') {
            return InputInsertionIdSimple();
        }
        unsigned char zero = '0';
        unsigned char val = str[2];
        assert(val >= zero && val <= zero + 10);
        WordIndex result = val - zero;
        for (size_t i = 3; i < size_ - 1; i++) {
            unsigned char val = str[i];
            result *= 10;
            result += val - zero;
        }
        return InputInsertionIdSimple(result);
    }
    
  private:
    WordIndex index_;
};

class Hypothesis;

class InputInsertionIdMulti {
  public:
    InputInsertionIdMulti() {}
    explicit InputInsertionIdMulti(WordIndex index) {
        index_vect_.insert(index);
        }
        
    explicit InputInsertionIdMulti(const std::vector<WordIndex>& v) {
        index_vect_.insert(v.begin(), v.end());
        assert(index_vect_.size() == v.size());
    }
     
    explicit InputInsertionIdMulti(const LatterallyDisconnectedAugmentedPattern& pat) {
        std::vector<WordIndex> v = pat.get_roots();
        index_vect_.insert(v.begin(), v.end());
        assert(index_vect_.size() == v.size());
    }
     
    InputInsertionIdMulti(const InputInsertionIdMulti& other): index_vect_(other.index_vect_) {}
    
    bool compatible_with(WordIndex index) const {return index_vect_.count(index) != 0;}
    
    std::vector<int> compatible_with_cnum_additionals(const std::vector<WordIndex>& cnum_list) {
        std::vector<int> result;
        for (int i = 0; i < static_cast<int>(cnum_list.size()); i++) {
            if (index_vect_.count(cnum_list[i]) == 1) result.push_back(i);
        }
        if (result.size() != index_vect_.size()) return std::vector<int>();
        return result;
    }
    
    bool is_valid() const {
        return index_vect_.size() > 0;
    }
    bool is_included_in_pattern(const PatternType& pat) {
        BOOST_FOREACH(WordIndex idx, index_vect_) {
            if (pat.count(idx) == 0) return false;
        }
        return true;
    }
    
    void add_refs_to_set(std::set<WordIndex>* set) const {
        set->insert(index_vect_.begin(), index_vect_.end());
    }
    
    bool has_intersection_with_set(const std::set<WordIndex>& set) const {
        BOOST_FOREACH(WordIndex idx, index_vect_) {
            if (set.count(idx) != 0) return true;
        }
        return false;
    }
    bool operator==(const InputInsertionIdMulti& other) const {return index_vect_ == other.index_vect_;}
    bool operator< (const InputInsertionIdMulti& other) const {return index_vect_ < other.index_vect_;}
    
    static InputInsertionIdMulti get_ref_id(const std::string& str) {
        // a bit hackish
        // since get_ref_id get called exceedingly often, it is actually
        // interesting to have a very optimized version
        size_t size_ = str.size();
        if (size_ < 3 || str[0] != '[' || str[size_ - 1] != ']') {
            return InputInsertionIdMulti();
        }
        unsigned char zero = '0';
        unsigned char comma = ',';
        WordIndex current_val = 0;
        InputInsertionIdMulti result;
        
        unsigned char val = str[2];
        assert(val >= zero && val <= zero + 10);
        current_val = val - zero;
        for (size_t i = 3; i < size_ - 1; i++) {
            unsigned char val = str[i];
            if (val == comma) {
                assert(result.index_vect_.count(current_val) == 0);
                result.index_vect_.insert(current_val);
                current_val = 0;
                continue;
            }
            current_val *= 10;
            current_val += val - zero;
        }
        assert(result.index_vect_.count(current_val) == 0);
        result.index_vect_.insert(current_val);
        return result;
    }

    
  private:
    std::set<WordIndex> index_vect_;
    friend void additional_cleaning(const InputInsertionIdMulti& insertion_point, Hypothesis* hyp);
    friend std::ostream& operator<<(std::ostream& stream, const InputInsertionIdMulti& id);
};

inline std::ostream& operator<<(std::ostream& stream, const InputInsertionIdMulti& id) {
    stream << "I<";
    BOOST_FOREACH(WordIndex i, id.index_vect_) {
        stream << i << ",";
    }
    stream << ">";
    return stream;
}
//typedef InputInsertionIdSimple InputInsertionId;
typedef InputInsertionIdMulti InputInsertionId;

#endif //__INPUT_INSERTION_ID_HPP__
