/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "lexical_translation_scorer.hpp"
#include <boost/lexical_cast.hpp>
#include <math.h>
#include <fstream>
#include <string>
#include <utility>
#include <map>
#include "utils.hpp"
#include "params.hpp"

double get_lexical_trans_prob(const std::map<std::pair<std::string, std::string>, double>& table, const std::string& word1, const std::string& word2) {
    std::map<std::pair<std::string, std::string>, double>::const_iterator it = table.find(std::make_pair(word1, word2));
    if (it == table.end()) {
        return 0.00000001;
    } else {
        return it->second;
    }
}

double LexicalTranslationScorer::calc_lexical_trans_prob_s_t(const TM& tm, const DataForTM* data_for_tm) const {
    if (!read_flag_) {
        return 0.0;
    }

    const Dtree& s_dtree = tm.para_sentence().s_dtree();
    const Dtree& t_dtree = tm.para_sentence().t_dtree();

    double s_given_t_prob = 0.0;
    BOOST_FOREACH(const DtreeNode* s_w, data_for_tm->s_body_ref) {
        const Alignment* align = tm.para_sentence().get_alignment_s(s_w);
        Lem s_w_lem = s_dtree.node_at(s_w->num)->word.lem;
        if (align == NULL) {
            s_given_t_prob += log(get_lexical_trans_prob(lexical_translation_table_s_t_, s_w_lem, "NULL"));
        } else {
            double tmp_sum = 0.0;
            int count = 0;
            BOOST_FOREACH(WordIndex t_index, align->p2) {
                tmp_sum += get_lexical_trans_prob(lexical_translation_table_s_t_, s_w_lem, t_dtree.node_at(t_index)->word.lem);
                count++;
            }
            s_given_t_prob += log(tmp_sum / boost::lexical_cast<double>(count));
        }
    }
    return s_given_t_prob;
}

double LexicalTranslationScorer::calc_lexical_trans_prob_t_s(const TM& tm, const DataForTM* data_for_tm) const {
    if (!read_flag_) {
        return 0.0;
    }

    const Dtree& s_dtree = tm.para_sentence().s_dtree();
    const Dtree& t_dtree = tm.para_sentence().t_dtree();

    double t_given_s_prob = 0;
    BOOST_FOREACH(const DtreeNode* t_w, data_for_tm->t_body_ref) {
        const Alignment* align = tm.para_sentence().get_alignment_t(t_w);
        Lem t_w_lem = t_dtree.node_at(t_w->num)->word.lem;
        if (align == NULL) {
            t_given_s_prob += log(get_lexical_trans_prob(lexical_translation_table_t_s_, t_w_lem, "NULL"));
        } else {
            double tmp_sum = 0.0;
            int count = 0;
            BOOST_FOREACH(WordIndex s_index, align->p1) {
                tmp_sum += get_lexical_trans_prob(lexical_translation_table_t_s_, t_w_lem, s_dtree.node_at(s_index)->word.lem);
                count++;
            }
            t_given_s_prob += log(tmp_sum / boost::lexical_cast<double>(count));
        }
    }
    return t_given_s_prob;
}

LexicalTranslationScorer::LexicalTranslationScorer(std::string filename) {
    std::ifstream ist(filename.c_str());
    if (ist.is_open()) {
        LOG(INFO) << "Reading lexical translation table from " << filename << "... ";
        std::string line_buf;
        while (!ist.eof()) {
            getline(ist, line_buf);
            if (line_buf == "") {
                continue;
            }
            VS data = split(line_buf, "\t");
            if (data[0] == "lex_i2j") {
                if (Params::Translator::tr_language == "1") {
                    lexical_translation_table_s_t_[std::make_pair(data[2], data[1])] = boost::lexical_cast<double>(data[3]);
                } else if (Params::Translator::tr_language == "2") {
                    lexical_translation_table_t_s_[std::make_pair(data[2], data[1])] = boost::lexical_cast<double>(data[3]);
                }
            } else if (data[0] == "lex_j2i") {
                if (Params::Translator::tr_language == "1") {
                    lexical_translation_table_t_s_[std::make_pair(data[2], data[1])] = boost::lexical_cast<double>(data[3]);
                } else if (Params::Translator::tr_language == "2") {
                    lexical_translation_table_s_t_[std::make_pair(data[2], data[1])] = boost::lexical_cast<double>(data[3]);
                }
            }
        }
        LOG(INFO) << "Done reading lexical translation table.";
        read_flag_ = true;
    } else {
        // LOG(FATAL) << "Error: can't open " << filename << ".";
        LOG(INFO) << "Error: can't open " << filename << ".";
        read_flag_ = false;
    }
}
