/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include <glog/logging.h>
#include <boost/scoped_ptr.hpp>

#include <iostream>
#include <string>
#include <vector>
#include "input_handler.hpp"
#include "params.hpp"
#include "translator.hpp"
#include "translation_manager.hpp"
#include "decoding/translation_log.hpp"

//#include "decoding/lattice_trans.hpp"

int main(int argc, char *argv[]) {
    srand(time(NULL));
    // First, the command line arguments and the config file are parsed to set the various parameters.
    Params::parse_options(argc, argv);

    // We initialize the logging after the parameter parsing to have a chance to set the logging options
    // before the initialization
    google::InitGoogleLogging("");
    LOG(INFO) << "Starting Translation with options:\n" << Params::ParamsInPrintableForm() << "\n";

//    latt_old::test2_lattice();
//    latt_old::test_lattice_reduction();
//    std::cerr << "Done test lattice" << std::endl;
    //exit(0);

    // incorrect translation direction
    if (Params::Translator::source_language == Params::Translator::target_language) {
        std::cerr << "Error: specified the same language for both source and target" << std::endl;
        return 1;
    }

    int start_setup_time = t_log::get_time_in_ms();


    // Since translator and input_handler need to be created differently depending on the input mode,
    // we need to create them dynamically with a new.
    // Using scoped_ptr allows us to be sure the dynamically created objects are correctly deleted at the end.
    // Big Warning: as it is now, in hypothesis mode, a null pointer is passed as the input_handler to the
    // translator_manager. It works since the translatot_manager do not need to access the input_handler in hypothesis mode.
    // But better be careful with this.
    boost::scoped_ptr<Translator> translator;
    boost::scoped_ptr<InputHandlerInterface> input_handler;

//    std::cout << "HERE\n" << std::endl;
    if (Params::Global::input_mode != "hypothesis") {
//        std::cout <<"non hyp" << std::endl;
        if (Params::Global::input_mode == "xml" || Params::Global::input_mode == "xml_nonparsed") {
//            std::cout << "xml mode" << std::endl;
            input_handler.reset(new ConcatenatedInputHandler(Params::Translator::input_filename_list));
        } else if (Params::Global::input_mode == "plain") {
//            std::cout << "plain mode" << std::endl;
            input_handler.reset(new InputHandler(&std::cin));
        } else {
            LOG(FATAL) << "Unrecognized non-hypothesis input mode: " << Params::Global::input_mode;
        }

        // We first get the TRInterface objects that will give us access to the examples
        std::vector<TRInterfacePtr> tr_list;
        BOOST_FOREACH(std::string tr_db_filename, Params::Translator::tr_db_filename_list) {
            LOG(INFO) << "Loading Example DB: " << tr_db_filename << "\n";
            TRInterfacePtr tr;
            if (Params::Translator::use_tr_server) {
                tr = get_client_TR(tr_db_filename,
                                   Params::Translator::tr_server_language,
                                   Params::Translator::tr_server_host,
                                   Params::Translator::tr_server_port);
            } else {
                if (Params::Translator::no_siblings_retrieval) {
                    tr = get_default_TR(tr_db_filename, Params::Translator::tr_language, Params::Translator::tr_max_nb_tms_per_pattern);
                } else {
                    tr = get_default_TR_WC(tr_db_filename, Params::Translator::tr_language, Params::Translator::tr_max_nb_tms_per_pattern, Params::Translator::tr_fast_siblings,
                    Params::Translator::use_wildcard_pair_cache);
                }
            }
            assert(tr);  // check the tr object was actually created
            LOG(INFO) << "Finished loading Example DB: " << tr_db_filename << "\n";
            tr_list.push_back(tr);
        }

        // We then create the translator object, that will produce the translations
        translator.reset(new Translator(tr_list, Params::Translator::feature_weight,
                                        Params::Translator::default_dpnd,
                                        Params::Translator::dic_punc_filename,
                                        Params::Translator::dic_words_filename,
                                        Params::Translator::lexical_translation_prob_filename,
                                        Params::Translator::source_language,
                                        Params::Translator::target_language) );
    } else {
        translator.reset(new Translator(Params::Translator::feature_weight));
    }

    LOG(INFO) << "Translator object has been created.\n";
    int end_setup_time = t_log::get_time_in_ms();

    int start_time = t_log::get_time_in_ms();

    // Create a new output handler
    // As the Translation Manager will be in charge of deleting the pointer, we will not have to do it here.
    OutputHandlerInterface* output_handler;
    boost::scoped_ptr<std::ofstream> alternative_output;

    boost::scoped_ptr<std::ofstream> time_log_file;

    if (Params::Global::output_mode == "bleu_eval") {
        alternative_output.reset(new std::ofstream("bleu_eval_sentences.log"));
        LOG_IF(FATAL, !(*alternative_output)) << "Could not open bleu eval file '" << "bleu_eval_sentences.log" << "'\n";
        OutputHandlerInterface* alternative_output_handler = new StreamPrinter(alternative_output.get());
        if (Params::Translator::nb_threads != 0) {
            // We decorate the output handler with a SerializeToOutputDecorator so that all output will be written in proper order
            alternative_output_handler = new SerializeToOutputDecorator(alternative_output_handler);
        }
        output_handler = get_bleu_evaluator(alternative_output_handler);
    //} else if (Params::Global::output_mode == "online_tuning") {
        //output_handler = new OnlineTuner(translator->get_weights());
    } else {
        if (Params::Global::time_log_file.size() > 0) {
            if (Params::Global::time_log_file == "stderr") {
                output_handler = new StreamPrinter(&std::cout, &std::cerr);
            } else {
                time_log_file.reset(new std::ofstream(Params::Global::time_log_file.c_str()));
                output_handler = new StreamPrinter(&std::cout, time_log_file.get());
            }
        } else {
            output_handler = new StreamPrinter(&std::cout);
        }
        if (Params::Translator::nb_threads != 0) {
            // We decorate the output handler with a SerializeToOutputDecorator so that all output will be written in proper order
            output_handler = new SerializeToOutputDecorator(output_handler);
        }
    }

    // A translator manager object is created, that will feed sentences to the translator and print the results
    // SerializeToOutputDecorator* serialized_output_handler = new SerializeToOutputDecorator(output_handler);
    AsyncTranslationManager translation_manager(input_handler.get(), *translator, output_handler);
    // The method translate_all makes the translator translate all of its input
    LOG(INFO) << "Starting translation (multithread).\n";
    translation_manager.translate_all(Params::Translator::nb_threads);

    if (Params::Translator::display_progress) { // for aesthetic when running evaluation script
        std::cerr << std::endl;
    }

    int end_time = t_log::get_time_in_ms();
    LOG(INFO) << "Finished translation work in " << end_time-start_time << " ms" <<
            " (+ " << end_setup_time-start_setup_time << " ms setup time  --> " <<
            end_time-start_setup_time << " ms total time)";

    return 0;
}

