/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*! @file
  This file contains the definition of ParaSentence class
*/
#ifndef PARASENTENCE_INCLUDED
#define PARASENTENCE_INCLUDED

#include <boost/serialization/string.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/utility.hpp>
#include <iostream>
#include <string>
#include <vector>
#include "dtree.hpp"
#include "alignment.hpp"

/*!
  The class ParaSentence represents a pair of parsed and aligned sentence pairs.
  It is made noncopyable by inheriting from boost::noncopyable. This means that 
  if one tries to copy or assign an object of this class, the compiler will complain
  with a message containing something like: 
  "Error: noncopyable(const noncopyable&) constructor is private" (+ a lot of template instanciation mess)
*/
class ParaSentence : public boost::noncopyable  {
    friend class TMBuilder;
  public:
    //! default constructor
    ParaSentence() {}

    //! constructor using sentence id and both source and target length
    ParaSentence(std::string s_id, int s_length, int t_length) : id_(s_id), s_dtree_(s_length), t_dtree_(t_length) {}
    
    //~ParaSentence();

    //! get sentence id
    const std::string& id() const {return id_;}
    //! get source-side dtree as const
    const Dtree& s_dtree() const {return s_dtree_;}
    //! get target-side dtree as const
    const Dtree& t_dtree() const {return t_dtree_;}
    //! get source-side dtree
    Dtree& get_modifiable_s_dtree() {return s_dtree_;}
    //! get target-side dtree
    Dtree& get_modifiable_t_dtree() {return t_dtree_;}
    //! accessor to the s_t_match as const
    const std::vector<Alignment>& s_t_match() const {return s_t_match_;}

    //! find the Alignment which includes the specified DtreeNode in the source side of the Alignment
    const Alignment* get_alignment_s(const DtreeNode *w) const;
    //! find the Alignment which includes the specified DtreeNode in the target side of the Alignment
    const Alignment* get_alignment_t(const DtreeNode *w) const;

    //! set sentence id
    void set_id(std::string i) {id_ = i;}

    //! print both source and target dtrees one by one
    void print_parallel_dtree();

    //! used from operator<< to print parallel dtree
    void print(std::ostream *os) const;
    void print(std::ostream *os, const DtreeNode *s_node_r, const DtreeNode *t_node_r, const int *s_bond, const int *t_bond) const;

    //! small function for sanity check
    //! (used in the construction of tm by the TR code)
    bool is_default_constructed() const {
        return id_.size() == 0 &&
            s_t_match_.size() == 0 &&
            t_dtree_.is_default_constructed() &&
            s_dtree_.is_default_constructed();
    }

    void set_alignment_for_OOV() {
        assert(s_t_match_.size() == 0);
        Alignment alignment;
        alignment.p1.push_back(0);
        alignment.p2.push_back(0);
        alignment.accept = YES; 
        s_t_match_.push_back(alignment);
    }
    
  private:
    std::string            id_;         //!< sentence id
    Dtree                  s_dtree_;    //!< source-side Dtree
    Dtree                  t_dtree_;    //!< target-side Dtree
    std::vector<Alignment> s_t_match_;  //!< alignment between source and target words

    // friend class boost::serialization::access;
    // template<class Archive>
    // void serialize(Archive &ar, const unsigned int version) {
    // ar & id_;
    // ar & s_dtree_;
    // ar & t_dtree_;
    // ar & s_t_match_;
    // }
};
std::ostream &operator<<(std::ostream &os, const ParaSentence &t);

#endif
