/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "params.hpp"
#include <boost/lexical_cast.hpp>
#include <boost/program_options.hpp>
#include <boost/foreach.hpp>
#include <glog/logging.h>
#include <set>
#include <string>
#include <vector>

#ifndef EBMT_INI
#define EBMT_INI "../sample/sample.ini"
#endif

std::vector<std::vector<std::string> > read_reference(const char* filename) {
    std::vector<std::vector<std::string> > result;
    std::ifstream infile(filename);
    if (!infile) return result;

    std::string line;
    while (std::getline(infile, line)) {
        std::stringstream ss(line);
        std::istream_iterator<std::string> begin(ss);
        std::istream_iterator<std::string> end;

        result.push_back(std::vector<std::string>(begin, end));
#ifndef NDEBUG
        BOOST_FOREACH(const std::string& word, result.back()) {
            assert(word.size() > 0);
            assert(word != "\n");
        }
        assert(result.back().size() > 0);
#endif //NDEBUG
    }

    return result;
}

namespace po = boost::program_options;

namespace Params {
    // Warning: the default values specified here will be overridden by the default values specified
    // to boost::program_options.
    // Therefore, to change the default values, change them in the parse_options function.
    namespace Global {
        std::string config_filename = EBMT_INI;
        bool web = false;
        bool web_verbose = false;
        bool web_all = false;
        std::string web_dir = ".";
        std::string output_mode = "tuning";
        std::string input_mode = "xml";
        int sentence_id = 0;
        //bool no_lm_postprocessing = false;
        bool color = false;
        bool keep_tms_in_memory = false;
        std::string time_log_file;
    }

    namespace Features {
        std::set<std::string> all_features;
        std::set<std::string> ignored_features;
    }

    namespace Translator {
        size_t nb_threads;
        //std::string input_filename = "";
        std::vector<std::string> input_filename_list;
        std::vector<std::string> tr_db_filename_list;
        std::string source_language = "ja";
        std::string target_language = "en";
        std::string dic_words_filename = "dic/words-je.dat";
        std::string dic_punc_filename = "dic/punc-je.dat";
        std::string lm_filename = "";
        std::string tr_result_filename = "";
        std::string lexical_translation_prob_filename = "";
        std::string default_dpnd = "";
        std::string feature_weight = "weights.ini";
        int beam_width = 10;
        bool no_sibling = false;
        bool no_head_switch = true;
        bool iso_dic_trans = false;
        std::string tr_mode = "cplus_auto";
        std::string tr_language = "1";
        bool server_mode = false;
        int port = 43251;
        std::string parse_command;
        int n_gram_length = 5;
        int n_best_length = 10;
        bool use_tr_server = false;
        std::string tr_server_language = "1";
        std::string tr_server_host = "localhost";
        std::string tr_server_port = "12223";
        std::string hypothesis_output_prefix = "";
        int leading_zeros = 5;
        int rerank_beam = 0;
//        bool use_old_lm_postprocess = false;
        std::string decoder = "lattice";
//        bool enable_lm_estimation = false;
        bool display_progress = false;
        bool tr_fast_siblings = false;
        int tr_max_nb_tms_per_pattern = 200;
        bool use_wildcard_pair_cache = false;
        std::string reference_file;
        std::vector<std::vector<std::string> > reference_translations;
        bool use_oracle_decoding = false;
        bool first_word_should_be_uppercased_for_lm_eval = false;
        double randomization_factor = 0.0;
//        bool lm_state_merging = false;
        int max_patterns = 4000;
        int beam_width_reduced = 100;
//        bool multiply_beam_width_by_pattern_size = false;
//        bool enable_new_features = false;
//        bool filter_additionals = false;
        bool no_keep_input_order_of_additionals = false;
//        bool average_lm = false;
//        int beam_width_final_multiplier = 5;
        bool shuffle_translation_order = false;
        bool verbose_features = false;
        bool disable_lattice_reduction = false;
        bool reconstruct_final_hypotheses = false;
        int lattice_beam_pre_truncation = 0;
        bool no_siblings_retrieval = false;
        bool use_nbest_parses = false;
}

    std::ostream& operator<<(std::ostream& stream, ParamsInPrintableForm) {
        std::string prefix("\t--");
        std::string suffix("\n");
        stream << prefix << "config_filename=" << Global::config_filename << suffix;
        stream << prefix << "web=" << Global::web << suffix;
        stream << prefix << "web_verbose=" << Global::web_verbose << suffix;
        stream << prefix << "web_all=" << Global::web_all << suffix;
        stream << prefix << "web_dir=" << Global::web_dir << suffix;
        stream << prefix << "output_mode=" << Global::output_mode << suffix;
        stream << prefix << "input_mode=" << Global::input_mode << suffix;
        stream << prefix << "sentence_id=" << Global::sentence_id << suffix;

        stream << prefix << "nb_threads=" << Translator::nb_threads << suffix;
        //stream << prefix << "input_filename=" << Translator::input_filename << suffix;

        stream << prefix << "input_filename_list=";
        BOOST_FOREACH(const std::string& str, Translator::input_filename_list) {
            stream << str << " ";
        }
        stream << suffix;

        stream << prefix << "tr_db_filename_list=";
        BOOST_FOREACH(const std::string& str, Translator::tr_db_filename_list) {
            stream << str << " ";
        }
        stream << suffix;

        stream << prefix << "decoder=" << Translator::decoder << suffix;

        stream << prefix << "source_language=" << Translator::source_language << suffix;
        stream << prefix << "target_language=" << Translator::target_language << suffix;
        stream << prefix << "dic_words_filename=" << Translator::dic_words_filename << suffix;
        stream << prefix << "dic_punc_filename=" << Translator::dic_punc_filename << suffix;

        stream << prefix << "lm_filename=" << Translator::lm_filename << suffix;
        stream << prefix << "tr_result_filename=" << Translator::tr_result_filename << suffix;
        stream << prefix << "lexical_translation_prob_filename=" << Translator::lexical_translation_prob_filename << suffix;
        stream << prefix << "default_dpnd=" << Translator::default_dpnd << suffix;
        stream << prefix << "feature_weight=" << Translator::feature_weight << suffix;
        stream << prefix << "beam_width=" << Translator::beam_width << suffix;
        stream << prefix << "no_sibling=" << Translator::no_sibling << suffix;
        stream << prefix << "no_head_switch=" << Translator::no_head_switch << suffix;
        stream << prefix << "iso_dic_trans=" << Translator::iso_dic_trans << suffix;
        stream << prefix << "tr_mode=" << Translator::tr_mode << suffix;
        stream << prefix << "tr_language=" << Translator::tr_language << suffix;
        stream << prefix << "server_mode=" << Translator::server_mode << suffix;
        stream << prefix << "port=" << Translator::port << suffix;
        stream << prefix << "n_gram_length=" << Translator::n_gram_length << suffix;
        stream << prefix << "n_best_length=" << Translator::n_best_length << suffix;
        stream << prefix << "use_tr_server=" << Translator::use_tr_server << suffix;
        stream << prefix << "tr_server_language=" << Translator::tr_server_language << suffix;
        stream << prefix << "tr_server_host=" << Translator::tr_server_host << suffix;
        stream << prefix << "tr_server_port=" << Translator::tr_server_port << suffix;
        stream << prefix << "randomization_factor=" << Translator::randomization_factor << suffix;
        stream << prefix << "max_patterns=" << Translator::max_patterns << suffix;
        stream << prefix << "beam_width_reduced=" << Translator::beam_width_reduced << suffix;
//        stream << prefix << "multiply_beam_width_by_pattern_size=" << Translator::multiply_beam_width_by_pattern_size << suffix;
//        stream << prefix << "filter_additionals=" << Translator::filter_additionals << suffix;
        stream << prefix << "no_keep_input_order_of_additionals=" << Translator::no_keep_input_order_of_additionals << suffix;
//        stream << prefix << "average_lm" << Translator::average_lm << suffix;
//        stream << prefix << "beam_width_final_multiplier" << Translator::beam_width_final_multiplier << suffix;

        stream << prefix << "FLAGS_minloglevel=" << FLAGS_minloglevel << suffix;
        stream << prefix << "FLAGS_logtostderr=" << FLAGS_logtostderr << suffix;
        stream << prefix << "FLAGS_v=" << FLAGS_v << suffix;
        //stream << prefix << "FLAGS_vmodule =" << FLAGS_vmodule  << suffix;
        stream << prefix << "FLAGS_log_dir=" << FLAGS_log_dir << suffix;
        stream << prefix << "FLAGS_stderrthreshold=" << FLAGS_stderrthreshold  << suffix;
        return stream;
    }


    void print_basic_usage(std::ostream* stream) {
        *stream << "********************************************************************************" << std::endl;
        *stream << "*****         KyotoEBMT: Example-based Machine Translation Decoder         *****" << std::endl;
        *stream << "*****                        Version 0.4 'Bourbon'                         *****" << std::endl;
        *stream << "********************************************************************************" << std::endl;
        *stream << "Usage: $ ./KyotoEBMT [-c CONFIG_FILE] [INPUT_FILE]" << std::endl;
    }

    /* 2 Auxiliary functions for checking input for validity. */

    /* Function used to check that 'opt1' and 'opt2' are not specified
       at the same time. */
    void conflicting_options(const po::variables_map& vm,
                             const char* opt1, const char* opt2) {
        if (vm.count(opt1) && !vm[opt1].defaulted() &&
                vm.count(opt2) && !vm[opt2].defaulted()) {
            throw std::logic_error(std::string("Conflicting options '")
                    + opt1 + "' and '" + opt2 + "'.");
        }
    }

    /* Function used to check that of 'for_what' is specified, then
       'required_option' is specified too. */
    void option_dependency(const po::variables_map& vm,
            const char* for_what, const char* required_option) {
        if (vm.count(for_what) && !vm[for_what].defaulted()) {
            if (vm.count(required_option) == 0 || vm[required_option].defaulted()) {
                throw std::logic_error(std::string("Option '") + for_what
                        + "' requires option '" + required_option + "'.");
            }
        }
    }

    bool add_input_filenames_from_file(std::string filename) {
        std::ifstream is(filename.c_str());
        if (is.bad() || is.eof()) return false;
        std::string line;
        Translator::input_filename_list = std::vector<std::string>();
        while (std::getline(is, line)) {
            if (line.size() > 0 && line[line.size()-1] == '\n') {
                line.resize(line.size()-1);
            }
            if (line.size() > 0) {
                Translator::input_filename_list.push_back(line);
            }
        }
        return true;
    }

    void parse_options(int argc, char* argv[]) {
        // NB : the logger is not expected to be initialized when this function is called.
        // Therefore, no message can be logged during this function call. (send everything to stderr...)
        po::options_description command_line_only_options("Command Line-specific Options");
        command_line_only_options.add_options()
            ("help,h", "List command-line options.")
            ("version,v", "Show current version.")
            ("config,c", po::value<std::string>(&Global::config_filename)->default_value(EBMT_INI), "Specify configuration file.")
            ("nb_threads", po::value<size_t>(&Translator::nb_threads)->default_value(1), "Specify number of threads to use.")
            ("display_params", "Just display the list of the set parameters and exit.")
            ("glog.minloglevel", po::value<int>(&FLAGS_minloglevel)->default_value(1), "Specify Logging Level (0:INFO, 1:WARNING, 2:ERROR, 3:FATAL).")
            ("glog.use_file", po::bool_switch()->default_value(false), "Send log to file instead of stderr.")
            ("glog.log_dir", po::value<std::string>(&FLAGS_log_dir)->default_value(""), "Specify log file.")
            ("glog.v", po::value<int>(&FLAGS_v)->default_value(0), "Specify verbosity for VLOG messages (0: less verbose).")
            ("glog.stderrthreshold", po::value<int>(&FLAGS_stderrthreshold)->default_value(2), "Specify threshold for sending to stderr even when logging to a file.")
            ("display_progress", po::bool_switch(&Translator::display_progress)->default_value(false), "Display translation progress on stderr (primitive and not totally thread safe: use only as rough indication).")
        ; // NOLINT

        po::options_description tr_options("TR-related Options");
        tr_options.add_options()
            ("tr_result_filename", po::value<std::string>(&Translator::tr_result_filename), "Specify tr_result filename.")
            ("tr_db_filenames", po::value<std::vector<std::string> >(&Translator::tr_db_filename_list), "Specify treelet database files.")
            ("tr_language", po::value<std::string>(&Translator::tr_language)->default_value("1"), "Specify translation direction of the TR.")
            ("use_tr_server", po::bool_switch(&Translator::use_tr_server), "Use TR server.")
            ("tr_server_language", po::value<std::string>(&Translator::tr_server_language)->default_value("1"), "Specify translation direction of the TR server.")
            ("tr_server_host", po::value<std::string>(&Translator::tr_server_host)->default_value("localhost"), "Specify URL of the TR server.")
            ("tr_server_port", po::value<std::string>(&Translator::tr_server_port)->default_value("12223"), "Specify port of the TR server.")
            ("tr_fast_siblings", po::bool_switch(&Translator::tr_fast_siblings)->default_value(false), "Faster sibling retrieving.")
            ("tr_max_nb_tms_per_pattern", po::value<int>(&Translator::tr_max_nb_tms_per_pattern)->default_value(200), "Specify max number of TMs loaded per pattern (-1 for no limit).")
            ("use_wildcard_pair_cache", po::bool_switch(&Translator::use_wildcard_pair_cache)->default_value(false), "Active cache of wildcard pairs (faster but more memory use).")
            ("no_siblings_retrieval", po::bool_switch(&Translator::no_siblings_retrieval)->default_value(false), "Disable siblings (faster).")

            ; // NOLINT

        po::options_description translation_options("Translation Options");
        translation_options.add_options()
            ("input_list", po::value<std::string>(), "Specify a file containing a list of input filenames.")
            ("input", po::value<std::vector<std::string> >(&Translator::input_filename_list), "Specify input filename (several can be used).")
            ("input_mode", po::value<std::string>(&Global::input_mode)->default_value("xml"), "Specify the input mode: xml/xml_nonparsed/plain/hypothesis.")
            ("default_dpnd", po::value<std::string>(&Translator::default_dpnd), "Specify default dependency filename.")
            ("weight,w", po::value<std::string>(&Translator::feature_weight)->default_value("weights.ini"), "Specify features weights filename.")
            ("lm,l", po::value<std::string>(&Translator::lm_filename), "Specify LM filename.")
            ("beam_width,b", po::value<int>(&Translator::beam_width)->default_value(10), "Specify the beam width.")
            ("rerank_beam,r", po::value<int>(&Translator::rerank_beam)->default_value(0), "Specify the max number of translations to be considered (set to n_best_length if lower than n_best_length).")
            ("source_language", po::value<std::string>(&Translator::source_language)->default_value("ja"), "Specify source language.")
            ("target_language", po::value<std::string>(&Translator::target_language)->default_value("en"), "Specify target language.")
            ("dic_words_filename", po::value<std::string>(&Translator::dic_words_filename)->default_value(""), "Specify OOV dictionary filename.")
            ("dic_punc_filename", po::value<std::string>(&Translator::dic_punc_filename)->default_value(""), "Specify punctuation dictionary filename.")
            ("lexical_translation_prob_filename", po::value<std::string>(&Translator::lexical_translation_prob_filename), "Specify lexical translation probability table filename.")
            ("decoder", po::value<std::string>(&Translator::decoder)->default_value("lattice"), "Specify the decoder: currently only 'lattice'.")
            ("server_mode", po::bool_switch(&Translator::server_mode)->default_value(false), "Server mode.")
            ("port", po::value<int>(&Translator::port)->default_value(0), "Specify port in server mode.")
            ("parse_command", po::value<std::string>(&Translator::parse_command), "Specify parsing command (for translating raw sentences).")
            ("no_sibling", po::bool_switch(&Translator::no_sibling)->default_value(false), "Do not use sibling examples for source side.")
            ("ref", po::value<std::string>(&Translator::reference_file), "File containing reference translations (should have exactly one line per input sentence).")
            ("no_keep_input_order_of_additionals", po::bool_switch(&Translator::no_keep_input_order_of_additionals)->default_value(false), "Keep input order of additionals.")
            ("disable_lattice_reduction", po::bool_switch(&Translator::disable_lattice_reduction)->default_value(false), "Disable lattice reduction.")
            ("lattice_beam_pre_truncation", po::value<int>(&Translator::lattice_beam_pre_truncation)->default_value(0), "Activate lattice beam pre truncation.")
            ("use_nbest_parses", po::bool_switch(&Translator::use_nbest_parses)->default_value(false), "Use n-best parses as input.")
            ; // NOLINT

        po::options_description output_options("Output Options");
        output_options.add_options()
            ("output_mode,m", po::value<std::string>(&Global::output_mode)->default_value("tuning"),
                    "Specify the output mode: tuning/eval/tuning-rerank*/hypothesis/hypo+eval/bleu_eval/online_tuning[experimental] (options with * require '--reconstruct_final_hypotheses').")
            ("hypothesis_output_prefix", po::value<std::string>(&Params::Translator::hypothesis_output_prefix)->default_value(""), "Specify output files prefix for hypothesis output mode")
            ("sentence_id", po::value<int>(&Global::sentence_id)->default_value(0), "sentence_id for tuning mode.")
            ("leading_zeros", po::value<int>(&Translator::leading_zeros)->default_value(5), "Number of leading zeros in the hypothesis filenames.")
            ("web", po::bool_switch(&Global::web)->default_value(false), "Web mode.")
            ("web_verbose", po::bool_switch(&Global::web_verbose)->default_value(false), "Web mode with verbose output.")
            ("web_all", po::bool_switch(&Global::web_all)->default_value(false), "Web mode with n-best output.")
            ("web_dir", po::value<std::string>(&Global::web_dir)->default_value("."), "Specify web directory (require --web).")
            ("n_best_length,n", po::value<int>(&Translator::n_best_length)->default_value(10), "Specify n-best list length.")
            ("shuffle_translation_order", po::bool_switch(&Translator::shuffle_translation_order)->default_value(false), "Shuffle the order in which sentences are translated (currently only work with --nb_threads >= 1 and --input_mode hypothesis.")
            ("verbose_features", po::bool_switch(&Translator::verbose_features)->default_value(false), "Output feature names alongside values in tuning mode.")
            ("time_log_file", po::value<std::string>(&Global::time_log_file), "File into which output timings of translations (give filename stderr to print to stderr instead of a file).")
            ("reconstruct_final_hypotheses", po::bool_switch(&Translator::reconstruct_final_hypotheses)->default_value(false), "Reconstruct final old-style hypotheses when using lattice decoding.")
            ; // NOLINT

        // This specify that all remaining arg should be treated as input filenames:
        po::positional_options_description p;
        p.add("input", -1);

        po::options_description cmdline_options;
        cmdline_options.add(command_line_only_options).add(tr_options).add(translation_options).add(output_options);

        po::options_description config_file_options;
        config_file_options.add(tr_options).add(translation_options).add(output_options);

        po::options_description all_options;
        all_options.add(command_line_only_options).add(tr_options).add(translation_options).add(output_options);

        po::variables_map vm;


        // First, we parse the command line.
        try {
            po::store(po::command_line_parser(argc, argv).
                options(cmdline_options).positional(p).run(), vm);

            option_dependency(vm, "web_dir", "web");
            //conflicting_options(vm, "web", "nb_threads"); // web output is not thread-safe for now

            po::notify(vm);
        }
        catch(const boost::program_options::unknown_option &exc) {
            std::cerr << "Command line parsing error: " << exc.what() << "\n";
            std::cerr << "Run with -h or --help for a list of valid usage and options." << std::endl;
            exit(1);
        }

        catch(const std::exception& e) {
            std::cerr << "Unexpected error during command line parsing: " << e.what() << "\n";
            exit(1);
        }

        // If help was requested, we print it and exit the program.
        if (vm.count("help")) {
            print_basic_usage(&std::cerr);
            std::cerr << all_options << "\n";
            exit(0);
        }
        if (vm.count("version")) {
            print_basic_usage(&std::cerr);
            exit(0);
        }

        // Then, we try to open the config file.
        std::ifstream ifs(Global::config_filename.c_str());
        if (!ifs) {
            std::cerr << "Error: Could not open configuration file '" << Global::config_filename<< "'!\n";
            exit(1);
        }

        // Then, we parse the config file.
        try {
            po::store(po::parse_config_file(ifs, config_file_options), vm);
            po::notify(vm);
        }
        catch(const boost::program_options::unknown_option &exc) {
            std::cerr << "Error while parsing config file '" << Global::config_filename<< "':" << exc.what() << "\n";
            std::cerr << "Run with -h or --help for a list of valid usage and options." << std::endl;
            exit(1);
        }
        catch(std::exception *e) {
            std::cerr << "Unexpected error while parsing config file: " << e->what() << "\n";
            exit(1);
        }

        // Some input validation:
        if (!(Global::output_mode == "tuning" || Global::output_mode == "tuning-rerank" || Global::output_mode == "eval" ||
                    Global::output_mode == "hypothesis" || Global::output_mode == "hypo+eval" || Global::output_mode == "bleu_eval" ||
                    Global::output_mode == "online_tuning")) {
            std::cerr << "Options parsing error: output_mode should be one of tuning/tuning-rerank/eval/hypothesis/hypo+eval/bleu_eval/online_tuning, not '" << Global::output_mode << "'!\n";
            exit(1);
        }

        if (Global::output_mode == "bleu_eval" && Translator::reference_file.size() == 0) {
            std::cerr << "Error: bleu_eval output mode requires a reference file to be given with --ref\n";
            exit(1);
        }

        if (Global::output_mode == "online_tuning") {
            if (Translator::reference_file.size() == 0) {
                    std::cerr << "Error: online_tuning mode require to specify a reference file with --ref\n";
                    exit(1);
            }
            if (Global::input_mode != "hypothesis") {
                std::cerr << "Error: online_tuning mode requires hypotheses input mode\n";
                exit(1);
            }
            if (Translator::nb_threads == 0) {
                std::cerr << "Error: online_tuning mode requires nb_threads to be set to >= 1 \n";
                exit(1);
            }
        }

        if (Global::output_mode == "tuning-rerank" && !Translator::reconstruct_final_hypotheses) {
            std::cerr << "Error: Must set --reconstruct_final_hypotheses to use this output mode." << std::endl;
            exit(1);
        }

        if (!(Global::input_mode == "xml" || Global::input_mode == "xml_nonparsed" || Global::input_mode == "plain" || Global::input_mode == "hypothesis")) {
            std::cerr << "Options parsing error: input_mode should be one of xml/xml_nonparsed/plain/hypothesis, not '" << Global::input_mode << "'!\n";
            exit(1);
        }

        // Some input validation:
        if (!(//Translator::decoder == "old" || Translator::decoder == "packed" || Translator::decoder == "lazy" ||
                    Translator::decoder == "lattice")) {
            std::cerr << "Options parsing error: Decoder should be 'lattice', not '" << Translator::decoder << "'!\n";
            exit(1);
        }

        // if input_list was set, replace the list of input files
        if (vm.count("input_list")) {
            bool success = add_input_filenames_from_file(vm["input_list"].as<std::string>());
            if (!success) {
                std::cerr << "Could not load input files list from '" << vm["input_list"].as<std::string>() <<"'.\n";
            }
        }

        bool log_to_file = vm["glog.use_file"].as<bool>();
        FLAGS_logtostderr = !log_to_file;

        if (Translator::rerank_beam == 0) {
            Translator::rerank_beam = Translator::n_best_length;
        }

        if (Global::web) {
            Global::keep_tms_in_memory = true;
        }

        if (Translator::reference_file.size() > 0) {
            Translator::reference_translations = read_reference(Translator::reference_file.c_str());
            if (Translator::reference_translations.size() == 0) {
                std::cerr << "Error: Could not read reference file '" << Translator::reference_file << "'\n";
                exit(1);
            }
        }

        if (vm.count("display_params")) {
            std::cerr << "With the provided options and config file, the parameters will be set in the following way:" << std::endl;
            std::cerr << ParamsInPrintableForm() << std::endl;
            exit(0);
        }
    }
}

