/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __PARTIAL_TRANSLATION__HPP__
#define __PARTIAL_TRANSLATION__HPP__

#include <string>
#include <vector>
#include "string_indexer.hpp"
#include "def.hpp"
#include "tr_interface.hpp"

class WordOrNT {
  public:
    WordOrNT() {}
    static WordOrNT make_ref(WordIndex ref) {
        WordOrNT result;
        result.word_or_ref.ref = ref;
        result.is_ref_flag = true;
        return result;
    }
    static WordOrNT make_string_id(StringIndex string_id) {
        WordOrNT result;
        result.word_or_ref.string_id = string_id;
        result.is_ref_flag = false;
        return result;
    }        
    bool is_ref() const {
        return is_ref_flag;
    }
        
    WordIndex get_ref() const {
        assert(is_ref_flag);
        return word_or_ref.ref;
    }
        
    StringIndex get_string_id() const {
        assert(!is_ref_flag);
        return word_or_ref.string_id;
    }
    
    bool operator<(const WordOrNT& other) const {
        if (is_ref_flag) {
            if (other.is_ref_flag) {
                return word_or_ref.ref < other.word_or_ref.ref;
            } else {
                return true; // ref<string_id
            }
        } else {
            if (!other.is_ref_flag) {
                return word_or_ref.string_id < other.word_or_ref.string_id;
            } else {
                return false; //string_id>ref
            }
        }
    }
    
    int compare3(const WordOrNT& other) const {
        if (is_ref_flag) {
            if (other.is_ref_flag) {
                if (word_or_ref.ref < other.word_or_ref.ref) return -1;
                if (word_or_ref.ref > other.word_or_ref.ref) return 1;
                return 0;
            } else {
                return -1; // ref<string_id
            }
        } else {
            if (!other.is_ref_flag) {
                if (word_or_ref.string_id < other.word_or_ref.string_id) return -1;
                if (word_or_ref.string_id > other.word_or_ref.string_id) return 1;
                return 0;
            } else {
                return 1; //string_id>ref
            }
        }
    }
        
  private:
    union StringRefUnion {
        StringIndex string_id;
        WordIndex ref;
    } word_or_ref;
    bool is_ref_flag;
};

class PartialTranslation {
    friend std::ostream& operator<<(std::ostream& out, const PartialTranslation& partial_translation);

  public:
    // construction from VS
    // PartialTranslation(const VS& t_string, StringIndexerPtr indexer);
    PartialTranslation() {
        indexer_ = new StringIndexer;
    };
    /*
    // construction by expanding one partial translation into another
    PartialTranslation(const PartialTranslation& partial_trans_outside, 
                       const PartialTranslation& partial_trans_inside,
                       size_t                    insertion_point,
                       bool                      replace_element);
    */
    
    int find_compatible_ref_id(const PatternType &list) const {
        for (int i = sequence_.size() - 1; i >= 0; i--) {
            if (sequence_[i].is_ref() && list.count(sequence_[i].get_ref()) != 0) {
                return i;
            }
        }
        return -1;
    }

    inline bool is_incomplete() const {
        for (size_t j = 0; j < sequence_.size(); j++) {
            if (sequence_[j].is_ref()) {
                return true;
            }
        }
        return false;
    }  
    
    size_t size() const {
        return sequence_.size();
    }
    
    std::string get_word_as_string(WordOrNT word) const {
        if (word.is_ref()) {
            std::string ref_string = "[X"+boost::lexical_cast<std::string>(word.get_ref())+"]";
            return ref_string;
        } else {
            return indexer_->getLabel(word.get_string_id());
        }    
    }
    
    VS to_VS() const {
        VS result;
        BOOST_FOREACH(WordOrNT word, sequence_) {
            if (word.is_ref()) {
                std::string ref_string = "[X"+boost::lexical_cast<std::string>(word.get_ref())+"]";
                result.push_back(ref_string);
            } else {
                result.push_back(indexer_->getLabel(word.get_string_id()));
            }
        }
        return result;
    }
    
    std::string operator[](int i) const {
        return get_word_as_string(sequence_[i]);
    }
    void hash(std::string *result) const;
    
    // bool operator<(const PartialTranslation& other) {
    // assert(&indexer==&(other.indexer));
    // if (sequence.size() < other.sequence.size()) return true;
    // if (sequence.size() > other.sequence.size()) return false;
    // for (size_t i=0; i<sequence.size(); i++) {
    // if (sequence[i]<other.sequence[i]) return true;
    // if (sequence[i]>other.sequence[i]) return false;
    // }
    // return false; //identical
    // }
    
    int compare3(const PartialTranslation& other) const {
        //assert(indexer.get()==other.indexer.get());
        if (sequence_.size() < other.sequence_.size()) return -1;
        if (sequence_.size() > other.sequence_.size()) return 1;
        for (size_t i = 0; i < sequence_.size(); i++) {
            int compare_int = sequence_[i].compare3(other.sequence_[i]);
            if (compare_int != 0) return compare_int;
        }
        return 0; //identical
    }
    std::vector<WordOrNT> &get_t_string_ref() {return sequence_;}

  private:
    std::vector<WordOrNT> sequence_;
    std::vector<WordIndex> dpnd_;
    StringIndexerPtr      indexer_;
};
std::ostream& operator<<(std::ostream& out, const PartialTranslation& partial_translation);

class PartialTranslationIndexedFactory {
  public:
    PartialTranslationIndexedFactory() : indexer(new StringIndexer) {
        assert(indexer);
    }
    PartialTranslation make_partial_translation(const VS& t_string) {
        return PartialTranslation(t_string, indexer);
    }
  private:
    StringIndexerPtr indexer;
};
#endif //__PARTIAL_TRANSLATION__HPP__

