/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include <boost/foreach.hpp>
#include <iostream>
#include <vector>
#include <string>
#include <set>
#include <cassert>
#include "def.hpp"

#include "input_insertion_id.hpp"

#ifndef __PARTIAL_TRANSLATION_STRING__HPP__
#define __PARTIAL_TRANSLATION_STRING__HPP__

class PartialTranslationString : public VS {
  public:
    friend std::ostream& operator<<(std::ostream& out,
            const PartialTranslationString& partial_translation);
    PartialTranslationString() : VS() {}
    explicit PartialTranslationString(const VS& other) : VS(other) {}
    template<typename IT1, typename IT2>
    PartialTranslationString(IT1 it1, IT2 it2) : VS(it1, it2) {}
    PartialTranslationString(const PartialTranslationString& outside,
                             const PartialTranslationString& inside,
                             int insertion_point, bool replace);

    int find_compatible_ref_id(const PatternType &list) const;

    int find_pos_of_ref_id(InputInsertionId requested_ref) const {
        for (int i = size() - 1; i >= 0; i--) {
            InputInsertionId ref_id = InputInsertionId::get_ref_id((*this)[i]);
            if (ref_id.is_valid() && ref_id == requested_ref) {
                return i;
            }
        }
        return -1;
    }

    
    int get_target_nb_of_words() const {
        int res = 0;
        BOOST_FOREACH(const std::string& word, *this) {
            if (!(InputInsertionId::get_ref_id(word).is_valid())) {
                res += 1;
            }
        }
        return res;
    }

    int get_nb_of_optional_words() const {
        int res = 0;
        BOOST_FOREACH(const std::string& word, *this) {
            if (!(InputInsertionId::get_ref_id(word).is_valid())) {
                if (is_word_optional(word)) {
                    res += 1;
                }
            }
        }
        return res;
    }

    inline bool is_incomplete() const {
        for (size_t j = 0; j < size(); j++) {
            if (InputInsertionId::get_ref_id((*this)[j]).is_valid()) {
                return true;
            }
        }
        return false;
    }
    
    static bool is_word_optional(const std::string& word) {
        return word.size() >= 2 && word[word.size()-1] == '*';
    }

    
    bool has_optional_words() const {
        BOOST_FOREACH(const std::string& word, *this) {
            if (is_word_optional(word)) {
                return true;
            }
        }
        return false;
    }
    

    const VS& to_VS() const {
        return *this;
    }

    const std::string to_string_no_asterisks() const {
        std::string result = "";
        for (size_t i = 0; i < size(); ++i) {
            const std::string &part = (*this)[i];
            if (part[part.size() - 1] == '*') {
                result += part.substr(0, part.size() - 1) + " ";
            } else {
                result += part + " ";
            }
        }
        return result;
    }
    
    PartialTranslationString& operator=(const VS& other) {
        //if (this == &other) return *this;
        VS& this_as_VS = *this;
        this_as_VS = other;
        return *this;
    }

    int compare3(const PartialTranslationString& other) const {
        //assert(indexer.get() == other.indexer.get());
        if (size() < other.size()) return -1;
        if (size() > other.size()) return 1;
        for (size_t i = 0; i < size(); i++) {
            int compare_int = (*this)[i].compare(other[i]);
            if (compare_int != 0) {
                return (compare_int < 0) ? -1 : 1;
            }
        }
        return 0; // identical
    }

    std::set<WordIndex> get_set_of_ref() const {
        std::set<WordIndex> result;
        BOOST_FOREACH(const std::string& w, *this) {
            InputInsertionId idx = InputInsertionId::get_ref_id(w);
            if (idx.is_valid()) {
                idx.add_refs_to_set(&result);
            }
        }
        return result;
    }

    bool check_no_duplicate_ref() const {
        std::set<WordIndex> set_of_ref;
        BOOST_FOREACH(const std::string& w, *this) {
            InputInsertionId idx = InputInsertionId::get_ref_id(w);
            if (idx.is_valid()) {
                if (idx.has_intersection_with_set(set_of_ref)) return false;
                idx.add_refs_to_set(&set_of_ref);
            }
        }
        return true;
    }

    void hash(std::string *result) const;
};

std::ostream& operator<<(std::ostream& out, const PartialTranslationString& partial_translation);

class PartialTranslationStringFactory {
  public:
    PartialTranslationString make_partial_translation(const VS& t_string) const {
        return PartialTranslationString(t_string);
    }
};

#endif

