/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdlib.h>
#include <string.h>
#include <glog/logging.h>
#include <boost/foreach.hpp>
#include <boost/algorithm/string.hpp>
#include <string>
#include <vector>
#include <map>
#include <set>
#include <functional>
#include <iomanip>
#include <limits> 

#ifndef __SERIALIZATION_HELPERS_HPP__
#define __SERIALIZATION_HELPERS_HPP__

/*******************************************
 *  Some utility functions
 */

inline void serialize_dpnd_vector(const DpndList &vect, char delimiter, std::ostream& stream) {
    for (size_t i = 0; i < vect.size(); i++) {
        stream << vect[i]; 
        if ((i + 1) < vect.size()) {
            stream << delimiter; 
        }
    }
}

inline void serialize_wordindex_vector(const std::vector<WordIndex> &vect, char delimiter, std::ostream& stream) {
    for (size_t i = 0; i < vect.size(); i++) {
        stream << static_cast<int>(vect[i]); 
        if ((i + 1) < vect.size()) {
            stream << delimiter; 
        }
    }
}

template<typename Container>
void serialize_wordindex_container(const Container &vect, char delimiter, std::ostream& stream) {
    size_t cursor = 0;
    BOOST_FOREACH(WordIndex idx, vect) {
        stream << static_cast<int>(idx);
        if (cursor + 1 < vect.size()) stream << delimiter; 
        cursor++;
    }
}

inline void deserialize_dpnd_vector(const std::string& input_string, char delimiter, DpndList *result) {
    VLOG(2) << "deserializing dpnd_vector:'" << input_string << "' delim:'" << delimiter << "'" << std::endl;
    std::vector<std::string> vector_of_strings;
    boost::split(vector_of_strings, input_string, std::bind1st(std::equal_to<char>(), delimiter), boost::token_compress_on);
    VLOG(2) << "deserializing dpnd_vector: splitted into " << vector_of_strings.size()  << std::endl;
    BOOST_FOREACH(const std::string& int_as_string, vector_of_strings) {
        VLOG(2) << "deserializing int_as_string:'" << int_as_string << "'" << std::endl;
        result->push_back(atoi(int_as_string.c_str()));
    }
}

inline void deserialize_wordindex_vector(const std::string& input_string, char delimiter, std::vector<WordIndex> *result) {
    VLOG(2) << "deserializing wordindex_vector:'" << input_string << "' delim:'" << delimiter << "'" << std::endl;
    std::vector<std::string> vector_of_strings;
    boost::split(vector_of_strings, input_string, std::bind1st(std::equal_to<char>(), delimiter), boost::token_compress_on);
    VLOG(2) << "deserializing wordindex_vector: splitted into " << vector_of_strings.size()  << std::endl;
    BOOST_FOREACH(const std::string& int_as_string, vector_of_strings) {
        VLOG(2) << "deserializing int_as_string:'" << int_as_string << "'" << std::endl;
        result->push_back((WordIndex) atoi(int_as_string.c_str()));
    }
}


/* These should work on utf8 sequences.
 * Apparently, utf8 multibyte sequences can not contain ascii (code <=127) char
 */

inline void escape_string(const std::string& string, std::string* out) {
     if (string.size() == 0) {
        out->push_back('\\');
        out->push_back('0');
        return;
     }

    out->reserve(string.size());
    
    BOOST_FOREACH(char c, string) {
        switch (c) {
            case '\n':
            out->push_back('\\');
            out->push_back('n');
            break;
            
            case '\\':
            out->push_back('\\');
            out->push_back('n');
            break;         
            
            case ':':
            out->push_back('\\');
            out->push_back('c');
            break;

            case '|':
            out->push_back('\\');
            out->push_back('p');
            break;  

            case ' ':
            out->push_back('\\');
            out->push_back('_');
            break;  

            case '\0':
            out->push_back('\\');
            out->push_back('z');
            break;  
            
            default:
            out->push_back(c);
            break;
        }   
    }
}

inline void de_escape_cstring(const char* string, std::string* out) {
    size_t cursor = 0;
    while (string[cursor] != '\0') {
        char c = string[cursor];
        if (c != '\\') {
            out->push_back(c);
        } else {
            LOG_IF(FATAL, string[cursor+1] == '\0') << "Error de-escaping string :" << string << "\n";
            cursor++;
            switch (string[cursor]) {
                case 'n':
                out->push_back('\n');
                break;
                
                case '\\':
                out->push_back('\\');
                break; 

                case 'c':
                out->push_back(':');
                break; 

                case 'p':
                out->push_back('|');
                break;

                case '_':
                out->push_back(' ');
                break; 

                case 'z':
                out->push_back('\0');
                break; 
                
                case '0':  //case of empty string
                break;
                
                default:
                LOG(FATAL) << "Error de-escaping string :" << string << "\n";
                break;
            }
        }
        cursor++;
    }
}

inline void de_escape_string(const std::string& string, std::string* out) {
    out->reserve(string.size());
    size_t cursor = 0;
    while (cursor < string.size()) {
        char c = string[cursor];
        if (c != '\\') {
            out->push_back(c);
        } else {
            LOG_IF(FATAL, cursor + 1 >= string.size()) << "Error de-escaping string :" << string << "\n";
            cursor++;
            switch (string[cursor]) {
                case 'n':
                out->push_back('\n');
                break;
                
                case '\\':
                out->push_back('\\');
                break; 

                case 'c':
                out->push_back(':');
                break; 

                case 'p':
                out->push_back('|');
                break;

                case '_':
                out->push_back(' ');
                break; 

                case 'z':
                out->push_back('\0');
                break; 
                
                case '0':  //case of empty string
                break;
                
                default:
                LOG(FATAL) << "Error de-escaping string :" << string << "\n";
                break;
            }
        }
        cursor++;
    }
}

inline void de_escape_string_inplace(std::string* pstring) {
    std::string& string = *pstring;
    size_t cursor = 0;
    while (cursor < string.size()) {
        if (string[cursor] == '\\') {
            LOG_IF(FATAL, cursor + 1 >= string.size()) << "Error de-escaping string :" << string << "\n";
            string.erase(cursor, 1);
            switch (string[cursor]) {
                case 'n':
                string[cursor] = '\n';
                break;
                
                case '\\':
                string[cursor] = '\\';
                break; 

                case 'c':
                string[cursor] = ':';
                break; 

                case 'p':
                string[cursor] = '|';
                break;

                case '_':
                string[cursor] = ' ';
                break; 

                case 'z':
                string[cursor] = '\0';
                break; 
                
                case '0':  //case of empty string
                string.erase(cursor, 1);
                break;
                
                default:
                LOG(FATAL) << "Error de-escaping string :" << string << "\n";
                break;
            }
        }
        cursor++;
    }
}


/*! my_strtok_r behave similarly to the standard C function strtok_r, with a few differences:
 *  Mainly, 3 differences: 1- sequences of delimiters are not seen as one delimiter, 2- rest is set to NULL when we reach the end-of-string, 3- the delimiter parameter is of type char instead of char* 
 * (I had originally misunderstood the behaviour of strtok_r, so I implemented a version that corresponds to my expectations)
 */
inline char* my_strtok_r(char* string, char delimiter, char** rest) {
    char *cursor;
    if (string != NULL) {
        cursor = string;
    } else {
        if (*rest == NULL) {
            DVLOG(3) << "my_strtok_r return NULL\n";
            return NULL;
        }
        cursor = *rest;
    }
    DVLOG(3) << "my_strtok_r tokenize '" << cursor << "'\n";
    char* token = cursor;
    while (*cursor != delimiter && *cursor != '\0') {
        cursor++;
    }
    if (*cursor != '\0') {
        *cursor = '\0';
        *rest = cursor + 1;
        DVLOG(3) << "my_strtok_r return '" << token << "'\n";
        return token;
    } else { // reached end of string
        *rest = NULL;
        DVLOG(3) << "my_strtok_r reached end of sring and return '" << token << "'\n";
        return token;
    }
}
 
inline int read_one_int_delimited_by(char* string, char delimiter, char** rest) {
    DVLOG(3) << "read int from '" << (string?string:*rest) << "'.\n";
    char* token = my_strtok_r(string, delimiter, rest);
    LOG_IF(FATAL, token == NULL) << "Error Parsing int" << "\n";
    return atoi(token);
}

inline std::string read_one_string_delimited_by(char* string, char delimiter, char** rest) {
    DVLOG(3) << "read string from '" << (string?string:*rest) << "'.\n";
    char* token = my_strtok_r(string, delimiter, rest);
    LOG_IF(FATAL, token == NULL) << "Error Parsing string" << "\n";
    return token;
}



inline void read_wordindex_vector(char* string, char delimiter, std::vector<WordIndex>* result) {
    DVLOG(3) << "read vect from '" << string << "'.\n";
    if (*string == '\0') return; //empty vector
    char* rest = NULL;
    char* token = my_strtok_r(string, delimiter, &rest);
    result->push_back(static_cast<WordIndex>(atoi(token)));
    while ((token = my_strtok_r(NULL, delimiter, &rest))) {
        result->push_back(static_cast<WordIndex>(atoi(token)));   
    }
}

inline void read_wordindex_set(char* string, char delimiter, std::set<WordIndex>* result) {
    DVLOG(3) << "read set from '" << string << "'.\n";
    if (*string == '\0') return; //empty set
    char* rest = NULL;
    char* token = my_strtok_r(string, delimiter, &rest);
    //if (token == NULL) return;
    result->insert(static_cast<WordIndex>(atoi(token)));
    while ((token = my_strtok_r(NULL, delimiter, &rest))) {
            result->insert(static_cast<WordIndex>(atoi(token)));   
    }
}

#endif //__SERIALIZATION_HELPERS_HPP__
