/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include <boost/intrusive_ptr.hpp>
#include <boost/foreach.hpp>
#include <map>
#include <string>
#include <vector>

#ifndef __STRING_INDEXER__HPP__
#define __STRING_INDEXER__HPP__

class IndexerException: public std::exception {
    virtual const char* what() const throw() {
        return "IndexerException";
    }
};

template<typename CodeType> 
class GenericStringIndexerWithRefCount;

namespace boost {
    template<typename CodeType>
    void intrusive_ptr_add_ref(GenericStringIndexerWithRefCount<CodeType> * p);
    template<typename CodeType>
    void intrusive_ptr_release(GenericStringIndexerWithRefCount<CodeType> * p);
};

template<typename CodeType> 
class GenericStringIndexerWithRefCount {
    std::vector<std::string> idToLabel;
    std::map<std::string, CodeType> labelToId;

    template<CodeType> 
    friend std::ostream &operator<< (std::ostream &stream, const GenericStringIndexerWithRefCount<CodeType> &li);

  public:
    size_t refcount;
    // template<CodeType> 
    // friend void ::boost::intrusive_ptr_add_ref(GenericStringIndexerWithRefCount<CodeType> * p);
    // template<CodeType> 
    // friend void ::boost::intrusive_ptr_release(GenericStringIndexerWithRefCount<CodeType> * p);
    
  public:
    const std::vector<std::string>& getIdToLabelVector() const {
        return idToLabel;
    }

    const std::string& getLabel(CodeType id) const {
        if (id > idToLabel.size()) {
            throw IndexerException();
        }
        return idToLabel[id];    
    }
        
    CodeType getId(const std::string& label) const {
        typename std::map<std::string, CodeType>::const_iterator it = labelToId.find(label);
        if (it == labelToId.end()) {
            throw IndexerException();
        }
        return it->second;
    }

    CodeType getIdAndUpdate(const std::string& label) {
        typename std::map<std::string, CodeType>::iterator it = labelToId.find(label);
        CodeType code;
        if (it == labelToId.end()) {
            idToLabel.push_back(label);
            code = idToLabel.size() - 1;
            labelToId[label] = code;
            // std::cout<<"not found "<<label<<" new id: "<<id<<std::endl;
        } else {
            // std::cout<<"found "<<label<<" id: "<<it->second<<std::endl;
            code= it->second;
        }
        return code;
    }
        
    std::vector<CodeType> convertVectorOfLabel(const std::vector<std::string>& vLabel) const {
        std::vector<CodeType> result;
        BOOST_FOREACH(const std::string& label, vLabel) {
            result.push_back(getId(label));
        }
        return result;
    }

    std::vector<CodeType> convertVectorOfLabelAndUpdate(const std::vector<std::string>& vLabel) {
        std::vector<CodeType> result;
        BOOST_FOREACH(const std::string& label, vLabel) {
            result.push_back(getIdAndUpdate(label));
        }
        return result;
    }        
};

    
template<typename CodeType>
std::ostream &operator<<(std::ostream &stream, const GenericStringIndexerWithRefCount<CodeType> &li) {
    stream << li.labelToId;
    return stream;
}

namespace boost {
    // NB: should be made critical in case of multithreading ...
    template<typename CodeType>
    inline void intrusive_ptr_add_ref(GenericStringIndexerWithRefCount<CodeType> * p) {
        // increment reference count of object *p
        ++(p->refcount);
    }
    template<typename CodeType>
    inline void intrusive_ptr_release(GenericStringIndexerWithRefCount<CodeType> * p) {
        // decrement reference count, and delete object when reference count reaches 0
        if (--(p->refcount) == 0) {
            delete p;
            p = NULL;
        }
    } 
} 


// class StringIndex {
    // unsigned int idx;
    // StringIndex(unsigned int i):idx(i){}
    // operator unsigned int() {
        // return idx;
    // }

// };
typedef unsigned int StringIndex;
typedef GenericStringIndexerWithRefCount<StringIndex> StringIndexer;
typedef boost::intrusive_ptr<StringIndexer> StringIndexerPtr;


#endif //__STRING_INDEXER__HPP__
