/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef TABLE_DISPLAY_OUTPUT
#define TABLE_DISPLAY_OUTPUT

#include <boost/lexical_cast.hpp>
#include <boost/regex.hpp>
#include <boost/range/adaptor/map.hpp>
#include <boost/foreach.hpp>
#include <iostream>
#include <map>
#include <set>
#include <sstream>
#include <fstream>
#include <string>
#include <utility>
#include <vector>
#include "params.hpp"
#include "tm.hpp"
#include "hypothesis.hpp"
#include "confidence_estimator.hpp"
#include "utils.hpp"

typedef std::pair<std::set<WordIndex>, std::set<WordIndex> > MapPair;
typedef std::map<const size_t, MapPair> InOutMapper;

class TableDisplayColor {
  public:
    TableDisplayColor() {
        table_display_color.push_back("Crimson");
        table_display_color.push_back("Blue");
        table_display_color.push_back("Green");
        table_display_color.push_back("Orange");
        table_display_color.push_back("Red");
        table_display_color.push_back("DarkSlateGray");
        table_display_color.push_back("Violet");
        table_display_color.push_back("Brown");
        table_display_color.push_back("Indigo");
        table_display_color.push_back("Lime");
    }
    std::string color(const int i) const {
        if (i < 0) {
            return "Black";
        } else {
            return table_display_color[i % 10];
        }
    }
  private:
    VS table_display_color;
};

class TableDisplayOutput {
  public:
    TableDisplayOutput() : output_stream_(NULL), output_string_stream_(NULL), table_(1), row_(1), col_(1), input_tree_(NULL), translation_confidence_(1.0) {}
    ~TableDisplayOutput() {
        if (output_stream_) {
            delete output_stream_;
            output_stream_ = NULL;
        }
    }

    void set_table(unsigned int t) {table_ = t;}
    void set_row(unsigned int r) {row_ = r;}
    void set_col(unsigned int c) {col_ = c;}
    void set_input_tree(Dtree* input_tree) {input_tree_ = input_tree;}
    void next_table() {
        table_++; 
        row_ = 1; 
        col_ = 1;
    }
    void next_row() {
        row_++; 
        col_ = 1;
    }
    void next_col() {col_++;}
    unsigned int table() const {return table_;}
    unsigned int row() const {return row_;}
    unsigned int col() const {return col_;}
    std::string color(const int i) const {return td_color_.color(i);}

    
    void initialize(const std::string& id, const Dtree* input_tree) {
        open(id + ".txt");
        println("%% title=Translation result ("+id+")");
        // const cast is a bit ugly, but since TableDisplayOutput modify the tree, 
        // there is no choice for now
        set_input_tree(const_cast<Dtree*>(input_tree));
    }
    
    void open(const std::string &filename) {
        if (Params::Translator::server_mode) {
            if (output_string_stream_) {
                delete output_string_stream_;
                output_string_stream_ = NULL;
                output_stream_ = NULL;
            }
            output_string_stream_ = new std::ostringstream();
            output_stream_ = output_string_stream_;
        } else {
            if (output_stream_) {
                delete output_stream_;
                output_stream_ = NULL;
            }
            if (Params::Global::web_dir == "")
                Params::Global::web_dir = ".";
            std::string file_path = Params::Global::web_dir+"/"+filename;
            output_stream_ = new std::ofstream(file_path.c_str());
        }
        table_ = 1;
        row_ = 1;
        col_ = 1;
    }

    void print(const std::string &str) const {
        if (output_stream_ == NULL) return;
        *output_stream_ << str << std::endl;
    }
    void print(const std::string &str, const unsigned int table, const unsigned int row, const unsigned int col, const std::string option = "") const {
        if (output_stream_ == NULL) return;
        *output_stream_ << "%% " << table << " " << row << " " << col << " " << option << std::endl;
        print(str);
    }

    void println(const std::string &str) {
        if (output_stream_ == NULL) return;
        print(str+"<br>");
    }
    void println(const std::string &str, const unsigned int table, const unsigned int row, const unsigned int col, const std::string option = "") const {
        if (output_stream_ == NULL) return;
        print(str+"<br>", table, row, col, option);
    }

    void print_tm(const int index, const TM* tm, const DataForTM *data_for_tm);

    void output_for_table_display(int m, const SingleNBestTranslationResult& tl_result);

    void print_buffer(std::ostream* stream) {
        if (output_stream_ != NULL && Params::Translator::server_mode) {
            std::string input_sentence = input_tree_->sentence();
            *stream << "# segmented input sentence start" << std::endl;
            convert_bracket_symbols(input_sentence);
            LOG(INFO) << input_sentence << "\n";
            *stream << input_sentence << std::endl;
            //LOG(INFO) << input_tree_->sentence() << "\n";
            //*stream << input_tree_->sentence() << std::endl;
            *stream << "# segmented input sentence end" << std::endl;
            *stream << "# in-out mapping start" << std::endl;
            BOOST_FOREACH(const MapPair v, in_out_mapper_ | boost::adaptors::map_values) {
                for (std::set<WordIndex>::iterator w = v.first.begin(); w != v.first.end(); ) {
                    *stream << *w;
                    w++;
                    if (w != v.first.end()) {
                        *stream << ",";
                    }
                }
                *stream << "-";
                for (std::set<WordIndex>::iterator w = v.second.begin(); w != v.second.end(); ) {
                    *stream << *w;
                    w++;
                    if (w != v.second.end()) {
                        *stream << ",";
                    }
                }
                *stream << " ";
            }
            *stream << std::endl;
            *stream << "# in-out mapping end" << std::endl;
            *stream << "# translation confidence start" << std::endl;
            *stream << translation_confidence_ << std::endl;
            *stream << "# translation confidence end" << std::endl;
            *stream << "# web output start" << std::endl;
            *stream << output_string_stream_->str();
            *stream << "# web output end" << std::endl;
        }
    }

  private:
    std::ostream *output_stream_;
    std::ostringstream* output_string_stream_;
    unsigned int table_;
    unsigned int row_;
    unsigned int col_;
    TableDisplayColor td_color_;
    Dtree *input_tree_;
    InOutMapper in_out_mapper_;
    ConfidenceEstimator confidence_estimator_;
    double translation_confidence_;
};


#endif
