/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*! @file
  This file contains the definitions of TM class and DataForTM class
*/
#ifndef TM_INCLUDED
#define TM_INCLUDED

#include <boost/serialization/vector.hpp>
#include <boost/utility.hpp>
#include <map>
#include <set>
#include <string>
#include <vector>
#include "para_sentence.hpp"
#include "alignment.hpp"
#include "def.hpp"

/*! This structure contains additional info for TM which are used in
 *  formatting the TM. The info of root word and bond words are stored
 *  in this class. */
struct DataForTM {
  public:
    //! default constructor
    DataForTM() :
        s_bond(NULL), t_bond(NULL), t_lm_check(NULL), match_i_of_t(NULL), additional_for_t(NULL),
        i_root(NULL), s_root(NULL), t_root(NULL), s_bond_root(NULL), t_bond_root(NULL), ref_count_(1) {}

    //! constructor with source and target size. This constructor
    //! allocates memory for several members (will be deleted in the
    //! destructor).
    DataForTM(size_t s_size, size_t t_size) :
        i_root(NULL), s_root(NULL), t_root(NULL), s_bond_root(NULL), t_bond_root(NULL), s_size_(s_size), t_size_(t_size), ref_count_(1) {
        s_bond = new int[s_size];
        t_bond = new int[t_size];
        t_lm_check = new int[t_size];
        match_i_of_t = new PatternType[t_size];
        additional_for_t = new PatternType[t_size];        
        for (size_t i = 0; i < s_size; i++) {
            s_bond[i] = -1;
        }
        for (size_t i = 0; i < t_size; i++) {
            t_bond[i] = -1;
            t_lm_check[i] = 0;
        }
    }

    //! destructor which frees allocated memory
    ~DataForTM() {
        if (s_bond != NULL) {
            delete[] s_bond;
            s_bond = NULL;
        }
        if (t_bond != NULL) {
            delete[] t_bond;
            t_bond = NULL;
        }
        if (t_lm_check != NULL) {
            delete[] t_lm_check;
            t_lm_check = NULL;
        }
        if (match_i_of_t != NULL) {
            delete[] match_i_of_t;
            match_i_of_t = NULL;
        }
        if (additional_for_t != NULL) {
            delete[] additional_for_t;
            additional_for_t = NULL;
        }
    }

    //! list of input words which are covered by this TM
    std::vector<const DtreeNode*> i_body_ref;
    //! list of source-side words which covers input treelet
    std::vector<const DtreeNode*> s_body_ref;
    //! list of target-side words which will be output
    std::vector<const DtreeNode*> t_body_ref;

    //! list of child words in the input connected to i_body
    std::vector<const DtreeNode*> i_child_ref;
    //! list of child words in the source-side which are connected to
    //! s_body and candidates of bond
    std::vector<const DtreeNode*> s_child_ref;
    //! list of child words in the source-side which are connected to
    //! s_body but *not* candidates of bond
    std::vector<const DtreeNode*> s_other_child_ref;
    //! s_child_ref + s_other_child_ref
    std::vector<const DtreeNode*> s_child_bond_cand;

    //! if the content of the target word is modified (like numerals),
    //! the info is stored here
    std::map<WordIndex, std::string> t_modified_content;

    /*! the bond info for each word in the source-side of the TM:
     * - -1: nothing to do with the translation
     * - 0: body of the tm
     * - 1: candidate of the bond
     * - 2: bond will actually be used
     */
    int* s_bond;
    /*! the bond info for each word in the target-side of the TM:
     * - -1: nothing to do with the translation
     * - 0: body of the tm
     * - 1: candidate of the bond
     * - 2: bond will actually be used
     */
    int* t_bond;
    /*! info for putting '*' mark indicating the word should be judged
     *  by LM:
     * - 0: not put '*' mark
     * - 1: put '*' mark
     */
    int* t_lm_check;
    //! corresponding input word index for each target word. this is
    //! vector to handle sibling matching
    PatternType* match_i_of_t;
    /*! list of additional words in the input which are not covered by
     *  the TM and do not have bonds available. The translations for
     *  these words will be attached to the specified target word
     *  heuristically, and the ordering between siblings will be
     *  judged by LM. */
    PatternType* additional_for_t;
    PatternType additional_for_empty;
    
    DtreeNode *i_root; //!< the root word of the input treelet
    DtreeNode *s_root; //!< the root word of the source-side of TM
    DtreeNode *t_root; //!< the root word of the target-side of TM
    /*! the bond root word of the source-side of TM. Basically, the
     *  bond root is the parent of the root of TM. If the root of TM
     *  is the root of the sentence, the bond_root is the same to the
     *  root */
    DtreeNode *s_bond_root;
    /*! the bond root word of the target-side of TM. Basically, the
     *  bond root is the parent of the root of TM. If the root of TM
     *  is the root of the sentence, the bond_root is the same to the
     *  root */
    DtreeNode *t_bond_root;

    size_t s_size_;
    size_t t_size_;

    /*! Functions for handling simply multiple ownership of a DataForTM pointer using s ref counter.
     * Warning: not thread safe (but OK to use with the multithread implemenentation of KyotoEBMT, since DataForTM objects are all thread-locals).
     */

    DataForTM* duplicate_with_incref() {
        assert(ref_count_ >= 1);
        ref_count_++;
        return this;
    }

    static void decref_and_maybe_delete(DataForTM* data) {
        assert(data != NULL);
        assert(data->ref_count_ >= 1);
        data->ref_count_--;
        if (data->ref_count_ == 0) {
            delete data;
        }
    }

  private:
    DataForTM(const DataForTM&); // do not allow copying

    int ref_count_; // simple reference count
};

std::ostream &operator<<(std::ostream &os, const DataForTM &t);

/*! This class contains Translation Memory (or simply example). TM is
 *  composed of one ParaSentence, matching info between input and
 *  source-side of ParaSentence, a list of corresponding target-side
 *  words and target-side neighboring words. The neighboring words
 *  will be judged whether they are necessary or not by Language Model
 *  in the end. 
 *  TM is made noncopyable by inheriting from boost::noncopyable. This means that 
 *  if one tries to copy or assign an object of this class, the compiler will complain
 *  with a message containing something like: 
 *  "Error: noncopyable(const noncopyable&) constructor is private" (+ a lot of template instanciation mess)
 */
class TM : public boost::noncopyable {
    friend class TMBuilder;
  public:
    //!  constructor
    explicit TM(ParaSentence &para_sentence) : para_sentence_(para_sentence) {} // NOLINT (because reference in TR++)

    //~TM();

    //! accessor to the para_sentence as const
    const ParaSentence& para_sentence() const {return para_sentence_;}
    //! getter of the id of the tm
    std::string id() const {return para_sentence_.id();}
    //! accessor to the i_s_match as const
    const std::vector<Alignment>& i_s_match() const {return i_s_match_;}
    //! getter of the t_string
    VS t_string() const {
        VS tmp; 
        tmp.push_back(id()); 
        return tmp;
    }
    //! accessor to the connected_image (targe-side body) as const
    const std::vector<WordIndex>& connected_image() const {return connected_image_;}
    //! accessor to the additional_nodes (additional words which will
    //! be judged by LM) as const
    const std::vector<WordIndex>& additional_nodes() const {return additional_nodes_;}
    //! accessor to wildcards_ as const
    const PatternType wildcards() const {return wildcards_;}
         
    void set_alignment_for_OOV(WordIndex position) {
        assert(i_s_match_.size() == 0);
        Alignment alignment;
        alignment.p1.push_back(position);
        alignment.p2.push_back(0);
        alignment.accept = YES; 
        i_s_match_.push_back(alignment);
    }

    void set_connected_image_for_OOV() {
        assert(connected_image_.size() == 0);
        connected_image_.push_back(0);
    }
    
    //! find the Alignment which includes the specified DtreeNode in the
    //! input side of the Alignment
    const Alignment* get_alignment_i(const DtreeNode *w) const;

    //! find the Alignment which includes the specified DtreeNode in the
    //! source side of TM of the Alignment
    const Alignment* get_alignment_s(const DtreeNode *w) const;
    const Alignment* get_alignment_s(const WordIndex& w) const;

    //! find the Alignment which includes the specified DtreeNode in the
    //! source side of TM of the Alignment
    const PatternType get_i_of_t(const DtreeNode *w) const;

    //! this function prints the values of i_s_match (for debugging)
    void print_i_s_match(std::ostream *os) const {
        std::vector<Alignment>::const_iterator it = i_s_match_.begin();
        for (; it != i_s_match_.end(); it++) {
            if (word_overlap(this->wildcards(), it->p1)) {
                *os << "i_s_match: wild ( ";
            } else {
                *os << "i_s_match: ( ";
            }
            for (std::vector<WordIndex>::const_iterator it1 = it->p1.begin(); it1 != it->p1.end(); it1++) {
                *os << *it1 << " ";
            }
            *os << ") ( ";
            for (std::vector<WordIndex>::const_iterator it1 = it->p2.begin(); it1 != it->p2.end(); it1++) {
                *os << *it1 << " ";
            }
            *os << ")" << std::endl;
        }
    }

    //! print dependency tree of the TM and match info between input and source
    void print(std::ostream *os) const;
    void print(std::ostream *os, DataForTM *data_for_tm) const;

    //! small function for sanity check
    //! (used in the construction of tm by the TR code)
    bool is_default_constructed() const {
        return para_sentence_.is_default_constructed() &&
            i_s_match_.size() == 0 &&
            connected_image_.size() == 0 &&
            additional_nodes_.size() == 0;
    }

    //! This is used for the in-place construction of OOV pseudo-TM
    ParaSentence &get_modifiable_para_sentence() {
        // The following assertion is to avoid abuse of this function.
        // In general, we do not want para_sentence to be modifiable.
        // So this function should be called only on uninitialized TM.
        assert(is_default_constructed());
        return para_sentence_;
    }
    
  private:
    ParaSentence           &para_sentence_;  //!< original parallel sentence pair of the TM
    std::vector<Alignment>  i_s_match_;      //!< alignment between input and source-side of TM 

    /*! List of target-side word indexes which will be output in the
      end. When we only search for match with connected image, this is
      the set of target position forming the minimum connected
      image: */
    std::vector<WordIndex> connected_image_;

    /*! List of target-side surrounding NULL-aligned word indexes
      which will be judged if they are necessary by LM. The target
      position in additional_nodes are non-aligned words that can be
      added to the connected image without making it disconnected: */
    std::vector<WordIndex> additional_nodes_;
    PatternType wildcards_;
    // friend class boost::serialization::access;
    // template<class Archive>
    // void serialize(Archive &ar, const unsigned int version) {
    // ar & para_sentence_;
    // ar & i_s_match_;
    // ar & connected_image_;
    // ar & additional_nodes_;
    // }
};
std::ostream &operator<<(std::ostream &os, const TM &t);

#endif
