/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/regex.hpp>
#include <glog/logging.h>
#include <algorithm>
#include <iostream>
#include <fstream>
#include <map>
#include <set>
#include <string>
#include <utility>
#include <vector>
#include "params.hpp"
#include "tm_processor.hpp"

/*
  functions for DefaultDependManager
*/
std::pair<bool, ParentBondRelation> DefaultDependManager::query_parent_bond_relation(const std::string& dpnd_pattern_str,
                                                                                     const POS& pos_of_root) const {
    std::pair<bool, ParentBondRelation> result;
    const_iterator it1 = find(dpnd_pattern_str);
    if (it1 != end()) {
        std::map<std::string, std::string>::const_iterator it2 = it1->second.find(pos_of_root);
        if (it2 != it1->second.end()) {
            if (it2->second == "pre") {
                result.first = true;
                result.second = PRE;
                return result;
            } else if (it2->second == "post") {
                result.first = true;
                result.second = POST;
                return result;
            } else {
                assert(false);
                result.first = false;
                return result;
            }
        } else {
            result.first = false;
            return result;
        }
    } else {
        result.first = false;
        return result;
    }
}

std::pair<bool, ParentBondRelation> DefaultDependManager::get_parent_bond_relation(DataForTM* data_for_tm) const {
    const DtreeNode *i_root = data_for_tm->i_root;
    std::string dpnd_pattern_str;

    if ((unsigned int)i_root->dpnd > i_root->num) {
        dpnd_pattern_str = i_root->word.lem + "->P:" + i_root->dpnd_ref->word.pos;
    } else {
        dpnd_pattern_str = "P:" + i_root->dpnd_ref->word.pos + "<-" + i_root->word.lem;
    }
    const POS& pos_of_root = data_for_tm->t_root->word.pos;
    std::pair<bool, ParentBondRelation> found_relation_pair = query_parent_bond_relation(dpnd_pattern_str, pos_of_root);
    if (found_relation_pair.first) {
        return found_relation_pair;
    }

    if ((unsigned int)i_root->dpnd > i_root->num) {
        dpnd_pattern_str = "P:" + i_root->word.pos + "->P:" + i_root->dpnd_ref->word.pos;
    } else {
        dpnd_pattern_str = "P:" + i_root->dpnd_ref->word.pos + "<-P:" + i_root->word.pos;
    }
    found_relation_pair = query_parent_bond_relation(dpnd_pattern_str, pos_of_root);
    if (found_relation_pair.first) {
        return found_relation_pair;
    }

    if (data_for_tm->i_root->dpnd != -1) {
        std::pair<bool, ParentBondRelation> result;
        result.first = true;
        if (data_for_tm->i_root->dpnd_ref->num > data_for_tm->i_root->num) {
            result.second = PRE;
        } else {
            result.second = POST;
        }
        return result;
    }
    std::pair<bool, ParentBondRelation> result;
    result.first = false;
    return result;
}

void DefaultDependManager::read_default_dpnd(std::string filename) {
    if (filename != "") {
        std::ifstream ist(filename.c_str());
        if (ist.is_open()) {
            LOG(INFO) << "Reading default dpnd from " << filename << "...";
            std::string line_buf;
            while (!ist.eof()) {
                getline(ist, line_buf);
                if (line_buf == "") continue;
                VS data = split(line_buf, "\t");
                (*this)[data[0]][data[1]] = data[2];
            }
            LOG(INFO) << "Done reading default dpnd." << std::endl;
        } else {
            LOG(FATAL) << "Error: can't open " << filename << ".";
        }
    }
}


/*
  functions for TMProcessor
*/
Hypothesis *TMProcessor::create_hypothesis_from_TM(const Dtree& input,
                                                   const TRResultPtr& result,
                                                   const PatternType& pattern,
                                                   const TM& tm,
                                                   std::map<std::string, double>& t_exp_freq,
                                                   std::map<std::string, std::map<std::string, double> >& t_lm_exp_freq,
                                                   const PartialTranslationFactory& partial_translation_factory,
                                                   const LexicalTranslationScorer& lex_trans_scorer) const {
    // std::cerr << "Creating Initial Hypothesis from " << tm;
    if (!check_compatibility(input, tm)) {
        return NULL;
    }
    Hypothesis *hypothesis = new Hypothesis;
    const Dtree& s_dtree = tm.para_sentence().s_dtree();
    const Dtree& t_dtree = tm.para_sentence().t_dtree();
    DataForTM* data_for_tm = new DataForTM(s_dtree.size(), t_dtree.size());

    // put * marks if the TM comes from OOV-punc
    if (tm.id().find("OOV-punc") != std::string::npos) {
        for (size_t i = 0; i < t_dtree.size(); i++) {
            data_for_tm->t_lm_check[i] = 1;
        }
    }
    
    if (tm.id().find("OOV") != std::string::npos) {
        hypothesis->set_feature_value("oov", hypothesis->get_feature_value("oov") + 1);
    }

    // set [i_body_ref, s_body_ref, t_body_ref, i_root, s_root, t_root, s_bond, t_bond]
    set_body(input, tm, data_for_tm);
    //hypothesis->set_feature_value("size", static_cast<double>(data_for_tm->i_body_ref.size() - 1));

    // set [s_bond, t_bond, s_bond_root, t_bond_root]
    set_bond(tm, data_for_tm);

    // check the condition of the root of TM
    if (!check_tm_root(hypothesis, data_for_tm)) {
        delete hypothesis;
        hypothesis = NULL;
        delete data_for_tm;
        data_for_tm = NULL;
        return NULL;
    }

    // TODO(Nakazawa): check the numerals
    if (!num_adaptation(input, tm, hypothesis, data_for_tm)) {
        delete hypothesis;
        hypothesis = NULL;
        delete data_for_tm;
        data_for_tm = NULL;
        return NULL;
    }

    // sibling or not
    if (tm.wildcards().size() != 0) {
        hypothesis->set_has_wildcard(true);
        hypothesis->set_feature_value("sibling", 1.0);
    }

    // parent bond
    if (data_for_tm->i_root->dpnd_ref && data_for_tm->s_root->dpnd_ref) {
        ParentBondRelation pbr = find_parent_bond(tm, hypothesis, data_for_tm);
        LOG(INFO) << "parent_bond_relation: " << pbr;
        hypothesis->set_parent_bond_relation(pbr);
    } else if (!(data_for_tm->i_root->dpnd_ref) && !(data_for_tm->s_root->dpnd_ref)) {
        hypothesis->set_feature_value("both_ROS", 1.0);
    } else {
        hypothesis->set_feature_value("one_side_ROS", 1.0);
    }
    // 句読点などはParentBondRelationを入力と一致させる
    if (tm_processor_source_->is_punctuation(data_for_tm->i_root->word.lem) ||
        tm_processor_pair_->is_skip_translation(data_for_tm->i_root)) {
        if (static_cast<WordIndex>(data_for_tm->i_root->dpnd) < data_for_tm->i_root->num) { // 後ろからかかる
            hypothesis->set_parent_bond_relation(POST);
        } else { // 前からかかる
            hypothesis->set_parent_bond_relation(PRE);
        }
    }

    // lexical translation prob
    hypothesis->set_feature_value("lex_trans_prob_s_t", lex_trans_scorer.calc_lexical_trans_prob_s_t(tm, data_for_tm));
    hypothesis->set_feature_value("lex_trans_prob_t_s", lex_trans_scorer.calc_lexical_trans_prob_t_s(tm, data_for_tm));

    // child bond
    find_child_bond(input, tm, hypothesis, data_for_tm);
    /*
    // Skip hypotheses with content words aligned to NULL.
    if (hypothesis->get_feature_value("NULL_content") != 0) {
    delete hypothesis;
    hypothesis = NULL;
    delete data_for_tm;
    data_for_tm = NULL;
    return NULL;
    }
    */

    VS t_string;
    std::vector<Additional> additional;
    VS t_real_string;
    VS t_lm_string;
    WordIndex local_num = 0;
    std::map<WordIndex, Dpnd> global_dpnd;
    std::map<WordIndex, Dpnd> global2local;
    DtreeNode* t_string_root = data_for_tm->t_bond_root;
    if (t_string_root->dpnd_ref != NULL) {
        t_string_root = t_string_root->dpnd_ref;
    }
    std::map<WordIndex, POS> bond_pos_map;
    make_t_string(input, t_string_root, tm, data_for_tm, &t_string, &additional,
                  &t_real_string, &t_lm_string, &local_num,
                  &global_dpnd, &global2local, &bond_pos_map);

    // make t_dpnd
    DpndList *t_dpnd = hypothesis->get_t_dpnd_ptr();
    // std::cout << "--- dpnd ---" << std::endl;
    for (size_t i = 0; i < t_string.size(); i++) {
        Dpnd dpnd;
        if (global2local.count(global_dpnd[i]) == 0) {
            dpnd = -1;
        } else {
            dpnd = global2local[global_dpnd[i]];
            Dpnd dpnd_global = global_dpnd[i];
            while (dpnd == -1 && dpnd_global != -1) {
                dpnd_global = t_dtree.node_at(dpnd_global)->dpnd;
                dpnd = global2local[dpnd_global];
            }
        }
        t_dpnd->push_back(dpnd);
        // std::cout << t_string[i] << " | " << i << " | " << (int)dpnd << "|" << global_dpnd[i] << std::endl;
    }

    //hypothesis->set_feature_value("target_size", t_real_string.size());
    hypothesis->set_real_size(t_real_string.size());
    add_count_of_t_exp(&t_real_string, &t_lm_string, &t_exp_freq, &t_lm_exp_freq);

    hypothesis->set_t_string(t_string, partial_translation_factory);
    hypothesis->set_additional(additional);
    // set parent bond relation learned from corpus
    if (hypothesis->parent_bond_relation() == UNK && data_for_tm->i_root->dpnd_ref != NULL) {
        std::pair<bool, ParentBondRelation>  found_relation_pair = default_dpnd_.get_parent_bond_relation(data_for_tm);
        if (found_relation_pair.first) hypothesis->set_parent_bond_relation(found_relation_pair.second);
    }
    hypothesis->set_tm(&tm);
    hypothesis->set_data_for_tm(data_for_tm);

    // set the DtreeNode* to the root word of the hypothesis
    hypothesis->set_root_pos(data_for_tm->t_root->word.pos);

    hypothesis->set_bond_pos_map(bond_pos_map);

    if (Params::Global::keep_tms_in_memory) {
        hypothesis->set_initial_hypotheses_as_self();
    }

    // std::cerr << "loglevel:" << FLAGS_minloglevel << std::endl;;
    LOG(INFO) << *hypothesis;
    
    // For correct working with GOOGLE_STRIP_LOG
#ifndef GOOGLE_STRIP_LOG 
    tm.print(&LOG(INFO), data_for_tm);
#else
    if (GOOGLE_STRIP_LOG < 1) tm.print(&LOG(INFO), data_for_tm);
#endif
    
    //hypothesis->set_feature_value("pattern_frequency", result->get_real_total_number_of_tm(pattern));
    double log_pattern_frequency = log(result->get_real_total_number_of_tm(pattern));
    hypothesis->set_feature_value("log_pattern_frequency", log_pattern_frequency);

//    if (Params::Translator::enable_new_features) {
    hypothesis->set_feature_value("size_x_log_pattern_frequency",
                                  static_cast<double>(data_for_tm->i_body_ref.size() * log_pattern_frequency));
//    }
    LOG(INFO) << *hypothesis;

    //std::cout << *hypothesis << std::endl;
    //std::cout << tm << std::endl;

    return hypothesis;
}

bool TMProcessor::check_compatibility(const Dtree& input, const TM& tm) const {
    if (tm.wildcards().size() == 0) {
        return true;
    } else if (Params::Translator::no_sibling) {
        return false;
    }

    // eliminate if wildcard is in the middle
    // eliminate if siblings includes grand-children
    int max_depth = 0;
    BOOST_FOREACH(WordIndex index, tm.wildcards()) {
        int tmp = input.node_at(index)->depth;
        if (max_depth < tmp) {
            max_depth = tmp;
        }
    }
    int sibling_depth = -1;
    unsigned int s_size = 0;
    WordIndex min_index = 100000;
    WordIndex max_index = 0;
    BOOST_FOREACH(Alignment i_s_match, tm.i_s_match()) {
        if (word_overlap(tm.wildcards(), i_s_match.p1)) {
            continue;
        }
        BOOST_FOREACH(WordIndex index, i_s_match.p1) {
            int tmp = input.node_at(index)->depth;
            if (max_depth > tmp) {
                return false;
            }
            if (sibling_depth == -1) {
                sibling_depth = tmp;
            } else if (sibling_depth != tmp) {
                return false;
            }
            s_size++;
            if (index < min_index) {
                min_index = index;
            }
            if (index > max_index) {
                max_index = index;
            }
        }
    }

    // eliminate if siblings are not contiguous
    if (max_index - min_index + 1 != s_size) {
        return false;
    }

    return true;
}

Hypothesis *TMProcessor::create_empty_hypothesis(const PatternType pattern,
                                                 const PartialTranslationFactory& partial_translation_factory) const {
    Hypothesis *hypothesis = new Hypothesis;
    hypothesis->set_feature_value("log_pattern_frequency", 0.0); // 0.0 = log 1

    VS t_string;
    std::vector<Additional> additional;
    hypothesis->set_t_string(t_string, partial_translation_factory);
    hypothesis->set_additional(additional);
    hypothesis->set_parent_bond_relation(PRE);
    hypothesis->set_initial_hypotheses_as_self();

    return hypothesis;
}

void TMProcessor::set_body(const Dtree& input, const TM& tm, DataForTM* data_for_tm) const {
    const std::vector<Alignment>& i_s_match = tm.i_s_match();
    const Dtree& s_dtree = tm.para_sentence().s_dtree();
    const Dtree& t_dtree = tm.para_sentence().t_dtree();

    // size of the example
    // input and source side body
    for (std::vector<Alignment>::const_iterator it = i_s_match.begin(); it != i_s_match.end(); it++) {
        // skip wildcard nodes
        if (word_overlap(tm.wildcards(), it->p1)) {
            continue;
        }
        for (std::vector<WordIndex>::const_iterator wd = it->p1.begin(); wd != it->p1.end(); wd++) {
            data_for_tm->i_body_ref.push_back(input.node_at(*wd));
        }
        for (std::vector<WordIndex>::const_iterator wd = it->p2.begin(); wd != it->p2.end(); wd++) {
            data_for_tm->s_body_ref.push_back(s_dtree.node_at(*wd));
            data_for_tm->s_bond[*wd] = 0;
        }
    }

    // target side body
    const std::vector<WordIndex>& t_body = tm.connected_image();
    for (std::vector<WordIndex>::const_iterator it = t_body.begin(); it != t_body.end(); it++) {
        data_for_tm->t_bond[*it] = 0;
        data_for_tm->t_body_ref.push_back(t_dtree.node_at(*it));
    }

    // target側の周りの機能語、および葉の内容語を取る(2段階まで)
    std::set<const DtreeNode*> t_lm_around;
    std::vector<const DtreeNode*> t_body_ref_tmp = data_for_tm->t_body_ref;
    t_dtree.get_neighbors(t_body_ref_tmp, &t_lm_around, true, true, true);
    std::set<const DtreeNode*>::iterator tmp = t_lm_around.begin();
    while (tmp != t_lm_around.end()) {
        if ((*tmp)->word.content_p == 0 ||
            ((*tmp)->pre_children_ref.size() == 0 &&
             (*tmp)->post_children_ref.size() == 0)) {
            t_body_ref_tmp.push_back(const_cast<const DtreeNode*>(*tmp));
            tmp++;
        } else {
            t_lm_around.erase(tmp++);
        }
    }
    std::set<const DtreeNode*> t_lm_around_second;
    t_dtree.get_neighbors(t_body_ref_tmp, &t_lm_around_second, true, true, true);
    for (std::set<const DtreeNode*>::iterator it = t_lm_around_second.begin(); it != t_lm_around_second.end(); it++) {
        if ((*it)->word.content_p == 0 ||
            ((*it)->pre_children_ref.size() == 0 &&
             (*it)->post_children_ref.size() == 0)) {
            t_lm_around.insert(*it);
        }
    }

    // target side around NULL
    const std::vector<WordIndex>& t_around = tm.additional_nodes();
    for (std::vector<WordIndex>::const_iterator it = t_around.begin(); it != t_around.end(); it++) {
        if (t_lm_around.find(const_cast<DtreeNode*>(t_dtree.node_at(*it))) != t_lm_around.end()) {
            data_for_tm->t_lm_check[*it] = 1;
            data_for_tm->t_bond[*it] = 0;
            data_for_tm->t_body_ref.push_back(t_dtree.node_at(*it));
        }
    }

    // root node
    data_for_tm->i_root = const_cast<DtreeNode*>(input.get_root_p(data_for_tm->i_body_ref));
    data_for_tm->s_root = const_cast<DtreeNode*>(s_dtree.get_root_p(data_for_tm->s_body_ref));
    data_for_tm->t_root = const_cast<DtreeNode*>(t_dtree.get_root_p(data_for_tm->t_body_ref));
}

void TMProcessor::set_bond(const TM& tm, DataForTM* data_for_tm) const {
    std::set<const DtreeNode*> s_around;
    tm.para_sentence().s_dtree().get_neighbors(data_for_tm->s_body_ref, &s_around, true, true, false); // no siblings
    std::set<const DtreeNode*> t_around;
    tm.para_sentence().t_dtree().get_neighbors(data_for_tm->t_body_ref, &t_around, true, true, false); // no siblings

    // get around words which are possible to be bonds
    BOOST_FOREACH(const DtreeNode *sw, s_around) {
        const Alignment* align = tm.para_sentence().get_alignment_s(sw);
        if (align != NULL && word_overlap(align->p2, t_around)) {
            // source-side bond
            for (std::vector<WordIndex>::const_iterator it = align->p1.begin(); it != align->p1.end(); it++) {
                data_for_tm->s_bond[*it] = 1;
            }
            // target-side bond
            // TODO(nakazawa): need to consider discontiguous alignment in target side
            for (std::vector<WordIndex>::const_iterator it = align->p2.begin(); it != align->p2.end(); it++) {
                data_for_tm->t_bond[*it] = 1;
            }
        }
    }

    // source bond root
    if (data_for_tm->s_root->dpnd_ref && data_for_tm->s_bond[data_for_tm->s_root->dpnd] == 1) {
        data_for_tm->s_bond_root = data_for_tm->s_root->dpnd_ref;
    } else {
        data_for_tm->s_bond_root = const_cast<DtreeNode*>(data_for_tm->s_root);
    }
    // target bond root
    if (data_for_tm->t_root && data_for_tm->t_root->dpnd_ref && data_for_tm->t_bond[data_for_tm->t_root->dpnd] == 1) {
        data_for_tm->t_bond_root = data_for_tm->t_root->dpnd_ref;
    } else {
        data_for_tm->t_bond_root = const_cast<DtreeNode*>(data_for_tm->t_root);
    }
}

void TMProcessor::compute_child_list(const Dtree& input,
                                     const TM& tm,
                                     DataForTM* data_for_tm) const {
    const Dtree& s_dtree = tm.para_sentence().s_dtree();
    const std::vector<Alignment>& i_s_match = tm.i_s_match();

    BOOST_FOREACH(const Alignment& align, i_s_match) {
        // skip wildcard nodes
        if (word_overlap(tm.wildcards(), align.p1)) {
            continue;
        }

        // input側の子供を集める
        BOOST_FOREACH(const WordIndex iw, align.p1) {
            BOOST_FOREACH(const DtreeNode *c, input.node_at(iw)->pre_children_ref) {
                if (grep(c, data_for_tm->i_body_ref).size() == 0) {
                    data_for_tm->i_child_ref.push_back(c);
                }
            }
            BOOST_FOREACH(const DtreeNode *c, input.node_at(iw)->post_children_ref) {
                if (grep(c, data_for_tm->i_body_ref).size() == 0) {
                    data_for_tm->i_child_ref.push_back(c);
                }
            }
        }

        // source側の子供を集める
        BOOST_FOREACH(const WordIndex sw, align.p2) {
            BOOST_FOREACH(const DtreeNode *c, s_dtree.node_at(sw)->pre_children_ref) {
                if (data_for_tm->s_bond[c->num] == 1) {
                    data_for_tm->s_child_ref.push_back(c);
                } else if (data_for_tm->s_bond[c->num] == -1) {
                    data_for_tm->s_other_child_ref.push_back(c);
                }
            }
            BOOST_FOREACH(const DtreeNode *c, s_dtree.node_at(sw)->post_children_ref) {
                if (data_for_tm->s_bond[c->num] == 1) {
                    data_for_tm->s_child_ref.push_back(c);
                } else if (data_for_tm->s_bond[c->num] == -1) {
                    data_for_tm->s_other_child_ref.push_back(c);
                }
            }
        }
    }
}

ParentBondRelation TMProcessor::find_parent_bond(const TM& tm, Hypothesis *hypothesis, DataForTM* data_for_tm)  const {
    // currently, no head-switched bonds are accepted. if you need
    // head-switched examples, you must change the conditions below.
    bool same_relative_position = (((unsigned int)data_for_tm->i_root->dpnd > data_for_tm->i_root->num &&
                                    (unsigned int)data_for_tm->s_root->dpnd > data_for_tm->s_root->num) ||
                                   (data_for_tm->i_root->num > (unsigned int)data_for_tm->i_root->dpnd &&
                                    data_for_tm->s_root->num > (unsigned int)data_for_tm->s_root->dpnd));
    const Alignment* align = tm.para_sentence().get_alignment_s(data_for_tm->s_root->dpnd_ref);
    bool head_compatibility = (align && grep(data_for_tm->t_bond_root->num, align->p2).size() > 0);
    LOG(INFO) << "find_parent_bond:" << std::endl;
    LOG(INFO) << "  same_relative_position: " << same_relative_position << std::endl;
    LOG(INFO) << "  head_compatibility: " << head_compatibility << std::endl;
    LOG(INFO) << "  bond(s_root->dpnd) == 1: " << data_for_tm->s_bond[data_for_tm->s_root->dpnd] << std::endl;
    bool big_condition = (same_relative_position && head_compatibility && data_for_tm->s_bond[data_for_tm->s_root->dpnd] == 1);
    if (big_condition) {
        double word_similarity = tm_processor_source_->calc_word_similarity(data_for_tm->i_root->dpnd_ref, data_for_tm->s_root->dpnd_ref);
        hypothesis->set_feature_value("parent_bond_available", 1.0);
        hypothesis->set_feature_value("parent_bond_similarity", word_similarity);
        hypothesis->set_feature_value("parent_bond_LD",
                                      tm_processor_source_->check_parent_bond(data_for_tm->i_root->dpnd_ref,
                                                                              data_for_tm->s_root->dpnd_ref));
        data_for_tm->s_bond[data_for_tm->s_root->dpnd] = 2;
        std::vector<const DtreeNode*> t_bond_match_ref;
        for (std::vector<WordIndex>::const_iterator it = align->p2.begin(); it != align->p2.end(); it++) {
            data_for_tm->t_bond[*it] = 2;
            t_bond_match_ref.push_back(tm.para_sentence().t_dtree().node_at(*it));
        }
        return self_bond_relation(data_for_tm->t_root, t_bond_match_ref);
    } else {
        std::string i_parent_cat = tm_processor_source_->get_abst_cat(data_for_tm->i_root->dpnd_ref);
        std::string s_parent_cat = tm_processor_source_->get_abst_cat(data_for_tm->s_root->dpnd_ref);
        LOG(INFO) << "check parent cat: input(" << i_parent_cat << ") vs source(" << s_parent_cat << ")";
        if (i_parent_cat == s_parent_cat) {
            hypothesis->set_feature_value("same_parent_cat", 1.0);
        } else {
            hypothesis->set_feature_value("different_parent_cat", 1.0);
        }
    }
    return UNK;
}

void TMProcessor::align_child_bond(const TM& tm,
                                   Hypothesis *hypothesis,
                                   DataForTM* data_for_tm) const {
    const Dtree& t_dtree = tm.para_sentence().t_dtree();
    int repeat_flag = 1;
    std::vector<const DtreeNode*> tmp_t_body = data_for_tm->t_body_ref;
  CHILD_BOND_LOOP:
    while (repeat_flag) {
        repeat_flag = 0;
        for (std::vector<const DtreeNode*>::iterator s_child = data_for_tm->s_child_bond_cand.begin();
             s_child != data_for_tm->s_child_bond_cand.end(); s_child++) {
            for (std::vector<const DtreeNode*>::iterator i_child = data_for_tm->i_child_ref.begin();
                 i_child != data_for_tm->i_child_ref.end(); i_child++) {
                // 句読点などはbondの対象にしない
                if (tm_processor_source_->is_punctuation((*i_child)->word.lem) ||
                    tm_processor_pair_->is_skip_translation(*i_child)) continue;
                // POS check
                if (!tm_processor_source_->is_child_bond(tm, *i_child, *s_child)) continue;
                // s_childがNULLでない場合のみOK
                const Alignment* align = tm.para_sentence().get_alignment_s(*s_child);
                if (align != NULL) {
                    std::set<const DtreeNode*> t_tmp_around;
                    // target側が兄弟となっている語はのりしろとして使わない
                    t_dtree.get_neighbors(align->p2, &t_tmp_around, true, true, false);
                    if (word_overlap(tmp_t_body, t_tmp_around) &&
                        // no head switch
                        grep(data_for_tm->t_bond_root->num, align->p2).size() == 0) {
                        double word_similarity = tm_processor_source_->calc_word_similarity(*i_child, *s_child);
                        hypothesis->set_feature_value("child_bond_count",
                                                      hypothesis->get_feature_value("child_bond_count") + 1.0);
                        hypothesis->set_feature_value("child_bond_similarity",
                                                      hypothesis->get_feature_value("child_bond_similarity") + word_similarity);
                        // abnormal child
                        WordIndex t_root_num = t_dtree.get_root(align->p2);
                        if (t_dtree.node_at(t_root_num)->dpnd_ref != NULL &&
                            data_for_tm->t_bond[t_dtree.node_at(t_root_num)->dpnd] != 0) {
                            hypothesis->set_feature_value("abnormal_child_bond",
                                                          hypothesis->get_feature_value("abnormal_child_bond") + 1.0);
                        }
                        
                        // removing lm_check for parents of the child bond
                        const DtreeNode* tmp_t_for_lm_check = t_dtree.node_at(t_root_num);
                        while (tmp_t_for_lm_check->dpnd_ref != NULL && data_for_tm->t_lm_check[tmp_t_for_lm_check->dpnd] == 1) {
                            data_for_tm->t_lm_check[tmp_t_for_lm_check->dpnd] = 0;
                            tmp_t_for_lm_check = tmp_t_for_lm_check->dpnd_ref;
                        }

                        data_for_tm->s_bond[(*s_child)->num] = 2;
                        for (std::vector<WordIndex>::const_iterator it = align->p2.begin(); it != align->p2.end(); it++) {
                            if (*it != data_for_tm->t_root->num) {
                                tmp_t_body.push_back(t_dtree.node_at(*it));
                            }
                            data_for_tm->t_bond[*it] = 2;
                            data_for_tm->match_i_of_t[*it].insert((*i_child)->num);
                        }
                        data_for_tm->s_child_bond_cand.erase(s_child);
                        data_for_tm->i_child_ref.erase(i_child);
                        repeat_flag = 1;
                        goto CHILD_BOND_LOOP;
                    }
                }
            }
        }
    }
}

void TMProcessor::set_feature_value_for_child_bond(const TM& tm,
                                                   Hypothesis *hypothesis,
                                                   DataForTM* data_for_tm) const {
    const std::vector<Alignment>& i_s_match = tm.i_s_match();
    const Dtree& s_dtree = tm.para_sentence().s_dtree();
    const Dtree& t_dtree = tm.para_sentence().t_dtree();

    // 余分な子供が全くいない
    if (data_for_tm->i_child_ref.size() == 0 && data_for_tm->s_child_bond_cand.size() == 0) {
        hypothesis->set_feature_value("no_redundant_child", 1.0);
    }
    // 入力側に糊代のない子供がある場合は（その個数分だけ）ペナルティ
    hypothesis->set_feature_value("redundant_child_in_input", static_cast<double>(data_for_tm->i_child_ref.size()));
    // ソート必要???
    BOOST_FOREACH(const DtreeNode *i, data_for_tm->i_child_ref) {
        // 子方向の糊しろがないものを記録
        DtreeNode *now = const_cast<DtreeNode*>(i);
        DtreeNode *tmp_t_root = NULL;
        while (grep(const_cast<const DtreeNode*>(now->dpnd_ref), data_for_tm->i_body_ref).size() > 0) {
            const Alignment* i_s_align = tm.get_alignment_i(now->dpnd_ref);
            if (i_s_align) {
                std::vector<const DtreeNode*> tmp_t;
                BOOST_FOREACH(const WordIndex s, i_s_align->p2) {
                    const Alignment* s_t_align = tm.para_sentence().get_alignment_s(s_dtree.node_at(s));
                    if (s_t_align) {
                        BOOST_FOREACH(const WordIndex t, s_t_align->p2) {
                            tmp_t.push_back(t_dtree.node_at(t));
                        }
                    }
                }
                if (tmp_t.size() > 0) {
                    tmp_t_root = const_cast<DtreeNode*>(t_dtree.get_root_p(tmp_t));
                    break;
                }
            }
            now = now->dpnd_ref;
        }
        if (tmp_t_root == NULL) {  // or defined $s_cp->{is_punc}
            tmp_t_root = data_for_tm->t_root;
        }
        if (tmp_t_root == NULL) {
            data_for_tm->additional_for_empty.insert(i->num);
        } else {
            data_for_tm->additional_for_t[tmp_t_root->num].insert(i->num);
        }
    }
    // source側に余分の子供がある場合は（その個数分だけ）ペナルティ
    hypothesis->set_feature_value("redundant_child_in_source",
                                  static_cast<double>(data_for_tm->s_child_bond_cand.size()));
    // NULL対応は減点
    BOOST_FOREACH(const Alignment& align, i_s_match) {
        // skip wildcard nodes
        if (word_overlap(tm.wildcards(), align.p1)) {
            continue;
        }
        // i側の子供を集める
        BOOST_FOREACH(const WordIndex sw, align.p2) {
            const Alignment* align_tmp = tm.para_sentence().get_alignment_s(s_dtree.node_at(sw));
            if (align_tmp == NULL) {
                if (s_dtree.node_at(sw)->word.content_p == 1 &&
                    (!tm_processor_source_->is_punctuation(s_dtree.node_at(sw)->word.lem) ||
                     !tm_processor_pair_->is_skip_translation(s_dtree.node_at(sw)))) {
                    hypothesis->set_feature_value("NULL_content",
                                                  hypothesis->get_feature_value("NULL_content") + 1.0);
                } else {
                    hypothesis->set_feature_value("NULL_function",
                                                  hypothesis->get_feature_value("NULL_function") + 1.0);
                }
            }
        }
    }
}

void TMProcessor::find_child_bond(const Dtree& input,
                                  const TM& tm,
                                  Hypothesis *hypothesis,
                                  DataForTM* data_for_tm) const {
    const Dtree& t_dtree = tm.para_sentence().t_dtree();
    // gather children
    compute_child_list(input, tm, data_for_tm);
    data_for_tm->s_child_bond_cand = data_for_tm->s_child_ref;
    for (std::vector<const DtreeNode*>::const_iterator it = data_for_tm->s_other_child_ref.begin();
         it != data_for_tm->s_other_child_ref.end(); it++) {
        data_for_tm->s_child_bond_cand.push_back(*it);
    }
    // child bond
    for (size_t i = 0; i < t_dtree.size(); i++) {
        data_for_tm->match_i_of_t[i].clear();
    }

    align_child_bond(tm, hypothesis, data_for_tm);

    set_feature_value_for_child_bond(tm, hypothesis, data_for_tm);
}

ParentBondRelation TMProcessor::self_bond_relation(const DtreeNode *self, std::vector<const DtreeNode*> bond) const {
    BOOST_FOREACH(const DtreeNode *bw, bond) {
        if (grep((DtreeNode*)(self), bw->pre_children_ref).size() > 0) { // NOLINT
            return PRE;
        } else if (grep((DtreeNode*)(self), bw->post_children_ref).size() > 0) { // NOLINT
            return POST;
        }
    }
    return UNK;
}

std::vector<WordIndex> select_sep(WordIndex self, const AdditionalPositions& sep) {
    std::vector<WordIndex> result;
    size_t size = sep.size();
    if (!Params::Translator::no_keep_input_order_of_additionals) {
        for (size_t i = 0; i < size; i++) {
            bool flag = true;
            for (size_t j = i+1; j < size; j++) {
                BOOST_FOREACH(const WordIndex w, sep[j].second) {
                    if (w < self) {
                        flag = false;
                        break;
                    }
                }
            }
            if (!flag) {
                continue;
            }
            for (size_t j = 1; j <= i; j++) { // j = 0 is always empty
                BOOST_FOREACH(const WordIndex w, sep[j].second) {
                    if (w > self) {
                        flag = false;
                        break;
                    }
                }
            }
            if (flag) {
                result.push_back(sep[i].first);
            }
        }
    }
    if (result.size() == 0) {
        for (size_t i = 0; i < size; i++) {
            result.push_back(sep[i].first);
        }
    }
    if (result.size() == 0) {
        std::cerr << "Error! no insertion list for additionals" << std::endl;
    }
    return result;
}

void TMProcessor::make_t_string(const Dtree& input, const DtreeNode *t_ref, const TM& tm, DataForTM *data_for_tm, VS *t_string,
                                std::vector<Additional> *additional, VS *t_real_string, VS *t_lm_string,
                                WordIndex *local_num, std::map<WordIndex, Dpnd> *global_dpnd,
                                std::map<WordIndex, Dpnd> *global2local,
                                std::map<WordIndex, POS> *bond_pos_map) const {
    // pre-children
    PatternType empty_type;
    AdditionalPositions pre_children_sep;
    pre_children_sep.push_back(std::make_pair((WordIndex)t_string->size(), empty_type));
    if (t_ref) {
        BOOST_FOREACH(const DtreeNode *c, t_ref->pre_children_ref) {
            make_t_string(input, c, tm, data_for_tm, t_string, additional, t_real_string, t_lm_string,
                          local_num, global_dpnd, global2local, bond_pos_map);
            if (pre_children_sep.rbegin()->first != (WordIndex)t_string->size()) {
                if (data_for_tm->match_i_of_t[c->num].size() != 0) {
                    pre_children_sep.push_back(std::make_pair((WordIndex)t_string->size(), data_for_tm->match_i_of_t[c->num]));
                } else {
                    pre_children_sep.push_back(std::make_pair((WordIndex)t_string->size(), tm.get_i_of_t(c)));
                }
            }
        }
    }

    // self
    WordIndex self_index = (WordIndex)t_string->size();
    bool added_new_t_string = false;
    bool keep_dependency_info = false;
    if (!t_ref) {
        t_real_string->push_back("");
    } else if (data_for_tm->t_bond[t_ref->num] == 0) {  // terminals
        std::string cont = pre_processor_.pre_process(t_ref->word.content);
        if (data_for_tm->t_modified_content.find(t_ref->num) != data_for_tm->t_modified_content.end())
            cont = data_for_tm->t_modified_content[t_ref->num];
        if (data_for_tm->t_lm_check[t_ref->num]) {  // optional words
            cont += '*';
            t_lm_string->push_back(cont);
        } else {  // real words
            t_real_string->push_back(cont);
        }
        t_string->push_back(cont);
        // debugging
        // std::cout << t_ref->num << "/" << cont << ": " << t_ref->dpnd << std::endl;
        added_new_t_string = true;
    } else if (data_for_tm->t_bond[t_ref->num] == 2 &&
               data_for_tm->match_i_of_t[t_ref->num].size() != 0) {  // non-terminals
        keep_dependency_info = true;
        std::string new_t_string = "[X" + pattern_type_str(data_for_tm->match_i_of_t[t_ref->num]) + "]";
        // 使うのりしろ
        if ((self_index == 0 || (*t_string)[self_index-1] != new_t_string) &&
            t_ref->dpnd_ref != NULL &&
            pattern_type_str(data_for_tm->match_i_of_t[t_ref->num]) != pattern_type_str(data_for_tm->match_i_of_t[t_ref->dpnd])) {
            t_string->push_back(new_t_string);
            // debugging
            // std::cout << ">> " << t_ref->num << "/" << new_t_string << ": " << t_ref->dpnd << std::endl;
            (*bond_pos_map)[t_string->size() - 1] = t_ref->word.pos;
            added_new_t_string = true;
        }
    }
    // Store dependency information for each word.
    if (added_new_t_string) {
        (*global_dpnd)[*local_num] = t_ref->dpnd;
        (*global2local)[t_ref->num] = *local_num;
        (*local_num)++;
    } else if (keep_dependency_info) {
        (*global2local)[t_ref->num] = -1;
    }

    // post-children
    AdditionalPositions post_children_sep;
    post_children_sep.push_back(std::make_pair((WordIndex)t_string->size(), empty_type));
    if (t_ref) {
        BOOST_FOREACH(const DtreeNode *c, t_ref->post_children_ref) {
            make_t_string(input, c, tm, data_for_tm, t_string, additional, t_real_string, t_lm_string,
                          local_num, global_dpnd, global2local, bond_pos_map);
            if (post_children_sep.rbegin()->first != (WordIndex)t_string->size()) {
                if (data_for_tm->match_i_of_t->find(c->num) != data_for_tm->match_i_of_t->end()) {
                    post_children_sep.push_back(std::make_pair((WordIndex)t_string->size(), data_for_tm->match_i_of_t[c->num]));
                } else {
                    post_children_sep.push_back(std::make_pair((WordIndex)t_string->size(), tm.get_i_of_t(c)));
                }
            }
        }
    }

    // additionals
    if (t_ref) {
        for (PatternType::iterator it = data_for_tm->additional_for_t[t_ref->num].begin();
             it != data_for_tm->additional_for_t[t_ref->num].end(); it++) {
            // 句読点などは、addの方向を固定
            if (tm_processor_source_->is_punctuation(input.node_at(*it)->word.lem) ||
                tm_processor_pair_->is_skip_translation(input.node_at(*it))) {
                std::vector<WordIndex> empty_vec;
                if (static_cast<WordIndex>(input.node_at(*it)->dpnd) < *it) { // 後ろからかかる
                    std::vector<WordIndex> selected_post_children_sep = select_sep(*it, post_children_sep);
                    Additional add(*it, empty_vec, selected_post_children_sep,
                                   (*global2local)[t_ref->num], t_ref->word.pos);
                    additional->push_back(add);
                } else { // 前からかかる
                    std::vector<WordIndex> selected_pre_children_sep = select_sep(*it, pre_children_sep);
                    Additional add(*it, selected_pre_children_sep, empty_vec,
                                   (*global2local)[t_ref->num], t_ref->word.pos);
                    additional->push_back(add);
                }
            } else {
                std::vector<WordIndex> selected_pre_children_sep = select_sep(*it, pre_children_sep);
                std::vector<WordIndex> selected_post_children_sep = select_sep(*it, post_children_sep);
                Additional add(*it, selected_pre_children_sep, selected_post_children_sep,
                               (*global2local)[t_ref->num], t_ref->word.pos);
                additional->push_back(add);
            }
        }
    } else {
        for (PatternType::iterator it = data_for_tm->additional_for_empty.begin();
             it != data_for_tm->additional_for_empty.end(); it++) {
            // 句読点などは、addの方向を固定
            if (tm_processor_source_->is_punctuation(input.node_at(*it)->word.lem) ||
                tm_processor_pair_->is_skip_translation(input.node_at(*it))) {
                std::vector<WordIndex> empty_vec;
                if (static_cast<WordIndex>(input.node_at(*it)->dpnd) < *it) { // 後ろからかかる
                    std::vector<WordIndex> selected_post_children_sep = select_sep(*it, post_children_sep);
                    Additional add(*it, empty_vec, selected_post_children_sep, -1, "");
                    additional->push_back(add);
                } else { // 前からかかる
                    std::vector<WordIndex> selected_pre_children_sep = select_sep(*it, pre_children_sep);
                    Additional add(*it, selected_pre_children_sep, empty_vec, -1, "");
                    additional->push_back(add);
                }
            } else {
                std::vector<WordIndex> selected_pre_children_sep = select_sep(*it, pre_children_sep);
                std::vector<WordIndex> selected_post_children_sep = select_sep(*it, post_children_sep);
                Additional add(*it, selected_pre_children_sep, selected_post_children_sep, -1, "");
                additional->push_back(add);
            }
        }
    }
}

void TMProcessor::add_count_of_t_exp(VS *t_real_string,
                                     VS *t_lm_string,
                                     std::map<std::string, double> *t_exp_freq,
                                     std::map<std::string, std::map<std::string, double> > *t_lm_exp_freq) const {
    sort(t_real_string->begin(), t_real_string->end());
    std::string t_exp = join(";", *t_real_string);
    (*t_exp_freq)[t_exp]++;

    sort(t_lm_string->begin(), t_lm_string->end());
    std::string t_lm_exp = join(";", *t_lm_string);
    std::map<std::string, std::map<std::string, double> >::iterator it = t_lm_exp_freq->find(t_exp);
    if (it == t_lm_exp_freq->end()) {
        std::map<std::string, double> t_lm_tmp;
        t_lm_tmp[t_lm_exp] = 1.0;
        t_lm_exp_freq->insert(std::map<std::string, std::map<std::string, double> >::value_type(t_exp, t_lm_tmp));
    } else {
        (it->second)[t_lm_exp]++;
    }
}

bool TMProcessor::check_tm_root(Hypothesis *hypothesis,
                                DataForTM* data_for_tm) const {
    // check if the content word of input and source are equal
    if (data_for_tm->i_root->word.content == data_for_tm->s_root->word.content) {
        hypothesis->set_feature_value("same_root_content", 1.0);
    }
    // language dependent check
    hypothesis->set_feature_value("root_LD", hypothesis->get_feature_value("root_LD") + tm_processor_source_->check_tm_root(data_for_tm));
    return true;
}

bool TMProcessor::num_adaptation(const Dtree& input,
                                 const TM& tm,
                                 Hypothesis *hypothesis,
                                 DataForTM* data_for_tm) const {
    BOOST_FOREACH(const Alignment& align, tm.i_s_match()) {
        // skip wildcard nodes
        if (word_overlap(tm.wildcards(), align.p1)) {
            continue;
        }
        if (!_num_adaptation(input, tm, align, hypothesis, data_for_tm)) {
            return false;
        }
    }
    return true;
}

bool TMProcessor::_num_adaptation(const Dtree& input,
                                  const TM& tm,
                                  const Alignment& align,
                                  Hypothesis *hypothesis,
                                  DataForTM* data_for_tm) const {
    const DtreeNode *i_w = input.node_at(align.p1[0]);
    const DtreeNode *s_w = tm.para_sentence().s_dtree().node_at(align.p2[0]);
    boost::regex num_regex("<\\@num\\-?(.*?):(.+?)>");
    boost::smatch input_results;
    boost::smatch source_results;
    // input has numerals
    if (boost::regex_search(i_w->word.lem, input_results, num_regex)) {
        // both input and source-side of TM have numerals
        if (boost::regex_search(s_w->word.lem, source_results, num_regex)) {
            // not use if source word is NULL-aligned
            const Alignment* s_t_align = tm.para_sentence().get_alignment_s(s_w);
            if (s_t_align == NULL) {
                return false;
            }
            // no need to process if two numerals are equal
            if (input_results.str(2) == source_results.str(2)) {
                return true;
            }
            // マッチしなければnum_type "nm" の用例は使えない(二人部屋/twin roomなど)
            // if not equal, can't use num_type "nm" (ex 二人部屋/twin room)
            if (source_results.str(1) == "nm") {
                return false;
            }
            // num_type の不一致は使わない
            if (input_results.str(1) != source_results.str(1)) {
                return false;
            }

            // 他はOKだが減点し、数字を合わせる
            hypothesis->set_feature_value("numeral_mismatch", 1.0);
            bool num_adapted_flag = false;
            BOOST_FOREACH(WordIndex t_w, s_t_align->p2) {
                boost::smatch target_results;
                const DtreeNode *t_w_ref = tm.para_sentence().t_dtree().node_at(t_w);
                if (boost::regex_search(t_w_ref->word.lem, target_results, num_regex)) {
                    if (num_adapted_flag) {  // there are more than one numerals aligned to one numeral
                        return false;
                    }
                    if (input_results.str(2) == target_results.str(2)) {
                        // do nothing
                    } else if (source_results.str(2) == target_results.str(2)) {
                        data_for_tm->t_modified_content.insert(
                            std::pair<WordIndex, std::string>(t_w, tm_processor_pair_->num_translation(t_w_ref->word.content, input_results.str(2), source_results.str(1))));
                    } else {
                        // アラインメント時にマッチしなかったものでも、とりあえず数字を置き換える
                        data_for_tm->t_modified_content.insert(
                            std::pair<WordIndex, std::string>(t_w, tm_processor_pair_->num_translation(t_w_ref->word.content, input_results.str(2), target_results.str(1))));
                    }
                    num_adapted_flag = true;
                }
            }
            if (!num_adapted_flag) { // not adapted
                return false;
            }
        } else {
            // only input has numerals
            return false;
        }
    } else {
        // only source-side of TM has numerals
        if (boost::regex_search(s_w->word.lem, source_results, num_regex)) {
            return false;
        }
        if (i_w->word.lem != s_w->word.lem) {
            return false;
        }
    }
    return true;
}

