/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef TM_PROCESSOR_INCLUDED
#define TM_PROCESSOR_INCLUDED

#include <map>
#include <set>
#include <string>
#include <utility>
#include <vector>
#include "para_sentence.hpp"
#include "tm.hpp"
#include "alignment.hpp"
#include "def.hpp"
#include "dtree.hpp"
#include "hypothesis.hpp"
#include "pre_processor.hpp"
#include "../TR++/tr_interface.hpp"

#include "TMProcessor/source.hpp"
#include "TMProcessor/target.hpp"
#include "TMProcessor/pair.hpp"
#include "TMProcessor/ja.hpp"
#include "TMProcessor/en.hpp"
#include "TMProcessor/zh.hpp"
#include "TMProcessor/ja_en.hpp"
#include "TMProcessor/en_ja.hpp"
#include "TMProcessor/ja_zh.hpp"
#include "TMProcessor/zh_ja.hpp"
#include "lexical_translation_scorer.hpp"

typedef std::vector<std::pair<WordIndex, std::set<WordIndex> > > AdditionalPositions;

/******************************************************************************
 * The class DefaultDependManager is in charge of finding default dependency
 * relations for examples for which the bond could not be computed.
 */
class DefaultDependManager : public std::map<std::string, std::map<std::string, std::string> > {
  public:
    /// \brief Create a DefaultDependManager by loading data from filename.
    explicit DefaultDependManager(const std::string& filename) {
        read_default_dpnd(filename);
    }
    
    /// Default constructor (mainly for allowing instanciating
    /// Translator object without TMProcessor for hypothesis input
    /// mode)
    DefaultDependManager() {}
    /// \brief Attempt to find an appropriate parent-bond relation for
    /// data_for_tm The result is a pair of a boolean and a
    /// ParentBondRelation. If the boolean is false, no
    /// ParentBondRelation could be found. If it is true, the second
    /// field of the pair contains a ParentBondRelation for
    /// data_for_tm.
    std::pair<bool, ParentBondRelation> get_parent_bond_relation(DataForTM* data_for_tm) const;
    
  private:
    std::pair<bool, ParentBondRelation> query_parent_bond_relation(const std::string& dpnd_pattern_str,
                                                                   const POS& pos_of_root) const;
    void read_default_dpnd(std::string filename);
};

class TMProcessor {
  public:
    /// Default constructor (mainly for allowing instanciating Translator object without TMProcessor for hypothesis input mode
    TMProcessor() : tm_processor_source_(NULL), tm_processor_target_(NULL), tm_processor_pair_(NULL) {}
    TMProcessor(const Language& s_lang, const Language& t_lang, std::string& dpnd_filename) :
        default_dpnd_(dpnd_filename), tm_processor_source_(NULL), tm_processor_target_(NULL), tm_processor_pair_(NULL) {
        create_source_tm_processor(s_lang);
        create_target_tm_processor(t_lang);
        create_bilingual_tm_processor(s_lang, t_lang);
    }

    ~TMProcessor() {
        if (tm_processor_source_ != NULL) {
            delete tm_processor_source_;
            tm_processor_source_ = NULL;
        }
        if (tm_processor_target_ != NULL) {
            delete tm_processor_target_;
            tm_processor_target_ = NULL;
        }
        if (tm_processor_pair_ != NULL) {
            delete tm_processor_pair_;
            tm_processor_pair_ = NULL;
        }
    }
    
    Hypothesis *create_hypothesis_from_TM(const Dtree& input,
                                          const TRResultPtr& result,
                                          const PatternType& pattern,
                                          const TM& tm,
                                          std::map<std::string, double>& t_exp_freq,
                                          std::map<std::string, std::map<std::string, double> >& t_lm_exp_freq,
                                          const PartialTranslationFactory& partial_translation_factory,
                                          const LexicalTranslationScorer& lex_trans_table) const;

    Hypothesis *create_empty_hypothesis(const PatternType pattern, 
            const PartialTranslationFactory& partial_translation_factory) const;
    
    bool is_punctuation(std::string p) const {
        return tm_processor_source_->is_punctuation(p);
    }

    bool is_skip_translation(const DtreeNode *w) const {
        return tm_processor_pair_->is_skip_translation(w);
    }
    
  private:
    DefaultDependManager default_dpnd_;
    TMProcessorSource*   tm_processor_source_;
    TMProcessorTarget*   tm_processor_target_;
    TMProcessorPair*     tm_processor_pair_;
    PreProcessor         pre_processor_;

    void create_source_tm_processor(const Language& lang) {
        if (lang == JA) {
            tm_processor_source_ = new TMProcessorSourceJa;
        } else if (lang == EN) {
            tm_processor_source_ = new TMProcessorSourceEn;
        } else if (lang == ZH) {
            tm_processor_source_ = new TMProcessorSourceZh;
        } else {
            tm_processor_source_ = new TMProcessorSource;
        }
    }

    void create_target_tm_processor(const Language& lang) {
        // if (lang == JA) {
        //     tm_processor_target_ = new TMProcessorTargetJa;
        // } else if (lang == EN) {
        //     tm_processor_target_ = new TMProcessorTargetEn;
        // } else if (lang == ZH) {
        //     tm_processor_target_ = new TMProcessorTargetZh;
        // } else {
            tm_processor_target_ = new TMProcessorTarget;
        // }
    }

    void create_bilingual_tm_processor(const Language& s_lang, const Language& t_lang) {
        if (s_lang == JA && t_lang == EN) {
            tm_processor_pair_ = new TMProcessorPairJaEn;
        } else if (s_lang == EN && t_lang == JA) {
            tm_processor_pair_ = new TMProcessorPairEnJa;
        } else if (s_lang == JA && t_lang == ZH) {
            tm_processor_pair_ = new TMProcessorPairJaZh;
        } else if (s_lang == ZH && t_lang == JA) {
            tm_processor_pair_ = new TMProcessorPairZhJa;
        } else {
            tm_processor_pair_ = new TMProcessorPair;
        }
    }

    void set_body(const Dtree& input, const TM& tm, DataForTM* data_for_tm) const;
    void set_bond(const TM& tm, DataForTM* data_for_tm) const;
    void compute_child_list(const Dtree& input,
                            const TM& tm,
                            DataForTM* data_for_tm) const;
    ParentBondRelation find_parent_bond(const TM& tm, Hypothesis *hypothesis, DataForTM* data_for_tm) const;
    void align_child_bond(const TM& tm,
                          Hypothesis *hypothesis,
                          DataForTM* data_for_tm) const;
    void set_feature_value_for_child_bond(const TM& tm,
                                          Hypothesis *hypothesis,
                                          DataForTM* data_for_tm) const;
    void find_child_bond(const Dtree& input, const TM& tm, Hypothesis *hypothesis, DataForTM* data_for_tm) const;
    double calc_score(const FeatureVector& fv) const;
    ParentBondRelation self_bond_relation(const DtreeNode *self, std::vector<const DtreeNode*> bond) const;
    void make_t_string(const Dtree& input, const DtreeNode *t_ref, const TM& tm, DataForTM *data_for_tm, VS *t_string, 
            std::vector<Additional> *additional, VS *t_real_string, VS *t_lm_string,
            WordIndex *local_num, std::map<WordIndex, Dpnd> *global_dpnd,
            std::map<WordIndex, Dpnd> *global2local,
            std::map<WordIndex, POS> *bond_pos_map) const;
    void add_count_of_t_exp(VS *t_real_string,
                            VS *t_lm_string,
                            std::map<std::string, double> *t_exp_freq,
                            std::map<std::string, std::map<std::string, double> > *t_lm_exp_freq) const;

    bool check_tm_root(Hypothesis *hypothesis, DataForTM* data_for_tm) const;
    bool num_adaptation(const Dtree& input, const TM& tm, Hypothesis *hypothesis, DataForTM* data_for_tm) const;
    bool _num_adaptation(const Dtree& input, const TM& tm, const Alignment& align, Hypothesis *hypothesis, DataForTM* data_for_tm) const;

    bool check_compatibility(const Dtree& input, const TM& tm) const;
};

#endif
