/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "translation_manager.hpp"
#include <boost/algorithm/string.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/regex.hpp>
#include <boost/scoped_ptr.hpp>
#include <boost/scoped_array.hpp>
#include <glog/logging.h>
#include <algorithm>
#include <fstream>
#include <iostream>
#include <limits>
#include <map>
#include <sstream>
#include <string>
#include <vector>

#include "hypothesis_serialization.hpp"
#include "table_display_output.hpp"
#include "translation_result.hpp"
#include "utils.hpp"

#include "decoding/translation_log.hpp"

using boost::math::isfinite;

void set_color(std::ostream *stream, int color) {
    int max_colors = 7;
    *stream << "\033[3" << color % max_colors + 1 << "m";
}

void reset_color(std::ostream *stream) {
    *stream << "\033[0m";
}

//! Compares two translation candidates by their bleu ( if computed).
class CompareBleu {
  public:
    bool operator()(const SingleNBestTranslationResult &a, const SingleNBestTranslationResult &b) {
        assert(a.bleu != -std::numeric_limits<double>::max());
        assert(b.bleu != -std::numeric_limits<double>::max());
        return (a.bleu > b.bleu);
    }
};

/*
void OnlineTuner::process_nbest(NBestTranslationResult* nbest, int num_sentence) {
        LOG_IF(FATAL, nbest->size() == 0) << "bug?";
        seen_sentences_++;

        ModifiedNgramPrecisionComputer ngram_computer(Params::Translator::reference_translations[num_sentence]);


        BOOST_FOREACH(SingleNBestTranslationResult& trans_info, *nbest) { // NOLINT
            VS segmented_translation = segment_sentence_along_spaces(trans_info.t_string_flat);
            //trans_info.bleu = ngram_computer.compute_BLEU(segmented_translation, true);
            trans_info.bleu = rolling_bleu_.current_bleu_info().get_delta_bleu(ngram_computer, segmented_translation);
        }

        SingleNBestTranslationResult best_model_trans_info = (*nbest)[0];

        VS segmented_best_translation = segment_sentence_along_spaces(best_model_trans_info.t_string_flat);
        rolling_bleu_.add_sentence(ngram_computer, segmented_best_translation);

        //std::cerr << "\r\t" << seen_sentences_ << ": Rolling bleu: " << rolling_bleu_;

        double learning_rate = 1.0/sqrt(seen_sentences_);

        std::stringstream ss;
        ss << rolling_bleu_;
        std::string message_bleu = ss.str();
        move(0, 0);
        clrtoeol(); // clear line
        printw("%i: %s %f", seen_sentences_, message_bleu.c_str(), learning_rate);
        refresh();

        CompareBleu compare_score;
        std::sort(nbest->begin(), nbest->end(), compare_score);
        SingleNBestTranslationResult best_bleu_trans_info = (*nbest)[0];

        int num_line = 1;
        BOOST_FOREACH(WeightVector::value_type& name_val, current_weight_) { // NOLINT
            num_line++;

            double best_model_val = 0;
            double best_bleu_val = 0;
            if (best_model_trans_info.features.count(name_val.first) != 0) {
                best_model_val = best_model_trans_info.features.at(name_val.first);
            }
            if (best_bleu_trans_info.features.count(name_val.first) != 0) {
                best_bleu_val = best_bleu_trans_info.features.at(name_val.first);
            }

            name_val.second += learning_rate * (best_bleu_val - best_model_val);

            move(num_line, 0);
            clrtoeol();  // clear line
            printw("%s: %f", name_val.first.c_str(), name_val.second);
        }
        refresh();

//
//
//        prec_comp.update_counts(segmented_translation, &modified_ngram_counts_, &total_ngram_counts_, &total_size_,
//                &total_ref_size_);

        //if (Params::Translator::display_progress) {
            //std::cerr << "\r\t\tTranslated " << nb_translated_ << " sentences with BLEU: " << compute_bleu_string() << std::flush;
        //}

        //std::cerr << "got translation\n";
        delete nbest;
    }
*/

void output_blank_sentence(std::ostream *stream, int sentence_id, const WeightVector &weights) {
    LOG(WARNING) << "Skipped sentence. Generating blank translation.";
    if (Params::Global::output_mode == "eval") {
        *stream << std::endl;
    } else {
        *stream << sentence_id << " ||| ";
        *stream << ""; // blank translation
        *stream << " ||| f= ";
        for (WeightVector::const_iterator it = weights.begin();
                it != weights.end(); it++) {
            *stream << "0 ";
        }
        *stream << " ||| ";
        *stream << "0.0"; // blank score
        *stream << std::endl;
    }
}

void print_to_stream_n_best_translations_in_tuning_format(std::ostream *stream,
        const NBestTranslationResult &n_best_translations,
        int sentence_id, const WeightVector& weights) {
    if (n_best_translations.size() == 0) {
        output_blank_sentence(stream, sentence_id, weights);
    }
    for (int i = 0; i < static_cast<int>(n_best_translations.size()) && 
            i < Params::Translator::n_best_length; i++) {
        *stream << sentence_id << " ||| ";
        std::map<const Hypothesis*, int> color_map;
        *stream << n_best_translations[i].t_string_flat;
        if (Params::Translator::verbose_features) {
            *stream << " ||| ";
            for (WeightVector::const_iterator it = weights.begin();
                    it != weights.end(); it++) {
                *stream << it->first << "= ";
                double val_feature = 0;
                if (n_best_translations[i].features.count(it->first) != 0) {
                    val_feature = n_best_translations[i].features.at(it->first);
                    LOG_IF(FATAL, !boost::math::isfinite(val_feature)) << "Feature " << it->first << " has non finite value";
                }
                if (Params::Features::ignored_features.count(it->first)) {
                    *stream << "0 ";
                } else {
                    *stream << val_feature << " ";
                }
            }
        } else {
            *stream << " ||| f= ";
            for (WeightVector::const_iterator it = weights.begin();
                    it != weights.end(); it++) {
                double val_feature = 0;
                if (n_best_translations[i].features.count(it->first) != 0) {
                    val_feature = n_best_translations[i].features.at(it->first);
                }
                if (Params::Features::ignored_features.count(it->first)) {
                    *stream << "0 ";
                } else {
                    *stream << val_feature << " ";
                }
            }
        }
        *stream << " ||| ";
        *stream << n_best_translations[i].score;
        if (n_best_translations[i].bleu != -std::numeric_limits<double>::max()) {
            *stream << " " << n_best_translations[i].bleu;
        }
        *stream << std::endl;
        if (Params::Global::output_mode == "tuning-rerank") {
            process_optionals_and_print(n_best_translations[i], stream);
        }
    }
}

void print_to_stream_n_best_translations_in_eval_format(std::ostream *stream,
        const NBestTranslationResult &n_best_translations) {
    if (n_best_translations.size() > 0) {
        *stream << n_best_translations[0].t_string_flat << std::endl;
        // if (n_best_translations[0].features.count("sibling") && n_best_translations[0].features.at("sibling")) {
            // LOG(WARNING) << "Sib: " << n_best_translations[0].t_string_flat << std::endl;
        // } else {
            // LOG(WARNING) << "no sib" << "\n";
        // }
    } else {
        *stream << std::endl; //no translation found: we print an empty line
        LOG(WARNING) << "Warning: No Translation Found.";
    }
}

void _translate_and_output(const Translator& translator_, InitialHypothesesStorage *cand, 
                           const std::vector<AugmentedPattern>& augmented_patterns, size_t input_size, 
                           size_t num_translated, 
                           std::ostream* output_stream, TableDisplayOutput* tdout, t_log::TranslationLog* translation_log, 
                           const VS* reference = NULL,
                           const WeightVector* specified_weights = NULL) {
    size_t nb_initial_patterns = cand->nb_patterns();
    if (Params::Global::output_mode == "hypothesis" || Params::Global::output_mode == "hypo+eval") {
        // std::stringstream ss;
        
        if (true) { // Params::Translator::hypothesis_output_prefix.size() != 0) {
            std::string hypothesis_file_id = boost::lexical_cast<std::string>(num_translated);
            if (static_cast<int>(hypothesis_file_id.size()) < Params::Translator::leading_zeros) {
                int nb_zeros = Params::Translator::leading_zeros - hypothesis_file_id.size();
                std::string zero_prefix(nb_zeros, '0');
                hypothesis_file_id = zero_prefix + hypothesis_file_id;
            }
            std::string output_filename = Params::Translator::hypothesis_output_prefix + hypothesis_file_id + ".hyp";
            std::ofstream outfile(output_filename.c_str());
            LOG_IF(FATAL, !outfile) << "Could not open hypothesis output file '" << output_filename <<"'\n";
            HumanReadableHypothesisSerialization::serialize_translation_candidate(*cand, translator_.get_weights(), augmented_patterns, input_size, outfile);
        } else {
            HumanReadableHypothesisSerialization::serialize_translation_candidate(*cand, translator_.get_weights(), augmented_patterns, input_size, *output_stream);
        }
        LOG(INFO) << "Completed hypothesis output of sentence " << num_translated << ".";
    } 
    if (Params::Global::output_mode != "hypothesis") {
//        if (Params::Global::output_mode == "debug") {
//            cand->compute_initial_hypothesis_to_pattern_map();
//        }
        LOG(INFO) << "Starting translation of sentence " << num_translated << ".";
        NBestTranslationResult n_best_translations;
//        if (Params::Translator::use_oracle_decoding) {
//            LOG_IF(FATAL, num_translated >= Params::Translator::reference_translations.size())
//                        << "Not enough references provided: need at least " << (num_translated + 1)
//                        << " but only " << Params::Translator::reference_translations.size() << "found.\n";
//            n_best_translations = translator_.translate_from_translation_candidates_with_oracle(
//                                augmented_patterns, input_size, cand, tdout,
//                                Params::Translator::reference_translations[num_translated]);
//        } else {
        n_best_translations = translator_.translate_from_translation_candidates(
                augmented_patterns, input_size, cand, tdout, translation_log, reference, specified_weights);
//        }
        if (Params::Global::output_mode == "tuning") {
            print_to_stream_n_best_translations_in_tuning_format(output_stream, n_best_translations,
                    Params::Global::sentence_id + num_translated, translator_.get_weights());
        } else if (Params::Global::output_mode == "tuning-rerank") {
            print_to_stream_n_best_translations_in_tuning_format(output_stream, n_best_translations,
                    Params::Global::sentence_id + num_translated, translator_.get_weights());
        } else if (Params::Global::output_mode == "eval" || Params::Global::output_mode == "hypo+eval" || Params::Global::output_mode == "bleu_eval") {
            print_to_stream_n_best_translations_in_eval_format(output_stream, n_best_translations);
        } else { 
            LOG(WARNING) << "Unrecognized output format.";
        }
        LOG(INFO) << "Completed translation of sentence " << num_translated << ".";
        LOG(INFO) << "Nb Patterns:" << nb_initial_patterns << " -> " << cand->nb_patterns();
        if (Params::Global::web) {
            for (int i = 0; i < static_cast<int>(n_best_translations.size()); i++) {
                tdout->output_for_table_display(i, n_best_translations[i]);
                if (!Params::Global::web_all) {
                    break;
                }
            }
            tdout->print_buffer(output_stream);
        }
    }
}

void load_hypothesis_file_into_string(const std::string& filename, std::string* contents) {
    std::ifstream hypothesis_file(filename.c_str(), std::ios::in | std::ios::binary);
    LOG_IF(FATAL, !hypothesis_file) << "Could not open hypothesis filename '" << filename << "'\n";
    LOG(INFO) << "Loading hypothesis from '" << filename << "'...\n";
    hypothesis_file.seekg(0, std::ios::end);
    contents->resize(hypothesis_file.tellg());
    hypothesis_file.seekg(0, std::ios::beg);
    hypothesis_file.read(&(*contents)[0], contents->size());
    hypothesis_file.close();
    LOG(INFO) << "Done loading hypothesis from '" << filename << "'. Starting parsing of hypothesis.\n";
}

char* load_hypothesis_file_into_cstring(const std::string& filename) {
    FILE* f = fopen(filename.c_str(), "rb");
    LOG_IF(FATAL, f == NULL) << "Could not open hypothesis filename '" << filename << "'\n";
    fseek(f, 0, SEEK_END);
    size_t length = ftell(f);
    fseek(f, 0, SEEK_SET);    
    LOG(INFO) << "Loading hypothesis from '" << filename << "'...\n";
    char* cstring = new char[length+1];
    fread(cstring, 1, length, f);
    cstring[length] = '\0';
    fclose(f);
    LOG(INFO) << "Done loading hypothesis from '" << filename << "'. Starting parsing of hypothesis.\n";
    return cstring;
}
                


///////////////////////////////////////////////////////////////
// The following functions are for multi-thread translations //
///////////////////////////////////////////////////////////////

void SerializeToOutputDecorator::process(OutputMessageInterface* msg, int num) {
    assert(num >= static_cast<int>(nb_outputed));
    if (num == static_cast<int>(nb_outputed)) {
        out->process(msg, num);
        nb_outputed++;
        try_to_flush_queue();
    } else {
        queue.push(MessageWithPriority(num, msg));
    }
    if (Params::Translator::display_progress) {
        std::cerr << "\r\t\tTranslated sent num " << num << " (" << nb_outputed << " outputed)." << std::flush;
    }
}

void SerializeToOutputDecorator::try_to_flush_queue() {
    while (queue.size() > 0 && queue.top().priority == nb_outputed) {
        out->process(queue.top().msg, queue.top().priority);
        queue.pop();
        nb_outputed++;
    }
}

void AsyncThreadSafeOutputDecorator::process(OutputMessageInterface* msg, int num) {
    if (Params::Translator::display_progress) {
        std::cerr << "\rD:" << num << std::flush;
    }
    io_.post(strand_.wrap(boost::bind(&OutputHandlerInterface::process, out_.get(), msg, num)));
}

void AsyncThreadSafeOutputDecorator::process_nbest(NBestTranslationResult* nbest, int num_sentence) {
    io_.post(strand_.wrap(boost::bind(&OutputHandlerInterface::process_nbest, out_.get(), nbest, num_sentence)));
}

WeightVector* AsyncThreadSafeOutputDecorator::get_current_weight_copy() const {
    // We do not wrap this one, so it has to be called in a thread-safe context ....
    return out_->get_current_weight_copy();
}

// DecodingWorker::Pointer DecodingWorker::create(size_t  num_sentence,
                                                     // const Translator&                 translator,
                                                     // SerializeToOutputAsyncThreadSafe& output,
                                                     // size_t                            id_sentence_for_nbest,
                                                     // const std::string& hyp_filename) {
                                                     
    // return boost::make_shared<DecodingWorker>(num_sentence, translator, boost::ref(output), id_sentence_for_nbest, hyp_filename);
// }

//class OutputHandlerInterface {
//    virtual ~OutputHandlerInterface(){}
//    virtual void process(const std::string& msg, int num_sentence) = 0;
//};

void BleuEvaluator::process(OutputMessageInterface* translation, int num) {
    assert(num < static_cast<int>(Params::Translator::reference_translations.size()));
    nb_translated_++;
    VS segmented_translation = segment_sentence_along_spaces(translation->get_main_message());

    ModifiedNgramPrecisionComputer prec_comp(Params::Translator::reference_translations[num]);

    bleu_info_.update_with_sentence(prec_comp, segmented_translation);
//    prec_comp.update_counts(segmented_translation, &modified_ngram_counts_, &total_ngram_counts_, &total_size_,
            //&total_ref_size_);

    //if (Params::Translator::display_progress) {
        std::cerr << "\r\t\tTranslated " << nb_translated_ << " sentences with BLEU: " << bleu_info_ << std::flush;
    //}

    if (alternative_output_) {
        alternative_output_->process(translation, num);
    } else {
        delete translation;
    }
}

//std::string BleuEvaluator::compute_bleu_string() const {
//    std::stringstream ss;
//    double current_bleu = compute_BLEU_from_counts(modified_ngram_counts_, total_ngram_counts_, total_ref_size_, total_size_);
//    ss << current_bleu << "(";
//    for (int n = 0; n < 4; ++n) {
//        ss << modified_ngram_counts_[n] << "/" << total_ngram_counts_[n] << " ";
//    }
//    ss << ") [" << total_size_ << ":" << total_ref_size_ <<"]";
//    return ss.str();
//}

TranslationWorker::Pointer TranslationWorker::create(size_t                            num_sentence,
                                                     const Translator&                 translator,
                                                     OutputHandlerInterface&                  output,
                                                     size_t                            id_sentence_for_nbest,
                                                     WeightVector*                     specified_weights) {
    return boost::make_shared<TranslationWorker>(num_sentence, translator, boost::ref(output), id_sentence_for_nbest, specified_weights);
}

void TranslationWorker::start() {
    InitialHypothesesStorage cand;
    std::vector<AugmentedPattern> augmented_patterns;
    size_t input_size;
    
    TableDisplayOutput tdout;
    if (Params::Global::web) {
        tdout.initialize(id, p_input_sent.get());
    }
    
    t_log::TranslationLog translation_log;

    int time0 = t_log::get_time_in_ms();
    std::vector<TRResultPtr> tr_results_life_maintaining_references;
    if (Params::Global::keep_tms_in_memory) {
        tr_results_life_maintaining_references = translator.generate_translation_candidates(*p_input_sent, &cand);
    } else {
        translator.generate_translation_candidates(*p_input_sent, &cand);
    }

    translation_log.set_tr_retrieving_time(time0);

    augmented_patterns = Translator::build_augmented_patterns(cand, *p_input_sent);
    input_size = p_input_sent->size();

    SimpleOutputMessage* ouput_msg = new SimpleOutputMessage();
    std::stringstream ss;
    _translate_and_output(translator, &cand, augmented_patterns, input_size, num_sentence, ouput_msg->main_stream(),
                                &tdout, &translation_log);
    output.process(ouput_msg, num_sentence);
}

void TranslationWorker::start_decoding() {
    //if (Params::Translator::display_progress) {
    //    std::cerr << "\rStart decoding " << id << " (" << num_sentence << ")" << std::flush;
    //}
    InitialHypothesesStorage cand;
    std::vector<AugmentedPattern> augmented_patterns;
    size_t input_size;

    std::string filename = Params::Translator::input_filename_list[num_sentence];

    // std::string contents;
    // load_hypothesis_file_into_string(filename, &contents);

    // HumanReadableHypothesisSerialization::deserialize_translation_candidate(contents, &cand, &augmented_patterns, &input_size);
    
    t_log::TranslationLog translation_log;

    int start_loading_time = t_log::get_time_in_ms();
    boost::scoped_array<char> content(load_hypothesis_file_into_cstring(filename));
    HumanReadableHypothesisSerialization::deserialize_translation_candidate(content.get(), &cand, &augmented_patterns, &input_size);
    translation_log.set_loading_time(start_loading_time);
    
//    bool skip = false; // TODO(fabien/john/nakazawa): remove this
//    if (cand.total_nb_hypotheses() > 100000) {
//        LOG(WARNING) << "Too many hypotheses (" << cand.total_nb_hypotheses() << ") loaded from file.";
//        skip = true;
//    }

    const VS* reference = NULL;
    if (Params::Translator::reference_translations.size() > 0) {
        LOG_IF(FATAL, num_sentence >=Params::Translator::reference_translations.size()) <<
                "Not enough reference translations:" << num_sentence << " >= " << Params::Translator::reference_translations.size();
        reference = &Params::Translator::reference_translations[num_sentence];
    }


    int start_decoding_time = t_log::get_time_in_ms();
    if (Params::Global::output_mode == "online_tuning") {
        NBestTranslationResult n_best_translations = translator.translate_from_translation_candidates(
                    augmented_patterns, input_size, &cand, NULL, &translation_log, reference, specific_weight_vector_.get());
        output.process_nbest(new NBestTranslationResult(n_best_translations), num_sentence);
        translation_log.set_total_decoding_time(start_decoding_time);
    } else {
        SimpleOutputMessage* output_msg = new SimpleOutputMessage; // TODO(fabien): use std::unique_ptr when switching to c++11
//        if (skip) {
//            output_blank_sentence(output_msg->main_stream(), Params::Global::sentence_id + num_sentence, translator.get_weights());
//        } else {
            _translate_and_output(translator, &cand, augmented_patterns, input_size, num_sentence, output_msg->main_stream(), NULL, &translation_log,
                    reference, specific_weight_vector_.get());
//        }

        translation_log.set_total_decoding_time(start_decoding_time);

        if (Params::Global::time_log_file.size() > 0) {
            *(output_msg->log_stream()) << num_sentence << ": " << filename << " " << translation_log << std::endl;
        }

        *(output_msg->translation_log()) = translation_log; // TODO(fabien): improve/optimize this

        output.process(output_msg, num_sentence);
    }
}

//////////////////////////////////////
using boost::asio::ip::tcp;

class tcp_connection : public boost::enable_shared_from_this<tcp_connection> {
  private:
    boost::asio::streambuf buffer_;
    tcp::socket socket_;
    std::string message_;
    
    const Translator& translator_;
    size_t num_sentence_;
    
  public:
    typedef boost::shared_ptr<tcp_connection> pointer;

    static pointer create(boost::asio::io_service& io_service, const Translator& translator) {
        return pointer(new tcp_connection(io_service, translator));
    }

    tcp::socket& socket() {
        return socket_;
    }

    void start() {
        boost::asio::async_read_until(socket_, buffer_, "</sentence>",
            boost::bind(&tcp_connection::handle_read, shared_from_this(),
            boost::asio::placeholders::error,
            boost::asio::placeholders::bytes_transferred));
    }

    virtual ~tcp_connection() {
        LOG(INFO) << "Connection destroyed";
    }

  private:
    tcp_connection(boost::asio::io_service& io_service, const Translator& translator)
        : socket_(io_service), translator_(translator), num_sentence_(0) {}

    void handle_write(boost::shared_ptr<boost::asio::streambuf> p_message,
                    const boost::system::error_code& error, size_t bytes_transferred) {
        //passing p_message as an argument here, so that it stays alive until the end of the copy
        //LOG(INFO) << "Finish transfered" << std::endl; 
        if (error) {
            LOG(WARNING) << "Send error: " << error.message();
        }
        start();
    }
    
    void handle_read(const boost::system::error_code& error, size_t bytes_transferred) {
        if (error == boost::asio::error::eof) {
            LOG(INFO) << "EOF. End of connection";
            return;
        }
        LOG(INFO) << "Read " << bytes_transferred << " bytes.";
        
        std::istream is(&buffer_);
        
        boost::shared_ptr<boost::asio::streambuf> p_message = boost::make_shared<boost::asio::streambuf>();
        //message=boost::asio::streambuf();
        std::ostream message_stream(p_message.get());

        InputHandler input_handler(&is);
        std::string best_result = "";
        double best_score = 0;
        while (1) {
            boost::scoped_ptr<Dtree> p_input_sent;
            p_input_sent.reset(new Dtree());
            std::string id;
            bool could_read_sentence = input_handler.read_one_sentence(p_input_sent.get(), &id);
            if (could_read_sentence) {
                InitialHypothesesStorage cand;
                std::vector<AugmentedPattern> augmented_patterns;
                size_t input_size;
            
                TableDisplayOutput tdout;
                if (Params::Global::web) {
                    tdout.initialize(id, p_input_sent.get());
                }
            
                t_log::TranslationLog translation_log;

                std::vector<TRResultPtr> tr_results_life_maintaining_references;
                if (Params::Global::keep_tms_in_memory) {
                    tr_results_life_maintaining_references = translator_.generate_translation_candidates(*p_input_sent, &cand);
                } else {
                    translator_.generate_translation_candidates(*p_input_sent, &cand);
                }
                augmented_patterns = Translator::build_augmented_patterns(cand, *p_input_sent);
                input_size = p_input_sent->size();
                std::stringstream ss;
                // _translate_and_output(translator_, &cand, augmented_patterns, input_size, num_sentence_, &message_stream, &tdout, &translation_log);
                _translate_and_output(translator_, &cand, augmented_patterns, input_size, num_sentence_, &ss, &tdout, &translation_log);
                num_sentence_++;
                // message_stream << "\nEOS\n";
                ss << "\nEOS\n";
                std::string line_buf;
                std::getline(ss, line_buf);
                std::cerr << line_buf << std::endl;
                VS tmp = split(line_buf, " ||| ");
                double score = boost::lexical_cast<double>(tmp[3]);
                if (best_score == 0 || best_score < score) {
                    best_result = ss.str();
                    best_score = score;
                }
                if (!Params::Translator::use_nbest_parses) {
                    break;
                }
            } else {
                break;
            }
        }
        message_stream << best_result;

        // synchronous write code
        boost::system::error_code send_error;
        boost::asio::write(socket_, *p_message, boost::asio::transfer_all(), send_error);
        LOG(INFO) << "Finish transfered"; 
        if (send_error) {
            LOG(ERROR) << "Send error: " << send_error.message();
        }
        start(); //process next request
    }
};

//////////////////////////////////////

class tcp_server {
  public:
    tcp::acceptor& acceptor_;
    const Translator& translator_;

  public:
    tcp_server(boost::asio::io_service& io_service, tcp::acceptor& acceptor, const Translator& translator)
     : acceptor_(acceptor), translator_(translator) {
        start_accept();
    }

    virtual ~tcp_server() {
        LOG(INFO) << "Server destroyed";
    }

  private:
    void start_accept() {
        tcp_connection::pointer new_connection =
            tcp_connection::create(acceptor_.get_io_service(), translator_);

        acceptor_.async_accept(new_connection->socket(),
            boost::bind(&tcp_server::handle_accept, this, new_connection, boost::asio::placeholders::error));
    }

    void handle_accept(tcp_connection::pointer new_connection, const boost::system::error_code& error) {
        if (!error) {
            LOG(INFO) << "Starting connection...";
            new_connection->start();
        } else {
            LOG(ERROR) << "Asynchronous connection error : " << error.message();
        }
        start_accept();
    }
};
//////////////////////////////////////

std::vector<int> get_random_permutation_vector(int nb_sent) {
    std::vector<int> vect(nb_sent);
    for (int i = 0; i < nb_sent; ++i) {
        vect[i] = i;
    }
    std::random_shuffle(vect.begin(), vect.end());
    return vect;
}

void AsyncTranslationManager::post_read_input() {
    io.post(strand_input.wrap(boost::bind(&AsyncTranslationManager::read_input_handler, this)));
}

void AsyncTranslationManager::read_input_handler() { 
    // <-not inherently thread-safe, need to be made "critical"
    TranslationWorker::Pointer t_worker = TranslationWorker::create(nb_sentences_read, translator_, output, nb_sentences_read);
    
    bool could_read_sentence = input_handler_->read_one_sentence(t_worker->p_input_sent.get(), &t_worker->id);

    if (could_read_sentence) {
        nb_sentences_read++;
        
        // For correct working with GOOGLE_STRIP_LOG
        #ifndef GOOGLE_STRIP_LOG 
        t_worker->p_input_sent->print(&LOG(INFO));
        #else
        if (GOOGLE_STRIP_LOG < 1) t_worker->p_input_sent->print(&LOG(INFO));
        #endif
        
        io.post(boost::bind(&TranslationWorker::start, t_worker));
        
        post_read_input();
    }
}

void AsyncTranslationManager::post_decoding_job() { 
    io.post(strand_input.wrap(boost::bind(&AsyncTranslationManager::post_decoding_job_thread_unsafe, this)));
}

void AsyncTranslationManager::post_decoding_job_thread_unsafe() {
    // <-not inherently thread-safe, need to be made "critical"
    if (nb_sentences_read < Params::Translator::input_filename_list.size()) {
        int shuffled_num_sentence;
        if (Params::Translator::shuffle_translation_order) {
            shuffled_num_sentence = shuffled_sentence_num_.at(nb_sentences_read);
        } else {
            shuffled_num_sentence = nb_sentences_read;
        }
        TranslationWorker::Pointer t_worker = TranslationWorker::create(shuffled_num_sentence, translator_, output, shuffled_num_sentence);
        nb_sentences_read++;
        io.post(boost::bind(&TranslationWorker::start_decoding, t_worker));
        post_decoding_job();
    }
}

void AsyncTranslationManager::post_online_tuning_job() {
    io.post(strand_input.wrap(boost::bind(&AsyncTranslationManager::post_online_tuning_job_thread_unsafe, this)));
}

void AsyncTranslationManager::post_online_tuning_job_thread_unsafe() {
/*
    // <-not inherently thread-safe, need to be made "critical"
    if (nb_sentences_read >= Params::Translator::input_filename_list.size()) {
        nb_sentences_read = 0;
        epoch_ += 1;
        shuffled_sentence_num_ = get_random_permutation_vector(Params::Translator::input_filename_list.size());
    }
    int shuffled_num_sentence;
    if (Params::Translator::shuffle_translation_order) {
        shuffled_num_sentence = shuffled_sentence_num_.at(nb_sentences_read);
    } else {
        shuffled_num_sentence = nb_sentences_read;
    }

    mvprintw(1, 10, "Epoch: %i  Sent. used this epoch: %i", epoch_, nb_sentences_read);

    TranslationWorker::Pointer t_worker = TranslationWorker::create(shuffled_num_sentence, translator_, output, shuffled_num_sentence,
                    output.get_current_weight_copy());
    nb_sentences_read++;
    io.post(boost::bind(&TranslationWorker::start_decoding, t_worker));
    post_online_tuning_job();
*/
}


void AsyncTranslationManager::translate_all(int nb_threads) {
    boost::scoped_ptr<tcp::acceptor> acceptor;
    boost::scoped_ptr<tcp_server> server;

    if (Params::Translator::server_mode) {
        try {
            acceptor.reset(new tcp::acceptor(io, tcp::endpoint(tcp::v4(), Params::Translator::port)));
            server.reset(new tcp_server(io, *acceptor, translator_));
            LOG(INFO) << "Ready to answer queries!";
        } catch(const std::exception& e) {
            LOG(FATAL) << "Exception: " << e.what();
            return;
        }
    } else {
       if (Params::Global::input_mode == "hypothesis") {
           if (Params::Global::output_mode == "online_tuning") {
               post_online_tuning_job();
           } else {
               post_decoding_job();
           }
       }  else {
           post_read_input();
       }
    }
    
    if (nb_threads > 1) {
        boost::thread_group threads;
        for (int i = 0; i < nb_threads; i++) {
            threads.create_thread(boost::bind(&boost::asio::io_service::run, &io));
        }
        io.run();
        threads.join_all();
    } else {
        io.run();
    }
}

