/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __TRANSLATION_MANAGER_HPP__
#define __TRANSLATION_MANAGER_HPP__

//#include <ncurses.h>
#include <boost/asio.hpp>
#include <boost/foreach.hpp>
#include <boost/make_shared.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/thread/thread.hpp>
#include <iostream>
#include <sstream>
#include <string>
#include <queue>
#include <vector>
#include "translator.hpp"
#include "score_computer.hpp"
#include "decoding/translation_log.hpp"
#include "translation_result.hpp"

class OutputMessageInterface {
  public:
    virtual ~OutputMessageInterface() {}
    virtual const std::string get_main_message() const = 0;
    virtual const std::string get_log_message() const {
        LOG(FATAL) << "This method should be either overridden or not called";
        return std::string();
    }

    virtual const t_log::TranslationLog& get_translation_log() const {
        LOG(FATAL) << "This method should be either overridden or not called";
    }
};

static const std::string kEmptyLogMessage = "NO LOG!";

class SimpleOutputMessage : public OutputMessageInterface {
  public:
    // explicit SimpleOutputMessage(std::string* msg): main_msg_(msg) {}
//    explicit SimpleOutputMessage(std::string* msg, std::string* log_msg): main_msg_(msg), log_msg_(log_msg) {}

    const std::string get_main_message() const {
        return main_stream_.str();
    }

    const std::string get_log_message() const {
        return log_stream_.str();
    }

    std::ostream* main_stream() {
        return &main_stream_;
    }

    std::ostream* log_stream() {
        return &log_stream_;
    }

    t_log::TranslationLog* translation_log() {
        return &translation_log_;
    }

    virtual const t_log::TranslationLog& get_translation_log() const {
        return translation_log_;
    }

  private:
    std::stringstream main_stream_;
    std::stringstream log_stream_;
    t_log::TranslationLog translation_log_;
    //boost::scoped_ptr<std::string> main_msg_;
    //boost::scoped_ptr<std::string> log_msg_;
};

/*!
 * The Interface OutputHandlerInterface is designed to process the result of a translation.
 * The Translation manager object will send the result of a translation through the process method of a
 * OutputHandlerInterface object. The most basic type of processing is to just send the result to std::cout.
 */
class OutputHandlerInterface {
  public:
    virtual ~OutputHandlerInterface() {}

    /*! Process an output result for sentence num_sentence.
     * process get ownership of the string pointer msg (ie. it should free it afer processing it)
     */
    virtual void process(OutputMessageInterface* msg, int num_sentence) {
        LOG(FATAL) << "This method should be either overridden or not called";
    }

    /*! The following two methods will be used in online tuning*/
    virtual void process_nbest(NBestTranslationResult* nbest, int num_sentence) {
        LOG(FATAL) << "This method should be either overridden or not called";
    }

    virtual WeightVector* get_current_weight_copy() const {
        LOG(FATAL) << "This method should be either overridden or not called";
        return NULL;
    }

    virtual void log(std::string* msg, int num_sentence) {
        LOG(FATAL) << "This method should be either overridden or not called";
    }
};

/*
class OnlineTuner: public OutputHandlerInterface {
  public:
    explicit OnlineTuner(const WeightVector& initial_weight): current_weight_(initial_weight),
                rolling_bleu_(200), seen_sentences_(0) {
        initscr(); //initialize ncurse
    }

    ~OnlineTuner() {
        endwin(); // clean ncurse
    }

    void process_nbest(NBestTranslationResult* nbest, int num_sentence);

    WeightVector* get_current_weight_copy() const {
        return new WeightVector(current_weight_);
    }

   private:
    WeightVector current_weight_;
    RollingBleuInfo rolling_bleu_;
    int seen_sentences_;
};
*/

class StreamPrinter: public OutputHandlerInterface {
  public:
    explicit StreamPrinter(std::ostream* out): out_(out), log_(NULL) {
        sum_log_ = t_log::TranslationLog::zero_translation_log();
    }

    StreamPrinter(std::ostream* out, std::ostream* log): out_(out), log_(log) {
        sum_log_ = t_log::TranslationLog::zero_translation_log();
    }

    ~StreamPrinter() {
        if (log_ != NULL) {
            *log_ << "Total: " << sum_log_ << std::endl;
        }
    }

    void process(OutputMessageInterface* msg, int num_sentence) {
        *out_ << msg->get_main_message();
        if (log_ != NULL) {
            *log_ << msg->get_log_message();
            sum_log_.add_in(msg->get_translation_log());
        }
        delete msg;
    }

  private:
    std::ostream* out_;
    std::ostream* log_;
    t_log::TranslationLog sum_log_;
};

//inline OutputHandlerInterface* get_stream_printer(std::ostream* stream) {
//    return new StreamPrinter(stream);
//}

class BleuEvaluator: public OutputHandlerInterface {
  public:
    explicit BleuEvaluator(OutputHandlerInterface *alternative_output = NULL) :
        nb_translated_(0), alternative_output_(alternative_output) {}

    void process(OutputMessageInterface* translation, int num);

    //std::string compute_bleu_string() const;

  private:
//    std::vector<int> modified_ngram_counts_;
//    std::vector<int> total_ngram_counts_;
//    int total_size_;
//    int total_ref_size_;
    BleuInfo bleu_info_;
    int nb_translated_;
    boost::scoped_ptr<OutputHandlerInterface> alternative_output_;
};

inline OutputHandlerInterface* get_bleu_evaluator(OutputHandlerInterface* alternative_output = NULL) {
    return new BleuEvaluator(alternative_output);
}

/// A translation manager organize the inputs and outputs around a translation
//class SimpleTranslationManager {
//  public:
//    SimpleTranslationManager(InputHandlerInterface *ih, const Translator& translator, OutputHandlerInterface* output) :
//        input_handler_(ih), translator_(translator), output_(output) {}
//    void translate_all();
//  private:
//    InputHandlerInterface* input_handler_;
//    const Translator&      translator_;
//    boost::scoped_ptr<OutputHandlerInterface> output_;
//};

/////////////////////////////////////////////////////////////
// The following classes are for multi-thread translations //
/////////////////////////////////////////////////////////////

struct MessageWithPriority {
    size_t      priority;
    OutputMessageInterface* msg;
    MessageWithPriority(size_t priority, OutputMessageInterface* msg):
        priority(priority), msg(msg) {}
    bool operator<(const MessageWithPriority& other) const {
        return priority > other.priority; //we actually want to pop lower priorities first
    }
};

class SerializeToOutputDecorator: public OutputHandlerInterface {
  public:
    explicit SerializeToOutputDecorator(OutputHandlerInterface *out): out(out), nb_outputed(0) {}
    
    void process(OutputMessageInterface* msg, int num);
    
  private:
    boost::scoped_ptr<OutputHandlerInterface> out;
    size_t nb_outputed;
    std::priority_queue<MessageWithPriority> queue;

    void try_to_flush_queue();
};

//class OutputInterface {
//  public:
//    void post_msg(const std::string& msg, size_t num);
//};

// protected inheritance cause we do not want the methods of SerializeToOutput
// to be called directly
class AsyncThreadSafeOutputDecorator: public OutputHandlerInterface {
  public:
    AsyncThreadSafeOutputDecorator(OutputHandlerInterface *out, boost::asio::io_service &io) // NOLINT
            : io_(io), strand_(io), out_(out) {}
    void process(OutputMessageInterface* msg, int num);

    void process_nbest(NBestTranslationResult* nbest, int num_sentence);

    // Warning: we do not wrap this one, so it has to be called in a thread-safe context ....
    virtual WeightVector* get_current_weight_copy() const;

  private:
    boost::asio::io_service &io_;
    boost::asio::strand strand_;
    boost::scoped_ptr<OutputHandlerInterface> out_;
};



//// protected inheritance cause we do not want the methods of SerializeToOutput
//// to be called directly
//class SerializeToOutputAsyncThreadSafe: public SerializeToOutput, public OutputHandlerInterface {
//  public:
//    SerializeToOutputAsyncThreadSafe(std::ostream *out, boost::asio::io_service &io) // NOLINT
//            : SerializeToOutput(out), io_(io), strand_(io) {}
//    void process(const std::string& msg, size_t num);
//  private:
//    boost::asio::io_service &io_;
//    boost::asio::strand strand_;
//};

class TranslationWorker : public boost::enable_shared_from_this<TranslationWorker> {
  public:
    std::string id;
    // careful: a bit dangerous that p_input_sent is first initialized to an 
    // inconsistent Dtree --> never call start() before setting up p_input_sent;
    boost::shared_ptr<Dtree> p_input_sent;
    
    typedef boost::shared_ptr<TranslationWorker> Pointer;

    TranslationWorker(size_t num_sentence, const Translator& translator, OutputHandlerInterface& output,
            size_t id_sentence_for_nbest, WeightVector* specific_weight_vector = NULL) :
        p_input_sent(boost::make_shared<Dtree>()), 
        num_sentence(num_sentence), id_sentence_for_nbest(id_sentence_for_nbest),
        translator(translator), 
        output(output), specific_weight_vector_(specific_weight_vector) {}
    
    static Pointer create(size_t num_sentence, const Translator& translator, OutputHandlerInterface& output, size_t id_sentence_for_nbest,
                                    WeightVector* specified_weights = NULL);
    
    void start();
    void start_decoding();

  private:    
    size_t num_sentence;
    size_t id_sentence_for_nbest;
    const Translator& translator;
    OutputHandlerInterface &output;
    boost::scoped_ptr<WeightVector> specific_weight_vector_;
};
   
// class DecodingWorker : public boost::enable_shared_from_this<DecodingWorker> {
  // public:
    // typedef boost::shared_ptr<TranslationWorker> Pointer;

    // TranslationWorker(size_t num_sentence, const Translator& translator, SerializeToOutputAsyncThreadSafe& output, size_t id_sentence_for_nbest, const std::string& hyp_filename) : 
        // num_sentence(num_sentence), id_sentence_for_nbest(id_sentence_for_nbest),
        // translator(translator), 
        // hyp_filename(hyp_filename), output(output) {}
    
    // static Pointer create(size_t num_sentence, const Translator& translator, SerializeToOutputAsyncThreadSafe& output, size_t id_sentence_for_nbest, const std::string& hyp_filename);
    
    // void start();
  
  // private:    
    // size_t num_sentence;
    // size_t id_sentence_for_nbest;
    // const Translator& translator;
    // std::string hyp_filename;
    // SerializeToOutputAsyncThreadSafe &output;
// };

std::vector<int> get_random_permutation_vector(int nb_sent);
   
class AsyncTranslationManager {
  public:    
    AsyncTranslationManager(InputHandlerInterface *ih,
                            const Translator& tr,
                            OutputHandlerInterface *ost) :
        io(), input_handler_(ih), translator_(tr), nb_sentences_read(0),
        //careful with initialization order here...
        strand_input(io), output(ost, io),
        shuffled_sentence_num_(get_random_permutation_vector(Params::Translator::input_filename_list.size())),
        epoch_(0) {}
    
    void translate_all(int nb_threads = 1);
    //void decode_all(int nb_threads = 1);
  private:
    boost::asio::io_service           io;
    InputHandlerInterface*            input_handler_;
    const Translator&                 translator_;
    size_t                            nb_sentences_read;
    boost::asio::strand               strand_input;
    AsyncThreadSafeOutputDecorator    output;
    std::vector<int>                  shuffled_sentence_num_;
    int epoch_;
    
    void post_read_input();
    void read_input_handler();
    void post_decoding_job();
    void post_decoding_job_thread_unsafe();
    void post_online_tuning_job();
    void post_online_tuning_job_thread_unsafe();
};

#endif //__TRANSLATION_MANAGER_HPP__

