/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "translator.hpp"
//#include <ncurses.h>
#include <boost/foreach.hpp>
#include <boost/regex.hpp>
#include <boost/lexical_cast.hpp>
#include <cassert>
#include <map>
#include <string>
#include <utility>
#include <vector>
#include "params.hpp"

template<typename T>
std::ostream &operator<< (std::ostream &stream, const std::vector<T> &vect) {
    stream << "Vect(" << vect.size() << "/" << vect.capacity() << ")<";
    for (unsigned int i = 0; i < vect.size(); ++i) {
        stream << vect[i] << " ";
    }
    stream << ">";
    return stream;
}

void Translator::read_feature_weight(std::string filename) {
    std::vector<std::string> names_ordered_list;
    if (filename != "") {
        std::ifstream ist(filename.c_str());
        if (ist.is_open()) {
            LOG(INFO) << "Reading feature weight from " << filename << "... ";
            std::string line_buf;
            while (!ist.eof()) {
                getline(ist, line_buf);
                if (line_buf == "") {
                    continue;
                }
                bool ignore_feature = false;
                for (size_t i = 0; i < line_buf.size(); i++) {
                    if (i == 0 && line_buf[0] == '#') {
                        ignore_feature = true;
                    } else if (line_buf[i] == ' ') {
                        int start_i = ignore_feature ? 1 : 0;
                        std::string name = line_buf.substr(start_i, i - start_i);
                        double val = boost::lexical_cast<double>(line_buf.substr(i + 1));
                        if (ignore_feature) {
                            val = 0.0;
                            Params::Features::ignored_features.insert(name);
                            LOG(INFO) << "W (ignore): " << name << " = " << val;
                        } else {
                            LOG(INFO) << "W: " << name << " = " << val;
                        }
                        weights_.insert(std::pair<std::string, double>(name, val));
                        Params::Features::all_features.insert(name);
                        names_ordered_list.push_back(name);
                        break;
                    }
                }
            }
            LOG(INFO) << "Done reading feature weight file.";
        } else {
            LOG(FATAL) << "Error: can't open " << filename << ".";
        }
    }
    int i = 0;
    for (std::map<std::string, double>::const_iterator it = weights_.begin();
            it != weights_.end(); ++it) {
        if (it->first != names_ordered_list[i]) {
            LOG(FATAL) << "Weight file is not in alphabetical order.";
        }
        i++;
    }
}

std::vector<TRResultPtr> Translator::generate_translation_candidates(const Dtree& input_sent, InitialHypothesesStorage* cand) const {
    // Results must be stored like this to remain in the same scope as the
    // call to CombiSearch::search_tm_combination, which references the results.
    std::vector<TRResultPtr> results;
    BOOST_FOREACH(TRInterfacePtr tr, tr_list_) {
        if (Params::Translator::tr_result_filename != "") {
            // results.push_back(load_tr_result_from_file());
        } else {
            TreeRetrieveQuery tr_query = create_tree_retrieve_query(input_sent);
            LOG(INFO) << "Retrieving examples...";
            results.push_back(tr->retrieve(tr_query.first, tr_query.second));
            LOG(INFO) << "Done retrieving examples.";
        }
    }
    // add pseudo TM for OOVs
    add_pseudo_TM(results, input_sent, oov_handler_);
    //results.push_back(results.back());
    // calculate scores and find bonds for all TM
    PartialTranslationFactory partial_translation_factory;
    LOG(INFO) << "Creating initial Hypotheses...";
    create_initial_hypothesis(input_sent, results, cand, partial_translation_factory, tm_processor_.get());
    LOG(INFO) << "Done creating initial Hypotheses.";
    if (cand->has_empty_stack()) {
        LOG(FATAL) << "empty stack in translator";
    }
    return results;
    // generate n-best translations
}

/* obsolete
bool check_consistency_of_initial_hypothesis(const std::vector<AugmentedPattern>& augmented_patterns, TranslationCandidate* cand) {
    BOOST_FOREACH(const AugmentedPattern& pattern, augmented_patterns) {
        if (!cand->check_hypothesis_compatible_with_outer(pattern, pattern.get_outer_frontier())) {
            LOG(ERROR) << "Inconsistent Hypothesis in Pattern " << pattern << "\n";
            return false;
        }
    }
    return true;
}
*/

// Generate n-best translations.
NBestTranslationResult Translator::translate_from_translation_candidates(
        const std::vector<AugmentedPattern>& augmented_patterns,
        size_t i_dtree_size, InitialHypothesesStorage* cand, TableDisplayOutput* tdout, t_log::TranslationLog* translation_log,
        const VS* reference, const WeightVector* specific_weight_vector) const {
    // Check consistency between initial hypothesis and patterns
    // assert(check_consistency_of_initial_hypothesis(augmented_patterns, cand));
    const WeightVector& weights_to_use = (specific_weight_vector == NULL) ? weights_ : (*specific_weight_vector);
    if (Params::Translator::decoder == "lattice") {
        return lattice_search( augmented_patterns,
                                            cand, lm_, weights_to_use, translation_log);
    } else {
        LOG(FATAL) << "Unknown decoder";
        return NBestTranslationResult(); // just to avoid compiler warnings about non-void function reaching end of block
    }
}

std::vector<AugmentedPattern> Translator::build_augmented_patterns(const InitialHypothesesStorage& cand, const Dtree& input_sent) {
    std::vector<PatternType> initial_patterns = cand.get_pattern_list();
    std::vector<AugmentedPattern> augmented_patterns;
    BOOST_FOREACH(const PatternType& pattern, initial_patterns) {
        augmented_patterns.push_back(AugmentedPattern(pattern, input_sent));
    }
    return augmented_patterns;
}

NBestTranslationResult Translator::translate(const Dtree& input_sent) const {
    InitialHypothesesStorage cand;

    TableDisplayOutput tdout;
    if (Params::Global::web) {
        tdout.initialize("id not set", &input_sent);
    }

    // The web logger makes use of the tms in the TRResult objects. Therefore, if --web is enabled, we have to keep a reference to TRResult
    // until the end of the translation. (so that the TRResult object smart pointer do not delete the tms).
    // Otherwise, it is fine to have the TRResults release memory when the call to generate_translation_candidates return.
    std::vector<TRResultPtr> results;
    if (Params::Global::keep_tms_in_memory) {
        results = generate_translation_candidates(input_sent, &cand);
    } else {
        generate_translation_candidates(input_sent, &cand);
    }

    std::vector<AugmentedPattern> augmented_patterns = build_augmented_patterns(cand, input_sent);

    return translate_from_translation_candidates(augmented_patterns, input_sent.size(), &cand,
                &tdout);
}

void Translator::create_initial_hypothesis(const Dtree& input_sent, const std::vector<TRResultPtr> &results, InitialHypothesesStorage* cand,
                                           PartialTranslationFactory& partial_translation_factory,
                                           const TMProcessor* tm_processor_) const {
    BOOST_FOREACH(const TRResultPtr& result, results) {
        BOOST_FOREACH(const PatternType& pattern, result->get_vector_of_pattern()) {
            std::string pattern_str = "";
            for (PatternType::const_iterator it = pattern.begin(); it != pattern.end(); it++) {
                pattern_str = pattern_str + boost::lexical_cast<std::string>(*it) + " ";
            }
            LOG(INFO) << "creating initial hypotheses for PatternType: " << pattern_str;
            // Skip translation
            if (pattern.size() == 1 &&
                input_sent.node_at(*pattern.begin())->pre_children_ref.size() == 0 &&
                input_sent.node_at(*pattern.begin())->post_children_ref.size() == 0 &&
                tm_processor_->is_skip_translation(input_sent.node_at(*pattern.begin()))) {
                LOG(INFO) << *pattern.begin() << " was skipped!";
                Hypothesis *hypothesis = tm_processor_->create_empty_hypothesis(pattern, partial_translation_factory);
                if (hypothesis->check_consistency()) {
                    cand->push(pattern, hypothesis);
                } else {
                    LOG(WARNING) << "Inconsistent initial hypothesis of size 1.";
                }
            } else {
                std::map<std::string, double> t_exp_freq;
                std::map<std::string, std::map<std::string, double> > t_lm_exp_freq;
                // std::cout << cand.get_stack(pattern) << std::endl;
                for (int i = result->get_nb_of_tm_for_pattern(pattern) - 1; i >= 0; i--) {
                    const TM& tm = result->get_tm(pattern, i);
                    LOG(INFO) << "TMID: " << tm.id();
                    Hypothesis *hypothesis = tm_processor_->create_hypothesis_from_TM(input_sent, result, pattern, tm, t_exp_freq, t_lm_exp_freq, partial_translation_factory, lex_trans_table_);
                    if (hypothesis != NULL) {
                        //hypothesis->set_feature_value("arity", hypothesis->get_nb_full_set_of_ref());
                        hypothesis->set_feature_value("nb_additionals", hypothesis->get_nb_additionals());
                        int nb_optionals = hypothesis->get_nb_of_optional_words();
                        hypothesis->set_feature_value("nb_optionals", nb_optionals);
                        int target_size = hypothesis->real_size();
                        if (target_size != 0) {
                            hypothesis->set_feature_value("ratio_optionals",
                                    static_cast<double>(nb_optionals) / (target_size+nb_optionals));
                        }
                        if (hypothesis->check_consistency()) {
                            cand->push(pattern, hypothesis);
                        } else {
                            LOG(WARNING) << "Inconsistent initial hypothesis.";
                        }
                    } else {
                        LOG(INFO) << "discarded.";
                    }
                }
                // cand.get_stack(pattern).calc_trans_prob(pattern, t_exp_freq, t_lm_exp_freq);
                cand->calc_trans_prob_for_pattern(pattern, t_exp_freq, t_lm_exp_freq);
                int pattern_frequency = result->get_real_total_number_of_tm(pattern);
                cand->calc_target_frequency_feature(pattern, pattern_frequency);
            }
        }
    }
}

std::vector<std::pair<const DtreeNode*, WordIndex> > Translator::find_oov_words(
        const std::vector<TRResultPtr> &results, const Dtree &input_sent) {
    // Actually makes an OOV item for every word to ensure that
    // there are no incomplete sentences.
    std::vector<std::pair<const DtreeNode*, WordIndex> > result;
    int n_words = input_sent.size();
    for (int i = 0; i < n_words; i++) {
        std::pair<const DtreeNode*, WordIndex> temp;
        temp.first = input_sent.node_at(i);
        temp.second = i;
        result.push_back(temp);
    }
    return result;
}

void Translator::add_pseudo_TM(std::vector<TRResultPtr> &results, const Dtree &input_sent,
        const OOVHandler& oov_handler_) {
    std::vector<std::pair<const DtreeNode*, WordIndex> > oov_words = find_oov_words(results, input_sent);
    for (size_t i = 0; i < oov_words.size(); i++) {
        std::string oov_word = oov_words[i].first->word.content;
        std::string oov_lem = oov_words[i].first->word.lem;
        std::string method = "unknown";
        std::string translation = oov_handler_.query(oov_word, method);
        WordIndex position = oov_words[i].second;
        add_oov_tm(oov_word, oov_lem, translation, position, results.back(), method);
    }
}

void Translator::add_oov_tm(const std::string &oov_word, const std::string &oov_lem,
                            const std::string &translation, WordIndex position,
                            TRResultPtr &result, std::string &method) {
    PatternType pattern;
    pattern.insert(position);

    TM& tm = result->push_one_empty_tm_and_return_ref(pattern);

    std::string sent_name = "OOV-" + method + "-" + boost::lexical_cast<std::string>(position);

    ParaSentence &para_sent = tm.get_modifiable_para_sentence();
    Dtree &s_dtree = para_sent.get_modifiable_s_dtree();
    Dtree &t_dtree = para_sent.get_modifiable_t_dtree();

    s_dtree.set_to_size_1_for_OOV();
    t_dtree.set_to_size_1_for_OOV();
    para_sent.set_id(sent_name);

    DtreeNode *s_word = s_dtree.get_modifiable_node_at(0);
    s_word->dpnd             = -1;
    s_word->num              = 0;
    s_word->align_index      = -1;
    s_word->word.content     = oov_word;
    s_word->word.lem         = oov_lem;
    s_word->word.content_p   = 1;
    s_dtree.format_dtree();

    DtreeNode *t_word = t_dtree.get_modifiable_node_at(0);
    t_word->dpnd             = -1;
    t_word->num              = 0;
    t_word->align_index      = -1;
    t_word->word.content     = translation;
    t_word->word.lem         = translation;
    t_word->word.content_p   = 1;
    t_dtree.format_dtree();

    // set alignment: 0-0
    para_sent.set_alignment_for_OOV();

    tm.set_alignment_for_OOV(position);
    tm.set_connected_image_for_OOV();
}

/*
  This function create a query for tree retrieve module from dtree
*/
TreeRetrieveQuery Translator::create_tree_retrieve_query(const Dtree &dtree) const {
    std::string label_list;
    std::string dpnd_list;
    BOOST_FOREACH(const DtreeNode &node, dtree.nodes()) {
        if (label_list != "")
            label_list += " ";
        if (tm_processor_->is_punctuation(node.word.lem)) {
            label_list += "<=PUNC=>";
        } else {
            label_list += node.word.lem;
        }
        if (dpnd_list != "")
            dpnd_list += " ";
        dpnd_list += boost::lexical_cast<std::string>(node.dpnd);
    }
    const boost::regex reg("<@([^:]+):.+?>");
    label_list = boost::regex_replace(label_list, reg, "<$1>", boost::format_all);
    return std::make_pair(label_list, dpnd_list);
}
