/*
KyotoEBMT Example-Based Machine Translation System
Copyright (C) 2014 John Richardson, Fabien Cromieres, 
Toshiaki Nakazawa and Sadao Kurohashi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*! @file
  This file contains the definition of Translator class.
*/
#ifndef TRANSLATOR_INCLUDED
#define TRANSLATOR_INCLUDED

#include <boost/lexical_cast.hpp>
#include <boost/scoped_ptr.hpp>
#include <boost/noncopyable.hpp>
#include <glog/logging.h>
#include <map>
#include <string>
#include <utility>
#include <vector>
#include "input_handler.hpp"
#include "tm.hpp"
#include "../TR++/tr_interface.hpp"
#include "hypothesis.hpp"
#include "tm_processor.hpp"
#include "lexical_translation_scorer.hpp"

#include "oov_handler.hpp"

#include "def.hpp"

#include "decoding/combi_search.hpp"
#include "hypotheses_storage.hpp"
#include "decoding/augmented_pattern.hpp"
#include "table_display_output.hpp"
#include "decoding/base_lattice.hpp" // for latt::LanguageModel

typedef std::pair<std::string, std::string> TreeRetrieveQuery;
typedef std::map<std::string, std::vector<TM> > TreeRetrieveResult;

class Translator {
  public:
    //! constructor 
    Translator(std::vector<TRInterfacePtr> tr_list, std::string weight_filename, std::string default_dpnd_filename,
               std::string dic_punc_filename, std::string dic_je_filename, std::string lex_trans_table_filename, std::string s_lang_str = "", std::string t_lang_str = "") 
            : tr_list_(tr_list), oov_handler_(dic_punc_filename, dic_je_filename), lm_(Params::Translator::lm_filename.c_str()), lex_trans_table_(lex_trans_table_filename) {
        set_language(s_lang_str, s_lang_);
        set_language(t_lang_str, t_lang_);
        tm_processor_.reset(new TMProcessor(s_lang_, t_lang_, default_dpnd_filename));
        read_feature_weight(weight_filename);
    }

    /// For when we just want to perform decoding
    explicit Translator(std::string weight_filename)
            : lm_(Params::Translator::lm_filename.c_str()) {
        read_feature_weight(weight_filename); 
    }
   
    //! main function for translation
    NBestTranslationResult translate(const Dtree& input_sent) const; 

    //! get current weights for features
    const WeightVector& get_weights() const {
        return weights_;
    }
    
    // 2-step translation 1/2: generation of the translation candidates
    std::vector<TRResultPtr>  generate_translation_candidates(const Dtree& input_sent, InitialHypothesesStorage* cand) const;
    
    // 2-step translation 2/2: generation of the translations from the translation candidates
    NBestTranslationResult translate_from_translation_candidates(
        const std::vector<AugmentedPattern>& augmented_patterns, size_t i_dtree_size, 
        InitialHypothesesStorage* cand, TableDisplayOutput* tdout, t_log::TranslationLog* translation_log = NULL,
        const VS* reference = NULL, const WeightVector* specific_weight_vector = NULL) const;
    
//    NBestTranslationResult translate_from_translation_candidates_with_oracle(
//            const std::vector<AugmentedPattern>& augmented_patterns,
//            size_t i_dtree_size, TranslationCandidate* cand, TableDisplayOutput* tdout,
//            const VS& reference) const;

    // build the Augmented version of each pattern in cand
    static std::vector<AugmentedPattern> build_augmented_patterns(const InitialHypothesesStorage& cand, const Dtree& input_sent);

    // set language for source or target
    void set_language(const std::string &lang_str, Language &lang) {
        if (lang_str == "ja") {
            lang = JA;
        } else if (lang_str == "en") {
            lang = EN;
        } else if (lang_str == "zh") {
            lang = ZH;
        } else {
            LOG(WARNING) << "Warning: unknown language specified (" << lang_str << ")";
            lang = DEFAULT_LANG;
        }
    }
    
  private:
    std::vector<TRInterfacePtr> tr_list_; //!< vector of tree retrieval objects
    boost::scoped_ptr<TMProcessor> tm_processor_; //!< pointer to the TMProcessor object
    OOVHandler                  oov_handler_; //!< OOVHandler object
    WeightVector                weights_; //!< weights to be used for the decoding
    Language                    s_lang_; //!< source language
    Language                    t_lang_; //!< target language
    latt::LanguageModel lm_;
    LexicalTranslationScorer lex_trans_table_;
    
    void read_feature_weight(std::string filename);
    
    //! generate a query for tree retrieval from dtree
    TreeRetrieveQuery create_tree_retrieve_query(const Dtree &dtree) const;
    
    //! add pseudo TM for OOV words
    static void add_pseudo_TM(std::vector<TRResultPtr> &results, const Dtree &input_sent, const OOVHandler& oov_handler_);
    //! find oov words by checking if each word has a tm available
    static std::vector<std::pair<const DtreeNode*, WordIndex> > find_oov_words(const std::vector<TRResultPtr> &results,
                                                                          const Dtree &input_sent);
    //! create pseudo TM for OOV words and add to TRResult
    static void add_oov_tm(const std::string &oov_word,
                           const std::string &oov_lem,
                           const std::string &translation,
                           WordIndex position, 
                           TRResultPtr& result, 
                           std::string &method);

    //! generate n-best translations for the input sentence using TRResult
    NBestTranslationResult compute_n_best_translations(const Dtree& input_sent, TRResultPtr result) const;

    //! create initial Hypothesis from retrieved TM
    void create_initial_hypothesis(const Dtree& input_sent,
                                   const std::vector<TRResultPtr>& results,
                                   InitialHypothesesStorage* cand,
                                   PartialTranslationFactory& partial_translation_factory,
                                   const TMProcessor* tm_processor_) const;
};

#endif
