# KyotoEBMT Example-Based Machine Translation System
# Copyright (C) 2014 John Richardson, Fabien Cromieres, 
# Toshiaki Nakazawa and Sadao Kurohashi

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

#!/usr/bin/env perl

use strict;
use Getopt::Long;
use XML::Writer;
use XML::Simple;
use utf8;
use Dumpvalue;
binmode STDIN, ':utf8';
binmode STDOUT, ':utf8';

unless($ENV{"XML_SIMPLE_PREFERRED_PARSER"}) {
    $ENV{"XML_SIMPLE_PREFERRED_PARSER"} = "XML::Parser";
}

my (%opt); GetOptions(\%opt, 'help', 'trans_dir=s', 'debug=i', 'print_all');

if ($opt{help}) {
    &usage();
    exit;
}

my $xml = new XML::Simple;

$opt{"trans_dir"} = defined $opt{"trans_dir"} ? $opt{"trans_dir"} : "ja-en";
$opt{"trans_dir"} =~ /^(.+?)-(.+?)$/;
my ($src_lang, $trg_lang) = ($1, $2);
my $src_lang_classname = ucfirst(lc($src_lang));
my $trg_lang_classname = ucfirst(lc($trg_lang));
die sprintf("Invalid trans_dir: \"%s\"\n", $opt{"trans_dir"}) if ($src_lang_classname !~ /^(Ja|En|Zh)$/ || $trg_lang_classname !~ /^(Ja|En|Zh)$/);

my $result = {};
while (1) {
    my $i_dtree;
    my $j_dtree;
    my $DTREE;
    my $flag = 0;
    my $buf = "";

    while (<STDIN>) {
	if (/\s*\<para_sentence id=\"(.+?)\"/) {
	    $buf .= $_;
	    $flag = 1;
	}
	elsif (/\s*\<\/para_sentence/) {
	    $buf .= $_;
	    $flag = 0;
	    last;
	}
	elsif ($flag) {
	    $buf .= $_ 
	}
    }

    next if ($flag);
    last if ($buf eq "");

    $DTREE = $xml->XMLin($buf, ForceArray => 1, keyattr => []);

    next unless (defined $DTREE->{match});

    my @i_words = ();
    foreach my $p (@{$DTREE->{i_data}[0]{phrase}}) {
	if (defined($p->{word})) {
	    push(@i_words, @{$p->{word}});
	} else {
	    # NOTE: It will suggest a bug of previous process to pass here.
	    #       But we make it continue next anyway.
	    # push(@i_words, undef);
	    warn "Warning: i_data's phrase without any words: id='".$DTREE->{id}."'";
	    if ($opt{debug} > 0) {
		print STDERR Encode::encode("utf-8", $buf);
	    }
	}
    }
    next if ($#i_words < 0);
    @i_words = sort {$a->{abs_wnum} <=> $b->{abs_wnum}} @i_words if ($src_lang_classname ne "Ja");
	
    my @j_words = ();
    foreach my $p (@{$DTREE->{j_data}[0]{phrase}}) {
	if (defined($p->{word})) {
	    push(@j_words, @{$p->{word}});
	} else {
	    # NOTE: It will suggest a bug of previous process to pass here.
	    #       But we make it continue next anyway.
	    # push(@j_words, undef);
	    warn "Warning: j_data's phrase without any words: id='".$DTREE->{id}."'";
	    if ($opt{debug} > 0) {
		print STDERR Encode::encode("utf-8", $buf);
	    }
	}
    }
    next if ($#j_words < 0);
    @j_words = sort {$a->{abs_wnum} <=> $b->{abs_wnum}} @j_words if ($trg_lang_classname ne "Ja");

    for (my $i = 0; $i < @{$DTREE->{match}}; $i++) {
	my $m = $DTREE->{match}[$i];
	foreach my $p (split(" ", $m->{i_p})) {
	    $DTREE->{i_data}[0]{phrase}[$p]{match} = $i;
	}
	foreach my $p (split(" ", $m->{j_p})) {
	    $DTREE->{j_data}[0]{phrase}[$p]{match} = $i;
	}
    }

    foreach my $m (@{$DTREE->{match}}) {
	my @i_p = split(/\s/, $m->{i_p});
	my @j_p = split(/\s/, $m->{j_p});

	my $i_root = &get_root($DTREE->{i_data}[0], -1, @i_p);
	my $i_dpnd_ref = $DTREE->{i_data}[0]{phrase}[$DTREE->{i_data}[0]{phrase}[$i_root]{dpnd}];
	my $i_root_pos = $DTREE->{i_data}[0]{phrase}[$i_root]{word}[0]{pos};
	my $j_root = &get_root($DTREE->{j_data}[0], -1, @j_p);
	my $j_dpnd_ref = $DTREE->{j_data}[0]{phrase}[$DTREE->{j_data}[0]{phrase}[$j_root]{dpnd}];
	my $j_root_pos = $DTREE->{j_data}[0]{phrase}[$j_root]{word}[0]{pos};

	next if ($DTREE->{i_data}[0]{phrase}[$i_root]{dpnd} == -1);
	next if ($DTREE->{j_data}[0]{phrase}[$j_root]{dpnd} == -1);

	my $dpnd_type_target;
	if ($DTREE->{j_data}[0]{phrase}[$j_root]{dpnd} > $j_root) {
	    $dpnd_type_target = 'pre';
	} elsif ($DTREE->{j_data}[0]{phrase}[$j_root]{dpnd} < $j_root) {
	    $dpnd_type_target = 'post';
	}

	my $str;
	if ($DTREE->{i_data}[0]{phrase}[$i_root]{dpnd} > $i_root) {
	    $str = sprintf "%s->P:%s\t%s", $DTREE->{i_data}[0]{phrase}[$i_root]{word}[0]{lem}, $i_dpnd_ref->{word}[0]{pos}, $j_root_pos;
	} else {
	    $str = sprintf "P:%s<-%s\t%s", $i_dpnd_ref->{word}[0]{pos}, $DTREE->{i_data}[0]{phrase}[$i_root]{word}[0]{lem}, $j_root_pos;
	}
	$result->{norm}{$str}{$dpnd_type_target}++;

	if ($DTREE->{i_data}[0]{phrase}[$i_root]{dpnd} > $i_root) {
	    $str = sprintf "P:%s->P:%s\t%s", $DTREE->{i_data}[0]{phrase}[$i_root]{word}[0]{pos}, $i_dpnd_ref->{word}[0]{pos}, $j_root_pos;
	} else {
	    $str = sprintf "P:%s<-P:%s\t%s", $i_dpnd_ref->{word}[0]{pos}, $DTREE->{i_data}[0]{phrase}[$i_root]{word}[0]{pos}, $j_root_pos;
	}
	$result->{norm}{$str}{$dpnd_type_target}++;


	my $dpnd_type_target;
	if ($DTREE->{i_data}[0]{phrase}[$i_root]{dpnd} > $i_root) {
	    $dpnd_type_target = 'pre';
	} elsif ($DTREE->{i_data}[0]{phrase}[$i_root]{dpnd} < $i_root) {
	    $dpnd_type_target = 'post';
	}

	my $str;
	if ($DTREE->{j_data}[0]{phrase}[$j_root]{dpnd} > $j_root) {
	    $str = sprintf "%s->P:%s\t%s", $DTREE->{j_data}[0]{phrase}[$j_root]{word}[0]{lem}, $j_dpnd_ref->{word}[0]{pos}, $i_root_pos;
	} else {
	    $str = sprintf "P:%s<-%s\t%s", $j_dpnd_ref->{word}[0]{pos}, $DTREE->{j_data}[0]{phrase}[$j_root]{word}[0]{lem}, $i_root_pos;
	}
	$result->{inv}{$str}{$dpnd_type_target}++;

	if ($DTREE->{j_data}[0]{phrase}[$j_root]{dpnd} > $j_root) {
	    $str = sprintf "P:%s->P:%s\t%s", $DTREE->{j_data}[0]{phrase}[$j_root]{word}[0]{pos}, $j_dpnd_ref->{word}[0]{pos}, $i_root_pos;
	} else {
	    $str = sprintf "P:%s<-P:%s\t%s", $j_dpnd_ref->{word}[0]{pos}, $DTREE->{j_data}[0]{phrase}[$j_root]{word}[0]{pos}, $i_root_pos;
	}
	$result->{inv}{$str}{$dpnd_type_target}++;
    }
}

if ($opt{"print_all"}) {
    open (OUT, ">:utf8", $ARGV[0]."_".$src_lang."_".$trg_lang.".txt") || die;
    foreach my $key (sort keys %{$result->{norm}}) {
	foreach my $dpnd (keys %{$result->{norm}{$key}}) {
	    printf OUT "%s\t%s\t%d\n", $key, $dpnd, $result->{norm}{$key}{$dpnd};
	}
    }
    close OUT;
    open (OUT, ">:utf8", $ARGV[0]."_".$trg_lang."_".$src_lang.".txt") || die;
    foreach my $key (sort keys %{$result->{inv}}) {
	foreach my $dpnd (keys %{$result->{inv}{$key}}) {
	    printf OUT "%s\t%s\t%d\n", $key, $dpnd, $result->{inv}{$key}{$dpnd};
	}
    }
    close OUT;
} else {
    foreach my $key (sort keys %$result) {
	my $max = 0;
	my $max_dpnd = "";
	foreach my $dpnd (keys %{$result->{$key}}) {
	    if ($result->{$key}{$dpnd} > $max) {
		$max_dpnd = $dpnd;
		$max = $result->{$key}{$dpnd};
	    }
	}
	printf "%s\t%s\n", $key, $max_dpnd;
    }
}

sub get_root
{
    my ($dtree, $self, @body) = @_;

    $self = $body[0] if ($self == -1);

    while (grep($dtree->{phrase}[$self]{dpnd} eq $_, @body)) {
	$self = $dtree->{phrase}[$self]{dpnd};
    }
    return $self;
}

sub usage {
    printf("usage: cat bitext_m*.xml | %s\n", $0);
    printf("  This script extracts function word expression patterns from bitext_m files.\n");
    printf("options:\n");
    printf("  --trans_dir                 SRC-TRG  specify the direction of translation\n");
    printf("                                   (default: 'ja-en')\n");
    printf("  --help                      show this message and exit\n");
    printf("  --debug          level      specify the debug level\n");
}
